/*************************************************************************
StrongNet/Contour Engine training code.
Copyright (c) 2020 Sergey Bochkanov (ALGLIB project).

>>> SOURCE LICENSE >>>
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation (www.fsf.org); either version 2 of the 
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

A copy of the GNU General Public License is available at
http://www.fsf.org/licensing/licenses
>>> END OF LICENSE >>>
*************************************************************************/
#include <stdio.h>
#include <experimental/filesystem>

#include "snetcommon.h"
#include "layer1.h"
#include "layer2.h"
#include "layer2c.h"
#include "layer4.h"
#include "snet_trn.h"
#include <atomic>
#include <mutex>

using namespace strongnet;

std::string path_sep = std::string(1, (char)std::experimental::filesystem::path::preferred_separator);

raw_dataset mnist_normalize(const raw_dataset &ds)
{
    raw_dataset out = ds;
    for(raw_dataset::entry &e: out.entries)
    {
        ASSERT(e.img.height==28 && e.img.width==28);
        e.img = e.img.append_border_v(2, 0.0f);
    }
    return out;
}

raw_dataset gtsrb_normalize(const raw_dataset &ds)
{
    raw_dataset out = ds;
    for(raw_dataset::entry &e: out.entries)
    {
        ASSERT(e.img.height==e.img.width);
        if( e.img.height!=32 || e.img.width!=32)
            e.img = e.img.lanczos_resize_to(32,32);
    }
    return out;
}

raw_dataset svhn_normalize(const raw_dataset &ds)
{
    raw_dataset r = ds.normalize_contrast();
    for(raw_dataset::entry &e: r.entries)
    {
        float avg = 0;
        for(int k=0; k<e.img.data.size(); k++)
            avg += e.img.data[k];
        avg /= e.img.data.size();
        if( avg<128.0f )
            e.img = e.img.invert();
    }
    return r;
}

template<class network_t>
class adversarial_fgsm_env
{
public:
    const raw_dataset &ds_study;
    const network_t &xnet;
    const int nsteps;
    const float eps;
    std::vector<int>  increase_targets;
    std::vector<bool> decrease_flags;

    raw_dataset ds_perturbed;

    std::mutex mutex;
    std::vector<float> raw_prob_label, raw_prob_adv;
    std::vector<float> perturbed_prob_label, perturbed_prob_adv;
    std::vector<bool> is_changed;

    adversarial_fgsm_env(const raw_dataset &ds,
                         const network_t &xn,
                         const float ns,
                         const float e):
        ds_study(ds), ds_perturbed(ds), xnet(xn),
        nsteps(ns),eps(e),
        increase_targets(ds.ntrn+ds.ntst,-1), decrease_flags(ds.ntrn+ds.ntst,true),
        raw_prob_label(ds.ntrn+ds.ntst,1.0f),raw_prob_adv(ds.ntrn+ds.ntst,0.0f),
        perturbed_prob_label(ds.ntrn+ds.ntst,1.0f),perturbed_prob_adv(ds.ntrn+ds.ntst,0.0f),
        is_changed(ds.ntrn+ds.ntst,false)
    {}

    float avg_changed() const
    { float v = 0.0f; for(bool vv: is_changed) if( vv ) v += 1; return v/(is_changed.size()+1.0e-20f); }

    float avg_label() const
    { float v = 0.0f; for(float vv: perturbed_prob_label) v += vv; return v/(perturbed_prob_label.size()+1.0e-20f); }

    float avg_adv() const
    { float v = 0.0f; for(float vv: perturbed_prob_adv) v += vv; return v/(perturbed_prob_adv.size()+1.0e-20f); }

    float percent_high_confidence() const
    { float v = 0.0f; for(float vv: perturbed_prob_adv) if( vv>0.5f ) v += 1; return v/(perturbed_prob_adv.size()+1.0e-20f); }
};

class adversarial_fgsm_result
{
public:
    int raw_label, /*raw_predicted_class, */perturbed_class;

    float raw_prob_label, raw_prob_adv/*, raw_predicted_prob*/;
    float perturbed_prob_adv, perturbed_prob_label/*, perturbed_prob_prev_predicted*/;
    raw_dataset::entry raw, perturbed;
};

template<class network_t, class cost_f>
std::vector<float> adversarial_pgd_gradient_gray(const network_t &xnet,
                                                 const raw_dataset &ds1,
                                                 cost_f cost,
                                                 const float diff_step)
{
    const int nclasses = ds1.nclasses;
    ASSERT(ds1.ntrn+ds1.ntst==1);

    //
    // Prepare trial dataset and per-column baselines
    //
    const int dim0 = ds1.entries[0].img.height;
    const int dim1 = ds1.entries[0].img.width;
    const int imdim = dim0*dim1;
    raw_dataset ds_trial = ds1.duplicate_all(imdim+1);
    std::vector<double> diff_steps(imdim);
    for(int i=0; i<imdim; i++)
    {
        int pos0 = i/dim1, pos1=i%dim1, r, g, b;
        ds_trial.entries[i].img.get_rgb(pos0,pos1,&r,&g,&b);
        float v = (r+g+b)/(3.0f*255.0f);
        int s = roundf((v>=0.5f ? -diff_step : +diff_step)*255);
        diff_steps[i] = s;
        r += s;
        g += s;
        b += s;
        ds_trial.entries[i].img.draw_rgb(pos0,pos1,r,g,b);
    }

    //
    // Process trial dataset
    // Analyze perturbation results
    //
    predictions pred_logprob  = xnet.apply_log(ds_trial, storage_format::f32);
    //const int target_class    = pred_logprob.labels[imdim];
    //const int predicted_class = (pred_logprob.predicted_classes())[imdim];
    std::vector<double> cost_func(imdim+1);
    for(int i=0; i<imdim+1; i++)
        cost_func[i] = cost(pred_logprob,i);
    /*{
        const double asymmetry = 0.0;
        double c = 0;
        for(int k=0; k<nclasses; k++)
            if( adv_target<0 || (k==target_class || k==predicted_class || k==adv_target) )
            {
                // important: computations are performed in double precision
                double target = (k==target_class || k==predicted_class) ? 1.0f : 0.0f;
                double prob   = exp(pred_logprob.predicted_prob(i,k));
                double asymm  = (k==target_class || k==predicted_class) ? 0.0f : asymmetry;
                c += (1+asymm)*pow(prob-target,2);
            }
        cost_func[i] = c;
    }*/
    std::vector<float> grad(imdim);
    for(int i=0; i<imdim; i++)
        grad[i] = (cost_func[i]-cost_func[imdim])/diff_steps[i];
    return grad;
}

adversarial_fgsm_result evaluate_result(const raw_dataset &ds_state2, predictions pred2, const int increase_target)
{
    adversarial_fgsm_result result;
    std::vector<int> pred_classes = pred2.predicted_classes();
    const int target_class = pred2.labels[1];
    const int predicted_class = pred_classes[1];
    float prob_adv = 0.0f, prob_raw_adv = 0.0f;
    if( increase_target<0 )
    {
        // untargeted attack
        for(int k=0; k<ds_state2.nclasses; k++)
            if( k!=target_class && pred2.predicted_prob(0,k)>prob_adv )
            {
                prob_adv     = pred2.predicted_prob(0,k);
                prob_raw_adv = pred2.predicted_prob(1,k);
            }
    }
    else
    {
        // targeted attack
        if( increase_target<ds_state2.nclasses )
        {
            prob_adv     = pred2.predicted_prob(0,increase_target);
            prob_raw_adv = pred2.predicted_prob(1,increase_target);
        }
    }
    result.raw                 = ds_state2.entries[1];
    result.raw_label           = target_class;
    result.raw_prob_label      = pred2.predicted_prob(1,target_class);
    result.raw_prob_adv        = prob_raw_adv;
    //result.raw_predicted_class = predicted_class;
    //result.raw_predicted_prob  = pred2.predicted_prob(1,predicted_class);
    result.perturbed           = ds_state2.entries[0];
    result.perturbed_class     = pred_classes[0];
    result.perturbed_prob_adv  = prob_adv;
    result.perturbed_prob_label= pred2.predicted_prob(0,target_class);
    //result.perturbed_prob_prev_predicted = pred2.predicted_prob(0,predicted_class);
    return result;
}

template<class network_t>
adversarial_fgsm_result adversarial_fgsm_produce_gray(const network_t &xnet,
                                                      const raw_dataset &ds1,
                                                      float diff_step,
                                                      int nsteps,
                                                      float eps,
                                                      const int increase_target,
                                                      const bool decrease_flag)
{
    const int max_bt = 8;
    const int stagnation_its = 4;
    const float stagnation_eps = 0.0001;
    const int nclasses = ds1.nclasses;
    ASSERT(ds1.ntrn+ds1.ntst==1);
    const int target_class = ds1.entries[0].class_id;
    raw_dataset ds_state2 = ds1.duplicate_all(2);

    //
    // Quick exit if no adversarial task is specified
    //
    if( increase_target>=nclasses && !decrease_flag )
        return evaluate_result(ds_state2, xnet.apply(ds_state2), increase_target);

    //
    // Perform iterations
    //
    const int dim0 = ds1.entries[0].img.height;
    const int dim1 = ds1.entries[0].img.width;
    const int imdim = dim0*dim1;
    adversarial_fgsm_result result;
    for(int step_idx=0, cnt_small=0; step_idx<nsteps; step_idx++)
    {
        //
        // Cost function
        //
        auto cost_func = [nclasses,target_class,increase_target,decrease_flag](predictions logprob, int i)->double
        {
            double c = 0;

            // cost for deviation from unperturbed prediction
            if( decrease_flag )
                c += pow(1-exp(logprob.predicted_prob(i,target_class)),2);

            // cost for increase in adversarial target
            double max_bad = 0;
            for(int k=0; k<nclasses; k++)
                if( k!=target_class && (increase_target<0 || k==increase_target) )
                    max_bad = fmax(max_bad, pow(exp(logprob.predicted_prob(i,k)),2));
            c+= max_bad;

            // done
            return c;
        };

        //
        // Compute gradient for current state
        //
        std::vector<float> grad = adversarial_pgd_gradient_gray(xnet, ds_state2.select_subsample({0}), cost_func, diff_step);

        //
        // Perform backtracking line search
        //
        bool step_found = false;
        float delta_cost = 0.0f;
        for(int bt_idx=0; bt_idx<max_bt; bt_idx++)
        {
            //
            // Prepare trial position
            //
            const float bt_step = eps*powf(2,-bt_idx);
            raw_dataset ds_cand2 = ds_state2.select_subsample({0}).duplicate_all(2);
            for(int i=0; i<imdim; i++)
            {
                int pos0 = i/dim1, pos1=i%dim1, r0, g0, b0;

                // read and perform step
                ds_state2.entries[0].img.get_rgb(pos0,pos1,&r0,&g0,&b0);
                int dir = 0;
                if( grad[i]>0 )
                    dir = +roundf(bt_step*255.0f);
                if( grad[i]<0 )
                    dir = -roundf(bt_step*255.0f);
                r0 = irange(r0+dir, 0, 255);
                g0 = irange(g0+dir, 0, 255);
                b0 = irange(b0+dir, 0, 255);

                // bound perturbation
                int r1, g1, b1, bnd = roundf(eps*255.0f);
                ds_state2.entries[1].img.get_rgb(pos0,pos1,&r1,&g1,&b1);
                r0 = irange(r0, r1-bnd, r1+bnd);
                g0 = irange(g0, g1-bnd, g1+bnd);
                b0 = irange(b0, b1-bnd, b1+bnd);

                // write to candidate
                ds_cand2.entries[0].img.draw_rgb(pos0,pos1,r0,g0,b0);
            }
            predictions logprob = xnet.apply_log(ds_cand2);
            double cost0 = cost_func(logprob,0), cost1 = cost_func(logprob,1);
            if( cost0>cost1 )
            {
                step_found = true;
                delta_cost = cost0-cost1;
                ds_state2.entries[0] = ds_cand2.entries[0];
                //printf("step found: delta=%.6f / %.6f vs %.6f\n",
                //       cost_func(logprob,0)-cost_func(logprob,1),
                //       cost_func(logprob,0),
                //       cost_func(logprob,1));
                //fflush(stdout);
                break;
            }
        }
        if( delta_cost<stagnation_eps )
            cnt_small++;
        else
            cnt_small = 0;

        //
        // Update result
        //
        /*predictions pred_prob = xnet.apply(ds_state2);
        std::vector<int> pred_classes = pred_prob.predicted_classes();
        const int target_class = pred_prob.labels[1];
        const int predicted_class = pred_classes[1];
        float prob_adv = 0.0f;
        for(int k=0; k<nclasses; k++)
            if( k!=target_class && (k==increase_target || increase_target<0) )
                prob_adv = fmaxf(prob_adv, pred_prob.predicted_prob(0,k));
        result.raw                 = ds_state2.entries[1];
        result.raw_target_class    = target_class;
        result.raw_target_prob     = pred_prob.predicted_prob(1,target_class);
        result.raw_predicted_class = predicted_class;
        result.raw_predicted_prob  = pred_prob.predicted_prob(1,predicted_class);
        result.perturbed           = ds_state2.entries[0];
        result.perturbed_class     = pred_classes[0];
        result.perturbed_prob      = prob_adv;
        result.perturbed_prob_prev_target = pred_prob.predicted_prob(0,target_class);
        result.perturbed_prob_prev_predicted = pred_prob.predicted_prob(0,predicted_class);
        //printf("adv(%02d): %5.3f vs %5.3f\n", (int)step_idx, (double)result.perturbed_prob, (double)result.perturbed_prob_prev_predicted);
        //fflush(stdout);*/
        result = evaluate_result(ds_state2, xnet.apply(ds_state2), increase_target);

        //
        // Break if no changes
        //
        if( !step_found || cnt_small>=stagnation_its )
            break;
    }
    return result;
}

template<class network_t>
void adversarial_fgsm_perturb_gray(void *param, int img_idx)
{
    adversarial_fgsm_env<network_t> &env = *((adversarial_fgsm_env<network_t>*)param);

    //
    // Analyze
    //
    int running_time = dtickcount();
    adversarial_fgsm_result result = adversarial_fgsm_produce_gray(env.xnet,
                                                                   env.ds_study.select_subsample({img_idx}),
                                                                   0.01f, // diff_step
                                                                   env.nsteps,
                                                                   env.eps,
                                                                   env.increase_targets[img_idx],
                                                                   env.decrease_flags[img_idx]);
    running_time = dtickcount()-running_time;

    //
    // Output to adversarial set
    //
    env.ds_perturbed.entries[img_idx] = result.perturbed;

    //
    // Output successful adversarial images
    //
    bool class_changed = result.perturbed_class!=result.raw_label;
    //if( class_changed )
    {
        bitmap bmp_raw = bitmap::combine_ltr({result.raw.img},       1, 0x00FF00);
        bitmap bmp_bad = bitmap::combine_ltr({result.perturbed.img}, 1, 0x0000FF);
        bitmap bmp_out = bitmap::combine_ltr({bmp_raw, bmp_bad},     1, 0x808080);
        char filename[1024];
        sprintf(filename, "%s/%s-class%02d_as_%02d-img%06d.png",
                OUT_DIR,
                class_changed ? "attack" : "badattack",
                (int)result.raw_label,
                (int)result.perturbed_class,
                (int)img_idx);
        bmp_out.to_png(filename);
    }

    //
    // Report results
    //
    env.mutex.lock();
    env.is_changed[img_idx] = class_changed;
    env.raw_prob_label[img_idx]       = result.raw_prob_label;
    env.raw_prob_adv[img_idx]         = result.raw_prob_adv;
    env.perturbed_prob_label[img_idx] = result.perturbed_prob_label;
    env.perturbed_prob_adv[img_idx]   = result.perturbed_prob_adv;
    printf("img=%06d    %s    P(label)=%6.4f->%6.4f    P(worst)=%6.4f->%6.4f    (done in %d ms)\n",
           (int)img_idx,
           class_changed ? "CHG" : "   ",
           (double)result.raw_prob_label,
           (double)result.perturbed_prob_label,
           (double)result.raw_prob_adv,
           (double)result.perturbed_prob_adv,
           running_time);
    fflush(stdout);
    env.mutex.unlock();
}

template<class network_t>
raw_dataset adversarial_fgsm_experiment_gray(const network_t &xnet,
                                      const raw_dataset &ds_full,
                                      const int nsteps,
                                      const float eps,
                                      const bool is_targeted = false)
{
    //
    // Run experiment
    //
    adversarial_fgsm_env<network_t> env(ds_full, xnet, nsteps, eps);
    if( is_targeted )
    {
        for(int i=0; i<ds_full.ntrn+ds_full.ntst; i++)
        {
            std::vector<int> possible_targets;
            for(int k=0; k<ds_full.nclasses; k++)
                if( k!=ds_full.entries[i].class_id )
                    possible_targets.push_back(k);
            env.increase_targets[i] = possible_targets.size()>0 ? possible_targets[i%possible_targets.size()] : -1;
        }
    }
    int running_time = dtickcount();
    parallel_for(adversarial_fgsm_perturb_gray<network_t>, &env, ds_full.ntrn+ds_full.ntst);
    running_time = dtickcount()-running_time;
    printf("Done in %d ms:\n* %4.1f%% susceptible to attack\n* %4.1f%% high-confidence\n* %.2f average adv prob\n* %.2f average original prob\n",
           running_time,
           double(100.0*env.avg_changed()),
           (double)(100*env.percent_high_confidence()),
           (double)env.avg_adv(),
           (double)env.avg_label());
    return env.ds_perturbed;
}

void demo_mnist_small()
{
    //
    // Load dataset
    //
    std::string filename = "mnist-all-all-raw.bin";
    if( !file_exists(filename) )
    {
        printf("File '%s' does not exist!\n", filename.c_str());
        fflush(stdout);
        abort();
    }
    int t_load = dtickcount();
    printf("Loading and augmenting dataset... ");
    fflush(stdout);
    raw_dataset ds_all(filename);
    printf("done in %d ms\n", dtickcount()-t_load);
    fflush(stdout);

    //
    // Train unsupervised feature detector
    //
    net_v1layer v1layer;
    v1layer.append_layer(get_v1c_lxg(ds_all, std::make_tuple(10,10)));

    //
    // Train sparse layers
    //
    const auto nc_w24 = std::make_tuple(8,50,50);
    const auto co_geo = std::make_tuple(0,split_t::entire,0,0);
    train_subcolumn_settings settings;
    settings.omega_rat   = 0.100f;
    settings.post_lambda1_reg = 0.00001f;
    settings.pre_lambda2_reg  = 0.00001f;
    settings.post_lambda2_reg = 0.00001f;
    settings.top_and_type = and_type_t::AND_MUL;
    settings.v2v4_batch_size = 10;
    settings.logit_batch_momentum = std::make_tuple(128,0.98f);
    settings.logit_learning_rates = std::make_tuple( 100,std::vector<float>{0.3f});
    x1_network net;
    predictions pred;
    const int seed = 0;
    std::tie(net,pred) = x1_train_baseline(
            ds_all,
            {},
            v1layer,
            settings,
            0.00f,
            nc_w24,
            co_geo,
            seed);
    serialize_to(&net,sPATH_OUT+path_sep+"mnist-small.x1");

    //
    // Re-evaluate test set error
    //
    predictions pred_tst = net.apply(ds_all.select_tst());
    printf("mean test loss   = %.4f\n", pred_tst.tst_rms());
    printf("test set cls err = %.4f\n", pred_tst.tst_cls());
    fflush(stdout);
}

void demo_mnist()
{
    //
    // Load dataset
    //
    std::string filename = "mnist-all-all-raw.bin";
    if( !file_exists(filename) )
    {
        printf("File '%s' does not exist!\n", filename.c_str());
        fflush(stdout);
        abort();
    }
    int t_load = dtickcount();
    printf("Loading and augmenting dataset... ");
    fflush(stdout);
    raw_dataset ds_all(filename);
    printf("done in %d ms\n", dtickcount()-t_load);
    fflush(stdout);

    //
    // Train unsupervised feature detector
    //
    net_v1layer v1layer;
    v1layer.append_layer(get_v1c_lxg(ds_all, std::make_tuple(25,25)));

    //
    // Train sparse layers
    //
    const auto nc_w24 = std::make_tuple(8,100,100);
    const auto co_geo = std::make_tuple(0,split_t::entire,0,0);
    train_subcolumn_settings settings;
    settings.omega_rat   = 0.100f;
    settings.post_lambda1_reg = 0.00001f;
    settings.pre_lambda2_reg  = 0.00001f;
    settings.post_lambda2_reg = 0.00001f;
    settings.top_and_type = and_type_t::AND_MUL;
    settings.v2v4_batch_size = 10;
    settings.logit_batch_momentum = std::make_tuple(128,0.98f);
    settings.logit_learning_rates = std::make_tuple( 100,std::vector<float>{0.3f});
    x1_network net;
    predictions pred;
    const int seed = 0;
    std::tie(net,pred) = x1_train_baseline(
            ds_all,
            {},
            v1layer,
            settings,
            0.00f,
            nc_w24,
            co_geo,
            seed);
    serialize_to(&net,sPATH_OUT+path_sep+"mnist.x1");

    //
    // Re-evaluate test set error
    //
    predictions pred_tst = net.apply(ds_all.select_tst());
    printf("mean test loss   = %.4f\n", pred_tst.tst_rms());
    printf("test set cls err = %.4f\n", pred_tst.tst_cls());
    fflush(stdout);
}

void demo_gtsrb()
{
    //
    // Load and augment dataset
    //
    std::string filename = "gtsrb-all-all-raw.bin";
    if( !file_exists(filename) )
    {
        printf("File '%s' does not exist!\n", filename.c_str());
        fflush(stdout);
        abort();
    }
    int t_load = dtickcount();
    printf("Loading and augmenting dataset... ");
    fflush(stdout);
    raw_dataset ds_aug(filename);
    ds_aug = gtsrb_normalize(ds_aug);
    ds_aug = ds_aug.augment_affine(5, std::make_tuple(0.10f,0.10f));
    printf("done in %d ms\n", dtickcount()-t_load);
    fflush(stdout);

    //
    // Train unsupervised feature extraction layer
    //
    net_v1layer v1layer;
    v1layer.append_layer(get_v1_w32p6_lxg(ds_aug, std::make_tuple(50,50), true));
    v1layer.append_layer(get_v1_w32p4_clx(ds_aug, std::make_tuple(10,10), true));

    //
    // Train network
    //
    const int seed = 0;
    const auto nc_w24 = std::make_tuple(8,200,200);
    const auto co_geo = std::make_tuple(0,split_t::entire,0,0);
    train_subcolumn_settings settings;
    settings.omega_rat   = 0.100f;
    settings.post_lambda1_reg = 0.00001f;
    settings.pre_lambda2_reg  = 0.00001f;
    settings.post_lambda2_reg = 0.00001f;
    settings.top_and_type = and_type_t::AND_MUL;
    settings.v2v4_batch_size = 10;
    settings.logit_batch_momentum = std::make_tuple(128,0.98f);
    settings.logit_learning_rates = std::make_tuple( 100,std::vector<float>{0.3f});
    x1_network net;
    predictions pred;
    std::tie(net,pred) = x1_train_baseline(
                ds_aug,
                {},
                v1layer,
                settings,
                0.00f, // clip level
                nc_w24,
                co_geo,
                seed);
    serialize_to(&net,sPATH_OUT+path_sep+"gtsrb.x1");
}

void demo_svhn()
{
    //
    // Dataset
    //
    std::string filename = "svhn-xtr-all-raw.bin";
    if( !file_exists(filename) )
    {
        printf("File '%s' does not exist!\n", filename.c_str());
        fflush(stdout);
        abort();
    }
    int t_load = dtickcount();
    printf("Loading and augmenting dataset... ");
    fflush(stdout);
    raw_dataset ds_aug(filename);
    ds_aug = svhn_normalize(ds_aug);
    printf("done in %d ms\n", dtickcount()-t_load);
    fflush(stdout);

    //
    // Train unsupervised feature detection layer
    //
    net_v1layer v1layer;
    v1layer.append_layer(get_v1_w32p6_lxg(ds_aug, std::make_tuple(50,50), true));

    //
    // Train network
    //
    const int seed = 1;
    const auto nc_w24 = std::make_tuple(8,800,800);
    const auto co_geo = std::make_tuple(0,split_t::entire,0,0);
    train_subcolumn_settings settings;
    settings.omega_rat   = 0.100f;
    settings.post_lambda1_reg = 0.00001f;
    settings.pre_lambda2_reg  = 0.00001f;
    settings.post_lambda2_reg = 0.00001f;
    settings.top_and_type = and_type_t::AND_MUL;
    settings.v2v4_batch_size = 10;
    settings.logit_batch_momentum = std::make_tuple(128,0.98f);
    settings.logit_learning_rates = std::make_tuple( 100,std::vector<float>{0.3f});
    x1_network net;
    predictions pred;
    std::tie(net,pred) = x1_train_baseline(
                ds_aug,
                {},
                v1layer,
                settings,
                0.00f, // clip level
                nc_w24,
                co_geo,
                seed);
    serialize_to(&net,sPATH_OUT+path_sep+"svhn.x1");
}

void demo_adversarial_svhn()
{
    //
    // Dataset
    //
    std::string filename = "svhn-xtr-all-raw.bin";
    if( !file_exists(filename) )
    {
        printf("File '%s' does not exist!\n", filename.c_str());
        fflush(stdout);
        abort();
    }
    int t_load = dtickcount();
    printf("Loading and augmenting dataset... ");
    fflush(stdout);
    raw_dataset ds_aug(filename);
    ds_aug = svhn_normalize(ds_aug);
    printf("done in %d ms\n", dtickcount()-t_load);
    fflush(stdout);

    //
    // Load network
    //
    x1_network net;
    unserialize_from(std::string("adv-svhn.x1"), &net);

    //
    // Adversarial experiment
    //
    raw_dataset ds_adv;
    ds_adv = ds_aug.select_tst();
    ds_adv = ds_adv.subsample(0, 1000);
    printf("Starting experiment...\n");
    fflush(stdout);
    adversarial_fgsm_experiment_gray(net, ds_adv, 40,  0.01f, false);
}

int main(int argc, char *argv[])
{
    if( argc!=2 )
    {
        printf("Run program as './demo EXPERIMENTNAME' where EXPERIMENTNAME can be :\n");
        printf("* mnist_small - small-scale MNIST experiment (can be done at laptop, 4 GB of RAM are required)\n");
        printf("* mnist       - MNIST experiment (can be done at laptop)\n");
        printf("* gtsrb       - GTSRB experiment (16 cores, 64 GB of RAM are required)\n");
        printf("* svhn        - SVHN  experiment (16 cores, 64 GB of RAM are required)\n");
        printf("* advsvhn     - adversarial SVHN  experiment (16 cores, 8 GB of RAM are required)\n");
        return 0;
    }
    printf("Path separator detected: '%c'\n", (char)std::experimental::filesystem::path::preferred_separator);
    printf("Output is saved to '.%c%s'\n", (char)std::experimental::filesystem::path::preferred_separator, sPATH_OUT.c_str());

    //
    // Prepare
    //
    int64_t t0 = strongnet::dtickcount();
    strongnet::parallelism_activate(); // Necessary for parallelism support, should be called only once from the parent thread
    printf("CORES COUNT: %d\n", (int)strongnet::get_cores_count());

    //
    // Action requested
    //
    std::string ename = argv[1];
    if( ename=="mnist_small" )
    {
        printf("=== Small-scale MNIST experiment, needs about a minute with 24 cores ===\n");
        demo_mnist_small();
    }
    else if( ename=="mnist" )
    {
        printf("=== MNIST experiment, needs about 5 minutes with 24 cores ===\n");
        demo_mnist();
    }
    else if( ename=="gtsrb" )
    {
        printf("=== GTSRB experiment, needs about an hour with 24 cores, 64 GB of RAM===\n");
        demo_gtsrb();
    }
    else if( ename=="svhn" )
    {
        printf("=== SVHN experiment, needs about 24 hours at 24 cores, 64 GB of RAM ===\n");
        demo_svhn();
    }
    else if( ename=="advsvhn" )
    {
        printf("=== adversarial SVHN experiment, needs a few hours ===\n");
        demo_adversarial_svhn();
    }
    else
    {
        printf("ERROR: unexpected experiment name!\n");
        return 1;
    }

    //
    // Done, report time
    //
    printf("TOTAL TIME: %8d ms\n",
           int(strongnet::dtickcount()-t0));
    fflush(stdout);
    return 0;
}

