/*************************************************************************
StrongNet/Contour Engine training framework.
Copyright (c) 2020 Sergey Bochkanov (ALGLIB project).

>>> SOURCE LICENSE >>>
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation (www.fsf.org); either version 2 of the 
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

A copy of the GNU General Public License is available at
http://www.fsf.org/licensing/licenses
>>> END OF LICENSE >>>
*************************************************************************/
#ifndef LAYER1_H
#define LAYER1_H

#include "snetcommon.h"

namespace strongnet
{

class filter_family
{
public:
    int size, cnt;
    std::vector<float> coeffs;

    filter_family():size(0){}

    filter_family normalize_sum2nz() const;
};
void serialize(FILE *f, const filter_family *ptr);
void unserialize(FILE *f, filter_family *ptr);

class filter_set
{
public:
    class rule
    {
    public:
        int chsize0, chsize1;
        filter_family filters;
    };
    std::vector<rule> rules;

    const filter_family& find_family(int chsize) const;
};
void serialize(FILE *f, const filter_set *ptr);
void serialize(FILE *f, const filter_set::rule *ptr);
void unserialize(FILE *f, filter_set *ptr);
void unserialize(FILE *f, filter_set::rule *ptr);

enum v1_flags_t { not_compete = 0x1, compete = 0x2 };

class v1_settings
{
public:
    float kmeans_wht_stab;
    float kmeans_rad_param;
    int   kmeans_trn_size;
    float kmeans_hardness;
    int   kmeans_clr_super;

    float lxg_alpha_param;
    float lxg_beta_param;
    float clx_alpha_param;

    v1_settings():
        kmeans_wht_stab(0.1f),
        kmeans_rad_param(0.4f),
        kmeans_trn_size(600000),
        kmeans_hardness(0.0f),
        kmeans_clr_super(1),
        lxg_alpha_param(1.40f),
        lxg_beta_param(2.80f),
        clx_alpha_param(2.80f)
    {}
};

class net_v1layer
{
public:
    enum class realm_t { none = 1,
                         lightness = 2,
                         color6_edges = 3,
                         color8_uni = 4,
                         gray = 5,
                         kmeans = 6,
                         kmeans_lgt  = 7,
                         kmeans_clr  = 8,
                         kmeans_gry  = 9,
                         kmeans_clg  = 10,
                         kmeans_clu  = 11,
                         kmeans_lxg  = 12,
                         kmeans_xxg  = 13,
                         kmeans_clx1 = 14};

    class v1_realm
    {
    public:
        realm_t     realm_type;
        std::string realm_name;
        filter_set  filters;
        v1_flags_t  flags;
        int         scale_idx;
        int         projected_size;
        int         downsample_min;
        int         downsample_max;

        // k-means and other settings
        float       kmeans_wht_stab;
        float       kmeans_hardness;
        float       kmeans_rad_param;
        float       lxg_alpha_param;
        float       lxg_beta_param;
        float       clx_alpha_param;
        float       filter_gain; // 0 for autonormalization
        float       kmeans_flt7;
        int         kmeans_int0;
        int         kmeans_int1;
        int         kmeans_int2;
        int         kmeans_int3;
        int         kmeans_int4;
        int         kmeans_int5;
        int         kmeans_int6;
        int         kmeans_int7;
        std::vector<float> kmeans_wht_matrix;

        v1_realm();

        bool quick_comparison(const v1_realm &rhs) const;

        void apply(const raw_dataset &ds_in, activities &act_out) const;
        void apply_append(const raw_dataset &ds_in, activities &act_out) const;

        void debug_output(std::string path) const;
    };

    net_v1layer(){}
    explicit net_v1layer(std::string filename);
    std::vector<v1_realm>           realms;

    bool operator==(const net_v1layer &rhs) const;

    // return input image width/height, presently dimensions are same
    int  get_img_width() const;
    int  get_img_height() const;

    // return lowest and highest "effective" downsampling factors;
    // effective means that downsampling input image counts for downsampling as well as downsampling layer outputs
    int  get_ds_lo(select_t select = select_t::all, int select_height = 0, int select_width = 0) const;
    int  get_ds_hi(select_t select = select_t::all, int select_height = 0, int select_width = 0) const;

    // returns maximum height/width of V1 realms selected using user-specified criteria
    int  get_max_height(select_t select = select_t::all, int select_height = 0, int select_width = 0) const;
    int  get_max_width(select_t select = select_t::all, int select_height = 0, int select_width = 0) const;

    void append_realm(const v1_realm &rr);
    void append_layer(const net_v1layer &layer);

    void apply(const raw_dataset &ds_in, activities &act_out, bool report = false, storage_format fmt = storage_format::i8) const;
    void apply_append(const raw_dataset &ds_raw, activities &act_out, bool report = false, storage_format fmt = storage_format::i8) const;

    void conv_apply(const bitmap &img, conv_activities &act_out, bool report = false, storage_format fmt = storage_format::i8) const;

    void run_dummy_optimization();
    void debug_output(std::string path) const;

    friend void serialize(FILE *f, const net_v1layer::v1_realm *ptr);
    friend void serialize(FILE *f, const net_v1layer *ptr);
    friend void unserialize(FILE *f, net_v1layer::v1_realm *ptr);
    friend void unserialize(FILE *f, net_v1layer *ptr);
};

void serialize(FILE *f, const net_v1layer::v1_realm *ptr);
void serialize(FILE *f, const net_v1layer *ptr);
void unserialize(FILE *f, net_v1layer::v1_realm *ptr);
void unserialize(FILE *f, net_v1layer *ptr);

// generate uniform color area
filter_family generate_uni_color_area(int size);

// generate "edges and angles" set of big filters
filter_family generate_ea_colorless(int size);

// generate "edges" set of micro-filters
filter_family generate_e_colorless_micro(int size);

// generate "edges and angles" set of micro-filters
filter_family generate_ea_colorless_micro(int size);

// generate "edges and angles" set of big filters
filter_family generate_ea_colored(int size);

// generate "edges and blobs" set of filters
filter_family generate_edges_blobs_colored(int size);

// generate femto filters
filter_family generate_femto_colored();

// train k-means filters
filter_family train_kmeans_filters(const raw_dataset &ds_raw,
                                   const int scale_idx,
                                   net_v1layer::realm_t kmeans_type,
                                   int patch_size,
                                   int sample_size,
                                   int nfilters,
                                   const v1_settings s,
                                   const char *rep_suffix = "med");
filter_family train_kmeans_filters_lxg(const raw_dataset &ds_raw,
                                       const int scale_idx,
                                       int patch_size,
                                       int sample_size,
                                       int nfilters,
                                       const v1_settings s,
                                       const char *rep_suffix = "med");


// applies "colorless" set of filters to the dataset
void apply_colorless_v1(const raw_dataset &ds_in,
                        const filter_set &filters,
                        int projected_size,
                        std::string realm_name,
                        activities &ds_out);

// colored uniform
void apply_colored_uniform_v1(const raw_dataset &ds_in,
                              const filter_set &filters,
                              int projected_size,
                              v1_flags_t flags,
                              std::string realm_name,
                              int downsample_min,
                              int downsample_max,
                              std::vector<activities> &vec_out);

//
void apply_colored_edges_v1(const raw_dataset &ds_in,
                        const filter_set &filters,
                        int projected_size,
                        v1_flags_t flags,
                        std::string realm_name,
                        bool need_black_and_white,
                        activities &ds_out);

//
void apply_lightness_v1(const raw_dataset &ds_in,
                         const filter_set &filters,
                         int projected_size,
                         v1_flags_t flags,
                         std::string realm_name,
                         int downsample_min,
                         int downsample_max,
                         std::vector<activities> &vec_out);

//
void apply_gray_v1(const raw_dataset &ds_in,
                   const filter_set &filters,
                   int projected_size,
                   v1_flags_t flags,
                   std::string realm_name,
                   int downsample_min,
                   int downsample_max,
                   std::vector<activities> &vec_out);

// k-means
void apply_kmeans_v1(const raw_dataset &ds_in,
                     const filter_set &filters,
                     int projected_size,
                     net_v1layer::realm_t kmeans_type,
                     const float wht_stab,
                     const float hardness,
                     const float rad_param,
                     std::string realm_name,
                     int downsample_min,
                     int downsample_max,
                     std::vector<activities> &vec_out);
void apply_kmeans_v1_xg(const raw_dataset &ds_in,
                         const filter_set &filters,
                         int projected_size,
                         const net_v1layer::realm_t rtype,
                         const float lxg_alpha,
                         const float lxg_beta,
                         const float filter_gain,
                         std::string realm_name,
                         int downsample_min,
                         int downsample_max,
                         storage_format fmt,
                         std::vector<activities> &vec_out);
void apply_kmeans_v1_xc(
        const raw_dataset &ds_in,
        const filter_set &filters,
        int expected_size,
        const net_v1layer::realm_t rtype,
        const float clx_alpha,
        std::string realm_name,
        int downsample_min,
        int downsample_max,
        storage_format fmt,
        std::vector<activities> &vec_out);

void downsample2_max(const activities &ds_in, const char *pfx, activities &ds_out);
void downsample2_avg(const activities &ds_in, const char *pfx, activities &ds_out);
void convolve(const std::vector<float> &input, int input_h, int input_w,
              const std::vector<float> &filter, int filter_h, int filter_w,
              std::vector<float> &out);

net_v1layer get_v1basic4x8_lxg(
        int img_width,
        const int scale_idx,
        const std::string layer_name,
        const int downsample_lo, const int downsample_hi,
        const v1_settings s,
        const float filter_gain);
net_v1layer get_v1basic4x12_lxg(
        int img_width,
        const int scale_idx,
        const std::string layer_name,
        const int downsample_lo, const int downsample_hi,
        const v1_settings s,
        const float filter_gain);
net_v1layer get_v1km_xg(
        const raw_dataset &ds,
        const net_v1layer::realm_t rtype,
        const int scale_idx,
        const std::string layer_name,
        const std::string rep_suffix,
        const int patch_size, const int nkernels,
        const int downsample_lo, const int downsample_hi,
        const v1_settings s);

net_v1layer get_v1c_lxg(const raw_dataset &ds,
                        std::tuple<int,int> kernels,
                        int kextra = 0,
                        bool just_top_level = false);
net_v1layer get_v1c_clr(const raw_dataset &ds,
                        std::tuple<int,int> kernels,
                        bool just_top_level = false);

net_v1layer get_v1_w32p4_lxg(const raw_dataset &ds,
                             std::tuple<int,int> kernels,
                             bool just_8x8_level = false);

net_v1layer get_v1_wnnp4_bas8lxg(int w, float filter_gain, int scale_hi = 1, bool just_top_level = true);
net_v1layer get_v1_wnnp4_bas12lxg(int w, float filter_gain, int scale_hi = 1, bool just_top_level = true);
net_v1layer get_v1_wnnp4_bas14lxg(int w, float filter_gain, int scale_hi = 1, bool just_top_level = true);

net_v1layer get_v1_w32p4_clx(const raw_dataset &ds,
                             std::tuple<int,int> kernels,
                             bool just_8x8_level = false);

net_v1layer get_v1_w32p6_lxg(const raw_dataset &ds,
                             std::tuple<int,int> kernels,
                             bool just_8x8_level = false);
net_v1layer get_v1_w32p6_clx(const raw_dataset &ds,
                             std::tuple<int,int> kernels,
                             bool just_8x8_level = false);

void v1_analyze(const net_v1layer &layer);
}

#endif
