/*************************************************************************
StrongNet/Contour Engine training framework.
Copyright (c) 2020 Sergey Bochkanov (ALGLIB project).

>>> SOURCE LICENSE >>>
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation (www.fsf.org); either version 2 of the 
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

A copy of the GNU General Public License is available at
http://www.fsf.org/licensing/licenses
>>> END OF LICENSE >>>
*************************************************************************/
#include "layer2.h"
#include <map>
#include <set>
#include <atomic>
#include <algorithm>
#include <random>

namespace strongnet
{

std::atomic<int32_t> net_midlayer::perf_mlapply(0);

float net_midlayer::get_flop_count() const
{
    float flops = 0;
    for(const realm<detector> &rr: realms)
        for(const detector &d: rr.elements)
        {
            int c = d.input0.size()+d.input1.size()+d.input2.size()+d.input3.size()+d.input4.size();
            flops  += (c+d.inputs_cnt())*imax(d.transform.size(),1);
        }
    return flops;
}

void net_midlayer::append1(const detector &d)
{
    ASSERT(realms.size()==1 && realms[0].grid_width==1 && realms[0].grid_height==1);
    ASSERT(realms[0].ch_offs==0);
    ASSERT(realms[0].grid_depth==ndetectors);
    realms[0].append_element(d);
    ndetectors++;
}

void net_midlayer::append_realm(const realm<detector> &r)
{
    // add and reallocate
    realms.push_back(r.rebase(ndetectors));
    ndetectors += r.get_nchannels();
}

void net_midlayer::merge_in(const net_midlayer &merged)
{
    // rebase realms
    int64_t cur_offs = ndetectors;
    for(const realm<detector> &r: merged.realms)
    {
        realms.push_back(r.rebase(cur_offs));
        cur_offs += r.elements.size();
    }
    ndetectors += merged.ndetectors;
}

void net_midlayer::apply(const activities &act_in, activities &act_out) const
{
    timer t(perf_mlapply);
    activity8_vector v_act8;
    activityf_vector v_actf;
    act_out = activities(act_in.ntrn, act_in.ntst, act_in.get_format());
    for(const realm<detector> &rr: realms)
    {
        // integrity check
        ASSERT(rr.grid_height==1 && rr.grid_width==1);

        // append realm
        int base_idx = act_out.append_realm(realm<channel>::create_grid(rr.name,rr.grid_height,rr.grid_width,rr.grid_depth));

        // overwrite activities
        for(int k=0; k<rr.elements.size(); k++)
        {
            const detector &d = rr.elements[k];
            if( act_in.get_format()==storage_format::i8 )
            {
                d.generate_activity_slow(act_in, v_act8);
                act_out.set_activity(base_idx+k, v_act8);
            }
            else if( act_in.get_format()==storage_format::f32 )
            {
                d.generate_activity_slow(act_in, v_actf);
                act_out.set_activity(base_idx+k, v_actf);
            }
            else
                ASSERT(0);
        }
    }
}

void net_midlayer::conv_apply(const conv_activities &act_in, conv_activities &act_out) const
{
    timer t(perf_mlapply);
    conv_activities act_tmp_out = conv_activities(act_in.get_img_height(), act_in.get_img_width(), act_in.get_format());

    // determine base (i.e. maximum) window width
    int base_win0 = 1, base_win1 = 1;
    for(const realm<detector> &rr: realms)
        if( rr.grid_depth>0 )
        {
            ASSERT(rr.grid_height==1 && rr.grid_width==1 && rr.grid_depth==rr.elements.size());
            ASSERT(rr.grid_depth>0 && rr.elements.size()>0 && rr.elements[0].input0.size()>0);
            const conv_activities::conv_realm &rr_alpha = act_in.realm_by_name(rr.elements[0].input0[0].realm_name);
            base_win0 = imax(base_win0,rr_alpha.window_dim0);
            base_win1 = imax(base_win1,rr_alpha.window_dim1);
        }

    // processs detectors realm by realm
    for(const realm<detector> &rr: realms)
    {
        // fast skip if needed
        if( rr.grid_depth==0 )
        {
            //act_tmp_out.append_realm(conv_activities::conv_realm::create_grid(rr.name,1,1,0,0));
            continue;
        }

        // integrity check - a lot of work
        ASSERT(rr.grid_height==1 && rr.grid_width==1 && rr.grid_depth==rr.elements.size());
        ASSERT(rr.grid_depth>0 && rr.elements.size()>0 && rr.elements[0].input0.size()>0);
        const conv_activities::conv_realm &rr_alpha = act_in.realm_by_name(rr.elements[0].input0[0].realm_name);
        int in_dim0 = rr_alpha.grid_dim0;
        int in_dim1 = rr_alpha.grid_dim1;
        int in_win0 = rr_alpha.window_dim0;
        int in_win1 = rr_alpha.window_dim1;
        int scale0 = get_scale_idx_safe(in_win0,base_win0);
        int scale1 = get_scale_idx_safe(in_win0,base_win0);
        ASSERT(scale0==scale1);
        int scale_idx = scale0;
        for(const detector &dd: rr.elements)
        {
            for(const detector::conn &c: dd.input0)
            {
                const conv_activities::conv_realm &rr_beta = act_in.realm_by_name(c.realm_name);
                ASSERT(rr_alpha.grid_dim0==rr_beta.grid_dim0);
                ASSERT(rr_alpha.grid_dim1==rr_beta.grid_dim1);
                ASSERT(rr_alpha.window_dim0==rr_beta.window_dim0);
                ASSERT(rr_alpha.window_dim1==rr_beta.window_dim1);
            }
            for(const detector::conn &c: dd.input1)
            {
                const conv_activities::conv_realm &rr_beta = act_in.realm_by_name(c.realm_name);
                ASSERT(rr_alpha.grid_dim0==rr_beta.grid_dim0);
                ASSERT(rr_alpha.grid_dim1==rr_beta.grid_dim1);
                ASSERT(rr_alpha.window_dim0==rr_beta.window_dim0);
                ASSERT(rr_alpha.window_dim1==rr_beta.window_dim1);
            }
            for(const detector::conn &c: dd.input2)
            {
                const conv_activities::conv_realm &rr_beta = act_in.realm_by_name(c.realm_name);
                ASSERT(rr_alpha.grid_dim0==rr_beta.grid_dim0);
                ASSERT(rr_alpha.grid_dim1==rr_beta.grid_dim1);
                ASSERT(rr_alpha.window_dim0==rr_beta.window_dim0);
                ASSERT(rr_alpha.window_dim1==rr_beta.window_dim1);
            }
            for(const detector::conn &c: dd.input3)
            {
                const conv_activities::conv_realm &rr_beta = act_in.realm_by_name(c.realm_name);
                ASSERT(rr_alpha.grid_dim0==rr_beta.grid_dim0);
                ASSERT(rr_alpha.grid_dim1==rr_beta.grid_dim1);
                ASSERT(rr_alpha.window_dim0==rr_beta.window_dim0);
                ASSERT(rr_alpha.window_dim1==rr_beta.window_dim1);
            }
            for(const detector::conn &c: dd.input4)
            {
                const conv_activities::conv_realm &rr_beta = act_in.realm_by_name(c.realm_name);
                ASSERT(rr_alpha.grid_dim0==rr_beta.grid_dim0);
                ASSERT(rr_alpha.grid_dim1==rr_beta.grid_dim1);
                ASSERT(rr_alpha.window_dim0==rr_beta.window_dim0);
                ASSERT(rr_alpha.window_dim1==rr_beta.window_dim1);
            }
        }

        // determine window dimensions
        ASSERT(rr.grid_depth>0 && rr.elements.size()>0);
        const int window_dim0 = rr.elements[0].get_realm_dim0();
        const int window_dim1 = rr.elements[0].get_realm_dim1();
        for(const detector &dd: rr.elements)
            ASSERT(dd.get_realm_dim0()==window_dim0 && dd.get_realm_dim1()==window_dim1);
        ASSERT(in_win0==window_dim0 && in_win1==window_dim1);

        // create detector realm and generate its activities
        int out_height = in_dim0-window_dim0+1;
        int out_width  = in_dim1-window_dim1+1;
        conv_activities act_rlm = conv_activities(act_in.get_img_height(), act_in.get_img_width(), act_in.get_format());
        act_rlm.append_realm(conv_activities::conv_realm::create_grid(rr.name,out_height,out_width,scale_idx,0));
        for(const detector &dd: rr.elements)
            act_rlm.append_planes_from1(dd.conv_generate_activity_slow(act_in,rr.name));

        // append to output
        act_tmp_out.merge_in(act_rlm);
    }

    // synchronize scales
    act_out = act_tmp_out.synchronize_scales();
}

void serialize(FILE *f, const net_midlayer *ptr)
{
    write_magic8(f, 22);
    serialize(f, &ptr->ndetectors);
    serialize(f, &ptr->realms);
    write_magic8(f, 97);
}

void unserialize(FILE *f, net_midlayer *ptr)
{
    check_magic8(f, 22);
    unserialize(f, &ptr->ndetectors);
    unserialize(f, &ptr->realms);
    check_magic8(f, 97);
}

v2_channel_scanner::v2_channel_scanner(
           const activities &ai,
           const predictions &bl,
           std::atomic<int32_t> &rs,
           double lr,
           v2_settings s):act_in(ai),settings(s),lambda_reg(lr),running_rng_seed(rs),local_eval(bl,s.ort_cnt,lr)
{
    ASSERT(settings.gain_levels>=1 && settings.gain_levels<=4);
    contest_size = act_in.realms.size()*settings.gain_levels;
    winners = std::vector<winner>(contest_size,winner());
    rng_seeds_to_use = std::vector<int>(contest_size, 0);
}

void* v2_channel_scanner_run_parallel(void *ptr)
{
    v2_channel_scanner      &scanner= *((v2_channel_scanner*)ptr);
    const activities        &act_in = scanner.act_in;
    const linear_evaluator  &lin_eval = scanner.local_eval;
    v2_settings             settings = scanner.settings;
    seed_evaluator          seed_eval(act_in,lin_eval);

    //
    // Select input realm to work with
    //
    float seed1cov2;
    activity8_vector seed_act;
    seed_detector seed1;
    for(int entry_idx=scanner.entries_examined.fetch_add(1);
            entry_idx<scanner.contest_size;
            entry_idx=scanner.entries_examined.fetch_add(1))
    {

        //
        // Internal RNGs initialized by iteration-specific seeds
        //
        std::uniform_real_distribution<> uniform(0.0, 1.0);
        std::mt19937  mt_rnd;
        std::seed_seq mt_seed{(unsigned int)scanner.rng_seeds_to_use[entry_idx], (unsigned int)settings.rng_seed+1};
        mt_rnd.seed(mt_seed);

        //
        // Determine entry properties: realm, gain, location within realm
        //
        const gain_t gain      = gain_t(1<<(entry_idx%settings.gain_levels));
        const int    realm_idx = entry_idx/settings.gain_levels;

        //
        // Scan channels
        //
        const realm<channel> &rr = act_in.realms[realm_idx];
        seed_eval.start_evaluation(and_type_t::AND_MIN);
        for(int d=0; d<rr.grid_depth; d++)
            for(int offs0=0; offs0<rr.grid_height; offs0++)
                for(int offs1=0; offs1<rr.grid_width; offs1++)
                {
                    if( uniform(mt_rnd)>settings.eval_prob )
                        continue;
                    seed_eval.evaluate_seed(seed_detector(rr.get_cidx(d,offs0,offs1), gain));
                }
        seed_eval.get_winner(&seed1cov2, &seed1, &seed_act);
        if( seed1cov2<=0 )
            continue;
        scanner.winners[entry_idx].is_set = true;
        scanner.winners[entry_idx].seed = seed1;
        scanner.winners[entry_idx].winner_act = seed_act;
        scanner.winners[entry_idx].rating = seed1cov2;//(1.0e-20f+i8nrm2(&seed_act[0],act_in.ntrn));
    }
    return NULL;
}

v2_channel_scanner::winner* v2_channel_scanner::run_competition()
{
    //
    // Prepare seeds, iteration counter, winners
    //
    entries_examined = 0;
    for(int i=0; i<contest_size; i++)
        rng_seeds_to_use[i] = running_rng_seed.fetch_add(1);
    for(winner &w: winners)
        w.invalidate();

    //
    // Parallel execution
    //
    parallel_invoke(v2_channel_scanner_run_parallel, this);

    //
    // Retrieve
    //
    return retrieve_next();
}

v2_channel_scanner::winner* v2_channel_scanner::retrieve_next()
{
    //
    // Reset linear evaluator if its memory is full
    //
    if( local_eval.is_full() )
    {
        predictions pred = local_eval.outputs();
        local_eval = linear_evaluator(pred, settings.ort_cnt, lambda_reg);
    }

    //
    // Handle results
    //
    int best_idx = -1;
    for(int i=0; i<contest_size; i++)
        if( winners[i].is_set && (best_idx<0 || winners[i].rating>winners[best_idx].rating) )
            best_idx = i;
    if( best_idx<0 )
        return NULL;
    local_eval.append_activity(winners[best_idx].winner_act);

    //
    // Update the rest after we retrieve winner
    //
    winners[best_idx].is_set = false;
    for(int i=0; i<contest_size; i++)
        if( winners[i].is_set )
            winners[i].rating = local_eval.estimate_trn_covariation(&winners[i].winner_act[0]);

    //
    // Return
    //
    return &winners[best_idx];
}

std::atomic<int32_t> v2_contest::pcnt_refresh(0);
const int v2_contest::line_width = 5;

v2_contest::v2_contest(const activities &ai,
               const linear_evaluator &le,
               const std::vector<seed_detector> &sd,
               std::atomic<int32_t> &rs,
               v2_settings s):act_in(ai),lin_eval(le),seeds(sd),running_rng_seed(rs),settings(s)
{
    winners          = std::vector<winner>(line_width*seeds.size(), winner());
    rng_seeds_to_use = std::vector<int>(seeds.size(), 0);
}

template<class generator_t, class center_t, class func_t>
void scan_neighborhood(const realm<channel> &rr,
                       center_t center,
                       int scan_rad,
                       float eval_prob,
                       const std::set<int> &skip_channels,
                       generator_t &generator,
                       func_t action)
{
    std::uniform_real_distribution<> uniform(0.0, 1.0);
    for(int offs0=-scan_rad; offs0<=+scan_rad; offs0++)
        for(int offs1=-scan_rad; offs1<=+scan_rad; offs1++)
        {
            int input_pos0 = center.pos0+offs0, input_pos1 = center.pos1+offs1;
            if( input_pos0<0 || input_pos0>=rr.grid_height )
                continue;
            if( input_pos1<0 || input_pos1>=rr.grid_width )
                continue;
            for(int depth_idx=0; depth_idx<rr.grid_depth; depth_idx++)
            {
                int new_channel = rr.get_cidx(depth_idx, input_pos0, input_pos1);
                if( skip_channels.count(new_channel)>0 )
                    continue;
                if( uniform(generator)>eval_prob )
                    continue;
                action(new_channel);
            }
        }
}

void* v2_contest_run_parallel(void *ptr)
{
    v2_contest              &contest = *((v2_contest*)ptr);
    const activities        &act_in = contest.act_in;
    const linear_evaluator  &lin_eval = contest.lin_eval;
    v2_settings             settings = contest.settings;
    seed_evaluator          seed_eval(act_in,lin_eval);
    ASSERT(settings.dmax>=1);
    ASSERT(settings.ncenters>=1);

    //
    // Select input seed to work with
    //
    activity8_vector d1_seed_act, d2_seed_act, d3_seed_act, d4_seed_act, d5_seed_act;
    for(int seed_idx=contest.lines_examined.fetch_add(1);
            seed_idx<contest.seeds.size();
            seed_idx=contest.lines_examined.fetch_add(1))
    {
        v2_contest::winner &out_d1 = contest.winners[seed_idx*v2_contest::line_width+0];
        v2_contest::winner &out_d2 = contest.winners[seed_idx*v2_contest::line_width+1];
        v2_contest::winner &out_d3 = contest.winners[seed_idx*v2_contest::line_width+2];
        v2_contest::winner &out_d4 = contest.winners[seed_idx*v2_contest::line_width+3];
        v2_contest::winner &out_d5 = contest.winners[seed_idx*v2_contest::line_width+4];
        const realm<channel> &rr = act_in.realm_by_channel(contest.seeds[seed_idx].channel_idx0);
        struct center_t
        {
            int pos0, pos1;
        };
        std::vector<center_t> centers;
        auto center_from_channel = [&](int cidx){ int c0, c1; rr.get_element_location(cidx, &c0, &c1); return center_t{c0,c1}; };

        //
        // Internal RNGs initialized by iteration-specific seeds
        //
        std::uniform_real_distribution<> uniform(0.0, 1.0);
        std::mt19937  mt_rnd;
        std::seed_seq mt_seed{(unsigned int)contest.rng_seeds_to_use[seed_idx], (unsigned int)settings.rng_seed};
        mt_rnd.seed(mt_seed);

        //
        // Skip evaluation with some probability
        //
        //if( uniform(mt_rnd)>settings.eval_prob )
        //    continue;

        //
        // Evaluate D1 detector
        //
        if( settings.dmax<1 )
            continue;
        seed_detector d1_seed_det;
        float d1_seed_cov2;
        seed_eval.start_evaluation(and_type_t::AND_MIN);
        seed_eval.evaluate_seed(seed_detector(contest.seeds[seed_idx]));
        seed_eval.get_winner(&d1_seed_cov2, &d1_seed_det, &d1_seed_act);
        if( d1_seed_cov2<=0 )
            continue;
        out_d1.is_set = true;
        out_d1.seed = d1_seed_det;
        out_d1.winner_act = d1_seed_act;
        out_d1.realm_name = rr.name;
        out_d1.anrm2  = i8nrm2(&d1_seed_act[0], act_in.ntrn);
        out_d1.rating = compute_rating(settings.con_rating, settings.omega_rat, d1_seed_act, d1_seed_cov2, out_d1.anrm2, lin_eval);
        centers.push_back(center_from_channel(d1_seed_det.channel_idx0));

        //
        // Evaluate D2 detector built upon D1 one
        //
        if( settings.dmax<2 )
            continue;
        seed_detector d2_seed_det;
        float d2_seed_cov2;
        seed_eval.start_evaluation(and_type_t::AND_MIN);
        for(int center_idx=0; center_idx<imin(centers.size(),settings.ncenters); center_idx++)
            scan_neighborhood(rr,
                          centers[center_idx],
                          settings.max_dist,
                          settings.d2d3_eval_prob,
                          {d1_seed_det.channel_idx0},
                          mt_rnd,
                          [&](int new_channel){seed_eval.evaluate_seed(seed_detector(d1_seed_det.channel_idx0,
                                                                                     new_channel,
                                                                                     d1_seed_det.ggain));});
        seed_eval.get_winner(&d2_seed_cov2, &d2_seed_det, &d2_seed_act);
        if( d2_seed_cov2<=0 )
            continue;
        out_d2.is_set = true;
        out_d2.seed = d2_seed_det;
        out_d2.winner_act = d2_seed_act;
        out_d2.realm_name = rr.name;
        out_d2.anrm2  = i8nrm2(&d2_seed_act[0], act_in.ntrn);
        out_d2.rating = compute_rating(settings.con_rating, settings.omega_rat, d2_seed_act, d2_seed_cov2, out_d2.anrm2, lin_eval);
        centers.push_back(center_from_channel(d2_seed_det.channel_idx1));

        //
        // Evaluate D3 detector built upon D2 one
        //
        if( settings.dmax<3 )
            continue;
        seed_detector d3_seed_det;
        float d3_seed_cov2;
        seed_eval.start_evaluation(and_type_t::AND_MIN);
        for(int center_idx=0; center_idx<imin(centers.size(),settings.ncenters); center_idx++)
            scan_neighborhood(rr,
                          centers[center_idx],
                          settings.max_dist,
                          settings.d2d3_eval_prob,
                          {d2_seed_det.channel_idx0,d2_seed_det.channel_idx1},
                          mt_rnd,
                          [&](int new_channel){seed_eval.evaluate_seed(seed_detector(d2_seed_det.channel_idx0,
                                                                                     d2_seed_det.channel_idx1,
                                                                                     new_channel,
                                                                                     d2_seed_det.ggain));});
        seed_eval.get_winner(&d3_seed_cov2, &d3_seed_det, &d3_seed_act);
        if( d3_seed_cov2<=0 )
            continue;
        out_d3.is_set = true;
        out_d3.seed = d3_seed_det;
        out_d3.winner_act = d3_seed_act;
        out_d3.realm_name = rr.name;
        out_d3.anrm2  = i8nrm2(&d3_seed_act[0], act_in.ntrn);
        out_d3.rating = compute_rating(settings.con_rating, settings.omega_rat, d3_seed_act, d3_seed_cov2, out_d3.anrm2, lin_eval);
        centers.push_back(center_from_channel(d3_seed_det.channel_idx2));

        //
        // Evaluate D4 detector built upon D3 one
        //
        if( settings.dmax<4 )
            continue;
        seed_detector d4_seed_det;
        float d4_seed_cov2;
        seed_eval.start_evaluation(and_type_t::AND_MIN);
        for(int center_idx=0; center_idx<imin(centers.size(),settings.ncenters); center_idx++)
            scan_neighborhood(rr,
                          centers[center_idx],
                          settings.max_dist,
                          settings.d2d3_eval_prob,
                          {d3_seed_det.channel_idx0,d3_seed_det.channel_idx1,d3_seed_det.channel_idx2},
                          mt_rnd,
                          [&](int new_channel){seed_eval.evaluate_seed(seed_detector(d3_seed_det.channel_idx0,
                                                                                     d3_seed_det.channel_idx1,
                                                                                     d3_seed_det.channel_idx2,
                                                                                     new_channel,
                                                                                     d3_seed_det.ggain));});
        seed_eval.get_winner(&d4_seed_cov2, &d4_seed_det, &d4_seed_act);
        if( d4_seed_cov2<=0 )
            continue;
        out_d4.is_set = true;
        out_d4.seed = d4_seed_det;
        out_d4.winner_act = d4_seed_act;
        out_d4.realm_name = rr.name;
        out_d4.anrm2  = i8nrm2(&d4_seed_act[0], act_in.ntrn);
        out_d4.rating = compute_rating(settings.con_rating, settings.omega_rat, d4_seed_act, d4_seed_cov2, out_d4.anrm2, lin_eval);
        centers.push_back(center_from_channel(d4_seed_det.channel_idx3));

        //
        // Evaluate D5 detector built upon D4 one
        //
        if( settings.dmax<5 )
            continue;
        seed_detector d5_seed_det;
        float d5_seed_cov2;
        seed_eval.start_evaluation(and_type_t::AND_MIN);
        for(int center_idx=0; center_idx<imin(centers.size(),settings.ncenters); center_idx++)
            scan_neighborhood(rr,
                          centers[center_idx],
                          settings.max_dist,
                          settings.d2d3_eval_prob,
                          {d4_seed_det.channel_idx0,d4_seed_det.channel_idx1,d4_seed_det.channel_idx2,d4_seed_det.channel_idx3},
                          mt_rnd,
                          [&](int new_channel){seed_eval.evaluate_seed(seed_detector(d4_seed_det.channel_idx0,
                                                                                     d4_seed_det.channel_idx1,
                                                                                     d4_seed_det.channel_idx2,
                                                                                     d4_seed_det.channel_idx3,
                                                                                     new_channel,
                                                                                     d4_seed_det.ggain));});
        seed_eval.get_winner(&d5_seed_cov2, &d5_seed_det, &d5_seed_act);
        if( d5_seed_cov2<=0 )
            continue;
        out_d5.is_set = true;
        out_d5.seed = d5_seed_det;
        out_d5.winner_act = d5_seed_act;
        out_d5.realm_name = rr.name;
        out_d5.anrm2  = i8nrm2(&d5_seed_act[0], act_in.ntrn);
        out_d5.rating = compute_rating(settings.con_rating, settings.omega_rat, d5_seed_act, d5_seed_cov2, out_d5.anrm2, lin_eval);
        centers.push_back(center_from_channel(d5_seed_det.channel_idx4));
    }
    return NULL;
}

void v2_contest::run_competition()
{
    lines_examined = 0;
    for(int i=0; i<seeds.size(); i++)
        rng_seeds_to_use[i] = running_rng_seed.fetch_add(1);
    for(winner &w: winners)
        w.invalidate();
    parallel_invoke(v2_contest_run_parallel, this);
}

v2_contest::winner* v2_contest::retrieve_winner()
{
    if( settings.con_rating==rating_t::cov2 ||
        settings.con_rating==rating_t::deltar2 ||
        settings.con_rating==rating_t::fastdr2 )
    {
        //
        // Select winner and return
        //
        int winner_idx = -1;
        for(int i=0; i<winners.size(); i++)
            if( winners[i].is_set && (winner_idx<0 || winners[i].rating>winners[winner_idx].rating) )
                winner_idx = i;
        if( winner_idx<0 )
            return NULL;
        winners[winner_idx].is_set = false;
        return &winners[winner_idx];
    }
    else if( settings.con_rating==rating_t::fastdr2_x5 ||
             settings.con_rating==rating_t::fastdr2_x25 )
    {
        const int cand_max = settings.con_rating==rating_t::fastdr2_x5 ? 5 : 25;

        //
        // Select several winners
        //
        std::set<int> candidates;
        for(int ncand=0; ncand<cand_max; ncand++)
        {
            int winner_idx = -1;
            for(int i=0; i<winners.size(); i++)
                if( winners[i].is_set && candidates.count(i)==0 && (winner_idx<0 || winners[i].rating>winners[winner_idx].rating) )
                    winner_idx = i;
            if( winner_idx>=0 )
                candidates.insert(winner_idx);
        }

        //
        // Choose one with best delta_r2
        //
        float winner_dr2 = 0.0f;
        int   winner_idx = -1;
        for(int idx: candidates)
        {
            float dr2 = lin_eval.deltar2(&winners[idx].winner_act[0]);
            if( winner_idx<0 || dr2>winner_dr2 )
            {
                winner_idx = idx;
                winner_dr2 = dr2;
            }
        }
        if( winner_idx<0 )
            return NULL;
        winners[winner_idx].is_set = false;
        return &winners[winner_idx];
    }
    else
        ASSERT(0);
    return NULL;
}

void parallel_v2_refresh_adapter(void *p, int idx)
{
    v2_contest &con = *((v2_contest*)p);
    con.refresh_ith_winner(idx);
}

void v2_contest::refresh_winners()
{
    timer tmr(pcnt_refresh);
    parallel_for(parallel_v2_refresh_adapter, this, winners.size());
}

void v2_contest::refresh_ith_winner(int winner_idx)
{
    winner &w = winners[winner_idx];
    if( !w.is_set )
        return;
    if( settings.con_rating==rating_t::cov2 )
    {
        w.rating = lin_eval.estimate_trn_covariation(&w.winner_act[0]);
    }
    else if( settings.con_rating==rating_t::deltar2 )
    {
        ASSERT(w.has_daa);
        lin_eval.analyze(&w.winner_act[0], &w.daa);
        w.rating = w.daa.deltar2;
        w.has_daa = true;
    }
    else if( settings.con_rating==rating_t::fastdr2 ||
             settings.con_rating==rating_t::fastdr2_x5 ||
             settings.con_rating==rating_t::fastdr2_x25 )
    {
        //lin_eval.estimate_trn_covariation(&w.winner_act[0])/(w.anrm2+1.0e-20f);
        w.rating = compute_rating(settings.con_rating, settings.omega_rat, w.winner_act, lin_eval);
    }
    else
        ASSERT(0);
}

void v2_contest::append_and_refresh_winners(linear_evaluator *eval, const activity8_vector &v_act)
{
    ASSERT(eval==&lin_eval);
    if( settings.con_rating==rating_t::cov2 )
    {
        eval->append_activity(v_act);
        refresh_winners();
    }
    else if( settings.con_rating==rating_t::deltar2 )
    {
        linear_evaluator::daa append;
        eval->analyze(&v_act[0], &append);
        std::vector<linear_evaluator::daa*> refresh;
        for(int i=0; i<winners.size(); i++)
            if( winners[i].is_set )
            {
                ASSERT(winners[i].has_daa);
                refresh.push_back(&winners[i].daa);
            }
        eval->append_and_refresh(append, refresh);
        for(int i=0; i<winners.size(); i++)
            if( winners[i].is_set )
                winners[i].rating = winners[i].daa.deltar2;
    }
    else if( settings.con_rating==rating_t::fastdr2 ||
             settings.con_rating==rating_t::fastdr2_x5 ||
             settings.con_rating==rating_t::fastdr2_x25 )
    {
        eval->append_activity(v_act);
        refresh_winners();
    }
    else
        ASSERT(0);
}

static void generalize(const activities &act_in,
                const linear_evaluator &lin_eval,
                generalizer_evaluator  &gen_eval,
                const v2_settings settings,
                const seed_detector &seed,
                detector *d_out, activity8_vector *act_out)
{
    //
    // Prepare detector and its activity
    //
    detector            &det = *d_out;
    activity8_vector    &act = *act_out;
    det = detector(seed, and_type_t::AND_MIN, or_type_t::OR_MAX, act_in);
    det.generate_activity_slow(act_in, act);
    int w_inputs_cnt = seed.inputs_cnt();

    //
    // Generalize
    //
    float rating = compute_rating(settings.gen_rating, settings.omega_rat, act, lin_eval);
    for(;;)
    {
        float            gen_cand_cov2;
        detector         gen_cand_det;
        activity8_vector gen_cand_act;
        bool something_changed = false;
        for(int input_idx=0; input_idx<w_inputs_cnt; input_idx++)
        {
            int input_sizes[5]       = {int(det.input0.size()),
                                        int(det.input1.size()),
                                        int(det.input2.size()),
                                        int(det.input3.size()),
                                        int(det.input4.size())};
            int channel_indexes[5]   = {seed.channel_idx0,
                                        seed.channel_idx1,
                                        seed.channel_idx2,
                                        seed.channel_idx3,
                                        seed.channel_idx4};
            int input_size           = input_sizes[input_idx];
            int channel_idx          = channel_indexes[input_idx];
            const realm<channel> &rr = act_in.realm_by_channel(channel_idx);
            ASSERT(input_size>=1);
            if( input_size<SNET_MAX_V2_GEN_COUNT )
            {
                int pos0, pos1;
                rr.get_element_location(channel_idx, &pos0, &pos1);
                std::vector<int> channels_to_try = rr.get_elements_at_location(pos0,pos1);
                gen_eval.start_evaluation(det, input_idx, and_type_t::AND_MIN, or_type_t::OR_MAX);
                for(int c: channels_to_try)
                    gen_eval.evaluate_generalization(c);
                gen_eval.get_winner(&gen_cand_cov2, &gen_cand_det, &gen_cand_act);
                float gen_cand_rating = compute_rating(settings.gen_rating,
                                                       settings.omega_rat,
                                                       gen_cand_act,
                                                       gen_cand_cov2,
                                                       vnrm2(&gen_cand_act[0],act_in.ntrn),
                                                       lin_eval);
                ASSERT(settings.gen_rating_growth>0);
                if( gen_cand_rating>rating*settings.gen_rating_growth )
                {
                    something_changed = true;
                    rating = gen_cand_rating;
                    det    = gen_cand_det;
                    act    = gen_cand_act;
                }
            }
        }
        if( !something_changed )
            break;
    }
}

static void transform(const activities &act_in,
               const linear_evaluator &lin_eval,
               const v2_settings settings,
               const seed_detector &seed,
               const detector &cand,
               detector *d_out, activity8_vector *act_out)
{

    //
    // Prepare detector and its activity
    //
    detector            &det = *d_out;
    activity8_vector    &act = *act_out;
    det = cand;
    det.generate_activity_slow(act_in, act);
    int w_inputs_cnt = seed.inputs_cnt();

    //
    // Generalize
    //
    float trf_rating = compute_rating(settings.gen_rating, settings.omega_rat, act, lin_eval);
    for(int search_rad=1; search_rad<=SNET_MAX_SEARCH_RAD; search_rad++)
    {
        // scan area, find possible transformations (ones which do not move us beyond boundaries)
        const realm<channel> &rr = act_in.realm_by_channel(seed.channel_idx0);
        detector trf_cand = det;
        int pos00, pos01, pos10, pos11, pos20, pos21, pos30, pos31, pos40, pos41;
        rr.get_element_location(seed.channel_idx0, &pos00, &pos01);
        if( w_inputs_cnt<2 )
        {
            pos10 = pos00;
            pos11 = pos01;
        }
        else
            rr.get_element_location(seed.channel_idx1, &pos10, &pos11);
        if( w_inputs_cnt<3 )
        {
            pos20 = pos00;
            pos21 = pos01;
        }
        else
            rr.get_element_location(seed.channel_idx2, &pos20, &pos21);
        if( w_inputs_cnt<4 )
        {
            pos30 = pos00;
            pos31 = pos01;
        }
        else
            rr.get_element_location(seed.channel_idx3, &pos30, &pos31);
        if( w_inputs_cnt<5 )
        {
            pos40 = pos00;
            pos41 = pos01;
        }
        else
            rr.get_element_location(seed.channel_idx4, &pos40, &pos41);
        for(int shift0=-search_rad; shift0<=+search_rad; shift0++)
            for(int shift1=-search_rad; shift1<=+search_rad; shift1++)
            {
                if( abs(shift0)!=search_rad && abs(shift1)!=search_rad )
                    continue;
                if( pos00+shift0<0 || pos00+shift0>=rr.grid_height )
                    continue;
                if( pos01+shift1<0 || pos01+shift1>=rr.grid_width )
                    continue;
                if( pos10+shift0<0 || pos10+shift0>=rr.grid_height )
                    continue;
                if( pos11+shift1<0 || pos11+shift1>=rr.grid_width )
                    continue;
                if( pos20+shift0<0 || pos20+shift0>=rr.grid_height )
                    continue;
                if( pos21+shift1<0 || pos21+shift1>=rr.grid_width )
                    continue;
                if( pos30+shift0<0 || pos30+shift0>=rr.grid_height )
                    continue;
                if( pos31+shift1<0 || pos31+shift1>=rr.grid_width )
                    continue;
                if( pos40+shift0<0 || pos40+shift0>=rr.grid_height )
                    continue;
                if( pos41+shift1<0 || pos41+shift1>=rr.grid_width )
                    continue;
                trf_cand.transform.push_back(detector::trf(shift0,shift1));
            }

        // produce activity vector, estimate covariation
        activity8_vector trf_cand_act;
        trf_cand.generate_activity_slow(act_in, trf_cand_act);
        float trf_cand_rating = compute_rating(settings.trf_rating, settings.omega_rat, trf_cand_act, lin_eval);
        ASSERT(settings.gen_rating_growth>0);
        if( search_rad>settings.min_transform_rad && trf_cand_rating<=trf_rating*settings.gen_rating_growth )
            break;
        trf_rating = trf_cand_rating;
        det  = trf_cand;
        act  = trf_cand_act;
    }
}

static std::string generate_report(const activities &act_in,
                            const seed_detector &seed,
                            const detector &det)
{
    if( seed.channel_idx1<0 )
    {
        const realm<channel> &rr = act_in.realm_by_channel(seed.channel_idx0);
        int grid_area    = rr.grid_stride0*rr.grid_stride1;
        int inner_idx    = seed.channel_idx0-rr.ch_offs;
        int plane_idx    = inner_idx/grid_area;
        int pos0         = (inner_idx%grid_area)/rr.grid_stride1;
        int pos1         = (inner_idx%grid_area)%rr.grid_stride1;
        char buf[1024];
        sprintf(buf, "%14s*%d D1[g=%d;t=%d] = c%d(%02d,%02d)",
                rr.name.c_str(),
                int(seed.ggain),
                int(det.input0.size()), int(det.transform.size()),
                plane_idx, pos0, pos1);
        return buf;
    }
    else if( seed.channel_idx2<0 )
    {
        const realm<channel> &rr = act_in.realm_by_channel(seed.channel_idx0);
        char buf[1024];
        int plane0, pos00, pos01;
        int plane1, pos10, pos11;
        rr.get_element_info(seed.channel_idx0, &plane0, &pos00, &pos01);
        rr.get_element_info(seed.channel_idx1, &plane1, &pos10, &pos11);
        sprintf(buf, "%14s*%d D2[g=%d,%d;t=%d] = c%02d(%02d,%02d)&c%02d(%02d,%02d)",
                rr.name.c_str(),
                int(seed.ggain),
                int(det.input0.size()), int(det.input1.size()), int(det.transform.size()),
                plane0, pos00, pos01,
                plane1, pos10, pos11);
        return buf;
    }
    else if( seed.channel_idx3<0 )
    {
        const realm<channel> &rr = act_in.realm_by_channel(seed.channel_idx0);
        char buf[1024];
        int plane0, pos00, pos01;
        int plane1, pos10, pos11;
        int plane2, pos20, pos21;
        rr.get_element_info(seed.channel_idx0, &plane0, &pos00, &pos01);
        rr.get_element_info(seed.channel_idx1, &plane1, &pos10, &pos11);
        rr.get_element_info(seed.channel_idx2, &plane2, &pos20, &pos21);
        sprintf(buf, "%14s*%d D3[g=%d,%d,%d;t=%d] = c%02d(%02d,%02d)&c%02d(%02d,%02d)&c%02d(%02d,%02d)",
                rr.name.c_str(),
                int(seed.ggain),
                int(det.input0.size()), int(det.input1.size()), int(det.input2.size()), int(det.transform.size()),
                plane0, pos00, pos01,
                plane1, pos10, pos11,
                plane2, pos20, pos21);
        return buf;
    }
    else if( seed.channel_idx4<0 )
    {
        const realm<channel> &rr = act_in.realm_by_channel(seed.channel_idx0);
        char buf[1024];
        int plane0, pos00, pos01;
        int plane1, pos10, pos11;
        int plane2, pos20, pos21;
        int plane3, pos30, pos31;
        rr.get_element_info(seed.channel_idx0, &plane0, &pos00, &pos01);
        rr.get_element_info(seed.channel_idx1, &plane1, &pos10, &pos11);
        rr.get_element_info(seed.channel_idx2, &plane2, &pos20, &pos21);
        rr.get_element_info(seed.channel_idx3, &plane3, &pos30, &pos31);
        sprintf(buf, "%14s*%d D4[g=%d,%d,%d,%d;t=%d] = c%02d(%02d,%02d)&c%02d(%02d,%02d)&c%02d(%02d,%02d)&c%02d(%02d,%02d)",
                rr.name.c_str(),
                int(seed.ggain),
                int(det.input0.size()), int(det.input1.size()), int(det.input2.size()), int(det.input3.size()),
                int(det.transform.size()),
                plane0, pos00, pos01,
                plane1, pos10, pos11,
                plane2, pos20, pos21,
                plane3, pos30, pos31);
        return buf;
    }
    else
    {
        const realm<channel> &rr = act_in.realm_by_channel(seed.channel_idx0);
        char buf[1024];
        int plane0, pos00, pos01;
        int plane1, pos10, pos11;
        int plane2, pos20, pos21;
        int plane3, pos30, pos31;
        int plane4, pos40, pos41;
        rr.get_element_info(seed.channel_idx0, &plane0, &pos00, &pos01);
        rr.get_element_info(seed.channel_idx1, &plane1, &pos10, &pos11);
        rr.get_element_info(seed.channel_idx2, &plane2, &pos20, &pos21);
        rr.get_element_info(seed.channel_idx3, &plane3, &pos30, &pos31);
        rr.get_element_info(seed.channel_idx4, &plane4, &pos40, &pos41);
        sprintf(buf, "%14s*%d D5[g=%d,%d,%d,%d,%d;t=%d] = c%02d(%02d,%02d)&c%02d(%02d,%02d)&c%02d(%02d,%02d)&c%02d(%02d,%02d)&c%02d(%02d,%02d)",
                rr.name.c_str(),
                int(seed.ggain),
                int(det.input0.size()), int(det.input1.size()), int(det.input2.size()), int(det.input3.size()), int(det.input4.size()),
                int(det.transform.size()),
                plane0, pos00, pos01,
                plane1, pos10, pos11,
                plane2, pos20, pos21,
                plane3, pos30, pos31,
                plane4, pos40, pos41);
        return buf;
    }
}

/*void v2_layer_xx(const predictions &input_predictions,
                const activities &ds_in,
                v2_settings settings,
                int max_channels,
                double lambda_reg,
                net_midlayer &ml_out,
                activities   &act_out,
                predictions  &output_predictions)
{
    ASSERT(max_channels>0);
    linear_evaluator lin_eval(input_predictions, max_channels, lambda_reg);
    v2_layer_xx(lin_eval, ds_in, settings, max_channels, lambda_reg, ml_out, act_out, output_predictions);
}*/

void v2_layer_xx(linear_evaluator &lin_eval,
              const activities    &act_in,
              v2_settings         settings,
              int                 max_channels,
              net_midlayer        &ml_out,
              activities          &act_out,
              predictions         &output_predictions)
{
    int t0 = dtickcount();

    ASSERT(lin_eval.outputs().ntrn==act_in.ntrn);
    ASSERT(lin_eval.outputs().ntst==act_in.ntst);
    ASSERT(act_in.realms.size()>=1);

    //
    const int64_t ntrn = act_in.ntrn;
    const int64_t ntst = act_in.ntst;
    const int64_t nclasses = lin_eval.outputs().nclasses;

    //
    // Internal RNG initialized by running seed
    //
    std::atomic<int32_t> running_seed(settings.rng_seed);
    std::uniform_real_distribution<> uniform(0.0, 1.0);
    std::mt19937  mt_rnd;
    std::seed_seq mt_seed{(unsigned int)999, (unsigned int)settings.rng_seed};
    mt_rnd.seed(mt_seed);

    //
    // Determine seed channels
    //
    std::vector<seed_detector> seeds_detectors;
    {
        ASSERT(settings.seeds_count>0);
        const int scanner_batch_size = 2;
        int t0_seed = dtickcount();
        v2_channel_scanner scanner(act_in, lin_eval.outputs(), running_seed, sqrtf(1.0e-5f*ntrn), settings);
        for(;seeds_detectors.size()<settings.seeds_count;)
        {
            for(int bidx=0; bidx<scanner_batch_size && seeds_detectors.size()<settings.seeds_count; bidx++)
            {
                v2_channel_scanner::winner *p_winner = bidx==0 ? scanner.run_competition() : scanner.retrieve_next();
                if( p_winner==NULL)
                    continue;
                char buf[1024];
                sprintf(buf, "%s", act_in.realm_by_channel(p_winner->seed.channel_idx0).name.c_str());
                seeds_detectors.push_back(p_winner->seed);
                printf("%s[SEED:%04d]  TrnRms=%.4f  TstRms=%.4f  TstCls=%.4f  %s\n",
                       settings.rep_prefix.c_str(),
                       int(seeds_detectors.size()-1),
                       (double)scanner.eval().trn_rms(), (double)scanner.eval().tst_rms(), (double)scanner.eval().tst_cls(), buf);
                fflush(stdout);
            }
        }
        t0_seed = dtickcount()-t0_seed;
        printf("Seed evaluation done in %d ms\n", t0_seed);
        fflush(stdout);
    }

    //
    // Prepare temporary storage for output realms
    //
    std::map<std::string,activities>   map_acout;
    std::map<std::string,net_midlayer> map_mlout;
    for(const realm<channel> &rr : act_in.realms)
    {
        ASSERT(map_acout.count(rr.name)==0);
        ASSERT(map_mlout.count(rr.name)==0);
        map_acout[rr.name] = activities(ntrn, ntst, storage_format::i8);
        map_acout[rr.name].append_realm(realm<channel>::create_grid(rr.name,1,1,0));
        map_mlout[rr.name] = net_midlayer();
        map_mlout[rr.name].append_realm(realm<detector>::create_grid(rr.name,1,1,0));
    }

    //
    // Start contest
    //
    const and_type_t AND_TYPE = and_type_t::AND_MIN;
    const or_type_t  OR_TYPE  = or_type_t::OR_MAX;
    generalizer_evaluator   gen_eval(act_in, lin_eval);
    v2_contest contest(act_in, lin_eval, seeds_detectors, running_seed, settings);
    int t0_det = dtickcount();
    for(int channels_done=0; channels_done<max_channels;)
    {
        contest.run_competition();
        for(int batch_idx=0; batch_idx<settings.batch_size && channels_done<max_channels; batch_idx++)
        {
            //
            // retrieve winner and prepare initial detector
            //
            v2_contest::winner *p_winner = contest.retrieve_winner();
            if( p_winner==NULL )
                continue;
            v2_contest::winner &w = *p_winner;

            //
            // Generalize
            //
            detector gen_det;
            activity8_vector gen_act;
            generalize(act_in, lin_eval, gen_eval, settings, w.seed, &gen_det, &gen_act);

            //
            // Transform
            //
            detector w_det;
            activity8_vector w_act;
            transform(act_in, lin_eval, settings, w.seed, gen_det, &w_det, &w_act);

            //
            // prepare report
            //
            std::string rep = generate_report(act_in, w.seed, w_det);

            // Process winner
            //lin_eval.append_activity(w_act);
            //contest.refresh_winners();
            contest.append_and_refresh_winners(&lin_eval, w_act);
            ASSERT(map_acout.count(w.realm_name)==1);
            ASSERT(map_mlout.count(w.realm_name)==1);
            map_acout[w.realm_name].append1(w_act,rep);
            map_mlout[w.realm_name].append1(w_det);

            printf("%s[V2:%04d]  TrnRms=%.4f  TstRms=%.4f  TstCls=%.4f  %s\n",
                   settings.rep_prefix.c_str(),
                   int(channels_done),
                   (double)lin_eval.trn_rms(), (double)lin_eval.tst_rms(), (double)lin_eval.tst_cls(),
                   rep.c_str());
            fflush(stdout);
            channels_done++;
        }
    }
    t0_det = dtickcount()-t0_det;
    printf("Detector generation is done in %d ms\n", t0_det);
    fflush(stdout);

    //
    // Combine everything
    //
    ml_out = net_midlayer();
    for(auto const &iter: map_mlout)
        ml_out.merge_in(iter.second);
    act_out = activities(ntrn, ntst, storage_format::i8);
    for(auto const &iter: map_acout)
        act_out.merge_in(iter.second);
    output_predictions = lin_eval.outputs();

    t0 = dtickcount()-t0;
    printf("Layer V2 is done in %d ms\n", t0);
    fflush(stdout);
}

#if 0
void v2_layer_exacting(
              linear_evaluator    &global_eval,
              const activities    &act_in,
              v2_settings         settings,
              int                 max_channels,
              double              lambda_reg,
              net_midlayer        &ml_out,
              activities          &act_out,
              predictions         &output_predictions)
{
    int t0 = dtickcount();

    ASSERT(global_eval.outputs().ntrn==act_in.ntrn);
    ASSERT(global_eval.outputs().ntst==act_in.ntst);
    ASSERT(act_in.realms.size()>=1);
    ASSERT(settings.batch_size>=1);

    //
    const int64_t ntrn = act_in.ntrn;
    const int64_t ntst = act_in.ntst;

    //
    // Seeds for session-local RNGs
    //
    std::atomic<int32_t> running_seed(settings.rng_seed);

    //
    // Determine seed channels
    //
    std::vector<seed_detector> seeds_detectors;
    {
        const int scanner_batch_size = 2;
        int t0_seed = dtickcount();
        v2_channel_scanner scanner(act_in, global_eval.outputs(), running_seed, sqrtf(1.0e-5f*ntrn), settings);
        for(;seeds_detectors.size()<settings.seeds_count;)
        {
            for(int bidx=0; bidx<scanner_batch_size && seeds_detectors.size()<settings.seeds_count; bidx++)
            {
                v2_channel_scanner::winner *p_winner = bidx==0 ? scanner.run_competition() : scanner.retrieve_next();
                if( p_winner==NULL)
                    continue;
                char buf[1024];
                sprintf(buf, "%s", act_in.realm_by_channel(p_winner->seed.channel_idx0).name.c_str());
                seeds_detectors.push_back(p_winner->seed);
                printf("%s[SEED:%04d]  TrnRms=%.4f  TstRms=%.4f  TstCls=%.4f  %s\n",
                       settings.rep_prefix.c_str(),
                       int(seeds_detectors.size()-1),
                       (double)scanner.eval().trn_rms(), (double)scanner.eval().tst_rms(), (double)scanner.eval().tst_cls(), buf);
                fflush(stdout);
            }
        }
        t0_seed = dtickcount()-t0_seed;
        printf("Seed evaluation done in %d ms\n", t0_seed);
        fflush(stdout);
    }

    //
    // Prepare temporary storage for output realms
    //
    std::map<std::string,activities>   map_acout;
    std::map<std::string,net_midlayer> map_mlout;
    for(const realm<channel> &rr : act_in.realms)
    {
        ASSERT(map_acout.count(rr.name)==0);
        ASSERT(map_mlout.count(rr.name)==0);
        map_acout[rr.name] = activities(ntrn, ntst, storage_format::i8);
        map_acout[rr.name].append_realm(realm<channel>::create_grid(rr.name,1,1,0));
        map_mlout[rr.name] = net_midlayer();
        map_mlout[rr.name].append_realm(realm<detector>::create_grid(rr.name,1,1,0));
    }

    //
    // Start contest
    //
    generalizer_evaluator   gen_eval(act_in, global_eval); // do we need local generalizer?? cons: memory reallocation, pro: more precise
    int t0_det = dtickcount();
    for(int channels_done=0; channels_done<max_channels;)
    {
        //
        // Local contest
        //
        class candidate_record
        {
        public:
            std::string realm_name;
            seed_detector seed;
            detector det;
            activity8_vector act;
        };
        std::vector<candidate_record> contest_winners;
        linear_evaluator local_eval(global_eval.outputs(), settings.batch_size, lambda_reg);
        v2_contest contest(act_in, local_eval, seeds_detectors, running_seed, settings);
        contest.run_competition();
        for(int batch_idx=0; batch_idx<settings.batch_size; batch_idx++)
        {
            v2_contest::winner *p_winner = contest.retrieve_winner();
            if( p_winner==NULL )
                break;
            v2_contest::winner &w = *p_winner;
            detector gen_det;
            activity8_vector gen_act;
            generalize(act_in, local_eval, gen_eval, settings, w.seed, &gen_det, &gen_act);
            detector w_det;
            activity8_vector w_act;
            transform(act_in, local_eval, settings, w.seed, gen_det, &w_det, &w_act);
            contest_winners.push_back(candidate_record{w.realm_name,w.seed,w_det,w_act});
            contest.append_and_refresh_winners(&local_eval, w_act);
        }
        if( contest_winners.size()==0 )
            continue;
        const int nbatch = contest_winners.size();

        //
        // Select best detectors from contest
        //
        class analysis_record
        {
        public:
            bool retrieved;
            linear_evaluator::daa daa;
            analysis_record():retrieved(false){}
        };
        std::vector<analysis_record> analysis_records(nbatch);
        parallel_for([&](void* _p,int _idx){global_eval.analyze(&contest_winners[_idx].act[0], &analysis_records[_idx].daa);}, NULL, nbatch);
        int nbest = roundf(settings.frac_chosen*settings.batch_size);
        nbest = imax(nbest,1);
        nbest = imin(nbest,contest_winners.size());
        for(int iter=0; iter<nbest && channels_done<max_channels; iter++)
        {
            int idx_best = -1;
            for(int i=0; i<nbatch; i++)
                if( !analysis_records[i].retrieved &&
                        (idx_best<0 ||
                         analysis_records[i].daa.deltar2>analysis_records[idx_best].daa.deltar2) )
                    idx_best = i;
            ASSERT(idx_best>=0);
            const candidate_record &best = contest_winners[idx_best];
            analysis_record &best_rec    = analysis_records[idx_best];

            best_rec.retrieved = true;
            std::vector<linear_evaluator::daa*> update_vec;
            for(int i=0; i<nbatch; i++)
                if( !analysis_records[i].retrieved )
                    update_vec.push_back(&analysis_records[i].daa);

            std::string rep = generate_report(act_in, best.seed, best.det);
            ASSERT(map_acout.count(best.realm_name)==1);
            ASSERT(map_mlout.count(best.realm_name)==1);
            map_acout[best.realm_name].append1(best.act,rep);
            map_mlout[best.realm_name].append1(best.det);
            global_eval.append_and_refresh(best_rec.daa, update_vec);
            printf("%s[V2:%04d]  TrnRms=%.4f  TstRms=%.4f  TstCls=%.4f  %s\n",
                   settings.rep_prefix.c_str(),
                   int(channels_done),
                   (double)global_eval.trn_rms(), (double)global_eval.tst_rms(), (double)global_eval.tst_cls(),
                   rep.c_str());
            fflush(stdout);
            channels_done++;
        }
    }
    t0_det = dtickcount()-t0_det;
    printf("Detector generation is done in %d ms\n", t0_det);
    fflush(stdout);

    //
    // Combine everything
    //
    ml_out = net_midlayer();
    for(auto const &iter: map_mlout)
        ml_out.merge_in(iter.second);
    act_out = activities(ntrn, ntst, storage_format::i8);
    for(auto const &iter: map_acout)
        act_out.merge_in(iter.second);
    output_predictions = global_eval.outputs();

    t0 = dtickcount()-t0;
    printf("Layer V2 is done in %d ms\n", t0);
    fflush(stdout);
}
#endif

}
