/*************************************************************************
StrongNet/Contour Engine training framework.
Copyright (c) 2020 Sergey Bochkanov (ALGLIB project).

>>> SOURCE LICENSE >>>
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation (www.fsf.org); either version 2 of the 
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

A copy of the GNU General Public License is available at
http://www.fsf.org/licensing/licenses
>>> END OF LICENSE >>>
*************************************************************************/
#ifndef LAYER2_H
#define LAYER2_H

#include "snetcommon.h"
#include <atomic>

namespace strongnet
{
class net_midlayer
{
public:
    static std::atomic<int32_t> perf_mlapply;

    int64_t                         ndetectors;
    std::vector<realm<detector> >   realms;

    net_midlayer():ndetectors(0){}

    const realm<detector>& realm_by_detector(int cidx) const;

    // info
    float get_flop_count() const;

    // appends just one detector; layer MUST have just one realm with 1x1 grid
    void append1(const detector &d);

    // appends (possibly non-empty) realm; layer may already have some realms, it will work
    void append_realm(const realm<detector> &r);

    // merges entire layer into this one
    void merge_in(const net_midlayer &merged);

    // apply to activities
    void apply(const activities &act_in, activities &act_out) const;
    void conv_apply(const conv_activities &act_in, conv_activities &act_out) const;
};
void serialize(FILE *f, const net_midlayer *ptr);
void unserialize(FILE *f, net_midlayer *ptr);

class v2_settings
{
public:
    int     min_dist;
    int     max_dist;
    int     seeds_count;
    float   eval_prob;
    float   d2d3_eval_prob;
    int     rng_seed;
    int     l23_batch;
    int     batch_size;
    float   frac_chosen;
    int     min_transform_rad;
    int     ort_cnt;
    int     gain_levels;
    int     dmax;
    int     ncenters;
    rating_t con_rating;
    rating_t gen_rating;
    rating_t trf_rating;
    float   omega_rat;
    std::string rep_prefix;
    float   gen_rating_growth;

    v2_settings():
        min_dist(0),
        max_dist(5),
        seeds_count(0),
        eval_prob(0.1f),
        d2d3_eval_prob(0.90f),
        rng_seed(1),
        l23_batch(20),
        batch_size(50),
        con_rating(rating_t::fastdr2),
        gen_rating(rating_t::fastdr2),
        trf_rating(rating_t::fastdr2),
        omega_rat(1.0e-6f),
        min_transform_rad(0),
        ort_cnt(25),
        gain_levels(4),
        dmax(3),
        ncenters(2),
        frac_chosen(0.1f),
        gen_rating_growth(0)
    {}
};

/*void v2_layer_xx(const predictions &input_predictions,
                const activities &ds_in,
                v2_settings settings,
                int max_channels,
                double lambda_reg,
                net_midlayer &ml_out,
                activities   &act_out,
                predictions  &output_predictions);*/

void v2_layer_xx(linear_evaluator &lin_eval,
                const activities &ds_in,
                v2_settings settings,
                int max_channels,
                net_midlayer &ml_out,
                activities   &act_out,
                predictions  &output_predictions);
#if 0
void v2_layer_exacting(
                linear_evaluator &global_eval,
                const activities &ds_in,
                v2_settings settings,
                int max_channels,
                double lambda_reg,
                net_midlayer &ml_out,
                activities   &act_out,
                predictions  &output_predictions);
#endif

/////////////////////// internal classes not intended for use outside of this file ///////////////////////////

class v2_channel_scanner
{
public:
    class winner
    {
    friend class v2_channel_scanner;
    friend void* v2_channel_scanner_run_parallel(void *ptr);
    private:
        bool             is_set;
        float            rating;
    public:
        seed_detector    seed;
        activity8_vector winner_act;

        void invalidate()
        { is_set = false; seed = seed_detector();  rating = 0.0f; }

        winner()
        { invalidate(); }
    };

    v2_channel_scanner(
               const activities &ai,
               const predictions &bl,
               std::atomic<int32_t> &rs,
               double lr,
               v2_settings s);

    winner* run_competition();
    winner* retrieve_next();
private:
    const activities            &act_in;
    const v2_settings           settings;
    std::atomic<int32_t>        &running_rng_seed;
    const double                lambda_reg;

    // competition
    linear_evaluator            local_eval;
    std::atomic<int32_t>        entries_examined;
    std::vector<winner>         winners;
    std::vector<int>            rng_seeds_to_use;
    int                         contest_size;

public:
    const linear_evaluator& eval() const { return local_eval; }

friend void* v2_channel_scanner_run_parallel(void *ptr);
};

class v2_contest
{
public:
    class winner
    {
    friend class v2_contest;
    friend void* v2_contest_run_parallel(void *ptr);
    private:
        bool             is_set;
        bool             has_daa;
        linear_evaluator::daa daa;
    public:
        seed_detector    seed;
        activity8_vector winner_act;
        float            anrm2;
        float            rating;
        std::string      realm_name;

        void invalidate()
        { is_set = has_daa = false; seed = seed_detector();  daa.invalidate(); realm_name = ""; rating = 0.0f; anrm2 = 0; }

        winner()
        { invalidate(); }

        const linear_evaluator::daa& get_daa() const { ASSERT(has_daa); return daa; }
    };

    static std::atomic<int32_t> pcnt_refresh;
    v2_contest(const activities &ai,
               const linear_evaluator &le,
               const std::vector<seed_detector> &sd,
               std::atomic<int32_t> &rs,
               v2_settings s);
    void run_competition();

    // retrieves current winner and marks it as non-set
    winner* retrieve_winner();

    // called after append() to linear evaluator, forces recomputation of winner metrics
    void refresh_winners();

    // appends activity given by its deep analysis to the linear solver (it must be the one you specified
    // during object creation) and performs quick update on the current winners
    void append_and_refresh_winners(linear_evaluator *eval, const activity8_vector &v_act);
private:
    friend void parallel_v2_refresh_adapter(void *p, int idx);
    void refresh_ith_winner(int winner_idx);

    static const int            line_width;
    const activities            &act_in;
    const linear_evaluator      &lin_eval;
    const std::vector<seed_detector> seeds;
    std::atomic<int32_t>        &running_rng_seed;
    const v2_settings           settings;

    // competition
    std::atomic<int32_t>        lines_examined;
    std::vector<winner>         winners;
    std::vector<int>            rng_seeds_to_use;

friend void* v2_contest_run_parallel(void *ptr);
};

}

#endif
