/*************************************************************************
StrongNet/Contour Engine training framework.
Copyright (c) 2020 Sergey Bochkanov (ALGLIB project).

>>> SOURCE LICENSE >>>
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation (www.fsf.org); either version 2 of the 
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

A copy of the GNU General Public License is available at
http://www.fsf.org/licensing/licenses
>>> END OF LICENSE >>>
*************************************************************************/
#include "layer2c.h"

#include <map>
#include <vector>
#include <algorithm>
#include <random>

#include <emmintrin.h>
#include <smmintrin.h>
#ifdef SNET_AVX2
#include <immintrin.h>
#endif

namespace strongnet
{

const int v2c_analysis::CHUNK_SIZE = 32;

static float compute_rating_omega(float cov2, float nrm2, float omega_rat, int64_t ntotal)
{
    return cov2/(omega_rat*ntotal+nrm2);
}

static float conv_rating(const activities &act_in, const linear_evaluator &lin_eval, float omega_rat)
{
#ifndef SNET_AVX2
#error AVX2 support is required (SSE4 code is not implemented yet)
#endif
    ASSERT(act_in.realms.size()==1);
    ASSERT(act_in.realms[0].ch_offs==0 && act_in.realms[0].grid_depth==1);
    const realm<channel> &rr = act_in.realms[0];
    const int grid_height = rr.grid_height, grid_width = rr.grid_width;
    v2c_analysis a;
    a.allocate(act_in.ntrn, act_in.ntst);
    const int64_t ntotal = a.nchunks*a.chunk_size;
    a.invalidate();
    float *p_dst_f32 = &a.trn_chunks[0];
    for(int k0=0; k0<grid_height; k0++)
        for(int k1=0; k1<grid_width; k1++)
        {
            const int8_t *p_src8 = act_in.get_i8(rr.get_cidx(0,k0,k1));
            for(int64_t i=0; i<ntotal; i++)
                p_dst_f32[i] += p_src8[i];
        }
    for(int64_t i=0; i<ntotal; i++)
        p_dst_f32[i] *= SNET_I2F;
    a.anrm2 = vnrm2(p_dst_f32,ntotal);
    a.is_set = true;
    a.compute_rating(lin_eval, omega_rat);
    return a.rating;
}

static void update_d2_sumact(
                const activities &act_a, const realm<channel> &rr_a,
                int plane_a, int offsa0, int offsa1,
                const activities &act_b, const realm<channel> &rr_b,
                int plane_b, int offsb0, int offsb1,
                const int chunk_size, const int chunk0, const int chunk1,
                activityf_vector &sum_act)
{
#ifndef SNET_AVX2
#error AVX2 support is required (SSE4 code is not implemented yet)
#endif
    ASSERT(chunk_size==32);
    ASSERT(act_a.ntrn==act_b.ntrn);
    ASSERT(act_a.ntst==act_b.ntst);
    ASSERT(rr_a.name==rr_b.name);
    ASSERT(rr_a.grid_height==rr_b.grid_height && rr_a.grid_width==rr_b.grid_width);
    const int grid_height = rr_a.grid_height, grid_width = rr_a.grid_width;
    const int64_t nchunks = chunk1-chunk0;
    int32_t *p_dst_i32 = (int32_t*)(&sum_act[chunk0*chunk_size]);

    //
    // Clear output range
    //
    memset(p_dst_i32, 0, nchunks*chunk_size*sizeof(p_dst_i32[0]));

    //
    // First step: total sum of activities is evaluated as int32_t
    // (we still use g16_sumact as storage because float and int32_t have same size).
    //
    for(int k0=0; k0<grid_height; k0++)
        for(int k1=0; k1<grid_width; k1++)
        {
            // check receptive fields are within grid
            int srca0 = k0+offsa0, srca1 = k1+offsa1;
            int srcb0 = k0+offsb0, srcb1 = k1+offsb1;
            if( srca0<0 || srca0>=grid_height || srca1<0 || srca1>=grid_width )
                continue;
            if( srcb0<0 || srcb0>=grid_height || srcb1<0 || srcb1>=grid_width )
                continue;

            // inputs and outputs
            const int8_t *p_src_a = act_a.get_i8(rr_a.get_cidx(plane_a, srca0, srca1))+chunk0*chunk_size;
            const int8_t *p_src_b = act_b.get_i8(rr_b.get_cidx(plane_b, srcb0, srcb1))+chunk0*chunk_size;

            // process
            alignas(32) int8_t arr8[32];
            for(int64_t i=0; i<nchunks; i++)
            {
                __m256i mm_a = _mm256_stream_load_si256((__m256i*)(p_src_a+i*chunk_size));
                __m256i mm_b = _mm256_stream_load_si256((__m256i*)(p_src_b+i*chunk_size));
                __m256i mm_v = _mm256_min_epi8(mm_a,mm_b);
                _mm256_store_si256((__m256i*)arr8, mm_v);
                int32_t *p_i32 = p_dst_i32+i*chunk_size;
                for(int k=0; k<4; k++)
                {
                    __m256i mm_dst = _mm256_load_si256((__m256i*)(p_i32+k*8));
                    __m256i mm_act = _mm256_cvtepi8_epi32(_mm_set1_epi64x(((const int64_t*)arr8)[k]));
                    mm_dst = _mm256_add_epi32(mm_dst, mm_act);
                    _mm256_store_si256((__m256i*)(p_i32+k*8), mm_dst);
                }
            }
        }

    //
    // Second step: int32_t sums are converted to floats
    //
    const int64_t nchunks_fp32 = nchunks*4;
    const int64_t chunk_fp32_size = 8;
    float scl = SNET_I2F;
    __m256 mm_scl = _mm256_set1_ps(scl);
    float *p_dst_fp32 = (float*)p_dst_i32;
    for(int64_t i=0; i<nchunks_fp32; i++)
    {
        __m256 mm_v;
        mm_v = _mm256_cvtepi32_ps(_mm256_load_si256((__m256i*)(p_dst_i32+i*chunk_fp32_size)));
        mm_v = _mm256_mul_ps(mm_v, mm_scl);
        _mm256_store_ps(p_dst_fp32+i*chunk_fp32_size, mm_v);
    }
}

const int64_t d2_chunks_to_process = 24;

static gain_t choose_best_gain(const activities &act_det, std::vector<gain_t> gain_vals, const linear_evaluator &lin_eval, v2c_settings s)
{
#ifndef SNET_AVX2
#error AVX2 support is required (SSE4 code is not implemented yet)
#endif
    ASSERT(act_det.realms.size()==1);
    ASSERT(act_det.realms[0].ch_offs==0 && act_det.realms[0].grid_depth==1);
    ASSERT(gain_vals.size()>=1);

    const realm<channel> &rr = act_det.realms[0];
    const int grid_height = rr.grid_height, grid_width = rr.grid_width;
    const int64_t ntotal = act_det.ntrn+act_det.ntst;
    gain_t best_gain = gain_t::gain1;
    float  best_dr2 = 0.0f;
    const int CHUNK_SIZE = 32;
    const int64_t nchunks = ntotal/CHUNK_SIZE;
    activityf_vector act_trn(ntotal);
    for(gain_t gain_cand: gain_vals)
    {
        float *p_dst_f32 = &act_trn[0];
        memset(p_dst_f32, 0, ntotal*sizeof(p_dst_f32[0]));
        for(int k0=0; k0<grid_height; k0++)
            for(int k1=0; k1<grid_width; k1++)
            {
                const int8_t *p_src8 = act_det.get_i8(rr.get_cidx(0,k0,k1));
                for(int64_t i=0; i<ntotal; i++) //AVX!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                    p_dst_f32[i] += imin(int(p_src8[i])*int(gain_cand),SNET_IMAX);
            }
        float dr2_cand = lin_eval.estimate_trn_covariation(p_dst_f32)/(s.omega_rat*act_det.ntrn+vnrm2(p_dst_f32,act_det.ntrn));
        if( dr2_cand>best_dr2 )
        {
            best_dr2 = dr2_cand;
            best_gain = gain_cand;
        }
    }
    return best_gain;
}

static void apply_gain(activities &act_det, gain_t gain)
{
#ifndef SNET_AVX2
#error AVX2 support is required (SSE4 code is not implemented yet)
#endif
    ASSERT(act_det.realms.size()==1);
    ASSERT(act_det.realms[0].ch_offs==0 && act_det.realms[0].grid_depth==1);
    const realm<channel> &rr = act_det.realms[0];
    const int grid_height = rr.grid_height, grid_width = rr.grid_width;
    const int64_t ntotal = act_det.ntrn+act_det.ntst;
    const int CHUNK_SIZE = 32;
    const int64_t nchunks = ntotal/CHUNK_SIZE;
    for(int k0=0; k0<grid_height; k0++)
        for(int k1=0; k1<grid_width; k1++)
        {
            int8_t *p_act8 = act_det.get_i8(rr.get_cidx(0,k0,k1));
            for(int64_t i=0; i<ntotal; i++) //AVX!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                p_act8[i] = imin(int(p_act8[i])*int(gain),SNET_IMAX);
        }
}

static void generate_activity_sum(const activities &act_det, activityf_vector &sum)
{
    ASSERT(act_det.realms.size()==1);
    ASSERT(act_det.realms[0].ch_offs==0 && act_det.realms[0].grid_depth==1);

    const realm<channel> &rr = act_det.realms[0];
    const int grid_height = rr.grid_height, grid_width = rr.grid_width;
    const int64_t ntotal = act_det.ntrn+act_det.ntst;
    sum = activityf_vector(ntotal,0.0f);
    for(int k0=0; k0<grid_height; k0++)
        for(int k1=0; k1<grid_width; k1++)
        {
            const int8_t *p_src = act_det.get_i8(rr.get_cidx(0,k0,k1));
            float *p_dst = &sum[0];
            for(int64_t i=0; i<ntotal; i++)
                p_dst[i] += p_src[i];
        }
    float scl = SNET_I2F;
    for(int64_t i=0; i<ntotal; i++)
        sum[i] *= scl;
}

static void generate_activity_max(const activities &act_det, activityf_vector &amax)
{
    ASSERT(act_det.realms.size()==1);
    ASSERT(act_det.realms[0].ch_offs==0 && act_det.realms[0].grid_depth==1);

    const realm<channel> &rr = act_det.realms[0];
    const int grid_height = rr.grid_height, grid_width = rr.grid_width;
    const int64_t ntotal = act_det.ntrn+act_det.ntst;
    amax = activityf_vector(ntotal,0.0f);
    for(int k0=0; k0<grid_height; k0++)
        for(int k1=0; k1<grid_width; k1++)
        {
            const int8_t *p_src = act_det.get_i8(rr.get_cidx(0,k0,k1));
            float *p_dst = &amax[0];
            for(int64_t i=0; i<ntotal; i++)
                p_dst[i] = fmaxf(p_dst[i],p_src[i]);
        }
    float scl = SNET_I2F;
    for(int64_t i=0; i<ntotal; i++)
        amax[i] *= scl;
}

static void generalize(const activities &act_in,
                       const linear_evaluator &lin_eval,
                       conv_generalizer_evaluator  &gen_eval,
                       const v2c_settings settings,
                       const conv_seed_detector &seed,
                       conv_detector *d_out,
                       activities *act_out)
{
    ASSERT(seed.inputs_cnt()>=1);
    for(int k=0; k<seed.inputs_cnt(); k++)
        ASSERT(seed.get_input(k).realm_name==seed.get_input(0).realm_name);

    //
    // Prepare detector and its activity
    //
    const std::string realm_name = seed.get_input(0).realm_name;
    conv_detector   &det = *d_out;
    activities      &act = *act_out;
    det = conv_detector(seed);
    det.generate_activity_slow(act_in, act);
    int w_inputs_cnt = seed.inputs_cnt();

    //
    // Generalize
    //
    float rating = conv_rating(act, lin_eval, settings.omega_rat);
    for(;;)
    {
        float         gen_cand_rating;
        conv_detector gen_cand_det;
        activities    gen_cand_act;
        bool something_changed = false;
        for(int input_idx=0; input_idx<w_inputs_cnt; input_idx++)
        {
            int input_size = det.get_input(input_idx).connections.size();
            const realm<channel> &rr = act_in.realm_by_name(realm_name);
            ASSERT(input_size>=1);
            if( input_size>=SNET_MAX_V2C_GEN_COUNT )
                continue;
            gen_eval.start_evaluation(det, input_idx, and_type_t::AND_MIN, or_type_t::OR_MAX, settings.use_weights);
            for(int c=0; c<rr.grid_depth; c++)
                gen_eval.evaluate_generalization(c);
            gen_eval.get_winner(&gen_cand_rating, &gen_cand_det, &gen_cand_act);
            ASSERT(settings.gen_rating_growth>0);
            if( gen_cand_rating<=rating*settings.gen_rating_growth )
                continue;
            something_changed = true;
            rating = gen_cand_rating;
            det    = gen_cand_det;
            act    = gen_cand_act;
        }
        if( !something_changed )
            break;
    }
}

int conv_seed_detector::inputs_cnt() const
{
    if( input0.plane_idx<0 )
        return 0;
    if( input1.plane_idx<0 )
        return 1;
    if( input2.plane_idx<0 )
        return 2;
    if( input3.plane_idx<0 )
        return 3;
    if( input4.plane_idx<0 )
        return 4;
    return 5;
}

const conv_seed_detector::input& conv_seed_detector::get_input(int i) const
{
    ASSERT(i<inputs_cnt());
    if( i==0 )
        return input0;
    if( i==1 )
        return input1;
    if( i==2 )
        return input2;
    if( i==3 )
        return input3;
    return input4;
}

conv_seed_detector::input& conv_seed_detector::get_input(int i)
{
    ASSERT(i<inputs_cnt());
    if( i==0 )
        return input0;
    if( i==1 )
        return input1;
    if( i==2 )
        return input2;
    if( i==3 )
        return input3;
    return input4;
}

// generate composite detector using two input ones
conv_seed_detector conv_seed_detector::compose(const conv_seed_detector &d0, int offs00, int offs01,
                                               const conv_seed_detector &d1, int offs10, int offs11)
{
    // check
    ASSERT(d0.inputs_cnt()>0 && d1.inputs_cnt()>0);
    ASSERT(d0.inputs_cnt()+d1.inputs_cnt()<=5);

    //
    conv_seed_detector r;
    conv_seed_detector::input *inputs[5] = { &r.input0, &r.input1, &r.input2, &r.input3, &r.input4 };
    int inputs_done = 0;
    for(int i=0; i<d0.inputs_cnt(); i++)
    {
        conv_seed_detector::input &in = *(inputs[inputs_done]);
        in = d0.get_input(i);
        in.offs0 += offs00;
        in.offs1 += offs01;
        inputs_done++;
    }
    for(int i=0; i<d1.inputs_cnt(); i++)
    {
        conv_seed_detector::input &in = *(inputs[inputs_done]);
        in = d1.get_input(i);
        in.offs0 += offs10;
        in.offs1 += offs11;
        inputs_done++;
    }
    return r;
}

void conv_seed_detector::generate_activity_slow(const activities &act_in, activities &act_out, int downsample) const
{
    ASSERT(inputs_cnt()>=1 && inputs_cnt()<=5);
    for(int k=0; k<inputs_cnt(); k++)
    {
        const realm<channel> &r0 = act_in.realm_by_name(input0.realm_name);
        const realm<channel> &rk = act_in.realm_by_name(get_input(k).realm_name);
        ASSERT(rk.name==r0.name && rk.grid_height==r0.grid_height && rk.grid_width==r0.grid_width);
        ASSERT(rk.grid_height%downsample==0 && rk.grid_width%downsample==0);
    }
    const realm<channel> &rr = act_in.realm_by_name(input0.realm_name);
    std::string realm_name = rr.name;
    const int64_t grid_height = rr.grid_height, grid_width = rr.grid_width;
    const int64_t ntotal = act_in.ntrn+act_in.ntst;


    act_out = activities(act_in.ntrn, act_in.ntst, storage_format::i8);
    act_out.append_realm(realm<channel>::create_grid(realm_name, grid_height/downsample, grid_width/downsample, 1));
    realm<channel> &rr_out = act_out.realms.back();

    std::vector<int8_t> buf(ntotal);
    for(int k0=0; k0<grid_height; k0++)
        for(int k1=0; k1<grid_width; k1++)
        {
            int8_t *p_buf8 = &buf[0];
            memset(p_buf8, SNET_IMAX, ntotal);
            for(int input_idx=0; input_idx<inputs_cnt(); input_idx++)
            {
                const input &in = get_input(input_idx);
                int src0 = k0+in.offs0, src1 = k1+in.offs1;
                if( src0<0 || src0>=grid_height || src1<0 || src1>=grid_width )
                {
                    memset(p_buf8, 0, ntotal);
                    continue;
                }
                const int8_t *p_src8 = act_in.get_i8(rr.get_cidx(in.plane_idx, src0, src1));
                for(int64_t i=0; i<ntotal; i++)
                {
                    int8_t va = p_src8[i], vb = p_buf8[i];
                    int8_t v = va<vb ? va : vb;
                    p_buf8[i] = v;
                }
            }
            int8_t *p_dst8 = act_out.get_i8(rr_out.get_cidx(0, k0/downsample, k1/downsample));
            for(int64_t i=0; i<ntotal; i++)
            {
                int8_t va = p_buf8[i], vb = p_dst8[i];
                p_dst8[i] = va>vb ? va : vb;
            }
        }
}

void conv_seed_detector::generate_activity_sum(const activities &act_in, activityf_vector &v_out) const
{
    ASSERT(inputs_cnt()>=1 && inputs_cnt()<=5);
    for(int k=0; k<inputs_cnt(); k++)
    {
        const realm<channel> &r0 = act_in.realm_by_name(input0.realm_name);
        const realm<channel> &rk = act_in.realm_by_name(get_input(k).realm_name);
        ASSERT(rk.name==r0.name && rk.grid_height==r0.grid_height && rk.grid_width==r0.grid_width);
    }
    const realm<channel> &rr = act_in.realm_by_name(input0.realm_name);
    const int64_t grid_height = rr.grid_height, grid_width = rr.grid_width;
    const int64_t ntotal = act_in.ntrn+act_in.ntst;

    v_out.resize(act_in.ntrn+act_in.ntst);
    memset_vector(v_out, 0);
    std::vector<int8_t> buf(ntotal);
    float *p_dstf = &v_out[0];
    for(int k0=0; k0<grid_height; k0++)
        for(int k1=0; k1<grid_width; k1++)
        {
            int8_t *p_buf8 = &buf[0];
            memset(p_buf8, SNET_IMAX, ntotal);
            for(int input_idx=0; input_idx<inputs_cnt(); input_idx++)
            {
                const input &in = get_input(input_idx);
                int src0 = k0+in.offs0, src1 = k1+in.offs1;
                if( src0<0 || src0>=grid_height || src1<0 || src1>=grid_width )
                {
                    memset(p_buf8, 0, ntotal);
                    continue;
                }
                act8_merge_min_simd(p_buf8, act_in.get_i8(rr.get_cidx(in.plane_idx, src0, src1)), ntotal);
            }
            for(int64_t i=0; i<ntotal; i++)
                p_dstf[i] += p_buf8[i];
        }
    for(int64_t i=0; i<ntotal; i++)
        p_dstf[i] *= SNET_I2F;
}

conv_detector::conv_detector(const conv_seed_detector &seed, gain_t g):
    gain(g), input0(seed.input0), input1(seed.input1), input2(seed.input2), input3(seed.input3), input4(seed.input4)
{
}

const conv_detector::input& conv_detector::get_input(int i) const
{
    ASSERT(i<inputs_cnt());
    if( i==0 )
        return input0;
    if( i==1 )
        return input1;
    if( i==2 )
        return input2;
    if( i==3 )
        return input3;
    return input4;
}

conv_detector::input& conv_detector::get_input(int i)
{
    ASSERT(i<inputs_cnt());
    if( i==0 )
        return input0;
    if( i==1 )
        return input1;
    if( i==2 )
        return input2;
    if( i==3 )
        return input3;
    return input4;
}

int conv_detector::inputs_cnt() const
{
    if( input0.connections.size()==0 )
        return 0;
    if( input1.connections.size()==0 )
        return 1;
    if( input2.connections.size()==0 )
        return 2;
    if( input3.connections.size()==0 )
        return 3;
    if( input4.connections.size()==0 )
        return 4;
    return 5;
}

void conv_detector::add_conn(int input_idx, int plane_idx, weight_type w)
{
    ASSERT(input_idx>=0 && input_idx<inputs_cnt());
    get_input(input_idx).connections.push_back(conn(plane_idx,w));
}

void conv_detector::generate_activity_slow(const activities &act_in, activities &act_out, int downsample) const
{
    std::vector<int> inputs;
    for(int k=0; k<inputs_cnt(); k++)
        inputs.push_back(k);
    generate_activity_x(act_in, inputs, act_out, downsample);
}

void conv_detector::generate_activity_sum(const activities &act_in, activityf_vector &v_out) const
{
    // integrity checks
    for(int k=0; k<inputs_cnt(); k++)
    {
        const realm<channel> &r0 = act_in.realm_by_name(input0.realm_name);
        const realm<channel> &rk = act_in.realm_by_name(get_input(k).realm_name);
        ASSERT(rk.name==r0.name && rk.grid_height==r0.grid_height && rk.grid_width==r0.grid_width);
    }

    // process
    const int64_t ntotal = act_in.ntrn+act_in.ntst;
    const realm<channel> &rr = act_in.realm_by_name(input0.realm_name);
    activity8_vector buf8(ntotal), act8(ntotal);
    activityf_vector buff(ntotal), actf(ntotal);
    std::vector<int> inputs;
    for(int k=0; k<inputs_cnt(); k++)
        inputs.push_back(k);
    v_out = activityf_vector(ntotal,0.0f);
    float *p_out_f32 = &v_out[0];
    for(int k0=0; k0<rr.grid_height; k0++)
        for(int k1=0; k1<rr.grid_width; k1++)
        {
            if( act_in.is_i8() )
            {
                // 8-bit input
                int8_t *p_act8 = &act8[0];
                generate_activity_for_location(inputs, act_in, rr, k0, k1, &buf8[0], p_act8);
                for(int64_t i=0; i<ntotal; i++) // TODO: AVX2!!!!!!!!!!!!!
                    p_out_f32[i] += p_act8[i];
            }
            else if( act_in.is_f32() )
            {
                // 32-bit input
                float *p_actf = &actf[0];
                generate_activity_for_location(inputs, act_in, rr, k0, k1, &buff[0], p_actf);
                for(int64_t i=0; i<ntotal; i++) // TODO: AVX2!!!!!!!!!!!!!
                    p_out_f32[i] += p_actf[i];

            }
            else
                ASSERT(0);
        }

    //
    // Postprocessing for 8-bit inputs
    //
    if( act_in.is_i8() )
    {
        for(int64_t i=0; i<ntotal; i++)
            p_out_f32[i] *= SNET_I2F;
    }
}

void conv_detector::generate_activity_except_for_input(const activities &act_in, int input_idx, activities &act_out, int downsample) const
{
    ASSERT(int(gain)==1);

    // integrity checks
    ASSERT(inputs_cnt()>=1 && inputs_cnt()<=5);
    for(int k=0; k<inputs_cnt(); k++)
    {
        const realm<channel> &r0 = act_in.realm_by_name(input0.realm_name);
        const realm<channel> &rk = act_in.realm_by_name(get_input(k).realm_name);
        ASSERT(rk.name==r0.name && rk.grid_height==r0.grid_height && rk.grid_width==r0.grid_width);
        ASSERT(rk.grid_height%downsample==0 && rk.grid_width%downsample==0);
    }
    ASSERT(input_idx>=0 && input_idx<inputs_cnt());

    //
    // Quick exit for just one input
    //
    if( inputs_cnt()==1 )
    {
        const realm<channel> &rr = act_in.realm_by_name(input0.realm_name);
        std::string realm_name = rr.name;
        const int64_t grid_height = rr.grid_height, grid_width = rr.grid_width;
        const int64_t ntotal = act_in.ntrn+act_in.ntst;
        act_out = activities(act_in.ntrn, act_in.ntst, storage_format::i8);
        act_out.append_realm(realm<channel>::create_grid(realm_name, grid_height/downsample, grid_width/downsample, 1));
        realm<channel> &rr_out = act_out.realms.back();
        for(int k0=0; k0<grid_height; k0++)
            for(int k1=0; k1<grid_width; k1++)
            {
                int8_t *p_dst8 = act_out.get_i8(rr_out.get_cidx(0, k0/downsample, k1/downsample));
                memset(p_dst8, SNET_IMAX, ntotal*sizeof(p_dst8[0]));
            }
        return;
    }

    //
    // More than one input
    //
    std::vector<int> inputs;
    for(int k=0; k<inputs_cnt(); k++)
        if( k!=input_idx )
            inputs.push_back(k);
    generate_activity_x(act_in, inputs, act_out, downsample);
}

void conv_detector::generate_activity_for_input(const activities &act_in, int input_idx, activities &act_out, int downsample) const
{
    ASSERT(inputs_cnt()>=1 && inputs_cnt()<=5);
    ASSERT(input_idx>=0 && input_idx<inputs_cnt());
    generate_activity_x(act_in, {input_idx}, act_out, downsample);
}

void conv_detector::generate_activity_for_location(
        std::vector<int> inputs_to_process,
        const activities &act_bottom,
        const realm<channel> &rr_work,
        const int k0, const int k1,
        int8_t *p_buf,
        int8_t *p_out) const
{
    const int igain = int(gain);
    const int64_t ntotal = act_bottom.ntrn+act_bottom.ntst;
    const int64_t grid_height = rr_work.grid_height, grid_width = rr_work.grid_width;
    memset(p_out, SNET_IMAX, ntotal*sizeof(p_out[0]));
    for(int ii=0; ii<inputs_to_process.size(); ii++)
    {
        ASSERT(inputs_to_process[ii]>=0 && inputs_to_process[ii]<inputs_cnt());
        const input &in = get_input(inputs_to_process[ii]);

        // check location of input with respect to grid bounds
        int src0 = k0+in.offs0, src1 = k1+in.offs1;
        if( src0<0 || src0>=grid_height || src1<0 || src1>=grid_width )
        {
            memset(p_out, 0, ntotal*sizeof(p_out[0]));
            continue;
        }

        // prepare input buffer, compute OR over buffer elements
        if( in.connections.size()!=1 || igain!=1)
        {
            memset(p_buf, 0, ntotal*sizeof(p_buf[0]));
            for(const conn &c: in.connections)
            {
                if( c.weight!=weight_type::w100 || gain==gain_t::gain1 )
                {
                    ASSERT(gain==gain_t::gain1);
                    act8_merge_wmax_simd(p_buf,
                                         act_bottom.get_i8(rr_work.get_cidx(c.plane_idx, src0, src1)),
                                         c.weight,
                                         ntotal);
                }
                else
                {
                    ASSERT(c.weight==weight_type::w100);
                    act8_merge_gmax_simd(p_buf,
                                         act_bottom.get_i8(rr_work.get_cidx(c.plane_idx, src0, src1)),
                                         gain,
                                         ntotal);
                }
            }
        }
        else
        {
            ASSERT(in.connections[0].weight==weight_type::w100 && igain==1);
            memcpy(p_buf,
                    act_bottom.get_i8(rr_work.get_cidx(in.connections[0].plane_idx, src0, src1)),
                    ntotal*sizeof(p_buf[0]));
        }

        // compute AND with total buffer
        act8_merge_min_simd(p_out, p_buf, ntotal);
    }
}

void conv_detector::generate_activity_for_location(
        std::vector<int> inputs_to_process,
        const activities &act_bottom,
        const realm<channel> &rr_work,
        const int k0, const int k1,
        float *p_buff,
        float *p_outf) const
{
    const int igain = int(gain);
    const int64_t ntotal = act_bottom.ntrn+act_bottom.ntst;
    const int64_t grid_height = rr_work.grid_height, grid_width = rr_work.grid_width;
    ASSERT(igain==1);
    ASSERT(act_bottom.is_f32());
    memset(p_outf, SNET_IMAX, ntotal*sizeof(p_outf[0]));
    for(int ii=0; ii<inputs_to_process.size(); ii++)
    {
        ASSERT(inputs_to_process[ii]>=0 && inputs_to_process[ii]<inputs_cnt());
        const input &in = get_input(inputs_to_process[ii]);

        // check location of input with respect to grid bounds
        int src0 = k0+in.offs0, src1 = k1+in.offs1;
        if( src0<0 || src0>=grid_height || src1<0 || src1>=grid_width )
        {
            memset(p_outf, 0, ntotal*sizeof(p_outf[0]));
            continue;
        }

        // prepare input buffer, compute OR over buffer elements
        if( in.connections.size()!=1 || igain!=1)
        {
            memset(p_buff, 0, ntotal*sizeof(p_buff[0]));
            for(const conn &c: in.connections)
            {
                float w = 1.0f;
                if( c.weight==weight_type::w050 )
                    w = 0.50f;
                else if( c.weight==weight_type::w075 )
                    w = 0.75f;
                else
                    ASSERT(c.weight==weight_type::w100);
                ASSERT(gain==gain_t::gain1);
                fmerge_wmax(p_buff,
                            act_bottom.get_f32(rr_work.get_cidx(c.plane_idx, src0, src1)),
                            w,
                            ntotal);
            }
        }
        else
        {
            ASSERT(in.connections[0].weight==weight_type::w100 && igain==1);
            memcpy(p_buff,
                    act_bottom.get_f32(rr_work.get_cidx(in.connections[0].plane_idx, src0, src1)),
                    ntotal*sizeof(p_buff[0]));
        }

        // compute AND with total buffer
        fmerge_min(p_outf, p_buff, ntotal);
    }
}

void conv_detector::generate_activity_x(const activities &act_in, std::vector<int> inputs, activities &act_out, int downsample) const
{
    ASSERT(inputs_cnt()>=1 && inputs_cnt()<=5);
    for(int k=0; k<inputs_cnt(); k++)
    {
        const realm<channel> &r0 = act_in.realm_by_name(input0.realm_name);
        const realm<channel> &rk = act_in.realm_by_name(get_input(k).realm_name);
        ASSERT(rk.name==r0.name && rk.grid_height==r0.grid_height && rk.grid_width==r0.grid_width);
        ASSERT(rk.grid_height%downsample==0 && rk.grid_width%downsample==0);
    }
    ASSERT(inputs.size()>=1 && inputs.size()<=inputs_cnt());
    for(int i: inputs)
        ASSERT(i>=0 && i<inputs_cnt());
    const realm<channel> &rr = act_in.realm_by_name(input0.realm_name);
    std::string realm_name = rr.name;
    const int64_t grid_height = rr.grid_height, grid_width = rr.grid_width;
    const int64_t ntotal = act_in.ntrn+act_in.ntst;

    act_out = activities(act_in.ntrn, act_in.ntst, storage_format::i8);
    act_out.append_realm(realm<channel>::create_grid(realm_name, grid_height/downsample, grid_width/downsample, 1));
    realm<channel> &rr_out = act_out.realms.back();

    activity8_vector input_buf(ntotal), total_buf(ntotal);
    for(int k0=0; k0<grid_height; k0++)
        for(int k1=0; k1<grid_width; k1++)
        {
            generate_activity_for_location(inputs, act_in, rr, k0, k1, &input_buf[0], &total_buf[0]);
            act8_merge_max_simd(act_out.get_i8(rr_out.get_cidx(0, k0/downsample, k1/downsample)), &total_buf[0], ntotal);
        }
}

void serialize(FILE *f, const conv_detector::conn *ptr)
{
    write_magic8(f, 17);
    serialize_as_int64(f, &ptr->weight);
    serialize(f, &ptr->plane_idx);
}

void serialize(FILE *f, const conv_detector::input *ptr)
{
    write_magic8(f, 9);
    serialize(f, &ptr->realm_name);
    serialize(f, &ptr->realm_depth);
    serialize(f, &ptr->realm_dim0);
    serialize(f, &ptr->realm_dim1);
    serialize(f, &ptr->offs0);
    serialize(f, &ptr->offs1);
    serialize(f, &ptr->connections);
    write_magic8(f, 52);
}

void serialize(FILE *f, const conv_detector *ptr)
{
    write_magic8(f, 69);
    serialize_as_int64(f, &ptr->gain);
    serialize(f, &ptr->input0);
    serialize(f, &ptr->input1);
    serialize(f, &ptr->input2);
    serialize(f, &ptr->input3);
    serialize(f, &ptr->input4);
    serialize(f, &ptr->desc);
    write_magic8(f, 97);
}

void unserialize(FILE *f, conv_detector::conn *ptr)
{
    check_magic8(f, 17);
    unserialize_as_int64(f, &ptr->weight);
    unserialize(f, &ptr->plane_idx);
}

void unserialize(FILE *f, conv_detector::input *ptr)
{
    check_magic8(f, 9);
    unserialize(f, &ptr->realm_name);
    unserialize(f, &ptr->realm_depth);
    unserialize(f, &ptr->realm_dim0);
    unserialize(f, &ptr->realm_dim1);
    unserialize(f, &ptr->offs0);
    unserialize(f, &ptr->offs1);
    unserialize(f, &ptr->connections);
    check_magic8(f, 52);
}

void unserialize(FILE *f, conv_detector *ptr)
{
    check_magic8(f, 69);
    unserialize_as_int64(f, &ptr->gain);
    unserialize(f, &ptr->input0);
    unserialize(f, &ptr->input1);
    unserialize(f, &ptr->input2);
    unserialize(f, &ptr->input3);
    unserialize(f, &ptr->input4);
    unserialize(f, &ptr->desc);
    check_magic8(f, 97);
}

void conv_midlayer::append1(const conv_detector &d)
{
    ASSERT(realms.size()==1 && realms[0].grid_width==1 && realms[0].grid_height==1);
    ASSERT(realms[0].ch_offs==0);
    ASSERT(realms[0].grid_depth==ndetectors);
    realms[0].append_element(d);
    ndetectors++;
}

void conv_midlayer::append_realm(const realm<conv_detector> &r)
{
    // add and reallocate
    realms.push_back(r.rebase(ndetectors));
    ndetectors += r.get_nchannels();
}

void conv_midlayer::merge_in(const conv_midlayer &merged)
{
    // rebase realms
    int64_t cur_offs = ndetectors;
    for(const realm<conv_detector> &r: merged.realms)
    {
        realms.push_back(r.rebase(cur_offs));
        cur_offs += r.elements.size();
    }
    ndetectors += merged.ndetectors;
}

void conv_midlayer::apply_and_sum_f32(const activities &act_in, activities &act_out) const
{
    activityf_vector v_actf;
    act_out = activities(act_in.ntrn, act_in.ntst, storage_format::f32);
    for(const realm<conv_detector> &rr: realms)
    {
        // integrity check
        ASSERT(rr.grid_height==1 && rr.grid_width==1 && rr.grid_depth==rr.elements.size());

        // append realm
        int base_idx = act_out.append_realm(realm<channel>::create_grid(rr.name,1,1,rr.grid_depth));

        // overwrite activities
        for(int k=0; k<rr.elements.size(); k++)
        {
            const conv_detector &d = rr.elements[k];
            d.generate_activity_sum(act_in, v_actf);
            act_out.set_activity(base_idx+k, v_actf);
        }
    }
}

void conv_midlayer::apply(const activities &act_in, split_t post_divide, accumulate_t accumulate, activities &act_out) const
{
    ASSERT(post_divide==split_t::entire && accumulate==accumulate_t::sum);
    apply_and_sum_f32(act_in, act_out);
}

void serialize(FILE *f, const conv_midlayer *ptr)
{
    write_magic8(f, 103);
    serialize(f, &ptr->ndetectors);
    serialize(f, &ptr->realms);
    write_magic8(f, 59);
}

void unserialize(FILE *f, conv_midlayer *ptr)
{
    check_magic8(f, 103);
    unserialize(f, &ptr->ndetectors);
    unserialize(f, &ptr->realms);
    check_magic8(f, 59);
}

void v2c_analysis::allocate(int trn, int tst)
{
    nchunks = trn/chunk_size;
    trn_chunks.resize(nchunks*chunk_size);
}

void v2c_analysis::invalidate()
{
    is_set = false;
    seed_det = conv_seed_detector();
    memset(&trn_chunks[0], 0, trn_chunks.size()*sizeof(trn_chunks[0]));
    rating = 0.0f;
    anrm2 = 0.0f;
    realm_name = "";
}

void v2c_analysis::analyze_plane(const activities &act_in, const realm<channel> &rr, int plane_idx, const linear_evaluator &lin_eval, float omega_rat)
{
    allocate(act_in.ntrn, act_in.ntst);
    invalidate();

    const int grid_height = rr.grid_height, grid_width = rr.grid_width;
    const int64_t ntotal = nchunks*chunk_size;
    float *p_dst_f32 = &trn_chunks[0];
    for(int k0=0; k0<grid_height; k0++)
        for(int k1=0; k1<grid_width; k1++)
        {
            const int8_t *p_src8 = act_in.get_i8(rr.get_cidx(plane_idx, k0, k1));
            for(int64_t i=0; i<ntotal; i++)
                p_dst_f32[i] += p_src8[i];
        }
    for(int64_t i=0; i<ntotal; i++)
        p_dst_f32[i] *= SNET_I2F;
    anrm2 = vnrm2(p_dst_f32,ntotal);
    is_set = true;
    compute_rating(lin_eval, omega_rat);
}

void v2c_analysis::analyze_planes_max(const activities &act0, const realm<channel> &rr0, int plane0,
                                      const activities &act1, const realm<channel> &rr1, int plane1,
                                      const linear_evaluator &lin_eval, float omega_rat)
{
#ifndef SNET_AVX2
#error AVX2 support is required (SSE4 code is not implemented yet)
#endif
    ASSERT(chunk_size==32);
    ASSERT(rr0.grid_height==rr1.grid_height && rr0.grid_width==rr1.grid_width);

    allocate(act0.ntrn, act0.ntst);
    invalidate();

    const int grid_height = rr0.grid_height, grid_width = rr0.grid_width;
    const int64_t ntotal = nchunks*chunk_size;
    int32_t *p_dst_i32 = (int32_t*)(&trn_chunks[0]);
    for(int k0=0; k0<grid_height; k0++)
        for(int k1=0; k1<grid_width; k1++)
        {
            const int8_t *p_src_a = act0.get_i8(rr0.get_cidx(plane0, k0, k1));
            const int8_t *p_src_b = act1.get_i8(rr1.get_cidx(plane1, k0, k1));
            alignas(32) int8_t arr8[32];
            for(int64_t i=0; i<nchunks; i++)
            {
                __m256i mm_a8 = _mm256_stream_load_si256((__m256i*)(p_src_a+i*32));
                __m256i mm_b8 = _mm256_stream_load_si256((__m256i*)(p_src_b+i*32));
                __m256i mm_act= _mm256_max_epi8(mm_a8,mm_b8);
                _mm256_store_si256((__m256i*)arr8, mm_act);
                int32_t *p = p_dst_i32+i*32;
                for(int k=0; k<4; k++)
                {
                    __m256i mm_dst32 = _mm256_load_si256((__m256i*)(p+k*8));
                    __m256i mm_act32 = _mm256_cvtepi8_epi32(_mm_set1_epi64x(((const int64_t*)arr8)[k]));
                    mm_dst32 = _mm256_add_epi32(mm_dst32, mm_act32);
                    _mm256_store_si256((__m256i*)(p+k*8), mm_dst32);
                }
            }
        }
    float *p_dst_f32 = &trn_chunks[0];
    for(int64_t i=0; i<ntotal; i++)
        p_dst_f32[i] = SNET_I2F*p_dst_i32[i];
    anrm2 = vnrm2(p_dst_f32,ntotal);
    is_set = true;
    compute_rating(lin_eval, omega_rat);
}

void v2c_analysis::compute_rating(const linear_evaluator &lin_eval, float omega_rat)
{
    rating = 0.0f;
    for(int i=0; i<lin_eval.get_nclasses(); i++)
    {
        const float *p_res = lin_eval.get_res32()+i*lin_eval.get_res_stride();
        const float *p_sum = &trn_chunks[0];
        const int64_t ntotal = chunk_size*nchunks;
        float cov = fdot_simd(p_res, p_sum, ntotal);
        rating += cov*cov/(omega_rat*ntotal+anrm2);
    }
}

void v2c_realm::invalidate_analyses()
{
    for(v2c_analysis &a: entry_winners)
        a.invalidate();
}

static std::string generate_report(const conv_detector &winner)
{
    auto vector_printer = [](const std::vector<conv_detector::conn> &c)->std::string
    {
        std::string r ="{";
        for(int k=0; k<c.size(); k++)
        {
            const conv_detector::conn &ck = c[k];
            std::string rk;
            if( ck.weight==weight_type::w100 )
                rk = "1.00";
            else if( ck.weight==weight_type::w075 )
                rk = "0.75";
            else if( ck.weight==weight_type::w050 )
                rk = "0.50";
            else
                ASSERT(0);
            r += rk+"*c"+std::to_string(ck.plane_idx);
            if( k<c.size()-1 )
                r += ",";
        }
        r += "}";
        return r;
    };

    char buf[4096];
    if( winner.inputs_cnt()==1 )
    {
        std::string conns0 = vector_printer(winner.input0.connections);
        sprintf(buf, "%14s*%d D1(g=%d) = %s[0,0]",
                winner.input0.realm_name.c_str(),
                int(winner.gain),
                int(winner.input0.connections.size()),
                conns0.c_str());
        return buf;
    }
    else if( winner.inputs_cnt()==2 )
    {
        std::string conns0 = vector_printer(winner.input0.connections);
        std::string conns1 = vector_printer(winner.input1.connections);
        sprintf(buf, "%14s*%d D2(g=%d,%d) =%s[%+d,%+d] & %s[%+d,%+d]",
                winner.input0.realm_name.c_str(),
                int(winner.gain),
                int(winner.input0.connections.size()), int(winner.input1.connections.size()),
                conns0.c_str(), int(winner.input0.offs0), int(winner.input0.offs1),
                conns1.c_str(), int(winner.input1.offs0), int(winner.input1.offs1));
        return buf;
    }
    else if( winner.inputs_cnt()==3 )
    {
        std::string conns0 = vector_printer(winner.input0.connections);
        std::string conns1 = vector_printer(winner.input1.connections);
        std::string conns2 = vector_printer(winner.input2.connections);
        sprintf(buf, "%14s*%d D3(g=%d,%d,%d) = %s[%+d,%+d] & %s[%+d,%+d] & %s[%+d,%+d]",
                winner.input0.realm_name.c_str(),
                int(winner.gain),
                int(winner.input0.connections.size()), int(winner.input1.connections.size()), int(winner.input2.connections.size()),
                conns0.c_str(), int(winner.input0.offs0), int(winner.input0.offs1),
                conns1.c_str(), int(winner.input1.offs0), int(winner.input1.offs1),
                conns2.c_str(), int(winner.input2.offs0), int(winner.input2.offs1));
        return buf;
    }
    else if( winner.inputs_cnt()==4 )
    {
        std::string conns0 = vector_printer(winner.input0.connections);
        std::string conns1 = vector_printer(winner.input1.connections);
        std::string conns2 = vector_printer(winner.input2.connections);
        std::string conns3 = vector_printer(winner.input3.connections);
        sprintf(buf, "%14s*%d D4(g=%d,%d,%d,%d) = %s[%+d,%+d] & %s[%+d,%+d] & %s[%+d,%+d] & %s[%+d,%+d]",
                winner.input0.realm_name.c_str(),
                int(winner.gain),
                int(winner.input0.connections.size()), int(winner.input1.connections.size()), int(winner.input2.connections.size()), int(winner.input3.connections.size()),
                conns0.c_str(), int(winner.input0.offs0), int(winner.input0.offs1),
                conns1.c_str(), int(winner.input1.offs0), int(winner.input1.offs1),
                conns2.c_str(), int(winner.input2.offs0), int(winner.input2.offs1),
                conns3.c_str(), int(winner.input3.offs0), int(winner.input3.offs1));
        return buf;
    }
    else
        ASSERT(0);
    return "";
}

static std::string generate_report(const std::vector<conv_seed_detector> &d_vec, int gain)
{
    //auto vector_printer = [](std::vector<int> v)->std::string
    //    { std::string r =" ["; for(int a: v) r += std::to_string(a)+" "; r += "]"; return r; };

    std::vector<std::string> v_rep;
    char buf[8192];
    for(const conv_seed_detector &d: d_vec)
    {
        if( d.inputs_cnt()==1 )
        {
            sprintf(buf, "%s{c%d[0,0]}",
                    d.input0.realm_name.c_str(),
                    int(d.input0.plane_idx));
            v_rep.push_back(buf);
        }
        else if( d.inputs_cnt()==2 )
        {
            sprintf(buf, "%s{c%d[%+d,%+d]&c%d[%+d,%+d]}",
                    d.input0.realm_name.c_str(),
                    int(d.input0.plane_idx), int(d.input0.offs0), int(d.input0.offs1),
                    int(d.input1.plane_idx), int(d.input1.offs0), int(d.input1.offs1));
            v_rep.push_back(buf);
        }
        else if( d.inputs_cnt()==3 )
        {
            sprintf(buf, "%s{c%d[%+d,%+d]&c%d[%+d,%+d]&c%d[%+d,%+d]}",
                    d.input0.realm_name.c_str(),
                    int(d.input0.plane_idx), int(d.input0.offs0), int(d.input0.offs1),
                    int(d.input1.plane_idx), int(d.input1.offs0), int(d.input1.offs1),
                    int(d.input2.plane_idx), int(d.input2.offs0), int(d.input2.offs1));
            v_rep.push_back(buf);
        }
        else
            ASSERT(0);
    }

    std::string rep;
    for(int k=0; k<v_rep.size(); k++)
        rep += v_rep[k]+(k<v_rep.size()-1 ? " / " : "");
    return rep;
}

conv_generalizer_evaluator::conv_generalizer_evaluator(const activities &_a, const linear_evaluator &_le, const v2c_settings s):
    act_in(_a), lin_eval(_le), settings(s)
{
    max_queue_size = 256;
    queue_size = 0;
    queue_conns.resize(max_queue_size);
    queue_buffers.resize(max_queue_size);
    queue_results.resize(max_queue_size);
    queue_ratings.resize(max_queue_size);
    for(int k=0; k<max_queue_size; k++)
        queue_buffers[k].allocate(act_in.ntrn, act_in.ntst);
    best_rating = 0;
    best_detector = conv_detector();
}

void conv_generalizer_evaluator::start_evaluation(const conv_detector &d, int input_idx, and_type_t andtype, or_type_t ortype, bool use_weights)
{
    // integrity checks
    ASSERT(d.input0.connections.size()>0);
    ASSERT(d.input2.connections.size()==0 || d.input1.connections.size()>0);
    ASSERT(d.input3.connections.size()==0 || d.input2.connections.size()>0);
    ASSERT(d.input4.connections.size()==0 || d.input3.connections.size()>0);
    ASSERT(input_idx>=0 && input_idx<d.inputs_cnt());
    ASSERT(d.gain==gain_t::gain1);

    // clear
    queue_size = 0;

    // store evaluation settings
    and_type = andtype;
    or_type  = ortype;
    input_studied = input_idx;
    weights = use_weights ? std::vector<weight_type>{weight_type::w100,weight_type::w075,weight_type::w050}
                          : std::vector<weight_type>{weight_type::w100};

    // baseline
    ASSERT(act_in.is_i8());
    ASSERT(and_type==and_type_t::AND_MIN);
    ASSERT(or_type==or_type_t::OR_MAX);
    base_detector = d;
    merge_realm_name = d.get_input(input_idx).realm_name;
    d.generate_activity_except_for_input(act_in, input_idx, base_act_a);
    d.generate_activity_for_input(act_in, input_idx, base_act_g);
    ASSERT(base_act_a.is_i8() && base_act_g.is_i8());

    // results
    best_rating = 0;
    best_detector = conv_detector();
}

void conv_generalizer_evaluator::evaluate_generalization(int plane_idx)
{
    ASSERT(queue_size<=max_queue_size);
    for(const weight_type w: weights)
    {
        if( queue_size==max_queue_size )
            process_queue();
        queue_conns[queue_size] = conv_detector::conn(plane_idx,w);
        queue_size++;
    }
}

void conv_generalizer_process_queue_item_adapter(void *p, int queue_idx)
{
    conv_generalizer_evaluator &gen = *((conv_generalizer_evaluator*)p);
    const activities &act_in = gen.act_in;
    const activities &base_act_a = gen.base_act_a;
    const activities &base_act_g = gen.base_act_g;


    ASSERT(gen.and_type==and_type_t::AND_MIN);
    ASSERT(gen.or_type==or_type_t::OR_MAX);
    ASSERT(act_in.is_i8());
    ASSERT(base_act_a.realms.size()==1 && base_act_g.realms.size()==1);

    const realm<channel> &basea_realm = base_act_a.realms[0];
    const realm<channel> &baseg_realm = base_act_g.realms[0];
    const realm<channel> &merge_realm = act_in.realm_by_name(gen.merge_realm_name);
    const int64_t grid_height = merge_realm.grid_height, grid_width = merge_realm.grid_width;
    v2c_analysis &a_buf = gen.queue_buffers[queue_idx];
    const int64_t ntotal = a_buf.nchunks*a_buf.chunk_size;
    const int64_t chunk_size = 32;
    const int64_t nchunks = ntotal/chunk_size;
    const conv_detector::conn &conn = gen.queue_conns[queue_idx];
    const weight_type wtype = conn.weight;
    ASSERT(ntotal%chunk_size==0);
    ASSERT(wtype==weight_type::w100 || wtype==weight_type::w075 || wtype==weight_type::w050);
    ASSERT(gen.base_detector.gain==gain_t::gain1);
    a_buf.invalidate();
    int32_t *p_sum_i32 = (int32_t*)(&a_buf.trn_chunks[0]);
    for(int k0=0; k0<grid_height; k0++)
        for(int k1=0; k1<grid_width; k1++)
        {
            const int8_t *p_src_a = base_act_a.get_i8(basea_realm.get_cidx(0, k0, k1));
            const int8_t *p_src_g = base_act_g.get_i8(baseg_realm.get_cidx(0, k0, k1));
            const int8_t *p_src_m = act_in.get_i8(merge_realm.get_cidx(conn.plane_idx, k0, k1));
            const __m256i mm_mask7 = _mm256_set1_epi16(0x7F7Ful);
            const __m256i mm_mask6 = _mm256_set1_epi16(0x3F3Ful);
            alignas(32) int8_t arr8[32];
            for(int64_t i=0; i<nchunks; i++)
            {
                __m256i mm_a8 = _mm256_load_si256((__m256i*)(p_src_a+i*chunk_size));
                __m256i mm_g8 = _mm256_load_si256((__m256i*)(p_src_g+i*chunk_size));
                __m256i mm_m8 = _mm256_load_si256((__m256i*)(p_src_m+i*chunk_size));
                if( wtype==weight_type::w075 )
                {
                    __m256i mm_050 = _mm256_and_si256(_mm256_srli_epi16(mm_m8, 0x1), mm_mask7);
                    __m256i mm_025 = _mm256_and_si256(_mm256_srli_epi16(mm_m8, 0x2), mm_mask6);
                    mm_m8 = _mm256_add_epi8(mm_050,mm_025);
                }
                else if( wtype==weight_type::w050 )
                {
                    mm_m8 = _mm256_srli_epi16(mm_m8, 0x1);
                    mm_m8 = _mm256_and_si256(mm_m8, mm_mask7);
                }
                mm_g8 = _mm256_max_epi8(mm_g8, mm_m8);
                mm_a8 = _mm256_min_epi8(mm_a8, mm_g8);
                _mm256_store_si256((__m256i*)arr8, mm_a8);
                int32_t *p = p_sum_i32+i*32;
                for(int k=0; k<4; k++)
                {
                    __m256i mm_dst32 = _mm256_load_si256((__m256i*)(p+k*8));
                    __m256i mm_act32 = _mm256_cvtepi8_epi32(_mm_set1_epi64x(((const int64_t*)arr8)[k]));
                    mm_dst32 = _mm256_add_epi32(mm_dst32, mm_act32);
                    _mm256_store_si256((__m256i*)(p+k*8), mm_dst32);
                }
            }
        }
    float *p_sum_f32 = (float*)p_sum_i32;
    for(int64_t i=0; i<ntotal; i++)
        p_sum_f32[i] = SNET_I2F*p_sum_i32[i];
    a_buf.anrm2 = vnrm2(p_sum_f32,ntotal);
    a_buf.is_set = true;
    a_buf.compute_rating(gen.lin_eval, gen.settings.omega_rat);
    gen.queue_ratings[queue_idx] = a_buf.rating;
    gen.queue_results[queue_idx] = gen.base_detector;
    gen.queue_results[queue_idx].add_conn(gen.input_studied, conn.plane_idx, conn.weight);
}

void conv_generalizer_evaluator::process_queue()
{
    //
    // Process elements
    //
    parallel_for(conv_generalizer_process_queue_item_adapter, this, queue_size);

    //
    // Select winner, clear queue
    //
    for(int k=0; k<queue_size; k++)
        if( queue_ratings[k]>best_rating )
        {
            best_rating = queue_ratings[k];
            best_detector = queue_results[k];
        }
    queue_size = 0;
}

void conv_generalizer_evaluator::get_winner(float *rating, conv_detector *winner, activities *activity)
{
    if( queue_size>0 )
        process_queue();
    if( best_rating==0 )
    {
        *rating = 0.0f;
        *winner = conv_detector();
        *activity = activities();
        return;
    }
    *rating = best_rating;
    *winner = best_detector;
    winner->generate_activity_slow(act_in, *activity);
}

v2c_contest_x::entry_info::entry_info(const conv_seed_detector &d,
                                      const activities &act_in,
                                      const linear_evaluator &eval,
                                      float omega_rat)
{
    is_set = true;
    seed = d;
    seed.generate_activity_sum(act_in, act);
    anrm2 = vnrm2(&act[0],act_in.ntrn);
    rating = 0;
    compute_rating(eval, omega_rat);
}

void v2c_contest_x::entry_info::compute_rating(const linear_evaluator &eval, float omega_rat)
{
    // TODO: 2-class code!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    rating = 0.0f;
    if( !is_set )
        return;
    const int64_t ntrn = eval.get_ntrn();
    ASSERT(act.size()>=ntrn);
    for(int i=0; i<eval.get_nclasses(); i++)
    {
        const float *p_res = eval.get_res32()+i*eval.get_res_stride();
        const float *p_sum = &act[0];
        float cov = fdot_simd(p_res, p_sum, ntrn);
        rating += cov*cov/(omega_rat*ntrn+anrm2);
    }
}

v2c_contest_x::v2c_contest_x(const activities &ai,
              const linear_evaluator &le,
              v2c_settings s,
              std::atomic<int32_t> &rs,
              int max_search_rad):act_in(ai),lin_eval(le), running_seed(rs), settings(s), search_rad(max_search_rad)
{
    //
    // Precompute analysis for D1 detectors
    //
    for(const realm<channel> &rr: act_in.realms)
            for(int plane_idx=0; plane_idx<rr.grid_depth; plane_idx++)
            {
                contest_result r;
                r.is_set = true;
                r.seed = conv_seed_detector(rr, plane_idx);
                round0_seeds.push_back(r);
            }
}

void _v2c_parallel_contest_adapter(void *p, int idx)
{
    v2c_contest_x &contest = *((v2c_contest_x*)p);
    const conv_seed_detector &seed_det = contest.ctst_input_seeds[idx];
    v2c_contest_x::contest_result &c_res = contest.ctst_output_seeds[idx];
    ASSERT(!c_res.is_set);
    if( seed_det.inputs_cnt()==0 )
        return;
    const realm<channel> &rr = contest.act_in.realm_by_name(seed_det.input0.realm_name);

    //
    // Internal RNGs initialized by iteration-specific seeds
    //
    std::uniform_real_distribution<> uniform(0.0, 1.0);
    std::mt19937  mt_rnd;
    std::seed_seq mt_seed{(unsigned int)contest.running_seed, (unsigned int)idx};
    mt_rnd.seed(mt_seed);

    //
    // Reference activity for detector
    //
    activities seed_act;
    seed_det.generate_activity_slow(contest.act_in, seed_act);

    //
    // Determine padding required for detector seed_det
    //
    int64_t padding = contest.search_rad;
    for(int k=0; k<seed_det.inputs_cnt(); k++)
    {
        int max_k = imax(abs(seed_det.get_input(k).offs0), abs(seed_det.get_input(k).offs1));
        padding = imax(padding, max_k);
    }

    //
    // Run analysis
    //
    float w_fast_rating = 0;
    for(int plane_idx=0; plane_idx<rr.grid_depth; plane_idx++)
    {
        //
        // skip according to eval_prob
        //
        if( uniform(mt_rnd)>contest.settings.eval_prob)
            continue;

        //
        // Evaluate all possible offsets
        //
        const int64_t nclasses = contest.lin_eval.get_nclasses();
        const int64_t microchunk_size = 8;
        const int64_t n_microchunks = contest.act_in.ntrn/microchunk_size;
        //const int64_t ntotal = n_microchunks*microchunk_size;
        const int64_t nvariants = seed_det.inputs_cnt()*(2*contest.search_rad+1)*(2*contest.search_rad+1);
        const int64_t grid_height   = rr.grid_height;
        const int64_t grid_width    = rr.grid_width;
        const int64_t grid_stride1_a= rr.grid_stride1;
        const int64_t grid_stride1_b= seed_act.realms[0].grid_stride1;
        const int64_t padded_height = grid_height+2*padding;
        const int64_t padded_width  = grid_width+2*padding;
        activityf_vector   grd_a(microchunk_size*padded_height*padded_width, 0.0f);
        activityf_vector   grd_b(microchunk_size*padded_height*padded_width, 0.0f);
        activityf_vector v8_nrm2(microchunk_size*nvariants, 0.0f);
        activityf_vector  v8_cov(microchunk_size*nvariants*nclasses, 0.0f);
        const int8_t *p_a = contest.act_in.get_i8(rr.get_cidx(plane_idx,0,0));
        const int8_t *p_b = seed_act.get_i8();
        const int64_t stride_a = contest.act_in.get_stride();
        const int64_t stride_b = seed_act.get_stride();
        const float *p_res = contest.lin_eval.get_res32();
        const int64_t stride_res = contest.lin_eval.get_res_stride();
        int64_t nactive_elements = 0;
        for(int64_t mc_idx=0; mc_idx<n_microchunks; mc_idx++)
        {
            //
            // Check fold status, perform computations below only for active folds
            //
            int64_t page_idx = (mc_idx*microchunk_size)/contest.settings.page_size;
            if( contest.page_flags[page_idx]!=0 )
            {
                //
                // Update counter
                //
                nactive_elements += microchunk_size;

                //
                // Fetch activity chunks into grids A and B; each grid is padded with zero chunks.
                // Advance p_a and p_b.
                //
                __m256 mm_scl = _mm256_set1_ps(SNET_I2F);
                for(int i0=0; i0<grid_height; i0++)
                    for(int i1=0; i1<grid_width; i1++)
                    {
                        const int8_t *pca = p_a+(i0*grid_stride1_a+i1)*stride_a;
                        const int8_t *pcb = p_b+(i0*grid_stride1_b+i1)*stride_b;
                        __m256 mm_a = _mm256_cvtepi32_ps(_mm256_cvtepi8_epi32(_mm_set1_epi64x(*((const int64_t*)pca))));
                        __m256 mm_b = _mm256_cvtepi32_ps(_mm256_cvtepi8_epi32(_mm_set1_epi64x(*((const int64_t*)pcb))));
                        mm_a = _mm256_mul_ps(mm_a,mm_scl);
                        mm_b = _mm256_mul_ps(mm_b,mm_scl);
                        const int64_t dst_idx = (i0+padding)*padded_width+(i1+padding);
                        _mm256_store_ps(&grd_a[microchunk_size*dst_idx], mm_a);
                        _mm256_store_ps(&grd_b[microchunk_size*dst_idx], mm_b);
                    }

                //
                // Now, compute norm2 and cov2 for all variants of input offset and center.
                // Advance p_res
                //
                for(int vidx=0; vidx<nvariants; vidx++)
                {
                    // determine offsets of new plane and seed
                    int center_idx = vidx%seed_det.inputs_cnt();
                    int loc_idx    = vidx/seed_det.inputs_cnt();
                    int c0 = seed_det.get_input(center_idx).offs0;
                    int c1 = seed_det.get_input(center_idx).offs1;
                    int offs0 = loc_idx/(2*contest.search_rad+1)-contest.search_rad;
                    int offs1 = loc_idx%(2*contest.search_rad+1)-contest.search_rad;

                    // scan grid, compute sum of activities
                    __m256 mm_sumact = _mm256_set1_ps(0.0f);
                    for(int i0=0; i0<grid_height; i0++)
                    {
                        const float *p_chunk_a = &grd_a[microchunk_size*((i0+offs0+padding)*padded_width+(offs1+padding))];
                        const float *p_chunk_b = &grd_b[microchunk_size*((   i0+c0+padding)*padded_width+(   c1+padding))];
                        for(int i1=0; i1<grid_width; i1++, p_chunk_a+=microchunk_size, p_chunk_b+=microchunk_size)
                        {
                            __m256 mm_a = _mm256_loadu_ps(p_chunk_a);
                            __m256 mm_b = _mm256_loadu_ps(p_chunk_b);
                            mm_sumact = _mm256_add_ps(mm_sumact, _mm256_min_ps(mm_a,mm_b));
                        }
                    }

                    // update nrm2
                    float *p_nrm2_chunk = &v8_nrm2[microchunk_size*vidx];
                    __m256 mm_sumact2 = _mm256_load_ps(p_nrm2_chunk);
                    mm_sumact2 = _mm256_fmadd_ps(mm_sumact, mm_sumact, mm_sumact2);
                    _mm256_store_ps(p_nrm2_chunk, mm_sumact2);

                    // update cov
                    float *p_cov_chunk = &v8_cov[microchunk_size*nclasses*vidx];
                    for(int k=0; k<nclasses; k++)
                    {
                        __m256 mm_res = _mm256_load_ps(p_res      +k*stride_res);
                        __m256 mm_cov = _mm256_load_ps(p_cov_chunk+k*microchunk_size);
                        mm_cov = _mm256_fmadd_ps(mm_sumact, mm_res, mm_cov);
                        _mm256_store_ps(p_cov_chunk+k*microchunk_size, mm_cov);
                    }
                }
            }

            //
            // Advance pointers and counters
            //
            p_a += microchunk_size;
            p_b += microchunk_size;
            p_res += microchunk_size;
        }

        //
        // Compute ratings and squared norms
        //
        activityf_vector v_nrm2(nvariants, 0.0f), v_cov(nvariants*nclasses, 0.0f), v_cov2(nvariants, 0.0f);
        for(int k=0; k<nvariants*microchunk_size; k++)
            v_nrm2[k/microchunk_size] += v8_nrm2[k];
        for(int k=0; k<nvariants*nclasses*microchunk_size; k++)
            v_cov[k/microchunk_size] += v8_cov[k];
        for(int k=0; k<nvariants*nclasses; k++)
            v_cov2[k/nclasses] += v_cov[k]*v_cov[k];
        activityf_vector v_ratings(nvariants);
        for(int k=0; k<nvariants; k++)
            v_ratings[k] = compute_rating_omega(v_cov2[k], v_nrm2[k], contest.settings.omega_rat, nactive_elements);

        //
        // Save winner
        //
        int k_best = 0;
        for(int k=0; k<nvariants; k++)
            if( v_ratings[k]>v_ratings[k_best] )
                k_best = k;
        if( !c_res.is_set || v_ratings[k_best]>w_fast_rating )
        {
            int center_idx = k_best%seed_det.inputs_cnt();
            int loc_idx    = k_best/seed_det.inputs_cnt();
            int c0 = seed_det.get_input(center_idx).offs0;
            int c1 = seed_det.get_input(center_idx).offs1;
            int offs0 = loc_idx/(2*contest.search_rad+1)-contest.search_rad;
            int offs1 = loc_idx%(2*contest.search_rad+1)-contest.search_rad;
            c_res.seed = conv_seed_detector::compose(conv_seed_detector(rr, plane_idx), offs0, offs1, seed_det, -c0, -c1);
            c_res.is_set = true;
            w_fast_rating = v_ratings[k_best];
        }
    }
}

void _v2c_parallel_refresh_adapter(void *p, int idx)
{
    v2c_contest_x &c = *((v2c_contest_x*)p);
    ASSERT(idx>=0 && idx<c.entries.size());
    if( c.entries[idx].is_set )
        c.entries[idx].compute_rating(c.lin_eval, c.settings.omega_rat);
}

void _v2c_parallel_merge_adapter(void *p, int idx)
{
    v2c_contest_x &c = *((v2c_contest_x*)p);
    ASSERT(idx>=0 && idx<c.entries.size());

    //
    // If incoming result is not set, skip
    //
    const v2c_contest_x::contest_result &r = c.ctst_results[idx];
    if( !r.is_set)
        return;

    //
    // Refresh rating for already existing entry
    //
    v2c_contest_x::entry_info &e = c.entries[idx];
    if( e.is_set )
        e.compute_rating(c.lin_eval, c.settings.omega_rat);

    //
    // Try to insert new entry
    //
    v2c_contest_x::entry_info e_new(r.seed, c.act_in, c.lin_eval, c.settings.omega_rat);
    if( !e.is_set || e_new.rating>e.rating )
        e = e_new;
}

void v2c_contest_x::run_competition()
{
    //
    // Increase seed used to initialize RNGs
    //
    running_seed++;

    //
    // Prepare fold information
    //
    ASSERT(settings.page_prob>=0);
    std::mt19937  mt_rnd;
    std::seed_seq mt_seed{(unsigned int)running_seed};
    mt_rnd.seed(mt_seed);
    const int total_pages = iDivUp(act_in.ntrn,settings.page_size);
    const int fold0_pages = iDivUp(imin(settings.fold0_len,act_in.ntrn),settings.page_size);
    const int active_pages= imin(roundf((total_pages-fold0_pages)*settings.page_prob)+1,total_pages-fold0_pages);
    page_flags = std::vector<int8_t>(total_pages, 0x0);
    for(int k=0; k<fold0_pages+active_pages; k++)
        page_flags[k] = 0x1;
    std::shuffle(page_flags.begin()+fold0_pages, page_flags.end(), mt_rnd);

    //
    // First, prepare D1 detectors
    //
    ASSERT(settings.dmax>=1);
    ctst_results = round0_seeds;

    //
    // Then, run D2 contest
    //
    if( settings.dmax>=2 )
    {
        ctst_input_seeds.clear();
        for(const contest_result &r: round0_seeds)
            ctst_input_seeds.push_back(r.is_set ? r.seed : conv_seed_detector());
        ctst_output_seeds.clear();
        for(const conv_seed_detector &s: ctst_input_seeds)
            ctst_output_seeds.push_back(contest_result());
        parallel_for(_v2c_parallel_contest_adapter, this, ctst_input_seeds.size());
        for(const contest_result &r: ctst_output_seeds)
            ctst_results.push_back(r);
    }

    //
    // Wow!!! D3 contest!!! Yeeeahhh!
    //
    if( settings.dmax>=3 )
    {
        ctst_input_seeds.clear();
        for(const contest_result &r: ctst_output_seeds)
            ctst_input_seeds.push_back(r.is_set ? r.seed : conv_seed_detector());
        ctst_output_seeds.clear();
        for(const conv_seed_detector &s: ctst_input_seeds)
            ctst_output_seeds.push_back(contest_result());
        parallel_for(_v2c_parallel_contest_adapter, this, ctst_input_seeds.size());
        for(const contest_result &r: ctst_output_seeds)
            ctst_results.push_back(r);
    }

    //
    // We won't stop! D4 contest!
    //
    if( settings.dmax>=4 )
    {
        ctst_input_seeds.clear();
        for(const contest_result &r: ctst_output_seeds)
            ctst_input_seeds.push_back(r.is_set ? r.seed : conv_seed_detector());
        ctst_output_seeds.clear();
        for(const conv_seed_detector &s: ctst_input_seeds)
            ctst_output_seeds.push_back(contest_result());
        parallel_for(_v2c_parallel_contest_adapter, this, ctst_input_seeds.size());
        for(const contest_result &r: ctst_output_seeds)
            ctst_results.push_back(r);
    }

    //
    // Compete with already existing entries
    //
    if( entries.size()==0 )
        entries.resize(ctst_results.size());
    ASSERT(entries.size()==ctst_results.size());
    parallel_for(_v2c_parallel_merge_adapter, this, entries.size());
}

conv_seed_detector v2c_contest_x::retrieve_winner()
{
    int best_idx = -1;
    for(int idx=0; idx<entries.size(); idx++)
        if( entries[idx].is_set && (best_idx<0 || entries[idx].rating>entries[best_idx].rating) )
            best_idx = idx;
    if( best_idx<0 )
        return conv_seed_detector();
    entries[best_idx].is_set = false;
    return entries[best_idx].seed;
}

void v2c_contest_x::append_and_refresh_winners(linear_evaluator *eval, const activityf_vector &sum_act)
{
    ASSERT(eval==&lin_eval);
    eval->append_activity(sum_act);
    parallel_for(_v2c_parallel_refresh_adapter, this, entries.size());
}

void v2c_layer_yy(const predictions baseline,
               const activities &act_in,
               v2c_settings settings,
               int max_search_rad,
               int max_detectors,
               conv_midlayer &ml_out,
               activities *p_act_out,
               predictions &pred_out)
{
    int t0 = dtickcount();
    int t_cmpt = 0, t_retr = 0, t_gnrl = 0, t_appr = 0;

    //
    // Compute initial value for running seed using improvised hash function
    //
    std::mt19937 seed_gen;
    seed_gen.seed(settings.rng_seed);
    int32_t local_seed = int32_t(seed_gen());
    std::atomic<int32_t> running_seed(local_seed);

    //
    // Integrity checks
    //
    ASSERT(act_in.is_i8());
    for(const realm<channel> &rr: act_in.realms)
    {
        //ASSERT(rr.grid_height>=3 && rr.grid_width>=3); // to make sure that search rad is at least 1
        //ASSERT(rr.grid_height>=4 && rr.grid_width>=4); // to make sure that we can apply A16
        //ASSERT(rr.grid_height%4==0 && rr.grid_width%4==0); // to make sure that we can apply A16
        //ASSERT(rr.grid_height%downsample_out==0 && rr.grid_width%downsample_out==0); // to make sure that we can downsample
    }

    //
    // Prepare temporary storage for output detectors/activities
    //
    std::map<std::string,activities> map_actout;
    std::map<std::string,conv_midlayer> map_mlout;
    for(const realm<channel> &rr : act_in.realms)
    {
        ASSERT(map_actout.count(rr.name)==0);
        map_actout[rr.name] = activities(act_in.ntrn, act_in.ntst, storage_format::f32);
        map_actout[rr.name].append_realm(realm<channel>::create_grid(rr.name,1,1,0));
        ASSERT(map_mlout.count(rr.name)==0);
        map_mlout[rr.name] = conv_midlayer();
        map_mlout[rr.name].append_realm(realm<conv_detector>::create_grid(rr.name,1,1,0));
    }

    //
    // Round #1 - generate uncompactified detectors
    //
    linear_evaluator local_eval(baseline, max_detectors, sqrtf(settings.lambda_reg*act_in.ntrn));
    {
        v2c_contest_x    contest_x(act_in, local_eval, settings, running_seed, max_search_rad);
        conv_generalizer_evaluator gen_eval(act_in, local_eval, settings);
        for(int detectors_done=0; detectors_done<max_detectors;)
        {
            t_cmpt -= dtickcount();
            contest_x.run_competition();
            t_cmpt += dtickcount();
            for(int bidx=0; bidx<settings.batch_size && detectors_done<max_detectors; bidx++)
            {
                // retrieve winner
                t_retr -= dtickcount();
                const conv_seed_detector winner = contest_x.retrieve_winner();
                t_retr += dtickcount();
                if( winner.inputs_cnt()==0 )
                    continue;
                for(int k=0; k<winner.inputs_cnt(); k++)
                    ASSERT(winner.get_input(k).realm_name==winner.input0.realm_name);
                const std::string realm_name = winner.input0.realm_name;

                // generalization (none as for now)
                t_gnrl -= dtickcount();
                conv_detector gen;
                activities gen_act;
                activityf_vector sum_act;
                //winner.generate_activity_slow(act_in, winner_act);
                generalize(act_in, local_eval, gen_eval, settings, winner, &gen, &gen_act);
                t_gnrl += dtickcount();
                generate_activity_sum(gen_act, sum_act);

                // add
                if( p_act_out!=NULL )
                {
                    ASSERT(map_actout.count(realm_name)==1);
                    //map_actout[realm_name].append_plane1(gen_act, downsample_out);
                    map_actout[realm_name].append1(sum_act, "");
                }
                ASSERT(map_mlout.count(realm_name)==1);
                map_mlout[realm_name].append1(gen);
                t_appr -= dtickcount();
                contest_x.append_and_refresh_winners(&local_eval, sum_act);
                //std::string rep = generate_report(std::vector<conv_seed_detector>{winner}, 1);
                std::string rep = generate_report(gen);
                t_appr += dtickcount();

                printf("%s[V2:%04d]  TrnRms=%.4f  TstRms=%.4f  TstCls=%.4f  %s\n",
                       settings.rep_prefix.c_str(),
                       int(detectors_done),
                       (double)local_eval.trn_rms(), (double)local_eval.tst_rms(), (double)local_eval.tst_cls(),
                       rep.c_str());
                fflush(stdout);
                detectors_done++;
            }
        }
    }

    //
    // Combine results
    //
    ml_out = conv_midlayer();
    for(auto const &iter: map_mlout)
        ml_out.merge_in(iter.second);
    if( p_act_out!=NULL )
    {
        *p_act_out = activities(act_in.ntrn, act_in.ntst, storage_format::f32);
        for(auto const &iter: map_actout)
            p_act_out->merge_in(iter.second);
    }
    pred_out = local_eval.outputs();

    //
    // Report
    //
    double worst_l1_expansion = 0;
    for(int class_idx=0; class_idx<baseline.nclasses; class_idx++)
    {
        double l1_expansion = 0;
        for(int k=0; k<local_eval.model().get_nchannels(); k++)
            l1_expansion += fabs(local_eval.model().get_coeff(k,class_idx));
        worst_l1_expansion = fmax(worst_l1_expansion, l1_expansion);
    }
    printf("L1-expansion = %.2f\n", worst_l1_expansion);
    fflush(stdout);
    t0 = dtickcount()-t0;
    printf("V2c is done in %d ms (compete=%d, retr=%d, gen=%d, append/refresh=%d)\n", t0, t_cmpt, t_retr, t_gnrl, t_appr);
    fflush(stdout);
}


void v2c_layer_yy(const predictions baseline,
               const activities &act_in,
               v2c_settings settings,
               int max_search_rad,
               int max_detectors,
               conv_midlayer &ml_out,
               activities &act_out,
               predictions &pred_out)
{
    v2c_layer_yy(baseline, act_in, settings, max_search_rad, max_detectors, ml_out, &act_out, pred_out);
}

void v2c_layer_yy(const predictions baseline,
               const activities &act_in,
               v2c_settings settings,
               int max_search_rad,
               int max_detectors,
               conv_midlayer &ml_out,
               predictions &pred_out)
{
    v2c_layer_yy(baseline, act_in, settings, max_search_rad, max_detectors, ml_out, NULL, pred_out);
}

void v2c_layer_zz_ds2(const predictions baseline,
               const activities &act_in,
               v2c_settings settings,
               int max_search_rad,
               int max_detectors,
               activities &act_out_sum,
               activities &act_out_ds2,
               predictions &pred_out)
{
    int t0 = dtickcount();
    int t_cmpt = 0, t_retr = 0, t_gnrl = 0, t_appr = 0;

    //
    // Compute initial value for running seed using improvised hash function
    //
    std::mt19937 seed_gen;
    seed_gen.seed(settings.rng_seed);
    int32_t local_seed = int32_t(seed_gen());
    std::atomic<int32_t> running_seed(local_seed);

    //
    // Integrity checks
    //
    ASSERT(act_in.is_i8());
    for(const realm<channel> &rr: act_in.realms)
    {
        //ASSERT(rr.grid_height>=3 && rr.grid_width>=3); // to make sure that search rad is at least 1
        //ASSERT(rr.grid_height>=4 && rr.grid_width>=4); // to make sure that we can apply A16
        //ASSERT(rr.grid_height%4==0 && rr.grid_width%4==0); // to make sure that we can apply A16
        //ASSERT(rr.grid_height%downsample_out==0 && rr.grid_width%downsample_out==0); // to make sure that we can downsample
    }

    //
    // Prepare temporary storage for output detectors/activities
    //
    std::map<std::string,activities> map_actout_sum;
    std::map<std::string,activities> map_actout_ds2;
    //std::map<std::string,conv_midlayer> map_mlout;
    for(const realm<channel> &rr : act_in.realms)
    {
        ASSERT(map_actout_sum.count(rr.name)==0);
        map_actout_sum[rr.name] = activities(act_in.ntrn, act_in.ntst, storage_format::f32);
        map_actout_sum[rr.name].append_realm(realm<channel>::create_grid(rr.name,1,1,0));

        ASSERT(map_actout_ds2.count(rr.name)==0);
        map_actout_ds2[rr.name] = activities(act_in.ntrn, act_in.ntst, storage_format::i8);
        map_actout_ds2[rr.name].append_realm(realm<channel>::create_grid(rr.name,iDivUp(rr.grid_height,2),iDivUp(rr.grid_width,2),0));

        //ASSERT(map_mlout.count(rr.name)==0);
        //map_mlout[rr.name] = conv_midlayer();
        //map_mlout[rr.name].append_realm(realm<conv_detector>::create_grid(rr.name,1,1,0));
    }

    //
    // Round #1 - generate uncompactified detectors
    //
    linear_evaluator local_eval(baseline, max_detectors, sqrtf(settings.lambda_reg*act_in.ntrn));
    {
        v2c_contest_x    contest_x(act_in, local_eval, settings, running_seed, max_search_rad);
        conv_generalizer_evaluator gen_eval(act_in, local_eval, settings);
        for(int detectors_done=0; detectors_done<max_detectors;)
        {
            t_cmpt -= dtickcount();
            contest_x.run_competition();
            t_cmpt += dtickcount();
            for(int bidx=0; bidx<settings.batch_size && detectors_done<max_detectors; bidx++)
            {
                // retrieve winner
                t_retr -= dtickcount();
                const conv_seed_detector winner = contest_x.retrieve_winner();
                t_retr += dtickcount();
                if( winner.inputs_cnt()==0 )
                    continue;
                for(int k=0; k<winner.inputs_cnt(); k++)
                    ASSERT(winner.get_input(k).realm_name==winner.input0.realm_name);
                const std::string realm_name = winner.input0.realm_name;

                // generalization (none as for now)
                t_gnrl -= dtickcount();
                conv_detector gen;
                activities gen_act;
                activityf_vector sum_act;
                //winner.generate_activity_slow(act_in, winner_act);
                generalize(act_in, local_eval, gen_eval, settings, winner, &gen, &gen_act);
                t_gnrl += dtickcount();
                generate_activity_sum(gen_act, sum_act);

                // add
                ASSERT(map_actout_sum.count(realm_name)==1);
                ASSERT(map_actout_ds2.count(realm_name)==1);
                map_actout_sum[realm_name].append1(sum_act, "");
                map_actout_ds2[realm_name].append_plane1(gen_act, 2);
                //ASSERT(map_mlout.count(realm_name)==1);
                //map_mlout[realm_name].append1(gen);
                t_appr -= dtickcount();
                contest_x.append_and_refresh_winners(&local_eval, sum_act);
                //std::string rep = generate_report(std::vector<conv_seed_detector>{winner}, 1);
                std::string rep = generate_report(gen);
                t_appr += dtickcount();

                printf("%s[V2:%04d]  TrnRms=%.4f  TstRms=%.4f  TstCls=%.4f  %s\n",
                       settings.rep_prefix.c_str(),
                       int(detectors_done),
                       (double)local_eval.trn_rms(), (double)local_eval.tst_rms(), (double)local_eval.tst_cls(),
                       rep.c_str());
                fflush(stdout);
                detectors_done++;
            }
        }
    }

    //
    // Combine results
    //
    //ml_out = conv_midlayer();
    //for(auto const &iter: map_mlout)
    //    ml_out.merge_in(iter.second);
    act_out_sum = activities(act_in.ntrn, act_in.ntst, storage_format::f32);
    for(auto const &iter: map_actout_sum)
        act_out_sum.merge_in(iter.second);
    act_out_ds2 = activities(act_in.ntrn, act_in.ntst, storage_format::i8);
    for(auto const &iter: map_actout_ds2)
        act_out_ds2.merge_in(iter.second);
    pred_out = local_eval.outputs();

    //
    // Report
    //
    double worst_l1_expansion = 0;
    for(int class_idx=0; class_idx<baseline.nclasses; class_idx++)
    {
        double l1_expansion = 0;
        for(int k=0; k<local_eval.model().get_nchannels(); k++)
            l1_expansion += fabs(local_eval.model().get_coeff(k,class_idx));
        worst_l1_expansion = fmax(worst_l1_expansion, l1_expansion);
    }
    printf("L1-expansion = %.2f\n", worst_l1_expansion);
    fflush(stdout);
    t0 = dtickcount()-t0;
    printf("V2c is done in %d ms (compete=%d, retr=%d, gen=%d, append/refresh=%d)\n", t0, t_cmpt, t_retr, t_gnrl, t_appr);
    fflush(stdout);
}

void v2c_layer_zz_ds(const predictions baseline,
                     activities &act_raw,
                     v2c_settings settings,
                     int max_search_rad,
                     int max_planes,
                     split_t split,
                     int ds_out,
                     activities &act_out_ds)
{
    int t0 = dtickcount();
    int t_cmpt = 0, t_retr = 0, t_gnrl = 0, t_appr = 0;

    //
    // Compute initial value for running seed using improvised hash function
    //
    std::mt19937 seed_gen;
    seed_gen.seed(settings.rng_seed);
    int32_t local_seed = int32_t(seed_gen());
    std::atomic<int32_t> running_seed(local_seed);

    //
    // Integrity checks
    //
    ASSERT(act_raw.is_i8());
    ASSERT(act_raw.realms.size()>=1);
    const int grid_height = act_raw.realms[0].grid_height;
    const int grid_width  = act_raw.realms[0].grid_width;
    const int out_height = grid_height/ds_out;
    const int out_width  = grid_width/ds_out;
    for(const realm<channel> &rr: act_raw.realms)
    {
        ASSERT(rr.grid_height%ds_out==0 && rr.grid_width%ds_out==0); // to make sure that we can downsample
        ASSERT(rr.grid_height==grid_height && rr.grid_width==grid_width); // all realms have same size
    }

    //
    // Prepare temporary storage for output detectors/activities
    //
    std::map<std::string,activities> map_actout;
    //std::map<std::string,conv_midlayer> map_mlout;
    for(const realm<channel> &rr : act_raw.realms)
    {
        ASSERT(map_actout.count(rr.name)==0);
        map_actout[rr.name] = activities(act_raw.ntrn, act_raw.ntst, storage_format::i8);
        map_actout[rr.name].append_realm(realm<channel>::create_grid(rr.name,out_height,out_width,0));
        //ASSERT(map_mlout.count(rr.name)==0);
        //map_mlout[rr.name] = conv_midlayer();
        //map_mlout[rr.name].append_realm(realm<conv_detector>::create_grid(rr.name,1,1,0));
    }

    //
    // Work subsplit of activities
    //
    std::map<std::string,std::string> realm_names_map;
    activities act_wrk;
    act_raw.create_view(act_wrk);
    act_wrk.subdivide(split, realm_names_map, max_search_rad);

    //
    // Generate planes
    //
    linear_evaluator            local_eval(baseline, max_planes*out_height*out_width, sqrtf(settings.lambda_reg*act_raw.ntrn));
    v2c_contest_x               contest_x(act_wrk, local_eval, settings, running_seed, max_search_rad);
    conv_generalizer_evaluator  gen_eval(act_wrk, local_eval, settings);
    for(int planes_done=0; planes_done<max_planes;)
    {
        t_cmpt -= dtickcount();
        contest_x.run_competition();
        t_cmpt += dtickcount();
        for(int bidx=0; bidx<settings.batch_size && planes_done<max_planes; bidx++)
        {
            // retrieve winner
            t_retr -= dtickcount();
            conv_seed_detector winner = contest_x.retrieve_winner();
            t_retr += dtickcount();
            if( winner.inputs_cnt()==0 )
                continue;
            for(int k=0; k<winner.inputs_cnt(); k++)
                ASSERT(winner.get_input(k).realm_name==winner.input0.realm_name);

            // generalization
            t_gnrl -= dtickcount();
            conv_detector gen;
            activities tmp_act;
            generalize(act_wrk, local_eval, gen_eval, settings, winner, &gen, &tmp_act);
            t_gnrl += dtickcount();

            // generate activities over original realm, add
            t_appr -= dtickcount();
            activities det_act;
            for(int k=0; k<gen.inputs_cnt(); k++)
                gen.get_input(k).realm_name = realm_names_map.at(gen.get_input(k).realm_name);
            gen.generate_activity_slow(act_raw, det_act, ds_out);
            ASSERT(det_act.realms.size()==1);
            ASSERT(det_act.realms[0].grid_depth==1);
            ASSERT(det_act.realms[0].grid_height==out_height);
            ASSERT(det_act.realms[0].grid_width==out_width);
            map_actout.at(det_act.realms[0].name).append_plane1(det_act);
            for(int i=0; i<out_height; i++)
                for(int j=0; j<out_width; j++)
                {
                    int8_t *p = det_act.get_i8(det_act.realms[0].get_cidx(0,i,j));
                    activityf_vector act_ij(det_act.ntrn+det_act.ntst);
                    for(int64_t k=0; k<det_act.ntrn+det_act.ntst; k++)
                        act_ij[k] = SNET_I2F*p[k];
                    contest_x.append_and_refresh_winners(&local_eval, act_ij);
                }
            std::string rep = generate_report(gen);
            t_appr += dtickcount();

            printf("%s[V2:%04d]  TrnRms=%.4f  TstRms=%.4f  TstCls=%.4f  %s\n",
                   settings.rep_prefix.c_str(),
                   int(planes_done),
                   (double)local_eval.trn_rms(), (double)local_eval.tst_rms(), (double)local_eval.tst_cls(),
                   rep.c_str());
            fflush(stdout);
            planes_done++;
        }
    }

    //
    // Combine results
    //
    //ml_out = conv_midlayer();
    //for(auto const &iter: map_mlout)
    //    ml_out.merge_in(iter.second);
    act_out_ds = activities(act_raw.ntrn, act_raw.ntst, storage_format::i8);
    for(auto const &iter: map_actout)
        act_out_ds.merge_in(iter.second);
    //pred_out = local_eval.outputs();

    //
    // Report
    //
    t0 = dtickcount()-t0;
    printf("V2c is done in %d ms (compete=%d, retr=%d, gen=%d, append/refresh=%d)\n", t0, t_cmpt, t_retr, t_gnrl, t_appr);
    fflush(stdout);
}

void v2c_layer_zz_splitacc(linear_evaluator &eval,
                           const activities &act_in,
                           v2c_settings settings,
                           int max_search_rad,
                           int max_planes,
                           split_t split,
                           accumulate_t accumulate,
                           conv_midlayer &ml_out,
                           activities &act_out_sa,
                           predictions &pred_out)
{
    int t0 = dtickcount();
    int t_cmpt = 0, t_retr = 0, t_gnrl = 0, t_appr = 0;

    //
    // Compute initial value for running seed using improvised hash function
    //
    std::mt19937 seed_gen;
    seed_gen.seed(settings.rng_seed);
    int32_t local_seed = int32_t(seed_gen());
    std::atomic<int32_t> running_seed(local_seed);

    //
    // Integrity checks
    //
    ASSERT(act_in.is_i8());
    for(const realm<channel> &rr: act_in.realms)
    {
        //ASSERT(rr.grid_height>=3 && rr.grid_width>=3); // to make sure that search rad is at least 1
        //ASSERT(rr.grid_height>=4 && rr.grid_width>=4); // to make sure that we can apply A16
        //ASSERT(rr.grid_height%4==0 && rr.grid_width%4==0); // to make sure that we can apply A16
        //ASSERT(rr.grid_height%downsample_out==0 && rr.grid_width%downsample_out==0); // to make sure that we can downsample
    }

    //
    // Prepare temporary storage for output detectors/activities
    //
    std::map<std::string,activities> map_actout;
    std::map<std::string,conv_midlayer> map_mlout;
    for(const realm<channel> &rr : act_in.realms)
    {
        //ASSERT(map_actout.count(rr.name)==0);
        //map_actout[rr.name] = activities(act_in.ntrn, act_in.ntst, storage_format::f32);
        //map_actout[rr.name].append_realm(realm<channel>::create_grid(rr.name,1,1,0));
        ASSERT(map_mlout.count(rr.name)==0);
        map_mlout[rr.name] = conv_midlayer();
        map_mlout[rr.name].append_realm(realm<conv_detector>::create_grid(rr.name,1,1,0));
    }

    //
    // Round #1 - generate uncompactified detectors
    //
    {
        v2c_contest_x    contest_x(act_in, eval, settings, running_seed, max_search_rad);
        conv_generalizer_evaluator gen_eval(act_in, eval, settings);
        for(int planes_done=0; planes_done<max_planes;)
        {
            t_cmpt -= dtickcount();
            contest_x.run_competition();
            t_cmpt += dtickcount();
            for(int bidx=0; bidx<settings.batch_size && planes_done<max_planes; bidx++)
            {
                // retrieve winner
                t_retr -= dtickcount();
                const conv_seed_detector winner = contest_x.retrieve_winner();
                t_retr += dtickcount();
                if( winner.inputs_cnt()==0 )
                    continue;
                for(int k=0; k<winner.inputs_cnt(); k++)
                    ASSERT(winner.get_input(k).realm_name==winner.input0.realm_name);
                const std::string realm_name = winner.input0.realm_name;

                // generalization and gain
                t_gnrl -= dtickcount();
                conv_detector gen;
                activities gen_act;
                activityf_vector entire_sum_act;
                generalize(act_in, eval, gen_eval, settings, winner, &gen, &gen_act);
                t_gnrl += dtickcount();
                if( accumulate==accumulate_t::sum )
                {
                    //gen.gain = choose_best_gain(gen_act, {gain_t::gain1, gain_t::gain2, gain_t::gain4, gain_t::gain8}, eval, settings);
                    //gen.generate_activity_slow(act_in, gen_act);
                    generate_activity_sum(gen_act, entire_sum_act);
                }
                else if( accumulate==accumulate_t::max )
                {
                    //ASSERT(0); // no gain support
                    generate_activity_max(gen_act, entire_sum_act);
                }
                else
                    ASSERT(0);
                activities sa_act;
                gen_act.split_and_accumulate(split, accumulate, sa_act);

                // add
                for(const realm<channel> &rr: sa_act.realms)
                {
                    ASSERT(rr.grid_width==1 && rr.grid_height==1);
                    if( map_actout.count(rr.name)==0 )
                    {
                        map_actout[rr.name] = activities(act_in.ntrn, act_in.ntst, storage_format::f32);
                        map_actout[rr.name].append_realm(realm<channel>::create_grid(rr.name,1,1,0));
                    }
                    map_actout[rr.name].append_planes_from_realm(sa_act, rr.name);
                }
                ASSERT(map_mlout.count(realm_name)==1);
                map_mlout[realm_name].append1(gen);
                t_appr -= dtickcount();
                contest_x.append_and_refresh_winners(&eval, entire_sum_act);
                std::string rep = generate_report(gen);
                t_appr += dtickcount();

                printf("%s[V2:%04d]  TrnRms=%.4f  TstRms=%.4f  TstCls=%.4f  %s\n",
                       settings.rep_prefix.c_str(),
                       int(planes_done),
                       (double)eval.trn_rms(), (double)eval.tst_rms(), (double)eval.tst_cls(),
                       rep.c_str());
                fflush(stdout);
                planes_done++;
            }
        }
    }

    //
    // Combine results
    //
    ml_out = conv_midlayer();
    for(auto const &iter: map_mlout)
        ml_out.merge_in(iter.second);
    act_out_sa = activities(act_in.ntrn, act_in.ntst, storage_format::f32);
    for(auto const &iter: map_actout)
        act_out_sa.merge_in(iter.second);
    pred_out = eval.outputs();

    //
    // Report
    //
    /*double worst_l1_expansion = 0;
    for(int class_idx=0; class_idx<eval.get_nclasses(); class_idx++)
    {
        double l1_expansion = 0;
        for(int k=0; k<eval.model().get_nchannels(); k++)
            l1_expansion += fabs(local_eval.model().get_coeff(k,class_idx));
        worst_l1_expansion = fmax(worst_l1_expansion, l1_expansion);
    }
    printf("L1-expansion = %.2f\n", worst_l1_expansion);
    fflush(stdout);*/
    t0 = dtickcount()-t0;
    printf("V2c is done in %d ms (compete=%d, retr=%d, gen=%d, append/refresh=%d)\n", t0, t_cmpt, t_retr, t_gnrl, t_appr);
    fflush(stdout);
}

}
