/*************************************************************************
StrongNet/Contour Engine training framework.
Copyright (c) 2020 Sergey Bochkanov (ALGLIB project).

>>> SOURCE LICENSE >>>
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation (www.fsf.org); either version 2 of the 
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

A copy of the GNU General Public License is available at
http://www.fsf.org/licensing/licenses
>>> END OF LICENSE >>>
*************************************************************************/
#ifndef LAYER2C_H
#define LAYER2C_H

#include <atomic>
#include "snetcommon.h"

namespace strongnet
{

class v2c_settings
{
public:
    std::string rep_prefix;

    float lambda_reg;
    float omega_rat;
    int batch_size;
    bool use_gain;
    bool use_weights;
    float eval_prob;
    int rng_seed;
    int calibrator_bins;
    int dmax;
    float gen_rating_growth;

    // paged subdivision of dataset
    int page_size;
    float page_prob;
    int fold0_len;

    v2c_settings():
        lambda_reg(1.0e-5f),
        omega_rat(1.0e-5f),
        batch_size(5),
        use_weights(false),
        use_gain(true),
        eval_prob(0.1f),
        rng_seed(1),
        page_size(512),
        page_prob(1.0f),
        fold0_len(0),
        calibrator_bins(21),
        dmax(3),
        gen_rating_growth(0)
    {}
};
class conv_seed_detector
{
public:
    struct input
    {
        std::string realm_name;
        int realm_depth, realm_dim0, realm_dim1;
        int offs0, offs1;
        int plane_idx;

        input():realm_depth(0),realm_dim0(0),realm_dim1(0),offs0(0),offs1(0),plane_idx(-1){}
        //input(std::string rname, int p):realm_name(rname),offs0(0),offs1(0),plane_idx(p){}
        //input(std::string rname, int o0, int o1, int p):realm_name(rname),offs0(o0),offs1(o1),plane_idx(p){}
        input(const realm<channel> &rr, int p):
            realm_name(rr.name),realm_depth(rr.grid_depth),realm_dim0(rr.grid_height),realm_dim1(rr.grid_width),
            offs0(0),offs1(0),plane_idx(p){}
        input(const realm<channel> &rr, int o0, int o1, int p):
            realm_name(rr.name),realm_depth(rr.grid_depth),realm_dim0(rr.grid_height),realm_dim1(rr.grid_width),
            offs0(o0),offs1(o1),plane_idx(p){}


    };

    conv_seed_detector(){}
    conv_seed_detector(const realm<channel> &rr, int pi):input0(rr,pi){}

    input input0;
    input input1;
    input input2;
    input input3;
    input input4;

    // stats
    const input& get_input(int i) const;
    input& get_input(int i);
    int inputs_cnt() const;

    // generate composite detector using two input ones
    static conv_seed_detector compose(const conv_seed_detector &d0, int offs00, int offs01,
                                      const conv_seed_detector &d1, int offs10, int offs11);

    void generate_activity_slow(const activities &act_in, activities &act_out, int downsample = 1) const;
    void generate_activity_sum(const activities &act_in, activityf_vector &v_out) const;
};

class conv_detector
{
public:
    class conn
    {
    public:
        weight_type weight;
        int plane_idx;

        conn():weight(weight_type::w100),plane_idx(-1){}
        conn(int p):weight(weight_type::w100),plane_idx(p){}
        conn(int p, weight_type w):weight(w),plane_idx(p){}
    };

    class input
    {
    public:
        std::string realm_name;
        int realm_depth, realm_dim0, realm_dim1;
        int offs0, offs1;
        std::vector<conn> connections;

        input():realm_depth(0),realm_dim0(0),realm_dim1(0),offs0(0),offs1(0){}
        //input(std::string rname, int p):realm_name(rname),offs0(0),offs1(0),connections{p}{}
        //input(std::string rname, int o0, int o1, int p):realm_name(rname),offs0(o0),offs1(o1),connections{p}{}
        input(const realm<channel> &rr, int p):
            realm_name(rr.name),realm_depth(rr.grid_depth),realm_dim0(rr.grid_height),realm_dim1(rr.grid_width),
            offs0(0),offs1(0),connections{p}{}
        input(const realm<channel> &rr, int o0, int o1, int p):
            realm_name(rr.name),realm_depth(rr.grid_depth),realm_dim0(rr.grid_height),realm_dim1(rr.grid_width),
            offs0(o0),offs1(o1),connections{p}{}
        input(conv_seed_detector::input in):
            realm_name(in.realm_name), offs0(in.offs0), offs1(in.offs1),
            connections(in.plane_idx>=0 ? std::vector<conn>{in.plane_idx} : std::vector<conn>{}){}
    };

    conv_detector():gain(gain_t::gain1){}
    conv_detector(const conv_seed_detector &seed, gain_t g = gain_t::gain1);

    gain_t gain;
    input input0;
    input input1;
    input input2;
    input input3;
    input input4;
    std::string desc;

    // stats
    const input& get_input(int i) const;
    input& get_input(int i);
    int inputs_cnt() const;

    // add connection to plane
    void add_conn(int input_idx, int plane_idx, weight_type w);

    // generate activities
    void generate_activity_slow(const activities &act_in, activities &act_out, int downsample = 1) const;
    void generate_activity_sum(const activities &act_in, activityf_vector &v_out) const;

    // generate activities for AND(input0,...,inputk) except for input_idx-th one
    // if detector has just one input, 1.0 is returned as activity
    void generate_activity_except_for_input(const activities &act_in, int input_idx, activities &act_out, int downsample = 1) const;

    // generate activity for input_idx-th input
    void generate_activity_for_input(const activities &act_in, int input_idx, activities &act_out, int downsample = 1) const;
private:
    // expects aligned p_buf and p_out, both are int8_t[ntotal]
    void generate_activity_for_location(
            std::vector<int> inputs_to_process,
            const activities &act_bottom,
            const realm<channel> &rr_work,
            const int k0, const int k1,
            int8_t *p_buf,
            int8_t *p_out) const;
    // expects aligned p_buf and p_out, both are float[ntotal]
    void generate_activity_for_location(
            std::vector<int> inputs_to_process,
            const activities &act_bottom,
            const realm<channel> &rr_work,
            const int k0, const int k1,
            float *p_buff,
            float *p_outf) const;
    void generate_activity_x(const activities &act_in, std::vector<int> inputs, activities &act_out, int downsample) const;
};
void serialize(FILE *f, const conv_detector::conn *ptr);
void serialize(FILE *f, const conv_detector::input *ptr);
void serialize(FILE *f, const conv_detector *ptr);
void unserialize(FILE *f, conv_detector::conn *ptr);
void unserialize(FILE *f, conv_detector::input *ptr);
void unserialize(FILE *f, conv_detector *ptr);

class conv_midlayer
{
public:
    int64_t                             ndetectors;
    std::vector<realm<conv_detector>>   realms;

    conv_midlayer():ndetectors(0){}

    const realm<detector>& realm_by_detector(int cidx) const;

    // appends just one detector; layer MUST have just one realm with 1x1 grid
    void append1(const conv_detector &d);

    // appends (possibly non-empty) realm; layer may already have some realms, it will work
    void append_realm(const realm<conv_detector> &r);

    // merges entire layer into this one
    void merge_in(const conv_midlayer &merged);

    // applies convolutional layer, computes summary activity; produces f32 activities with 1x1 grid
    void apply_and_sum_f32(const activities &act_in, activities &act_out) const;

    // generic apply-and-postprocess
    void apply(const activities &act_in, split_t post_divide, accumulate_t accumulate, activities &act_out) const;
};
void serialize(FILE *f, const conv_midlayer *ptr);
void unserialize(FILE *f, conv_midlayer *ptr);

class v2c_analysis
{
public:
    static const int                CHUNK_SIZE;

    bool                            is_set;
    int64_t                         chunk_size, nchunks;

    std::string                     realm_name;
    conv_seed_detector              seed_det;
    activityf_vector                trn_chunks;
    float                           rating;
    float                           anrm2;

    v2c_analysis():is_set(false),chunk_size(CHUNK_SIZE){}

    void allocate(int trn, int tst);
    void invalidate();
    void analyze_plane(const activities &act_in, const realm<channel> &rr, int plane_idx, const linear_evaluator &lin_eval, float omega_rat);
    void analyze_planes_max(const activities &act0, const realm<channel> &rr0, int plane0,
                            const activities &act1, const realm<channel> &rr1, int plane1,
                            const linear_evaluator &lin_eval, float omega_rat);
    void compute_rating(const linear_evaluator &lin_eval, float omega_rat);
};

class v2c_realm
{
public:
    std::string                         realm_name;
    int                                 search_rad;
    int                                 search_area;

    //std::vector<v2c_candidate> candidate_entries;
    activities                          act_new;
    std::vector<conv_seed_detector>     seeds;

    // analysis results from the last contest
    std::vector<v2c_analysis>           entry_winners;
    void invalidate_analyses();
};

class conv_generalizer_evaluator
{
friend void conv_generalizer_process_queue_item_adapter(void *p, int queue_idx);
private:
    const v2c_settings          settings;
    const activities            &act_in;
    const linear_evaluator      &lin_eval;

    // evaluation queue
    class queue_entry
    {
    public:
        int merge_idx;
        float rating;
        queue_entry():merge_idx(-1),rating(0){}
        queue_entry(int mi):merge_idx(mi),rating(0){}
    };
    std::vector<queue_entry>    eval_queue;
    and_type_t                  and_type;
    or_type_t                   or_type;
    std::vector<weight_type>    weights;
    int                         max_queue_size;
    //activity8_vector            queue_cand_act_i8; // thread-safe buffer for process_queue_item()

    // detector info
    conv_detector               base_detector;
    activities                  base_act_a;
    activities                  base_act_g;
    int                         input_studied;
    std::string                 merge_realm_name;

    // evaluation queue
    int                         queue_size;
    std::vector<conv_detector::conn> queue_conns;
    std::vector<v2c_analysis>   queue_buffers;
    std::vector<conv_detector>  queue_results;
    std::vector<float>          queue_ratings;
    float                       best_rating;
    conv_detector               best_detector;
    void process_queue();
public:
    //static std::atomic<int32_t> pcnt_gen;
    conv_generalizer_evaluator(const activities &_a, const linear_evaluator &_le, const v2c_settings s);
    void start_evaluation(const conv_detector &d, int input_idx, and_type_t andtype, or_type_t ortype, bool use_weights);
    void evaluate_generalization(int plane_idx);
    void get_winner(float *rating, conv_detector *winner, activities *activity);
};

class v2c_contest_x
{
friend void _v2c_parallel_refresh_adapter(void *p, int idx);
friend void _v2c_parallel_contest_adapter(void *p, int idx);
friend void _v2c_parallel_merge_adapter(void *p, int idx);
public:
    v2c_contest_x(const activities &ai,
                  const linear_evaluator &le,
                  v2c_settings s,
                  std::atomic<int32_t> &rs,
                  int max_search_rad);

    void run_competition();

    // retrieves current winner and marks it as non-set
    conv_seed_detector retrieve_winner();

    // appends activity and performs quick update on the current winners
    void append_and_refresh_winners(linear_evaluator *eval, const activityf_vector &sum_act);
private:
    class contest_result
    {
    public:
        bool is_set;
        conv_seed_detector seed;

        contest_result():is_set(false){}
    };

    class entry_info
    {
    public:
        bool is_set;
        conv_seed_detector seed;
        activityf_vector act;
        float anrm2;
        float rating;

        entry_info():is_set(false),rating(0){}
        entry_info(const conv_seed_detector &d, const activities &act_in, const linear_evaluator &eval, float omega_rat);

        void compute_rating(const linear_evaluator &eval, float omega_rat);
    };

    const activities            &act_in;
    const linear_evaluator      &lin_eval;
    std::atomic<int32_t>        &running_seed;
    const v2c_settings          settings;
    const int                   search_rad;

    // winner entries
    std::vector<entry_info>     entries;

    // parallel D2/D3 contest
    std::vector<contest_result>     round0_seeds;
    std::vector<int8_t>             page_flags;
    std::vector<conv_seed_detector> ctst_input_seeds;
    std::vector<contest_result>     ctst_output_seeds;
    std::vector<contest_result>     ctst_results;
};

void v2c_layer_yy(const predictions baseline,
               const activities &act_in,
               v2c_settings settings,
               int max_search_rad,
               int max_detectors,
               conv_midlayer &ml_out,
               activities *p_act_out,
               predictions &pred_out);

void v2c_layer_yy(const predictions baseline,
               const activities &act_in,
               v2c_settings settings,
               int max_search_rad,
               int max_detectors,
               conv_midlayer &ml_out,
               activities &act_out,
               predictions &pred_out);

void v2c_layer_yy(const predictions baseline,
               const activities &act_in,
               v2c_settings settings,
               int max_search_rad,
               int max_detectors,
               conv_midlayer &ml_out,
               predictions &pred_out);

void v2c_layer_zz_ds2(const predictions baseline,
               const activities &act_in,
               v2c_settings settings,
               int max_search_rad,
               int max_detectors,
               activities &act_out_sum,
               activities &act_out_ds2,
               predictions &pred_out);

void v2c_layer_zz_ds(const predictions baseline,
                     activities &act_raw,
                     v2c_settings settings,
                     int max_search_rad,
                     int max_planes,
                     split_t split,
                     int ds_out,
                     activities &act_out_ds);

void v2c_layer_zz_splitacc(linear_evaluator &eval,
                           const activities &act_in,
                           v2c_settings settings,
                           int max_search_rad,
                           int max_planes,
                           split_t split,
                           accumulate_t accumulate,
                           conv_midlayer &ml_out,
                           activities &act_out_sa,
                           predictions &pred_out);
}

#endif
