/*************************************************************************
StrongNet/Contour Engine training framework.
Copyright (c) 2020 Sergey Bochkanov (ALGLIB project).

>>> SOURCE LICENSE >>>
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation (www.fsf.org); either version 2 of the 
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

A copy of the GNU General Public License is available at
http://www.fsf.org/licensing/licenses
>>> END OF LICENSE >>>
*************************************************************************/
#include "layer4.h"
#include <algorithm>
#include <atomic>
#include <tuple>
#include <set>
#include <random>

namespace strongnet
{

template<class activity_v_t>
void generalize(const activities &act_in,
                const linear_evaluator &lin_eval,
                generalizer_evaluator  &gen_eval,
                const v4_settings settings,
                const seed_detector &seed,
                and_type_t and_type, or_type_t or_type,
                detector *d_out, activity_v_t *act_out)
{
    ASSERT(seed.channel_idx0>=0);
    ASSERT(seed.inputs_cnt()>=1 && seed.inputs_cnt()<=5);

    //
    // Try to generalize
    //
    const int inputs_cnt = seed.inputs_cnt();
    const realm<channel> *p_realm0 = &(act_in.realm_by_channel(seed.channel_idx0));
    //const realm<channel> *p_realm1 = seed.channel_idx1>=0 ? &(act_in.realm_by_channel(seed.channel_idx1)) : NULL;
    //const realm<channel> *p_realm2 = seed.channel_idx2>=0 ? &(act_in.realm_by_channel(seed.channel_idx2)) : NULL;
    float                 gen_rating;
    detector              gen_det(seed,and_type,or_type,act_in);
    activity_v_t          gen_act;
    gen_det.generate_activity_slow(act_in, gen_act);
    gen_rating = compute_rating(settings.gen_rating, settings.omega_rat, gen_act, lin_eval);
    for(;;)
    {
        float            cand_cov;
        detector         cand;
        activity_v_t     cand_act;
        bool             something_changed = false;
        for(int input_idx=0; input_idx<inputs_cnt; input_idx++)
        {
            int input_sizes[5]       = {int(gen_det.input0.size()),
                                        int(gen_det.input1.size()),
                                        int(gen_det.input2.size()),
                                        int(gen_det.input3.size()),
                                        int(gen_det.input4.size())};
            int channel_indexes[5]   = {seed.channel_idx0,
                                        seed.channel_idx1,
                                        seed.channel_idx2,
                                        seed.channel_idx3,
                                        seed.channel_idx4};
            int input_size           = input_sizes[input_idx];
            int channel_idx          = channel_indexes[input_idx];
            ASSERT(input_size>=1);
            if( input_size<SNET_MAX_V4_GEN_COUNT )
            {
                gen_eval.start_evaluation(gen_det, input_idx, and_type, or_type);
                if( settings.restrict_gen_by_realm )
                {
                    for(int d=0; d<p_realm0->grid_depth; d++)
                        for(int i0=0; i0<p_realm0->grid_height; i0++)
                            for(int i1=0; i1<p_realm0->grid_width; i1++)
                                gen_eval.evaluate_generalization(p_realm0->get_cidx(d, i0, i1));
                }
                else
                {
                    for(int64_t channel_idx=0; channel_idx<act_in.nchannels; channel_idx++)
                        gen_eval.evaluate_generalization(channel_idx);
                }
                gen_eval.get_winner(&cand_cov, &cand, &cand_act);
                float cand_rating = compute_rating(settings.gen_rating, settings.omega_rat, cand_act, lin_eval);
                ASSERT(settings.gen_rating_growth>0);
                if( cand_rating>gen_rating*settings.gen_rating_growth )
                {
                    gen_det = cand;
                    gen_act = cand_act;
                    gen_rating = cand_rating;
                    something_changed = true;
                }
            }
        }
        if( !something_changed )
            break;
    }

    //
    // Done
    //
    *d_out   = gen_det;
    *act_out = gen_act;
}

template<class activity_t>
const int v4_contest<activity_t>::line_width = 5;

template<class activity_t>
void* v4_contest_run_parallel(void *ptr)
{
    v4_contest<activity_t>  &contest = *((v4_contest<activity_t>*)ptr);
    const activities        &act_in = contest.act_in;
    const linear_evaluator  &lin_eval = contest.lin_eval;
    v4_settings             settings = contest.settings;
    seed_evaluator          seed_eval(act_in,lin_eval);

    ASSERT(settings.dmax>=1 && settings.dmax<=5);
    ASSERT(settings.ncenters>=1 && settings.ncenters<=5);

    //
    // Select input seed to work with
    //
    activity_t d1_seed_act, d2_seed_act, d3_seed_act, d4_seed_act, d5_seed_act;
    for(int line_idx=contest.lines_examined.fetch_add(1); line_idx<act_in.nchannels; line_idx=contest.lines_examined.fetch_add(1))
    {
        typename v4_contest<activity_t>::winner &out_d1 = contest.winners[line_idx*v4_contest<activity_t>::line_width+0];
        typename v4_contest<activity_t>::winner &out_d2 = contest.winners[line_idx*v4_contest<activity_t>::line_width+1];
        typename v4_contest<activity_t>::winner &out_d3 = contest.winners[line_idx*v4_contest<activity_t>::line_width+2];
        typename v4_contest<activity_t>::winner &out_d4 = contest.winners[line_idx*v4_contest<activity_t>::line_width+3];
        typename v4_contest<activity_t>::winner &out_d5 = contest.winners[line_idx*v4_contest<activity_t>::line_width+4];
        std::vector<int> centers;

        //
        // Internal RNGs initialized by iteration-specific seeds
        //
        std::uniform_real_distribution<> uniform(0.0, 1.0);
        std::mt19937  mt_rnd;
        std::seed_seq mt_seed{(unsigned int)contest.rng_seeds_to_use[line_idx], (unsigned int)settings.rng_seed};
        mt_rnd.seed(mt_seed);

        //
        // Evaluate D1 detector
        //
#warning We do not need seed evaluator here because input seeds may have zero covariance due to orthogonalization
        if( settings.dmax<1 )
            continue;
        seed_detector d1_seed_det;
        float d1_seed_cov2;
        seed_eval.start_evaluation(contest.and_type);
        seed_eval.evaluate_seed(seed_detector(line_idx,contest.and_type));
        seed_eval.get_winner(&d1_seed_cov2, &d1_seed_det, &d1_seed_act);
        if( d1_seed_cov2<=0 )
            continue;
        out_d1.is_set = true;
        out_d1.seed = d1_seed_det;
        out_d1.winner_act = d1_seed_act;
        out_d1.anrm2  = vnrm2(&d1_seed_act[0], act_in.ntrn);
        out_d1.rating = compute_rating(settings.con_rating, settings.omega_rat, d1_seed_act, d1_seed_cov2, out_d1.anrm2, lin_eval);
        centers.push_back(d1_seed_det.channel_idx0);

        //
        // Evaluate D2 detector built upon D1 one
        //
        if( settings.dmax<2 )
            continue;
        float d2_seed_cov2;
        seed_detector d2_seed_det;
        seed_eval.start_evaluation(contest.and_type);
        for(int center_idx=0; center_idx<imin(centers.size(),settings.ncenters); center_idx++)
            for(int neighbor: contest.most_correlated_neighbors[centers[center_idx]])
                if( uniform(mt_rnd)<=settings.eval_prob )
                    seed_eval.evaluate_seed(seed_detector(d1_seed_det.channel_idx0,neighbor,contest.and_type));
        seed_eval.get_winner(&d2_seed_cov2, &d2_seed_det, &d2_seed_act);
        if( d2_seed_cov2<=0 )
            continue;
        out_d2.is_set = true;
        out_d2.seed = d2_seed_det;
        out_d2.winner_act = d2_seed_act;
        out_d2.anrm2  = vnrm2(&d2_seed_act[0], act_in.ntrn);
        out_d2.rating = compute_rating(settings.con_rating, settings.omega_rat, d2_seed_act, d2_seed_cov2, out_d2.anrm2, lin_eval);
        centers.push_back(d2_seed_det.channel_idx1);

        //
        // Evaluate D3 detector built upon D2 one
        //
        if( settings.dmax<3 )
            continue;
        float d3_seed_cov2;
        seed_detector d3_seed_det;
        seed_eval.start_evaluation(contest.and_type);
        for(int center_idx=0; center_idx<imin(centers.size(),settings.ncenters); center_idx++)
            for(int neighbor: contest.most_correlated_neighbors[centers[center_idx]])
                if( uniform(mt_rnd)<=settings.eval_prob )
                    seed_eval.evaluate_seed(seed_detector(d2_seed_det.channel_idx0,
                                                          d2_seed_det.channel_idx1,
                                                          neighbor,
                                                          contest.and_type));
        seed_eval.get_winner(&d3_seed_cov2, &d3_seed_det, &d3_seed_act);
        if( d3_seed_cov2<=0 )
            continue;
        out_d3.is_set = true;
        out_d3.seed = d3_seed_det;
        out_d3.winner_act = d3_seed_act;
        out_d3.anrm2  = vnrm2(&d3_seed_act[0], act_in.ntrn);
        out_d3.rating = compute_rating(settings.con_rating, settings.omega_rat, d3_seed_act, d3_seed_cov2, out_d3.anrm2, lin_eval);
        centers.push_back(d3_seed_det.channel_idx2);

        //
        // Evaluate D4 detector built upon D3 one
        //
        if( settings.dmax<4 )
            continue;
        float d4_seed_cov2;
        seed_detector d4_seed_det;
        seed_eval.start_evaluation(contest.and_type);
        for(int center_idx=0; center_idx<imin(centers.size(),settings.ncenters); center_idx++)
            for(int neighbor: contest.most_correlated_neighbors[centers[center_idx]])
                if( uniform(mt_rnd)<=settings.eval_prob )
                    seed_eval.evaluate_seed(seed_detector(d3_seed_det.channel_idx0,
                                                          d3_seed_det.channel_idx1,
                                                          d3_seed_det.channel_idx2,
                                                          neighbor,
                                                          contest.and_type));
        seed_eval.get_winner(&d4_seed_cov2, &d4_seed_det, &d4_seed_act);
        if( d4_seed_cov2<=0 )
            continue;
        out_d4.is_set = true;
        out_d4.seed = d4_seed_det;
        out_d4.winner_act = d4_seed_act;
        out_d4.anrm2  = vnrm2(&d4_seed_act[0], act_in.ntrn);
        out_d4.rating = compute_rating(settings.con_rating, settings.omega_rat, d4_seed_act, d4_seed_cov2, out_d4.anrm2, lin_eval);
        centers.push_back(d4_seed_det.channel_idx3);

        //
        // Evaluate D5 detector built upon D4 one
        //
        if( settings.dmax<5 )
            continue;
        float d5_seed_cov2;
        seed_detector d5_seed_det;
        seed_eval.start_evaluation(contest.and_type);
        for(int center_idx=0; center_idx<imin(centers.size(),settings.ncenters); center_idx++)
            for(int neighbor: contest.most_correlated_neighbors[centers[center_idx]])
                if( uniform(mt_rnd)<=settings.eval_prob )
                    seed_eval.evaluate_seed(seed_detector(d4_seed_det.channel_idx0,
                                                          d4_seed_det.channel_idx1,
                                                          d4_seed_det.channel_idx2,
                                                          d4_seed_det.channel_idx3,
                                                          neighbor,
                                                          contest.and_type));
        seed_eval.get_winner(&d5_seed_cov2, &d5_seed_det, &d5_seed_act);
        if( d5_seed_cov2<=0 )
            continue;
        out_d5.is_set = true;
        out_d5.seed = d5_seed_det;
        out_d5.winner_act = d5_seed_act;
        out_d5.anrm2  = vnrm2(&d5_seed_act[0], act_in.ntrn);
        out_d5.rating = compute_rating(settings.con_rating, settings.omega_rat, d5_seed_act, d5_seed_cov2, out_d5.anrm2, lin_eval);
        centers.push_back(d5_seed_det.channel_idx4);
    }
    return NULL;
}

template<class activity_t>
v4_contest<activity_t>::v4_contest(const activities &ai,
               const linear_evaluator &le,
               const std::vector<std::vector<int> > &mcn,
               std::atomic<int32_t> &rs,
               v4_settings s, and_type_t at, or_type_t ot):act_in(ai),lin_eval(le),most_correlated_neighbors(mcn),running_rng_seed(rs),
    settings(s),and_type(at),or_type(ot)
{
    winners          = std::vector<winner>(line_width*ai.nchannels, winner());
    rng_seeds_to_use = std::vector<int>(ai.nchannels, 0);
}

template<class activity_t>
void v4_contest<activity_t>::run_competition()
{
    lines_examined = 0;
    for(int i=0; i<act_in.nchannels; i++)
        rng_seeds_to_use[i] = running_rng_seed.fetch_add(1);
    for(winner &w: winners)
        w.invalidate();
    parallel_invoke(v4_contest_run_parallel<activity_t>, this);
}

template<class activity_t>
typename v4_contest<activity_t>::winner* v4_contest<activity_t>::retrieve_winner()
{
    if( settings.con_rating==rating_t::cov2 ||
        settings.con_rating==rating_t::deltar2 ||
        settings.con_rating==rating_t::fastdr2 )
    {
        //
        // Select winner and return
        //
        int winner_idx = -1;
        for(int i=0; i<winners.size(); i++)
            if( winners[i].is_set && (winner_idx<0 || winners[i].rating>winners[winner_idx].rating) )
                winner_idx = i;
        if( winner_idx<0 )
            return NULL;
        winners[winner_idx].is_set = false;
        return &winners[winner_idx];
    }
    else if( settings.con_rating==rating_t::fastdr2_x5 ||
             settings.con_rating==rating_t::fastdr2_x25 )
    {
        const int cand_max = settings.con_rating==rating_t::fastdr2_x5 ? 5 : 25;
        //
        // Select several winners
        //
        std::set<int> candidates;
        for(int ncand=0; ncand<cand_max; ncand++)
        {
            int winner_idx = -1;
            for(int i=0; i<winners.size(); i++)
                if( winners[i].is_set && candidates.count(i)==0 && (winner_idx<0 || winners[i].rating>winners[winner_idx].rating) )
                    winner_idx = i;
            if( winner_idx>=0 )
                candidates.insert(winner_idx);
        }

        //
        // Choose one with best delta_r2
        //
        float winner_dr2 = 0.0f;
        int   winner_idx = -1;
        for(int idx: candidates)
        {
            float dr2 = lin_eval.deltar2(&winners[idx].winner_act[0]);
            if( winner_idx<0 || dr2>winner_dr2 )
            {
                winner_idx = idx;
                winner_dr2 = dr2;
            }
        }
        if( winner_idx<0 )
            return NULL;
        winners[winner_idx].is_set = false;
        return &winners[winner_idx];
    }
    else
        ASSERT(0);
    return NULL;
}

template<class activity_t>
void v4_contest<activity_t>::append_and_refresh_winners(linear_evaluator *eval, const activity_t &v_act)
{
    ASSERT(eval==&lin_eval);
    if( settings.con_rating==rating_t::cov2 )
    {
        eval->append_activity(v_act);
        refresh_winners();
    }
    else if( settings.con_rating==rating_t::deltar2 )
    {
        linear_evaluator::daa append;
        eval->analyze(&v_act[0], &append);
        std::vector<linear_evaluator::daa*> refresh;
        for(int i=0; i<winners.size(); i++)
            if( winners[i].is_set )
            {
                ASSERT(winners[i].has_daa);
                refresh.push_back(&winners[i].daa);
            }
        eval->append_and_refresh(append, refresh);
        for(int i=0; i<winners.size(); i++)
            if( winners[i].is_set )
                winners[i].rating = winners[i].daa.deltar2;
    }
    else if( settings.con_rating==rating_t::fastdr2 ||
             settings.con_rating==rating_t::fastdr2_x5 ||
             settings.con_rating==rating_t::fastdr2_x25 )
    {
        eval->append_activity(v_act);
        refresh_winners();
    }
    else
        ASSERT(0);
}

template<class a_t>
void parallel_v4_refresh_adapter(void *p, int idx)
{
    v4_contest<a_t> &con = *((v4_contest<a_t>*)p);
    con.refresh_ith_winner(idx);
}

template<class activity_t>
void v4_contest<activity_t>::refresh_winners()
{
    timer tmr(pcnt_refresh);
    parallel_for(parallel_v4_refresh_adapter<activity_t>, this, winners.size());
}

template<class activity_t>
void v4_contest<activity_t>::refresh_ith_winner(int winner_idx)
{
    winner &w = winners[winner_idx];
    if( !w.is_set )
        return;
    if( settings.con_rating==rating_t::cov2 )
    {
        w.rating = lin_eval.estimate_trn_covariation(&w.winner_act[0]);
    }
    else if( settings.con_rating==rating_t::deltar2 )
    {
        ASSERT(w.has_daa);
        lin_eval.analyze(&w.winner_act[0], &w.daa);
        w.rating = w.daa.deltar2;
        w.has_daa = true;
    }
    else if( settings.con_rating==rating_t::fastdr2 ||
             settings.con_rating==rating_t::fastdr2_x5 ||
             settings.con_rating==rating_t::fastdr2_x25 )
    {
        //lin_eval.estimate_trn_covariation(&w.winner_act[0])/(w.anrm2+1.0e-20f);
        w.rating = compute_rating(settings.con_rating, settings.omega_rat, w.winner_act, lin_eval);
    }
    else
        ASSERT(0);
}

class v4_covariates_scanner_env
{
public:
    const activities &act_in;
    const std::vector<float> &means;
    const v4_settings settings;
    std::vector<std::vector<int>> &v_out;

    v4_covariates_scanner_env(const activities &ai, const std::vector<float> &me, v4_settings s, std::vector<std::vector<int>> &vo):
        act_in(ai), means(me), settings(s), v_out(vo){}
};

static void parallel_select_covariates(void *p, const int channel_idx)
{
    v4_covariates_scanner_env &env = *((v4_covariates_scanner_env*)p);
    const activities &act_in = env.act_in;
    const std::vector<float> &means = env.means;
    const v4_settings &settings = env.settings;
    const int64_t ntrn = act_in.ntrn;

    //
    // Compute excess covariations
    //
    std::vector<std::tuple<int,float> > excess_cov;
    for(int64_t n_idx=0; n_idx<act_in.nchannels; n_idx++)
    {
        float r = 0.0f;
        const float expected = means[channel_idx]*means[n_idx];
        if( act_in.is_i8() )
        {
            const int8_t *p1 = act_in.get_i8(channel_idx);
            const int8_t *p2 = act_in.get_i8(n_idx);
            for(int64_t k=0; k<ntrn; k++)
                r += SNET_I2F*SNET_I2F*float(p1[k])*float(p2[k])-expected;
        }
        else
        {
            ASSERT(act_in.is_f32());
            const float *p1 = act_in.get_f32(channel_idx);
            const float *p2 = act_in.get_f32(n_idx);
            for(int64_t k=0; k<ntrn; k++)
                r += p1[k]*p2[k]-expected;
        }
        if( r>0.0f )
            excess_cov.push_back(std::make_tuple(int(n_idx),float(r)));
    }

    //
    // Leave just most correlated
    //
    std::sort(std::begin(excess_cov), std::end(excess_cov),
              [](std::tuple<int,float> a, std::tuple<int,float> b) { return std::get<1>(a) > std::get<1>(b); });
    while( excess_cov.size()>settings.nneighbors )
        excess_cov.pop_back();

    //
    // Output
    //
    for(auto a: excess_cov)
        env.v_out[channel_idx].push_back(std::get<0>(a));
}

static std::string generate_report(const activities &act_in,
                            const seed_detector &seed,
                            const detector &det)
{
    if( seed.channel_idx1<0 )
    {
        char buf[1024];
        sprintf(buf, "D1[g=%d] = c%d",
                int(det.input0.size()), int(seed.channel_idx0));
        return buf;
    }
    else if( seed.channel_idx2<0 )
    {
        char buf[1024];
        sprintf(buf, "D2[g=%d,%d] = c%d & c%d",
                int(det.input0.size()), int(det.input1.size()),
                int(seed.channel_idx0), int(seed.channel_idx1));
        return buf;
    }
    else if( seed.channel_idx3<0 )
    {
        char buf[1024];
        sprintf(buf, "D3[g=%d,%d,%d] = c%d & c%d & c%d",
                int(det.input0.size()), int(det.input1.size()), int(det.input2.size()),
                int(seed.channel_idx0), int(seed.channel_idx1), int(seed.channel_idx2));
        return buf;
    }
    else if( seed.channel_idx4<0 )
    {
        char buf[1024];
        sprintf(buf, "D4[g=%d,%d,%d,%d] = c%d & c%d & c%d & c%d",
                int(det.input0.size()), int(det.input1.size()), int(det.input2.size()), int(det.input3.size()),
                int(seed.channel_idx0), int(seed.channel_idx1), int(seed.channel_idx2), int(seed.channel_idx3));
        return buf;
    }
    else
    {
        char buf[1024];
        sprintf(buf, "D5[g=%d,%d,%d,%d,%d] = c%d & c%d & c%d & c%d & c%d",
                int(det.input0.size()), int(det.input1.size()), int(det.input2.size()), int(det.input3.size()), int(det.input4.size()),
                int(seed.channel_idx0), int(seed.channel_idx1), int(seed.channel_idx2), int(seed.channel_idx3), int(seed.channel_idx4));
        return buf;

    }
}

template<class avector_t>
void v4_run(linear_evaluator  &lin_eval,
            const activities  &act_in,
            v4_settings settings,
            const std::vector<std::vector<int>> &most_correlated,
            const and_type_t AND_TYPE,
            std::atomic<int32_t> &running_seed,
            int out_channels,
            net_midlayer &ml_out,
            activities   &act_out)
{
    const or_type_t  OR_TYPE  = or_type_t::OR_MAX;
    generalizer_evaluator gen_eval(act_in, lin_eval);
    v4_contest<avector_t> contest(act_in, lin_eval, most_correlated, running_seed, settings, AND_TYPE, OR_TYPE);
    for(int channels_done=0; channels_done<out_channels;)
    {
        contest.run_competition();
        for(int attempt_idx=0; attempt_idx<settings.batch_size && channels_done<out_channels; attempt_idx++)
        {
            // retrieve winner
            typename v4_contest<avector_t>::winner *p_winner = contest.retrieve_winner();
            if( p_winner==NULL )
                continue;
            typename v4_contest<avector_t>::winner &w = *p_winner;

            // perform generalization on winner
            detector  gen_det;
            avector_t gen_act;
            generalize<avector_t>(act_in, lin_eval, gen_eval, settings,
                                  w.seed, AND_TYPE, OR_TYPE,
                                  &gen_det, &gen_act);

            // prepare report
            std::string rep = generate_report(act_in, w.seed, gen_det);

            // Process winner
            contest.append_and_refresh_winners(&lin_eval, gen_act);
            act_out.append1(gen_act,rep);
            ml_out.append1(gen_det);
            printf("%s[V4:%04d]  TrnRms=%.4f  TstRms=%.4f  TstCls=%.4f  %s\n",
                   settings.rep_prefix.c_str(),
                   int(channels_done),
                   (double)lin_eval.trn_rms(), (double)lin_eval.tst_rms(), (double)lin_eval.tst_cls(),
                   rep.c_str());
            fflush(stdout);
            channels_done++;
        }
    }
}

void v4_layer(linear_evaluator  &lin_eval,
              const activities  &act_in,
              v4_settings settings,
              int out_channels,
              net_midlayer &ml_out,
              activities   &act_out,
              predictions  &pred_out)
{
    int t0 = dtickcount();
    const int64_t ntrn = act_in.ntrn;
    bool is_i8 = act_in.is_i8();

    //
    // Internal RNGs
    // Running seed for internal RNGs, initialized with global seed provided by settings
    //
    std::atomic<int32_t> running_seed(settings.rng_seed);

    //
    // Prepare outputs
    //
    act_out = activities(act_in.ntrn, act_in.ntst, is_i8 ? storage_format::i8 : storage_format::f32);
    act_out.append_realm(realm<channel>::create_grid("all",1,1,0));
    ml_out = net_midlayer();
    ml_out.append_realm(realm<detector>::create_grid("all",1,1,0));

    //
    // First, build and analyze correlation matrix
    //
    std::vector<std::vector<int>> most_correlated(act_in.nchannels);
    {
        int t_corr = dtickcount();

        //
        // Compute vector of mean activity values
        //
        std::vector<float> means;
        for(int64_t channel_idx=0; channel_idx<act_in.nchannels; channel_idx++)
        {
            float r = 0.0f;
            if( is_i8 )
            {
                const int8_t *p1 = act_in.get_i8(channel_idx);
                for(int64_t k=0; k<ntrn; k++)
                    r += p1[k];
                means.push_back(SNET_I2F*r/ntrn);
            }
            else
            {
                ASSERT(act_in.is_f32());
                const float *p1 = act_in.get_f32(channel_idx);
                for(int64_t k=0; k<ntrn; k++)
                    r += p1[k];
                means.push_back(r/ntrn);
            }
        }

        //
        // For each activity compute its excess covariation (covariation minus
        // product of means) with each of the other activities and select most
        // correlated channels.
        //
        v4_covariates_scanner_env env(act_in, means, settings, most_correlated);
        parallel_for(parallel_select_covariates, &env, most_correlated.size());
        t_corr = dtickcount()-t_corr;
        printf("V4: correlation matrix built in %d ms\n", t_corr);
        fflush(stdout);
    }

    //
    //
    //
    const and_type_t AND_TYPE = is_i8 ? and_type_t::AND_MIN : and_type_t::AND_MUL;
    if( is_i8 )
        v4_run<activity8_vector>(lin_eval, act_in, settings,
                                 most_correlated, AND_TYPE, running_seed,
                                 out_channels,
                                 ml_out, act_out);
    else
        v4_run<activityf_vector>(lin_eval, act_in, settings,
                                 most_correlated, AND_TYPE, running_seed,
                                 out_channels,
                                 ml_out, act_out);
#if 0
            const or_type_t  OR_TYPE  = or_type_t::OR_MAX;
    generalizer_evaluator   gen_eval(act_in, lin_eval);
    v4_contest<activity8_vector> contest(act_in, lin_eval, most_correlated, running_seed, settings, AND_TYPE, OR_TYPE);
    for(int channels_done=0; channels_done<out_channels;)
    {
        contest.run_competition();
        for(int attempt_idx=0; attempt_idx<settings.batch_size && channels_done<out_channels; attempt_idx++)
        {
            // retrieve winner
            //typename v4_contest<activityf_vector>::winner *p_winner = contest.retrieve_winner_and_append(&lin_eval);
            typename v4_contest<activity8_vector>::winner *p_winner = contest.retrieve_winner();
            if( p_winner==NULL )
                continue;
            typename v4_contest<activity8_vector>::winner &w = *p_winner;

            // perform generalization on winner
            detector         gen_det;
            activity8_vector gen_act;
            generalize<activity8_vector>(act_in, lin_eval, gen_eval, settings,
                                         w.seed, AND_TYPE, OR_TYPE,
                                         &gen_det, &gen_act);

            // prepare report
            std::string rep = generate_report(act_in, w.seed, gen_det);

            // Process winner
            contest.append_and_refresh_winners(&lin_eval, gen_act);
            act_out.append1(gen_act,rep);
            ml_out.append1(gen_det);
            printf("%s[V4:%04d]  TrnRms=%.4f  TstRms=%.4f  TstCls=%.4f  %s\n",
                   settings.rep_prefix.c_str(),
                   int(channels_done),
                   (double)lin_eval.trn_rms(), (double)lin_eval.tst_rms(), (double)lin_eval.tst_cls(),
                   rep.c_str());
            fflush(stdout);
            channels_done++;
        }
    }
#endif
    pred_out = lin_eval.outputs();

    t0 = dtickcount()-t0;
    printf("V4: layer is done in %d ms\n", t0);
    fflush(stdout);
}

std::vector<int> select_best_features(const predictions &baseline_col,
                                      const activities &act_full,
                                      int nfeatures)
{
    int t0 = dtickcount();
    std::vector<int> chosen;

    ASSERT(act_full.ntrn==baseline_col.ntrn);
    ASSERT(act_full.ntst==baseline_col.ntst);
    int ctot = 0;
    for(const realm<channel> &rr: act_full.realms)
    {
        ASSERT(rr.is_densely_packed());
        ctot += rr.grid_depth*rr.grid_height*rr.grid_width;
    }
    ASSERT(ctot==act_full.nchannels);

    //
    // Settings object for V4 contest
    //
    v4_settings settings;
    settings.nneighbors = 1;
    settings.eval_prob = 1.0;
    settings.rng_seed  = 1;
    settings.batch_size = 5;
    settings.dmax = 1;
    settings.con_rating = rating_t::fastdr2;
    settings.gen_rating = rating_t::fastdr2;
    settings.omega_rat = 1.0e-5;

    //
    // Dummy neighbors array for V4 contest
    //
    std::vector<std::vector<int>> dummy_neighbors;
    for(int i=0; i<act_full.nchannels; i++)
        dummy_neighbors.push_back({i});

    //
    // Channel selection by means of V4 contest object
    //
    std::atomic<int32_t> running_seed(0);
    linear_evaluator lin_eval(baseline_col, nfeatures, 1.0e-6);
    v4_contest<activityf_vector> contest(act_full, lin_eval, dummy_neighbors, running_seed, settings, and_type_t::AND_MIN, or_type_t::OR_MAX);
    for(; chosen.size()<nfeatures;)
    {
        contest.run_competition();
        typename v4_contest<activityf_vector>::winner *p_winner = contest.retrieve_winner();
        ASSERT(p_winner!=NULL);
        typename v4_contest<activityf_vector>::winner &w = *p_winner;
        ASSERT(w.seed.inputs_cnt()==1 && w.seed.ggain==gain_t::gain1);
        chosen.push_back(w.seed.channel_idx0);
        contest.append_and_refresh_winners(&lin_eval, w.winner_act);
        //printf("cls = %.4f\n", lin_eval.tst_cls());
        fflush(stdout);
    }
    t0 = dtickcount()-t0;
    //printf("Feature selection is done in %d ms\n", t0);
    fflush(stdout);
    return chosen;
}

/*void v4_layer(const predictions &baseline,
              const activities  &act_in,
              v4_settings settings,
              int out_channels,
              double lambda_reg,
              net_midlayer &ml_out,
              activities   &act_out,
              predictions  &pred_out)
{
    linear_evaluator lin_eval(baseline, out_channels, lambda_reg);
    v4_layer(lin_eval, act_in, settings, out_channels, lambda_reg, ml_out, act_out, pred_out);
}*/

#if 0
void v4_layer_exacting(linear_evaluator  &global_eval,
              const activities  &act_in,
              v4_settings settings,
              int out_channels,
              double lambda_reg,
              net_midlayer &ml_out,
              activities   &act_out,
              predictions  &pred_out)
{
    int t0 = dtickcount();
    const int64_t ntrn = act_in.ntrn;
    ASSERT(settings.batch_size>=1);
    ASSERT(global_eval.outputs().ntrn==act_in.ntrn);
    ASSERT(global_eval.outputs().ntst==act_in.ntst);
    ASSERT(act_in.realms.size()>=1);

    //
    // Internal RNGs
    // Running seed for internal RNGs, initialized with global seed provided by settings
    //
    std::atomic<int32_t> running_seed(settings.rng_seed);

    //
    // Prepare outputs
    //
    act_out = activities(act_in.ntrn, act_in.ntst, storage_format::i8);
    act_out.append_realm(realm<channel>::create_grid("all",1,1,0));
    ml_out = net_midlayer();
    ml_out.append_realm(realm<detector>::create_grid("all",1,1,0));

    //
    // First, build and analyze correlation matrix
    //
    std::vector<std::vector<int>> most_correlated(act_in.nchannels);
    {
        int t_corr = dtickcount();

        //
        // Compute vector of mean activity values
        //
        std::vector<float> means;
        for(int64_t channel_idx=0; channel_idx<act_in.nchannels; channel_idx++)
        {
            float r = 0.0f;
            const int8_t *p1 = act_in.get_i8(channel_idx);
            for(int64_t k=0; k<ntrn; k++)
                r += p1[k];
            means.push_back(SNET_I2F*r/ntrn);
        }

        //
        // For each activity compute its excess covariation (covariation minus
        // product of means) with each of the other activities and select most
        // correlated channels.
        //
        v4_covariates_scanner_env env(act_in, means, settings, most_correlated);
        parallel_for(parallel_select_covariates, &env, most_correlated.size());
        t_corr = dtickcount()-t_corr;
        printf("V4: correlation matrix built in %d ms\n", t_corr);
        fflush(stdout);
    }

    //
    //
    //
    const and_type_t AND_TYPE = and_type_t::AND_MIN;
    const or_type_t  OR_TYPE  = or_type_t::OR_MAX;
    generalizer_evaluator   gen_eval(act_in, global_eval); // do we need local generalizer?? cons: memory reallocation, pro: more precise
    for(int channels_done=0; channels_done<out_channels;)
    {
        //
        // Local contest
        //
        class candidate_record
        {
        public:
            seed_detector seed;
            detector det;
            activity8_vector act;
        };
        std::vector<candidate_record> contest_winners;
        linear_evaluator local_eval(global_eval.outputs(), settings.batch_size, lambda_reg);
        v4_contest<activity8_vector> contest(act_in, local_eval, most_correlated, running_seed, settings, AND_TYPE, OR_TYPE);
        contest.run_competition();
        for(int idx=0; idx<settings.batch_size; idx++)
        {
            typename v4_contest<activity8_vector>::winner *p_winner = contest.retrieve_winner();
            if( p_winner==NULL )
                break;
            typename v4_contest<activity8_vector>::winner &w = *p_winner;
            detector         gen_det;
            activity8_vector gen_act;
            generalize<activity8_vector>(act_in, local_eval, gen_eval, settings,
                                         w.seed, AND_TYPE, OR_TYPE,
                                         &gen_det, &gen_act);
            contest_winners.push_back(candidate_record{w.seed,gen_det,gen_act});
            contest.append_and_refresh_winners(&local_eval, gen_act);
        }
        if( contest_winners.size()==0 )
            continue;
        const int nbatch = contest_winners.size();

        //
        // Select best detectors from contest
        //
        class analysis_record
        {
        public:
            bool retrieved;
            linear_evaluator::daa daa;
            analysis_record():retrieved(false){}
        };
        std::vector<analysis_record> analysis_records(nbatch);
        parallel_for([&](void* _p,int _idx){global_eval.analyze(&contest_winners[_idx].act[0], &analysis_records[_idx].daa);}, NULL, nbatch);
        int nbest = roundf(settings.frac_chosen*settings.batch_size);
        nbest = imax(nbest,1);
        nbest = imin(nbest,nbatch);
        for(int iter=0; iter<nbest && channels_done<out_channels; iter++)
        {
            int idx_best = -1;
            for(int i=0; i<nbatch; i++)
                if( !analysis_records[i].retrieved &&
                        (idx_best<0 ||
                         analysis_records[i].daa.deltar2>analysis_records[idx_best].daa.deltar2) )
                    idx_best = i;
            ASSERT(idx_best>=0);
            const candidate_record &best = contest_winners[idx_best];
            analysis_record &best_rec    = analysis_records[idx_best];

            best_rec.retrieved = true;
            std::vector<linear_evaluator::daa*> update_vec;
            for(int i=0; i<nbatch; i++)
                if( !analysis_records[i].retrieved )
                    update_vec.push_back(&analysis_records[i].daa);

            std::string rep = generate_report(act_in, best.seed, best.det);
            global_eval.append_and_refresh(best_rec.daa, update_vec);
            act_out.append1(best.act,rep);
            ml_out.append1(best.det);
            printf("%s[V4:%04d]  TrnRms=%.4f  TstRms=%.4f  TstCls=%.4f  %s\n",
                   settings.rep_prefix.c_str(),
                   int(channels_done),
                   (double)global_eval.trn_rms(), (double)global_eval.tst_rms(), (double)global_eval.tst_cls(),
                   rep.c_str());
            fflush(stdout);
            channels_done++;
        }
    }
    pred_out = global_eval.outputs();

    t0 = dtickcount()-t0;
    printf("V4: layer is done in %d ms\n", t0);
    fflush(stdout);
}
#endif

}
