/*************************************************************************
StrongNet/Contour Engine training framework.
Copyright (c) 2020 Sergey Bochkanov (ALGLIB project).

>>> SOURCE LICENSE >>>
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation (www.fsf.org); either version 2 of the 
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

A copy of the GNU General Public License is available at
http://www.fsf.org/licensing/licenses
>>> END OF LICENSE >>>
*************************************************************************/
#ifndef LAYER4_H
#define LAYER4_H

#include "snetcommon.h"
#include "layer1.h"
#include "layer2.h"

namespace strongnet
{
class v4_settings
{
public:
    int     nneighbors;
    float   eval_prob;
    int     rng_seed;
    bool    restrict_gen_by_realm;
    int     batch_size;
    float   frac_chosen;
    int     dmax;
    int     ncenters;
    std::string rep_prefix;
    rating_t con_rating;
    rating_t gen_rating;
    float   omega_rat;
    float   gen_rating_growth;

    v4_settings():
        nneighbors(300),
        eval_prob(0.9f),
        rng_seed(1),
        restrict_gen_by_realm(false),
        batch_size(50),
        frac_chosen(0.1f),
        dmax(3),
        ncenters(1),
        con_rating(rating_t::fastdr2),
        gen_rating(rating_t::fastdr2),
        omega_rat(1.0e-6f),
        gen_rating_growth(0)
    {}
};

void v4_layer(linear_evaluator  &lin_eval,
              const activities  &act_in,
              v4_settings settings,
              int out_channels,
              net_midlayer &ml_out,
              activities   &act_out,
              predictions  &pred_out);

std::vector<int> select_best_features(const predictions &baseline_col,
                                      const activities &act_full,
                                      int nfeatures);

/*void v4_layer(const predictions &baseline,
              const activities  &act_in,
              v4_settings settings,
              int out_channels,
              double lambda_reg,
              net_midlayer &ml_out,
              activities   &act_out,
              predictions  &pred_out);*/

#if 0
void v4_layer_exacting(linear_evaluator  &global_eval,
              const activities  &act_in,
              v4_settings settings,
              int out_channels,
              double lambda_reg,
              net_midlayer &ml_out,
              activities   &act_out,
              predictions  &pred_out);
#endif

//----------------- internal classes not intended for external use ------------------------------
template<class activity_t>
class v4_contest
{
public:
    class winner
    {
    friend class v4_contest;
    template<class a_t>
    friend void* v4_contest_run_parallel(void *ptr);
    private:
        bool                  is_set;
        bool                  has_daa;
        linear_evaluator::daa daa;
        float                 anrm2;
    public:
        seed_detector    seed;
        activity_t       winner_act;
        float            rating;

        void invalidate()
        { is_set = has_daa = false; daa.invalidate(); seed = seed_detector(); rating = 0; anrm2 = 0; }

        winner()
        { invalidate(); }

        //const linear_evaluator::daa& daa() const { return deep_analysis; }
    };

    static std::atomic<int32_t> pcnt_refresh;
    v4_contest(const activities &ai,
               const linear_evaluator &le,
               const std::vector<std::vector<int> > &mcn,
               std::atomic<int32_t> &rs,
               v4_settings s, and_type_t at, or_type_t ot);
    void run_competition();

    // retrieves current winner and marks it as non-set
    winner* retrieve_winner();

    // appends activity given by its deep analysis to the linear solver (it must be the one you specified
    // during object creation) and performs quick update on the current winners
    void append_and_refresh_winners(linear_evaluator *eval, const activity_t &v_act);

    // called after append() to linear evaluator, forces recomputation of winner rms-es
    // (VERY SLOW because we can not use fast updates provided by deep analysis feature;
    //  we recommend you to stick to retrieve_winner_and_append() ).
    void refresh_winners();
private:
    template<class a_t> friend void parallel_v4_refresh_adapter(void *p, int idx);
    void refresh_ith_winner(int winner_idx);

    static const int            line_width;
    const activities            &act_in;
    const linear_evaluator      &lin_eval;
    const std::vector<std::vector<int> > &most_correlated_neighbors;
    std::atomic<int32_t>        &running_rng_seed;
    const v4_settings           settings;
    const and_type_t            and_type;
    const or_type_t             or_type;

    // competition
    std::atomic<int32_t>        lines_examined;
    std::vector<winner>         winners;
    std::vector<int>            rng_seeds_to_use;
template<class a_t>
friend void* v4_contest_run_parallel(void *ptr);
};

template<class activity_t>
std::atomic<int32_t> v4_contest<activity_t>::pcnt_refresh(0);

}

#endif
