/*************************************************************************
StrongNet/Contour Engine training framework.
Copyright (c) 2020 Sergey Bochkanov (ALGLIB project).

>>> SOURCE LICENSE >>>
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation (www.fsf.org); either version 2 of the 
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

A copy of the GNU General Public License is available at
http://www.fsf.org/licensing/licenses
>>> END OF LICENSE >>>
*************************************************************************/
#include "snet.h"
#include <numeric>
#include <algorithm>
#include <random>
#include <mutex>
#include "3rdparty/alglib/optimization.h"
#include "3rdparty/alglib/dataanalysis.h"

namespace strongnet
{

float invsigma0_safe(float x)
{
    float L = 1.0e-7f;
    if( x<=L )
        x = L;
    if( x>=1-L )
        x = 1-L;
    ASSERT(x>0 && x<1);
    return logf(x/(1-x));
}

predictions invsigma0(const predictions &p)
{
    predictions r = p;
    for(int64_t k=0; k<r.nclasses; k++)
        for(int64_t i=0; i<r.ntrn+r.ntst; i++)
            r.fpredictions[i+k*r.stride] = invsigma0_safe(r.fpredictions[i+k*r.stride]);
    return r;
}

void singlecolumn(const predictions &baseline,
                 const activities &act_v1,
                 int v2_width,
                 int v4_width,
                 double lambda_reg,
                 int seed,
                 //net_midlayer     &ml_out,
                 //activities       &act_out,
                 predictions &pred_out)
{
    linear_evaluator eval_local(baseline, v2_width+v4_width, lambda_reg);

    // V2 processing
    net_midlayer ml_v2;
    activities   act_v2;
    predictions  pred_v2;
    v2_settings  s2;
    s2.seeds_count = v2_width;
    s2.rng_seed = seed;
    s2.batch_size = 50;
    v2_layer_xx(eval_local, act_v1, s2, v2_width, ml_v2, act_v2, pred_v2);

    // V4 processing
    net_midlayer ml_v4;
    activities   act_v4;
    predictions  pred_v4;
    v4_settings  s4;
    s4.rng_seed = seed;
    s4.batch_size = 50;
    v4_layer(eval_local, act_v2, s4, v4_width, ml_v4, act_v4, pred_v4);

    pred_out = pred_v4;
}

predictions net_multicolumn::apply(const raw_dataset &ds_raw) const
{
    //
    // Integrity checks
    //
    ASSERT(nclasses>=1);
    ASSERT(v2_columns.size()==nclasses);
    ASSERT(v4_columns.size()==nclasses);
    ASSERT(predictors.size()==nclasses);

    //
    // V1 processing, then apply
    //
    activities v1_act;
    v1.apply(ds_raw, v1_act);
    return apply(ds_raw, v1_act);
}

net_multicolumn::net_multicolumn(std::string filename)
{
    unserialize_from(filename, this);
}

void net_multicolumn::params_and_flops(const net_midlayer &layer, double &params, double &flops) const
{
    flops = 0;
    params = 0;
    for(const realm<detector> &rr: layer.realms)
        for(const detector &d: rr.elements)
        {
            int c = d.input0.size()+d.input1.size()+d.input2.size()+d.input3.size()+d.input4.size();
            params += c;
            flops  += c*imax(d.transform.size(),1);
        }
}

std::vector<predictions> net_multicolumn::train(
        const raw_dataset &ds_raw,
        const net_v1layer &layer1,
        const std::vector<int> columns_to_train,
        int v2_width,
        int v4_width,
        settings s)
{
    //
    // Check
    //
    ASSERT(columns_to_train.size()>=1);
    for(int c: columns_to_train)
        ASSERT(c>=0 && c<ds_raw.nclasses);

    //
    // V1 processing
    //
    activities act_v1;
    layer1.apply(ds_raw, act_v1, true);

    //
    // Train
    //
    predictions zero_baseline(ds_raw);
    std::vector<predictions> vp_out;
    nclasses = columns_to_train.size();
    v1 = layer1;
    v2_columns.clear();
    v4_columns.clear();
    predictors.clear();
    for(int column_id: columns_to_train)
    {
        char sbuf[1024];

        // prepare 1-vs-all problem
        predictions      pred_local = predictions::onevsall(zero_baseline, column_id);

        // V2 processing
        net_midlayer ml_v2_local;
        activities   act_v2_local;
        predictions  pred_v2_local;
        linear_evaluator eval_v2(pred_local, v2_width, s.lambda_reg);
        v2_settings  s2;
        sprintf(sbuf, "[COL%02d]", column_id);
        s2.seeds_count = v2_width;
        s2.rng_seed = s.seed;
        s2.rep_prefix = sbuf;
        s2.batch_size = s.batch_size;
        s2.omega_rat = s.omega_rat;
        if( s.v2_maxdist>0 )
            s2.max_dist = s.v2_maxdist;
        ASSERT(!s.use_exacting_layer);
        v2_layer_xx(eval_v2, act_v1, s2, v2_width, ml_v2_local, act_v2_local, pred_v2_local);

        // recompute linear predictor (eval_v2 accepts channels in the shuffled order because it reorders channels prior to output)
        linear_evaluator eval_final(pred_local, v2_width+v4_width, s.lambda_reg);
        for(int i=0; i<act_v2_local.nchannels; i++)
        {
            ASSERT(act_v2_local.is_i8());
            const int8_t *p = act_v2_local.get_i8(i);
            eval_final.append_activity(activity8_vector(p, p+act_v2_local.ntrn+act_v2_local.ntst));
        }

        // V4 processing
        net_midlayer ml_v4_local;
        activities   act_v4_local;
        predictions  pred_v4_local;
        v4_settings  s4;
        s4.rep_prefix = sbuf;
        s4.batch_size = s.batch_size;
        s4.omega_rat = s.omega_rat;
        ASSERT(!s.use_exacting_layer);
        v4_layer(eval_final, act_v2_local, s4, v4_width, ml_v4_local, act_v4_local, pred_v4_local);

        // out
        v2_columns.push_back(ml_v2_local);
        v4_columns.push_back(ml_v4_local);
        predictors.push_back(eval_final.model());
        vp_out.push_back(pred_v4_local);
    }
    return vp_out;
}

predictions net_multicolumn::train(
        const raw_dataset &ds_raw,
        const net_v1layer &layer1,
        int net_width,
        settings s)
{
    std::vector<int> columns_to_train;
    for(int k=0; k<ds_raw.nclasses; k++)
        columns_to_train.push_back(k);
    std::vector<predictions> vp_out = train(ds_raw, layer1, columns_to_train, net_width, net_width, s);
    predictions pred_out = predictions::merge_onevsall(vp_out);

    //
    // Report
    //
    predictions baseline(ds_raw);
    float trn_rms, tst_rms, tst_cls;
    pred_out.compute_errors(&trn_rms, &tst_rms, &tst_cls);
    printf("[CONCAT-OUT] TrnRms=%.4f    TstRms=%.4f    TstCls=%.4f\n",
           (double)trn_rms, (double)tst_rms, (double)tst_cls);
    linear_evaluator final_eval(baseline, baseline.nclasses, s.lambda_reg);
    for(int64_t k=0; k<baseline.nclasses; k++)
        final_eval.append_activity(pred_out.kth_as_activity(k));
    final_predictor = final_eval.model();
    predictions final_out = final_eval.outputs();
    final_out.compute_errors(&trn_rms, &tst_rms, &tst_cls);
    printf("[FINAL-OUT]  TrnRms=%.4f    TstRms=%.4f    TstCls=%.4f\n",
           (double)trn_rms, (double)tst_rms, (double)tst_cls);

    //
    // Done
    //
    return pred_out;
}

// concatenation
net_multicolumn net_multicolumn::concat(const std::vector<net_multicolumn> &v)
{
    //
    // Integrity checks
    //
    ASSERT(v.size()>=1);
    for(const net_multicolumn &mc: v)
        ASSERT(mc.v1==v[0].v1);

    //
    // Concat
    //
    net_multicolumn r;
    r.nclasses = 0;
    r.v1 = v[0].v1;
    for(const net_multicolumn &mc: v)
    {
        r.nclasses += mc.nclasses;
        r.v2_columns.insert(r.v2_columns.end(), mc.v2_columns.begin(), mc.v2_columns.end());
        r.v4_columns.insert(r.v4_columns.end(), mc.v4_columns.begin(), mc.v4_columns.end());
        r.predictors.insert(r.predictors.end(), mc.predictors.begin(), mc.predictors.end());
    }
    r.final_predictor = linear_predictor();
    return r;
}

predictions net_multicolumn::apply(const raw_dataset &ds_raw, const activities &v1_act) const
{
    //
    // Integrity checks
    //
    ASSERT(nclasses>=1);
    ASSERT(v2_columns.size()==nclasses);
    ASSERT(v4_columns.size()==nclasses);
    ASSERT(predictors.size()==nclasses);

    //
    // V2, V4 prediction
    //
    predictions zero_baseline(ds_raw);
    std::vector<predictions> vp_out;
    for(int class_idx=0; class_idx<nclasses; class_idx++)
    {
        predictions baselinei = predictions::onevsall(zero_baseline, class_idx);
        const net_midlayer &v2i = v2_columns[class_idx];
        const net_midlayer &v4i = v4_columns[class_idx];
        const linear_predictor &p = predictors[class_idx];
        activities v2_act, v4_act;
        v2i.apply(v1_act, v2_act);
        v4i.apply(v2_act, v4_act);
        vp_out.push_back(p.predict(baselinei, {&v2_act, &v4_act}));
    }
    predictions pred_concat = predictions::merge_onevsall(vp_out);
    // TODO: final predictor!!!!
    return pred_concat;
}

void net_multicolumn::dbg_analyze_stats() const
{
    //
    // Integrity checks
    //
    ASSERT(nclasses>=1);
    ASSERT(v2_columns.size()==nclasses);
    ASSERT(v4_columns.size()==nclasses);
    ASSERT(predictors.size()==nclasses);

    //
    // Analyze
    //
    double params_cnt = 0, flops_cnt = 0, worst_l1_expansion = 0;
    for(int class_idx=0; class_idx<nclasses; class_idx++)
    {
        double flp, prm;

        // V2
        params_and_flops(v2_columns[class_idx], prm, flp);
        params_cnt += prm;
        flops_cnt  += flp;

        // V4
        params_and_flops(v4_columns[class_idx], prm, flp);
        params_cnt += prm;
        flops_cnt  += flp;

        // output
        params_cnt += predictors[class_idx].get_nchannels();
        flops_cnt  += predictors[class_idx].get_nchannels();

        // L1 expansion coefficient
        double l1_expansion = 0;
        for(int k=0; k<predictors[class_idx].get_nchannels(); k++)
            l1_expansion += fabs(predictors[class_idx].get_coeff(k,0));
        worst_l1_expansion = fmax(worst_l1_expansion, l1_expansion);
    }
    printf("Parameters count: %8.1fK\n", double(0.001*params_cnt));
    printf("FLOPS count:      %8.1fK\n", double(0.001*flops_cnt));
    printf("L1-exp:           %8.1f\n",  double(worst_l1_expansion));
}

void serialize(FILE *f, const net_multicolumn *ptr)
{
    write_magic8(f, 231);
    serialize(f, &ptr->nclasses);
    serialize(f, &ptr->v1);
    serialize(f, &ptr->v2_columns);
    serialize(f, &ptr->v4_columns);
    serialize(f, &ptr->predictors);
    serialize(f, &ptr->final_predictor);
    write_magic8(f, 96);
}

void unserialize(FILE *f, net_multicolumn *ptr)
{
    check_magic8(f, 231);
    unserialize(f, &ptr->nclasses);
    unserialize(f, &ptr->v1);
    unserialize(f, &ptr->v2_columns);
    unserialize(f, &ptr->v4_columns);
    unserialize(f, &ptr->predictors);
    unserialize(f, &ptr->final_predictor);
    check_magic8(f, 96);
}

rbf_predictor rbf_predictor::create(const raw_dataset &ds,
                                    int class_idx,
                                    const activities &act_in,
                                    const std::vector<float> &ref_dir,
                                    int rbf_centers,
                                    float rbf_rad,
                                    int rbf_top_dim,
                                    int rbf_bottom_dim)
{
    ASSERT(act_in.is_f32());
    rbf_predictor out;

    // analyze
    int class_size = 0;
    for(int i=0; i<ds.ntrn; i++)
        if( ds.entries[i].class_id==class_idx )
            class_size++;
    ASSERT(class_size>=1);
    rbf_centers = rbf_centers>0 ? imin(rbf_centers,class_size) : class_size;

    // input, prescaler
    out.nchannels_in = act_in.nchannels;
    out.prescaler.clear(); //out.prescaler.resize(out.nchannels_in, 1.0f);

    //
    // Centers: either copy all centers or use k-means to reduce dimensionality
    //
    out.ncenters = rbf_centers;
    out.centers.resize(out.ncenters*out.nchannels_in);
    if( rbf_centers==class_size )
    {
        // centers count = subsample size, just copy centers
        for(int64_t i=0, t=0; i<ds.ntrn; i++)
            if( ds.entries[i].class_id==class_idx )
            {
                const float   *p_src = act_in.get_f32()+i;
                const int64_t stride = act_in.get_stride();
                ASSERT(t<out.ncenters);
                for(int64_t j=0; j<out.nchannels_in; j++)
                    out.centers[t*out.nchannels_in+j] = p_src[j*stride];
                t++;
            }
    }
    else
    {
        // run k-means analysis
        alglib::real_2d_array xy;
        xy.setlength(class_size, out.nchannels_in);
        int item_idx = 0;
        for(int64_t i=0, t=0; i<ds.ntrn; i++)
            if( ds.entries[i].class_id==class_idx )
            {
                const float   *p_src = act_in.get_f32()+i;
                const int64_t stride = act_in.get_stride();
                for(int64_t j=0; j<out.nchannels_in; j++)
                    xy[item_idx][j] = p_src[j*stride];
                item_idx++;
            }
        ASSERT(item_idx==class_size);
        alglib::clusterizerstate cs;
        alglib::kmeansreport crep;
        alglib::clusterizercreate(cs);
        alglib::clusterizersetpoints(cs, xy, 2);
        alglib::clusterizersetkmeanslimits(cs, 1, 50);
        alglib::clusterizerrunkmeans(cs, rbf_centers, crep, alglib::parallel);
        ASSERT(crep.terminationtype>0);
        for(int64_t i=0; i<rbf_centers; i++)
            for(int64_t j=0; j<out.nchannels_in; j++)
                out.centers[i*out.nchannels_in+j] = crep.c[i][j];
    }

    //
    // Projections: reference one suggested by user, and ones selected via PCA
    //
    ASSERT(out.nchannels_in>=2);
    ASSERT(ref_dir.size()==out.nchannels_in);
    if( rbf_top_dim+rbf_bottom_dim+1>=out.nchannels_in )
    {
        rbf_top_dim = out.nchannels_in-1;
        rbf_bottom_dim = 0;
    }
    out.nprojections = rbf_top_dim+rbf_bottom_dim+1;
    out.prj.resize(out.nprojections*out.nchannels_in);
    float v = 1.0e-20f;
    for(int i=0; i<out.nchannels_in; i++)
    {
        float vv = ref_dir[i];
        v += vv*vv;
        out.prj[i] = vv;
    }
    v = 1.0f/sqrtf(v);
    for(int i=0; i<out.nchannels_in; i++)
        out.prj[i] *= v;
    if( rbf_top_dim+rbf_bottom_dim>0 )
    {
        // perform PCA
        alglib::real_2d_array xx;
        xx.setlength(class_size, out.nchannels_in);
        int item_idx = 0;
        for(int64_t i=0, t=0; i<ds.ntrn; i++)
            if( ds.entries[i].class_id==class_idx )
            {
                const float   *p_src = act_in.get_f32()+i;
                const int64_t stride = act_in.get_stride();
                float v = 0;
                for(int64_t j=0; j<out.nchannels_in; j++)
                    v += p_src[j*stride]*out.prj[j];
                for(int64_t j=0; j<out.nchannels_in; j++)
                    xx[item_idx][j] = p_src[j*stride]-v*out.prj[j];
                item_idx++;
            }
        ASSERT(item_idx==class_size);
        alglib::real_1d_array s2;
        alglib::real_2d_array pca;
        alglib::ae_int_t info;
        alglib::pcabuildbasis(xx, xx.rows(), xx.cols(), info, s2, pca, alglib::parallel);

        // offload
        ASSERT(rbf_top_dim+rbf_bottom_dim+1<=out.nchannels_in);
        int prj_idx = 1;
        for(int k=0; k<rbf_top_dim; k++)
        {
            float *p_dst = &out.prj[prj_idx*out.nchannels_in];
            for(int i=0; i<out.nchannels_in; i++)
                p_dst[i] = pca[i][k];
            prj_idx++;
        }
        for(int k=0; k<rbf_bottom_dim; k++)
        {
            float *p_dst = &out.prj[prj_idx*out.nchannels_in];
            for(int i=0; i<out.nchannels_in; i++)
                p_dst[i] = pca[i][out.nchannels_in-2-k];
            prj_idx++;
        }
    }

    // rad
    out.rad = rbf_rad;

    // coeffs
    out.coeffs.resize(out.ncenters+1, 0.0f);//!!!!!!!!!!!!!!!!!!
    //out.coeffs.resize(out.ncenters+1, 1.0f);
    //out.coeffs.back() = 0.0f;

    // done
    return out;
}

rbf_predictor rbf_predictor::create(const raw_dataset &ds,
                                    int class_idx,
                                    const activities &act_in,
                                    int rbf_centers,
                                    float rbf_rad)
{
    ASSERT(act_in.is_f32());
    rbf_predictor out;

    // analyze
    int class_size = 0;
    for(int i=0; i<ds.ntrn; i++)
        if( ds.entries[i].class_id==class_idx )
            class_size++;
    ASSERT(class_size>=1);
    rbf_centers = rbf_centers>0 ? imin(rbf_centers,class_size) : class_size;

    // input, prescaler
    out.nchannels_in = act_in.nchannels;
    out.prescaler.clear(); //out.prescaler.resize(out.nchannels_in, 1.0f);

    //
    // Centers: either copy all centers or use k-means to reduce dimensionality
    //
    out.ncenters = rbf_centers;
    out.centers.resize(out.ncenters*out.nchannels_in);
    if( rbf_centers==class_size )
    {
        // centers count = subsample size, just copy centers
        for(int64_t i=0, t=0; i<ds.ntrn; i++)
            if( ds.entries[i].class_id==class_idx )
            {
                const float   *p_src = act_in.get_f32()+i;
                const int64_t stride = act_in.get_stride();
                ASSERT(t<out.ncenters);
                for(int64_t j=0; j<out.nchannels_in; j++)
                    out.centers[t*out.nchannels_in+j] = p_src[j*stride];
                t++;
            }
    }
    else
    {
        // run k-means analysis
        alglib::real_2d_array xy;
        xy.setlength(class_size, out.nchannels_in);
        int item_idx = 0;
        for(int64_t i=0, t=0; i<ds.ntrn; i++)
            if( ds.entries[i].class_id==class_idx )
            {
                const float   *p_src = act_in.get_f32()+i;
                const int64_t stride = act_in.get_stride();
                for(int64_t j=0; j<out.nchannels_in; j++)
                    xy[item_idx][j] = p_src[j*stride];
                item_idx++;
            }
        ASSERT(item_idx==class_size);
        alglib::clusterizerstate cs;
        alglib::kmeansreport crep;
        alglib::clusterizercreate(cs);
        alglib::clusterizersetpoints(cs, xy, 2);
        alglib::clusterizersetkmeanslimits(cs, 1, 50);
        alglib::clusterizerrunkmeans(cs, rbf_centers, crep, alglib::parallel);
        ASSERT(crep.terminationtype>0);
        for(int64_t i=0; i<rbf_centers; i++)
            for(int64_t j=0; j<out.nchannels_in; j++)
                out.centers[i*out.nchannels_in+j] = crep.c[i][j];
    }

    // No projections
    out.nprojections = 0;
    out.prj.clear();

    // rad
    out.rad = rbf_rad;

    // coeffs
    out.coeffs.resize(out.ncenters+1, 0.0f);

    // done
    return out;
}

void rbf_predictor::compute_basis_function(const activities& act_in, int center_idx, activityf_vector &out, activityf_vector &buf) const
{
    const int64_t ntotal = act_in.ntrn+act_in.ntst;

    // clear accumulator
    out.resize(ntotal);
    for(int64_t i=0; i<ntotal; i++)
        out[i] = 0.0f;

    // no input prescaling is supported
    ASSERT(prescaler.size()==0);

#if 0
    // for all projections
    buf.resize(ntotal);
    for(int64_t proj_idx=0; proj_idx<nprojections; proj_idx++)
    {
        float proj_ctr = fdot_simd(&centers[center_idx*nchannels_in], &prj[proj_idx*nchannels_in], nchannels_in);
        for(int64_t i=0; i<ntotal; i++)
            buf[i] = -proj_ctr;
        for(int64_t i=0; i<nchannels_in; i++)
            fadd_simd(&buf[0], act_in.get_f32(i), ntotal, prj[proj_idx*nchannels_in+i]);
        for(int64_t i=0; i<ntotal; i++)
            out[i] += buf[i]*buf[i];
    }

    // nonlinearity
    for(int64_t i=0; i<ntotal; i++)
        out[i] = expf(-out[i]/(rad*rad));
#endif

    //
    ASSERT(nprojections==0);
    for(int i=0; i<nchannels_in; i++)
        fadd_diffc2(&out[0], act_in.get_f32(i), centers[center_idx*nchannels_in+i], ntotal);
    for(int64_t i=0; i<ntotal; i++)
        out[i] = expf(-out[i]/(rad*rad));
}

class _rbf_predictor_parallel_predict
{
public:
    const rbf_predictor &rbfpred;
    const predictions   &baseline;
    const activities    &act_in;
    predictions         &pred_out;
    std::mutex          mutex;
    _rbf_predictor_parallel_predict(const rbf_predictor &p, const predictions &bl, const activities &ai, predictions &po):
        rbfpred(p),baseline(bl),act_in(ai),pred_out(po)
    {}
};

static void _parallel_rbf_predict(void *p, int idx)
{
    _rbf_predictor_parallel_predict &env = *((_rbf_predictor_parallel_predict*)p);
    const int64_t ntotal = env.act_in.ntrn+env.act_in.ntst;

    // compute
    activityf_vector buf, func;
    env.rbfpred.compute_basis_function(env.act_in, idx, func, buf);

    // output
    env.mutex.lock();
    fadd_simd(&env.pred_out.fpredictions[0], &func[0], ntotal, env.rbfpred.coeffs[idx]);
    env.mutex.unlock();
}

predictions rbf_predictor::predict(const predictions &baseline, const activities& act, predict_t ptype) const
{
    const int64_t ntotal = act.ntrn+act.ntst;
    const float intercept = coeffs[coeffs.size()-1];
    predictions out = baseline;
    for(int64_t i=0; i<ntotal; i++)
    {
        out.fpredictions[i] = intercept;
        out.fpredictions[i+out.stride] = 0;
    }

    //
    // Parallel processing
    //
    /*activityf_vector buf, func;
    for(int64_t center_idx=0; center_idx<ncenters; center_idx++)
    {
        compute_basis_function(act, center_idx, func, buf);
        for(int64_t i=0; i<ntotal; i++)
            out.fpredictions[i] += func[i]*coeffs[center_idx];
    }*/
    _rbf_predictor_parallel_predict env(*this, baseline, act, out);
    parallel_for(_parallel_rbf_predict, &env, ncenters);

    //
    // Postprocessing
    //
    if( ptype==predict_t::prob)
    {
        for(int64_t i=0; i<ntotal; i++)
            out.fpredictions[i] = sigma0(out.fpredictions[i]);
    }
    else if( ptype==predict_t::logit )
    {
        // do nothing, logits are already here
    }
    else if( ptype==predict_t::logprob)
    {
        for(int64_t i=0; i<ntotal; i++)
        {
            float mx = fmaxf(out.fpredictions[i],0.0f);
            float emx = expf(-mx);
            out.fpredictions[i] = (out.fpredictions[i]-mx)-log(expf(out.fpredictions[i]-mx)+emx);
        }
    }
    else
        ASSERT(0);
    return out;
}

class _rbf_predictor_parallel_bf
{
public:
    const rbf_predictor &rbfpred;
    const activities& act_in;
    activities &act_bf;
    _rbf_predictor_parallel_bf(const rbf_predictor &p, const activities &ai, activities &ab):
        rbfpred(p),act_in(ai),act_bf(ab)
    {}
};

static void _parallel_bf(void *p, int idx)
{
    _rbf_predictor_parallel_bf &env = *((_rbf_predictor_parallel_bf*)p);

    // compute
    activityf_vector buf, func;
    env.rbfpred.compute_basis_function(env.act_in, idx, func, buf);

    // output
    const int64_t ntotal = env.act_in.ntrn+env.act_in.ntst;
    float *p_out = env.act_bf.get_f32(idx);
    for(int64_t i=0; i<ntotal; i++)
        p_out[i] = func[i];
}

void rbf_predictor::compute_bf(const activities& act_in, activities &act_bf) const
{
    const int64_t ntotal = act_in.ntrn+act_in.ntst;

    // prepare
    act_bf = activities(act_in.ntrn, act_in.ntst, storage_format::f32);
    act_bf.append_realm(realm<channel>::create_grid("@bf",1,1,ncenters));

    //
    /*activityf_vector buf, func;
    for(int64_t center_idx=0; center_idx<ncenters; center_idx++)
    {
        float *p_out = act_bf.get_f32(center_idx);
        compute_basis_function(act_in, center_idx, func, buf);
        for(int64_t i=0; i<ntotal; i++)
            p_out[i] = func[i];
    }*/
    _rbf_predictor_parallel_bf env(*this, act_in, act_bf);
    parallel_for(_parallel_bf, &env, ncenters);
}


void serialize(FILE *f, const rbf_predictor *ptr)
{
    write_magic8(f, 34);
    serialize(f, &ptr->nchannels_in);
    serialize(f, &ptr->prescaler);
    serialize(f, &ptr->ncenters);
    serialize(f, &ptr->centers);
    serialize(f, &ptr->nprojections);
    serialize(f, &ptr->prj);
    serialize(f, &ptr->rad);
    serialize(f, &ptr->coeffs);
    write_magic8(f, 11);
}

void unserialize(FILE *f, rbf_predictor *ptr)
{
    check_magic8(f, 34);
    unserialize(f, &ptr->nchannels_in);
    unserialize(f, &ptr->prescaler);
    unserialize(f, &ptr->ncenters);
    unserialize(f, &ptr->centers);
    unserialize(f, &ptr->nprojections);
    unserialize(f, &ptr->prj);
    unserialize(f, &ptr->rad);
    unserialize(f, &ptr->coeffs);
    check_magic8(f, 11);
}

void knn_predictor::finalize_construction()
{
    alglib::real_2d_array xy;
    alglib::integer_1d_array tags;
    xy.setlength(ncenters, nchannels_in);
    tags.setlength(ncenters);
    for(int i=0; i<ncenters; i++)
    {
        for(int j=0; j<nchannels_in; j++)
            xy[i][j] = centers[i*nchannels_in+j];
        tags[i] = labels[i];
    }
    alglib::kdtreebuildtagged(xy, tags, nchannels_in, 0, 2, tree);
}

knn_predictor knn_predictor::create(const raw_dataset &ds,
                                    int class_idx,
                                    const activities &act_in,
                                    int nk)
{
    ASSERT(act_in.is_f32());
    knn_predictor out;

    // input, no prescaler
    out.nchannels_in = act_in.nchannels;
    out.prescaler.clear();

    //
    // Centers
    //
    out.ncenters = ds.ntrn;
    out.centers.resize(out.ncenters*out.nchannels_in);
    out.labels.resize(out.ncenters);
    for(int64_t i=0; i<ds.ntrn; i++)
    {
        const float   *p_src = act_in.get_f32()+i;
        const int64_t stride = act_in.get_stride();
        for(int64_t j=0; j<out.nchannels_in; j++)
            out.centers[i*out.nchannels_in+j] = p_src[j*stride];
        out.labels[i] = ds.entries[i].class_id==class_idx ? 0 : 1;
    }

    //
    // Params
    //
    out.nneighbors = nk;

    //
    // done
    //
    out.finalize_construction();
    return out;
}

class _knn_predictor_parallel_predict
{
public:
    const knn_predictor &knnpred;
    const predictions   &baseline;
    const activities    &act_in;
    predictions         &pred_out;
    _knn_predictor_parallel_predict(const knn_predictor &p, const predictions &bl, const activities &ai, predictions &po):
        knnpred(p),baseline(bl),act_in(ai),pred_out(po)
    {}
};

static void _parallel_knn_predict(void *p, int img_idx)
{
    _knn_predictor_parallel_predict &env = *((_knn_predictor_parallel_predict*)p);
    const int64_t ntotal = env.act_in.ntrn+env.act_in.ntst;
    const float *p_src = env.act_in.get_f32()+img_idx;
    const int64_t stride = env.act_in.get_stride();

    // compute
    alglib::kdtreerequestbuffer buf;
    alglib::real_1d_array x;
    alglib::integer_1d_array tags;
    x.setlength(env.knnpred.nchannels_in);
    for(int i=0; i<env.knnpred.nchannels_in; i++)
        x[i] = p_src[stride*i];
    alglib::kdtreecreaterequestbuffer(env.knnpred.tree, buf);
    int nn = alglib::kdtreetsqueryknn(env.knnpred.tree, buf, x, env.knnpred.nneighbors);
    //int nn = alglib::kdtreetsqueryaknn(env.knnpred.tree, buf, x, env.knnpred.nneighbors, 0.5);
    alglib::kdtreetsqueryresultstags(env.knnpred.tree, buf, tags);
    float prob = 0.0f;
    for(int i=0; i<nn; i++)
        if( tags[i]==0 )
            prob += 1;
    prob /= nn+1.0e-20f;

    // output
    env.pred_out.fpredictions[img_idx] = prob;
    env.pred_out.fpredictions[img_idx+env.pred_out.stride] = 1.0f-prob;
}

predictions knn_predictor::predict(const predictions &baseline, const activities& act, predict_t ptype) const
{
    ASSERT(ptype==predict_t::prob);

    const int64_t ntotal = act.ntrn+act.ntst;
    predictions out = baseline;
    _knn_predictor_parallel_predict env(*this, baseline, act, out);
    parallel_for(_parallel_knn_predict, &env, ntotal);

    //
    // Postprocessing
    //
    if( ptype==predict_t::prob)
    {
        // do nothing, probabilities are already here
    }
    else if( ptype==predict_t::logit )
    {
        ASSERT(0);
    }
    else if( ptype==predict_t::logprob)
    {
        ASSERT(0);
    }
    else
        ASSERT(0);
    return out;

}

void serialize(FILE *f, const knn_predictor *ptr)
{
    write_magic8(f, 44);
    serialize(f, &ptr->nchannels_in);
    serialize(f, &ptr->prescaler);
    serialize(f, &ptr->ncenters);
    serialize(f, &ptr->centers);
    serialize(f, &ptr->labels);
    serialize(f, &ptr->nneighbors);
    write_magic8(f, 82);
}

void unserialize(FILE *f, knn_predictor *ptr)
{
    *ptr = knn_predictor();
    check_magic8(f, 44);
    unserialize(f, &ptr->nchannels_in);
    unserialize(f, &ptr->prescaler);
    unserialize(f, &ptr->ncenters);
    unserialize(f, &ptr->centers);
    unserialize(f, &ptr->labels);
    unserialize(f, &ptr->nneighbors);
    check_magic8(f, 82);
    ptr->finalize_construction();
}

float superlogit2_predictor::get_l1_factor() const
{
    float r = 0.0f;
    for(const std::vector<float> &v: coeffs)
    {
        ASSERT(v.size()%(nchannels_in+1)==0);
        ASSERT(v.size()/(nchannels_in+1)>=1);
        for(int k=0; k<v.size()/(nchannels_in+1); k++)
        {
            float s = 0.0f;
            const float *p = &v[k*(nchannels_in+1)];
            for(int j=0; j<nchannels_in+1; j++)
                s += fabs(p[j]);
            r = fmaxf(r, s);
        }
    }
    return r;
}

superlogit2_predictor superlogit2_predictor::create_empty(const activities &act_in)
{
    superlogit2_predictor r;
    r.nchannels_in = act_in.nchannels;
    return r;
}

void superlogit2_predictor::append1(const std::vector<float> &subpredictor)
{
    ASSERT(subpredictor.size()==nchannels_in+1);
    coeffs.push_back(subpredictor);
}

void superlogit2_predictor::append1(const linear_predictor &subpredictor)
{
    ASSERT(subpredictor.get_logit_flag() && subpredictor.get_nclasses()==2 && subpredictor.get_nchannels()==nchannels_in);
    std::vector<float> cc;
    for(int i=0; i<nchannels_in+1; i++)
        cc.push_back(subpredictor.get_coeff(i,0));
    coeffs.push_back(cc);
}

void superlogit2_predictor::appendk(const std::vector<float> &subpredictor)
{
    ASSERT(subpredictor.size()%(nchannels_in+1)==0);
    ASSERT(subpredictor.size()/(nchannels_in+1)>=1);
    coeffs.push_back(subpredictor);
}

predictions superlogit2_predictor::predict(const predictions &baseline, const activities& act, predict_t ptype) const
{
    // integrity checks
    ASSERT(act.nchannels==nchannels_in);
    ASSERT(baseline.nclasses==2);
    ASSERT(baseline.ntrn==act.ntrn);
    ASSERT(baseline.ntst==act.ntst);

    //
    int64_t ntotal = act.ntrn+act.ntst;

    //
    predictions r = baseline;
    std::vector<float> tmp0(ntotal), tmp1(ntotal);
    for(int64_t i=0; i<ntotal; i++)
        r.fpredictions[i] = -1.0e20f;

    // process all partial models
    for(const std::vector<float> &xv: coeffs)
    {
        ASSERT(xv.size()%(nchannels_in+1)==0);
        ASSERT(xv.size()/(nchannels_in+1)>=1);
        const int xk = xv.size()/(nchannels_in+1);
        for(int64_t i=0; i<ntotal; i++)
            tmp1[i] = +1.0e20f;
        for(int64_t xx=0; xx<xk; xx++)
        {
            const float *pv = &xv[xx*(nchannels_in+1)];
            for(int64_t i=0; i<ntotal; i++)
                tmp0[i] = pv[nchannels_in];
            for(int64_t k=0; k<nchannels_in; k++)
            {
                if( act.is_i8() )
                {
                    const float  v = pv[k]*SNET_I2F;
                    const int8_t *p_src = act.get_i8(k);
                    for(int64_t i=0; i<ntotal; i++)
                        tmp0[i] += v*p_src[i];
                }
                else if( act.is_f32() )
                {
                    const float  v = pv[k];
                    const float *p_src = act.get_f32(k);
                    for(int64_t i=0; i<ntotal; i++)
                        tmp0[i] += v*p_src[i];
                }
                else
                    ASSERT(0);
            }
            for(int64_t i=0; i<ntotal; i++)
                tmp1[i] = fminf(tmp1[i], tmp0[i]);
        }
        for(int64_t i=0; i<ntotal; i++)
            r.fpredictions[i] = fmaxf(r.fpredictions[i], tmp1[i]);
    }

    // ptype-specific processing
    if( ptype==predict_t::logprob )
    {
        for(int64_t i=0; i<ntotal; i++)
        {
            float mx = fmaxf(r.fpredictions[i],0);
            float v  = r.fpredictions[i]-mx;
            float lsum = logf(expf(v)+expf(-mx));
            r.fpredictions[i] = v-lsum;
            r.fpredictions[i+r.stride] = -mx-lsum;
        }
    }
    else if( ptype==predict_t::prob )
    {
        for(int64_t i=0; i<ntotal; i++)
        {
            r.fpredictions[i] = sigma0(r.fpredictions[i]);
            r.fpredictions[i+r.stride] = 1.0f-r.fpredictions[i];
        }
    }
    else if( ptype==predict_t::logit )
    {
        // logits are already here
    }
    else
        ASSERT(0);
    return r;
}

conv_predictions superlogit2_predictor::conv_predict1(const conv_activities &act, const calibrator &cal) const
{
    // integrity checks
    int ch_total = 0;
    for(const conv_activities::conv_realm &rr: act.realms)
        ch_total += rr.grid_depth;
    ASSERT(ch_total==nchannels_in);
    ASSERT(act.realms.size()>=1);
    int64_t dim0 = act.realms[0].grid_dim0;
    int64_t dim1 = act.realms[0].grid_dim1;
    int64_t out_total = dim0*dim1;

    //
    conv_predictions r_pred = conv_predictions(dim0,dim1,1);
    float *p_pred = r_pred.get_ptr();
    std::vector<float> tmp0(out_total), tmp1(out_total);
    for(int64_t i=0; i<out_total; i++)
        p_pred[i] = -1.0e20f;

    // process all partial models
    for(const std::vector<float> &xv: coeffs)
    {
        ASSERT(xv.size()%(nchannels_in+1)==0);
        ASSERT(xv.size()/(nchannels_in+1)>=1);
        const int xk = xv.size()/(nchannels_in+1);
        for(int64_t i=0; i<out_total; i++)
            tmp1[i] = +1.0e20f;
        for(int64_t xx=0; xx<xk; xx++)
        {
            const float *pv = &xv[xx*(nchannels_in+1)];
            for(int64_t i=0; i<out_total; i++)
                tmp0[i] = pv[nchannels_in];
            int cidx = 0;
            for(const conv_activities::conv_realm &rr: act.realms)
            {
                ASSERT(rr.grid_dim0==dim0 && rr.grid_dim1==dim1);
                ASSERT(rr.window_dim0==1 && rr.window_dim1==1);
                for(int plane_idx=0; plane_idx<rr.grid_depth; plane_idx++)
                {
                    if( pv[cidx]!=0 )
                        fadd_simd(&tmp0[0], act.get_f32(rr.get_offs(plane_idx,0,0)), out_total, pv[cidx]);
                    cidx++;
                }
            }
            ASSERT(cidx==nchannels_in);
            fmerge_min(&tmp1[0], &tmp0[0], out_total);
        }
        fmerge_max(p_pred, &tmp1[0], out_total);
    }
    for(int64_t i=0; i<out_total; i++)
        p_pred[i] = sigma0(p_pred[i]);
    if( !cal.is_identity() )
    {
        for(int64_t i=0; i<out_total; i++)
            p_pred[i] = cal.calibrate(p_pred[i]);
    }
    return r_pred;
}

void serialize(FILE *f, const superlogit2_predictor *ptr)
{
    write_magic8(f, 52);
    serialize(f, &ptr->nchannels_in);
    serialize(f, &ptr->coeffs);
    write_magic8(f, 94);
}

void unserialize(FILE *f, superlogit2_predictor *ptr)
{
    check_magic8(f, 52);
    unserialize(f, &ptr->nchannels_in);
    unserialize(f, &ptr->coeffs);
    check_magic8(f, 94);
}

conv_subcolumn::conv_subcolumn(std::string filename)
{
    unserialize_from(filename, this);
}

void conv_subcolumn::process_v2v4(activities &_v1_act, activities &act_total) const
{
    const bool has_both = nonconv_selector!=select_t::none && conv_selector!=select_t::none;
    act_total = activities(_v1_act.ntrn, _v1_act.ntst, _v1_act.get_format());

    //
    // Nonconvolutional pathway is computed in i8
    //
    if( nonconv_selector!=select_t::none )
    {
        activities v1_view_nc, act_v2nc, act_v4nc;
        _v1_act.select_view(nonconv_selector, nonconv_select0, nonconv_select1, v1_view_nc);
        if( has_both)
            v1_view_nc.add_prefix("[N]");
        nonconv_v2.apply(v1_view_nc, act_v2nc);
        protect_activities(act_v2nc);
        act_total.merge_in(act_v2nc);
        if( nonconv_v4.ndetectors>0 )
        {
            nonconv_v4.apply(act_v2nc, act_v4nc);
            protect_activities(act_v4nc);
            act_total.merge_in(act_v4nc);
        }
    }

    //
    // Convert to fp32
    //
    act_total.convert_to_f32();

    //
    // Convolutional pathway
    //
    if( conv_selector!=select_t::none )
    {
        ASSERT(conv_v2c.ndetectors>0);
        activities v1_view_conv, act_v2c, act_v4c;
        _v1_act.select_view(conv_selector, conv_select0, conv_select1, v1_view_conv);
        v1_view_conv.subdivide(conv_subdivide_in, conv_overlap);
        if( has_both)
            v1_view_conv.add_prefix("[C]");
        conv_v2c.apply(v1_view_conv, conv_subdivide_out, conv_accumulate_out, act_v2c);
        protect_activities(act_v2c);
        act_total.merge_in(act_v2c);
        if( conv_v4.ndetectors>0 )
        {
            conv_v4.apply(act_v2c, act_v4c);
            protect_activities(act_v4c);
            act_total.merge_in(act_v4c);
        }
    }

    //
    // Done, integrity check
    //
    ASSERT(act_total.is_f32());
}

void conv_subcolumn::process_v2v4(conv_activities &_v1_act, conv_activities &act_total) const
{
    const bool has_both = nonconv_selector!=select_t::none && conv_selector!=select_t::none;
    act_total = conv_activities(_v1_act.get_img_height(), _v1_act.get_img_width(), _v1_act.get_format());

    //
    // Nonconvolutional pathway is computed in i8
    //
    if( nonconv_selector!=select_t::none )
    {
        conv_activities v1_view_nc, act_v2nc, act_v4nc;
        _v1_act.select_view(nonconv_selector, nonconv_select0, nonconv_select1, v1_view_nc);
        if( has_both)
            v1_view_nc.add_prefix("[N]");
        nonconv_v2.conv_apply(v1_view_nc, act_v2nc);
        protect_activities(act_v2nc);
        act_total.merge_in(act_v2nc);
        if( nonconv_v4.ndetectors>0 )
        {
            nonconv_v4.conv_apply(act_v2nc, act_v4nc);
            protect_activities(act_v4nc);
            act_total.merge_in(act_v4nc);
        }
    }

    //
    // Convert to fp32
    //
    act_total.convert_to_f32();

    //
    // Convolutional pathway
    //
    if( conv_selector!=select_t::none )
    {
        ASSERT(conv_selector==select_t::none);
#if 0
        ASSERT(conv_v2c.ndetectors>0);
        activities v1_view_conv, act_v2c, act_v4c;
        _v1_act.select_view(conv_selector, conv_select0, conv_select1, v1_view_conv);
        v1_view_conv.subdivide(conv_subdivide_in, conv_overlap);
        if( has_both)
            v1_view_conv.add_prefix("[C]");
        conv_v2c.apply(v1_view_conv, conv_subdivide_out, conv_accumulate_out, act_v2c);
        protect_activities(act_v2c);
        act_total.merge_in(act_v2c);
        if( conv_v4.ndetectors>0 )
        {
            conv_v4.apply(act_v2c, act_v4c);
            protect_activities(act_v4c);
            act_total.merge_in(act_v4c);
        }
#endif
    }

    //
    // Done, integrity check
    //
    ASSERT(act_total.is_f32());
}

void conv_subcolumn::protect_activities(activities &act) const
{
    protect_activities(act, std::make_tuple(adv_clip_level,adv_frac_chosen));
}

void conv_subcolumn::protect_activities(conv_activities &act) const
{
    ASSERT(adv_frac_chosen==0);
    protect_activities(act, adv_clip_level);
}

void conv_subcolumn::protect_activities(activities &act, const std::tuple<float,float> protection_coeffs)
{
    const float clip_level  = std::get<0>(protection_coeffs);
    const float frac_chosen = std::get<1>(protection_coeffs);
    const int64_t ntotal = act.ntrn+act.ntst;

    //
    // Integrity check
    //
    ASSERT(clip_level>=0 && clip_level<=1);
    ASSERT(frac_chosen>=0 && frac_chosen<=1);
    int _cnt0 = 0;
    for(const realm<channel> &rr: act.realms)
        _cnt0 += rr.grid_depth*rr.grid_height*rr.grid_width;
    ASSERT(act.nchannels==_cnt0); // contiguous storage required

    //
    // Protection is inactive
    //
    if( clip_level==0.0f && (frac_chosen==0 || frac_chosen==1) )
        return;

    //
    // Compute per-item clipping level
    //
    activityf_vector threshold(ntotal,0.0f);
    if( clip_level>0.0f )
    {
        if( act.is_i8() )
        {
            activity8_vector actmax(ntotal,0);
            for(int64_t k=0; k<act.nchannels; k++)
                act8_merge_max_simd(&actmax[0], act.get_i8(k), ntotal);
            for(int64_t i=0; i<ntotal; i++)
                threshold[i] = SNET_I2F*actmax[i];
        }
        else if( act.is_f32() )
        {
            for(int64_t k=0; k<act.nchannels; k++)
                fmerge_max(&threshold[0], act.get_f32(k), ntotal);
        }
        else
            ASSERT(0);
        for(int64_t i=0; i<ntotal; i++)
            threshold[i] *= clip_level;
    }


    //
    // Compute fraction-based protection
    //
    if( frac_chosen>0 && frac_chosen<1 ) // 0 and 1 are special values both meaning absence of defense
    {
        std::vector<float> buf(act.nchannels);
        for(int64_t i=0; i<ntotal; i++)
        {
            int64_t stride = act.get_stride();
            if( act.is_i8() )
            {
                const int8_t *p_act = act.get_i8()+i;
                for(int64_t k=0; k<act.nchannels; k++)
                    buf[k] = SNET_I2F*p_act[k*stride];
            }
            else if( act.is_f32() )
            {
                const float *p_act = act.get_f32()+i;
                for(int64_t k=0; k<act.nchannels; k++)
                    buf[k] = p_act[k*stride];
            }
            else
                ASSERT(0);
            int elem_idx = irange(roundf((1.0f-frac_chosen)*(buf.size()-1)), 0, buf.size()-1);
            std::nth_element(buf.begin(), buf.begin()+elem_idx, buf.end());
            threshold[i] = fmaxf(threshold[i],buf[elem_idx]);
        }
    }

    //
    // Apply protection
    //
    if( act.is_i8() )
    {
        for(int64_t k=0; k<act.nchannels; k++)
        {
            const float s = 1.0f/SNET_I2F;
            int8_t *p_act = act.get_i8(k);
            for(int64_t i=0; i<ntotal; i++)
            {
                int level=roundf(threshold[i]*s);
                int v = p_act[i];
                v = v>=level ? v : 0;
                p_act[i] = v;
            }
        }
    }
    else if( act.is_f32() )
    {
        for(int64_t k=0; k<act.nchannels; k++)
        {
            float *p_act = act.get_f32(k);
            for(int64_t i=0; i<ntotal; i++)
            {
                float v = p_act[i];
                v = v>=threshold[i] ? v : 0.0f;
                p_act[i] = v;
            }
        }
    }
    else
        ASSERT(0);
}

void conv_subcolumn::protect_activities(conv_activities &cact, float clip_level)
{
    ASSERT(clip_level==0);
}

float conv_subcolumn::get_l1_factor() const
{
    if( predictor_type==0 )
        return linpred.get_l1_factor();
    if( predictor_type==3 )
        return superpred.get_l1_factor();
    return 1.0f;
}

float conv_subcolumn::get_sparse_flops() const
{
    ASSERT(conv_selector==select_t::none);
    ASSERT(nonconv_selector!=select_t::none);
    float r = 0;
    r += nonconv_v2.get_flop_count();
    r += nonconv_v4.get_flop_count();
    return r;
}

int conv_subcolumn::get_v1_window_dim0(const net_v1layer &v1) const
{
    return v1.get_img_height();
}

int conv_subcolumn::get_v1_window_dim1(const net_v1layer &v1) const
{
    return v1.get_img_width();
}

int conv_subcolumn::get_v1_downsample_lo(const net_v1layer &v1) const
{
    ASSERT(conv_selector==select_t::none);
    ASSERT(nonconv_selector==select_t::not_larger);
    return v1.get_ds_lo(nonconv_selector, nonconv_select0, nonconv_select1);
}

int conv_subcolumn::get_v1_downsample_hi(const net_v1layer &v1) const
{
    ASSERT(conv_selector==select_t::none);
    ASSERT(nonconv_selector==select_t::not_larger);
    return v1.get_ds_hi(nonconv_selector, nonconv_select0, nonconv_select1);
}

int conv_subcolumn::get_v2_window_height(const net_v1layer &v1) const
{
    ASSERT(conv_selector==select_t::none);
    ASSERT(nonconv_selector==select_t::not_larger);
    return v1.get_max_height(nonconv_selector, nonconv_select0, nonconv_select1);
}

int conv_subcolumn::get_v2_window_width(const net_v1layer &v1) const
{
    ASSERT(conv_selector==select_t::none);
    ASSERT(nonconv_selector==select_t::not_larger);
    return v1.get_max_width(nonconv_selector, nonconv_select0, nonconv_select1);
}

predictions conv_subcolumn::apply(const predictions &zero_baseline, activities &_v1_act) const
{
    ASSERT(zero_baseline.nclasses==2);

    //
    // Process middle part of the subcolumn
    //
    activities act_total;
    process_v2v4(_v1_act, act_total);

    //
    // Combine and predict
    //
    ASSERT(act_total.is_f32());
    if( predictor_type==0 )
        return predictions::calibrate_onevsall(linpred.predict(zero_baseline, {&act_total}), cal);
    if( predictor_type==1 )
        return predictions::calibrate_onevsall(rbfpred.predict(zero_baseline, act_total), cal);
    if( predictor_type==2 )
        return predictions::calibrate_onevsall(knnpred.predict(zero_baseline, act_total), cal);
    if( predictor_type==3 )
        return predictions::calibrate_onevsall(superpred.predict(zero_baseline, act_total), cal);
    ASSERT(0);
}

predictions conv_subcolumn::apply_log(const predictions &zero_baseline, activities &_v1_act) const
{
    ASSERT(zero_baseline.nclasses==2);

    //
    // Process middle part of the subcolumn
    //
    activities act_total;
    process_v2v4(_v1_act, act_total);

    //
    // Combine and predict
    //
    ASSERT(act_total.is_f32());
    if( predictor_type==0 )
        return linpred.predict_log(zero_baseline, {&act_total});
    if( predictor_type==1 )
        return rbfpred.predict(zero_baseline, act_total, predict_t::logprob);
    if( predictor_type==3 )
        return superpred.predict(zero_baseline, act_total, predict_t::logprob);
    ASSERT(0);
}

predictions conv_subcolumn::apply_logit(const predictions &zero_baseline, activities &_v1_act) const
{
    ASSERT(zero_baseline.nclasses==2);

    //
    // Process middle part of the subcolumn
    //
    activities act_total;
    process_v2v4(_v1_act, act_total);

    //
    // Combine and predict
    //
    ASSERT(act_total.is_f32());
    if( predictor_type==0 )
        return linpred.predict_logit(zero_baseline, {&act_total});
    if( predictor_type==3 )
        return superpred.predict(zero_baseline, act_total, predict_t::logit);
    ASSERT(0);
}

conv_predictions conv_subcolumn::conv_apply0(const net_v1layer &v1, conv_activities &_v1_act) const
{
    //
    // Process middle part of the subcolumn
    //
    conv_activities act_total;
    process_v2v4(_v1_act, act_total);

    //
    // Predict
    //
    conv_predictions t;
    ASSERT(act_total.is_f32());
    if( predictor_type==0 )
        t = linpred.conv_predict1(act_total, cal);
    if( predictor_type==3 )
        t = superpred.conv_predict1(act_total, cal);

    //
    // Convert to output dimensions
    //
    int img_dim0 = _v1_act.get_img_height(), img_dim1 = _v1_act.get_img_width();
    int v1_ds = get_v1_downsample_lo(v1);
    ASSERT(img_dim0%v1_ds==0 && img_dim1%v1_ds==0);
    int windim0 = get_v2_window_height(v1), windim1 = get_v2_window_width(v1);
    int pred_dim0 = img_dim0/v1_ds-windim0+1, pred_dim1 = img_dim1/v1_ds-windim1+1;
    conv_predictions r(std::make_tuple(img_dim0,img_dim1), v1_ds, std::make_tuple(pred_dim0,pred_dim1), std::make_tuple(0,0), 1);
    ASSERT(r.get_pred_height()==t.get_pred_height() && r.get_pred_width()==t.get_pred_width());
    int ph = r.get_pred_height(), pw = r.get_pred_width();
    for(int i0=0; i0<ph; i0++)
        for(int i1=0; i1<pw; i1++)
            r.set_pred(i0, i1, 0, t.get_pred(i0,i1)); //r.set_pred(i0, i1, 0, 0.5f*((i0+i1)%2));
    return r;
}

void conv_subcolumn::init(split_t split_in, int overlap, const conv_midlayer &v2c, split_t split_out, accumulate_t acc_out,
                          net_midlayer v4,
                          const linear_predictor &pred, const calibrator &c)
{
    ASSERT(pred.get_nclasses()==2);

    // init convolutional pathway
    conv_selector = select_t::all;
    conv_select0 = 0;
    conv_select1 = 0;
    conv_subdivide_in = split_in;
    conv_overlap = overlap;
    conv_v2c = v2c;
    conv_subdivide_out = split_out;
    conv_accumulate_out = acc_out;
    conv_v4 = v4;

    // init nonconvolutional pathway
    nonconv_selector = select_t::none;
    nonconv_select0 = 0;
    nonconv_select1 = 0;
    nonconv_v2 = net_midlayer();
    nonconv_v4 = net_midlayer();

    // protection
    adv_clip_level = 0;
    adv_frac_chosen = 0;
    adv_param0 =  adv_param1 =  adv_param2 = adv_param3 = adv_param4 = adv_param5 = 0;

    // init predictor/calibrator
    predictor_type = 0;
    linpred = pred;
    cal = c;
}

void conv_subcolumn::init(int window_size,
                          select_t select_nonconv,
                          std::tuple<net_midlayer,net_midlayer> nonconv, // v2,v4
                          select_t select_conv,
                          std::tuple<split_t,int,conv_midlayer,split_t,accumulate_t,net_midlayer> conv, // <split_in,overlap,v2c,split_out,acc_out,v4c>
                          const std::tuple<float,float> protection_coeffs, // clip_lo, clip_hi
                          const linear_predictor &pred, const calibrator &c)
{
    ASSERT(pred.get_nclasses()==2);

    // init nonconvolutional pathway
    nonconv_selector = select_nonconv;
    if( select_nonconv!=select_t::none )
    {
        nonconv_select0 = window_size;
        nonconv_select1 = window_size;
        nonconv_v2      = std::get<0>(nonconv);
        nonconv_v4      = std::get<1>(nonconv);
    }
    else
    {
        nonconv_select0 = 0;
        nonconv_select1 = 0;
        nonconv_v2      = net_midlayer();
        nonconv_v4      = net_midlayer();
    }

    // init convolutional pathway
    conv_selector = select_conv;
    if( select_conv!=select_t::none )
    {
        conv_select0        = window_size;
        conv_select1        = window_size;
        conv_subdivide_in   = std::get<0>(conv);
        conv_overlap        = std::get<1>(conv);
        conv_v2c            = std::get<2>(conv);
        conv_subdivide_out  = std::get<3>(conv);
        conv_accumulate_out = std::get<4>(conv);
        conv_v4             = std::get<5>(conv);
    }
    else
    {
        conv_select0        = 0;
        conv_select1        = 0;
        conv_v2c            = conv_midlayer();
        conv_v4             = net_midlayer();
    }

    // protection
    adv_clip_level  = std::get<0>(protection_coeffs);
    adv_frac_chosen = std::get<1>(protection_coeffs);
    adv_param0 =  adv_param1 =  adv_param2 = adv_param3 = adv_param4 = adv_param5 = 0;

    // init predictor/calibrator
    predictor_type = 0;
    linpred = pred;
    cal = c;
}

void conv_subcolumn::init(int window_size,
                          select_t select_nonconv,
                          std::tuple<net_midlayer,net_midlayer> nonconv, // v2,v4
                          select_t select_conv,
                          std::tuple<split_t,int,conv_midlayer,split_t,accumulate_t,net_midlayer> conv, // <split_in,overlap,v2c,split_out,acc_out,v4c>
                          const rbf_predictor &pred, const calibrator &c)
{
    // init nonconvolutional pathway
    nonconv_selector = select_nonconv;
    if( select_nonconv!=select_t::none )
    {
        nonconv_select0 = window_size;
        nonconv_select1 = window_size;
        nonconv_v2      = std::get<0>(nonconv);
        nonconv_v4      = std::get<1>(nonconv);
    }
    else
    {
        nonconv_select0 = 0;
        nonconv_select1 = 0;
        nonconv_v2      = net_midlayer();
        nonconv_v4      = net_midlayer();
    }

    // init convolutional pathway
    conv_selector = select_conv;
    if( select_conv!=select_t::none )
    {
        conv_select0        = window_size;
        conv_select1        = window_size;
        conv_subdivide_in   = std::get<0>(conv);
        conv_overlap        = std::get<1>(conv);
        conv_v2c            = std::get<2>(conv);
        conv_subdivide_out  = std::get<3>(conv);
        conv_accumulate_out = std::get<4>(conv);
        conv_v4             = std::get<5>(conv);
    }
    else
    {
        conv_select0        = 0;
        conv_select1        = 0;
        conv_v2c            = conv_midlayer();
        conv_v4             = net_midlayer();

    }

    // protection
    adv_clip_level = 0;
    adv_frac_chosen = 0;
    adv_param0 =  adv_param1 =  adv_param2 = adv_param3 = adv_param4 = adv_param5 = 0;

    // init predictor/calibrator
    predictor_type = 1;
    rbfpred = pred;
    cal = c;
}

void conv_subcolumn::init(int window_size,
                          select_t select_nonconv,
                          std::tuple<net_midlayer,net_midlayer> nonconv, // v2,v4
                          select_t select_conv,
                          std::tuple<split_t,int,conv_midlayer,split_t,accumulate_t,net_midlayer> conv, // <split_in,overlap,v2c,split_out,acc_out,v4c>
                          const knn_predictor &pred, const calibrator &c)
{
    // init nonconvolutional pathway
    nonconv_selector = select_nonconv;
    if( select_nonconv!=select_t::none )
    {
        nonconv_select0 = window_size;
        nonconv_select1 = window_size;
        nonconv_v2      = std::get<0>(nonconv);
        nonconv_v4      = std::get<1>(nonconv);
    }
    else
    {
        nonconv_select0 = 0;
        nonconv_select1 = 0;
        nonconv_v2      = net_midlayer();
        nonconv_v4      = net_midlayer();
    }

    // init convolutional pathway
    conv_selector = select_conv;
    if( select_conv!=select_t::none )
    {
        conv_select0        = window_size;
        conv_select1        = window_size;
        conv_subdivide_in   = std::get<0>(conv);
        conv_overlap        = std::get<1>(conv);
        conv_v2c            = std::get<2>(conv);
        conv_subdivide_out  = std::get<3>(conv);
        conv_accumulate_out = std::get<4>(conv);
        conv_v4             = std::get<5>(conv);
    }
    else
    {
        conv_select0        = 0;
        conv_select1        = 0;
        conv_v2c            = conv_midlayer();
        conv_v4             = net_midlayer();

    }

    // protection
    adv_clip_level = 0;
    adv_frac_chosen = 0;
    adv_param0 =  adv_param1 =  adv_param2 = adv_param3 = adv_param4 = adv_param5 = 0;

    // init predictor/calibrator
    predictor_type = 2;
    knnpred = pred;
    cal = c;
}

void conv_subcolumn::init_partial(
                          select_t select_nonconv,
                          std::tuple<int,net_midlayer,net_midlayer> nonconv, // windowsize, v2,v4
                          select_t select_conv,
                          std::tuple<int,split_t,int,conv_midlayer,split_t,accumulate_t,net_midlayer> conv) // <split_in,overlap,v2c,split_out,acc_out,v4c>
{
    // init nonconvolutional pathway
    nonconv_selector = select_nonconv;
    if( select_nonconv!=select_t::none )
    {
        nonconv_select0 = std::get<0>(nonconv);
        nonconv_select1 = std::get<0>(nonconv);
        nonconv_v2      = std::get<1>(nonconv);
        nonconv_v4      = std::get<2>(nonconv);
    }
    else
    {
        nonconv_select0 = 0;
        nonconv_select1 = 0;
        nonconv_v2      = net_midlayer();
        nonconv_v4      = net_midlayer();
    }

    // init convolutional pathway
    conv_selector = select_conv;
    if( select_conv!=select_t::none )
    {
        conv_select0        = std::get<0>(conv);
        conv_select1        = std::get<0>(conv);
        conv_subdivide_in   = std::get<1>(conv);
        conv_overlap        = std::get<2>(conv);
        conv_v2c            = std::get<3>(conv);
        conv_subdivide_out  = std::get<4>(conv);
        conv_accumulate_out = std::get<5>(conv);
        conv_v4             = std::get<6>(conv);
    }
    else
    {
        conv_select0        = 0;
        conv_select1        = 0;
        conv_v2c            = conv_midlayer();
        conv_v4             = net_midlayer();

    }

    // protection
    adv_clip_level = 0;
    adv_frac_chosen = 0;
    adv_param0 =  adv_param1 =  adv_param2 = adv_param3 = adv_param4 = adv_param5 = 0;

    // init predictor/calibrator
    predictor_type = -1;
    cal = calibrator();
}

void conv_subcolumn::rewrite_predictor(const linear_predictor &pred, const calibrator &c)
{
    ASSERT(pred.get_nclasses()==2);
    predictor_type = 0;
    linpred = pred;
    cal = c;
}

void conv_subcolumn::rewrite_predictor(const rbf_predictor &pred, const calibrator &c)
{
    predictor_type = 1;
    rbfpred = pred;
    cal = c;
}

void conv_subcolumn::rewrite_predictor(const knn_predictor &pred, const calibrator &c)
{
    predictor_type = 2;
    knnpred = pred;
    cal = c;
}

void conv_subcolumn::rewrite_predictor(const superlogit2_predictor &pred, const calibrator &c)
{
    predictor_type = 3;
    superpred = pred;
    cal = c;
}

void conv_subcolumn::set_clipping(float level)
{
    ASSERT(level>=0 && level<=1);
    adv_clip_level = level;
}

const linear_predictor& conv_subcolumn::get_predictor() const
{
    ASSERT(predictor_type==0);
    return linpred;
}

const rbf_predictor& conv_subcolumn::get_rbf_predictor() const
{
    ASSERT(predictor_type==1);
    return rbfpred;
}

const knn_predictor& conv_subcolumn::get_knn_predictor() const
{
    ASSERT(predictor_type==2);
    return knnpred;
}

const superlogit2_predictor& conv_subcolumn::get_superlogit2_predictor() const
{
    ASSERT(predictor_type==3);
    return superpred;
}

static const int8_t conv_subcolumn_version0 = 65;
static const int8_t conv_subcolumn_version1 = 84;

void serialize(FILE *f, const conv_subcolumn *ptr)
{
    write_magic8(f, conv_subcolumn_version1);
    serialize_as_int64(f, &ptr->conv_selector);
    if( ptr->conv_selector!=select_t::none )
    {
        serialize(f, &ptr->conv_select0);
        serialize(f, &ptr->conv_select1);
        serialize_as_int64(f, &ptr->conv_subdivide_in);
        serialize(f, &ptr->conv_overlap);
        serialize(f, &ptr->conv_v2c);
        serialize_as_int64(f, &ptr->conv_subdivide_out);
        serialize_as_int64(f, &ptr->conv_accumulate_out);
        serialize(f, &ptr->conv_v4);
    }
    serialize_as_int64(f, &ptr->nonconv_selector);
    if( ptr->nonconv_selector!=select_t::none )
    {
        serialize(f, &ptr->nonconv_select0);
        serialize(f, &ptr->nonconv_select1);
        serialize(f, &ptr->nonconv_v2);
        serialize(f, &ptr->nonconv_v4);
    }
    serialize(f, &ptr->predictor_type);
    if( ptr->predictor_type==-1 )
    {
        // do nothing
    }
    else if( ptr->predictor_type==0 )
        serialize(f, &ptr->linpred);
    else if( ptr->predictor_type==1 )
        serialize(f, &ptr->rbfpred);
    else if( ptr->predictor_type==2 )
        serialize(f, &ptr->knnpred);
    else if( ptr->predictor_type==3 )
        serialize(f, &ptr->superpred);
    else
        ASSERT(0);
    serialize(f, &ptr->cal);

    // adversarial part
    write_magic8(f, 1);
    serialize(f, &ptr->adv_clip_level);
    serialize(f, &ptr->adv_frac_chosen);
    serialize(f, &ptr->adv_param0);
    serialize(f, &ptr->adv_param1);
    serialize(f, &ptr->adv_param2);
    serialize(f, &ptr->adv_param3);
    serialize(f, &ptr->adv_param4);
    serialize(f, &ptr->adv_param5);

    // done
    write_magic8(f, 111);
}

void unserialize(FILE *f, conv_subcolumn *ptr)
{
    *ptr = conv_subcolumn();
    int8_t version = check_magic8(f, conv_subcolumn_version0, conv_subcolumn_version1);
    unserialize_as_int64(f, &ptr->conv_selector);
    if( ptr->conv_selector!=select_t::none )
    {
        unserialize(f, &ptr->conv_select0);
        unserialize(f, &ptr->conv_select1);
        unserialize_as_int64(f, &ptr->conv_subdivide_in);
        unserialize(f, &ptr->conv_overlap);
        unserialize(f, &ptr->conv_v2c);
        unserialize_as_int64(f, &ptr->conv_subdivide_out);
        unserialize_as_int64(f, &ptr->conv_accumulate_out);
        unserialize(f, &ptr->conv_v4);
    }
    unserialize_as_int64(f, &ptr->nonconv_selector);
    if( ptr->nonconv_selector!=select_t::none )
    {
        unserialize(f, &ptr->nonconv_select0);
        unserialize(f, &ptr->nonconv_select1);
        unserialize(f, &ptr->nonconv_v2);
        unserialize(f, &ptr->nonconv_v4);
    }
    if( version==conv_subcolumn_version0 )
    {
        // older storage format
        ptr->predictor_type = 0;
        unserialize(f, &ptr->linpred);
    }
    else if( version==conv_subcolumn_version1 )
    {
        // current storage format
        unserialize(f, &ptr->predictor_type);
        if( ptr->predictor_type==-1 )
        {
            // do nothing
        }
        else if( ptr->predictor_type==0 )
            unserialize(f, &ptr->linpred);
        else if( ptr->predictor_type==1 )
            unserialize(f, &ptr->rbfpred);
        else if( ptr->predictor_type==2 )
            unserialize(f, &ptr->knnpred);
        else if( ptr->predictor_type==3 )
            unserialize(f, &ptr->superpred);
        else
            ASSERT(0);
    }
    else
        ASSERT(0);
    unserialize(f, &ptr->cal);
    if( check_magic8(f, 1, 111)==1 )
    {
        unserialize(f, &ptr->adv_clip_level);
        unserialize(f, &ptr->adv_frac_chosen);
        unserialize(f, &ptr->adv_param0);
        unserialize(f, &ptr->adv_param1);
        unserialize(f, &ptr->adv_param2);
        unserialize(f, &ptr->adv_param3);
        unserialize(f, &ptr->adv_param4);
        unserialize(f, &ptr->adv_param5);
        check_magic8(f, 111);
    }
    else
        ptr->adv_clip_level = ptr->adv_frac_chosen = ptr->adv_param0 =  ptr->adv_param1 =  ptr->adv_param2 = ptr->adv_param3 = ptr->adv_param4 = ptr->adv_param5 = 0;
}

float x1_column::get_l1_factor() const
{
    float r = 0.0f;
    for(const conv_subcolumn &s: subcolumns)
        r = fmaxf(r, s.get_l1_factor());
    return r;
}

float x1_column::get_sparse_flops() const
{
    float r = 0.0f;
    for(const conv_subcolumn &s: subcolumns)
        r += s.get_sparse_flops();
    return r;
}

int x1_column::get_nsubcolumns() const
{
    return subcolumns.size();
}

int x1_column::get_v1_window_dim0(const net_v1layer &v1) const
{
    ASSERT(subcolumns.size()>=1);
    int r = subcolumns[0].get_v1_window_dim0(v1);
    for(const conv_subcolumn &s: subcolumns)
        if( &s!=&subcolumns[0] )
            ASSERT(s.get_v1_window_dim0(v1)==r);
    return r;
}

int x1_column::get_v1_window_dim1(const net_v1layer &v1) const
{
    ASSERT(subcolumns.size()>=1);
    int r = subcolumns[0].get_v1_window_dim1(v1);
    for(const conv_subcolumn &s: subcolumns)
        if( &s!=&subcolumns[0] )
            ASSERT(s.get_v1_window_dim1(v1)==r);
    return r;
}

int x1_column::get_v1_downsample_lo(const net_v1layer &v1) const
{
    ASSERT(subcolumns.size()>=1);
    int r = subcolumns[0].get_v1_downsample_lo(v1);
    for(const conv_subcolumn &s: subcolumns)
        if( &s!=&subcolumns[0] )
            ASSERT(s.get_v1_downsample_lo(v1)==r);
    return r;
}

int x1_column::get_v1_downsample_hi(const net_v1layer &v1) const
{
    ASSERT(subcolumns.size()>=1);
    int r = subcolumns[0].get_v1_downsample_hi(v1);
    for(const conv_subcolumn &s: subcolumns)
        if( &s!=&subcolumns[0] )
            ASSERT(s.get_v1_downsample_hi(v1)==r);
    return r;
}

int x1_column::get_v2_window_height(const net_v1layer &v1) const
{
    ASSERT(subcolumns.size()>=1);
    int r = subcolumns[0].get_v2_window_height(v1);
    for(const conv_subcolumn &s: subcolumns)
        if( &s!=&subcolumns[0] )
            ASSERT(s.get_v2_window_height(v1)==r);
    return r;
}

int x1_column::get_v2_window_width(const net_v1layer &v1) const
{
    ASSERT(subcolumns.size()>=1);
    int r = subcolumns[0].get_v2_window_width(v1);
    for(const conv_subcolumn &s: subcolumns)
        if( &s!=&subcolumns[0] )
            ASSERT(s.get_v2_window_width(v1)==r);
    return r;
}


conv_subcolumn& x1_column::get_subcolumn(int col_idx)
{
    ASSERT(col_idx>=0 && col_idx<subcolumns.size());
    return subcolumns[col_idx];
}

predictions x1_column::_apply(const predictions &zero_baseline, activities &_v1_act, predict_t ptype) const
{
    ASSERT(zero_baseline.nclasses==2);
    predictions pred_out(zero_baseline);
    for(int i=0; i<pred_out.stride; i++)
        pred_out.fpredictions[i] = ptype!=predict_t::prob ? -1.0e20f : 0;
    for(const conv_subcolumn &subcol: subcolumns)
    {
        predictions pred_tmp;
        if( ptype==predict_t::prob )
            pred_tmp = subcol.apply(zero_baseline, _v1_act);
        if( ptype==predict_t::logprob )
            pred_tmp = subcol.apply_log(zero_baseline, _v1_act);
        if( ptype==predict_t::logit )
            pred_tmp = subcol.apply_logit(zero_baseline, _v1_act);
        ASSERT(pred_tmp.stride==pred_out.stride);
        for(int i=0; i<pred_out.stride; i++)
            pred_out.fpredictions[i] = fmaxf(pred_out.fpredictions[i], pred_tmp.fpredictions[i]);
    }
    if( ptype==predict_t::prob )
    {
        for(int i=0; i<pred_out.stride; i++)
            pred_out.fpredictions[i+pred_out.stride] = 1.0f-pred_out.fpredictions[i];
    }
    else if( ptype==predict_t::logprob )
    {
        for(int i=0; i<pred_out.stride; i++)
            pred_out.fpredictions[i+pred_out.stride] = 1.0f-expf(pred_out.fpredictions[i]);
    }
    else if( ptype==predict_t::logit )
    {
        for(int i=0; i<pred_out.stride; i++)
            pred_out.fpredictions[i+pred_out.stride] = 0.0;
    }
    else
        ASSERT(0);
    return pred_out;
}

predictions x1_column::apply(const predictions &zero_baseline, activities &_v1_act) const
{
    return _apply(zero_baseline, _v1_act, predict_t::prob);
}

predictions x1_column::apply_log(const predictions &zero_baseline, activities &_v1_act) const
{
    return _apply(zero_baseline, _v1_act, predict_t::logprob);
}

predictions x1_column::apply_logit(const predictions &zero_baseline, activities &_v1_act) const
{
    return _apply(zero_baseline, _v1_act, predict_t::logit);
}

conv_predictions x1_column::conv_apply0(const net_v1layer &v1, conv_activities &_v1_act) const
{
    int img_dim0 = _v1_act.get_img_height(), img_dim1 = _v1_act.get_img_width();
    int v1_ds    = get_v1_downsample_lo(v1);
    ASSERT(img_dim0%v1_ds==0 && img_dim1%v1_ds==0);
    int windim0 = get_v2_window_height(v1), windim1 = get_v2_window_width(v1);
    int pred_dim0 = img_dim0/v1_ds-windim0+1, pred_dim1 = img_dim1/v1_ds-windim1+1;
    conv_predictions r(std::make_tuple(img_dim0,img_dim1), v1_ds, std::make_tuple(pred_dim0,pred_dim1), std::make_tuple(0,0), 1);
    for(const conv_subcolumn &subcol: subcolumns)
        r.merge_max(subcol.conv_apply0(v1, _v1_act));
    return r;
}

void x1_column::append_subcolumn(const conv_subcolumn &c)
{
    subcolumns.push_back(c);
};

void serialize(FILE *f, const x1_column *ptr)
{
    write_magic8(f, 32);
    serialize(f, &ptr->subcolumns);
    write_magic8(f, 23);
}

void unserialize(FILE *f, x1_column *ptr)
{
    *ptr = x1_column();
    check_magic8(f, 32);
    unserialize(f, &ptr->subcolumns);
    check_magic8(f, 23);
}

float x1_network::get_l1_factor() const
{
    float r = 0.0f;
    for(const x1_column &c: columns)
        r = fmaxf(r, c.get_l1_factor());
    return r;
}

float x1_network::get_sparse_flops() const
{
    float r = 0.0f;
    for(const x1_column &c: columns)
        r += c.get_sparse_flops();
    return r;
}

int x1_network::get_nclasses() const
{
    return nclasses;
}

x1_column& x1_network::get_column(int col_idx)
{
    ASSERT(columns.size()==nclasses && col_idx>=0 && col_idx<nclasses);
    return columns[col_idx];
}


conv_subcolumn& x1_network::get_subcolumn1(int col_idx)
{
    ASSERT(get_column(col_idx).get_nsubcolumns()==1);
    return get_column(col_idx).get_subcolumn(0);
}

net_v1layer& x1_network::get_v1()
{
    return v1;
}

predictions x1_network::_apply(const raw_dataset &ds_raw, storage_format fmt, predict_t ptype) const
{
    //
    // Integrity checks
    //
    ASSERT(nclasses>=1);
    ASSERT(columns.size()==nclasses);

    //
    // V1
    //
    activities act_v1;
    v1.apply(ds_raw, act_v1, false, fmt);

    //
    // Columns
    //
    predictions zero_baseline(ds_raw);
    std::vector<predictions> vp_out;
    for(int class_idx=0; class_idx<nclasses; class_idx++)
    {
        predictions baselinei = predictions::onevsall(zero_baseline, class_idx);
        const x1_column &col_i = columns[class_idx];
        if( ptype==predict_t::prob )
            vp_out.push_back(col_i.apply(baselinei,act_v1));
        if( ptype==predict_t::logprob )
            vp_out.push_back(col_i.apply_log(baselinei,act_v1));
        if( ptype==predict_t::logit )
            vp_out.push_back(col_i.apply_logit(baselinei,act_v1));
    }
    return predictions::merge_onevsall(vp_out);
}

predictions x1_network::apply(const raw_dataset &ds_raw, storage_format fmt) const
{
    return _apply(ds_raw, fmt, predict_t::prob);
}

predictions x1_network::apply_log(const raw_dataset &ds_raw, storage_format fmt) const
{
    return _apply(ds_raw, fmt, predict_t::logprob);
}

predictions x1_network::apply_logit(const raw_dataset &ds_raw, storage_format fmt) const
{
    return _apply(ds_raw, fmt, predict_t::logit);
}

void x1_network::start_initialization(int classes, const net_v1layer &input_layer)
{
    ASSERT(nclasses==0 && columns.size()==0);
    nclasses = classes;
    v1 = input_layer;
};

void x1_network::append_column(const x1_column &c)
{
    ASSERT(nclasses>0);
    ASSERT(columns.size()<nclasses);
    columns.push_back(c);
}

void serialize(FILE *f, const x1_network *ptr)
{
    write_magic8(f, 63);
    serialize(f, &ptr->nclasses);
    serialize(f, &ptr->v1);
    serialize(f, &ptr->columns);
    write_magic8(f, 37);
}

void unserialize(FILE *f, x1_network *ptr)
{
    *ptr = x1_network();
    check_magic8(f, 63);
    unserialize(f, &ptr->nclasses);
    unserialize(f, &ptr->v1);
    unserialize(f, &ptr->columns);
    check_magic8(f, 37);
}

float x1_binary::get_sparse_flops() const
{
    return column0.get_sparse_flops();
}

int x1_binary::get_v1_window_dim0() const
{
    return column0.get_v1_window_dim0(v1);
}

int x1_binary::get_v1_window_dim1() const
{
    return column0.get_v1_window_dim1(v1);
}

predictions x1_binary::_apply(const raw_dataset &ds_raw, storage_format fmt, predict_t ptype) const
{
    //
    // Integrity checks
    //
    ASSERT(ds_raw.nclasses==2);
    ASSERT(initialized);
    ASSERT(ptype==predict_t::prob || ptype==predict_t::logit);

    //
    // V1
    //
    activities act_v1;
    v1.apply(ds_raw, act_v1, false, fmt);

    //
    // Columns
    //
    const predictions baseline(ds_raw);
    predictions pred0 = baseline;
    predictions pred_all = baseline;
    if( ptype==predict_t::prob )
    {
        pred0 = column0.apply(baseline,act_v1);
        for(int i=0; i<pred_all.ntrn+pred_all.ntst; i++)
        {
            pred_all.fpredictions[i] = pred0.fpredictions[i];
            pred_all.fpredictions[i+pred_all.stride] = 1-pred_all.fpredictions[i];
        }
    }
    if( ptype==predict_t::logit )
    {
        pred0 = column0.apply_logit(baseline,act_v1);
        for(int i=0; i<pred_all.ntrn+pred_all.ntst; i++)
        {
            pred_all.fpredictions[i] = pred0.fpredictions[i];
            pred_all.fpredictions[i+pred_all.stride] = invsigma0_safe(1-sigma0(pred_all.fpredictions[i]));
        }
    }
    return pred_all;
}

predictions x1_binary::apply(const raw_dataset &ds_raw, storage_format fmt) const
{
    return _apply(ds_raw, fmt, predict_t::prob);
}

predictions x1_binary::apply_logit(const raw_dataset &ds_raw, storage_format fmt) const
{
    return _apply(ds_raw, fmt, predict_t::logit);
}

float x1_binary::apply0(const bitmap &img, storage_format fmt) const
{
    raw_dataset ds_tmp(2);
    ds_tmp.append_trn(raw_dataset::entry(img,0));
    predictions p = apply(ds_tmp, fmt);
    return p.fpredictions[0];
}

conv_predictions x1_binary::conv_apply(const bitmap &raw_img, padding_t pad, storage_format fmt) const
{
    ASSERT(initialized);
    int raw_img_dim0 = raw_img.height, raw_img_dim1 = raw_img.width;
    int windim0 = column0.get_v2_window_height(v1);
    int windim1 = column0.get_v2_window_width(v1);
    int v1_ds_lo= column0.get_v1_downsample_lo(v1);
    int v1_ds_hi= column0.get_v1_downsample_hi(v1);
    ASSERT(raw_img_dim0%v1_ds_lo==0 && raw_img_dim1%v1_ds_lo==0);
    ASSERT(raw_img_dim0%v1_ds_hi==0 && raw_img_dim1%v1_ds_hi==0);

    //
    // Padding if needed
    //
    int pred_shift0=windim0/2, pred_shift1=windim1/2;
    bitmap pad_img;
    ASSERT(pad==padding_t::none || pad==padding_t::gray);
    if( pad==padding_t::gray )
    {
        int pad0_top  = (windim0/2)*v1_ds_lo, pad0_bottom = (windim0-windim0/2)*v1_ds_lo;
        int pad1_left = (windim1/2)*v1_ds_lo, pad1_right  = (windim1-windim1/2)*v1_ds_lo;
        ASSERT(pad0_top>=0 && pad0_bottom>=0 && pad1_left>=0 && pad1_right>=0);
        pad_img = raw_img.append_border_v(pad0_top, pad0_bottom, pad1_left, pad1_right, 0.5f);
        pred_shift0 = 0;
        pred_shift1 = 0;
    }
    const bitmap &wrk_img = pad==padding_t::none ? raw_img : pad_img;
    int wrk_img_dim0 = wrk_img.height, wrk_img_dim1 = wrk_img.width;
    ASSERT(v1.get_img_height()<=wrk_img_dim0 && v1.get_img_width()<=wrk_img_dim1);
    ASSERT(wrk_img_dim0%v1_ds_lo==0 && wrk_img_dim1%v1_ds_lo==0);
    ASSERT(wrk_img_dim0%v1_ds_hi==0 && wrk_img_dim1%v1_ds_hi==0);

    //
    // V1
    //
    conv_activities act_v1;
    v1.conv_apply(wrk_img, act_v1, false, fmt);

    //
    // V2-V4
    //
    int pred_dim0 = wrk_img_dim0/v1_ds_lo-windim0+1, pred_dim1 = wrk_img_dim1/v1_ds_lo-windim1+1;
    conv_predictions r0 = column0.conv_apply0(v1, act_v1);
    conv_predictions r(std::make_tuple(raw_img_dim0,raw_img_dim1), v1_ds_lo, std::make_tuple(pred_dim0,pred_dim1), std::make_tuple(pred_shift0,pred_shift1), 2);
    ASSERT(r0.get_pred_height()>=r.get_pred_height());
    ASSERT(r0.get_pred_width()>=r.get_pred_width());
    for(int i=0; i<pred_dim0; i++)
        for(int j=0; j<pred_dim1; j++)
            r.set_pred01(i, j, r0.get_pred(i,j));
    return r;
}

void x1_binary::initialize(const net_v1layer &input_layer, const x1_column &c)
{
    initialized = true;
    v1 = input_layer;
    column0 = c;
}

void serialize(FILE *f, const x1_binary *ptr)
{
    ASSERT(ptr->initialized);
    write_magic8(f, 38);
    write_magic8(f, 79);
    serialize(f, &ptr->v1);
    serialize(f, &ptr->column0);
    write_magic8(f, 37);
}

void unserialize(FILE *f, x1_binary *ptr)
{
    check_magic8(f, 38);
    check_magic8(f, 79);
    unserialize(f, &ptr->v1);
    unserialize(f, &ptr->column0);
    check_magic8(f, 37);
    ptr->initialized = true;
}

const x1_binary& x1_binary_cascade::last() const
{
    ASSERT(levels.size()>0);
    return std::get<1>(levels.back());
}

int x1_binary_cascade::get_last_scale() const
{
    ASSERT(levels.size()>0);
    return std::get<0>(levels.back());
}

void x1_binary_cascade::add_level(int scale_idx, const x1_binary &net, float lev)
{
    levels.push_back(std::make_tuple(scale_idx,net,lev));
}

conv_predictions x1_binary_cascade::conv_apply(const bitmap &img, padding_t pad, storage_format fmt) const
{
    //
    // Quick exit for empty cascade - return 0.5
    //
    if( levels.size()==0 )
    {
        ASSERT(pad==padding_t::gray);
        conv_predictions pred(img.height, img.width, 2);
        for(int i0=0; i0<img.height; i0++)
            for(int i1=0; i1<img.width; i1++)
                pred.set_original_pixel01(i0,i1,0.5f);
        return pred;
    }

    //
    // Non-empty cascade
    //
    std::vector<conv_predictions> v_out;
    std::vector<float> f_lev;
    for(const std::tuple<int,x1_binary,float> &tt: levels)
    {
        int scale_idx = std::get<0>(tt);
        ASSERT(scale_idx>=0);
        bitmap scaled_img;
        const bitmap &wrk_img = scale_idx==0 ? img : scaled_img;
        if( scale_idx>0 )
        {
            scaled_img = img;
            for(int k=0; k<scale_idx; k++)
                scaled_img = scaled_img.downsample2();
        }
        v_out.push_back(std::get<1>(tt).conv_apply(wrk_img, pad, fmt).upscale2k(scale_idx));
        f_lev.push_back(std::get<2>(tt));
    }
    for(int i=1; i<v_out.size(); i++)
    {
        ASSERT(v_out[i].get_pred_width()>=v_out[i-1].get_pred_width());
        ASSERT(v_out[i].get_pred_height()>=v_out[i-1].get_pred_height());
    }
    conv_predictions pred_r = v_out.back();
    for(int i0=0; i0<pred_r.get_img_height(); i0++)
        for(int i1=0; i1<pred_r.get_img_width(); i1++)
        {
            float v = 0.0f;
            for(int idx=0; idx<v_out.size(); idx++)
            {
                float vv = v_out[idx].get_original_pixel(i0,i1);
                if( vv<f_lev[idx] )
                {
                    v = 0;
                    break;
                }
                v = vv;
            }
            pred_r.set_original_pixel01(i0,i1,v);
        }
    return pred_r;
}

conv_predictions x1_binary_cascade::conv_apply(const bitmap &img, int nscales, padding_t pad, storage_format fmt) const
{
    ASSERT(nscales>=1);
    conv_predictions p0 = conv_apply(img, pad, fmt);
    if( nscales>1 )
    {
        bitmap wrk = img.downsample2();
        for(int k=1; k<nscales; k++)
        {
            conv_predictions pk = conv_apply(wrk, pad, fmt);
            for(int i0=0; i0<img.height; i0++)
                for(int i1=0; i1<img.width; i1++)
                    p0.set_original_pixel01(i0,i1,fmaxf(p0.get_original_pixel(i0,i1),
                                                        pk.get_original_pixel(i0>>k,i1>>k)));
            wrk = wrk.downsample2();
        }
    }
    return p0;
}

std::vector<conv_predictions> x1_binary_cascade::conv_apply_ml(const bitmap &img, padding_t pad, storage_format fmt) const
{
    //
    // Quick exit for empty cascade - return 0.5
    //
    if( levels.size()==0 )
    {
        ASSERT(pad==padding_t::gray);
        conv_predictions pred(img.height, img.width, 2);
        for(int i0=0; i0<img.height; i0++)
            for(int i1=0; i1<img.width; i1++)
                pred.set_original_pixel01(i0,i1,0.5f);
        return {pred};
    }

    //
    // Non-empty cascade
    //
    std::vector<conv_predictions> v_out;
    std::vector<float> f_lev;
    for(const std::tuple<int,x1_binary,float> &tt: levels)
    {
        int scale_idx = std::get<0>(tt);
        ASSERT(scale_idx>=0);
        bitmap scaled_img;
        const bitmap &wrk_img = scale_idx==0 ? img : scaled_img;
        if( scale_idx>0 )
        {
            scaled_img = img;
            for(int k=0; k<scale_idx; k++)
                scaled_img = scaled_img.downsample2();
        }
        v_out.push_back(std::get<1>(tt).conv_apply(wrk_img, pad, fmt).upscale2k(scale_idx));
        f_lev.push_back(std::get<2>(tt));
    }
    for(int i=1; i<v_out.size(); i++)
    {
        ASSERT(v_out[i].get_pred_width()>=v_out[i-1].get_pred_width());
        ASSERT(v_out[i].get_pred_height()>=v_out[i-1].get_pred_height());
    }
    for(int i0=0; i0<v_out.back().get_img_height(); i0++)
        for(int i1=0; i1<v_out.back().get_img_width(); i1++)
        {
            float v = 0.0f;
            for(int idx=0; idx<v_out.size(); idx++)
                if( v_out[idx].get_original_pixel(i0,i1)<f_lev[idx] )
                {
                    for(int idx2=idx; idx2<v_out.size(); idx2++)
                        v_out[idx2].set_original_pixel01(i0,i1,0.0f);
                }
        }
    return v_out;
}

std::vector<conv_predictions> x1_binary_cascade::conv_apply_ml(const bitmap &img, int nscales, padding_t pad, storage_format fmt) const
{
    ASSERT(nscales>=1);
    const int nlevels = levels.size();
    std::vector<conv_predictions> p0 = conv_apply_ml(img, pad, fmt);
    if( nscales>1 )
    {
        bitmap wrk = img.downsample2();
        for(int k=1; k<nscales; k++)
        {
            std::vector<conv_predictions> pk = conv_apply_ml(wrk, pad, fmt);
            for(int t=0; t<nlevels; t++)
                for(int i0=0; i0<img.height; i0++)
                    for(int i1=0; i1<img.width; i1++)
                        p0[t].set_original_pixel01(i0,i1,fmaxf(p0[t].get_original_pixel(i0,i1),
                                                               pk[t].get_original_pixel(i0>>k,i1>>k)));
            wrk = wrk.downsample2();
        }
    }
    return p0;
}

dataset_subsplit dataset_subsplit::train_subsplit(const raw_dataset &ds_raw,
                                                  const activities &act_v1,
                                                  const int col,
                                                  const int kpower,
                                                  const bool rep)
{
    dataset_subsplit  subsplit_out;
    int t_pca = 0, t_kmeans = 0, t_process = 0;
    ASSERT(kpower>=1);

    //
    // Init subsplit with default partition
    //
    subsplit_out.kfolds = kpower;
    subsplit_out.split = std::vector<int>(ds_raw.ntrn+ds_raw.ntst,-1);

    //
    // Do we need to perform actual analysis?
    //
    if( kpower>1 )
    {
        int idx_out;
        std::vector<int> items;

        //
        // Perform PCA on elements of class #col
        //
        t_pca -= dtickcount();
        items.clear();
        for(int i=0; i<ds_raw.ntrn; i++)
            if( ds_raw.entries[i].class_id==col )
                items.push_back(i);
        ASSERT(items.size()>=1);
        const int pca_dim = imin3(5*kpower,act_v1.nchannels,items.size());
        alglib::real_2d_array xx;
        xx.setlength(items.size(),act_v1.nchannels);
        idx_out = 0;
        for(int i: items)
        {
            for(int j=0; j<act_v1.nchannels; j++)
                xx[idx_out][j] = act_v1.get_normalized(j,i);
            idx_out++;
        }
        alglib::real_1d_array s2;
        alglib::real_2d_array pca, xx_proj;
        alglib::pcatruncatedsubspace(xx, xx.rows(), xx.cols(), pca_dim, 0.001, 50, s2, pca, alglib::parallel);
        xx_proj.setlength(items.size(), pca_dim);
        alglib::rmatrixgemm(xx_proj.rows(), xx_proj.cols(), xx.cols(),
                            1.0,
                            xx, 0, 0, 0,
                            pca, 0, 0, 0,
                            0.0,
                            xx_proj, 0, 0);
        t_pca += dtickcount();

        //
        // Run k-means analysis
        //
        t_kmeans -= dtickcount();
        alglib::clusterizerstate cs;
        alglib::kmeansreport crep;
        alglib::clusterizercreate(cs);
        alglib::clusterizersetpoints(cs, xx_proj, 2);
        alglib::clusterizersetkmeanslimits(cs, 1, 100);
        alglib::clusterizerrunkmeans(cs, kpower, crep, alglib::parallel);
        ASSERT(crep.terminationtype>0);
        t_kmeans += dtickcount();

        //
        // Process entire dataset
        //
        t_process -= dtickcount();
        items.clear();
        for(int i=0; i<ds_raw.ntrn+ds_raw.ntst; i++)
            if( ds_raw.entries[i].class_id==col )
                items.push_back(i);
        ASSERT(items.size()>=1);
        xx.setlength(items.size(),act_v1.nchannels);
        idx_out = 0;
        for(int i: items)
        {
            for(int j=0; j<act_v1.nchannels; j++)
                xx[idx_out][j] = act_v1.get_normalized(j,i);
            idx_out++;
        }
        xx_proj.setlength(items.size(), pca_dim);
        alglib::rmatrixgemm(xx_proj.rows(), xx_proj.cols(), xx.cols(),
                            1.0,
                            xx, 0, 0, 0,
                            pca, 0, 0, 0,
                            0.0,
                            xx_proj, 0, 0);
        for(int i=0; i<items.size(); i++)
        {
            const double *p_prj = xx_proj[i];
            int idx_min = -1;
            double dist_min = 0;
            for(int k=0; k<kpower; k++)
            {
                double *p_cluster = crep.c[k];
                double d2 = 0.0f;
                for(int j=0; j<pca_dim; j++)
                {
                    double v = p_cluster[j]-p_prj[j];
                    d2 += v*v;
                }
                if( idx_min<0 || d2<dist_min )
                {
                    idx_min = k;
                    dist_min = d2;
                }
            }
            ASSERT(idx_min>=0);
            subsplit_out.split[items[i]] = idx_min;
        }
        t_process += dtickcount();
    }
    else
    {
        //
        // Everything is fold 0
        //
        for(int i=0; i<ds_raw.ntrn+ds_raw.ntst; i++)
            if( ds_raw.entries[i].class_id==col )
                subsplit_out.split[i] = 0;
    }

    //
    // Output report
    //
    if( rep )
    {
        std::mt19937  mt_rnd;
        mt_rnd.seed(1);
        for(int widx=0; widx<kpower; widx++)
        {
            std::vector<int> arr;
            for(int i=ds_raw.ntrn; i<ds_raw.ntrn+ds_raw.ntst; i++)
                if( subsplit_out.split[i]==widx )
                    arr.push_back(i);
            std::shuffle(arr.begin(), arr.end(), mt_rnd);
            std::vector<bitmap> bmp_vec;
            for(int i=0; i<imin(arr.size(),100); i++)
                bmp_vec.push_back(ds_raw.entries[arr[i]].img);
            bitmap bmp_out = bitmap::combine_grid_w(bmp_vec, 10, 1, 0x808080);
            char filename[1024];
            sprintf(filename, "%s/class%02d_split%02d.png", OUT_DIR, col, widx);
            bmp_out.to_png(filename);
        }
        printf("pca=%d ms, km=%d ms, proc=%d ms", t_pca, t_kmeans, t_process);
        fflush(stdout);
    }

    //
    // Done
    //
    return subsplit_out;
}

class _maxksigma2_env
{
public:
    const activities &act_in;
    const predictions baseline;
    const int64_t kparam;
    //const float lambda_reg;
    const float leak_exp;
    const float leak_max;
    const int64_t stride;

    // forward pass
    activityf_vector sigmakbuf, sigmamaxbuf;
    std::vector<int> sigmaidx;
    //activityf_vector expbuf;
    //activityf_vector sumbuf;
    //activityf_vector dedfbuf;
    //activityf_vector sumfidedfbuf;
    //activityf_vector dednkbuf;

    // derivative of error with respect to sigmoid input (stored in sigmamaxbuf)
    activityf_vector dedn, dednk;

    int it_cnt;

    _maxksigma2_env(const activities &a, const predictions &p, int kp):
        leak_exp(0.05f),leak_max(0.15f),
        act_in(a), baseline(p), kparam(kp),// lambda_reg(lr),
        stride(a.ntrn+a.ntst),
        sigmakbuf(stride*kparam), sigmamaxbuf(stride), sigmaidx(stride),
        dedn(stride), dednk(stride*kparam),
        //expbuf(stride*p.nclasses,0.0f),  sumbuf(stride,0.0f),
        //dedfbuf(stride*p.nclasses,0.0f), sumfidedfbuf(stride,0.0f),
        //dednkbuf(stride*p.nclasses,0.0f),
        it_cnt(0)
    {}
};

template<class fptype>
void maxksigma2_produce_predictions(_maxksigma2_env &env, const fptype *params, predictions &pred)
{
    const int64_t nchannels = env.act_in.nchannels;
    const int64_t ntotal    = env.act_in.ntrn+env.act_in.ntst;

    // integrity checks
    ASSERT(env.baseline.nclasses==2);
    ASSERT(env.act_in.is_f32());
    ASSERT(env.kparam>0);

    // allocate memory
    pred = predictions(env.baseline.ntrn, env.baseline.ntst, env.baseline.nclasses, env.baseline.labels);

    // compute linear part
    for(int64_t kk=0; kk<env.kparam; kk++)
    {
        const float fintercept = params[nchannels+kk*(nchannels+1)];
        for(int64_t i=0; i<ntotal; i++)
            env.sigmakbuf[i+kk*env.stride] = fintercept;
        for(int64_t t=0; t<nchannels; t++)
            fadd_simd(&env.sigmakbuf[kk*env.stride], env.act_in.get_f32(t), ntotal, params[t+kk*(nchannels+1)]);
    }
    for(int64_t i=0; i<ntotal; i++)
    {
        env.sigmamaxbuf[i] = env.sigmakbuf[i];
        env.sigmaidx[i] = 0;
    }
    for(int64_t kk=1; kk<env.kparam; kk++)
        for(int64_t i=0; i<ntotal; i++)
            if( env.sigmakbuf[i+kk*env.stride]>env.sigmamaxbuf[i] )
            {
                env.sigmamaxbuf[i] = env.sigmakbuf[i+kk*env.stride];
                env.sigmaidx[i] = kk;
            }

    // sigmoid
    for(int64_t i=0; i<ntotal; i++)
    {
        float v0 = env.sigmamaxbuf[i], v1 = 0;
        float vm = fmaxf(v0,v1);
        v0 = expf(v0-vm);
        v1 = expf(v1-vm);
        pred.fpredictions[i+0*pred.stride] = v0/(v0+v1);
        pred.fpredictions[i+1*pred.stride] = v1/(v0+v1);
    }
}

class _maxksigma2_sgd
{
private:
    const activities    &act_in;
    const predictions   pred_target;
    const float         lambda_reg;

    std::mt19937        mt_rnd;
    maxout_postlayer    layer;
    std::vector<float>  ww;

    _maxksigma2_env     glbl_env;
public:
    _maxksigma2_sgd(const activities &ai, const predictions &trg, const maxout_postlayer lyr, const float lreg):
        act_in(ai), pred_target(trg), layer(lyr), lambda_reg(lreg),
        glbl_env(act_in, pred_target, layer.strength),
        ww(layer.get_dim(),0.0f)
        {}

    maxout_postlayer get_layer() const { return layer; }
    void run_epoch(float alpha, float mu, int batch_size);
};

void _maxksigma2_sgd::run_epoch(float alpha, float mu, int batch_size)
{
    ASSERT(batch_size>1);
    ASSERT(act_in.ntrn>=batch_size);

    //
    // Prepare structure for batch evaluation
    //
    activities act_batch;
    predictions pred_batch;
    std::vector<int64_t> idx;
    for(int64_t i=0; i<act_in.ntrn; i++)
        idx.push_back(i);
    std::shuffle(idx.begin(), idx.end(), mt_rnd);

    //
    // Prepare batch activities
    //
    predictions tmp_pred;
    for(int64_t bidx=0; bidx<act_in.ntrn/batch_size; bidx++)
    {
        //
        // select subsample to process
        //
        std::vector<int> items(idx.begin()+bidx*batch_size, idx.begin()+(bidx+1)*batch_size);
        std::sort(items.begin(), items.end());
        act_in.select_subsample(items, act_batch);
        pred_batch = pred_target.select_subsample(items);
        _maxksigma2_env lcl_env(act_batch, pred_batch, layer.strength);

        //
        // Compute predictions for current mini-batch
        //
        maxksigma2_produce_predictions(lcl_env, &layer.coeffs[0], tmp_pred);

        //
        // Compute target function and temporaries used to calculate its derivative
        //
        const float *pred0 = &tmp_pred.fpredictions[0];
        memset(&lcl_env.dednk[0], 0, lcl_env.dednk.size()*sizeof(lcl_env.dednk[0]));
        for(int64_t i=0; i<tmp_pred.ntrn; i++)
        {
            float v = pred0[i]-(tmp_pred.labels[i]==0 ? 1 : 0);
            lcl_env.dedn[i] = v*(lcl_env.leak_exp+pred0[i]*(1.0f-pred0[i]));
            lcl_env.dednk[i+lcl_env.sigmaidx[i]*lcl_env.stride] = lcl_env.dedn[i];
            if( tmp_pred.labels[i]==0 && pred0[i]<0.5f )
                lcl_env.dednk[i+(lcl_env.kparam-1)*lcl_env.stride] += lcl_env.leak_max*lcl_env.dedn[i];
        }

        //
        // Compute gradient
        //
        activityf_vector ones(tmp_pred.ntrn, 1.0f);
        std::vector<float> grad(layer.get_dim(),0.0f);
        for(int64_t kk=0; kk<lcl_env.kparam; kk++)
        {
            const float *p_dednk = &lcl_env.dednk[kk*lcl_env.stride];
            for(int64_t i=0; i<act_batch.nchannels+1; i++)
            {
                const float *p_act_i = i<act_batch.nchannels ? act_batch.get_f32(i) : &ones[0];
                grad[i+kk*(act_batch.nchannels+1)] = fdot_simd(p_dednk, p_act_i, tmp_pred.ntrn)/tmp_pred.ntrn;
            }
        }

        //
        // Add regularizer
        //
        for(int64_t i=0; i<layer.get_dim(); i++)
            grad[i] += lambda_reg*layer.coeffs[i];

        //
        // Update x0
        //
        for(int i=0; i<layer.get_dim(); i++)
        {
            ww[i]            = mu*ww[i]+alpha*grad[i];
            layer.coeffs[i] -= ww[i];
        }
    }

    //
    // Produce report
    //
    maxksigma2_produce_predictions(glbl_env, &layer.coeffs[0], tmp_pred);
    float mx_act_last = -1000.0f, avg_act_last = 0.0f, rat_won_last = 0.0f;
    for(int64_t i=0; i<act_in.ntrn; i++)
    {
        float v_last = glbl_env.sigmakbuf[i+(glbl_env.kparam-1)*glbl_env.stride];
        mx_act_last   = fmaxf(mx_act_last, v_last);
        avg_act_last += v_last/act_in.ntrn;
        rat_won_last += ((glbl_env.sigmaidx[i]==glbl_env.kparam-1 && v_last>0.0f) ? 1.0f : 0.0f)/act_in.ntrn;
    }
    printf("[maxksigma2-trn:%04d] TrnRms=%.4f    TstRms=%.4f    TstCls=%.4f    (mx_lst=%+.4f, av_lst=%+.6f, w_lst=%.6f)\n",
           glbl_env.it_cnt,
           (double)tmp_pred.trn_rms(),
           (double)tmp_pred.tst_rms(),
           (double)tmp_pred.tst_cls(),
           (double)mx_act_last,
           (double)avg_act_last,
           (double)rat_won_last);
    fflush(stdout);
    glbl_env.it_cnt++;
}

maxout_postlayer extend_maxksigma2_model(const activities &act_v2c,
                                         const predictions &pred_target,
                                         maxout_postlayer layer,
                                         const int its_cnt,
                                         const float lambda_reg)
{
    ASSERT(pred_target.nclasses==2);

    //const int nclasses  = pred_target.nclasses;
    const int nchannels = act_v2c.nchannels;
    std::mt19937  mt_rnd;
    std::seed_seq mt_seed{(unsigned int)1};
    mt_rnd.seed(mt_seed);

    try
    {
        //
        // Extend model
        //
        ASSERT(layer.coeffs.size()==(layer.nchannels+1)*layer.strength);
        layer.coeffs.resize((layer.nchannels+1)*(layer.strength+1),0.0f);
        layer.coeffs[layer.nchannels+(layer.nchannels+1)*layer.strength] = -10.0;
        layer.strength++;

        //
        // SGD training
        //
        const float mu    = 0.98f;
        const float alpha = 0.003f;
        _maxksigma2_sgd sgd(act_v2c, pred_target, layer, lambda_reg);

        for(int it_idx=0; it_idx<its_cnt; it_idx++)
            sgd.run_epoch(alpha, mu, 128);
        return sgd.get_layer();

#if 0
        const float mu = 0.95f;
        const float alpha = 0.2f/act_v2c.ntrn;
        const double dmu = mu, dalpha = alpha;
        _maxksigma2_env env(act_v2c, pred_target, layer.strength);
        alglib::real_1d_array x0, grad, momentum;
        predictions tmp_pred;
        x0.setlength(layer.get_dim());
        grad.setlength(layer.get_dim());
        momentum.setlength(layer.get_dim());
        for(int i=0; i<x0.length(); i++)
        {
            x0[i] = layer.coeffs[i];
            momentum[i] = 0.0;
        }
        for(int it_idx=0; it_idx<1500; it_idx++)
        {
            double func;

            // compute predictions
            maxksigma2_produce_predictions(env, &x0[0], tmp_pred);

            // compute target function and temporaries used to calculate its derivative
            const float *pred0 = &tmp_pred.fpredictions[0];
            func = 0;
            memset(&env.dednk[0], 0, env.dednk.size()*sizeof(env.dednk[0]));
            for(int64_t i=0; i<tmp_pred.ntrn; i++)
            {
                float v = pred0[i]-(tmp_pred.labels[i]==0 ? 1 : 0);
                func += 0.5*v*v;
                env.dedn[i] = v*(env.leak_exp+pred0[i]*(1.0f-pred0[i]));
                env.dednk[i+env.sigmaidx[i]*env.stride] = env.dedn[i];
                if( tmp_pred.labels[i]==0 && pred0[i]<0.5f )
                    env.dednk[i+(env.kparam-1)*env.stride] += env.leak_max*env.dedn[i];
            }

            // compute derivative
            activityf_vector ones(tmp_pred.ntrn, 1.0f);
            memset(&grad[0], 0, grad.length()*sizeof(grad[0]));
            for(int64_t kk=0; kk<env.kparam; kk++)
            {
                const float *p_dednk = &env.dednk[kk*env.stride];
                for(int64_t i=0; i<env.act_in.nchannels+1; i++)
                {
                    const float *p_act_i = i<env.act_in.nchannels ? env.act_in.get_f32(i) : &ones[0];
                    grad[i+kk*(env.act_in.nchannels+1)]  = fdot_simd(p_dednk, p_act_i, env.act_in.ntrn);
                }
            }

            // add regularizer
            for(int64_t i=0; i<x0.length(); i++)
            {
                float s = lambda_reg*env.act_in.ntrn;
                func    += 0.5*x0[i]*x0[i];
                grad[i] += s*x0[i];
            }

            // update x0
            for(int i=0; i<x0.length(); i++)
            {
                momentum[i] = dmu*momentum[i]+dalpha*grad[i];
                x0[i] -= momentum[i];
            }

            // report
            float mx_act_last = -1000.0f, avg_act_last = 0.0f, rat_won_last = 0.0f;
            for(int64_t i=0; i<act_v2c.ntrn; i++)
            {
                mx_act_last   = fmaxf(mx_act_last, env.sigmakbuf[i+(env.kparam-1)*env.stride]);
                avg_act_last += env.sigmakbuf[i+(env.kparam-1)*env.stride]/act_v2c.ntrn;
                rat_won_last += (env.sigmaidx[i]==env.kparam-1 ? 1.0f : 0.0f)/act_v2c.ntrn;
            }
            printf("[maxksigma2-trn:%04d] TrnRms=%.4f    TstRms=%.4f    TstCls=%.4f    (loss=%.8e, mx_lst=%+.4f, av_lst=%+.6f, w_lst=%.6f)\n",
                   env.it_cnt,
                   (double)tmp_pred.trn_rms(),
                   (double)tmp_pred.tst_rms(),
                   (double)tmp_pred.tst_cls(),
                   (double)func,
                   (double)mx_act_last,
                   (double)avg_act_last,
                   (double)rat_won_last);
            fflush(stdout);
            env.it_cnt++;
        }

#if 0
        for(int it_idx=0; it_idx<3000; it_idx++)
        {
            double func;
            maxksigma2_func_grad(x0, func, grad, &env);
            for(int i=0; i<x0.length(); i++)
            {
                momentum[i] = mu*momentum[i]+alpha*grad[i];
                x0[i] -= momentum[i];
            }
            //if( it_idx==0 || it_idx%10==9 )
            {
                maxksigma2_produce_predictions(env, &x0[0], tmp_pred);
                printf("[maxksigma2-trn:%04d] TrnRms=%.8f    TstRms=%.4f    TstCls=%.4f\n",
                       env.it_cnt,
                       (double)tmp_pred.trn_rms(),
                       (double)tmp_pred.tst_rms(),
                       (double)tmp_pred.tst_cls());
                fflush(stdout);
            }
            env.it_cnt++;
        }
#endif
        for(int i=0; i<x0.length(); i++)
            layer.coeffs[i] = x0[i];
        return layer;
#endif

#if 0
        //
        // Prepare initial point
        //
        //std::normal_distribution<> weight_distr(0.0, 1/sqrtf(nchannels+1));
        alglib::real_1d_array x0;
        x0.setlength(layer.get_dim());
        for(int i=0; i<x0.length(); i++)
            x0[i] = layer.coeffs[i];

        //
        // Prepare LBFGS optimizer
        //
        alglib::minlbfgsstate optimizer;
        alglib::minlbfgsreport rep;
        alglib::minlbfgscreate(x0.length(), nchannels+1, x0, optimizer);
        alglib::minlbfgssetxrep(optimizer, true);

        _maxksigma2_env env(act_v2c, pred_target, layer.strength, lambda_reg);
        alglib::minlbfgssetcond(optimizer, 0.0f, 0.0f, 0.0f, nchannels+100);
        alglib::minlbfgsrestartfrom(optimizer, x0);
        alglib::minlbfgsoptimize(optimizer, maxksigma2_func_grad, maxksigma2_rep, &env);
        alglib::minlbfgsresults(optimizer, x0, rep);

        //
        // Unpack and return
        //
        for(int i=0; i<x0.length(); i++)
            layer.coeffs[i] = x0[i];
        return layer;
#endif
    }
    catch(alglib::ap_error e)
    {
        printf("[exception]: %s\n", e.msg.c_str());
        throw;
    }
}

class gmaxout_buf
{
public:
    // forward pass
    activityf_vector sigmakbuf, sigmamaxbuf;
    std::vector<int> sigmaidx;

    // derivative of error with respect to sigmoid input (stored in sigmamaxbuf)
    activityf_vector dedn, dednk;

    //
    void alloc_and_clear(int n, int kpower)
    {
        sigmakbuf.resize(n*kpower);
        sigmamaxbuf.resize(n);
        sigmaidx.resize(n);
        dedn.resize(n);
        dednk.resize(n*kpower);

        memset(&sigmakbuf[0],   0,   sigmakbuf.size()*sizeof(sigmakbuf[0]));
        memset(&sigmamaxbuf[0], 0, sigmamaxbuf.size()*sizeof(sigmamaxbuf[0]));
        memset(&sigmaidx[0],    0,    sigmaidx.size()*sizeof(sigmaidx[0]));
        memset(&dedn[0],        0,        dedn.size()*sizeof(dedn[0]));
        memset(&dednk[0],       0,       dednk.size()*sizeof(dednk[0]));
    }
};

static void gmaxout_produce_predictions(const gmaxout_postlayer &layer, const predictions &baseline, std::vector<const activities*> vv_act, gmaxout_buf &buf, predictions &pred)
{
    ASSERT(baseline.nclasses==2);
    ASSERT(vv_act.size()>=1);
    for(const activities *p_act: vv_act)
        ASSERT(p_act->is_f32());
    ASSERT(layer.strength>=1);

    const int64_t ntrn   = vv_act[0]->ntrn;
    const int64_t ntst   = vv_act[0]->ntst;
    const int64_t ntotal = ntrn+ntst;

    // allocate
    buf.alloc_and_clear(ntotal, layer.strength*layer.nsubclassifiers);
    pred = predictions(ntrn, ntst, 2, baseline.labels);

    // forward pass
    for(int64_t kidx=0; kidx<layer.strength*layer.nsubclassifiers; kidx++)
    {
        const activities *p_act = vv_act[kidx/layer.strength];
        const int64_t nchannels_cur = p_act->nchannels;
        const float *p_params = layer.get_coeff_ptr(kidx/layer.strength,kidx%layer.strength);

        float *p_sigmabuf = &buf.sigmakbuf[kidx*ntotal];
        for(int64_t i=0; i<ntotal; i++) // intercept
            p_sigmabuf[i] = p_params[nchannels_cur];
        for(int64_t t=0; t<nchannels_cur; t++) // linear terms
            fadd_simd(p_sigmabuf, p_act->get_f32(t), ntotal, p_params[t]);
    }
    for(int64_t i=0; i<ntotal; i++)
    {
        buf.sigmamaxbuf[i] = buf.sigmakbuf[i];
        buf.sigmaidx[i] = 0;
    }
    for(int64_t kidx=1; kidx<layer.strength*layer.nsubclassifiers; kidx++)
    {
        float *p_sigmabuf = &buf.sigmakbuf[kidx*ntotal];
        float *p_sigmamax = &buf.sigmamaxbuf[0];
        int   *p_sigmaidx = &buf.sigmaidx[0];
        for(int64_t i=0; i<ntotal; i++)
            if( p_sigmabuf[i]>p_sigmamax[i] )
            {
                p_sigmamax[i] = p_sigmabuf[i];
                p_sigmaidx[i] = kidx;
            }
    }
    for(int64_t i=0; i<ntotal; i++)
    {
        float v0 = buf.sigmamaxbuf[i], v1 = 0;
        float vm = fmaxf(v0,v1);
        v0 = expf(v0-vm);
        v1 = expf(v1-vm);
        pred.fpredictions[i+0*pred.stride] = v0/(v0+v1);
        pred.fpredictions[i+1*pred.stride] = v1/(v0+v1);
    }
}

gmaxout_postlayer::gmaxout_postlayer(std::vector<const activities*> vv_act, int s):gmaxout_postlayer()
{
    ASSERT(vv_act.size()>=1);
    for(const activities *p: vv_act)
        ASSERT(p->ntrn==vv_act[0]->ntrn && p->ntst==vv_act[0]->ntst);

    nsubclassifiers = vv_act.size();
    strength = s;
    int ctotal = 0;
    for(const activities *p_act: vv_act)
    {
        nchannels_cls.push_back(p_act->nchannels);
        ctotal += (p_act->nchannels+1)*strength;
    }
    coeffs.resize(ctotal, 0.0f);

    std::uniform_real_distribution<> uniform(-1.0e-5, +1.0e-5);
    std::mt19937  mt_rnd;
    std::seed_seq mt_seed{(unsigned int)1};
    mt_rnd.seed(mt_seed);
    for(int k=0; k<coeffs.size(); k++)
        coeffs[k] = uniform(mt_rnd);
}

float* gmaxout_postlayer::get_coeff_ptr(int classifier_idx, int sub_idx)
{
    return &coeffs[get_coeff_offset(classifier_idx, sub_idx)];
}

const float* gmaxout_postlayer::get_coeff_ptr(int classifier_idx, int sub_idx) const
{
    return &coeffs[get_coeff_offset(classifier_idx, sub_idx)];
}

int    gmaxout_postlayer::get_coeff_offset(int classifier_idx, int sub_idx) const
{
    ASSERT(classifier_idx>=0 && classifier_idx<nsubclassifiers);
    ASSERT(sub_idx>=0 && sub_idx<strength);
    int offs = 0;
    for(int k=0; k<classifier_idx; k++)
        offs += (1+nchannels_cls[k])*strength;
    offs += (1+nchannels_cls[classifier_idx])*sub_idx;
    return offs;
}

linear_predictor gmaxout_postlayer::convert_to_linpred() const
{
    ASSERT(nsubclassifiers==1 && strength==1);
    return linear_predictor::create_logit(coeffs, 2, nchannels_cls[0]);
}

linear_predictor gmaxout_postlayer::convert_to_linpred(int classifier_idx, int sub_idx) const
{
    const float *p_coeff = get_coeff_ptr(classifier_idx,sub_idx);
    std::vector<float> v_coeff(p_coeff, p_coeff+nchannels_cls[classifier_idx]+1);
    return linear_predictor::create_logit(v_coeff, 2, nchannels_cls[classifier_idx]);
}

predictions gmaxout_postlayer::predict(const predictions &baseline, const activities &act_in) const
{
    gmaxout_buf buf;
    predictions r_pred;
    ASSERT(nsubclassifiers==1);
    gmaxout_produce_predictions(*this, baseline, {&act_in}, buf, r_pred);
    return r_pred;
}

predictions gmaxout_postlayer::predict(const predictions &baseline, std::vector<const activities*> vv_act) const
{
    gmaxout_buf buf;
    predictions r_pred;
    ASSERT(nsubclassifiers==vv_act.size());
    gmaxout_produce_predictions(*this, baseline, vv_act, buf, r_pred);
    return r_pred;
}

std::vector<int> gmaxout_postlayer::get_winner_idx(const predictions &baseline, const activities &act_in) const
{
    gmaxout_buf buf;
    predictions r_pred;
    ASSERT(nsubclassifiers==1);
    gmaxout_produce_predictions(*this, baseline, {&act_in}, buf, r_pred);
    return buf.sigmaidx;
}

gmaxout_postlayer gmaxout_postlayer::apply_scaling(const std::vector<std::vector<float>> &vv_means, const std::vector<std::vector<float>> &vv_sigmas) const
{
    ASSERT(vv_means.size()==nsubclassifiers && vv_sigmas.size()==nsubclassifiers);

    gmaxout_postlayer out = *this;
    for(int k=0; k<nsubclassifiers; k++)
        for(int i=0; i<strength; i++)
        {
            ASSERT(vv_means[k].size()==nchannels_cls[k] && vv_sigmas[k].size()==nchannels_cls[k]);
            float cshift = 0;
            float *p_coeffs = out.get_coeff_ptr(k,i);
            for(int j=0; j<nchannels_cls[k]; j++)
            {
                cshift      -= p_coeffs[j]*vv_means[k][j]/vv_sigmas[k][j];
                p_coeffs[j] /= vv_sigmas[k][j];
            }
            p_coeffs[nchannels_cls[k]] += cshift;
        }
    return out;
}

gmaxout_postlayer gmaxout_postlayer::remove_scaling(const std::vector<std::vector<float>> &vv_means, const std::vector<std::vector<float>> &vv_sigmas) const
{
    ASSERT(vv_means.size()==nsubclassifiers && vv_sigmas.size()==nsubclassifiers);

    gmaxout_postlayer out = *this;
    for(int k=0; k<nsubclassifiers; k++)
        for(int i=0; i<strength; i++)
        {
            ASSERT(vv_means[k].size()==nchannels_cls[k] && vv_sigmas[k].size()==nchannels_cls[k]);
            float cshift = 0;
            float *p_coeffs = out.get_coeff_ptr(k,i);
            for(int j=0; j<nchannels_cls[k]; j++)
            {
                cshift      += p_coeffs[j]*vv_means[k][j];
                p_coeffs[j] *= vv_sigmas[k][j];
            }
            p_coeffs[nchannels_cls[k]] += cshift;
        }
    return out;
}

void gmaxout_postlayer::extend_with(const activities* p_act)
{
    nsubclassifiers++;
    nchannels_cls.push_back(p_act->nchannels);
    coeffs.resize(coeffs.size()+strength*(p_act->nchannels+1), 0.0f);
    for(int i=0; i<strength; i++)
    {
        float *p = get_coeff_ptr(nsubclassifiers-1,i);
        p[p_act->nchannels] = -10;
    }
}

void gmaxout_postlayer::extend_with(const gmaxout_postlayer &model)
{
    ASSERT(nsubclassifiers==1 && model.nsubclassifiers==1);
    ASSERT(nchannels_cls[0]==model.nchannels_cls[0]);
    for(int i=0; i<model.coeffs.size(); i++)
        coeffs.push_back(model.coeffs[i]);
    strength += model.strength;
}

void gmaxout_postlayer::increase_strength(int snew)
{
    int sold = strength;
    ASSERT(nsubclassifiers==1);
    ASSERT(snew>strength);
    coeffs.resize(coeffs.size()+(snew-strength)*(nchannels_cls.back()+1), 0.0f);
    strength = snew;
    for(int i=sold; i<snew; i++)
    {
        float *p = get_coeff_ptr(0,i);
        p[nchannels_cls.back()] = -10;
    }
}

void gmaxout_postlayer::import_predictor(int classifier_idx, int sub_idx, const linear_predictor &predictor)
{
    ASSERT(classifier_idx>=0 && classifier_idx<nsubclassifiers);
    ASSERT(sub_idx>=0 && sub_idx<strength);
    ASSERT(predictor.get_nclasses()==2 && predictor.get_logit_flag() && predictor.get_nchannels()==nchannels_cls[classifier_idx]);
    int offs = get_coeff_offset(classifier_idx,sub_idx);
    for(int k=0; k<nchannels_cls[classifier_idx]+1; k++)
        coeffs[offs+k] = predictor.get_coeff(k,0);
}


void train_gmaxout(gmaxout_postlayer &model,
                   std::vector<const activities*> vv_act,
                   const predictions &pred_target,
                   const int epoch_cnt,
                   const int batch_size,
                   const float mu,
                   const float alpha,
                   const std::tuple<float,float> reg12,
                   const std::string rep)
{
    train_gmaxout(model, vv_act, pred_target, std::vector<float>(pred_target.ntrn,1.0f),
                  epoch_cnt, batch_size, mu, alpha, reg12, rep);
}

void train_gmaxout(gmaxout_postlayer &model,
                   std::vector<const activities*> vv_act,
                   const predictions &pred_target,
                   std::vector<float> pred_w,
                   const int epoch_cnt,
                   const int batch_size,
                   const float mu,
                   const float alpha,
                   const std::tuple<float,float> reg12,
                   const std::string rep)
{
    ASSERT(vv_act.size()>=1);
    ASSERT(pred_target.nclasses==2);
    ASSERT(pred_w.size()==pred_target.ntrn);

    const int64_t ntrn = vv_act[0]->ntrn;
    const int64_t ntst = vv_act[0]->ntst;

    std::mt19937  mt_rnd;
    std::seed_seq mt_seed{(unsigned int)1};
    mt_rnd.seed(mt_seed);

    //
    // SGD - prepare data structures
    //
    ASSERT(batch_size>1);
    ASSERT(ntrn>=batch_size);
    std::vector<int64_t> idx;
    for(int64_t i=0; i<ntrn; i++)
        idx.push_back(i);
    std::vector<std::vector<float>> vv_means(vv_act.size()), vv_sigmas(vv_act.size());
    for(int k=0; k<vv_act.size(); k++)
    {
        const activities   &act_k = *(vv_act[k]);
        std::vector<float> &means_k = vv_means[k], &sigmas_k = vv_sigmas[k];
        for(int i=0; i<act_k.nchannels; i++)
        {
            const float *p_act = act_k.get_f32(i);
            double mean = 0;
            for(int64_t j=0; j<ntrn; j++)
                mean += p_act[j];
            mean /= ntrn;
            double sigma = 0;
            for(int64_t j=0; j<ntrn; j++)
            {
                double v = p_act[j]-mean;
                sigma += v*v;
            }
            sigma = sqrt(sigma/ntrn)+1.0e-6;
            means_k.push_back(mean);
            sigmas_k.push_back(sigma);
        }
    }

    //
    // Create initial model
    //
    gmaxout_postlayer model_nrm = model.remove_scaling(vv_means, vv_sigmas);
    const int total_power = model_nrm.strength*model_nrm.nsubclassifiers;

    //
    // Run SGD
    //
    gmaxout_buf glbl_buf;
    std::vector<float> ww(model_nrm.coeffs.size(),0.0f);
    for(int epoch_idx=0; epoch_idx<epoch_cnt; epoch_idx++)
    {
        //
        // Run SGD for one epoch
        //
        gmaxout_buf lcl_buf;
        std::vector<activities> act_batch(vv_act.size());
        std::vector<float> batch_w;
        predictions pred_batch, tmp_pred;
        activityf_vector ones(batch_size, 1.0f);
        std::shuffle(idx.begin(), idx.end(), mt_rnd);
        for(int64_t bidx=0; bidx<ntrn/batch_size; bidx++)
        {
            //
            // select subsample to process
            //
            std::vector<int> items(idx.begin()+bidx*batch_size, idx.begin()+(bidx+1)*batch_size);
            std::sort(items.begin(), items.end());
            for(int aidx=0; aidx<vv_act.size(); aidx++)
            {
                vv_act[aidx]->select_subsample(items, act_batch[aidx]);
                activities &act_k = act_batch[aidx];
                std::vector<float> &means_k = vv_means[aidx], &sigmas_k = vv_sigmas[aidx];
                for(int i=0; i<act_k.nchannels; i++)
                {
                    float *p_act = act_k.get_f32(i);
                    for(int64_t j=0; j<batch_size; j++)
                        p_act[j] = (p_act[j]-means_k[i])/sigmas_k[i];
                }
            }
            std::vector<const activities*> vv_batch;
            for(int aidx=0; aidx<vv_act.size(); aidx++)
                vv_batch.push_back(&act_batch[aidx]);
            pred_batch = pred_target.select_subsample(items);
            batch_w.clear();
            for(int i: items)
                batch_w.push_back(pred_w[i]);

            //
            // Compute predictions for current mini-batch
            //
            // NOTE: this function internally sets to zero all fields of lcl_buf
            //
            gmaxout_produce_predictions(model_nrm, pred_batch, vv_batch, lcl_buf, tmp_pred);

            //
            // Compute gradient
            //
            const float *pred0 = &tmp_pred.fpredictions[0];
            for(int64_t i=0; i<batch_size; i++)
            {
                float v = pred0[i]-(tmp_pred.labels[i]==0 ? 1 : 0);
                v *= sqrtf(batch_w[i]);
                v = v*(model_nrm.leak_exp+pred0[i]*(1.0f-pred0[i]));
                lcl_buf.dedn[i] = v;
                lcl_buf.dednk[i+lcl_buf.sigmaidx[i]*batch_size] = v;
                for(int64_t k=0; k<total_power; k++)
                    lcl_buf.dednk[i+k*batch_size] += model_nrm.leak_max*v;
            }
            std::vector<float> grad(model_nrm.coeffs.size(),0.0f);
            for(int64_t kidx=0; kidx<total_power; kidx++)
            {
                const activities *p_act = vv_batch[kidx/model_nrm.strength];
                const float *p_dednk = &lcl_buf.dednk[kidx*batch_size];
                float *p_grad_k = &grad[model_nrm.get_coeff_offset(kidx/model_nrm.strength,kidx%model_nrm.strength)];
                for(int64_t i=0; i<p_act->nchannels+1; i++)
                {
                    const float *p_act_i = i<p_act->nchannels ? p_act->get_f32(i) : &ones[0];
                    p_grad_k[i] = fdot_simd(p_dednk, p_act_i, batch_size)/batch_size;
                }
            }

            //
            // Add regularizer
            //
            float reg_lambda1, reg_lambda2;
            std::tie(reg_lambda1, reg_lambda2) = reg12;
            reg_lambda1 /= model.strength*model.nsubclassifiers;
            reg_lambda2 /= model.strength*model.nsubclassifiers;
            for(int64_t i=0; i<grad.size(); i++)
            {
                float c  = model_nrm.coeffs[i];
                float sc = c>=0.0f ? +1.0f : -1.0f;
                grad[i] += reg_lambda1*sc;
                grad[i] += reg_lambda2*c;
            }

            //
            // Gradient with momentum
            //
            for(int i=0; i<model_nrm.coeffs.size(); i++)
                ww[i] = mu*ww[i]+(1.0-mu)*(alpha*grad[i]);
            for(int i=0; i<model_nrm.coeffs.size(); i++)
                model_nrm.coeffs[i] -= ww[i];
        }


        //
        // Report
        //
        model = model_nrm.apply_scaling(vv_means, vv_sigmas);
        gmaxout_produce_predictions(model, pred_target, vv_act, glbl_buf, tmp_pred);
        std::vector<float> wrate(total_power,0.0f);
        std::vector<float> wsepr(total_power,0.0f);
        std::vector<float> wtotl(total_power,1.0e-20f);
        int wtotal = 0;
        for(int64_t i=0; i<ntrn; i++)
            if( pred_target.labels[i]==0 )
            {
                float mx = glbl_buf.sigmakbuf[i+glbl_buf.sigmaidx[i]*(ntrn+ntst)];
                float sv = -999999.0;
                for(int k=0; k<total_power; k++)
                {
                    float vv = glbl_buf.sigmakbuf[i+k*(ntrn+ntst)];
                    if( vv<mx )
                        sv = fmaxf(sv, vv);
                }
                float em = sigma0(mx), es = sigma0(sv);
                wrate[glbl_buf.sigmaidx[i]] += 1.0f;
                wsepr[glbl_buf.sigmaidx[i]] += em-es;
                wtotl[glbl_buf.sigmaidx[i]] += 1.0f;
                wtotal++;
            }
        printf("[%12s:%04d]  TrnRms=%.4f  TstRms=%.4f  TstCls=%.4f\n",
               rep.c_str(),
               int(epoch_idx),
               (double)tmp_pred.trn_rms(),
               (double)tmp_pred.tst_rms(),
               (double)tmp_pred.tst_cls());
        fflush(stdout);
    }
}

class grbf_buf
{
public:
    // forward pass
    std::vector<activityf_vector> net_i;
    std::vector<activityf_vector> dednet_i;
    activityf_vector net_mx;
    std::vector<int> mx_idx;

    // derivative of error with respect to sigmoid input (stored in sigmamaxbuf)
    //activityf_vector dedn, dednk;

    //
    void alloc_and_clear(int n, int nmodels)
    {
        net_i.resize(nmodels);
        dednet_i.resize(nmodels);
        for(int k=0; k<nmodels; k++)
        {
            net_i[k].resize(n);
            dednet_i[k].resize(n);
            memset_vector(net_i[k],    0);
            memset_vector(dednet_i[k], 0);
        }
        net_mx.resize(n);
        mx_idx.resize(n);
        memset_vector(net_mx, 0);
        memset_vector(mx_idx, 0);
    }
};

void export_coeffs(const std::vector<rbf_predictor> &models, std::vector<float> &coeffs)
{
    coeffs.clear();
    for(const rbf_predictor &model: models)
        for(float c: model.coeffs)
            coeffs.push_back(c);
}

void import_coeffs(const std::vector<float> &coeffs, std::vector<rbf_predictor> &models)
{
    int _cnt = 0;
    for(const rbf_predictor &model: models)
        _cnt += model.coeffs.size();
    ASSERT(_cnt==coeffs.size());

    int src_offs = 0;
    for(rbf_predictor &model: models)
    {
        for(int i=0; i<model.coeffs.size(); i++)
            model.coeffs[i] = coeffs[src_offs+i];
        src_offs += model.coeffs.size();
    }
    ASSERT(src_offs==coeffs.size());
}

std::vector<float> unscale_coeffs(const std::vector<float> &coeffs,
                                  const std::vector<std::vector<float>> &vv_means,
                                  const std::vector<std::vector<float>> &vv_sigmas)
{
    ASSERT(vv_means.size()==vv_sigmas.size());

    const int nmodels = vv_means.size();
    int coeff_offs = 0;
    std::vector<float> out;
    for(int k=0; k<nmodels; k++)
    {
        ASSERT(vv_means[k].size()==vv_sigmas[k].size());
        const int nchannels = vv_means[k].size();
        const float *p_src = &coeffs[coeff_offs];
        float cshift = 0;
        for(int j=0; j<nchannels; j++)
        {
            out.push_back(p_src[j]*vv_sigmas[k][j]);
            cshift += p_src[j]*vv_means[k][j];
        }
        out.push_back(p_src[nchannels]+cshift);
        coeff_offs += nchannels+1;
    }
    ASSERT(coeff_offs==coeffs.size());
    return out;
}

std::vector<float>   scale_coeffs(const std::vector<float> &coeffs,
                                  const std::vector<std::vector<float>> &vv_means,
                                  const std::vector<std::vector<float>> &vv_sigmas)
{
    ASSERT(vv_means.size()==vv_sigmas.size());

    const int nmodels = vv_means.size();
    int coeff_offs = 0;
    std::vector<float> out;
    for(int k=0; k<nmodels; k++)
    {
        ASSERT(vv_means[k].size()==vv_sigmas[k].size());
        const int nchannels = vv_means[k].size();
        const float *p_src = &coeffs[coeff_offs];
        float cshift = 0;
        for(int j=0; j<nchannels; j++)
        {
            out.push_back(p_src[j]/vv_sigmas[k][j]);
            cshift -= p_src[j]*vv_means[k][j]/vv_sigmas[k][j];
        }
        out.push_back(p_src[nchannels]+cshift);
        coeff_offs += nchannels+1;
    }
    ASSERT(coeff_offs==coeffs.size());
    return out;
}

static void grbf_produce_predictions(const std::vector<float> &coeffs,
                                     const predictions &baseline,
                                     std::vector<const activities*> v_pact,
                                     grbf_buf &buf,
                                     predictions &pred)
{
    ASSERT(baseline.nclasses==2);
    ASSERT(v_pact.size()>=1);
    for(const activities *p_act: v_pact)
        ASSERT(p_act->is_f32());
    int _cnt = 0;
    for(int64_t kidx=0; kidx<v_pact.size(); kidx++)
        _cnt += v_pact[kidx]->nchannels+1;
    ASSERT(_cnt==coeffs.size());

    const int64_t nmodels= v_pact.size();
    const int64_t ntrn   = v_pact[0]->ntrn;
    const int64_t ntst   = v_pact[0]->ntst;
    const int64_t ntotal = ntrn+ntst;

    // allocate
    buf.alloc_and_clear(ntotal, nmodels);
    pred = predictions(ntrn, ntst, 2, baseline.labels);

    // forward pass
    int coeff_offs = 0;
    for(int64_t kidx=0; kidx<nmodels; kidx++)
    {
        const activities *p_act = v_pact[kidx];
        const int64_t nchannels_cur = p_act->nchannels;
        const float *p_params = &coeffs[coeff_offs];
        ASSERT(coeff_offs+nchannels_cur-1<=coeffs.size()-1);

        float *p_net = &buf.net_i[kidx][0];
        for(int64_t i=0; i<ntotal; i++)        // intercept
            p_net[i] = p_params[nchannels_cur];
        for(int64_t t=0; t<nchannels_cur; t++) // linear terms
            fadd_simd(p_net, p_act->get_f32(t), ntotal, p_params[t]);

        coeff_offs += nchannels_cur+1;
    }
    for(int64_t i=0; i<ntotal; i++)
    {
        float *p_net    = &buf.net_i[0][0];
        float *p_net_mx = &buf.net_mx[0];
        int   *p_idx    = &buf.mx_idx[0];
        p_net_mx[i] = p_net[i];
        p_idx[i] = 0;
    }
    for(int64_t kidx=1; kidx<nmodels; kidx++)
    {
        float *p_net    = &buf.net_i[0][0];
        float *p_net_mx = &buf.net_mx[0];
        int   *p_idx    = &buf.mx_idx[0];
        for(int64_t i=0; i<ntotal; i++)
            if( p_net[i]>p_net_mx[i] )
            {
                p_net_mx[i] = p_net[i];
                p_idx[i] = kidx;
            }
    }
    for(int64_t i=0; i<ntotal; i++)
    {
        //pred.fpredictions[i] = buf.net_mx[i];
        pred.fpredictions[i] = sigma0(buf.net_mx[i]);//!!!!!!!!!!!!!!!!!!
        pred.fpredictions[i+pred.stride] = 1.0f-pred.fpredictions[i];
    }
}

void train_grbf(std::vector<rbf_predictor> &models,
                std::vector<const activities*> vv_act_in,
                const predictions &pred_target,
                const int epoch_cnt,
                const int batch_size,
                const float mu,
                const float alpha,
                const std::tuple<float,float> reg12,
                const std::tuple<float,float> leak_me)
{
    ASSERT(vv_act_in.size()>=1);
    ASSERT(pred_target.nclasses==2);
    ASSERT(models.size()==vv_act_in.size());

    const int64_t nmodels = models.size();
    const int64_t ntrn = vv_act_in[0]->ntrn;
    const int64_t ntst = vv_act_in[0]->ntst;

#if 0
    ASSERT(nmodels==1);
    rbf_predictor &model = models[0];
    const activities &act_in = *(vv_act_in[0]);
    activities act_bf;
    model.compute_bf(act_in, act_bf);
    float lambda_reg = std::get<1>(reg12);
    linear_evaluator lin_eval(pred_target,act_bf.nchannels,sqrtf(lambda_reg*ntrn));
    for(int i=0; i<act_bf.nchannels; i++)
    {
        const float *p_data = act_bf.get_f32(i);
        activityf_vector act(p_data, p_data+ntrn+ntst);
        lin_eval.append_activity(act);
        printf("[%04d]  TrnRms=%.4f  TstRms=%.4f  TstCls=%.4f\n",
               int(i),
               (double)lin_eval.trn_rms(),
               (double)lin_eval.tst_rms(),
               (double)lin_eval.tst_cls());
        fflush(stdout);
    }
    ASSERT(0);
#endif

    const float   leak_max = std::get<0>(leak_me);
    const float   leak_exp = std::get<1>(leak_me);

    std::mt19937  mt_rnd;
    std::seed_seq mt_seed{(unsigned int)1};
    mt_rnd.seed(mt_seed);

    //
    // Compute activities of basis functions
    //
    std::vector<activities> act_bf(nmodels);
    std::vector<const activities*> act_pbf;
    for(int i=0; i<nmodels; i++)
    {
        models[i].compute_bf(*(vv_act_in[i]), act_bf[i]);
        act_pbf.push_back(&act_bf[i]);
    }

    //
    // SGD - prepare data structures
    //
    ASSERT(batch_size>1);
    ASSERT(ntrn>=batch_size);
    std::vector<int64_t> idx;
    for(int64_t i=0; i<ntrn; i++)
        idx.push_back(i);
    std::vector<std::vector<float>> vv_means(nmodels), vv_sigmas(nmodels);
    for(int k=0; k<nmodels; k++)
    {
        const activities   &act_k = act_bf[k];
        std::vector<float> &means_k = vv_means[k], &sigmas_k = vv_sigmas[k];
        for(int i=0; i<act_k.nchannels; i++)
        {
            const float *p_act = act_k.get_f32(i);
            double mean = 0;
            for(int64_t j=0; j<ntrn; j++)
                mean += p_act[j];
            mean /= ntrn;
            double sigma = 0;
            for(int64_t j=0; j<ntrn; j++)
            {
                double v = p_act[j]-mean;
                sigma += v*v;
            }
            sigma = sqrt(sigma/ntrn)+1.0e-6;
            means_k.push_back(mean);
            sigmas_k.push_back(sigma);
        }
    }

    //
    // Initial model, its gradient, mu
    //
    int model_dim = 0;
    for(const rbf_predictor &p: models)
        model_dim += p.ncenters+1;
    std::vector<float> coeffs_nrm, coeffs_raw, ww(model_dim,0.0f);
    export_coeffs(models, coeffs_raw);
    coeffs_nrm = unscale_coeffs(coeffs_raw, vv_means, vv_sigmas);

    //
    // Run SGD
    //
    grbf_buf glbl_buf;
    for(int epoch_idx=0; epoch_idx<epoch_cnt; epoch_idx++)
    {
        //
        // Run SGD for one epoch
        //
        grbf_buf lcl_buf;
        std::vector<activities> act_batch(nmodels);
        predictions pred_batch, tmp_pred;
        activityf_vector ones(batch_size, 1.0f);
        std::shuffle(idx.begin(), idx.end(), mt_rnd);
        for(int64_t bidx=0; bidx<ntrn/batch_size; bidx++)
        {
            //
            // select subsample to process
            //
            std::vector<int> items(idx.begin()+bidx*batch_size, idx.begin()+(bidx+1)*batch_size);
            std::sort(items.begin(), items.end());
            for(int aidx=0; aidx<nmodels; aidx++)
            {
                act_bf[aidx].select_subsample(items, act_batch[aidx]);
                activities &act_k = act_batch[aidx];
                std::vector<float> &means_k = vv_means[aidx], &sigmas_k = vv_sigmas[aidx];
                for(int i=0; i<act_k.nchannels; i++)
                {
                    float *p_act = act_k.get_f32(i);
                    for(int64_t j=0; j<batch_size; j++)
                        p_act[j] = (p_act[j]-means_k[i])/sigmas_k[i];
                }
            }
            std::vector<const activities*> vv_batch;
            for(int aidx=0; aidx<nmodels; aidx++)
                vv_batch.push_back(&act_batch[aidx]);
            pred_batch = pred_target.select_subsample(items);

            //
            // Compute predictions for current mini-batch
            //
            // NOTE: this function internally sets to zero all fields of lcl_buf
            //
            grbf_produce_predictions(coeffs_nrm, pred_batch, vv_batch, lcl_buf, tmp_pred);

            //
            // Compute gradient
            //
            std::vector<float> grad(model_dim,0.0f);
            const float *pred0 = &tmp_pred.fpredictions[0];
            for(int64_t i=0; i<batch_size; i++)
            {
                float v = pred0[i]-(tmp_pred.labels[i]==0 ? 1 : 0);
                v = v*(leak_exp+pred0[i]*(1.0f-pred0[i]));//!!!!!!!!!!!!!!!!!!
                lcl_buf.dednet_i[lcl_buf.mx_idx[i]][i] += v;
                for(int64_t k=0; k<nmodels; k++)
                    lcl_buf.dednet_i[k][i] += leak_max*v;
            }
            for(int64_t kidx=0,coeff_offs=0; kidx<nmodels; kidx++)
            {
                const activities *p_act = vv_batch[kidx];
                const float *p_dednet = &lcl_buf.dednet_i[kidx][0];
                float *p_grad_k = &grad[coeff_offs];
                for(int64_t i=0; i<p_act->nchannels+1; i++)
                {
                    const float *p_act_i = i<p_act->nchannels ? p_act->get_f32(i) : &ones[0];
                    p_grad_k[i] = fdot_simd(p_dednet, p_act_i, batch_size)/batch_size;
                }
                coeff_offs += models[kidx].coeffs.size()+1;
            }

            //
            // Add regularizer
            //
            float reg_lambda1, reg_lambda2;
            std::tie(reg_lambda1, reg_lambda2) = reg12;
            reg_lambda1 /= nmodels;
            reg_lambda2 /= nmodels;
            for(int64_t i=0; i<model_dim; i++)
            {
                float c  = coeffs_nrm[i];
                float sc = c>=0.0f ? +1.0f : -1.0f;
                grad[i] += reg_lambda1*sc;
                grad[i] += reg_lambda2*c;
            }

            //
            // Gradient descent with momentum
            //
            for(int i=0; i<model_dim; i++)
            {
                ww[i] = mu*ww[i]+(1.0-mu)*(alpha*grad[i]);
                coeffs_nrm[i] -= ww[i];
            }
        }

        //
        // Export
        //
        coeffs_raw = scale_coeffs(coeffs_nrm, vv_means, vv_sigmas);
        import_coeffs(coeffs_raw, models);

        //
        // Report
        //
        grbf_produce_predictions(coeffs_raw, pred_target, act_pbf, glbl_buf, tmp_pred);
        std::vector<float> wtotl(nmodels,1.0e-20f);
        std::vector<float> wsepr(nmodels,0.0f);
        int wtotal = 0;
        for(int64_t i=0; i<ntrn; i++)
            if( pred_target.labels[i]==0 )
            {
                float mx = glbl_buf.net_mx[i];
                float sv = -999999.0;
                for(int k=0; k<nmodels; k++)
                {
                    float vv = glbl_buf.net_i[k][i];
                    if( vv<mx )
                        sv = fmaxf(sv, vv);
                }
                float em = sigma0(mx), es = sigma0(sv);
                wsepr[glbl_buf.mx_idx[i]] += em-es;
                wtotl[glbl_buf.mx_idx[i]] += 1.0f;
                wtotal++;
            }
        printf("[maxksigma2-trn:%04d]  TrnRms=%.4f  TstRms=%.4f  TstCls=%.4f    (",
               int(epoch_idx),
               (double)tmp_pred.trn_rms(),
               (double)tmp_pred.tst_rms(),
               (double)tmp_pred.tst_cls());
        for(int j=0; j<nmodels; j++)
            printf("%s%4.1f%%/sep%.2f", j==0 ? "" : "  ",
                   double(100.0*wtotl[j])/double(wtotal),
                   double(wsepr[j])/double(wtotl[j]));
        printf(")\n");
        fflush(stdout);
    }
}

class multiscale_dataset_binary_env
{
public:
    const raw_dataset       &ds_src;
    const std::vector<std::tuple<int,int>>  &info;
    const int               dst_width;
    raw_dataset             &ds_dst;

    multiscale_dataset_binary_env(const raw_dataset &src, const std::vector<std::tuple<int,int>> &ii, int dw, raw_dataset &dst):
        ds_src(src),info(ii),dst_width(dw),ds_dst(dst){}
};

static void* multiscale_dataset_binary_apply(void *xptr, int idx)
{
    multiscale_dataset_binary_env &env = *((multiscale_dataset_binary_env*)xptr);

    const raw_dataset::entry &e_src = env.ds_src.entries[std::get<0>(env.info[idx])];
    raw_dataset::entry &e = env.ds_dst.entries[idx];
    e.class_id = e_src.class_id;
    e.img = e_src.img.central_crop(std::get<1>(env.info[idx])).lanczos_resize_to(env.dst_width,env.dst_width);
    return NULL;
}


std::tuple<raw_dataset,std::vector<int>> multiscale_dataset_binary(
        const raw_dataset &ds_src,
        int src_width, int src_aoi,
        int dst_width, std::vector<int> dst_aoi)
{
    ASSERT(ds_src.nclasses==2);
    raw_dataset ds_out;
    std::vector<int> splits;
    ds_out.nclasses = ds_src.nclasses;
#if 0
    ds_out.ntrn = 0;
    ds_out.ntst = 0;
    for(int img_idx=0; img_idx<ds_src.ntrn+ds_src.ntst; img_idx++)
    {
        const raw_dataset::entry &e_src = ds_src.entries[img_idx];
        ASSERT(e_src.img.width==src_width && e_src.img.height==src_width);
        for(int scale_idx=0; scale_idx<dst_aoi.size(); scale_idx++)
        {
            int aoi = dst_aoi[scale_idx];
            int dst_unscaled_size = roundf(float(dst_width)/aoi*src_aoi);
            raw_dataset::entry e;
            e.class_id = e_src.class_id;
            e.img = e_src.img.central_crop(dst_unscaled_size).lanczos_resize_to(dst_width,dst_width);
            if( img_idx<ds_src.ntrn )
                ds_out.append_trn(e);
            else
                ds_out.append_tst(e);
            splits.push_back(e_src.class_id==0 ? scale_idx : -1);
        }
    }
#endif

    std::vector<std::tuple<int,int>> info;
    ds_out.ntrn = ds_src.ntrn*dst_aoi.size();
    ds_out.ntst = ds_src.ntst*dst_aoi.size();
    ds_out.entries.resize(ds_out.ntrn+ds_out.ntst);
    for(int img_idx=0; img_idx<ds_src.ntrn+ds_src.ntst; img_idx++)
    {
        const raw_dataset::entry &e_src = ds_src.entries[img_idx];
        ASSERT(e_src.img.width==src_width && e_src.img.height==src_width);
        for(int scale_idx=0; scale_idx<dst_aoi.size(); scale_idx++)
        {
            int aoi = dst_aoi[scale_idx];
            int dst_unscaled_size = roundf(float(dst_width)/aoi*src_aoi);
            splits.push_back(e_src.class_id==0 ? scale_idx : -1);
            info.push_back(std::make_tuple(img_idx,dst_unscaled_size));
        }
    }
    multiscale_dataset_binary_env env(ds_src, info, dst_width, ds_out);
    //parallel_invoke(multiscale_dataset_binary_apply, &env);
    parallel_for(multiscale_dataset_binary_apply, &env, info.size());

    return std::make_tuple(ds_out,splits);
}

}
