/*************************************************************************
StrongNet/Contour Engine training framework.
Copyright (c) 2020 Sergey Bochkanov (ALGLIB project).

>>> SOURCE LICENSE >>>
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation (www.fsf.org); either version 2 of the 
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

A copy of the GNU General Public License is available at
http://www.fsf.org/licensing/licenses
>>> END OF LICENSE >>>
*************************************************************************/
#ifndef SNET_H
#define SNET_H

#include "snetcommon.h"
#include "layer1.h"
#include "layer2.h"
#include "layer2c.h"
#include "layer4.h"
#include "3rdparty/alglib/dataanalysis.h"

namespace strongnet
{

void singlecolumn(const predictions &baseline,
                 const activities &act_v1,
                 int v2_width,
                 int v4_width,
                 double lambda_reg,
                 int seed,
                 //net_midlayer     &ml_out,
                 //activities       &act_out,
                 predictions &pred_out);

class net_multicolumn
{
friend void serialize(FILE *f, const net_multicolumn *ptr);
friend void unserialize(FILE *f, net_multicolumn *ptr);
private:
    int nclasses;
    net_v1layer v1;
    std::vector<net_midlayer>       v2_columns;
    std::vector<net_midlayer>       v4_columns;
    std::vector<linear_predictor>   predictors;
    linear_predictor                final_predictor;
    void params_and_flops(const net_midlayer &layer, double &params, double &flops) const;
public:
    class settings
    {
    public:
        int seed;
        int batch_size;
        float lambda_reg;
        float omega_rat;
        bool use_exacting_layer;
        int v2_maxdist;

        settings():
            seed(1),batch_size(50),use_exacting_layer(false),lambda_reg(1),omega_rat(0.000001),
            v2_maxdist(-1){}
    };

    //
    net_multicolumn(){ nclasses = 0; }
    net_multicolumn(std::string filename);

    // training functions
    std::vector<predictions> train(
            const raw_dataset &ds_raw,
            const net_v1layer &layer1,
            const std::vector<int> columns_to_train,
            int v2_width,
            int v4_width,
            settings s);
    predictions train(
            const raw_dataset &ds_raw,
            const net_v1layer &layer1,
            int net_width,
            settings s);

    // predict
    predictions apply(const raw_dataset &ds_raw, const activities &v1_act) const;
    predictions apply(const raw_dataset &ds_raw) const;

    // concatenation
    static net_multicolumn concat(const std::vector<net_multicolumn> &v);

    // debug reports
    void dbg_analyze_stats() const;
};
void serialize(FILE *f, const net_multicolumn *ptr);
void unserialize(FILE *f, net_multicolumn *ptr);

class rbf_predictor
{
public:
    int64_t                         nchannels_in;
    std::vector<float>              prescaler;
    int64_t                         ncenters;
    std::vector<float>              centers;
    int                             nprojections;
    std::vector<float>              prj;
    float                           rad;
    std::vector<float>              coeffs;

public:
    rbf_predictor():nchannels_in(0),ncenters(0),nprojections(0),rad(0.0f){}

    // creates predictor
    void init(int nin, const std::vector<float> &ps, int nc, const std::vector<float> &ct, int np, const std::vector<float> &p, float rad, const std::vector<float> &cf);

    // creates untrained predictor (centers/projections are prepared, but coefficients are zero)
    static rbf_predictor create(const raw_dataset &ds,
                                int class_idx,
                                const activities &act_in,
                                const std::vector<float> &ref_dir,
                                int rbf_centers,
                                float rbf_rad,
                                int rbf_top_dim,
                                int rbf_bottom_dim);
    static rbf_predictor create(const raw_dataset &ds,
                                int class_idx,
                                const activities &act_in,
                                int rbf_centers,
                                float rbf_rad);

    //
    predictions predict(const predictions &baseline, const activities& act, predict_t ptype = predict_t::prob) const;

    //
    void compute_bf(const activities& act_in, activities &act_bf) const;
    void compute_basis_function(const activities& act_in, int center_idx, activityf_vector &out, activityf_vector &buf) const;

    friend void serialize(FILE *f, const rbf_predictor *ptr);
    friend void unserialize(FILE *f, rbf_predictor *ptr);
};
void serialize(FILE *f, const rbf_predictor *ptr);
void unserialize(FILE *f, rbf_predictor *ptr);

class knn_predictor
{
public:
    int64_t                         nchannels_in;
    std::vector<float>              prescaler;
    int64_t                         ncenters;
    std::vector<float>              centers;
    std::vector<int>                labels;
    int64_t                         nneighbors;

    //
    alglib::kdtree tree;
    void finalize_construction();
public:
    knn_predictor():nchannels_in(0),ncenters(0),nneighbors(0){}

    // creates untrained predictor (centers/projections are prepared, but coefficients are zero)
    static knn_predictor create(const raw_dataset &ds,
                                int class_idx,
                                const activities &act_in,
                                int nk);

    //
    predictions predict(const predictions &baseline, const activities& act, predict_t ptype = predict_t::prob) const;

    friend void serialize(FILE *f, const knn_predictor *ptr);
    friend void unserialize(FILE *f, knn_predictor *ptr);
};
void serialize(FILE *f, const knn_predictor *ptr);
void unserialize(FILE *f, knn_predictor *ptr);

class superlogit2_predictor
{
public:
    int64_t                         nchannels_in;
    std::vector<std::vector<float>> coeffs;
public:
    superlogit2_predictor():nchannels_in(0){}

    // info
    float get_l1_factor() const;

    // creates untrained predictor (centers/projections are prepared, but coefficients are zero)
    static superlogit2_predictor create_empty(const activities &act_in);

    // appends coefficients
    void append1(const std::vector<float> &subpredictor);
    void append1(const linear_predictor   &subpredictor);
    void appendk(const std::vector<float> &subpredictor);

    //
    predictions predict(const predictions &baseline, const activities& act, predict_t ptype = predict_t::prob) const;
    conv_predictions conv_predict1(const conv_activities &act, const calibrator &cal) const;

    friend void serialize(FILE *f, const superlogit2_predictor *ptr);
    friend void unserialize(FILE *f, superlogit2_predictor *ptr);
};
void serialize(FILE *f, const superlogit2_predictor *ptr);
void unserialize(FILE *f, superlogit2_predictor *ptr);

class conv_subcolumn
{
friend void serialize(FILE *f, const conv_subcolumn *ptr);
friend void unserialize(FILE *f, conv_subcolumn *ptr);
private:
    // convolutional pathway
    select_t                        conv_selector;
    int                             conv_select0, conv_select1;
    split_t                         conv_subdivide_in;
    int                             conv_overlap;
    conv_midlayer                   conv_v2c;
    split_t                         conv_subdivide_out;
    accumulate_t                    conv_accumulate_out;
    net_midlayer                    conv_v4;

    // nonconvolutional pathway
    select_t                        nonconv_selector;
    int                             nonconv_select0, nonconv_select1;
    net_midlayer                    nonconv_v2;
    net_midlayer                    nonconv_v4;

    // protection
    float                           adv_clip_level;
    float                           adv_frac_chosen;
    float                           adv_param0;
    float                           adv_param1;
    float                           adv_param2;
    float                           adv_param3;
    float                           adv_param4;
    float                           adv_param5;

    // predictor
    int                             predictor_type; // 0 for linear_predictor, 1 for rbf_predictor, 2 for knn predictor, 3 for superlogit2
    linear_predictor                linpred;
    rbf_predictor                   rbfpred;
    knn_predictor                   knnpred;
    superlogit2_predictor           superpred;

    // calibrator
    calibrator                      cal;
public:
    conv_subcolumn():conv_selector(select_t::none),nonconv_selector(select_t::none)
    { adv_clip_level = adv_frac_chosen = adv_param0 =  adv_param1 =  adv_param2 = adv_param3 = adv_param4 = adv_param5 = 0; }
    explicit conv_subcolumn(std::string filename);

    // info
    float get_l1_factor() const;
    float get_sparse_flops() const;
    int   get_v1_window_dim0(const net_v1layer &v1) const;
    int   get_v1_window_dim1(const net_v1layer &v1) const;
    int   get_v1_downsample_lo(const net_v1layer &v1) const;
    int   get_v1_downsample_hi(const net_v1layer &v1) const;
    int   get_v2_window_height(const net_v1layer &v1) const;
    int   get_v2_window_width(const net_v1layer &v1) const;

    // predict/process
    predictions       apply(const predictions &zero_baseline, activities &_v1_act) const;
    predictions   apply_log(const predictions &zero_baseline, activities &_v1_act) const;
    predictions apply_logit(const predictions &zero_baseline, activities &_v1_act) const;
    conv_predictions conv_apply0(const net_v1layer &v1, conv_activities &_v1_act) const;    // returns single-class prediction plane
    void process_v2v4(activities &_v1_act, activities &act_total) const;
    void process_v2v4(conv_activities &_v1_act, conv_activities &act_total) const;
    void protect_activities(activities &act) const;
    void protect_activities(conv_activities &act) const;
    static void protect_activities(activities      &act,  const std::tuple<float,float> protection_coeffs); // clip_level, frac_chosen
    static void protect_activities(conv_activities &cact, float clip_level); // clip_level

    // external initialization
    void init(split_t split_in, int overlap, const conv_midlayer &v2c, split_t split_out, accumulate_t acc_out,
              net_midlayer v4,
              const linear_predictor &pred, const calibrator &c);
    void init(int window_size,
              select_t select_nonconv,
              std::tuple<net_midlayer,net_midlayer> nonconv, // v2,v4
              select_t select_conv,
              std::tuple<split_t,int,conv_midlayer,split_t,accumulate_t,net_midlayer> conv, // <split_in,overlap,v2c,split_out,acc_out,v4c>
              const std::tuple<float,float> protection_coeffs, // clip_lo, clip_hi
              const linear_predictor &pred, const calibrator &c);
    void init(int window_size,
              select_t select_nonconv,
              std::tuple<net_midlayer,net_midlayer> nonconv, // v2,v4
              select_t select_conv,
              std::tuple<split_t,int,conv_midlayer,split_t,accumulate_t,net_midlayer> conv, // <split_in,overlap,v2c,split_out,acc_out,v4c>
              const rbf_predictor &pred, const calibrator &c);
    void init(int window_size,
              select_t select_nonconv,
              std::tuple<net_midlayer,net_midlayer> nonconv, // v2,v4
              select_t select_conv,
              std::tuple<split_t,int,conv_midlayer,split_t,accumulate_t,net_midlayer> conv, // <split_in,overlap,v2c,split_out,acc_out,v4c>
              const knn_predictor &pred, const calibrator &c);
    void init_partial(select_t select_nonconv,
              std::tuple<int, net_midlayer, net_midlayer> nonconv, // windowsize,v2,v4
              select_t select_conv,
              std::tuple<int, split_t,int,conv_midlayer,split_t,accumulate_t,net_midlayer> conv); // <window_size,split_in,overlap,v2c,split_out,acc_out,v4c>
    void rewrite_predictor(const linear_predictor &pred, const calibrator &c);
    void rewrite_predictor(const rbf_predictor &pred, const calibrator &c);
    void rewrite_predictor(const knn_predictor &pred, const calibrator &c);
    void rewrite_predictor(const superlogit2_predictor &pred, const calibrator &c);
    void set_clipping(float level);

    // access
    const linear_predictor&      get_predictor() const;
    const rbf_predictor&         get_rbf_predictor() const;
    const knn_predictor&         get_knn_predictor() const;
    const superlogit2_predictor& get_superlogit2_predictor() const;
};
void serialize(FILE *f, const conv_subcolumn *ptr);
void unserialize(FILE *f, conv_subcolumn *ptr);

class x1_column
{
private:
    std::vector<conv_subcolumn> subcolumns;

    predictions _apply(const predictions &zero_baseline, activities &_v1_act, predict_t ptype) const;
public:
    friend void serialize(FILE *f, const x1_column *ptr);
    friend void unserialize(FILE *f, x1_column *ptr);

    x1_column(){}

    // info
    float get_l1_factor() const;
    float get_sparse_flops() const;
    int   get_nsubcolumns() const;
    int   get_v1_window_dim0(const net_v1layer &v1) const;      // returns size of input image (V1 input)
    int   get_v1_window_dim1(const net_v1layer &v1) const;      // returns size of input image (V1 input)
    int   get_v1_downsample_lo(const net_v1layer &v1) const;    // returns lowest downsampling  (by parts of V1 used by columns)
    int   get_v1_downsample_hi(const net_v1layer &v1) const;    // returns highest downsampling  (by parts of V1 used by columns)
    int   get_v2_window_height(const net_v1layer &v1) const;    // returns effective window height of V2 input (downsampled V1 output subject to select conditions)
    int   get_v2_window_width(const net_v1layer &v1) const;     // returns effective window width  of V2 input (downsampled V1 output subject to select conditions)
    conv_subcolumn& get_subcolumn(int col_idx);

    // predict
    predictions       apply(const predictions &zero_baseline, activities &_v1_act) const;
    predictions   apply_log(const predictions &zero_baseline, activities &_v1_act) const;
    predictions apply_logit(const predictions &zero_baseline, activities &_v1_act) const;
    conv_predictions conv_apply0(const net_v1layer &v1, conv_activities &_v1_act) const;    // returns single-class prediction plane

    // external initialization
    void append_subcolumn(const conv_subcolumn &c);
};
void serialize(FILE *f, const x1_column *ptr);
void unserialize(FILE *f, x1_column *ptr);

class x1_network
{
private:
    int                    nclasses;
    net_v1layer            v1;
    std::vector<x1_column> columns;
    predictions _apply(const raw_dataset &ds_raw, storage_format fmt, predict_t ptype) const;
public:
    friend void serialize(FILE *f, const x1_network *ptr);
    friend void unserialize(FILE *f, x1_network *ptr);

    x1_network():nclasses(0){}

    // info
    float           get_l1_factor() const;
    float           get_sparse_flops() const;
    int             get_nclasses()  const;
    x1_column&      get_column(int col_idx);
    conv_subcolumn& get_subcolumn1(int col_idx);
    net_v1layer&    get_v1();

    // predict
    predictions       apply(const raw_dataset &ds_raw, storage_format fmt = storage_format::i8) const;
    predictions   apply_log(const raw_dataset &ds_raw, storage_format fmt = storage_format::i8) const;
    predictions apply_logit(const raw_dataset &ds_raw, storage_format fmt = storage_format::i8) const;

    // external initialization
    void start_initialization(int classes, const net_v1layer &input_layer);
    void append_column(const x1_column &c);
};
void serialize(FILE *f, const x1_network *ptr);
void unserialize(FILE *f, x1_network *ptr);

class x1_binary
{
private:
    bool            initialized;
    net_v1layer     v1;
    x1_column       column0;
    predictions _apply(const raw_dataset &ds_raw, storage_format fmt, predict_t ptype) const;
public:
    friend void serialize(FILE *f, const x1_binary *ptr);
    friend void unserialize(FILE *f, x1_binary *ptr);

    x1_binary():initialized(false){}

    // create dummy network which always return 0.5 as output (no processing is performed on input,
    // although network expects it to be window_dim0*window_dim1 image)
    static x1_binary init_dummy_05(int window_dim0, int window_dim1);

    // info
    float           get_l1_factor() const;
    float           get_sparse_flops() const;
    x1_column&      get_column();
    conv_subcolumn& get_subcolumn1();
    net_v1layer&    get_v1();
    int             get_v1_window_dim0() const;
    int             get_v1_window_dim1() const;

    // predict
    predictions       apply(const raw_dataset &ds_raw, storage_format fmt = storage_format::i8) const;
    predictions   apply_log(const raw_dataset &ds_raw, storage_format fmt = storage_format::i8) const;
    predictions apply_logit(const raw_dataset &ds_raw, storage_format fmt = storage_format::i8) const;
    float            apply0(const bitmap &img, storage_format fmt = storage_format::i8) const;

    // convolutional apply
    conv_predictions conv_apply(const bitmap &raw_img, padding_t pad = padding_t::none, storage_format fmt = storage_format::i8) const;

    // external initialization
    void initialize(const net_v1layer &input_layer, const x1_column &c);
};
void serialize(FILE *f, const x1_binary *ptr);
void unserialize(FILE *f, x1_binary *ptr);

class x1_binary_cascade
{
private:
    std::vector<std::tuple<int,x1_binary,float>> levels;
public:

    // info
    const x1_binary& last() const;
    int get_last_scale() const;

    // initialize
    void add_level(int scale_idx, const x1_binary &net, float lev);

    // predict
    predictions       apply(const raw_dataset &ds_raw, storage_format fmt = storage_format::i8) const;
    float            apply0(const bitmap &img, storage_format fmt = storage_format::i8) const;

    // convolutional apply
    conv_predictions              conv_apply(const bitmap &img, padding_t pad = padding_t::none, storage_format fmt = storage_format::i8) const;
    conv_predictions              conv_apply(const bitmap &img, int nscales, padding_t pad = padding_t::none, storage_format fmt = storage_format::i8) const;
    std::vector<conv_predictions> conv_apply_ml(const bitmap &img, padding_t pad = padding_t::none, storage_format fmt = storage_format::i8) const;
    std::vector<conv_predictions> conv_apply_ml(const bitmap &img, int nscales, padding_t pad = padding_t::none, storage_format fmt = storage_format::i8) const;
};

template<class subcolumn_t>
predictions predict_max1(const std::vector<subcolumn_t> &models, predictions baseline_col, std::vector<const activities*> v_pact)
{
    int64_t ntotal = baseline_col.ntrn+baseline_col.ntst;
    ASSERT(models.size()==v_pact.size());
    predictions out = baseline_col;
    for(int64_t i=0; i<models.size(); i++)
    {
        predictions p = models[i].predict(baseline_col, *(v_pact[i]));
        for(int64_t j=0; j<ntotal; j++)
            out.fpredictions[j] = fmaxf(out.fpredictions[j], p.fpredictions[j]);
    }
    for(int64_t j=0; j<ntotal; j++)
        out.fpredictions[j+out.stride] = 1.0f-out.fpredictions[j];
    return out;
}


class conv_multicolumn
{
friend void serialize(FILE *f, const conv_multicolumn *ptr);
friend void unserialize(FILE *f, conv_multicolumn *ptr);
private:
    int                             nclasses;
    net_v1layer                     v1;
    int64_t                         post_v1_split;
    std::vector<conv_midlayer>      v2_columns;
    int64_t                         post_v2_stage;
    std::vector<linear_predictor>   predictors;
    std::vector<calibrator>         calibrators;
public:
    conv_multicolumn(){ nclasses = 0; }
    explicit conv_multicolumn(std::string filename);

    /*// training functions
    std::vector<predictions> train(
            const raw_dataset &ds_raw,
            const net_v1layer &layer1,
            const std::vector<int> columns_to_train,
            int v2_width,
            int v4_width,
            settings s);
    predictions train(
            const raw_dataset &ds_raw,
            const net_v1layer &layer1,
            int net_width,
            settings s);

    // concatenation
    static conv_multicolumn concat(const std::vector<conv_multicolumn> &v);

    // debug reports
    void dbg_analyze_stats() const;*/

    // predict
    predictions apply(const raw_dataset &ds_raw, const activities &v1_act) const;
    predictions apply(const raw_dataset &ds_raw) const;

    // external initialization
    void init_v1(const net_v1layer &v1);
    void add_column(const conv_midlayer &v2_col, const linear_predictor &pred, const calibrator &cal);
};
void serialize(FILE *f, const conv_multicolumn *ptr);
void unserialize(FILE *f, conv_multicolumn *ptr);

class dataset_subsplit
{
public:
    int kfolds;
    std::vector<int> split;

    static dataset_subsplit train_subsplit(const raw_dataset &ds_raw,
                                           const activities &act_v1,
                                           const int col,
                                           const int kpower,
                                           const bool rep = false);
};

template<class model_t>
class ensemble
{
private:
    int                     nclasses;
    bool                    shared_v1;
    std::vector<model_t>    models;
    std::vector<calibrator> cal;
public:
    ensemble() { nclasses = -1; }

    // initialization
    void init(int _nclasses, bool _shared_v1 = true)
    {
        ASSERT(_nclasses>0);

        nclasses  = _nclasses;
        shared_v1 = _shared_v1;
        models.clear();
        cal = std::vector<calibrator>(nclasses,calibrator());
    }

    // add model to the initialized ensemble
    void add_model(const model_t &m)
    {
        models.push_back(m);
    }

    // predict
    predictions apply(const raw_dataset &ds_raw) const
    {
        ASSERT(shared_v1);
        ASSERT(models.size()>0);

        activities act_v1;
        std::vector<predictions> vp;
        models[0].v1_stage(ds_raw, act_v1);
        for(const model_t &m: models)
            vp.push_back(m.v2v4_stage(ds_raw,act_v1));
        return predictions::calibrate(predictions::average(vp), cal);
    }

    void _serialize(FILE *f) const;
    void _unserialize(FILE *f);
};

template<class model_t>
void serialize(FILE *f, const ensemble<model_t> *ptr)
{
    ptr->_serialize(f);
}

template<class model_t>
void unserialize(FILE *f, ensemble<model_t> *ptr)
{
    ptr->_unserialize(f);
}

class maxout_postlayer
{
public:
    int64_t nchannels, strength;
    std::vector<float> coeffs;

    int64_t get_dim() const { return (nchannels+1)*strength; }
    maxout_postlayer(const activities &act_in):nchannels(act_in.nchannels),strength(0){}
};

maxout_postlayer extend_maxksigma2_model(const activities &act_v2c,
                                         const predictions &pred_target,
                                         maxout_postlayer layer,
                                         const int its_cnt,
                                         const float lambda_reg);
class gmaxout_postlayer
{
public:
    float leak_exp;
    float leak_max;

    int64_t nsubclassifiers, strength;
    std::vector<int> nchannels_cls;
    std::vector<float> coeffs;

    gmaxout_postlayer():nsubclassifiers(0),strength(0),leak_exp(0.01f),leak_max(0.00f){}
    gmaxout_postlayer(std::vector<const activities*> vv_act, int s);

    float*       get_coeff_ptr(int classifier_idx, int sub_idx);
    const float* get_coeff_ptr(int classifier_idx, int sub_idx) const;
    int          get_coeff_offset(int classifier_idx, int sub_idx) const;
    linear_predictor convert_to_linpred() const;
    linear_predictor convert_to_linpred(int classifier_idx, int sub_idx) const;

    predictions  predict(const predictions &baseline, const activities &act_in) const;
    predictions  predict(const predictions &baseline, std::vector<const activities*> vv_act) const;
    std::vector<int> get_winner_idx(const predictions &baseline, const activities &act_in) const;

    gmaxout_postlayer  apply_scaling(const std::vector<std::vector<float>> &vv_means, const std::vector<std::vector<float>> &vv_sigmas) const;
    gmaxout_postlayer remove_scaling(const std::vector<std::vector<float>> &vv_means, const std::vector<std::vector<float>> &vv_sigmas) const;

    void extend_with(const activities* p_act);
    void extend_with(const gmaxout_postlayer &model);
    void increase_strength(int snew);
    void import_predictor(int classifier_idx, int sub_idx, const linear_predictor &predictor);
};


void train_gmaxout(gmaxout_postlayer &model,
                   std::vector<const activities*> vv_act,
                   const predictions &pred_target,
                   const int epoch_cnt,
                   const int batch_size,
                   const float mu,
                   const float alpha,
                   const std::tuple<float,float> reg12,
                   const std::string rep);
void train_gmaxout(gmaxout_postlayer &model,
                   std::vector<const activities*> vv_act,
                   const predictions &pred_target,
                   std::vector<float> pred_w,
                   const int epoch_cnt,
                   const int batch_size,
                   const float mu,
                   const float alpha,
                   const std::tuple<float,float> reg12,
                   const std::string rep);

void train_grbf(std::vector<rbf_predictor> &models,
                std::vector<const activities*> vv_act_in,
                const predictions &pred_target,
                const int epoch_cnt,
                const int batch_size,
                const float mu,
                const float alpha,
                const std::tuple<float,float> reg12,
                const std::tuple<float,float> leak_me);

std::tuple<raw_dataset,std::vector<int>> multiscale_dataset_binary(
        const raw_dataset &ds_src,
        int src_width, int src_aoi,
        int dst_width, std::vector<int> dst_aoi);

float invsigma0_safe(float x);
predictions invsigma0(const predictions &p);

}

#endif
