/*************************************************************************
StrongNet/Contour Engine training framework.
Copyright (c) 2020 Sergey Bochkanov (ALGLIB project).

>>> SOURCE LICENSE >>>
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation (www.fsf.org); either version 2 of the 
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

A copy of the GNU General Public License is available at
http://www.fsf.org/licensing/licenses
>>> END OF LICENSE >>>
*************************************************************************/
#include "snet_trn.h"
#include <numeric>
#include <algorithm>
#include <random>
#include <mutex>
#include <set>
#include "3rdparty/alglib/optimization.h"
#include "3rdparty/alglib/dataanalysis.h"

namespace strongnet
{

static void _train_conv_subcolumn_v2v4(const raw_dataset &ds_wrk,
                                       const int col,
                                       const int split,
                                       activities &act_v1,
                                       const train_subcolumn_settings settings,
                                       const float clip_level,
                                       const std::tuple<int,int,int> nc_w24, // window size, v2 size, v4 size
                                       const std::tuple<int,split_t,int,int> co_geo, // window size, v1-split, v2 planes, v4 detectors
                                       const int seed,
                                       conv_subcolumn &out_subcol,
                                       activities &out_actmerged)
//const std::tuple<split_t,int,split_t,accumulate_t,int> conv_geometry, // subdivision, v2planes, post-split, post-accumulate, v4detectors
//const std::tuple<int,int> conv_batch_sizes,
//const float conv_page_prob,
{
    // dummy conv_geometry
    //const std::tuple<int,split_t,int,split_t,accumulate_t,int> conv_geometry = std::make_tuple(0,split_t::h16g64,0,split_t::entire,accumulate_t::sum,0);
    const std::tuple<int,int> conv_batch_sizes = std::make_tuple(5,5);
    const float conv_page_prob = 0.5f;

    //
    const int nc_window_size                = std::get<0>(nc_w24);
    const int v2nc_detectors                = std::get<1>(nc_w24);
    const int v4nc_detectors                = std::get<2>(nc_w24);
    const int co_window_size                = std::get<0>(co_geo);
    const split_t conv_subdivision          = std::get<1>(co_geo);
    const int v2c_planes                    = std::get<2>(co_geo);
    const split_t conv_post_split           = split_t::entire;
    const accumulate_t conv_post_accumulate = accumulate_t::sum;
    const int v4c_detectors                 = std::get<3>(co_geo);

    const int v2c_batch_size        = std::get<0>(conv_batch_sizes);
    const int v4c_batch_size        = std::get<1>(conv_batch_sizes);

    const bool has_nc   = v2nc_detectors+v4nc_detectors>0;
    const bool has_conv = v2c_planes+v4c_detectors>0;
    const bool has_both = has_nc && has_conv;

    //
    // Secondary parameters
    //
    const int search_rad = 2;
    const int conv_overlap = search_rad;
    v2_settings  s2;
    v4_settings  s4;
    v2c_settings s2c;
    v4_settings s4c;

    s2.seeds_count = v2nc_detectors;
    s2.rng_seed = seed;
    s2.batch_size = settings.v2v4_batch_size;
    s2.omega_rat = settings.omega_rat;
    s2.max_dist = 2;
    s2.gain_levels = settings.gain_levels;
    s2.gen_rating_growth = settings.gen_rating_growth;

    s4.batch_size = settings.v2v4_batch_size;
    s4.omega_rat  = settings.omega_rat;
    s4.gen_rating_growth = settings.gen_rating_growth;

    s2c.page_prob  = conv_page_prob;
    s2c.batch_size = v2c_batch_size;
    s2c.lambda_reg = settings.pre_lambda2_reg;
    s2c.omega_rat  = settings.omega_rat;
    s2c.rng_seed   = seed;
    s2c.dmax       = 3;
    s2c.gen_rating_growth = settings.gen_rating_growth;
    s4c.batch_size = v4c_batch_size;
    s4c.omega_rat  = settings.omega_rat;
    s4c.nneighbors = 100;
    s4c.dmax       = 3;
    s4c.gen_rating_growth = settings.gen_rating_growth;

    //
    // Save/process V1 layer
    //
    activities act_v1_nc, act_v1_conv;
    if( has_nc )
    {
        ASSERT(nc_window_size>0);
        act_v1.select_view(select_t::not_larger, nc_window_size, nc_window_size, act_v1_nc);
        if( has_both )
            act_v1_nc.add_prefix("[N]");
        ASSERT(act_v1_nc.realms.size()>0);
    }
    if( has_conv )
    {
        act_v1.select_view(select_t::not_smaller, co_window_size, co_window_size, act_v1_conv);
        act_v1_conv.subdivide(conv_subdivision, conv_overlap);
        if( has_both )
            act_v1_conv.add_prefix("[C]");
        ASSERT(act_v1_conv.realms.size()>0);
    }

    //
    // Some common data
    //
    const predictions baseline(ds_wrk);
    const predictions pred_col =  predictions::onevsall(baseline, col);
    linear_evaluator eval_all(pred_col, v2nc_detectors+v4nc_detectors+v2c_planes+v4c_detectors, sqrtf(s2c.lambda_reg*baseline.ntrn));
    out_actmerged = activities(baseline.ntrn, baseline.ntst, storage_format::i8);

    //
    // Train nonconvolutional part using i8 storage format, convert to fp32
    //
    net_midlayer  ml_v2nc;
    net_midlayer  ml_v4nc;
    if( has_nc)
    {
        activities    act_v2nc, act_v4nc;
        predictions   pred_v2nc, pred_v4nc;
        char prefix[256];
        printf("--- column %d, seed %d, training V2 ---\n", col, seed);
        fflush(stdout);
        sprintf(prefix, "[COL%02d#%d]", int(col), int(split));
        s2.rep_prefix = prefix;
        s4.rep_prefix = prefix;

        v2_layer_xx(eval_all,  act_v1_nc, s2, v2nc_detectors, ml_v2nc, act_v2nc, pred_v2nc);
        out_actmerged.merge_in(act_v2nc);
        if( v4nc_detectors>0 )
        {
            v4_layer(eval_all, act_v2nc,  s4, v4nc_detectors, ml_v4nc, act_v4nc, pred_v4nc);
            out_actmerged.merge_in(act_v4nc);
        }
    }
    out_actmerged.convert_to_f32();

    //
    // Train convolutional part
    //
    conv_midlayer ml_v2c;
    net_midlayer  ml_v4c;
    if( has_conv )
    {
        char prefix[256];
        activities    act_v2c, act_v4c;
        predictions   pred_v2c_all, pred_v4c;
        printf("--- column %d, seed %d, training V2c ---\n", col, seed);
        fflush(stdout);
        sprintf(prefix, "[COL%02dc#%d]", int(col), int(split));
        s2c.rep_prefix = prefix;
        v2c_layer_zz_splitacc(eval_all, act_v1_conv, s2c, search_rad, v2c_planes, conv_post_split, conv_post_accumulate, ml_v2c, act_v2c, pred_v2c_all);
        out_actmerged.merge_in(act_v2c);
        if( v4c_detectors>0 )
        {
            v4_layer(eval_all, act_v2c, s4c, v4c_detectors, ml_v4c, act_v4c, pred_v4c);
            out_actmerged.merge_in(act_v4c);
        }
    }

    //
    // Column is ready
    //
    out_subcol.init_partial(has_nc   ? select_t::not_larger  : select_t::none, std::make_tuple(nc_window_size,ml_v2nc,ml_v4nc),
                            has_conv ? select_t::not_smaller : select_t::none, std::make_tuple(co_window_size,conv_subdivision,conv_overlap,ml_v2c,conv_post_split,conv_post_accumulate,ml_v4c));

    //
    // Add clipping
    //
    out_subcol.set_clipping(clip_level);
    out_subcol.protect_activities(out_actmerged);
}

static std::tuple<conv_subcolumn,predictions> train_subcolumn_baseline(const raw_dataset &ds_wrk,
                            const int col,
                            activities &act_v1,
                            const train_subcolumn_settings settings,
                            const float clip_level,
                            const std::tuple<int,int,int> nc_w24, // window size, v2 size, v4 size
                            const std::tuple<int,split_t,int,int> co_geo, // window size, v1-split, v2 planes, v4 detectors
                            const int seed)
//const std::tuple<split_t,int,split_t,accumulate_t,int> conv_geometry, // subdivision, v2planes, post-split, post-accumulate, v4detectors
//const std::tuple<int,int> conv_batch_sizes,
//const float conv_page_prob,
{
    //
    // Train V2-V4 part
    //
    conv_subcolumn subcolumn;
    activities act_merged;
    _train_conv_subcolumn_v2v4(ds_wrk, col, 0, act_v1, settings, clip_level, nc_w24, co_geo, seed, subcolumn, act_merged);

    //
    // Train gmaxout model
    //
    const predictions baseline(ds_wrk);
    const predictions pred_col =  predictions::onevsall(baseline, col);
    const int batch_size = std::get<0>(settings.logit_batch_momentum);
    const float mu = std::get<1>(settings.logit_batch_momentum);
    int epoch_cnt;
    std::vector<float> rates;
    gmaxout_postlayer gmaxout_model({&act_merged},1);
    std::tie(epoch_cnt,rates) = settings.logit_learning_rates;
    for(float rate: rates)
        if( rate!=0 )
            train_gmaxout(gmaxout_model, {&act_merged}, pred_col, epoch_cnt, batch_size, mu, rate, std::make_tuple(0.0f,settings.pre_lambda2_reg), "");

    //
    // Column is ready
    //
    subcolumn.rewrite_predictor(gmaxout_model.convert_to_linpred(), calibrator());
    predictions pred_u = subcolumn.apply(pred_col, act_v1);
    printf("Subcolumn L1 factor = %.2f\n", (double)subcolumn.get_l1_factor());
    return std::make_tuple(subcolumn,pred_u);
}


static std::tuple<conv_subcolumn,predictions> train_subcolumn_multiflavor(const raw_dataset &ds_wrk,
                            const int col,
                            activities &act_v1,
                            const train_subcolumn_settings settings,
                            const float clip_level,
                            const std::tuple<int,int,int> nc_w24, // window size, v2 size, v4 size
                            const std::tuple<int,split_t,int,int> co_geo, // window size, v1-split, v2 planes, v4 detectors
                            const int seed)
{

    //
    // Analyze flavors
    //
    std::vector<std::string> flavors;
    std::set<std::string> sf;
    bool all_flavors_set=true, all_flavors_empty=true;
    for(const raw_dataset::entry &e: ds_wrk.entries)
        if( e.class_id==col )
        {
            all_flavors_set   = all_flavors_set   && (e.flavor!="");
            all_flavors_empty = all_flavors_empty && (e.flavor=="");
        }
    ASSERT(all_flavors_set || all_flavors_empty);
    if( all_flavors_empty )
        return train_subcolumn_baseline(ds_wrk, col, act_v1, settings, clip_level, nc_w24, co_geo, seed);
    for(const raw_dataset::entry &e: ds_wrk.entries)
        if( e.class_id==col )
            sf.insert(e.flavor);
    for(const std::string &s: sf)
        flavors.push_back(s);
    printf("[train_subcolumn_multiflavor] %d flavors specified, training subcolumns\n", (int)sf.size());
    fflush(stdout);

    //
    // Train V2-V4 part (no flavor info is used at this stage)
    //
    conv_subcolumn subcolumn;
    activities act_merged;
    _train_conv_subcolumn_v2v4(ds_wrk, col, 0, act_v1, settings, clip_level, nc_w24, co_geo, seed, subcolumn, act_merged);

    //
    // Train flavor-specific parts of the output classifier
    //
    const predictions baseline(ds_wrk);
    const predictions pred_col =  predictions::onevsall(baseline, col);
    const int batch_size = std::get<0>(settings.logit_batch_momentum);
    const float mu = std::get<1>(settings.logit_batch_momentum);
    int epoch_cnt;
    std::vector<float> rates;
    std::tie(epoch_cnt,rates) = settings.logit_learning_rates;
    superlogit2_predictor super_predictor = superlogit2_predictor::create_empty(act_merged);
    for(const std::string flavor: flavors)
    {
        //
        // Select balanced sample
        //
        int col_total=0, col_flavor=0;
        for(int i=0; i<ds_wrk.ntrn; i++)
            if( ds_wrk.entries[i].class_id==col )
            {
                if( ds_wrk.entries[i].flavor==flavor )
                    col_flavor++;
                col_total++;
            }
        ASSERT(col_flavor>0);
        int nmul = roundf(float(col_total)/float(col_flavor));
        std::vector<int> items;
        for(int i=0; i<ds_wrk.ntrn; i++)
            if( ds_wrk.entries[i].class_id==col )
            {
                if( ds_wrk.entries[i].flavor==flavor )
                {
                    for(int k=0; k<nmul; k++)
                        items.push_back(i);
                }
            }
            else
                items.push_back(i);
        for(int i=ds_wrk.ntrn; i<ds_wrk.ntrn+ds_wrk.ntst; i++)
            items.push_back(i);
        predictions pred_flavor = pred_col.select_subsample(items);
        activities act_flavor;
        act_merged.select_subsample(items, act_flavor);

        //
        // Train gmaxout model
        //
        gmaxout_postlayer gmaxout_model({&act_flavor},1);
        for(float rate: rates)
            if( rate!=0 )
                train_gmaxout(gmaxout_model, {&act_flavor}, pred_flavor, epoch_cnt, batch_size, mu, rate, std::make_tuple(0.0f,settings.pre_lambda2_reg), flavor);
        super_predictor.append1(gmaxout_model.convert_to_linpred());
    }
    subcolumn.rewrite_predictor(super_predictor, calibrator());
    predictions pred_u = subcolumn.apply(pred_col, act_v1);
    printf("Subcolumn L1 factor = %.2f\n", (double)subcolumn.get_l1_factor());
    return std::make_tuple(subcolumn,pred_u);
}

static const double PROB_EPS = 0.0001; // probabilities below this level are considered all same
static const float BIG_LOGIT = log((1-PROB_EPS)/PROB_EPS);
static const int MAX_CLASS0_REPEAT = 50;

static void retrain_subcolumn_subspace_from_target(
        const raw_dataset &ds_wrk,
        const predictions &baseline_col,
        activities  &act_v1_raw,
        std::vector<float> target_logits_full,
        const std::tuple<int,int,float> xparams,
        const std::tuple<int,float> xstop_cond,
        conv_subcolumn &subcolumn,
        predictions &pred)
{
    const int   XWIDTH = std::get<0>(xparams);
    const int   NN_CNT = std::get<1>(xparams);
    const float ADVERSARIAL_CLIP_LEVEL = std::get<2>(xparams);
    const int   STOP_K    = std::get<0>(xstop_cond);
    const float STOP_PROB = std::get<1>(xstop_cond);
    const int64_t ntrn_full   = ds_wrk.ntrn;
    const int64_t ntotal_full = ds_wrk.ntrn+ds_wrk.ntst;

    //
    // RNG
    //
    std::mt19937  mt_rnd;
    std::seed_seq mt_seed{(unsigned int)1};
    mt_rnd.seed(mt_seed);

    //
    // Set clipping
    //
    subcolumn.set_clipping(ADVERSARIAL_CLIP_LEVEL);

    //
    // Compute V2-V4 activities and build a kd-tree
    //
    activities act_v2v4_full;
    subcolumn.process_v2v4(act_v1_raw, act_v2v4_full);
    const int64_t nchannels_full = act_v2v4_full.nchannels;
    alglib::real_2d_array    kdt_xy;
    alglib::integer_1d_array kdt_tags;
    alglib::kdtree kdtree;
    int class0_size = 0;
    int class1_size = 0;
    kdt_xy.setlength(ntrn_full, nchannels_full);
    kdt_tags.setlength(ntrn_full);
    for(int i=0; i<ntrn_full; i++)
        if( baseline_col.labels[i]==0 )
        {
            for(int j=0; j<nchannels_full; j++)
                kdt_xy[class0_size][j] = act_v2v4_full.get_normalized(j,i);
            kdt_tags[class0_size] = i;
            class0_size++;
        }
    else
            class1_size++;
    alglib::kdtreebuildtagged(kdt_xy, kdt_tags, class0_size, nchannels_full, 0, 2, kdtree);

    //
    // Prepare targets
    //
    std::vector<float> target_probs_full(ntrn_full,0.0f);
    target_logits_full.resize(ntrn_full);
    for(int i=0; i<ntrn_full; i++)
    {
        // relax constraint on logits
        if( baseline_col.labels[i]!=0 )
            target_logits_full[i] = fmaxf(target_logits_full[i], -BIG_LOGIT);
        else
            target_logits_full[i] = fminf(target_logits_full[i], +BIG_LOGIT);

        // compute target probability
        target_probs_full[i]  = sigma0(target_logits_full[i]);
    }
    std::vector<float> current_probs_full(ntotal_full,0.0f);


    //
    // Iteratively select worst point until error in output probabilities decreases below PROB_TOL
    //
    superlogit2_predictor predictor = superlogit2_predictor::create_empty(act_v2v4_full);
    predictions current_pred = baseline_col;
    for(int ncenters=0; ; ncenters++)
    {
        int t0 = dtickcount();

        //
        // Select target point and its neighbors
        //
        int target_idx    = -1;
        float target_err =  0;
        for(int i=0; i<ntrn_full; i++)
        {
            float err = target_probs_full[i]-current_probs_full[i];
            if( baseline_col.labels[i]==0 && err>target_err )
            {
                target_idx = i;
                target_err = err;
            }
        }
        if( target_idx<0 || target_err<=0 )
            break;
        float prev_central_prob = current_probs_full[target_idx];
        alglib::real_1d_array kdt_x;
        alglib::integer_1d_array kdt_idx;
        kdt_x.setlength(nchannels_full);
        for(int i=0; i<nchannels_full; i++)
            kdt_x[i] = act_v2v4_full.get_normalized(i,target_idx);
        int kneighbors = alglib::kdtreequeryknn(kdtree, kdt_x, NN_CNT>0 ? NN_CNT : class0_size);
        alglib::kdtreequeryresultstags(kdtree, kdt_idx);
        std::vector<int> targets;
        for(int i=0; i<kneighbors; i++)
            targets.push_back(kdt_idx[i]);
        ASSERT(targets.size()>=1);

        //
        // Determine subspace to work with
        //
        std::vector<int> subspace;
        //for(int i=0; i<nchannels_full; i++)
        //    subspace.push_back(i);
        //const int nchannels_subspace = subspace.size();
        if( XWIDTH>0 && XWIDTH<nchannels_full )
        {
            int time_select = dtickcount();

            // determine set of class #1 items
            std::vector<int> class1_items;
            for(int i=0; i<ntrn_full; i++)
                if( baseline_col.labels[i]!=0 )
                    class1_items.push_back(i);
            ASSERT(class1_items.size()>0);
            std::uniform_int_distribution<> rnd1(0,class1_items.size()-1);

            // determine number of class #0 repetitions and class #1 subsample size
            // (these numbers are used to retain proportion of class #0 items)
            int class0_repeat    = imax(class0_size/targets.size(),1);
            int class1_subsample = class1_items.size();
            if( class0_repeat>MAX_CLASS0_REPEAT )
            {
                float downscale = float(MAX_CLASS0_REPEAT)/float(class0_repeat);
                class0_repeat    = MAX_CLASS0_REPEAT;
                class1_subsample = imax(class1_subsample*downscale,1);
            }

            // generate set of items which are used to determine best features
            std::vector<int> xx_items;
            for(int k: targets)
                for(int i=0; i<class0_repeat; i++)
                    xx_items.push_back(k);
            for(int i=0; i<class1_subsample; i++)
                xx_items.push_back(class1_items[rnd1(mt_rnd)]);
            ASSERT(ds_wrk.ntst>=1);
            xx_items.push_back(ds_wrk.ntrn);
#if 0
            std::vector<int> xx_items;
            int xx_repeat_factor = imax(class0_size/targets.size(),1);
            for(int i=0; i<ntrn_full; i++)
                if( baseline_col.labels[i]!=0 )
                    xx_items.push_back(i);
            for(int k: targets)
                for(int i=0; i<xx_repeat_factor; i++)
                    xx_items.push_back(k);
            for(int i=ntrn_full; i<ntotal_full; i++)
                xx_items.push_back(i);
#endif

            // sort items, select subsample and determine best features
            std::sort(xx_items.begin(), xx_items.end());
            predictions xx_baseline = baseline_col.select_subsample(xx_items);
            activities xx_act_v2v4;
            act_v2v4_full.select_subsample(xx_items, xx_act_v2v4);
            subspace = select_best_features(xx_baseline, xx_act_v2v4, XWIDTH);

            // debug report
            time_select = dtickcount()-time_select;
            //printf("selected in %d ms\n", time_select);
            //fflush(stdout);
        }
        else
        {
            ASSERT(XWIDTH==0 || XWIDTH==nchannels_full);
            for(int i=0; i<nchannels_full; i++)
                subspace.push_back(i);
        }
        const int nchannels_subspace = subspace.size();

        //
        // Select subspace of activities
        //
        activities act_v2v4_subspace(act_v2v4_full.ntrn,act_v2v4_full.ntst,storage_format::f32);
        act_v2v4_subspace.append_realm(realm<channel>::create_grid("*",1,1,0));
        for(int cidx: subspace)
        {
            float *ptr = act_v2v4_full.get_f32(cidx);
            act_v2v4_subspace.append1(activityf_vector(ptr,ptr+ntotal_full),"");
        }

        //
        // Generate constraints
        //
        const int xdim = nchannels_subspace+1;
        alglib::integer_1d_array ct;
        alglib::real_2d_array linconstr;
        alglib::real_1d_array bndl, bndu, bterm;
        int nconstr = 0;
        bndl.setlength(xdim);
        bndu.setlength(xdim);
        bterm.setlength(xdim);
        for(int i=0; i<xdim; i++)
        {
            bndl[i] = alglib::fp_neginf;
            bndu[i] = alglib::fp_posinf;
            bterm[i] = 0;
        }
        linconstr.setlength(ntrn_full,xdim+1);
        ct.setlength(ntrn_full);
        for(int i=0; i<ntrn_full; i++)
            if( baseline_col.labels[i]!=0 )
            {
                for(int j=0; j<xdim-1; j++)
                    linconstr[nconstr][j] = act_v2v4_subspace.get_normalized(j,i);
                linconstr[nconstr][xdim-1] = 1.0;
                linconstr[nconstr][xdim] = target_logits_full[i];
                ct[nconstr] = -1;
                nconstr++;
            }
        for(int i: targets)
        {
            ASSERT(baseline_col.labels[i]==0);
            ASSERT(nconstr<linconstr.rows());
            for(int j=0; j<xdim-1; j++)
                linconstr[nconstr][j] = act_v2v4_subspace.get_normalized(j,i);
            linconstr[nconstr][xdim-1] = 1.0;
            linconstr[nconstr][xdim] = target_logits_full[i];
            ct[nconstr] = +1;
            nconstr++;
        }

        //
        // No L1 penalty
        //
        const int xwrk = xdim;
        const int nconstr_wrk = nconstr;
        // optimize
        alglib::real_1d_array xf;
        try
        {
            alglib::minqpstate  qpsolver;
            alglib::minqpreport qprep;
            alglib::real_2d_array qterm;
            qterm.setlength(xwrk, xwrk);
            for(int i=0; i<xwrk; i++)
                for(int j=0; j<xwrk; j++)
                    qterm[i][j] = i==j ? 1 : 0;
            alglib::minqpcreate(xwrk, qpsolver);
            alglib::minqpsetalgodenseaul(qpsolver, 1.0e-7, 2000.0, 10);
            alglib::minqpsetquadraticterm(qpsolver, qterm);
            alglib::minqpsetlinearterm(qpsolver, bterm);
            alglib::minqpsetbc(qpsolver, bndl, bndu);
            alglib::minqpsetlc(qpsolver, linconstr, ct, nconstr_wrk);
            alglib::minqpoptimize(qpsolver, alglib::parallel);
            alglib::minqpresults(qpsolver, xf, qprep);
            ASSERT(qprep.terminationtype>0);
        }
        catch(alglib::ap_error e)
        {
            printf("[EXCEPTION]: '%s'\n", e.msg.c_str());
            fflush(stdout);
            throw;
        }
#if 0
        //
        // Reformulate constraints in order to add L1 penalty
        //
        const int xpen = xdim-1;
        const int xwrk = xdim+xpen;
        const int nconstr_wrk = nconstr+2*xpen;
        {
            alglib::real_2d_array prev_linconstr = linconstr;
            alglib::real_1d_array prev_bndl = bndl, prev_bndu = bndu, prev_bterm = bterm;
            alglib::integer_1d_array prev_ct = ct;

            // extend bldl/bndu/bterm
            bndl.setlength(xwrk);
            bndu.setlength(xwrk);
            bterm.setlength(xwrk);
            for(int i=0; i<xdim; i++)
            {
                bndl[i]  = prev_bndl[i];
                bndu[i]  = prev_bndu[i];
                bterm[i] = prev_bterm[i];
            }
            for(int i=xdim; i<xwrk; i++)
            {
                bndl[i] = 0;
                bndu[i] = alglib::fp_posinf;
                bterm[i] = 100;
            }

            // extend constraints
            linconstr.setlength(nconstr_wrk, xwrk+1);
            ct.setlength(nconstr_wrk);
            for(int i=0; i<nconstr; i++)
            {
                for(int j=0; j<xdim; j++)
                    linconstr[i][j] = prev_linconstr[i][j];
                for(int j=xdim; j<xwrk; j++)
                    linconstr[i][j] = 0;
                linconstr[i][xwrk] = prev_linconstr[i][xdim];
                ct[i] = prev_ct[i];
            }
            for(int i=0; i<xpen; i++)
            {
                int base_row = nconstr+2*i;
                for(int j=0; j<xwrk+1; j++)
                {
                    linconstr[base_row+0][j] = 0;
                    linconstr[base_row+1][j] = 0;
                }
                linconstr[base_row+0][i]      = +1;
                linconstr[base_row+0][i+xdim] = -1;
                ct[base_row+0] = -1;
                linconstr[base_row+1][i]      = +1;
                linconstr[base_row+1][i+xdim] = +1;
                ct[base_row+1] = +1;
            }
        }

        // optimize
        alglib::real_1d_array xf;
        try
        {
            alglib::minqpstate  qpsolver;
            alglib::minqpreport qprep;
            alglib::real_2d_array qterm;
            qterm.setlength(xwrk, xwrk);
            for(int i=0; i<xwrk; i++)
                for(int j=0; j<xwrk; j++)
                    qterm[i][j] = i==j ? 1 : 0;
            alglib::minqpcreate(xwrk, qpsolver);
            alglib::minqpsetalgodenseaul(qpsolver, 1.0e-7, 2000.0, 10);
            alglib::minqpsetquadraticterm(qpsolver, qterm);
            alglib::minqpsetlinearterm(qpsolver, bterm);
            alglib::minqpsetbc(qpsolver, bndl, bndu);
            alglib::minqpsetlc(qpsolver, linconstr, ct, nconstr_wrk);
            alglib::minqpoptimize(qpsolver, alglib::parallel);
            alglib::minqpresults(qpsolver, xf, qprep);
            ASSERT(qprep.terminationtype>0);
        }
        catch(alglib::ap_error e)
        {
            printf("[EXCEPTION]: '%s'\n", e.msg.c_str());
            fflush(stdout);
            throw;
        }
#endif
#if 0
        //
        // Reformulate constraints in order to add L1 penalty using LP
        //
        const int xpen = xdim;
        const int xwrk = xdim+xpen;
        const int nconstr_wrk = nconstr+2*xpen;
        {
            alglib::real_2d_array prev_linconstr = linconstr;
            alglib::real_1d_array prev_bndl = bndl, prev_bndu = bndu, prev_bterm = bterm;
            alglib::integer_1d_array prev_ct = ct;

            // extend bldl/bndu/bterm
            bndl.setlength(xwrk);
            bndu.setlength(xwrk);
            bterm.setlength(xwrk);
            for(int i=0; i<xdim; i++)
            {
                bndl[i]  = alglib::fp_neginf;
                bndu[i]  = alglib::fp_posinf;
                bterm[i] = 0;
            }
            for(int i=xdim; i<xwrk; i++)
            {
                bndl[i] = 0;
                bndu[i] = alglib::fp_posinf;
                bterm[i] = 1.0;
            }

            // extend constraints
            linconstr.setlength(nconstr_wrk, xwrk+1);
            ct.setlength(nconstr_wrk);
            for(int i=0; i<nconstr; i++)
            {
                for(int j=0; j<xdim; j++)
                    linconstr[i][j] = prev_linconstr[i][j];
                for(int j=xdim; j<xwrk; j++)
                    linconstr[i][j] = 0;
                linconstr[i][xwrk] = prev_linconstr[i][xdim];
                ct[i] = prev_ct[i];
            }
            for(int i=0; i<xpen; i++)
            {
                int base_row = nconstr+2*i;
                for(int j=0; j<xwrk+1; j++)
                {
                    linconstr[base_row+0][j] = 0;
                    linconstr[base_row+1][j] = 0;
                }
                linconstr[base_row+0][i]      = +1;
                linconstr[base_row+0][i+xdim] = -1;
                ct[base_row+0] = -1;
                linconstr[base_row+1][i]      = +1;
                linconstr[base_row+1][i+xdim] = +1;
                ct[base_row+1] = +1;
            }
        }

        // optimize
        alglib::real_1d_array xf;
        try
        {
            alglib::minlpstate  lpsolver;
            alglib::minlpreport lprep;
            alglib::minlpcreate(xwrk, lpsolver);
            alglib::minlpsetcost(lpsolver, bterm);
            alglib::minlpsetbc(lpsolver, bndl, bndu);
            alglib::minlpsetlc(lpsolver, linconstr, ct, nconstr_wrk);
            alglib::minlpoptimize(lpsolver, alglib::parallel);
            alglib::minlpresults(lpsolver, xf, lprep);
            ASSERT(lprep.terminationtype>0);
        }
        catch(alglib::ap_error e)
        {
            printf("[EXCEPTION]: '%s'\n", e.msg.c_str());
            fflush(stdout);
            throw;
        }
#endif

        // append to predictor
        std::vector<float> coeffs_full(nchannels_full+1,0.0f);
        for(int i=0; i<nchannels_subspace; i++)
            coeffs_full[subspace[i]] = xf[i];
        coeffs_full[nchannels_full] = xf[nchannels_subspace];
        predictor.append1(coeffs_full);

        // update logits, report
        float feas_err = 0, prob_err = 0;
        float target_l2 = 0, target_l1 = 0;
        for(int i=0; i<nconstr; i++)
        {
            double v = 0;
            for(int j=0; j<xwrk; j++)
                v += linconstr[i][j]*xf[j];
            v -= linconstr[i][xwrk];
            ASSERT(ct[i]!=0);
            if( ct[i]>0 )
                feas_err = fmax(-v, feas_err);
            else
                feas_err = fmax(+v, feas_err);
            //if( feas_err>1.0e-2 )
            //{
            //    printf("LC=");
            //    for(int j=0; j<=xdim; j++)
            //        printf("%6.2f ", linconstr[i][j]);
            //    printf("\n");
            //    printf("CT=%d\n", (int)ct[i]);
            //    printf("V=%.3e\n", v);
            //    fflush(stdout);
            //    ASSERT(0);
            //}
        }
        for(int i=0; i<xdim-1; i++)
        {
            target_l1 += fabs(xf[i]);
            target_l2 += xf[i]*xf[i];
        }
        current_pred = baseline_col;
        for(int i=0; i<ntotal_full; i++)
        {
            float v = xf[xdim-1];
            for(int j=0; j<xdim-1; j++)
                v += act_v2v4_subspace.get_normalized(j,i)*xf[j];
            v = sigma0(v);
            current_probs_full[i] = fmaxf(v,current_probs_full[i]);
            current_pred.fpredictions[i] = current_probs_full[i];
            current_pred.fpredictions[i+current_pred.stride] = 1.0f-current_probs_full[i];
        }
        for(int i=0; i<ntrn_full; i++)
            if( baseline_col.labels[i]==0 )
                prob_err  = fmaxf(prob_err, target_probs_full[i]-current_probs_full[i]);
        float new_central_prob = current_probs_full[target_idx];
        printf("[%03d] prob(trg)=%.4f->%.4f  L1=%6.1f  L2=%6.1f  prob_err=%.4f  trn_rms=%.4f  tst_rms=%.4f  tst_cls=%.4f (%6d ms, feas_err=%.2e)\n",
               ncenters,
               prev_central_prob, new_central_prob,
               target_l1, target_l2,
               prob_err,
               current_pred.trn_rms(), current_pred.tst_rms(), current_pred.tst_cls(),
               dtickcount()-t0, feas_err);
        fflush(stdout);
        if( feas_err>0.05 )
        {
            printf("[error] unable to solve QP problem with constraints, maybe we do not have enough meaningful features to separate classes\n");
            fflush(stdout);
            ASSERT(0);
        }
        if( prob_err<STOP_PROB )
            break;
        if( STOP_K>0 && ncenters>=STOP_K-1 )
            break;
    }

    //
    // Update subcolumn and predictions
    //
    subcolumn.rewrite_predictor(predictor, calibrator());
    pred = current_pred;
}

std::tuple<x1_network,predictions> x1_train_baseline(
        const raw_dataset &ds_wrk,
        std::vector<int> columns,
        const net_v1layer &v1layer,
        const train_subcolumn_settings settings,
        const float clip_level,
        const std::tuple<int,int,int> nc_w24, // window size, v2 size, v4 size
        const std::tuple<int,split_t,int,int> co_geo, // window size, v1-split, v2 planes, v4 detectors
        const int seed)
{
    //
    // normalize params
    //
    if( columns.size()==0 )
    {
        for(int i=0; i<ds_wrk.nclasses; i++)
            columns.push_back(i);
    }

    //
    // Save/process V1 layer
    //
    activities act_v1_raw;
    printf("--- seed %d, processing V1 ---\n", seed);
    fflush(stdout);
    v1layer.apply(ds_wrk, act_v1_raw, true);
    //serialize_to(&v1layer, std::string(OUT_DIR)+"/v1-layer.bin");

    //
    // Process columns
    //
    std::vector<predictions> v_pred;
    std::vector<x1_column>   v_x1;
    for(int ii=0; ii<columns.size(); ii++)
    {
        const int col = columns[ii];
        conv_subcolumn subcolumn;
        predictions pred;
        std::tie(subcolumn,pred) = train_subcolumn_baseline(ds_wrk,
                                                            col,
                                                            act_v1_raw,
                                                            settings,
                                                            clip_level,
                                                            nc_w24,
                                                            co_geo,
                                                            seed);

        x1_column x1col;
        x1col.append_subcolumn(subcolumn);

        v_x1.push_back(x1col);
        v_pred.push_back(pred);

        //serialize_to(&subcolumn, std::string(OUT_DIR)+"/col"  +std::to_string(col)+"-seed"+std::to_string(seed)+".bin");
        //serialize_to(&x1col,     std::string(OUT_DIR)+"/x1col"+std::to_string(col)+"-seed"+std::to_string(seed)+".x1c");
        //serialize_to(&pred,      std::string(OUT_DIR)+"/col"  +std::to_string(col)+"-seed"+std::to_string(seed)+".pred");

        printf("Combined L1 factor = %.2f\n", (double)x1col.get_l1_factor());
        //predictions tmp = x1col.apply(baseline_col, act_v1_raw);
        //printf("cls=%.4f\n", (double)tmp.tst_cls());
        //fflush(stdout);
    }

    //
    // Combine (if we have exactly as many columns as classes)
    //
    if( v_pred.size()==ds_wrk.nclasses )
    {
        //
        // Merge predictions
        //
        const predictions pred_merged = predictions::merge_onevsall(v_pred);
        //serialize_to(&pred_merged, std::string(OUT_DIR)+"/allmerged-seed"+std::to_string(seed)+".pred");
        std::vector<int> cls_err;
        float trn_rms, tst_rms, tst_cls;
        printf("=== printing out combined predictions ===\n");
        pred_merged.compute_errors(&trn_rms, &tst_rms, &tst_cls, &cls_err);
        printf("[MERGED] TrnRms=%.4f    TstRms=%.4f    TstCls=%.4f\n", (double)trn_rms, (double)tst_rms, (double)tst_cls);
        for(int k=0; k<cls_err.size(); k++)
            printf("err-cnt[%02d] = %4d\n", k, cls_err[k]);
        fflush(stdout);

        //
        // Create X1 network
        //
        x1_network x1net;
        x1net.start_initialization(ds_wrk.nclasses, v1layer);
        for(const x1_column &x1col: v_x1)
            x1net.append_column(x1col);
        //serialize_to(&x1net, std::string(OUT_DIR)+"/x1net-seed"+std::to_string(seed)+".x1");
        //predictions tmp = x1net.apply(ds_wrk);
        //printf("cls=%.4f\n", (double)tmp.tst_cls());
        //fflush(stdout);
        return std::make_tuple(x1net, pred_merged);
    }
    else
        return std::make_tuple(x1_network(),predictions());
}

std::tuple<x1_binary,predictions> x1_train_multiflavor_binary(
        const raw_dataset &ds_wrk,
        const net_v1layer &v1layer,
        const train_subcolumn_settings settings,
        const float clip_level,
        const std::tuple<int,int,int> nc_w24, // window size, v2 size, v4 size
        const std::tuple<int,split_t,int,int> co_geo, // window size, v1-split, v2 planes, v4 detectors
        const int seed)
{
    ASSERT(ds_wrk.nclasses==2);


    //
    // Save/process V1 layer
    //
    activities act_v1_raw;
    printf("--- seed %d, processing V1 ---\n", seed);
    fflush(stdout);
    v1layer.apply(ds_wrk, act_v1_raw, true);
    //serialize_to(&v1layer, std::string(OUT_DIR)+"/v1-layer.bin");

    //
    // Train different subcolumns
    //
    x1_column x0;
    predictions pred_merged = predictions(ds_wrk);
    conv_subcolumn subcolumn0;
    predictions pred0;
    std::tie(subcolumn0,pred0) = train_subcolumn_multiflavor(ds_wrk,
                                                             0,
                                                             act_v1_raw,
                                                             settings,
                                                             clip_level,
                                                             nc_w24,
                                                             co_geo,
                                                             seed);
    x0.append_subcolumn(subcolumn0);
    for(int i=0; i<pred0.ntrn+pred0.ntst; i++)
    {
        pred_merged.fpredictions[i] = pred0.fpredictions[i];
        pred_merged.fpredictions[i+pred_merged.stride] = 1-pred_merged.fpredictions[i];
    }
    printf("L1 factor = %.2f\n", (double)x0.get_l1_factor());
    //serialize_to(&pred_merged, std::string(OUT_DIR)+"/allmerged-seed"+std::to_string(seed)+".pred");

    //
    // Print combined predictions
    //
    std::vector<int> cls_err;
    float trn_rms, tst_rms, tst_cls;
    printf("=== printing out combined predictions ===\n");
    pred_merged.compute_errors(&trn_rms, &tst_rms, &tst_cls, &cls_err);
    printf("[MERGED] TrnRms=%.4f    TstRms=%.4f    TstCls=%.4f\n", (double)trn_rms, (double)tst_rms, (double)tst_cls);
    for(int k=0; k<cls_err.size(); k++)
        printf("err-cnt[%02d] = %4d\n", k, cls_err[k]);
    fflush(stdout);

    //
    // Create X1 network
    //
    x1_binary x1net;
    x1net.initialize(v1layer, x0);
    //serialize_to(&x1net, std::string(OUT_DIR)+"/x1bin-seed"+std::to_string(seed)+".x1b");
    //predictions tmp = x1net.apply(ds_wrk);
    //printf("cls=%.4f\n", (double)tmp.tst_cls());
    //fflush(stdout);
    return std::make_tuple(x1net, pred_merged);
}

std::tuple<x1_network,predictions> x1_retrain_baseline_notransfer(
        x1_network ref_net,
        const raw_dataset &ds_wrk,
        std::vector<int> columns,
        const train_subcolumn_settings settings,
        const float clip_level)
{
    ASSERT(ref_net.get_nclasses()==ds_wrk.nclasses);

    //
    // normalize params
    //
    if( columns.size()==0 )
    {
        for(int i=0; i<ds_wrk.nclasses; i++)
            columns.push_back(i);
    }

    //
    // Save/process V1 layer
    //
    activities act_v1_raw;
    printf("--- processing V1 ---\n");
    fflush(stdout);
    ref_net.get_v1().apply(ds_wrk, act_v1_raw, true);

    //
    // Process columns
    //
    std::vector<predictions> vv_pred;
    std::vector<x1_column>   vv_x1;
    predictions baseline(ds_wrk);
    for(const int col: columns)
    {
        //
        // Obtain reference subcolumn
        //
        const predictions baseline_col = predictions::onevsall(baseline,col);
        conv_subcolumn subcolumn = ref_net.get_subcolumn1(col);

        //
        // Set clip level
        //
        subcolumn.set_clipping(clip_level);

        //
        // Obtain V2-V4 outputs
        //
        activities act_v2v4;
        int t0_v2v4 = dtickcount();
        printf("[COL%02d:V2V4] computing V2-V4...", col);
        fflush(stdout);
        subcolumn.process_v2v4(act_v1_raw, act_v2v4);
        printf(" done in %d ms\n", int(dtickcount()-t0_v2v4));
        fflush(stdout);

        //
        // Train gmaxout model
        //
        const int batch_size = std::get<0>(settings.logit_batch_momentum);
        const float mu = std::get<1>(settings.logit_batch_momentum);
        int epoch_cnt;
        std::vector<float> rates;
        gmaxout_postlayer gmaxout_model({&act_v2v4},1);
        std::tie(epoch_cnt,rates) = settings.logit_learning_rates;
        for(float rate: rates)
            if( rate!=0 )
                train_gmaxout(gmaxout_model, {&act_v2v4}, baseline_col, epoch_cnt, batch_size, mu, rate, std::make_tuple(0.0f,settings.pre_lambda2_reg), "");

        //
        // Column is ready
        //
        linear_predictor linpred = gmaxout_model.convert_to_linpred();
        predictions pred = linpred.predict(baseline_col,act_v2v4);
        subcolumn.rewrite_predictor(linpred, calibrator());

        //
        // Save results
        //
        x1_column x1col;
        x1col.append_subcolumn(subcolumn);
        vv_x1.push_back(x1col);
        vv_pred.push_back(pred);
        //serialize_to(&x1col,     std::string(OUT_DIR)+"/x1col"+std::to_string(col)+"-seed0.x1c");
        //serialize_to(&pred,      std::string(OUT_DIR)+"/col"  +std::to_string(col)+"-seed0.pred");

        printf("Final combined L1 factor = %.2f\n", (double)x1col.get_l1_factor());

        //predictions tmp = x1col.apply(baseline_col, act_v1_raw);
        //printf("cls=%.4f\n", (double)tmp.tst_cls());
        //fflush(stdout);
    }

    //
    // Combine (if we have exactly as many columns as classes)
    //
    if( vv_pred.size()==ds_wrk.nclasses )
    {
        //
        // Merge predictions
        //
        const predictions pred_merged = predictions::merge_onevsall(vv_pred);
        //serialize_to(&pred_merged, std::string(OUT_DIR)+"/allmerged-seed0.pred");
        std::vector<int> cls_err;
        float trn_rms, tst_rms, tst_cls;
        printf("=== printing out combined predictions ===\n");
        pred_merged.compute_errors(&trn_rms, &tst_rms, &tst_cls, &cls_err);
        printf("[MERGED] TrnRms=%.4f    TstRms=%.4f    TstCls=%.4f\n", (double)trn_rms, (double)tst_rms, (double)tst_cls);
        for(int k=0; k<cls_err.size(); k++)
            printf("err-cnt[%02d] = %4d\n", k, cls_err[k]);
        fflush(stdout);

        //
        // Create X1 network
        //
        x1_network x1net;
        x1net.start_initialization(ds_wrk.nclasses, ref_net.get_v1());
        for(const x1_column &x1col: vv_x1)
            x1net.append_column(x1col);
        //serialize_to(&x1net, std::string(OUT_DIR)+"/x1net-seed0.x1");
        printf("Combined L1 factor = %.2f\n", (double)x1net.get_l1_factor());
        fflush(stdout);
        //predictions tmp = x1net.apply(ds_wrk);
        //printf("cls=%.4f\n", (double)tmp.tst_cls());
        //fflush(stdout);
        return std::make_tuple(x1net,pred_merged);
    }
    else
        return std::make_tuple(x1_network(),predictions());
}

std::tuple<x1_network,predictions> x1_retrain_subspace(
        x1_network ref_net,
        const raw_dataset &ds_wrk,
        std::vector<int> columns,
        const std::tuple<int,int,float> xparams,
        const std::tuple<int,float> xstop_cond)
{
    ASSERT(ref_net.get_nclasses()==ds_wrk.nclasses);

    //
    // normalize params
    //
    if( columns.size()==0 )
    {
        for(int i=0; i<ds_wrk.nclasses; i++)
            columns.push_back(i);
    }

    //
    // Save/process V1 layer
    //
    activities act_v1_raw;
    printf("--- processing V1 ---\n");
    fflush(stdout);
    ref_net.get_v1().apply(ds_wrk, act_v1_raw, true);

    //
    // Process columns
    //
    std::vector<predictions> vv_pred;
    std::vector<x1_column>   vv_x1;
    predictions baseline(ds_wrk);
    for(const int col: columns)
    {
        //
        // Obtain reference subcolumn
        //
        const predictions baseline_col = predictions::onevsall(baseline,col);
        conv_subcolumn subcolumn = ref_net.get_subcolumn1(col);

        //
        // Obtain reference predictions
        //
        int t0_refpred = dtickcount();
        printf("[COL%02d:V2V4] computing reference predictions...", col);
        fflush(stdout);
        predictions col_logits = subcolumn.apply_logit(baseline_col, act_v1_raw);
        std::vector<float> target_logits_full = col_logits.kth_as_fvector(0);
        printf(" done in %d ms\n", int(dtickcount()-t0_refpred));
        fflush(stdout);

        //
        // Retrain subcolumn using reference predictions as targets
        //
        predictions pred;
        retrain_subcolumn_subspace_from_target(
                    ds_wrk,
                    baseline_col,
                    act_v1_raw,
                    target_logits_full,
                    xparams,
                    xstop_cond,
                    subcolumn,
                    pred);

        //
        // Save results
        //
        x1_column x1col;
        x1col.append_subcolumn(subcolumn);
        vv_x1.push_back(x1col);
        vv_pred.push_back(pred);
        //serialize_to(&x1col,     std::string(OUT_DIR)+"/x1col"+std::to_string(col)+"-seed0.x1c");
        //serialize_to(&pred,      std::string(OUT_DIR)+"/col"  +std::to_string(col)+"-seed0.pred");

        printf("Final combined L1 factor = %.2f\n", (double)x1col.get_l1_factor());

        //predictions tmp = x1col.apply(baseline_col, act_v1_raw);
        //printf("cls=%.4f\n", (double)tmp.tst_cls());
        //fflush(stdout);
    }

    //
    // Combine (if we have exactly as many columns as classes)
    //
    if( vv_pred.size()==ds_wrk.nclasses )
    {
        //
        // Merge predictions
        //
        const predictions pred_merged = predictions::merge_onevsall(vv_pred);
        //serialize_to(&pred_merged, std::string(OUT_DIR)+"/allmerged-seed0.pred");
        std::vector<int> cls_err;
        float trn_rms, tst_rms, tst_cls;
        printf("=== printing out combined predictions ===\n");
        pred_merged.compute_errors(&trn_rms, &tst_rms, &tst_cls, &cls_err);
        printf("[MERGED] TrnRms=%.4f    TstRms=%.4f    TstCls=%.4f\n", (double)trn_rms, (double)tst_rms, (double)tst_cls);
        for(int k=0; k<cls_err.size(); k++)
            printf("err-cnt[%02d] = %4d\n", k, cls_err[k]);
        fflush(stdout);

        //
        // Create X1 network
        //
        x1_network x1net;
        x1net.start_initialization(ds_wrk.nclasses, ref_net.get_v1());
        for(const x1_column &x1col: vv_x1)
            x1net.append_column(x1col);
        //serialize_to(&x1net, std::string(OUT_DIR)+"/x1net-seed0.x1");
        printf("Combined L1 factor = %.2f\n", (double)x1net.get_l1_factor());
        fflush(stdout);
        //predictions tmp = x1net.apply(ds_wrk);
        //printf("cls=%.4f\n", (double)tmp.tst_cls());
        //fflush(stdout);
        return std::make_tuple(x1net,pred_merged);
    }
    else
        return std::make_tuple(x1_network(),predictions());
}


x1_network x1_train_twostep(const raw_dataset &ds_wrk,
                            std::vector<int> columns,
                            const net_v1layer &v1layer,
                            const train_subcolumn_settings settings,
                            const std::tuple<int,int,int> nc_w24, // windowsize, v2 size, v4 size
                            const std::tuple<int,int,float> xparams,   // subclassifier width, NN_CNT, ADV_CLIP
                            const std::tuple<int,float> xstop_cond,
                            const int seed)
{
    //
    // normalize params
    //
    if( columns.size()==0 )
    {
        for(int i=0; i<ds_wrk.nclasses; i++)
            columns.push_back(i);
    }

    //
    //
    //
    const int64_t ntrn_full   = ds_wrk.ntrn;
    const int64_t ntotal_full = ds_wrk.ntrn+ds_wrk.ntst;
    const int   XWIDTH = std::get<0>(xparams);
    const int   NN_CNT = std::get<1>(xparams);
    const float ADVERSARIAL_CLIP_LEVEL = std::get<2>(xparams);
    const int   STOP_K    = std::get<0>(xstop_cond);
    const float STOP_PROB = std::get<1>(xstop_cond);
    ASSERT(XWIDTH>=0);
    ASSERT(NN_CNT>=1);

    //
    // Save/process V1 layer
    //
    activities act_v1_raw;
    printf("--- seed %d, processing V1 ---\n", seed);
    fflush(stdout);
    v1layer.apply(ds_wrk, act_v1_raw, true);
    //serialize_to(&v1layer, std::string(OUT_DIR)+"/v1-layer.bin");

    //
    // Process columns
    //
    std::vector<predictions> v_pred;
    std::vector<x1_column>   v_x1;
    predictions baseline(ds_wrk);
    for(int ii=0; ii<columns.size(); ii++)
    {
        //
        // Train baseline column
        //
        const int col = columns[ii];
        const predictions baseline_col = predictions::onevsall(baseline,col);
        conv_subcolumn subcolumn;
        predictions pred;
        std::tie(subcolumn,pred) = train_subcolumn_baseline(ds_wrk,
                                                            col,
                                                            act_v1_raw,
                                                            settings,
                                                            ADVERSARIAL_CLIP_LEVEL,
                                                            nc_w24,
                                                            std::make_tuple(0,split_t::entire,0,0),
                                                            seed);

        //
        // Compute V2-V4 activities and build a kd-tree
        //
        activities  act_v2v4_full;
        subcolumn.process_v2v4(act_v1_raw, act_v2v4_full);
        const int64_t nchannels_full = act_v2v4_full.nchannels;
        alglib::real_2d_array    kdt_xy;
        alglib::integer_1d_array kdt_tags;
        alglib::kdtree kdtree;
        int class0_size = 0;
        kdt_xy.setlength(ntrn_full, nchannels_full);
        kdt_tags.setlength(ntrn_full);
        for(int i=0; i<ntrn_full; i++)
            if( baseline_col.labels[i]==0 )
            {
                for(int j=0; j<nchannels_full; j++)
                    kdt_xy[class0_size][j] = act_v2v4_full.get_normalized(j,i);
                kdt_tags[class0_size] = i;
                class0_size++;
            }
        alglib::kdtreebuildtagged(kdt_xy, kdt_tags, class0_size, nchannels_full, 0, 2, kdtree);

        //
        // Prepare data for subclassifier training:
        // * V2V4 activities
        // * target output logits and probabilities (after applying BIG_LOGIT cap)
        // * current output probabilities (zero)
        //
        std::vector<float> target_logits_full = subcolumn.apply_logit(baseline_col, act_v1_raw).kth_as_fvector(0);
        std::vector<float> target_probs_full  = pred.kth_as_fvector(0);
        std::vector<float> current_probs_full(ntotal_full,0.0f);
        target_logits_full.resize(ntrn_full);
        target_probs_full.resize(ntrn_full);
        for(int i=0; i<ntrn_full; i++)
            if( baseline_col.labels[i]!=0 )
            {
                target_logits_full[i] = fmaxf(target_logits_full[i], -BIG_LOGIT); // relax constraint
                target_probs_full[i]  = fmaxf(target_probs_full[i],  PROB_EPS);   // relax constraint
            }
            else
            {
                target_logits_full[i] = fminf(target_logits_full[i], +BIG_LOGIT); // relax constraint
                target_probs_full[i]  = fminf(target_probs_full[i],  1-PROB_EPS); // relax constraint
            }

        //
        // Iteratively select worst point until error in output probabilities decreases below PROB_TOL
        //
        superlogit2_predictor predictor = superlogit2_predictor::create_empty(act_v2v4_full);
        predictions current_pred = baseline_col;
        for(int ncenters=0; ; ncenters++)
        {
            int t0 = dtickcount();

            //
            // Select target point and its neighbors
            //
            int target_idx    = -1;
            float target_err =  0;
            for(int i=0; i<ntrn_full; i++)
            {
                float err = target_probs_full[i]-current_probs_full[i];
                if( baseline_col.labels[i]==0 && err>target_err )
                {
                    target_idx = i;
                    target_err = err;
                }
            }
            if( target_idx<0 || target_err<=0 )
                break;
            float prev_central_prob = current_probs_full[target_idx];
            alglib::real_1d_array kdt_x;
            alglib::integer_1d_array kdt_idx;
            kdt_x.setlength(nchannels_full);
            for(int i=0; i<nchannels_full; i++)
                kdt_x[i] = act_v2v4_full.get_normalized(i,target_idx);
            int kneighbors = alglib::kdtreequeryknn(kdtree, kdt_x, NN_CNT);
            alglib::kdtreequeryresultstags(kdtree, kdt_idx);
            std::vector<int> targets;
            for(int i=0; i<kneighbors; i++)
                targets.push_back(kdt_idx[i]);
            ASSERT(targets.size()>=1);

            //
            // Determine subspace to work with
            //
            std::vector<int> subspace;
            //for(int i=0; i<nchannels_full; i++)
            //    subspace.push_back(i);
            //const int nchannels_subspace = subspace.size();
            if( XWIDTH>0 && XWIDTH<nchannels_full )
            {
                std::vector<int> xx_items;
                int xx_repeat_factor = imax(class0_size/targets.size(),1);
                for(int i=0; i<ntrn_full; i++)
                    if( baseline_col.labels[i]!=0 )
                        xx_items.push_back(i);
                for(int k: targets)
                    for(int i=0; i<xx_repeat_factor; i++)
                        xx_items.push_back(k);
                for(int i=ntrn_full; i<ntotal_full; i++)
                    xx_items.push_back(i);
                std::sort(xx_items.begin(), xx_items.end());
                predictions xx_baseline = baseline_col.select_subsample(xx_items);
                activities xx_act_v2v4;
                act_v2v4_full.select_subsample(xx_items, xx_act_v2v4);
                subspace = select_best_features(xx_baseline, xx_act_v2v4, XWIDTH);
            }
            else
            {
                ASSERT(XWIDTH==0 || XWIDTH==nchannels_full);
                for(int i=0; i<nchannels_full; i++)
                    subspace.push_back(i);
            }
            const int nchannels_subspace = subspace.size();

            //
            // Select subspace of activities
            //
            activities act_v2v4_subspace(act_v2v4_full.ntrn,act_v2v4_full.ntst,storage_format::f32);
            act_v2v4_subspace.append_realm(realm<channel>::create_grid("*",1,1,0));
            for(int cidx: subspace)
            {
                float *ptr = act_v2v4_full.get_f32(cidx);
                act_v2v4_subspace.append1(activityf_vector(ptr,ptr+ntotal_full),"");
            }

            // generate constraints
            const int xdim = nchannels_subspace+1;
            alglib::integer_1d_array ct;
            alglib::real_2d_array linconstr;
            int nconstr = 0;
            linconstr.setlength(ntrn_full,xdim+1);
            ct.setlength(ntrn_full);
            for(int i=0; i<ntrn_full; i++)
                if( baseline_col.labels[i]!=0 )
                {
                    for(int j=0; j<xdim-1; j++)
                        linconstr[nconstr][j] = act_v2v4_subspace.get_normalized(j,i);
                    linconstr[nconstr][xdim-1] = 1.0;
                    linconstr[nconstr][xdim] = target_logits_full[i];
                    ct[nconstr] = -1;
                    nconstr++;
                }
            for(int i: targets)
            {
                ASSERT(baseline_col.labels[i]==0);
                ASSERT(nconstr<linconstr.rows());
                for(int j=0; j<xdim-1; j++)
                    linconstr[nconstr][j] = act_v2v4_subspace.get_normalized(j,i);
                linconstr[nconstr][xdim-1] = 1.0;
                linconstr[nconstr][xdim] = target_logits_full[i];
                ct[nconstr] = +1;
                nconstr++;
            }

            // optimize
            alglib::real_1d_array xf;
            try
            {
                alglib::minqpstate  qpsolver;
                alglib::minqpreport qprep;
                alglib::real_2d_array qterm;
                qterm.setlength(xdim, xdim);
                for(int i=0; i<xdim; i++)
                    for(int j=0; j<xdim; j++)
                        qterm[i][j] = i==j ? 1 : 0;
                alglib::minqpcreate(xdim, qpsolver);
                alglib::minqpsetalgodenseaul(qpsolver, 1.0e-6, 1000.0, 10);
                alglib::minqpsetquadraticterm(qpsolver, qterm);
                alglib::minqpsetlc(qpsolver, linconstr, ct, nconstr);
                alglib::minqpoptimize(qpsolver, alglib::parallel);
                alglib::minqpresults(qpsolver, xf, qprep);
                ASSERT(qprep.terminationtype>0);
            }
            catch(alglib::ap_error e)
            {
                printf("[EXCEPTION]: '%s'\n", e.msg.c_str());
                fflush(stdout);
                throw;
            }

            // append to predictor
            std::vector<float> coeffs_full(nchannels_full+1,0.0f);
            for(int i=0; i<nchannels_subspace; i++)
                coeffs_full[subspace[i]] = xf[i];
            coeffs_full[nchannels_full] = xf[nchannels_subspace];
            predictor.append1(coeffs_full);

            // update logits, report
            float feas_err = 0, prob_err = 0;
            float target_l2 = 0, target_l1 = 0;
            for(int i=0; i<nconstr; i++)
            {
                double v = 0;
                for(int j=0; j<xdim; j++)
                    v += linconstr[i][j]*xf[j];
                v -= linconstr[i][xdim];
                ASSERT(ct[i]!=0);
                if( ct[i]>0 )
                    feas_err = fmax(-v, feas_err);
                else
                    feas_err = fmax(+v, feas_err);
                //if( feas_err>1.0e-2 )
                //{
                //    printf("LC=");
                //    for(int j=0; j<=xdim; j++)
                //        printf("%6.2f ", linconstr[i][j]);
                //    printf("\n");
                //    printf("CT=%d\n", (int)ct[i]);
                //    printf("V=%.3e\n", v);
                //    fflush(stdout);
                //    ASSERT(0);
                //}
            }
            for(int i=0; i<xdim-1; i++)
            {
                target_l1 += fabs(xf[i]);
                target_l2 += xf[i]*xf[i];
            }
            current_pred = baseline_col;
            for(int i=0; i<ntotal_full; i++)
            {
                float v = xf[xdim-1];
                for(int j=0; j<xdim-1; j++)
                    v += act_v2v4_subspace.get_normalized(j,i)*xf[j];
                v = sigma0(v);
                current_probs_full[i] = fmaxf(v,current_probs_full[i]);
                current_pred.fpredictions[i] = current_probs_full[i];
                current_pred.fpredictions[i+current_pred.stride] = 1.0f-current_probs_full[i];
            }
            for(int i=0; i<ntrn_full; i++)
                if( baseline_col.labels[i]==0 )
                    prob_err  = fmaxf(prob_err, target_probs_full[i]-current_probs_full[i]);
            float new_central_prob = current_probs_full[target_idx];
            printf("[%03d] prob(trg)=%.4f->%.4f  L1=%6.1f  L2=%6.1f  prob_err=%.4f  trn_rms=%.4f  tst_rms=%.4f  tst_cls=%.4f (%6d ms, feas_err=%.2e)\n",
                   ncenters,
                   prev_central_prob, new_central_prob,
                   target_l1, target_l2,
                   prob_err,
                   current_pred.trn_rms(), current_pred.tst_rms(), current_pred.tst_cls(),
                   dtickcount()-t0, feas_err);
            fflush(stdout);
            if( feas_err>1.0e-2 )
            {
                printf("[error] unable to solve QP problem with constraints, maybe we do not have enough meaningful features to separate classes\n");
                fflush(stdout);
                ASSERT(0);
            }
            if( prob_err<STOP_PROB )
                break;
            if( STOP_K>0 && ncenters>=STOP_K-1 )
                break;
        }

        //
        // Update subcolumn and predictions
        //
        subcolumn.rewrite_predictor(predictor, calibrator());
        pred = current_pred;

        //
        // Save results
        //
        x1_column x1col;
        x1col.append_subcolumn(subcolumn);

        v_x1.push_back(x1col);
        v_pred.push_back(pred);

        //serialize_to(&subcolumn, std::string(OUT_DIR)+"/col"  +std::to_string(col)+"-seed"+std::to_string(seed)+".bin");
        //serialize_to(&x1col,     std::string(OUT_DIR)+"/x1col"+std::to_string(col)+"-seed"+std::to_string(seed)+".x1c");
        //serialize_to(&pred,      std::string(OUT_DIR)+"/col"  +std::to_string(col)+"-seed"+std::to_string(seed)+".pred");

        printf("Final combined L1 factor = %.2f\n", (double)x1col.get_l1_factor());

        //predictions tmp = x1col.apply(baseline_col, act_v1_raw);
        //printf("cls=%.4f\n", (double)tmp.tst_cls());
        //fflush(stdout);
    }

    //
    // Combine (if we have exactly as many columns as classes)
    //
    if( v_pred.size()==ds_wrk.nclasses )
    {
        //
        // Merge predictions
        //
        const predictions pred_merged = predictions::merge_onevsall(v_pred);
        //serialize_to(&pred_merged, std::string(OUT_DIR)+"/allmerged-seed"+std::to_string(seed)+".pred");
        std::vector<int> cls_err;
        float trn_rms, tst_rms, tst_cls;
        printf("=== printing out combined predictions ===\n");
        pred_merged.compute_errors(&trn_rms, &tst_rms, &tst_cls, &cls_err);
        printf("[MERGED] TrnRms=%.4f    TstRms=%.4f    TstCls=%.4f\n", (double)trn_rms, (double)tst_rms, (double)tst_cls);
        for(int k=0; k<cls_err.size(); k++)
            printf("err-cnt[%02d] = %4d\n", k, cls_err[k]);
        fflush(stdout);

        //
        // Create X1 network
        //
        x1_network x1net;
        x1net.start_initialization(ds_wrk.nclasses, v1layer);
        for(const x1_column &x1col: v_x1)
            x1net.append_column(x1col);
        //serialize_to(&x1net, std::string(OUT_DIR)+"/x1net-seed"+std::to_string(seed)+".x1");
        printf("Combined L1 factor = %.2f\n", (double)x1net.get_l1_factor());
        fflush(stdout);
        //predictions tmp = x1net.apply(ds_wrk);
        //printf("cls=%.4f\n", (double)tmp.tst_cls());
        //fflush(stdout);
        return x1net;
    }
    else
        return x1_network();
}

#if 0
x1_network x1_train_twostep_hardened(const raw_dataset &ds_wrk,
                            std::vector<int> columns,
                            const net_v1layer &v1layer,
                            const int window_size,
                            const train_subcolumn_settings settings,
                            const std::tuple<int,int> nc_geometry, // v2 size, v4 size
                            const std::tuple<int,int,int,int> xparams,   // subclassifier width, NN_CNT, S-harden, REP-harden
                            const int seed)
{
    //
    // normalize params
    //
    if( columns.size()==0 )
    {
        for(int i=0; i<ds_wrk.nclasses; i++)
            columns.push_back(i);
    }

    //
    //
    //
    const int64_t ntrn_full   = ds_wrk.ntrn;
    const int64_t ntotal_full = ds_wrk.ntrn+ds_wrk.ntst;
    const int   XWIDTH   = std::get<0>(xparams);
    const int   NN_CNT   = std::get<1>(xparams);
    const int   S_HARDEN = std::get<2>(xparams);
    const int   REP_HARDEN = std::get<3>(xparams);
    const float PROB1_LEVEL = 0.05f;

    ASSERT(XWIDTH>=1);
    ASSERT(NN_CNT>=1);

    std::mt19937  mt_rnd;
    std::seed_seq mt_seed{(unsigned int)seed};
    mt_rnd.seed(mt_seed);

    //
    // Save/process V1 layer
    //
    activities act_v1_raw;
    printf("--- seed %d, processing V1 ---\n", seed);
    fflush(stdout);
    v1layer.apply(ds_wrk, act_v1_raw, true);
    //serialize_to(&v1layer, std::string(OUT_DIR)+"/v1-layer.bin");

    //
    // Process columns
    //
    std::vector<predictions> v_pred;
    std::vector<x1_column>   v_x1;
    predictions baseline(ds_wrk);
    for(int ii=0; ii<columns.size(); ii++)
    {
        //
        // Train baseline column
        //
        const int col = columns[ii];
        const predictions baseline_col = predictions::onevsall(baseline,col);
        conv_subcolumn subcolumn;
        predictions pred;
        std::tie(subcolumn,pred) = train_subcolumn_baseline(ds_wrk,
                                                            col,
                                                            0, // split idx
                                                            act_v1_raw,
                                                            window_size,
                                                            settings,
                                                            0.0f, // ADVERSARIAL_CLIP_LEVEL,
                                                            nc_geometry,
                                                            seed);

        //
        // Retrain using baseline predictions as targets
        //
        std::vector<float> target_logits_full = subcolumn.apply_logit(baseline_col, act_v1_raw).kth_as_fvector(0);
        retrain_subcolumn_subspace_from_target(
                    ds_wrk,
                    baseline_col,
                    act_v1_raw,
                    target_logits_full,
                    std::make_tuple(XWIDTH,NN_CNT,0.0f),
                    subcolumn,
                    pred);

        //
        // Save results
        //
        x1_column x1col;
        x1col.append_subcolumn(subcolumn);

        v_x1.push_back(x1col);
        v_pred.push_back(pred);

        //serialize_to(&subcolumn, std::string(OUT_DIR)+"/col"  +std::to_string(col)+"-seed"+std::to_string(seed)+".bin");
        //serialize_to(&x1col,     std::string(OUT_DIR)+"/x1col"+std::to_string(col)+"-seed"+std::to_string(seed)+".x1c");
        //serialize_to(&pred,      std::string(OUT_DIR)+"/col"  +std::to_string(col)+"-seed"+std::to_string(seed)+".pred");

        printf("Final combined L1 factor = %.2f\n", (double)x1col.get_l1_factor());

        //predictions tmp = x1col.apply(baseline_col, act_v1_raw);
        //printf("cls=%.4f\n", (double)tmp.tst_cls());
        //fflush(stdout);
    }

    //
    // Combine (if we have exactly as many columns as classes)
    //
    if( v_pred.size()==ds_wrk.nclasses )
    {
        //
        // Merge predictions
        //
        const predictions pred_merged = predictions::merge_onevsall(v_pred);
        //serialize_to(&pred_merged, std::string(OUT_DIR)+"/allmerged-seed"+std::to_string(seed)+".pred");
        std::vector<int> cls_err;
        float trn_rms, tst_rms, tst_cls;
        printf("=== printing out combined predictions ===\n");
        pred_merged.compute_errors(&trn_rms, &tst_rms, &tst_cls, &cls_err);
        printf("[MERGED] TrnRms=%.4f    TstRms=%.4f    TstCls=%.4f\n", (double)trn_rms, (double)tst_rms, (double)tst_cls);
        for(int k=0; k<cls_err.size(); k++)
            printf("err-cnt[%02d] = %4d\n", k, cls_err[k]);
        fflush(stdout);

        //
        // Create X1 network
        //
        x1_network x1net;
        x1net.start_initialization(ds_wrk.nclasses, v1layer);
        for(const x1_column &x1col: v_x1)
            x1net.append_column(x1col);
        //serialize_to(&x1net, std::string(OUT_DIR)+"/x1net-seed"+std::to_string(seed)+".x1");
        printf("Combined L1 factor = %.2f\n", (double)x1net.get_l1_factor());
        fflush(stdout);
        //predictions tmp = x1net.apply(ds_wrk);
        //printf("cls=%.4f\n", (double)tmp.tst_cls());
        //fflush(stdout);
        return x1net;
    }
    else
        return x1_network();
}
#endif

conv_subcolumn train_universal_subcolumn(raw_dataset &ds_wrk,
                                         const net_v1layer &v1layer,
                                         const train_subcolumn_settings settings,
                                         const std::tuple<int,int,int> nc_w24, // windowsize, v2 size, v4 size
                                         const int seed)
{
    std::mt19937  mt_rnd;
    std::seed_seq mt_seed{(unsigned int)seed};
    mt_rnd.seed(mt_seed);
    std::uniform_int_distribution<> rnd01(0,1);

    //
    // Save original entry indexes, rewrite by random noise, set class count to 2
    //
    std::vector<int> original_class_indexes;
    int original_class_count = ds_wrk.nclasses;
    for(int i=0; i<ds_wrk.ntrn+ds_wrk.ntst; i++)
    {
        original_class_indexes.push_back(ds_wrk.entries[i].class_id);
        ds_wrk.entries[i].class_id = rnd01(mt_rnd);
    }
    ds_wrk.nclasses = 2;

    //
    // Process V1 layer
    //
    activities act_v1_raw;
    printf("--- seed %d, processing V1 ---\n", seed);
    fflush(stdout);
    v1layer.apply(ds_wrk, act_v1_raw, true);

    //
    // Train universal subcolumn
    //
    conv_subcolumn out_subcol;
    activities out_actmerged;
    _train_conv_subcolumn_v2v4(ds_wrk, 0, 0, act_v1_raw, settings, 0.0f, nc_w24, std::make_tuple(0,split_t::entire,0,0), seed, out_subcol, out_actmerged);

    //
    // Restore class indexes
    //
    ds_wrk.nclasses = original_class_count;
    for(int i=0; i<ds_wrk.ntrn+ds_wrk.ntst; i++)
        ds_wrk.entries[i].class_id = original_class_indexes[i];

    //
    // Done
    //
    return out_subcol;
}

x1_network x1_use_universal_subcolumn(const raw_dataset &ds_wrk,
                                      std::vector<int> columns,
                                      const net_v1layer &v1layer,
                                      const train_subcolumn_settings settings,
                                      const conv_subcolumn &universal,
                                      const std::tuple<int,int,int,float> xparams,   // column-dim, subclassifier width, NN_CNT, ADV_CLIP
                                      const int seed)
{
    ASSERT(std::get<1>(xparams)==0 && std::get<2>(xparams)==0);

    //
    // normalize params
    //
    if( columns.size()==0 )
    {
        for(int i=0; i<ds_wrk.nclasses; i++)
            columns.push_back(i);
    }
    const int COL_DIM = std::get<0>(xparams);

    //
    // Save/process V1 layer
    //
    activities act_v1_raw;
    printf("--- seed %d, processing V1 ---\n", seed);
    fflush(stdout);
    v1layer.apply(ds_wrk, act_v1_raw, true);
    //serialize_to(&v1layer, std::string(OUT_DIR)+"/v1-layer.bin");

    //
    // Produce activities of the universal subcolumn
    //
    activities act_universal;
    universal.process_v2v4(act_v1_raw, act_universal);

    //
    // Process columns
    //
    std::vector<predictions> v_pred;
    std::vector<x1_column>   v_x1;
    predictions baseline(ds_wrk);
    for(int ii=0; ii<columns.size(); ii++)
    {
        const int col = columns[ii];
        const predictions baseline_col = predictions::onevsall(baseline,col);
        printf("--- column %d ---\n", col);
        fflush(stdout);

        //
        // Select subspace of V2V4 activities to be used in column #col
        //
        printf("Selecting subspace...");
        fflush(stdout);
        std::vector<int> subspace = select_best_features(baseline_col, act_universal, COL_DIM);
        activities act_subspace(act_universal.ntrn,act_universal.ntst,storage_format::f32);
        act_subspace.append_realm(realm<channel>::create_grid("*",1,1,0));
        for(int cidx: subspace)
        {
            float *ptr = act_universal.get_f32(cidx);
            act_subspace.append1(activityf_vector(ptr,ptr+act_universal.ntrn+act_universal.ntst),"");
        }
        printf(" DONE\n");
        fflush(stdout);

        //
        // Train gmaxout model
        //
        const int batch_size = std::get<0>(settings.logit_batch_momentum);
        const float mu = std::get<1>(settings.logit_batch_momentum);
        int epoch_cnt;
        std::vector<float> rates;
        gmaxout_postlayer gmaxout_model({&act_subspace},1);
        std::tie(epoch_cnt,rates) = settings.logit_learning_rates;
        for(float rate: rates)
            if( rate!=0 )
                train_gmaxout(gmaxout_model, {&act_subspace}, baseline_col, epoch_cnt, batch_size, mu, rate, std::make_tuple(0.0f,settings.pre_lambda2_reg), "");
        superlogit2_predictor predictor = superlogit2_predictor::create_empty(act_universal);
        std::vector<float> coeffs_full(act_universal.nchannels+1,0.0f);
        for(int i=0; i<subspace.size(); i++)
            coeffs_full[subspace[i]] = gmaxout_model.coeffs[i];
        coeffs_full[act_universal.nchannels] = gmaxout_model.coeffs[subspace.size()];
        predictor.append1(coeffs_full);

        //
        // Subcolumn is ready
        //
        conv_subcolumn subcolumn = universal;
        subcolumn.rewrite_predictor(predictor, calibrator());
        predictions pred = subcolumn.apply(baseline_col, act_v1_raw);
        printf("Subcolumn L1 factor = %.2f\n", (double)subcolumn.get_l1_factor());

        //
        // Column is ready
        //
        x1_column x1col;
        x1col.append_subcolumn(subcolumn);

        v_x1.push_back(x1col);
        v_pred.push_back(pred);

        //serialize_to(&subcolumn, std::string(OUT_DIR)+"/col"  +std::to_string(col)+"-seed"+std::to_string(seed)+".bin");
        //serialize_to(&x1col,     std::string(OUT_DIR)+"/x1col"+std::to_string(col)+"-seed"+std::to_string(seed)+".x1c");
        //serialize_to(&pred,      std::string(OUT_DIR)+"/col"  +std::to_string(col)+"-seed"+std::to_string(seed)+".pred");

        printf("Combined L1 factor = %.2f\n", (double)x1col.get_l1_factor());
        //predictions tmp = x1col.apply(baseline_col, act_v1_raw);
        //printf("cls=%.4f\n", (double)tmp.tst_cls());
        //fflush(stdout);
    }

    //
    // Combine (if we have exactly as many columns as classes)
    //
    if( v_pred.size()==ds_wrk.nclasses )
    {
        //
        // Merge predictions
        //
        const predictions pred_merged = predictions::merge_onevsall(v_pred);
        //serialize_to(&pred_merged, std::string(OUT_DIR)+"/allmerged-seed"+std::to_string(seed)+".pred");
        std::vector<int> cls_err;
        float trn_rms, tst_rms, tst_cls;
        printf("=== printing out combined predictions ===\n");
        pred_merged.compute_errors(&trn_rms, &tst_rms, &tst_cls, &cls_err);
        printf("[MERGED] TrnRms=%.4f    TstRms=%.4f    TstCls=%.4f\n", (double)trn_rms, (double)tst_rms, (double)tst_cls);
        for(int k=0; k<cls_err.size(); k++)
            printf("err-cnt[%02d] = %4d\n", k, cls_err[k]);
        fflush(stdout);

        //
        // Create X1 network
        //
        x1_network x1net;
        x1net.start_initialization(ds_wrk.nclasses, v1layer);
        for(const x1_column &x1col: v_x1)
            x1net.append_column(x1col);
        //serialize_to(&x1net, std::string(OUT_DIR)+"/x1net-seed"+std::to_string(seed)+".x1");
        //predictions tmp = x1net.apply(ds_wrk);
        //printf("cls=%.4f\n", (double)tmp.tst_cls());
        //fflush(stdout);
        return x1net;
    }
    else
        return x1_network();
}

std::tuple<x1_network,predictions> x1_train_twostep_with_distillation(
        const raw_dataset &ds_wrk,
        const predictions &targets_to_fit,
        std::vector<int> columns,
        const net_v1layer &v1layer,
        const train_subcolumn_settings settings,
        const std::tuple<int,int,int> nc_w24, // windowsize, v2 size, v4 size
        const std::tuple<int,int,float> xparams,   // subclassifier width, NN_CNT, ADV_CLIP
        const std::tuple<int,float> xstop_cond,    // stop after K centers of prob-err below EPS
        const int seed)
{
    //
    // normalize params
    //
    if( columns.size()==0 )
    {
        for(int i=0; i<ds_wrk.nclasses; i++)
            columns.push_back(i);
    }
    predictions logits_to_fit = invsigma0(targets_to_fit);

    //
    //
    //
    const int   XWIDTH = std::get<0>(xparams);
    const int   NN_CNT = std::get<1>(xparams);
    const float ADVERSARIAL_CLIP_LEVEL = std::get<2>(xparams);
    ASSERT(XWIDTH>=0);
    ASSERT(NN_CNT>=1);

    //
    // Save/process V1 layer
    //
    activities act_v1_raw;
    printf("--- seed %d, processing V1 ---\n", seed);
    fflush(stdout);
    v1layer.apply(ds_wrk, act_v1_raw, true);
    //serialize_to(&v1layer, std::string(OUT_DIR)+"/v1-layer.bin");

    //
    // Process columns
    //
    std::vector<predictions> v_pred;
    std::vector<x1_column>   v_x1;
    predictions baseline(ds_wrk);
    for(int ii=0; ii<columns.size(); ii++)
    {
        //
        // Train baseline column
        //
        const int col = columns[ii];
        const predictions baseline_col = predictions::onevsall(baseline,col);
        conv_subcolumn subcolumn;
        predictions pred;
        std::tie(subcolumn,pred) = train_subcolumn_baseline(ds_wrk,
                                                            col,
                                                            act_v1_raw,
                                                            settings,
                                                            ADVERSARIAL_CLIP_LEVEL,
                                                            nc_w24,
                                                            std::make_tuple(0,split_t::entire,0,0),
                                                            seed);

        //
        // Retrain using baseline predictions as targets
        //
        std::vector<float> target_logits_full = logits_to_fit.kth_as_fvector(col);
        retrain_subcolumn_subspace_from_target(
                    ds_wrk,
                    baseline_col,
                    act_v1_raw,
                    target_logits_full,
                    std::make_tuple(XWIDTH,NN_CNT,ADVERSARIAL_CLIP_LEVEL),
                    xstop_cond,
                    subcolumn,
                    pred);

        //
        // Save results
        //
        x1_column x1col;
        x1col.append_subcolumn(subcolumn);

        v_x1.push_back(x1col);
        v_pred.push_back(pred);

        //serialize_to(&subcolumn, std::string(OUT_DIR)+"/col"  +std::to_string(col)+"-seed"+std::to_string(seed)+".bin");
        //serialize_to(&x1col,     std::string(OUT_DIR)+"/x1col"+std::to_string(col)+"-seed"+std::to_string(seed)+".x1c");
        //serialize_to(&pred,      std::string(OUT_DIR)+"/col"  +std::to_string(col)+"-seed"+std::to_string(seed)+".pred");

        printf("Final combined L1 factor = %.2f\n", (double)x1col.get_l1_factor());

        //predictions tmp = x1col.apply(baseline_col, act_v1_raw);
        //printf("cls=%.4f\n", (double)tmp.tst_cls());
        //fflush(stdout);
#if 0
        void retrain_subcolumn_subspace_from_target(
                const raw_dataset &ds_wrk,
                const predictions &baseline_col,
                activities  &act_v1_raw,
                std::vector<float> target_logits_full,
                const std::tuple<int,int> xparams,
                conv_subcolumn &subcolumn,
                predictions &pred)

        //
        // Compute V2-V4 activities and build a kd-tree
        //
        activities  act_v2v4_full;
        subcolumn.process_v2v4(act_v1_raw, act_v2v4_full);
        const int64_t nchannels_full = act_v2v4_full.nchannels;
        alglib::real_2d_array    kdt_xy;
        alglib::integer_1d_array kdt_tags;
        alglib::kdtree kdtree;
        int class0_size = 0;
        kdt_xy.setlength(ntrn_full, nchannels_full);
        kdt_tags.setlength(ntrn_full);
        for(int i=0; i<ntrn_full; i++)
            if( baseline_col.labels[i]==0 )
            {
                for(int j=0; j<nchannels_full; j++)
                    kdt_xy[class0_size][j] = act_v2v4_full.get_normalized(j,i);
                kdt_tags[class0_size] = i;
                class0_size++;
            }
        alglib::kdtreebuildtagged(kdt_xy, kdt_tags, class0_size, nchannels_full, 0, 2, kdtree);

        //
        // Prepare data for subclassifier training:
        // * V2V4 activities
        // * target output logits and probabilities (after applying BIG_LOGIT cap)
        // * current output probabilities (zero)
        //
        std::vector<float> target_logits_full = subcolumn.apply_logit(baseline_col, act_v1_raw).kth_as_fvector(0);
        std::vector<float> target_probs_full  = pred.kth_as_fvector(0);
        std::vector<float> current_probs_full(ntotal_full,0.0f);
        target_logits_full.resize(ntrn_full);
        target_probs_full.resize(ntrn_full);
        for(int i=0; i<ntrn_full; i++)
            if( baseline_col.labels[i]!=0 )
            {
                target_logits_full[i] = fmaxf(target_logits_full[i], -BIG_LOGIT); // relax constraint
                target_probs_full[i]  = fmaxf(target_probs_full[i],  PROB_EPS);   // relax constraint
            }
            else
            {
                target_logits_full[i] = fminf(target_logits_full[i], +BIG_LOGIT); // relax constraint
                target_probs_full[i]  = fminf(target_probs_full[i],  1-PROB_EPS); // relax constraint
            }

        //
        // Iteratively select worst point until error in output probabilities decreases below PROB_TOL
        //
        superlogit2_predictor predictor = superlogit2_predictor::create_empty(act_v2v4_full);
        predictions current_pred = baseline_col;
        for(int ncenters=0; ; ncenters++)
        {
            int t0 = dtickcount();

            //
            // Select target point and its neighbors
            //
            int target_idx    = -1;
            float target_err =  0;
            for(int i=0; i<ntrn_full; i++)
            {
                float err = target_probs_full[i]-current_probs_full[i];
                if( baseline_col.labels[i]==0 && err>target_err )
                {
                    target_idx = i;
                    target_err = err;
                }
            }
            if( target_idx<0 || target_err<=0 )
                break;
            float prev_central_prob = current_probs_full[target_idx];
            alglib::real_1d_array kdt_x;
            alglib::integer_1d_array kdt_idx;
            kdt_x.setlength(nchannels_full);
            for(int i=0; i<nchannels_full; i++)
                kdt_x[i] = act_v2v4_full.get_normalized(i,target_idx);
            int kneighbors = alglib::kdtreequeryknn(kdtree, kdt_x, NN_CNT);
            alglib::kdtreequeryresultstags(kdtree, kdt_idx);
            std::vector<int> targets;
            for(int i=0; i<kneighbors; i++)
                targets.push_back(kdt_idx[i]);
            ASSERT(targets.size()>=1);

            //
            // Determine subspace to work with
            //
            std::vector<int> subspace;
            //for(int i=0; i<nchannels_full; i++)
            //    subspace.push_back(i);
            //const int nchannels_subspace = subspace.size();
            if( XWIDTH>0 && XWIDTH<nchannels_full )
            {
                std::vector<int> xx_items;
                int xx_repeat_factor = imax(class0_size/targets.size(),1);
                for(int i=0; i<ntrn_full; i++)
                    if( baseline_col.labels[i]!=0 )
                        xx_items.push_back(i);
                for(int k: targets)
                    for(int i=0; i<xx_repeat_factor; i++)
                        xx_items.push_back(k);
                for(int i=ntrn_full; i<ntotal_full; i++)
                    xx_items.push_back(i);
                std::sort(xx_items.begin(), xx_items.end());
                predictions xx_baseline = baseline_col.select_subsample(xx_items);
                activities xx_act_v2v4;
                act_v2v4_full.select_subsample(xx_items, xx_act_v2v4);
                subspace = select_best_features(xx_baseline, xx_act_v2v4, XWIDTH);
            }
            else
            {
                ASSERT(XWIDTH==0 || XWIDTH==nchannels_full);
                for(int i=0; i<nchannels_full; i++)
                    subspace.push_back(i);
            }
            const int nchannels_subspace = subspace.size();

            //
            // Select subspace of activities
            //
            activities act_v2v4_subspace(act_v2v4_full.ntrn,act_v2v4_full.ntst,storage_format::f32);
            act_v2v4_subspace.append_realm(realm<channel>::create_grid("*",1,1,0));
            for(int cidx: subspace)
            {
                float *ptr = act_v2v4_full.get_f32(cidx);
                act_v2v4_subspace.append1(activityf_vector(ptr,ptr+ntotal_full),"");
            }

            // generate constraints
            const int xdim = nchannels_subspace+1;
            alglib::integer_1d_array ct;
            alglib::real_2d_array linconstr;
            int nconstr = 0;
            linconstr.setlength(ntrn_full,xdim+1);
            ct.setlength(ntrn_full);
            for(int i=0; i<ntrn_full; i++)
                if( baseline_col.labels[i]!=0 )
                {
                    for(int j=0; j<xdim-1; j++)
                        linconstr[nconstr][j] = act_v2v4_subspace.get_normalized(j,i);
                    linconstr[nconstr][xdim-1] = 1.0;
                    linconstr[nconstr][xdim] = target_logits_full[i];
                    ct[nconstr] = -1;
                    nconstr++;
                }
            for(int i: targets)
            {
                ASSERT(baseline_col.labels[i]==0);
                ASSERT(nconstr<linconstr.rows());
                for(int j=0; j<xdim-1; j++)
                    linconstr[nconstr][j] = act_v2v4_subspace.get_normalized(j,i);
                linconstr[nconstr][xdim-1] = 1.0;
                linconstr[nconstr][xdim] = target_logits_full[i];
                ct[nconstr] = +1;
                nconstr++;
            }

            // optimize
            alglib::real_1d_array xf;
            try
            {
                alglib::minqpstate  qpsolver;
                alglib::minqpreport qprep;
                alglib::real_2d_array qterm;
                qterm.setlength(xdim, xdim);
                for(int i=0; i<xdim; i++)
                    for(int j=0; j<xdim; j++)
                        qterm[i][j] = i==j ? 1 : 0;
                alglib::minqpcreate(xdim, qpsolver);
                alglib::minqpsetalgodenseaul(qpsolver, 1.0e-6, 1000.0, 10);
                alglib::minqpsetquadraticterm(qpsolver, qterm);
                alglib::minqpsetlc(qpsolver, linconstr, ct, nconstr);
                alglib::minqpoptimize(qpsolver, alglib::parallel);
                alglib::minqpresults(qpsolver, xf, qprep);
                ASSERT(qprep.terminationtype>0);
            }
            catch(alglib::ap_error e)
            {
                printf("[EXCEPTION]: '%s'\n", e.msg.c_str());
                fflush(stdout);
                throw;
            }

            // append to predictor
            std::vector<float> coeffs_full(nchannels_full+1,0.0f);
            for(int i=0; i<nchannels_subspace; i++)
                coeffs_full[subspace[i]] = xf[i];
            coeffs_full[nchannels_full] = xf[nchannels_subspace];
            predictor.append1(coeffs_full);

            // update logits, report
            float feas_err = 0, prob_err = 0;
            float target_l2 = 0, target_l1 = 0;
            for(int i=0; i<nconstr; i++)
            {
                double v = 0;
                for(int j=0; j<xdim; j++)
                    v += linconstr[i][j]*xf[j];
                v -= linconstr[i][xdim];
                ASSERT(ct[i]!=0);
                if( ct[i]>0 )
                    feas_err = fmax(-v, feas_err);
                else
                    feas_err = fmax(+v, feas_err);
                //if( feas_err>1.0e-2 )
                //{
                //    printf("LC=");
                //    for(int j=0; j<=xdim; j++)
                //        printf("%6.2f ", linconstr[i][j]);
                //    printf("\n");
                //    printf("CT=%d\n", (int)ct[i]);
                //    printf("V=%.3e\n", v);
                //    fflush(stdout);
                //    ASSERT(0);
                //}
            }
            for(int i=0; i<xdim-1; i++)
            {
                target_l1 += fabs(xf[i]);
                target_l2 += xf[i]*xf[i];
            }
            current_pred = baseline_col;
            for(int i=0; i<ntotal_full; i++)
            {
                float v = xf[xdim-1];
                for(int j=0; j<xdim-1; j++)
                    v += act_v2v4_subspace.get_normalized(j,i)*xf[j];
                v = sigma0(v);
                current_probs_full[i] = fmaxf(v,current_probs_full[i]);
                current_pred.fpredictions[i] = current_probs_full[i];
                current_pred.fpredictions[i+current_pred.stride] = 1.0f-current_probs_full[i];
            }
            for(int i=0; i<ntrn_full; i++)
                if( baseline_col.labels[i]==0 )
                    prob_err  = fmaxf(prob_err, target_probs_full[i]-current_probs_full[i]);
            float new_central_prob = current_probs_full[target_idx];
            printf("[%03d] prob(trg)=%.4f->%.4f  L1=%6.1f  L2=%6.1f  prob_err=%.4f  trn_rms=%.4f  tst_rms=%.4f  tst_cls=%.4f (%6d ms, feas_err=%.2e)\n",
                   ncenters,
                   prev_central_prob, new_central_prob,
                   target_l1, target_l2,
                   prob_err,
                   current_pred.trn_rms(), current_pred.tst_rms(), current_pred.tst_cls(),
                   dtickcount()-t0, feas_err);
            fflush(stdout);
            if( feas_err>1.0e-2 )
            {
                printf("[error] unable to solve QP problem with constraints, maybe we do not have enough meaningful features to separate classes\n");
                fflush(stdout);
                ASSERT(0);
            }
            if( prob_err<PROB_TOL )
                break;
        }

        //
        // Update subcolumn and predictions
        //
        subcolumn.rewrite_predictor(predictor, calibrator());
        pred = current_pred;

        //
        // Save results
        //
        x1_column x1col;
        x1col.append_subcolumn(subcolumn);

        v_x1.push_back(x1col);
        v_pred.push_back(pred);

        //serialize_to(&subcolumn, std::string(OUT_DIR)+"/col"  +std::to_string(col)+"-seed"+std::to_string(seed)+".bin");
        //serialize_to(&x1col,     std::string(OUT_DIR)+"/x1col"+std::to_string(col)+"-seed"+std::to_string(seed)+".x1c");
        //serialize_to(&pred,      std::string(OUT_DIR)+"/col"  +std::to_string(col)+"-seed"+std::to_string(seed)+".pred");

        printf("Final combined L1 factor = %.2f\n", (double)x1col.get_l1_factor());

        //predictions tmp = x1col.apply(baseline_col, act_v1_raw);
        //printf("cls=%.4f\n", (double)tmp.tst_cls());
        //fflush(stdout);
#endif
    }

    //
    // Combine (if we have exactly as many columns as classes)
    //
    if( v_pred.size()==ds_wrk.nclasses )
    {
        //
        // Merge predictions
        //
        const predictions pred_merged = predictions::merge_onevsall(v_pred);
        //serialize_to(&pred_merged, std::string(OUT_DIR)+"/allmerged-seed"+std::to_string(seed)+".pred");
        std::vector<int> cls_err;
        float trn_rms, tst_rms, tst_cls;
        printf("=== printing out combined predictions ===\n");
        pred_merged.compute_errors(&trn_rms, &tst_rms, &tst_cls, &cls_err);
        printf("[MERGED] TrnRms=%.4f    TstRms=%.4f    TstCls=%.4f\n", (double)trn_rms, (double)tst_rms, (double)tst_cls);
        for(int k=0; k<cls_err.size(); k++)
            printf("err-cnt[%02d] = %4d\n", k, cls_err[k]);
        fflush(stdout);

        //
        // Create X1 network
        //
        x1_network x1net;
        x1net.start_initialization(ds_wrk.nclasses, v1layer);
        for(const x1_column &x1col: v_x1)
            x1net.append_column(x1col);
        //serialize_to(&x1net, std::string(OUT_DIR)+"/x1net-seed"+std::to_string(seed)+".x1");
        printf("Combined L1 factor = %.2f\n", (double)x1net.get_l1_factor());
        fflush(stdout);
        //predictions tmp = x1net.apply(ds_wrk);
        //printf("cls=%.4f\n", (double)tmp.tst_cls());
        //fflush(stdout);
        return std::make_tuple(x1net,pred_merged);
    }
    else
        return std::make_tuple(x1_network(),predictions());
}

}
