/*************************************************************************
StrongNet/Contour Engine training framework.
Copyright (c) 2020 Sergey Bochkanov (ALGLIB project).

>>> SOURCE LICENSE >>>
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation (www.fsf.org); either version 2 of the 
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

A copy of the GNU General Public License is available at
http://www.fsf.org/licensing/licenses
>>> END OF LICENSE >>>
*************************************************************************/
#ifndef SNETTRN_H
#define SNETTRN_H

#include "snet.h"

namespace strongnet
{

class train_subcolumn_settings
{
public:
    int gain_levels;
    float omega_rat;

    float pre_lambda2_reg;
    float post_lambda2_reg;
    float post_lambda1_reg;
    float rbf_ort_reg;

    and_type_t top_and_type;
    float top_leak_max;
    float top_leak_and;
    float top_leak_exp;

    // minimum growth of rating for generalizer
    float gen_rating_growth;

    // batch size for V2-V4
    int v2v4_batch_size;

    // batch size and momentum for logit training
    std::tuple<int,float> logit_batch_momentum;

    // vector of learning rates and shared epoch count (same for each rate)
    std::tuple<int,std::vector<float>> logit_learning_rates;

    train_subcolumn_settings():
        gain_levels(1),
        omega_rat(0.100f),
        pre_lambda2_reg( 0.00001f),
        post_lambda2_reg(0.00001f),
        post_lambda1_reg(0.0f),
        top_and_type(and_type_t::AND_MIN),
        top_leak_max(0.001f),
        top_leak_and(0.001f),
        top_leak_exp(0.001f),
        rbf_ort_reg(0.01f),
        gen_rating_growth(1.1f),
        v2v4_batch_size(10),
        logit_batch_momentum(std::make_tuple(128,0.98)),
        logit_learning_rates(std::make_tuple( 100,std::vector<float>{0.3f}))
    {}
};

// train network with simple linear classifier layer
std::tuple<x1_network,predictions> x1_train_baseline(
        const raw_dataset &ds_wrk,
        std::vector<int> columns,
        const net_v1layer &v1layer,
        const train_subcolumn_settings settings,
        const float clip_level,
        const std::tuple<int,int,int> nc_w24, // window size, v2 size, v4 size
        const std::tuple<int,split_t,int,int> co_geo, // window size, v1-split, v2 planes, v4 detectors
        const int seed);

// train binary network
std::tuple<x1_binary,predictions> x1_train_multiflavor_binary(
        const raw_dataset &ds_wrk,
        const net_v1layer &v1layer,
        const train_subcolumn_settings settings,
        const float clip_level,
        const std::tuple<int,int,int> nc_w24, // window size, v2 size, v4 size
        const std::tuple<int,split_t,int,int> co_geo, // window size, v1-split, v2 planes, v4 detectors
        const int seed);

// baseline retraining: we use V2V4 part of network passed to us and retrain just final linear classifier;
// no explicit knowledge transfer is performed
std::tuple<x1_network,predictions> x1_retrain_baseline_notransfer(
        x1_network ref_net,
        const raw_dataset &ds_wrk,
        std::vector<int> columns,
        const train_subcolumn_settings settings,
        const float clip_level);

// subspace retraining: first we use network passed to us to generate target logits,
// then we train new classifier layer on top of same V2-V4 with QP-based multi-max training
std::tuple<x1_network,predictions> x1_retrain_subspace(
        x1_network ref_net,
        const raw_dataset &ds_wrk,
        std::vector<int> columns,
        const std::tuple<int,int,float> xparams,
        const std::tuple<int,float> xstop_cond);

// two-step training: first we generate V2-V4 and train flat "teacher" model,
// then we transfer knowledge from teacher to student with QP-based multi-max training
x1_network x1_train_twostep(const raw_dataset &ds_wrk,
                            std::vector<int> columns,
                            const net_v1layer &v1layer,
                            const train_subcolumn_settings settings,
                            const std::tuple<int, int, int> nc_w24, // window size, v2 size, v4 size
                            const std::tuple<int,int,float> xparams,   // subclassifier width, NN_CNT, ADV_CLIP,
                            const std::tuple<int,float> xstop_cond,    // stop after K centers of prob-err below EPS
                            const int seed);

// two-step hardened training: first we generate V2-V4 and train flat "teacher" model,
// then we transfer knowledge from teacher to student with QP-based multi-max-of-min training
x1_network x1_train_twostep_hardened(const raw_dataset &ds_wrk,
                            std::vector<int> columns,
                            const net_v1layer &v1layer,
                            const int window_size,
                            const train_subcolumn_settings settings,
                            const std::tuple<int,int> nc_geometry, // v2 size, v4 size
                            const std::tuple<int,int,int,int> xparams,   // subclassifier width, NN_CNT, S-harden, REP-harden
                            const int seed);

// train universal subcolumn, which is trained on a random noise in class channel
conv_subcolumn train_universal_subcolumn(raw_dataset &ds_wrk,
                                         const net_v1layer &v1layer,
                                         const train_subcolumn_settings settings,
                                         const std::tuple<int, int, int> nc_w24, // window size, v2 size, v4 size
                                         const int seed);

// two-step training: first we generate V2-V4 and train flat "teacher" model,
// then we transfer knowledge from teacher to student with QP-based multi-max training
x1_network x1_use_universal_subcolumn(const raw_dataset &ds_wrk,
                                      std::vector<int> columns,
                                      const net_v1layer &v1layer,
                                      const train_subcolumn_settings settings,
                                      const conv_subcolumn &universal,
                                      const std::tuple<int,int,int,float> xparams,   // subclassifier width, NN_CNT, ADV_CLIP
                                      const int seed);

// two-step training with distillation:
// * first we generate V2-V4 and train flat "primer" model
// * then we train final model which tries to fit targets supplied by external teacher
//   with QP-based multi-max training
std::tuple<x1_network,predictions> x1_train_twostep_with_distillation(const raw_dataset &ds_wrk,
        const predictions &targets_to_fit,
        std::vector<int> columns,
        const net_v1layer &v1layer,
        const train_subcolumn_settings settings,
        const std::tuple<int, int, int> nc_w24, // windowsize, v2 size, v4 size
        const std::tuple<int,int,float> xparams,   // subclassifier width, NN_CNT, ADV_CLIP
        const std::tuple<int,float> xstop_cond,    // stop after K centers of prob-err below EPS
        const int seed);

}

#endif
