/*************************************************************************
StrongNet/Contour Engine training framework.
Copyright (c) 2020 Sergey Bochkanov (ALGLIB project).

>>> SOURCE LICENSE >>>
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation (www.fsf.org); either version 2 of the 
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

A copy of the GNU General Public License is available at
http://www.fsf.org/licensing/licenses
>>> END OF LICENSE >>>
*************************************************************************/
#ifndef SNETCOMMON_H
#define SNETCOMMON_H

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <cstddef>
#include <math.h>
#include <vector>
#include "3rdparty/lodepng/lodepng.h"
#include <atomic>
#include <map>
#include "3rdparty/alglib/ap.h"

#define SNET_MAX_SEARCH_RAD         3
#define SNET_UCOLOR_AMPLIFY         1.0f
#define SNET_MAX_V2_GEN_COUNT       3
#define SNET_MAX_V2C_GEN_COUNT      10
#define SNET_MAX_V4_GEN_COUNT       5

#define SNET_EVAL_QUEUE_SIZE        4
#define SNET_I2F                    0.01f
#define SNET_IMAX                   100
#define FLT_PI                      3.14159265358979323846f

#define ASSERT(x) if(!(x)) { printf("Assertion failed at line %d, file %s\n", (int)__LINE__, __FILE__); fflush(stdout); abort(); }
#define DEFAULT_ALIGNMENT   64

#define sPATH_OUT  std::string(OUT_DIR)
#define sPATH_WORK std::string(WORK_DIR)

namespace strongnet
{
int dtickcount();
void* aligned_malloc(size_t size, size_t alignment);
void aligned_free(void *block);

/************************************************************************
Performance timer
************************************************************************/
class timer
{
private:
    std::atomic<int32_t> &counter;
    double started;
public:
    timer(std::atomic<int32_t> &_c):counter(_c){ started=dtickcount(); }
    ~timer() { counter += dtickcount()-started; }
};

template <typename T, std::size_t Alignment = DEFAULT_ALIGNMENT>
class aligned_allocator
{
    public:

        // The following will be the same for virtually all allocators.
        typedef T * pointer;
        typedef const T * const_pointer;
        typedef T& reference;
        typedef const T& const_reference;
        typedef T value_type;
        typedef std::size_t size_type;
        typedef ptrdiff_t difference_type;

        T * address(T& r) const
        {
            return &r;
        }

        const T * address(const T& s) const
        {
            return &s;
        }

        std::size_t max_size() const
        {
            // The following has been carefully written to be independent of
            // the definition of size_t and to avoid signed/unsigned warnings.
            return (static_cast<std::size_t>(0) - static_cast<std::size_t>(1)) / sizeof(T);
        }


        // The following must be the same for all allocators.
        template <typename U>
        struct rebind
        {
            typedef aligned_allocator<U, Alignment> other;
        } ;

        bool operator!=(const aligned_allocator& other) const
        {
            return !(*this == other);
        }

        void construct(T * const p, const T& t) const
        {
            void * const pv = static_cast<void *>(p);

            new (pv) T(t);
        }

        void destroy(T * const p) const
        {
            p->~T();
        }

        // Returns true if and only if storage allocated from *this
        // can be deallocated from other, and vice versa.
        // Always returns true for stateless allocators.
        bool operator==(const aligned_allocator& other) const
        {
            return true;
        }


        // Default constructor, copy constructor, rebinding constructor, and destructor.
        // Empty for stateless allocators.
        aligned_allocator() { }

        aligned_allocator(const aligned_allocator&) { }

        template <typename U> aligned_allocator(const aligned_allocator<U, Alignment>&) { }

        ~aligned_allocator() { }


        // The following will be different for each allocator.
        T * allocate(const std::size_t n) const
        {
            // The return value of allocate(0) is unspecified.
            // Mallocator returns NULL in order to avoid depending
            // on malloc(0)'s implementation-defined behavior
            // (the implementation can define malloc(0) to return NULL,
            // in which case the bad_alloc check below would fire).
            // All allocators can return NULL in this case.
            if (n == 0)
            {
                return NULL;
            }

            // All allocators should contain an integer overflow check.
            // The Standardization Committee recommends that std::length_error
            // be thrown in the case of integer overflow.
            if (n > max_size())
            {
                throw std::bad_alloc();//throw std::length_error("aligned_allocator<T>::allocate() - Integer overflow.");
            }

            // Mallocator wraps malloc().
            void * const pv = aligned_malloc(n*sizeof(T), Alignment);

            // Allocators should throw std::bad_alloc in the case of memory allocation failure.
            if (pv == NULL)
            {
                throw std::bad_alloc();
            }

            return static_cast<T *>(pv);
        }

        void deallocate(T * const p, const std::size_t n) const
        {
            aligned_free(p);

            // just to avoid notes about unreferenced variables
            volatile std::size_t k = n+1;
            k--;
        }


        // The following will be the same for all allocators that ignore hints.
        template <typename U>
        T * allocate(const std::size_t n, const U * /* const hint */) const
        {
            return allocate(n);
        }


        // Allocators are not required to be assignable, so
        // all allocators should have a private unimplemented
        // assignment operator. Note that this will trigger the
        // off-by-default (enabled under /Wall) warning C4626
        // "assignment operator could not be generated because a
        // base class assignment operator is inaccessible" within
        // the STL headers, but that warning is useless.
    private:
        aligned_allocator& operator=(const aligned_allocator&);
};

float vnrm2(const int8_t *p, int64_t n);
float vnrm2(const float *p, int64_t n);
float vdist2(const float *p0, const float *p1, int64_t n);
float i8nrm2(const int8_t *p, int64_t n);
void fadd_simd(float *p0, const float *p1, int64_t n, float s);
float fdot_simd(const float *p0, const float *p1, int64_t n);
void fadd_diffc2(float *dst, const float *p0, float c, int64_t n);
void fadd_diff2(float *dst, const float *p0, const float *p1, int64_t n);

void fmerge_min(float *dst, const float *src, int64_t n);
void fmerge_max(float *dst, const float *src, int64_t n);
void fmerge_mul(float *dst, const float *src, int64_t n);
void fmerge_wmax(float *dst, const float *src, float w, int64_t n);

void fmin_inplace(float *dst, float v, int64_t n);
void fmul_inplace(float *dst, float v, int64_t n);
void fmul_inplace(float *dst, const float *src, int64_t n);
void fmul_add_inplace(float *srcdst, const float *mul, const float *add, int64_t n);
void fadd_mul_inplace(float *srcdst, const float *add0, const float *add1, int64_t n);

typedef std::vector<int8_t,aligned_allocator<int8_t> > activity8_vector;
typedef std::vector<float, aligned_allocator<float>  > activityf_vector;
typedef std::vector<float,aligned_allocator<float> >   fvector;

static inline int irange(int x, int a, int b)
{
    if( x<a )
        return a;
    if( x>b )
        return b;
    return x;
}
static inline float frange(float x, float a, float b)
{
    if( x<a )
        return a;
    if( x>b )
        return b;
    return x;
}

static inline int imin(int x, int y)
{
    return x<y ? x : y;
}
static inline int imax(int x, int y)
{
    return x>y ? x : y;
}
static inline int imax3(int x, int y, int z)
{
    if( x>y )
        return x>z ? x : z;
    else
        return y>z ? y : z;
}
static inline int imin3(int x, int y, int z)
{
    if( x<y )
        return x<z ? x : z;
    else
        return y<z ? y : z;
}
static inline int64_t imin64(int64_t x, int64_t y)
{
    return x<y ? x : y;
}
static inline int64_t imax64(int64_t x, int64_t y)
{
    return x>y ? x : y;
}
static inline float fmax3f(float x, float y, float z)
{
    if( x>y )
        return x>z ? x : z;
    else
        return y>z ? y : z;
}
static inline int8_t flt_to_int8(float v)
{
    v = v<-1.0f ? -1.0f : v;
    v = v>+1.0f ? +1.0f : v;
    return (int8_t)roundf(100.0f*v);
}
static inline int8_t posflt_to_int8(float v)
{
    if( v<=0 )
        return 0;
    if( v>=1 )
        return 100;
    return (int8_t)roundf(100.0f*v);
}
static inline float int8_to_flt(int8_t v)
{
    return 0.01f*v;
}
static inline int iDivUp(int x, int y)
{
    int result = x/y;
    if( x%y )
        result++;
    return result;
}

void serialize_long_blob(FILE *f, const void *blob, int64_t len);
void serialize64(FILE *f, int64_t v);
void write_magic8(FILE *f, int8_t magic);
void serialize(FILE *f, const std::string *s);
void serialize(FILE *f, const unsigned char *p_i);
void serialize(FILE *f, const int8_t *p_i);
void serialize(FILE *f, const int *p_i);
void serialize(FILE *f, const int64_t *p_i);
void serialize(FILE *f, const float *p_v);
void unserialize_long_blob(FILE *f, void *blob, int64_t len);
int64_t unserialize64(FILE *f);
int8_t check_magic8(FILE *f, int8_t magic);
int8_t check_magic8(FILE *f, int8_t magic0, int8_t magic1);
void unserialize(FILE *f, std::string *s);
void unserialize(FILE *f, unsigned char *p_i);
void unserialize(FILE *f, int8_t *p_i);
void unserialize(FILE *f, int *p_i);
void unserialize(FILE *f, int64_t *p_i);
void unserialize(FILE *f, float *p_v);

class rgb
{
public:
    float r;       // percent
    float g;       // percent
    float b;       // percent
};

class hsv
{
public:
    float h;       // angle in degrees
    float s;       // percent
    float v;       // percent
};

hsv hsvfromi24(int irgb);
int i24fromhsv(hsv v);
hsv rgb2hsv(float r, float g, float b);
rgb hsv2rgb(float h, float s, float v);
hsv rgb2hsv(const rgb &in);
rgb hsv2rgb(const hsv &in);

class bitmap
{
private:
public:
    int height, width;
    std::vector<unsigned char> data;

    enum class project_t { clip=0, mirror=1 };

    class trf
    {
    friend class bitmap;
    private:
        enum class type_t { none = 0, pipeline1 = 1};

        type_t trf_type;

        // pipeline #1
        bool  flip_h, flip_v;
        int crop_h, crop_w, crop_offs0, crop_offs1;
        float rot;
        float fscale0, fscale1;
        float fshift0, fshift1;

        // elastic pipeline
        float elastic_rad, elastic_mag;
        bool  elastic_fixed_bnd;
        int   elastic_seed;
    public:
        trf();
        void set_flips(bool h, bool v);
        void set_scale(float s);
        void set_fshift(float s0, float s1);
        void set_rot(float rad_rot);
        void set_crop(int h, int w, int offs0, int offs1);
    };

    bitmap():height(0),width(0){}
    bitmap(int h, int w){resize(h,w);}

    void resize(int h, int w)
    {
        height = h;
        width  = w;
        data.clear();
        data.resize(h*w*3, 0x0);
    }

    void resize_v(int h, int w, float val)
    {
        height = h;
        width  = w;
        data.clear();
        data.resize(h*w*3, irange(val*255, 0, 255));
    }

    void draw_v(int pos_h, int pos_w, float val)
    {
        if( pos_h<0 || pos_h>=height || pos_w<0 || pos_w>=width )
            return;
        unsigned char c = irange(val*255, 0, 255);
        data[(pos_h*width+pos_w)*3+0] = c;
        data[(pos_h*width+pos_w)*3+1] = c;
        data[(pos_h*width+pos_w)*3+2] = c;
    }

    void draw_rgb(int pos_h, int pos_w, int r, int g, int b)
    {
        if( pos_h<0 || pos_h>=height || pos_w<0 || pos_w>=width )
            return;
        data[(pos_h*width+pos_w)*3+0] = irange(r, 0, 255);
        data[(pos_h*width+pos_w)*3+1] = irange(g, 0, 255);
        data[(pos_h*width+pos_w)*3+2] = irange(b, 0, 255);
    }

    void draw_rgb(int pos_h, int pos_w, rgb z)
    {
        draw_rgb(pos_h, pos_w, roundf(z.r*255), roundf(z.g*255), roundf(z.b*255));
    }

    void get_rgb(int pos_h, int pos_w, int *r, int *g, int *b) const
    {
        if( pos_h<0 || pos_h>=height || pos_w<0 || pos_w>=width )
        {
            *r = 0;
            *g = 0;
            *b = 0;
            return;
        }
        *r = data[(pos_h*width+pos_w)*3+0];
        *g = data[(pos_h*width+pos_w)*3+1];
        *b = data[(pos_h*width+pos_w)*3+2];
    }

    void draw_24(int pos_h, int pos_w, int rgb)
    {
        if( pos_h<0 || pos_h>=height || pos_w<0 || pos_w>=width )
            return;
        data[(pos_h*width+pos_w)*3+0] = rgb&0xFF;
        data[(pos_h*width+pos_w)*3+1] = (rgb>>8)&0xFF;
        data[(pos_h*width+pos_w)*3+2] = (rgb>>16)&0xFF;
    }

    int get_24(int pos_h, int pos_w) const
    {
        pos_h = irange(pos_h, 0, height-1);
        pos_w = irange(pos_w, 0, width-1);
        return int(data[(pos_h*width+pos_w)*3+0]) | (int(data[(pos_h*width+pos_w)*3+1])<<8) | (int(data[(pos_h*width+pos_w)*3+2])<<16);
    }

    hsv get_hsv(int pos_h, int pos_w) const
    {
        return hsvfromi24(get_24(pos_h,pos_w));
    }

    int get_24_lanczos(float pos_h, float pos_w) const;
    int get_24_bilin(float pos_h, float pos_w) const;

    void to_png(std::string s) const
    {
        to_png(s.c_str());
    }
    void to_png(const char *filename) const
    {
        unsigned error = lodepng_encode24_file(filename, &data[0], (unsigned int)width, (unsigned int)height);
        ASSERT(!error);
    }
    void to_jpg(std::string filename, int compress = 95) const;

    static bitmap from_ppm(const char *filename);
    static bitmap from_jpg(std::string sfilename);
    static bitmap from_png(std::string sfilename);

    bitmap rough_resize_min(int min_width, int min_height) const;
    bitmap rough_resize_to (int new_width, int new_height) const;
    bitmap lanczos_resize_to (int new_width, int new_height) const;
    bitmap bilin_resize_to (int new_width, int new_height) const;
    bitmap smooth_resize_to(int new_width, int new_height) const;

    bitmap append_border_v(int border, float v) const;
    bitmap append_border_v(int top, int bottom, int left, int right, float v) const;
    bitmap central_transform_f(float fshift0, float fshift1, float fscale0, float fscale1, float rad_rot) const;
    bitmap central_crop(int size) const;
    bitmap invert() const;
    bitmap normalize_contrast() const;
    bitmap transform(const trf &t) const;
    bitmap shift_by(int shift0, int shift1) const;
    bitmap select_pos_xy(int x0, int y0, int x1, int y1) const;

    bitmap even_odd_upsample2() const;
    bitmap downsample2() const;
    bitmap downsample2k(int k) const;
    static bitmap combine_ltr(const std::vector<bitmap> &v, int border = 1, int border_color = 0xFFFFFF);
    static bitmap combine_grid_w(const std::vector<bitmap> &v, int grid_width, int border = 1, int border_color = 0xFFFFFF);
    static float project(float x, project_t prj, float a, float b);
};
void serialize(FILE *f, const bitmap *ptr);
void unserialize(FILE *f, bitmap *ptr);

class raw_dataset
{
private:
    static bitmap transform(const bitmap &img, float rad_rot);
public:
    class entry
    {
    public:
        bitmap img;
        int class_id;
        std::string flavor;

        entry():class_id(0){}
        entry(const bitmap &i, int c):img(i),class_id(c){}
        entry(const bitmap &i, int c, const std::string &f):img(i),class_id(c),flavor(f){}
    };

    int64_t ntrn;
    int64_t ntst;
    int64_t nclasses;

    std::vector<entry> entries;

    raw_dataset():ntrn(0),ntst(0),nclasses(0){}
    raw_dataset(int nc):ntrn(0),ntst(0),nclasses(nc){}
    raw_dataset(std::string filename);

    void append(const raw_dataset &ds);
    void append_trn(const entry &e);
    void append_tst(const entry &e);
    static raw_dataset merge(std::vector<const raw_dataset*> datasets);

    int64_t get_ntotal() const;
    int64_t count_trn_class(int class_id) const;

    void debug_visualize(std::string spath, int max_files = 0);
    void debug_visualize_all(const char *path);

    void transform_inplace(const std::vector<bitmap::trf> &v);
    void hue_rotate_scale_shift_inplace(float deg_hue, float deg_rot, float scale_dev, float fshift_mag, bitmap::project_t prj = bitmap::project_t::mirror, int seed=1);
    void rotate_shift_inplace(float deg_rot, float fshift_mag, bitmap::project_t prj = bitmap::project_t::mirror, int seed=1);

    // these functions return new copy of the dataset, original one is unchanged
    raw_dataset duplicate_trn(int factor) const;
    raw_dataset duplicate_all(int factor) const;
    raw_dataset shift_by(int shift0, int shift1) const;
    raw_dataset subsample(int new_trn, int new_tst, int seed = 1) const;
    raw_dataset augment(int afactor, float shift_mag, float scale_mag_lo, float deg_rot) const;
    raw_dataset augment_hflip() const;
    raw_dataset augment_hflip_crop(int afactor, int crop_to) const;
    raw_dataset augment_hflip_crop_rot(int afactor, int crop_to, float deg_rot) const;
    raw_dataset augment_shear(int afactor, float mag, float deg_rot) const;
    raw_dataset augment_affine(int afactor, std::tuple<float,float> mag_ctr_sft, int seed = 1) const;
    raw_dataset downsample2() const;
    raw_dataset multiply_class_trn(int class_idx, int factor) const;
    raw_dataset normalize_contrast() const;
    raw_dataset reorder_class_to_beginning(int class_idx) const;
    raw_dataset select_entries_of_class(int c) const;
    raw_dataset select_classes(std::vector<int> classes) const;
    raw_dataset select_subsample(const std::vector<int> &idx) const;
    raw_dataset select_trn() const;
    raw_dataset select_tst() const;
    raw_dataset set_class_to(int cls) const;
    raw_dataset shuffle(int seed = 1) const;
    raw_dataset shuffle_trn(int seed = 1) const;

    raw_dataset select_kth_fold(int folds_count, int fold) const;
    static raw_dataset recombine_folds(const std::vector<raw_dataset> &datasets);
};
void serialize(FILE *f, const raw_dataset::entry *ptr);
void serialize(FILE *f, const raw_dataset *ptr);
void unserialize(FILE *f, raw_dataset::entry *ptr);
void unserialize(FILE *f, raw_dataset *ptr);

/*************************************************************************
Output calibrator (calibrates just one scalar output)
*************************************************************************/
class calibrator
{
friend void serialize(FILE *f, const calibrator *ptr);
friend void unserialize(FILE *f, calibrator *ptr);
private:
    std::vector<float>      x, y;
public:
    calibrator();
    calibrator(const std::vector<float> &_x, const std::vector<float> &_y);

    bool  is_identity() const;
    float calibrate(float vx) const;
};
void serialize(FILE *f, const calibrator *ptr);
void unserialize(FILE *f, calibrator *ptr);

/*************************************************************************
Model predictions matrix
*************************************************************************/
class predictions
{
public:
    struct tfpn_rec
    {
        int true_pos_cnt, true_neg_cnt, false_pos_cnt, false_neg_cnt, total_cnt;
        float true_pos_rate, true_neg_rate, false_pos_rate, false_neg_rate;
    };

    int64_t             ntrn;
    int64_t             ntst;
    int64_t             nclasses;
    std::vector<int>    labels;
    fvector             fpredictions;
    int64_t             stride;

    predictions():ntrn(0),ntst(0),nclasses(0),stride(){}
    predictions(int64_t trn, int64_t tst, int64_t classes, const std::vector<int> &lbl);

    static predictions zero_baseline(int64_t trn, int64_t tst, int64_t classes);
    explicit predictions(const raw_dataset &ds);
    explicit predictions(std::string filename);

    void append(const predictions &pred);

    activityf_vector   kth_as_activity(int64_t k) const;
    std::vector<float> kth_as_fvector(int64_t k) const;
    void compute_errors(float *trn_rms, float *tst_rms, float *tst_cls, std::vector<int> *p_class_errors = NULL) const;
    float trn_rms() const;
    float tst_rms() const;
    float tst_cls() const;
    int   tst_false_positives(int class_idx) const;
    int   tst_false_negatives(int class_idx) const;
    tfpn_rec tst_class_detection_by_threshold(float threshold, int class_idx = 0) const;
    float predicted_prob(int img_idx, int class_idx) const;
    std::vector<int> predicted_classes() const;
    predictions select_subsample(const std::vector<int> items) const;

    std::vector<calibrator> recalibrate(int nbins);
    static predictions calibrate_onevsall(const predictions &p, const calibrator &c);
    static predictions calibrate(const predictions &p, const std::vector<calibrator> &c);
    static predictions average(const std::vector<predictions> &vec);
    static predictions onevsall(const predictions &all, int class_id);
    static predictions merge_onevsall(const std::vector<predictions> &vec);
    static predictions merge_onevsall_partial(const std::vector<predictions> &vec);

    void output_bad_cases(const raw_dataset &raw_ds, std::string dir) const;
    void print_calibration_curve(int class_idx) const;
};
void serialize(FILE *f, const predictions *ptr);
void unserialize(FILE *f, predictions *ptr);

/*************************************************************************
Convolutional predictions
*************************************************************************/
class conv_predictions
{
private:
    int64_t             img_width, img_height;
    int64_t             pred_width, pred_height;
    int64_t             shift_width, shift_height;
    int64_t             ds_factor;
    int64_t             nclasses;
    fvector             fpredictions;
public:

    conv_predictions():img_width(0),img_height(0),pred_width(0),pred_height(0),ds_factor(1),nclasses(0){}
    conv_predictions(int ih, int iw, int nc);
    conv_predictions(std::tuple<int,int> img_hw, int ds, std::tuple<int,int> pred_hw, std::tuple<int,int> shift_hw, int nc);
    //conv_predictions(int ih, int iw, int ds, int wh, int ww, int nc);

    int64_t get_img_width() const;
    int64_t get_img_height() const;
    int64_t get_pred_width() const;
    int64_t get_pred_height() const;

    // access to data
    float* get_ptr();
    const float* get_ptr() const;

    // get and set prediction using downsampled indexes
    float   get_pred(int pos0, int pos1, int class_idx = 0) const;
    void    set_pred(int pos0, int pos1, int class_idx, float pred);
    void    set_pred01(int pos_h, int pos_w, float pred);               // sets prediction for class #0 of 2-class problem, sets class #1 to 1-pred

    // get data using non-downsampled location
    float   get_original_pixel(int pos_h, int pos_w, int class_idx = 0) const;
    void    set_original_pixel01(int pos_h, int pos_w, float pred);

    // conversion between prediction and original indexes
    std::tuple<int,int> ds2ori(int pos_h, int pos_w) const;

    // merge with maximum
    void    merge_max(const conv_predictions &src);

    // upscaling
    conv_predictions upscale2k(int k) const;

    // debug functionality
    void    debug_output(const bitmap &orig_image, std::string filename);
    void    debug_output(const bitmap &orig_image, float mark_above, std::string filename);
};

/*************************************************************************
Channel, realm, entire dataset
*************************************************************************/
class channel
{
public:
    channel(){}
    channel(const std::string &d):desc(d){}

    std::string                  desc;
};
void serialize(FILE *f, const channel *ptr);
void unserialize(FILE *f, channel *ptr);

template<class element_t>
class realm
{
public:
    std::string name;
    int ch_offs;
    int grid_depth;
    int grid_stride0, grid_stride1;
    int grid_height, grid_width;
    std::vector<element_t> elements;

    // appends element to the current realm; the realm MUST be 1x1 grid
    void append_element(const element_t &c)
    {
        ASSERT(grid_width==1 && grid_height==1);
        grid_depth++;
        elements.push_back(c);
    }

    // appends plane to the current realm
    void append_plane(const element_t &c)
    {
        grid_depth++;
        for(int k=0; k<grid_height*grid_width; k++)
            elements.push_back(c);
    }

    // total channels count
    int get_nchannels() const
    { return grid_depth*grid_height*grid_width; }

    // checks whether realm is densely packed
    bool is_densely_packed() const
    { return grid_height==grid_stride0 && grid_width==grid_stride1; }

    // get channel index in the original dataset by plane/offs0/offs1
    int get_cidx(int plane_idx, int offs0, int offs1) const
    { return ch_offs+plane_idx*grid_stride0*grid_stride1+offs0*grid_stride1+offs1; }

    // informational queries
    void get_element_location(int cidx, int *pos0, int *pos1) const
    { int depth; get_element_info(cidx, &depth, pos0, pos1); }

    void get_element_info(int cidx, int *depth, int *pos0, int *pos1) const
    {
        ASSERT(cidx>=ch_offs);
        cidx-=ch_offs;
        *depth = cidx / (grid_stride0*grid_stride1);
        cidx   = cidx % (grid_stride0*grid_stride1);
        *pos0  = cidx/grid_stride1;
        *pos1  = cidx%grid_stride1;
        ASSERT(*pos0>=0 && *pos0<grid_height);
        ASSERT(*pos1>=0 && *pos1<grid_width);
    }

    std::vector<int> get_elements_at_location(int pos0, int pos1) const
    {
        std::vector<int> r;
        ASSERT(pos0>=0 && pos0<grid_height);
        ASSERT(pos1>=0 && pos1<grid_width);
        for(int k=0; k<grid_depth; k++)
            r.push_back(ch_offs+k*grid_stride0*grid_stride1+pos0*grid_stride1+pos1);
        return r;
    }

    // creation of new realms
    static realm create_grid(std::string name, int dim0, int dim1, int nfilters)
    {
        realm rr;
        rr.name = name;
        rr.ch_offs = 0;
        rr.grid_stride0 = dim0;
        rr.grid_stride1 = dim1;
        rr.grid_height  = dim0;
        rr.grid_width   = dim1;
        rr.grid_depth  = nfilters;
        rr.elements = std::vector<element_t>(dim0*dim1*nfilters);
        for(int k=0; k<nfilters; k++)
            for(int i=0; i<dim0; i++)
                for(int j=0; j<dim1; j++)
                {
                    element_t &cc = rr.elements[k*dim0*dim1+i*dim1+j];
                    char buf[256];
                    sprintf(buf, "c%02d[%02d,%02d]", k, i, j);
                    cc.desc = buf;
                }
        return rr;
    }

    realm rebase(int new_offs) const
    {
        ASSERT(is_densely_packed());
        realm r = *this;
        r.ch_offs = new_offs;
        return r;
    }
};

template<class element_t>
void serialize(FILE *f, const realm<element_t> *ptr)
{
    write_magic8(f, 23);
    serialize(f, &ptr->name);
    serialize(f, &ptr->ch_offs);
    serialize(f, &ptr->grid_depth);
    serialize(f, &ptr->grid_stride0);
    serialize(f, &ptr->grid_stride1);
    serialize(f, &ptr->grid_height);
    serialize(f, &ptr->grid_width);
    serialize(f, &ptr->elements);
    write_magic8(f, 101);
}

template<class element_t>
void unserialize(FILE *f, realm<element_t> *ptr)
{
    check_magic8(f, 23);
    unserialize(f, &ptr->name);
    unserialize(f, &ptr->ch_offs);
    unserialize(f, &ptr->grid_depth);
    unserialize(f, &ptr->grid_stride0);
    unserialize(f, &ptr->grid_stride1);
    unserialize(f, &ptr->grid_height);
    unserialize(f, &ptr->grid_width);
    unserialize(f, &ptr->elements);
    check_magic8(f, 101);
}

enum class storage_format { i8 = 1, f32 = 2 };
enum class padding_t      { none = 0, gray = 1};
enum class select_t       { none = 0, all = 1, larger = 2, smaller = 3, not_larger = 4, not_smaller = 5 };
enum class split_t        { entire = 0,  q4 = 1, q9 = 2,  h16 = 4, g64 = 8, p25 = 16,
                            q9h16 = 6, q9h16g64 = 14, q9h16p25 = 22, h16g64 = 12 };
enum class accumulate_t   { sum = 1, max = 2 };

class activities
{
private:
    storage_format                  format;
    activity8_vector                *pwrk_activities8;
    activityf_vector                *pwrk_activitiesf;
    activity8_vector                lcl_activities8;
    activityf_vector                lcl_activitiesf;
    int64_t                         stride;
public:
    int64_t                         ntrn;
    int64_t                         ntst;
    //int64_t                         nclasses;
    int64_t                         nchannels;
    std::vector<realm<channel> >    realms;

    activities():
        ntrn(0),ntst(0),nchannels(0),stride(0),format(storage_format::i8),
        pwrk_activities8(NULL),pwrk_activitiesf(NULL){}
    activities(int trn, int tst, storage_format fmt);
    explicit activities(std::string filename);

    const realm<channel>& realm_by_channel(int cidx) const;
    const realm<channel>& realm_by_name(std::string name) const;
    bool is_i8() const;
    bool is_f32() const;
    storage_format get_format() const { return format; }
    int8_t* get_i8(int64_t cidx = 0);
    float*  get_f32(int64_t cidx = 0);
    const int8_t* get_i8(int64_t cidx = 0) const;
    const float*  get_f32(int64_t cidx = 0) const;
    int64_t get_stride() const;
    float get_normalized(int64_t cidx, int64_t img_idx) const;
    int64_t get_memory_usage() const;

    // overwrites activity of the channel with user-specified vector
    void set_activity(int64_t cidx, const activity8_vector &a);
    void set_activity(int64_t cidx, const activityf_vector &a);

    // appends just one activity; dataset MUST have just one realm with 1x1 grid
    void append1(const activity8_vector &a, const std::string &desc);
    void append1(const activityf_vector &a, const std::string &desc);

    // appends just one plane; dataset MUST have just one realm with MxN grid
    void append_plane1(const activities &a, const std::string desc = "");
    void append_plane1(const activities &a, int downsample, const std::string desc = "");

    // appends all plane from realm in the source dataset;
    // destination must be (a) completely empty or (b) have exactly one realm with exactly same grid width/height
    void append_planes_from_realm(const activities &a, const std::string realm_name);

    // appends realm; dataset may already have some realms, it will work; newly added elements will be 0's.
    // returns index of the first channel in the realm (realm.ch_offs value)
    int append_realm(const realm<channel> &r);

    // extracts activities associated with realm into separate storage
    void extract_realm(std::string realm_name, activities &act_out) const;

    // merges entire dataset into this one
    void merge_in(const activities &merged, std::string name_prefix = "", std::string name_suffix = "");

    // merge in with gain applied (now supported only for i8 data)
    void merge_in_with_gain(const activities &merged, const int gain, std::string name_prefix = "", std::string name_suffix = "");

    // converts activities to f32 format
    void convert_to_f32(activities &out) const;
    void convert_to_f32();

    // splits into 6 quadrants
    void split_q6(int overlap = 0);

    // splits into 10 quadrants
    void split_q10(int max_overlap = 0);

    // split and sum over 4x4, 2x2 (plus central quadrant) and 1x1 grids
    void split_and_sum_h22(activities &act_out) const;

    // splits grids into areas, performs accumulation within areas
    void split_and_accumulate(split_t split, accumulate_t accumulate, activities &act_out) const;

    // inplace subdivision of all realms, new realms reuse common data stored in the activities object
    void subdivide(split_t split, std::map<std::string,std::string> &realms_map, int max_overlap = 0);
    void subdivide(split_t split, int max_overlap = 0);

    // select training set subsample
    void select_subsample(const std::vector<int> items, activities &act_out) const;

    // creates view into the activity
    void create_view(activities &act_out);
    void create_view_window(int plane_height, int plane_width,
                            int w_offs0, int w_offs1,
                            int w_height, int w_width,
                            activities &act_out);
    void select_view(select_t select, int window_height, int window_width, activities &act_out);

    // detaches view, CREATING FULL COPY of the original activity storage - INCLUDING ACTIVITIES NOT PRESENT IN THE VIEW
    void detach_view();

    // multiplies activity by 2
    void gain2(activities &out, std::string prefix = "") const;

    // multiplies activity by k
    void gaink(int k, activities &out) const;

    // multiplies activity by 2
    void gain2();

    // adds prefix/suffix to all realm names
    void add_prefix(std::string s);
    void add_suffix(std::string s);

    // reserves storage for additional channels
    void reserve_storage(int add_channels);

    // drops storage, leaving activities object in the invalid state
    void drop_storage();

    // works slowly, do not overuse it!
    void estimate_separability(const predictions &baseline, double lambdav, float *trn_rms, float *tst_rms, float *tst_cls, bool report = false);

    void visualize_item(int item_idx, std::string filename_prefix) const;

    // our friends!
    friend void serialize(FILE *f, const activities *ptr);
    friend void unserialize(FILE *f, activities *ptr);
};
void serialize(FILE *f, const activities *ptr);
void unserialize(FILE *f, activities *ptr);


/*************************************************************************
Convolutional activities
*************************************************************************/
class conv_activities
{
private:
    storage_format                  format;
    activityf_vector                *pwrk_activitiesf;
    activityf_vector                lcl_activitiesf;
    int64_t                         img_height, img_width;
    int64_t                         ndata;
public:
    class conv_realm
    {
    public:
        std::string name;
        int offs;
        int grid_depth;
        int grid_stride0, grid_stride1;
        int grid_dim0, grid_dim1;
        int window_dim0, window_dim1;
        int scale_idx;

        // informational
        int  get_ndata() const;
        bool is_densely_packed() const;
        int get_offs(int plane_idx, int offs0, int offs1) const;

        // create grid, either with 1x1 window or wh*ww one
        static conv_realm create_grid(const std::string &sname, int gh, int gw, int sidx, int nk);
        static conv_realm create_grid(const std::string &sname, int gh, int gw, int sidx, int wh, int ww, int nk);

        conv_realm rebase(int new_offs) const;
    };
    std::vector<conv_realm>    realms;

    conv_activities():
        img_height(0),img_width(0),ndata(0),format(storage_format::f32),pwrk_activitiesf(NULL){}
    conv_activities(int ih, int iw, storage_format ff):
        img_height(ih),img_width(iw),ndata(0),format(ff),pwrk_activitiesf(NULL){}

    // informational functions
    int get_img_width() const;
    int get_img_height() const;
    storage_format get_format() const;
    int64_t get_memory_usage() const;
    bool is_f32() const;
    float*  get_f32(int64_t offs = 0);
    const float*  get_f32(int64_t cidx = 0) const;
    const conv_realm& realm_by_channel(int cidx) const;
    const conv_realm& realm_by_name(const std::string &name) const;
    bool has_realm(const std::string &name) const;
    int get_grid_dim0(const std::vector<std::string> &names) const;   // checks that all realms have same grid_dim0 and returns it
    int get_grid_dim1(const std::vector<std::string> &names) const;   // checks that all realms have same grid_dim1 and returns it
    int get_window_dim0(const std::vector<std::string> &names) const; // checks that all realms have same window_dim0 and returns it
    int get_window_dim1(const std::vector<std::string> &names) const; // checks that all realms have same window_dim1 and returns it

    // appends all planes from realm in the source activities set;
    // destination must be (a) completely empty or (b) have exactly one realm with exactly same grid width/height
    void append_planes_from_realm(const conv_activities &a, const std::string realm_name);

    // appends all planes from the only realm in the source activities set;
    // source must have exactly one realm
    // destination must be (a) completely empty or (b) have exactly one realm with exactly same grid width/height as source
    void append_planes_from1(const conv_activities &a);

    // appends realm (rebasing it to the end of the present storage) and allocates new storage for this realm;
    // dataset may already have some realms, it will work; newly added elements will be 0's.
    // returns index of the first channel in the realm (realm.ch_offs value)
    int append_realm(const conv_realm &r);

    // merges entire dataset into this one
    void merge_in(const conv_activities &merged, std::string name_prefix = "", std::string name_suffix = "");

    // reserves storage for additional data
    void reserve_storage(int add_data);

    // adds prefix/suffix to all realm names
    void add_prefix(std::string s);
    void add_suffix(std::string s);

    // converts activities to f32 format
    void convert_to_f32(conv_activities &out) const;
    void convert_to_f32();

    // creating views into the activity
    void select_view(select_t select, int window_height, int window_width, conv_activities &act_out);

    // synchronize realm scales (set to the lowest one) and grid sizes, upsampling if needed;
    // requires dense packaging of the realm and window_dim=window_dim1=1
    conv_activities synchronize_scales() const;

#if 0
    activities(int trn, int tst, storage_format fmt);
    explicit activities(std::string filename);

    bool is_i8() const;
    storage_format get_format() const { return format; }
    int8_t* get_i8(int64_t cidx = 0);
    const int8_t* get_i8(int64_t cidx = 0) const;
    int64_t get_stride() const;
    float get_normalized(int64_t cidx, int64_t img_idx) const;

    // overwrites activity of the channel with user-specified vector
    void set_activity(int64_t cidx, const activity8_vector &a);
    void set_activity(int64_t cidx, const activityf_vector &a);

    // appends just one activity; dataset MUST have just one realm with 1x1 grid
    void append1(const activity8_vector &a, const std::string &desc);
    void append1(const activityf_vector &a, const std::string &desc);

    // appends just one plane; dataset MUST have just one realm with MxN grid
    void append_plane1(const activities &a, const std::string desc = "");
    void append_plane1(const activities &a, int downsample, const std::string desc = "");

    // extracts activities associated with realm into separate storage
    void extract_realm(std::string realm_name, activities &act_out) const;

    // merge in with gain applied (now supported only for i8 data)
    void merge_in_with_gain(const activities &merged, const int gain, std::string name_prefix = "", std::string name_suffix = "");

    // splits into 6 quadrants
    void split_q6(int overlap = 0);

    // splits into 10 quadrants
    void split_q10(int max_overlap = 0);

    // split and sum over 4x4, 2x2 (plus central quadrant) and 1x1 grids
    void split_and_sum_h22(activities &act_out) const;

    // splits grids into areas, performs accumulation within areas
    void split_and_accumulate(split_t split, accumulate_t accumulate, activities &act_out) const;

    // inplace subdivision of all realms, new realms reuse common data stored in the activities object
    void subdivide(split_t split, std::map<std::string,std::string> &realms_map, int max_overlap = 0);
    void subdivide(split_t split, int max_overlap = 0);

    // select training set subsample
    void select_subsample(const std::vector<int> items, activities &act_out) const;

    // creates view into the activity
    void create_view(activities &act_out);
    void create_view_window(int plane_height, int plane_width,
                            int w_offs0, int w_offs1,
                            int w_height, int w_width,
                            activities &act_out);

    // detaches view, CREATING FULL COPY of the original activity storage - INCLUDING ACTIVITIES NOT PRESENT IN THE VIEW
    void detach_view();

    // multiplies activity by 2
    void gain2(activities &out, std::string prefix = "") const;

    // multiplies activity by k
    void gaink(int k, activities &out) const;

    // multiplies activity by 2
    void gain2();

    // drops storage, leaving activities object in the invalid state
    void drop_storage();

    // works slowly, do not overuse it!
    void estimate_separability(const predictions &baseline, double lambdav, float *trn_rms, float *tst_rms, float *tst_cls, bool report = false);

#endif
};


/*************************************************************************
Seed detector, final detector, network layer
*************************************************************************/
enum class predict_t { prob = 0, logprob = 1, logit = 2 };

class linear_predictor
{
private:
    bool    is_logit;
    int64_t nchannels;
    int64_t nclasses;
    std::vector<float> coeffs;

    predictions _predict(const predictions &baseline, const std::vector<const activities*> v_act, predict_t pred) const;
public:
    linear_predictor();
    linear_predictor(const predictions &_baseline);

    // info
    int   get_nclasses()  const { return nclasses; }
    int   get_nchannels() const { return nchannels; }
    float get_coeff(int cidx, int cls_idx) const;
    float get_l1_factor() const;
    bool  get_logit_flag() const;

    // creates logit predictor, accepts (nchannels+1)*(nclasses-1) array
    static linear_predictor create_logit(const std::vector<float> &c, int nclasses, int nchannels);

    // coeffs must be array[(nchannels_new+1)*nclasses], with first nclasses
    // elements being coeffs before constant term
    void update_coeffs(const std::vector<float> &c, int nchannels_new);

    predictions predict(const predictions &baseline, const activities& act) const;
    predictions predict(const predictions &baseline, const std::vector<const activities*> v_act) const;
    predictions predict_log(const predictions &baseline, const std::vector<const activities*> v_act) const;
    predictions predict_logit(const predictions &baseline, const std::vector<const activities*> v_act) const;

    // returns convolutional predictions with img_width=pred_width, img_height=pred_height, downsample=1
    conv_predictions conv_predict1(const conv_activities &act, const calibrator &cal) const;

    friend void serialize(FILE *f, const linear_predictor *ptr);
    friend void unserialize(FILE *f, linear_predictor *ptr);
};
void serialize(FILE *f, const linear_predictor *ptr);
void unserialize(FILE *f, linear_predictor *ptr);

/*************************************************************************
Rating
*************************************************************************/
class linear_evaluator;
enum class rating_t   { cov2 = 1, deltar2 = 2, fastdr2 = 3, fastdr2_x5 = 4, fastdr2_x25 = 5 };
float compute_rating(rating_t r, float omega_rat, const activity8_vector &act, float cov2, float nrm2, const linear_evaluator &eval);
float compute_rating(rating_t r, float omega_rat, const activityf_vector &act, float cov2, float nrm2, const linear_evaluator &eval);
float compute_rating(rating_t r, float omega_rat, const activity8_vector &act, const linear_evaluator &eval);
float compute_rating(rating_t r, float omega_rat, const activityf_vector &act, const linear_evaluator &eval);

/*************************************************************************
Seed detector, final detector, network layer
*************************************************************************/
enum class and_type_t { AND_MIN=7301, AND_MUL=7302 };
enum class  or_type_t { OR_MAX =7401, OR_ADD =7402 };
enum class gain_t     { gain1 = 1, gain2 = 2, gain4 = 4, gain8 = 8 };
enum class weight_type { w100 = 100, w075 = 75, w050 = 50 };

void serialize(FILE *f, const and_type_t *ptr);
void serialize(FILE *f, const or_type_t  *ptr);
void unserialize(FILE *f, and_type_t *ptr);
void unserialize(FILE *f, or_type_t  *ptr);

//void act_f32_to_i8(int8_t *p_target, const float *p_source, int64_t n);
void act8_merge_max_simd(int8_t *p_target, const int8_t *p_merged, int64_t n);
void act8_merge_wmax_simd(int8_t *p_target, const int8_t *p_merged, weight_type w, int64_t n);
void act8_merge_gmax_simd(int8_t *p_target, const int8_t *p_merged, gain_t g, int64_t n);
void act8_merge_min_simd(int8_t *p_target, const int8_t *p_merged, int64_t n);

class seed_detector
{
public:
    and_type_t and_type;
    gain_t ggain;
    int channel_idx0, channel_idx1, channel_idx2, channel_idx3, channel_idx4;

    seed_detector():
        and_type(and_type_t::AND_MIN),ggain(gain_t::gain1),channel_idx0(-1),channel_idx1(-1),channel_idx2(-1),channel_idx3(-1),channel_idx4(-1){}
    seed_detector(int c0, gain_t g):
        and_type(and_type_t::AND_MIN),ggain(g),channel_idx0(c0),channel_idx1(-1),channel_idx2(-1),channel_idx3(-1),channel_idx4(-1){}
    seed_detector(int c0, int c1, gain_t g):
        and_type(and_type_t::AND_MIN),ggain(g),channel_idx0(c0),channel_idx1(c1),channel_idx2(-1),channel_idx3(-1),channel_idx4(-1){}
    seed_detector(int c0, int c1, int c2, gain_t g):
        and_type(and_type_t::AND_MIN),ggain(g),channel_idx0(c0),channel_idx1(c1),channel_idx2(c2),channel_idx3(-1),channel_idx4(-1){}
    seed_detector(int c0, int c1, int c2, int c3, gain_t g):
        and_type(and_type_t::AND_MIN),ggain(g),channel_idx0(c0),channel_idx1(c1),channel_idx2(c2),channel_idx3(c3),channel_idx4(-1){}
    seed_detector(int c0, int c1, int c2, int c3, int c4, gain_t g):
        and_type(and_type_t::AND_MIN),ggain(g),channel_idx0(c0),channel_idx1(c1),channel_idx2(c2),channel_idx3(c3),channel_idx4(c4){}

    seed_detector(int c0, and_type_t atype, gain_t g=gain_t::gain1):
        and_type(atype),ggain(g),channel_idx0(c0),channel_idx1(-1),channel_idx2(-1),channel_idx3(-1),channel_idx4(-1){}
    seed_detector(int c0, int c1, and_type_t atype, gain_t g=gain_t::gain1):
        and_type(atype),ggain(g),channel_idx0(c0),channel_idx1(c1),channel_idx2(-1),channel_idx3(-1),channel_idx4(-1){}
    seed_detector(int c0, int c1, int c2, and_type_t atype, gain_t g=gain_t::gain1):
        and_type(atype),ggain(g),channel_idx0(c0),channel_idx1(c1),channel_idx2(c2),channel_idx3(-1),channel_idx4(-1){}
    seed_detector(int c0, int c1, int c2, int c3, and_type_t atype, gain_t g=gain_t::gain1):
        and_type(atype),ggain(g),channel_idx0(c0),channel_idx1(c1),channel_idx2(c2),channel_idx3(c3),channel_idx4(-1){}
    seed_detector(int c0, int c1, int c2, int c3, int c4, and_type_t atype, gain_t g=gain_t::gain1):
        and_type(atype),ggain(g),channel_idx0(c0),channel_idx1(c1),channel_idx2(c2),channel_idx3(c3),channel_idx4(c4){}

    bool is_empty() const { return channel_idx0<0; }
    int  inputs_cnt() const
    {
        if( channel_idx0<0 )
            return 0;
        if( channel_idx1<0 )
            return 1;
        if( channel_idx2<0 )
            return 2;
        if( channel_idx3<0 )
            return 3;
        if( channel_idx4<0 )
            return 4;
        return 5;
    }
};

class detector
{
public:
    static const int WEIGHT_100;

    struct trf
    {
        int shift0, shift1;
        trf():shift0(0),shift1(0){}
        trf(int s0, int s1):shift0(s0),shift1(s1){}
    };

    struct conn
    {
        std::string realm_name;
        int realm_depth, realm_dim0, realm_dim1;
        int plane_idx, pos0, pos1;
        int weight;

        conn():weight(WEIGHT_100),plane_idx(-1),pos0(-1),pos1(-1),realm_depth(-1),realm_dim0(-1),realm_dim1(-1){}
        conn(int pp, int p0, int p1, int w, const realm<detector> &rr);
        conn(int cc, int w, const activities &ref_act);
    };

    detector();

    // creates detector from seed, uses ref_act to convert channel indexes into (plane,pos0,pos1) format
    detector(const seed_detector &seed, and_type_t andtype, or_type_t ortype, const activities &ref_act);

    std::string       desc;
    and_type_t        and_type;
    or_type_t         or_type;
    std::vector<trf>  transform;
    int               gain;
    std::vector<conn> input0;
    std::vector<conn> input1;
    std::vector<conn> input2;
    std::vector<conn> input3;
    std::vector<conn> input4;

    //
    bool is_empty() const { return input0.size()==0; }
    int  inputs_cnt() const;
    std::vector<std::string> get_realm_names() const;
    std::string get_realm_name() const;
    int         get_realm_dim0() const;
    int         get_realm_dim1() const;
    const std::vector<conn>& get_input(int i) const;
    std::vector<conn>& get_input(int i);

    // adds connection to specified input
    void add_conn(int input_idx, const conn &c);

    // fills ds.ntrn+ds.ntst elements of input0/input1/input2 by inputs of the detector;
    // if detector has less than 3 inputs, the rest is filled by 100's;
    // elements beyond [0,ntrn+ntst) range are not modified.
    // does not work for transform!={0}
    void generate_inputs5(const activities &ds, int8_t *p0, int8_t *p1, int8_t *p2, int8_t *p3, int8_t *p4) const;
    void generate_inputs5(const activities &ds, float  *p0, float  *p1, float  *p2, float  *p3, float  *p4) const;

    // fills ds.ntrn+ds.ntst elements of activity by detector values.
    // elements beyond [0,ntrn+ntst) range are not modified.
    // "SLOW" function which allocates internal temporaries - not recommended for heavy usage.
    void generate_activity_slow(const activities &ds, activity8_vector &activity) const;
    void generate_activity_slow(const activities &ds, activityf_vector &activity) const;
    conv_activities conv_generate_activity_slow(const conv_activities &act_src, std::string out_realm_name) const;
private:
    void generate_input(const activities &ds, const std::vector<conn> &input, trf channel_transform, int8_t *activity) const;
    void generate_input(const activities &ds, const std::vector<conn> &input, trf channel_transform, float  *activity) const;
    void conv_generate_input(const conv_activities &ds, int src_dim0, int src_dim1,
                             const std::vector<conn> &input, trf channel_transform, int window_dim0, int window_dim1,
                             float *activity) const;
};
void serialize(FILE *f, const detector::trf  *ptr);
void serialize(FILE *f, const detector::conn *ptr);
void serialize(FILE *f, const detector *ptr);
void unserialize(FILE *f, detector::trf  *ptr);
void unserialize(FILE *f, detector::conn *ptr);
void unserialize(FILE *f, detector *ptr);

/*************************************************************************
Various evaluators
*************************************************************************/
class linear_evaluator
{
public:
    // deep activity analysis results
    class daa
    {
    friend class linear_evaluator;
    private:
        // these fields are used for integrity checks (in order to avoid updates on invalidated analysis records)
        int64_t signature;
        const linear_evaluator *signed_by;

        void alloc(int trn, int tst, int baserows, int qcols, int nclasses);
    public:
        // original activity stored in fp32 format
        activityf_vector activity;

        // activity orthogonalized with current orthogonal base, NOT normalized
        activityf_vector qu;
        float qnrm;

        // dot products of the original activity with current orthogonal base
        std::vector<float> coeffs;

        // products of qu (unnormalized!) and residual rows
        std::vector<float> rqu;

        // sum of squared rq
        float deltar2;

        daa():signature(0),signed_by(NULL){}

        void invalidate();
    };

private:
    int64_t max_channels;
    int64_t max_rows;
    int64_t nrows;
    int64_t ncols;
    double lambdav;

    // baseline predictions (input to the model) and corrected output computed by the linear model
    predictions         baseline;
    predictions         output;

    // predictor used to construct output
    linear_predictor    predictor;

    // training set residuals, in single precision and fixed point formats; same stride is applied for both arrays
    activityf_vector        residuals_f32;
    activity8_vector        residuals_i8;
    activityf_vector        residuals_squared;
    int64_t                 residuals_stride;

    // test set activities, rowwise storage
    alglib::real_2d_array   test_set;

    // LQ factorization: orthogonal factor stored as single precision vector with stride
    fvector                 lq_ort32;
    int64_t                 lq_stride;

    // LQ factorization: triangular factor
    alglib::real_2d_array   lq_trf;

    // Q'*residuals, name is "qb" (not "qr") in order to avoid confusion with QR-decomposition
    alglib::real_2d_array qb;

    // error metrics
    float _trn_rms, _tst_rms, _tst_cls;

    // current signature
    int64_t signature;

    // member functions
    void update_predictor_output_errors();

    // updates residuals with new row; the row is assumed to be orthogonalized
    // to the already existing ones and NOT included in the working set yet;
    // if p_qr is specified, it contains products of q with residuals; if it
    // is not specified, products are computed internally.
    void update_residuals(const float *p_q, const float *p_qr = NULL);

    // runs analysis on instance of daa which is preallocated and preloaded with activity values
    void run_analysis(daa *p_result) const;

    void orthogonalize_row(float *p_row32, float *p_coeffs = NULL) const;
    void orthonormalize_row(float *p_row32, float *nrm, float *p_coeffs = NULL) const;
    void  estimate_trn_covariation(const float *activity_batch_f32, const int8_t *activity_batch_i8, int64_t batch_size, int64_t stride, int64_t *winner_idx, float *winner_cov) const;
    float true_trn_rms(const float *activity_f32, const int8_t *activity_i8) const;
    void append_activity(const float *activity_f32, const int8_t *activity_i8, int64_t activity_size);

    void compute_covariations(const float *activity, float *cov_vec) const;
    void compute_covariations(const int8_t *activity, float *cov_vec) const;
    float fastmetric(const float *cov_vec) const;
public:
    static std::atomic<int32_t> pcnt_truerms;
    static std::atomic<int32_t> pcnt_append;
    static std::atomic<int32_t> pcnt_cov;
    static std::atomic<int32_t> pcnt_da;
    static std::atomic<int32_t> pcnt_ar;

    linear_evaluator(const predictions &_pred, int64_t _max_channels, double _lambdav);
    ~linear_evaluator();

    float         get_lambdareg() const { return lambdav; }
    const int8_t* get_res8() const;
    const float*  get_res32() const;
    int64_t       get_res_stride() const;
    int           get_nclasses() const;
    int           get_ntrn() const;
    void analyze(const float  *activity, daa *p_result) const;
    void analyze(const int8_t *activity, daa *p_result) const;
    float deltar2(const float  *activity) const;
    float deltar2(const int8_t *activity) const;
    float true_trn_rms(const int8_t *activity) const;
    float true_trn_rms(const float  *activity) const;
    float estimate_trn_covariation(const int8_t *activity) const;
    float estimate_trn_covariation(const float *activity) const;
    void  estimate_trn_covariation(const int8_t *activity_batch, int64_t batch_size, int64_t stride, int64_t *winner_idx, float *winner_cov) const;
    void  estimate_trn_covariation(const float *activity_batch, int64_t batch_size, int64_t stride, int64_t *winner_idx, float *winner_cov) const;
    float trn_rms() const { return _trn_rms; }
    float tst_rms() const { return _tst_rms; }
    float tst_cls() const { return _tst_cls; }
    const predictions& outputs() const { return output; }
    const linear_predictor& model() const { return predictor; }
    bool is_full() const { return nrows==max_rows; }

    void append_activity(const activity8_vector &activity);
    void append_activity(const activityf_vector &activity);
    void append_and_refresh(const daa &append, const std::vector<daa*> &refresh);
};

class seed_evaluator
{
private:
    // original storage and linear evaluator
    const activities            &ds;
    const linear_evaluator      &lin_eval;

    // evaluation queue
    activity8_vector            eval_act8;
    activityf_vector            eval_act32;
    int64_t                     eval_stride;
    and_type_t                  eval_and_type;

    // current candidate
    //activity8_vector            cand_act8;

    // winner
    float                       best_cov2;
    seed_detector               best_seed;
    activityf_vector            best_actf;

    //
    void get_winner(float *cov, seed_detector *seed, activityf_vector *activity_f32, activity8_vector *activity_i8);
public:
    static std::atomic<int32_t> pcnt_seedeval;
    seed_evaluator(const activities &_ds, const linear_evaluator &_lin_eval);
    ~seed_evaluator();
    void start_evaluation(and_type_t atype);
    void evaluate_seed(const seed_detector &seed);
    void get_winner(float *cov, seed_detector *seed, activity8_vector *activity);
    void get_winner(float *cov, seed_detector *seed, activityf_vector *activity);
};

class generalizer_evaluator
{
friend void generalizer_process_queue_item_adapter(void *p, int idx);
private:
    // original storage and linear evaluator
    const activities            &ds;
    const linear_evaluator      &lin_eval;

    // evaluation queue
    class queue_entry
    {
    public:
        int merge_idx;
        float cov2;
        queue_entry():merge_idx(-1),cov2(0){}
        queue_entry(int mi):merge_idx(mi),cov2(0){}
    };
    std::vector<queue_entry>    eval_queue;
    and_type_t                  and_type;
    or_type_t                   or_type;
    int                         max_queue_size;
    activity8_vector            queue_cand_act_i8;  // thread-safe buffer for process_queue_item()
    activityf_vector            queue_cand_act_f32; // thread-safe buffer for process_queue_item()

    // detector info
    detector                    base_detector;
    activity8_vector            base_act_a_i8;
    activity8_vector            base_act_g_i8;
    activityf_vector            base_act_a_f32;
    activityf_vector            base_act_g_f32;
    int                         input_studied;

    // winner
    float                       best_cov;
    detector                    best_detector;
    activity8_vector            best_act_i8;
    activityf_vector            best_act_f32;

    // thread-safe temporaries for process_queue_item()
    //activity8_vector

    // temporaries
    activity8_vector            tmpa0_i8, tmpa1_i8, tmpa2_i8, tmpa3_i8, tmpa4_i8;
    activityf_vector            tmpa0_f32, tmpa1_f32, tmpa2_f32, tmpa3_f32, tmpa4_f32;

    //
    void process_queue();
    void process_queue_item(int item_idx);
public:
    static std::atomic<int32_t> pcnt_gen;
    generalizer_evaluator(const activities &_ds, const linear_evaluator &_lin_eval);
    void start_evaluation(const detector &d, int input_idx, and_type_t andtype, or_type_t ortype);
    void evaluate_generalization(int channel_idx);
    void get_winner(float *cov, detector *winner, activity8_vector *activity);
    void get_winner(float *cov, detector *winner, activityf_vector *activity);
};

float sigma0(float x);
int get_scale_idx_safe(int win_dim, int base_dim);

std::vector<std::string> explode(const std::string &s, char sep);
std::string trim(const std::string &s);
std::vector<std::string> trim(const std::vector<std::string> &s);

bool file_exists(std::string filename);
std::string file_get_contents(std::string filename);
void file_put_contents(std::string filename, std::string data);
void file_append_contents(std::string filename, std::string data);

template<class T>
void serialize_as_int64(FILE *f, const T *p)
{
    int64_t v = (int64_t)(*p);
    serialize(f, &v);
}

template<class T>
void unserialize_as_int64(FILE *f, T *p)
{
    int64_t v;
    unserialize(f, &v);
    *p = (T)v;
}

template<class string_t, class T>
void serialize_to(const T *ptr, string_t filename)
{
    FILE *f = fopen(filename.c_str(), "wb");
    ASSERT(f!=NULL);
    serialize(f, ptr);
    fclose(f);
}

template<class string_t, class T>
void unserialize_from(string_t filename, T *ptr)
{
    FILE *f = fopen(filename.c_str(), "rb");
    ASSERT(f!=NULL);
    unserialize(f, ptr);
    fclose(f);
}

template<class T>
void serialize(FILE *f, const std::vector<T> *vec)
{
    serialize64(f, vec->size());
    for(int i=0; i<vec->size(); i++)
        serialize(f, &((*vec)[i]));
}

template<class T>
void serialize(FILE *f, const std::vector<T, aligned_allocator<T> > *vec)
{
    serialize64(f, vec->size());
    for(int i=0; i<vec->size(); i++)
        serialize(f, &((*vec)[i]));
}

template<class T>
void unserialize(FILE *f, std::vector<T> *vec)
{
    int64_t vs = unserialize64(f);
    vec->resize(vs);
    for(int i=0; i<vs; i++)
        unserialize(f, &((*vec)[i]));
}

template<class T>
void unserialize(FILE *f, std::vector<T, aligned_allocator<T> > *vec)
{
    int64_t vs = unserialize64(f);
    vec->resize(vs);
    for(int i=0; i<vs; i++)
        unserialize(f, &((*vec)[i]));
}

template<> void serialize<int8_t>(FILE *f, const std::vector<int8_t,aligned_allocator<int8_t> > *vec);
template<> void serialize<float>(FILE *f, const std::vector<float,aligned_allocator<float> > *vec);
template<> void serialize<float>(FILE *f, const std::vector<float> *vec);
template<> void serialize<int>(FILE *f, const std::vector<int> *vec);

template<> void unserialize<int8_t>(FILE *f, std::vector<int8_t, aligned_allocator<int8_t> > *vec);
template<> void unserialize<float>(FILE *f, std::vector<float, aligned_allocator<float> > *vec);
template<> void unserialize<float>(FILE *f, std::vector<float> *vec);
template<> void unserialize<int>(FILE *f, std::vector<int> *vec);

static void memset_vector(activityf_vector &vec, int v)
{
    memset(&vec[0], v, vec.size()*sizeof(vec[0]));
}

template<class elem_t>
void memset_vector(std::vector<elem_t> &vec, int v)
{
    memset(&vec[0], v, vec.size()*sizeof(vec[0]));
}

extern thread_local int _parallel_flag;
void parallelism_activate();
int get_cores_count();

void parallel_invoke(void* (*func)(void*), void *param);

template<class f_t>
class _parallel_for_env
{
public:
    std::atomic<int32_t> counter;
    int32_t range;
    void *param;
    f_t func;
    _parallel_for_env(f_t f, void *p, int r):func(f), counter(0), range(r), param(p){}
};

template<class f_t>
void* _parallel_for_adapter(void *p)
{
    _parallel_for_env<f_t> &env = *((_parallel_for_env<f_t>*)p);
    for(int32_t idx=env.counter.fetch_add(1);
                idx<env.range;
                idx=env.counter.fetch_add(1))
        env.func(env.param, idx);
    return NULL;
}

template<class f_t>
void parallel_for(f_t func, void *param, int range)
{
    ASSERT(range>=0);

    //
    // Quick exit
    //
    // NOTE: checking _parallel_flag also prevents us from spawning new workers
    //       from already running worker threads.
    //
    if( range==0 )
        return;
    if( _parallel_flag==0 || range<=1 )
    {
        for(int idx=0; idx<range; idx++)
            func(param, idx);
        return;
    }

    //
    // Try to perform parallel processing
    //
#if !defined(SNET_SERIAL)
    ASSERT(get_cores_count()>=1);
    int nthreads = imin(get_cores_count(),range);
    _parallel_for_env<f_t> env(func, param, range);
    std::vector<pthread_t> threads(nthreads);
    for(int k=0; k<nthreads; k++)
    {
        int iret1 = pthread_create(&threads[k], NULL, _parallel_for_adapter<f_t>, &env);
        ASSERT(iret1==0);
    }
    for(int k=0; k<nthreads; k++)
        pthread_join(threads[k], NULL);
#else
    for(int idx=0; idx<range; idx++)
        func(param, idx);
#endif
}

}
#endif
