/*************************************************************************
StrongNet/Contour Engine training framework.
Copyright (c) 2020 Sergey Bochkanov (ALGLIB project).

>>> SOURCE LICENSE >>>
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation (www.fsf.org); either version 2 of the 
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

A copy of the GNU General Public License is available at
http://www.fsf.org/licensing/licenses
>>> END OF LICENSE >>>
*************************************************************************/
#include "layer1.h"
#include <algorithm>
#include <numeric>
#include <atomic>
#include <random>
#include <map>
#include "3rdparty/alglib/dataanalysis.h"

namespace strongnet
{

class downsampler
{
private:
    int depth, dim0, dim1, downsample_max;
    std::vector<std::vector<float>> ds_vec;
public:
    downsampler(int dpth, int d0, int d1, int ds_max);

    void apply_downsampling(const std::vector<float> &flt_in);
    void output_kth_layer(int64_t layer_idx, activities &act_out, int64_t img_idx, int64_t start_channel) const;
    void output_kth_layer(int64_t layer_idx, activities &act_out, int64_t img_idx, const realm<channel> &rr_out, bool append_not) const;
    void output_kth_layer(int64_t layer_idx, conv_activities &c_out, const conv_activities::conv_realm &rr_out) const;
};

downsampler::downsampler(int dpth, int d0, int d1, int ds_max)
{
    depth = dpth;
    dim0 = d0;
    dim1 = d1;
    downsample_max = ds_max;

    // integrity checks
    ASSERT(downsample_max>=1);
    for(int i=downsample_max; i>=1; i/=2)
        ASSERT(i%2==0 || i==1);
    for(int k=1; k<=downsample_max; k*=2)
        ASSERT(dim0%k==0 && dim1%k==0);

    // allocate
    for(int k=1; k<=downsample_max; k*=2)
        ds_vec.push_back(std::vector<float>());
}

void downsampler::apply_downsampling(const std::vector<float> &flt_in)
{
    for(int ds_k=0, ds=1; ds<=downsample_max; ds*=2,ds_k++)
    {
        std::vector<float> &flt_cur = ds_vec[ds_k];
        const int cur_dim0 = dim0/ds;
        const int cur_dim1 = dim1/ds;

        // downsample using previous downsampling layer
        if( ds_k>=1 )
        {
            const std::vector<float> &flt_prev = ds_vec[ds_k-1];
            const int prev_dim0 = dim0/(ds/2);
            const int prev_dim1 = dim1/(ds/2);
            ASSERT(flt_prev.size()==depth*prev_dim0*prev_dim1);
            flt_cur.clear();
            flt_cur.resize(depth*cur_dim0*cur_dim1, 0.0f);
            for(int64_t channel_idx=0; channel_idx<depth; channel_idx++)
                for(int64_t pos0=0; pos0<prev_dim0; pos0++)
                    for(int64_t pos1=0; pos1<prev_dim1; pos1++)
                    {
                        int64_t src_pos = channel_idx*prev_dim0*prev_dim1 +    pos0*prev_dim1 + pos1;
                        int64_t dst_pos = channel_idx*cur_dim0*cur_dim1   + (pos0/2)*cur_dim1 + (pos1/2);
                        flt_cur[dst_pos] = fmaxf(flt_cur[dst_pos], flt_prev[src_pos]);
                    }
        }
        else
        {
            flt_cur.clear();
            flt_cur.resize(depth*cur_dim0*cur_dim1, 0.0f);
            ASSERT(flt_cur.size()==flt_in.size());
            memcpy(&flt_cur[0], &flt_in[0], flt_cur.size()*sizeof(flt_cur[0]));
        }
        ASSERT(flt_cur.size()==depth*cur_dim0*cur_dim1);
    }
}

void downsampler::output_kth_layer(int64_t layer_idx, activities &act_out, int64_t img_idx, int64_t start_channel) const
{
    ASSERT(layer_idx>=0 && layer_idx<ds_vec.size());
    const std::vector<float> &flt_cur = ds_vec[layer_idx];
    const int64_t vsize = flt_cur.size();
    const int64_t out_stride = act_out.get_stride();
    int8_t      *p_dst = act_out.get_i8()+img_idx+start_channel*out_stride;
    const float *p_src = &flt_cur[0];
    ASSERT(start_channel+vsize<=act_out.nchannels);
    for(int64_t k=0; k<vsize; k++)
        p_dst[k*out_stride] = posflt_to_int8(p_src[k]);
}

void downsampler::output_kth_layer(int64_t layer_idx, activities &act_out, int64_t img_idx, const realm<channel> &rr_out, bool append_not) const
{
    ASSERT(layer_idx>=0 && layer_idx<ds_vec.size());
    const std::vector<float> &flt_cur = ds_vec[layer_idx];
    const int64_t cur_dim0 = dim0/(1<<layer_idx), cur_dim1 = dim1/(1<<layer_idx), out_depth = depth+(append_not ? 1 : 0);
    const int64_t vsize = flt_cur.size();
    const int64_t out_stride = act_out.get_stride();
    ASSERT(vsize==cur_dim0*cur_dim1*depth);
    ASSERT(rr_out.grid_height==cur_dim0 && rr_out.grid_width==cur_dim1 && rr_out.grid_depth==out_depth);
    const float *p_src = &flt_cur[0];

    //
    // Output channels
    //
    ASSERT(rr_out.ch_offs+vsize<=act_out.nchannels);
    if( act_out.is_i8() )
    {
        int8_t *p_dst_prim = act_out.get_i8()+img_idx+rr_out.ch_offs*out_stride;
        for(int64_t k=0; k<vsize; k++)
            p_dst_prim[k*out_stride] = posflt_to_int8(p_src[k]);
    }
    else if( act_out.is_f32() )
    {
        float *p_dst_prim = act_out.get_f32()+img_idx+rr_out.ch_offs*out_stride;
        for(int64_t k=0; k<vsize; k++)
            p_dst_prim[k*out_stride] = p_src[k];
    }
    else
        ASSERT(0);

    //
    // Output not-channel
    //
    if( append_not )
    {
        ASSERT(0);
        /*std::vector<float> vmx(cur_dim0*cur_dim1,0.0f);
        for(int64_t k=0; k<vsize; k++)
            vmx[k%(cur_dim0*cur_dim1)] = fmaxf(vmx[k%(cur_dim0*cur_dim1)], p_src[k]);
        int8_t *p_dst_not = act_out.get_i8()+img_idx+(rr_out.ch_offs+vsize)*out_stride;
        for(int64_t k=0; k<cur_dim0*cur_dim1; k++)
            p_dst_not[k*out_stride] = posflt_to_int8(1.0f-vmx[k]);*/
    }
}

void downsampler::output_kth_layer(int64_t layer_idx, conv_activities &c_out, const conv_activities::conv_realm &rr_out) const
{
    ASSERT(layer_idx>=0 && layer_idx<ds_vec.size());
    const std::vector<float> &flt_cur = ds_vec[layer_idx];
    const int64_t cur_dim0 = dim0/(1<<layer_idx), cur_dim1 = dim1/(1<<layer_idx), out_depth = depth;
    const int64_t vsize = flt_cur.size();
    ASSERT(vsize==cur_dim0*cur_dim1*depth);
    ASSERT(rr_out.grid_dim0==cur_dim0 && rr_out.grid_dim1==cur_dim1 && rr_out.grid_depth==out_depth);
    const float *p_src = &flt_cur[0];

    //
    // Output channels
    //
    if( c_out.is_f32() )
    {
        float *p_dst_prim = c_out.get_f32(rr_out.get_offs(0,0,0));
        for(int64_t k=0; k<vsize; k++)
            p_dst_prim[k] = p_src[k];
    }
    else
        ASSERT(0);
}


static float relu(float x)
{
    return x>=0 ? x : 0;
}

static std::vector<float> rotate90(std::vector<float> &src, int size)
{
    std::vector<float> r;
    ASSERT(src.size()==size*size);
    r.resize(size*size);
    for(int i=0; i<size; i++)
        for(int j=0; j<size; j++)
            r[(size-1-j)*size+i] = src[i*size+j];
    return r;
}

template<class pixel_t>
void rotate90(pixel_t *p_patch, int size, int nchannels)
{
    ASSERT(size>=1 && size<=32 && nchannels>=1 && nchannels<=3);

    pixel_t tmp[size*size*nchannels];
    for(int i=0; i<size; i++)
        for(int j=0; j<size; j++)
            for(int k=0; k<nchannels; k++)
                tmp[((size-1-j)*size+i)*nchannels+k] = p_patch[(i*size+j)*nchannels+k];
    for(int i=0; i<size*size*nchannels; i++)
        p_patch[i] = tmp[i];
}

static void normalize_sums(filter_family &f)
{
    int size = f.size;
    for(int i=0; i<f.cnt; i++)
    {
        float *p = &f.coeffs[i*size*size];
        float sp = 1.0e-20f, sn = 1.0e-20f;
        for(int k=0; k<size*size; k++)
            if( p[k]>=0 )
                sp += p[k];
            else
                sn -= p[k];
        //ASSERT(sp>0 && sn>0);
        for(int k=0; k<size*size; k++)
            p[k] /= p[k]>=0 ? sp : sn;
    }
}

static void soft_normalization(std::vector<float> &data, int plane_size)
{
    const float CLIP = 0.05f;
    const float Q = 0.5f;
    const float QV = 4.0;

    //
    ASSERT(data.size()%plane_size==0);

    // determine global maximum and location-specific maximum values
    float mx = 1.0e-20f;
    std::vector<float> vmx(plane_size, 0);
    for(int64_t k=0; k<data.size(); k++)
    {
        vmx[k%plane_size] = fmaxf(vmx[k%plane_size], data[k]);
        mx = fmaxf(mx, vmx[k%plane_size]);
    }

    // sample activities above CLIP*mx
    std::vector<float> sample;
    for(float v: vmx)
        if( v>CLIP*mx )
            sample.push_back(v);
    if( sample.size()==0 )
    {
        for(int64_t k=0; k<data.size(); k++)
            data[k] /= mx;
        return;
    }
    int qpos = Q*sample.size();
    //std::sort(sample.begin(), sample.end());
    std::nth_element(sample.begin(), sample.begin()+qpos, sample.end());
    ASSERT(sample[qpos]>0.0f);
    float scale_by = QV/sample[qpos];
    for(int64_t k=0; k<data.size(); k++)
    {
        float v = data[k]*scale_by;
        data[k] = v/(1.0f+v);
    }
}

static void white_balance(std::vector<float> &vr, std::vector<float> &vg, std::vector<float> &vb)
{
    const float MAX_DIFFERENCE = 0.75f;

    ASSERT(vr.size()==vg.size() && vr.size()==vb.size());
    int n = vr.size();

    // estimate maximum values
    float mx_r = 0.0f, mx_g = 0.0f, mx_b = 0.0f;
    for(int64_t i=0; i<n; i++)
    {
        mx_r = fmaxf(mx_r, vr[i]);
        mx_g = fmaxf(mx_g, vg[i]);
        mx_b = fmaxf(mx_b, vb[i]);
    }

    // prepare scaling
    float mx_all = fmaxf(mx_r,fmaxf(mx_g,mx_b));
    if( mx_all==0 )
        return;
    float s_r = 1.0f/fmaxf(mx_r, mx_all*MAX_DIFFERENCE);
    float s_g = 1.0f/fmaxf(mx_g, mx_all*MAX_DIFFERENCE);
    float s_b = 1.0f/fmaxf(mx_b, mx_all*MAX_DIFFERENCE);
    for(int64_t i=0; i<n; i++)
    {
        vr[i] *= s_r;
        vg[i] *= s_g;
        vb[i] *= s_b;
    }
}


filter_family filter_family::normalize_sum2nz() const
{
    const int size2 = size*size;
    filter_family r = *this;
    ASSERT(r.coeffs.size()==cnt*size2);
    for(int k=0; k<r.cnt; k++)
    {
        float *p = &r.coeffs[k*size2];
        float s=0, nnz=0;
        for(int i=0; i<size2; i++)
            if( p[i]!=0 )
            {
                s   += fabsf(p[i]);
                nnz += 1.0f;
            }
        if( nnz==0 )
            continue;
        float v = (2.0f/nnz)/s;
        for(int i=0; i<size2; i++)
            p[i] *= v;
    }
    return r;
}

void serialize(FILE *f, const filter_family *ptr)
{
    write_magic8(f, 71);
    serialize(f, &ptr->size);
    serialize(f, &ptr->cnt);
    serialize(f, &ptr->coeffs);
    write_magic8(f, 81);
}

void unserialize(FILE *f, filter_family *ptr)
{
    check_magic8(f, 71);
    unserialize(f, &ptr->size);
    unserialize(f, &ptr->cnt);
    unserialize(f, &ptr->coeffs);
    check_magic8(f, 81);
}

const filter_family& filter_set::find_family(int chsize) const
{
    for(const rule &rr: rules)
        if( rr.chsize0<=chsize && chsize<rr.chsize1 )
            return rr.filters;
    printf("No rule for size=%d\n", chsize);
    fflush(stdout);
    ASSERT(0);
    abort();
}

void serialize(FILE *f, const filter_set *ptr)
{
    write_magic8(f, 11);
    serialize(f, &ptr->rules);
    write_magic8(f, 19);
}

void serialize(FILE *f, const filter_set::rule *ptr)
{
    serialize(f, &ptr->chsize0);
    serialize(f, &ptr->chsize1);
    serialize(f, &ptr->filters);
}

void unserialize(FILE *f, filter_set *ptr)
{
    check_magic8(f, 11);
    unserialize(f, &ptr->rules);
    check_magic8(f, 19);
}

void unserialize(FILE *f, filter_set::rule *ptr)
{
    unserialize(f, &ptr->chsize0);
    unserialize(f, &ptr->chsize1);
    unserialize(f, &ptr->filters);
}

net_v1layer::v1_realm::v1_realm()
{
    realm_type = realm_t::none;
    flags = v1_flags_t::compete;
    scale_idx = 0;
    projected_size = 0;
    downsample_min = downsample_max = 1;
    kmeans_wht_stab = 0.01f;
    kmeans_rad_param = 0.3f;
    kmeans_hardness = 0.0f;

    lxg_alpha_param = lxg_beta_param = clx_alpha_param = 0.0f;
    filter_gain = 0.0f;
    kmeans_flt7 = 0.0f;
    kmeans_int0 = kmeans_int1 = kmeans_int2 = kmeans_int3 = kmeans_int4 = kmeans_int5 = kmeans_int6 = kmeans_int7 = 0;
}

bool net_v1layer::v1_realm::quick_comparison(const net_v1layer::v1_realm &rhs) const
{
    if( realm_type!=rhs.realm_type )
        return false;
    if( realm_name!=rhs.realm_name )
        return false;
    if( flags!=rhs.flags )
        return false;
    if( projected_size!=rhs.projected_size )
        return false;
    if( downsample_min!=rhs.downsample_min )
        return false;
    if( downsample_max!=rhs.downsample_max )
        return false;
    return true;
}

void net_v1layer::v1_realm::apply(const raw_dataset &ds_in, activities &act_out) const
{
    net_v1layer tmp;
    tmp.append_realm(*this);
    tmp.apply(ds_in, act_out);
}

void net_v1layer::v1_realm::apply_append(const raw_dataset &ds_in, activities &act_out) const
{
    net_v1layer tmp;
    tmp.append_realm(*this);
    tmp.apply_append(ds_in, act_out);
}

void net_v1layer::v1_realm::debug_output(std::string path) const
{
    ASSERT(filters.rules.size()==1);
    const filter_set::rule &rule = filters.rules[0];
    int nfilters = rule.filters.cnt;
    int patch_size = rule.filters.size;

    std::vector<bitmap> bitmaps;
    for(int k=0; k<nfilters; k++)
    {
        const int scale = 128;
        bitmap img;
        img.resize(patch_size, patch_size);
        for(int i=0; i<patch_size; i++)
            for(int j=0; j<patch_size; j++)
            {
                if( realm_type==net_v1layer::realm_t::kmeans_lgt ||
                    realm_type==net_v1layer::realm_t::kmeans_lxg ||
                    realm_type==net_v1layer::realm_t::kmeans_gry )
                {
                    ASSERT(rule.filters.coeffs.size()==rule.filters.cnt*rule.filters.size*rule.filters.size);
                    float mx = 1.0e-3f;
                    for(int t=0; t<patch_size*patch_size; t++)
                        mx = fmaxf(mx, fabsf(rule.filters.coeffs[k*patch_size*patch_size+t]));
                    int v = rule.filters.coeffs[k*patch_size*patch_size+i*patch_size+j]/mx*scale+128;
                    img.draw_rgb(i, j, v, v, v);
                    continue;
                }
                if( realm_type==net_v1layer::realm_t::kmeans_clx1 ||
                    realm_type==net_v1layer::realm_t::kmeans_clr ||
                    realm_type==net_v1layer::realm_t::kmeans_clg ||
                    realm_type==net_v1layer::realm_t::kmeans_clu )
                {
                    ASSERT(rule.filters.coeffs.size()==3*rule.filters.cnt*rule.filters.size*rule.filters.size);
                    float mx = 1.0e-3f;
                    for(int t=0; t<3*patch_size*patch_size; t++)
                        mx = fmaxf(mx, fabsf(rule.filters.coeffs[k*3*patch_size*patch_size+t]));
                    img.draw_rgb(i, j,
                                 rule.filters.coeffs[3*(k*patch_size*patch_size+i*patch_size+j)+0]/mx*scale+128,
                                 rule.filters.coeffs[3*(k*patch_size*patch_size+i*patch_size+j)+1]/mx*scale+128,
                                 rule.filters.coeffs[3*(k*patch_size*patch_size+i*patch_size+j)+2]/mx*scale+128);
                    continue;
                }
                ASSERT(0);
            }
        char out_name[1024];
        sprintf(out_name, "%s/%s-c%03d.png", path.c_str(), realm_name.c_str(), k);
        //img.to_png(out_name);
        bitmaps.push_back(img);
    }
    bitmap grid = bitmap::combine_grid_w(bitmaps, 5, 1, 0x808080);
    char out_name[1024];
    sprintf(out_name, "%s/%s-all.png", path.c_str(), realm_name.c_str());
    grid.to_png(out_name);
}

net_v1layer::net_v1layer(std::string filename)
{
    unserialize_from(filename, this);
}

bool net_v1layer::operator==(const net_v1layer &rhs) const
{
    if( realms.size()!=rhs.realms.size() )
        return false;
    for(int k=0; k<realms.size(); k++)
        if( !realms[k].quick_comparison(rhs.realms[k]) )
            return false;
    return true;
}

int  net_v1layer::get_img_width() const
{
    ASSERT(realms.size()>=1);
    ASSERT(realms[0].filters.rules.size()==1);
    ASSERT(realms[0].filters.rules[0].chsize0==realms[0].filters.rules[0].chsize1-1);
    int out_size = realms[0].filters.rules[0].chsize0;
    for(const v1_realm &rr: realms)
    {
        ASSERT(rr.filters.rules.size()==1);
        ASSERT(rr.filters.rules[0].chsize0==rr.filters.rules[0].chsize1-1);
        out_size = imax(out_size,rr.filters.rules[0].chsize0);
    }
    return out_size;
}

int net_v1layer::get_img_height() const
{
    return get_img_width();
}

int net_v1layer::get_ds_lo(select_t select, int select_height, int select_width) const
{
    ASSERT(select_height==select_width);
    ASSERT(select==select_t::all || select==select_t::not_larger);
    ASSERT(realms.size()>=1);
    int out_ds = 0;
    for(const v1_realm &rr: realms)
        for(int current_ds=rr.downsample_min; current_ds<=rr.downsample_max; current_ds*=2)
        {
            // filter
            bool passed = false;
            if( select==select_t::all )
                passed = true;
            if( select==select_t::not_larger && rr.projected_size/current_ds<=select_height )
                passed = true;

            // output
            if( passed )
            {
                int effective_ds = (1<<rr.scale_idx)*current_ds;
                out_ds = out_ds>0 ? imin(out_ds, effective_ds) : effective_ds;
            }
        }
    return out_ds>0 ? out_ds : 1;
}

int net_v1layer::get_ds_hi(select_t select, int select_height, int select_width) const
{
    ASSERT(select_height==select_width);
    ASSERT(select==select_t::all || select==select_t::not_larger);
    ASSERT(realms.size()>=1);
    int out_ds = 0;
    for(const v1_realm &rr: realms)
        for(int current_ds=rr.downsample_min; current_ds<=rr.downsample_max; current_ds*=2)
        {
            // filter
            bool passed = false;
            if( select==select_t::all )
                passed = true;
            if( select==select_t::not_larger && rr.projected_size/current_ds<=select_height )
                passed = true;

            // output
            if( passed )
            {
                int effective_ds = (1<<rr.scale_idx)*current_ds;
                out_ds = out_ds>0 ? imax(out_ds, effective_ds) : effective_ds;
            }
        }
    return out_ds>0 ? out_ds : 1;
}

int  net_v1layer::get_max_height(select_t select, int select_height, int select_width) const
{
    ASSERT(select_height==select_width);
    ASSERT(select==select_t::all || select==select_t::not_larger);
    ASSERT(realms.size()>=1);
    int out_height = 0;
    for(const v1_realm &rr: realms)
        for(int current_ds=rr.downsample_min; current_ds<=rr.downsample_max; current_ds*=2)
        {
            // filter
            bool passed = false;
            if( select==select_t::all )
                passed = true;
            if( select==select_t::not_larger && rr.projected_size/current_ds<=select_height )
                passed = true;

            // output
            if( passed )
            {
                int cur_height = rr.projected_size/current_ds;
                out_height = out_height>0 ? imax(out_height, cur_height) : cur_height;
            }
        }
    return out_height>0 ? out_height : 1;
}

int  net_v1layer::get_max_width(select_t select,  int select_height, int select_width) const
{
    return get_max_height(select, select_height, select_width);
}

void net_v1layer::append_realm(const v1_realm &rr)
{
    ASSERT(rr.projected_size>=1);
    realms.push_back(rr);
}

void net_v1layer::append_layer(const net_v1layer &layer)
{
    for(const v1_realm &rr: layer.realms)
        append_realm(rr);
}

void net_v1layer::apply(const raw_dataset &ds_in, activities &act_out, bool report, storage_format fmt) const
{
    act_out = activities(ds_in.ntrn, ds_in.ntst, fmt);
    apply_append(ds_in, act_out, report, fmt);
}

void net_v1layer::apply_append(const raw_dataset &ds_raw, activities &act_out, bool report, storage_format fmt) const
{
    int t0 = dtickcount();

    //
    // Prepare downscaled versions of dataset
    //
    int max_scale_idx = 0;
    for(const v1_realm &rr: realms)
        max_scale_idx = imax(max_scale_idx, rr.scale_idx);
    std::vector<raw_dataset> ds_downscaled(max_scale_idx);
    if( max_scale_idx>0 )
    {
        ds_downscaled[0] = ds_raw.downsample2();
        for(int k=1; k<ds_downscaled.size(); k++)
            ds_downscaled[k] = ds_downscaled[k-1].downsample2();
    }

    //
    // Prepare activity map
    //
    std::map<std::string,activities*> act_map;

    //
    // Generate activities
    //
    for(const v1_realm &rr: realms)
    {
        ASSERT(rr.realm_type!=realm_t::none);
        ASSERT(rr.downsample_min>=1 && rr.downsample_min<=rr.downsample_max);
        for(int i=rr.downsample_min; i>=1; i/=2)
            ASSERT(i%2==0 || i==1);
        for(int i=rr.downsample_max; i>=1; i/=2)
            ASSERT(i%2==0 || i==1);

        // select dataset to work with
        const raw_dataset &ds_wrk = rr.scale_idx==0 ? ds_raw : ds_downscaled[rr.scale_idx-1];

        // determine number of downsample layers
        int downsample_layers_lo = 0;
        for(int i=rr.downsample_min; i>=1; i/=2)
            downsample_layers_lo++;
        int downsample_layers_hi = 0;
        for(int i=rr.downsample_max; i>=1; i/=2)
            downsample_layers_hi++;

        //
        // Initial processing
        //
        if( report )
        {
            printf("[V1] realm %s... ", rr.realm_name.c_str());
            fflush(stdout);
        }
        std::vector<activities> tmp_va(downsample_layers_hi);
        if( rr.realm_type==realm_t::lightness )
        {
            ASSERT(0);
            ASSERT(fmt==storage_format::i8);
            apply_lightness_v1(ds_wrk, rr.filters, rr.projected_size, rr.flags, rr.realm_name, rr.downsample_min, rr.downsample_max, tmp_va);
        }
        else if( rr.realm_type==realm_t::gray )
        {
            ASSERT(0);
            ASSERT(fmt==storage_format::i8);
            apply_gray_v1(ds_wrk, rr.filters, rr.projected_size, rr.flags, rr.realm_name, rr.downsample_min, rr.downsample_max, tmp_va);
        }
        else if( rr.realm_type==realm_t::color6_edges )
        {
            ASSERT(0);
            ASSERT(fmt==storage_format::i8);
            apply_colored_edges_v1(ds_wrk, rr.filters, rr.projected_size, rr.flags, rr.realm_name, false, tmp_va[0]);
            for(int i=1; i<downsample_layers_hi; i++)
                downsample2_max(tmp_va[i-1], "", tmp_va[i]);
        }
        else if( rr.realm_type==realm_t::color8_uni )
        {
            ASSERT(0);
            ASSERT(fmt==storage_format::i8);
            apply_colored_uniform_v1(ds_wrk, rr.filters, rr.projected_size, rr.flags, rr.realm_name, rr.downsample_min, rr.downsample_max, tmp_va);
        }
        else if( rr.realm_type==realm_t::kmeans ||
                 rr.realm_type==realm_t::kmeans_lgt ||
                 rr.realm_type==realm_t::kmeans_clr ||
                 rr.realm_type==realm_t::kmeans_gry ||
                 rr.realm_type==realm_t::kmeans_clu ||
                 rr.realm_type==realm_t::kmeans_clg )
        {
            ASSERT(0);
            ASSERT(fmt==storage_format::i8);
            apply_kmeans_v1(ds_wrk, rr.filters, rr.projected_size, rr.realm_type, rr.kmeans_wht_stab, rr.kmeans_hardness, rr.kmeans_rad_param, rr.realm_name, rr.downsample_min, rr.downsample_max, tmp_va);
        }
        else if( rr.realm_type==realm_t::kmeans_lxg || rr.realm_type==realm_t::kmeans_xxg )
        {
            apply_kmeans_v1_xg(ds_wrk, rr.filters, rr.projected_size, rr.realm_type, rr.lxg_alpha_param, rr.lxg_beta_param, rr.filter_gain, rr.realm_name, rr.downsample_min, rr.downsample_max, fmt, tmp_va);
        }
        else if( rr.realm_type==realm_t::kmeans_clx1 )
        {
            apply_kmeans_v1_xc(ds_wrk, rr.filters, rr.projected_size, rr.realm_type, rr.clx_alpha_param, rr.realm_name, rr.downsample_min, rr.downsample_max, fmt, tmp_va);
        }
        else
            ASSERT(0);

        // modify realm names with downsampling prefixes
        for(int i=downsample_layers_lo-1, ds=1<<(downsample_layers_lo-1); i<downsample_layers_hi; i++, ds*=2)
            for(realm<channel> &rx: tmp_va[i].realms)
            {
                char rname[1024];
                sprintf(rname, "ds%d.%s", ds, rx.name.c_str());
                rx.name = rname;
            }

        // append
        for(int i=downsample_layers_lo-1, ds=1<<(downsample_layers_lo-1); i<downsample_layers_hi; i++, ds*=2)
            for(const realm<channel> &rx: tmp_va[i].realms)
            {
                if( act_map.count(rx.name)==0 )
                    act_map[rx.name] = new activities(ds_raw.ntrn, ds_raw.ntst, fmt);
                act_map[rx.name]->append_planes_from_realm(tmp_va[i], rx.name);
            }
        if( report )
        {
            printf("OK\n");
            fflush(stdout);
        }
    }

    // merge
    int nchannels_total = 0;
    for(auto &iter: act_map)
        nchannels_total += iter.second->nchannels;
    act_out.reserve_storage(nchannels_total);
    for(auto &iter: act_map)
    {
        act_out.merge_in(*(iter.second));
        delete iter.second;
        iter.second = NULL;
    }

    t0 = dtickcount()-t0;
    if( report )
    {
        printf("[V1] layer is done in %d ms (memory usage %.2f GB)\n", t0, double(act_out.get_memory_usage()*1.0e-9));
        fflush(stdout);
    }
}

#warning DECIDE ON V1 NORMALIZATION!!!

void conv_apply_v1_xg(int raw_height,
                      int raw_width,
                      const bitmap &img_in,
                      const filter_set &filters,
                      int projected_size,
                      const float lxg_alpha,
                      const float lxg_beta,
                      const float filter_gain,
                      std::string realm_name,
                      int downsample_min,
                      int downsample_max,
                      storage_format fmt,
                      std::vector<conv_activities> &vec_out)
{
    // perform integrity check for filter set, determine filters count
    ASSERT(filters.rules.size()==1);
    ASSERT(filters.rules.front().chsize0==projected_size && filters.rules.front().chsize1==projected_size+1);
    const filter_family &family = filters.rules[0].filters;
    const int nkernels = family.cnt;
    const int patch_size = family.size;
    const int patch_size2= patch_size*patch_size;
    const int patches0 =img_in.height-patch_size+1, patches1 = img_in.width-patch_size+1;
    const int shift = patch_size/2;
    ASSERT(img_in.height>=patch_size && img_in.width>=patch_size);

    // allocation, basic initialization by zeros
    vec_out = std::vector<conv_activities>();
    for(int ds=1; ds<=downsample_max; ds*=2)
    {
        ASSERT(ds==downsample_max || ds<=downsample_max/2);
        ASSERT(projected_size%ds==0);
        ASSERT(img_in.height%ds==0);
        ASSERT(img_in.width%ds==0);
        vec_out.push_back(conv_activities(raw_height, raw_width, fmt));
    }
    for(int k=0, ds=1, wrk0=img_in.height, wrk1=img_in.width, ws=projected_size;
        ds<=downsample_max;
        ds*=2, wrk0=iDivUp(wrk0,2), wrk1=iDivUp(wrk1,2), ws/=2, k++)
        if( ds>=downsample_min && ds<=downsample_max )
        {
            vec_out[k].append_realm(conv_activities::conv_realm::create_grid(realm_name+".gry", wrk0, wrk1, k, ws, ws, nkernels));
            vec_out[k].append_realm(conv_activities::conv_realm::create_grid(realm_name+".lgt", wrk0, wrk1, k, ws, ws, 2*nkernels));
        }

    // fetch image
    std::vector<float> _v_r(img_in.height*img_in.width), _v_g(img_in.height*img_in.width), _v_b(img_in.height*img_in.width);
    for(int64_t pos0=0; pos0<img_in.height; pos0++)
    {
        float s = 1.0f/255.0f;
        for(int64_t pos1=0; pos1<img_in.width; pos1++)
        {
            int ir, ig, ib;
            img_in.get_rgb(pos0, pos1, &ir, &ig, &ib);
            _v_r[pos0*img_in.width+pos1] = ir*s;
            _v_g[pos0*img_in.width+pos1] = ig*s;
            _v_b[pos0*img_in.width+pos1] = ib*s;
        }
    }

    // prepare internal output buffer
    downsampler dspl_lgt(2*nkernels, img_in.height, img_in.width, downsample_max);
    downsampler dspl_gry(  nkernels, img_in.height, img_in.width, downsample_max);

    // process image rows
    activityf_vector dotk_r(nkernels*img_in.width), dotk_g(nkernels*img_in.width), dotk_b(nkernels*img_in.width);
    activityf_vector wrk_lgt(2*nkernels*img_in.width), wrk_gry(nkernels*img_in.width);
    activityf_vector mean_lgt(img_in.width), mean_gry(img_in.width);
    std::vector<float> padded_lgt(2*nkernels*img_in.height*img_in.width), padded_gry(nkernels*img_in.height*img_in.width);
    float scl_lgt=1.0e-20f, scl_gry=1.0e-20f;
    memset(&padded_lgt[0], 0, sizeof(padded_lgt[0])*padded_lgt.size());
    memset(&padded_gry[0], 0, sizeof(padded_gry[0])*padded_gry.size());
    for(int uidx0=0; uidx0<patches0; uidx0++)
    {
        // compute dot products of r/g/b with kernels, store with padding
        memset(&dotk_r[0], 0, dotk_r.size()*sizeof(dotk_r[0]));
        memset(&dotk_g[0], 0, dotk_g.size()*sizeof(dotk_g[0]));
        memset(&dotk_b[0], 0, dotk_b.size()*sizeof(dotk_b[0]));
        for(int k=0; k<nkernels; k++)
            for(int j=0; j<patch_size2; j++)
            {
                int src_offs = (uidx0+j/patch_size)*img_in.width+(j%patch_size);
                fadd_simd(&dotk_r[k*img_in.width+shift], &_v_r[src_offs], patches1, family.coeffs[k*patch_size2+j]);
                fadd_simd(&dotk_g[k*img_in.width+shift], &_v_g[src_offs], patches1, family.coeffs[k*patch_size2+j]);
                fadd_simd(&dotk_b[k*img_in.width+shift], &_v_b[src_offs], patches1, family.coeffs[k*patch_size2+j]);
            }

        // produce activities
        memset(&mean_lgt[0],     0, mean_lgt.size()*sizeof(mean_lgt[0]));
        memset(&mean_gry[0],     0, mean_gry.size()*sizeof(mean_gry[0]));
        for(int k=0; k<nkernels; k++)
        {
            const float *p_src_r = &dotk_r[k*img_in.width];
            const float *p_src_g = &dotk_g[k*img_in.width];
            const float *p_src_b = &dotk_b[k*img_in.width];
            float *p_dst_lgt0 = &wrk_lgt[(2*k+0)*img_in.width];
            float *p_dst_lgt1 = &wrk_lgt[(2*k+1)*img_in.width];
            float *p_dst_gry  = &wrk_gry[k*img_in.width];
            for(int i=0; i<img_in.width; i++)
            {
                float v_lgt = 0.333333333f*(p_src_r[i]+p_src_g[i]+p_src_b[i]);
                float v_lgt_0 = fmaxf(+v_lgt, 0.0f);
                float v_lgt_1 = fmaxf(-v_lgt, 0.0f);
                p_dst_lgt0[i] = v_lgt_0;
                p_dst_lgt1[i] = v_lgt_1;
                float v_gry = fmax3f(fabsf(p_src_r[i]), fabsf(p_src_g[i]), fabsf(p_src_b[i]));
                p_dst_gry[i]  = v_gry;
            }
            fadd_simd(&mean_lgt[0], p_dst_lgt0, img_in.width, 0.5f/nkernels);
            fadd_simd(&mean_lgt[0], p_dst_lgt1, img_in.width, 0.5f/nkernels);
            fadd_simd(&mean_gry[0], p_dst_gry,  img_in.width, 1.0f/nkernels);
        }
        for(int k=0; k<nkernels; k++)
        {
            const float *p_mean_lgt = &mean_lgt[0];
            const float *p_mean_gry = &mean_gry[0];
            const float *p_src_lgt0 = &wrk_lgt[(2*k+0)*img_in.width];
            const float *p_src_lgt1 = &wrk_lgt[(2*k+1)*img_in.width];
            const float *p_src_gry  = &wrk_gry[k*img_in.width];
            float *p_dst_lgt0 = &padded_lgt[(2*k+0)*img_in.height*img_in.width+(uidx0+shift)*img_in.width];
            float *p_dst_lgt1 = &padded_lgt[(2*k+1)*img_in.height*img_in.width+(uidx0+shift)*img_in.width];
            float *p_dst_gry  = &padded_gry[k*img_in.height*img_in.width+(uidx0+shift)*img_in.width];
            for(int i=0; i<img_in.width; i++)
            {
                float v0 = fmaxf(p_src_lgt0[i]-lxg_beta*p_mean_lgt[i], 0.0f);
                float v1 = fmaxf(p_src_lgt1[i]-lxg_beta*p_mean_lgt[i], 0.0f);
                float vg = fmaxf( p_src_gry[i]-lxg_alpha*p_mean_gry[i], 0.0f);
                p_dst_lgt0[i] = v0;
                p_dst_lgt1[i] = v1;
                p_dst_gry[i]  = vg;
                scl_lgt = fmaxf(scl_lgt, v0);
                scl_lgt = fmaxf(scl_lgt, v1);
                scl_gry = fmaxf(scl_gry, vg);
            }
        }
    }

    // normalize padded results
    if( filter_gain==0 )
    {
        fmul_inplace(&padded_lgt[0], 1.0f/scl_lgt, padded_lgt.size());
        fmul_inplace(&padded_gry[0], 1.0f/scl_gry, padded_gry.size());
    }
    else
    {
        fmul_inplace(&padded_lgt[0], filter_gain, padded_lgt.size());
        fmul_inplace(&padded_gry[0], filter_gain, padded_gry.size());
        fmin_inplace(&padded_lgt[0], 1.0f, padded_lgt.size());
        fmin_inplace(&padded_gry[0], 1.0f, padded_gry.size());
    }

    // downsample and offload to activities vector
    dspl_lgt.apply_downsampling(padded_lgt);
    dspl_gry.apply_downsampling(padded_gry);
    for(int ds_k=0, ds=1; ds<=downsample_max; ds*=2,ds_k++)
    {
        ASSERT(ds_k<vec_out.size());
        conv_activities &act_out = vec_out[ds_k];
        if( ds>=downsample_min && ds<=downsample_max )
        {
            // gry realm
            ASSERT(act_out.realms.size()==1 || act_out.realms.size()==2);
            ASSERT(act_out.realms[0].grid_depth==nkernels);
            const conv_activities::conv_realm &rr_gry = act_out.realms[0];
            dspl_gry.output_kth_layer(ds_k, act_out, rr_gry);

            // lgt realm
            ASSERT(act_out.realms.size()==2);
            ASSERT(act_out.realms[1].grid_depth==2*nkernels);
            const conv_activities::conv_realm &rr_lgt = act_out.realms[1];
            dspl_lgt.output_kth_layer(ds_k, act_out, rr_lgt);
        }
    }
}

void net_v1layer::conv_apply(const bitmap &img, conv_activities &act_out, bool report, storage_format fmt) const
{
    act_out = conv_activities(img.height, img.width, fmt);
    int t0 = dtickcount();

    //
    // Prepare downscaled versions of dataset
    //
    int max_scale_idx = 0;
    for(const v1_realm &rr: realms)
        max_scale_idx = imax(max_scale_idx, rr.scale_idx);
    std::vector<bitmap> img_downscaled(max_scale_idx);
    if( max_scale_idx>0 )
    {
        img_downscaled[0] = img.downsample2();
        for(int k=1; k<img_downscaled.size(); k++)
            img_downscaled[k] = img_downscaled[k-1].downsample2();
    }

    //
    // Prepare activity map
    //
    std::map<std::string,conv_activities*> act_map;

    //
    // Generate activities
    //
    for(const v1_realm &rr: realms)
    {
        ASSERT(rr.realm_type!=realm_t::none);
        ASSERT(rr.downsample_min>=1 && rr.downsample_min<=rr.downsample_max);
        for(int i=rr.downsample_min; i>=1; i/=2)
            ASSERT(i%2==0 || i==1);
        for(int i=rr.downsample_max; i>=1; i/=2)
            ASSERT(i%2==0 || i==1);

        // select image to work with
        const bitmap &img_wrk = rr.scale_idx==0 ? img : img_downscaled[rr.scale_idx-1];

        // determine number of downsample layers
        int downsample_layers_lo = 0;
        for(int i=rr.downsample_min; i>=1; i/=2)
            downsample_layers_lo++;
        int downsample_layers_hi = 0;
        for(int i=rr.downsample_max; i>=1; i/=2)
            downsample_layers_hi++;

        //
        // Initial processing
        //
        if( report )
        {
            printf("[V1] realm %s... ", rr.realm_name.c_str());
            fflush(stdout);
        }
        std::vector<conv_activities> tmp_va(downsample_layers_hi);
        if( rr.realm_type==realm_t::kmeans_lxg )
        {
            conv_apply_v1_xg(img.height, img.width, img_wrk, rr.filters, rr.projected_size, rr.lxg_alpha_param, rr.lxg_beta_param, rr.filter_gain, rr.realm_name, rr.downsample_min, rr.downsample_max, fmt, tmp_va);
        }
        else
            ASSERT(0);

        // modify realm names with downsampling prefixes
        for(int i=downsample_layers_lo-1, ds=1<<(downsample_layers_lo-1); i<downsample_layers_hi; i++, ds*=2)
            for(conv_activities::conv_realm &rx: tmp_va[i].realms)
            {
                char rname[1024];
                sprintf(rname, "ds%d.%s", ds, rx.name.c_str());
                rx.name = rname;
            }

        // append
        for(int i=downsample_layers_lo-1, ds=1<<(downsample_layers_lo-1); i<downsample_layers_hi; i++, ds*=2)
            for(const conv_activities::conv_realm &rx: tmp_va[i].realms)
            {
                if( act_map.count(rx.name)==0 )
                    act_map[rx.name] = new conv_activities(img.height, img.width, fmt);
                act_map[rx.name]->append_planes_from_realm(tmp_va[i], rx.name);
            }
        if( report )
        {
            printf("OK\n");
            fflush(stdout);
        }
    }

    // merge
    int64_t ndata_total = 0;
    for(auto &iter: act_map)
        ndata_total += iter.second->get_memory_usage();
    act_out.reserve_storage(ndata_total);
    for(auto &iter: act_map)
    {
        act_out.merge_in(*(iter.second));
        delete iter.second;
        iter.second = NULL;
    }

    t0 = dtickcount()-t0;
    if( report )
    {
        printf("[V1] layer is done in %d ms (memory usage %.2f GB)\n", t0, double(act_out.get_memory_usage()*1.0e-9));
        fflush(stdout);
    }
}

void net_v1layer::run_dummy_optimization()
{
    for(v1_realm &rr: realms)
    {
        // skip realms which can't be optimized
        if( rr.realm_type!=realm_t::kmeans_lxg )
        {
            printf("[V1-dummy-opt] realm %s ignored\n", rr.realm_name.c_str());
            fflush(stdout);
            continue;
        }

        // integrity checks
        ASSERT(rr.filters.rules.size()==1);
        filter_set::rule &fr = rr.filters.rules[0];
        filter_family    &ff = fr.filters;
        ASSERT(ff.size%2==0);
        ASSERT(ff.cnt>=1);
        ASSERT(fr.chsize0==fr.chsize1-1);
        int fsize  = ff.size;
        int fsize2 = fsize/2;

        // report
        printf("[V1-dummy-opt] Analyzing realm %s, filter count is %d, filter size is %d^2\n", rr.realm_name.c_str(), ff.cnt, fsize);
        fflush(stdout);

        //
        // Split filters into fsize2*fsize2 chunks
        //
        std::vector<std::vector<float>> all_chunks;
        for(int k=0; k<ff.cnt; k++)
        {
            const float *coeffs = &ff.coeffs[k*ff.size*ff.size];
            for(int k0=0; k0<2; k0++)
                for(int k1=0; k1<2; k1++)
                {
                    std::vector<float> chunk;
                    for(int i=0; i<fsize2; i++)
                        for(int j=0; j<fsize2; j++)
                            chunk.push_back(coeffs[(k0*fsize2+i)*fsize+(k1*fsize2+j)]);
                    all_chunks.push_back(chunk);
                }
        }
        //printf("> all_chunks = %d items\n", int(all_chunks.size()));

        //
        // Create dictionary of reference chunks
        //
        std::vector<std::vector<float>> unique_chunks;
        const int   MAX_CNT = ff.cnt;
        const float MAX_ERR = 0.01f;
        std::vector<float> max_errors(all_chunks.size(),1.0e20f);
        std::vector<int>   ref_indexes(all_chunks.size(),-1);
        std::vector<float> ref_scales(all_chunks.size(),0.0f);
        float mx_err = 0;
        for(;;)
        {
            const int n = fsize2*fsize2;

            // find chunk with minimum error
            int mx_idx = 0;
            mx_err = max_errors[mx_idx];
            for(int i=0; i<all_chunks.size(); i++)
                if( max_errors[i]>max_errors[mx_idx] )
                {
                    mx_idx = i;
                    mx_err = max_errors[mx_idx];
                }
            if( unique_chunks.size()>=MAX_CNT || mx_err<MAX_ERR )
                break;
            std::vector<float> worst_chunk = all_chunks[mx_idx];

            // add normalized worst chunk to unique list
            {
                std::vector<float> nrm_chunk = worst_chunk;
                fmul_inplace(&nrm_chunk[0], 1/sqrtf(vnrm2(&nrm_chunk[0], n)), n);
                unique_chunks.push_back(nrm_chunk);
            }

            // update errors
            for(int k=0; k<all_chunks.size(); k++)
            {
                float best_err  = max_errors[k];
                int   ref_idx   = ref_indexes[k];
                float ref_scale = ref_scales[k];

                // find best single-vector approximation
                for(int uidx=0; uidx<unique_chunks.size(); uidx++)
                {
                    const std::vector<float> &cand0 = unique_chunks[uidx];
                    std::vector<float> wrk_chunk = all_chunks[k];
                    float cand_scale = fdot_simd(&cand0[0], &wrk_chunk[0], n);
                    fadd_simd(&wrk_chunk[0], &cand0[0], n, -cand_scale);
                    float cand_err = 0;
                    for(int i=0; i<n; i++)
                        cand_err = fmaxf(cand_err, fabsf(wrk_chunk[i]));
                    if( cand_err<best_err )
                    {
                        best_err  = cand_err;
                        ref_idx   = uidx;
                        ref_scale = cand_scale;
                    }
                }

                // update
                max_errors[k]  = best_err;
                ref_indexes[k] = ref_idx;
                ref_scales[k]  = ref_scale;
            }
        }

        //
        // Replace chunks stored in the layer by artificially constructed ones
        //
        float global_mx = 0;
        for(const auto &c: all_chunks)
            for(float v: c)
                global_mx = fmaxf(global_mx, fabsf(v));
        float max_restored_error = 0.0f;
        for(int k=0; k<ff.cnt; k++)
        {
            float *coeffs = &ff.coeffs[k*ff.size*ff.size];
            int base_cidx = 4*k;
            std::vector<float> wrk_coeffs(fsize*fsize,-1.0e20f);
            for(int k0=0; k0<2; k0++)
                for(int k1=0; k1<2; k1++)
                {
                    ASSERT(ref_indexes[base_cidx]>=0);
                    float scl = ref_scales[base_cidx];
                    std::vector<float> ref_chunk = unique_chunks[ref_indexes[base_cidx]];
                    for(int i=0; i<fsize2; i++)
                        for(int j=0; j<fsize2; j++)
                            wrk_coeffs[(k0*fsize2+i)*fsize+(k1*fsize2+j)] = scl*ref_chunk[i*fsize2+j];
                    base_cidx++;
                }
            for(int i=0; i<fsize*fsize; i++)
                max_restored_error = fmaxf(max_restored_error, fabsf(coeffs[i]-wrk_coeffs[i]));
            for(int i=0; i<fsize*fsize; i++)
                coeffs[i] = wrk_coeffs[i];
        }
        printf("[V1-dummy-opt] > dict_size =%d, max_err = %.3f\n", (int)unique_chunks.size(), max_restored_error/global_mx);

    }
}

void net_v1layer::debug_output(std::string path) const
{
    for(const v1_realm &r: realms)
        r.debug_output(path);
}

void serialize(FILE *f, const net_v1layer::v1_realm *ptr)
{
    write_magic8(f, 66);
    serialize_as_int64(f, &ptr->realm_type);
    serialize(f, &ptr->realm_name);
    serialize(f, &ptr->filters);
    serialize_as_int64(f, &ptr->flags);
    serialize(f, &ptr->scale_idx);
    serialize(f, &ptr->projected_size);
    serialize(f, &ptr->downsample_min);
    serialize(f, &ptr->downsample_max);
    serialize(f, &ptr->kmeans_wht_stab);
    serialize(f, &ptr->kmeans_hardness);
    serialize(f, &ptr->kmeans_rad_param);
    serialize(f, &ptr->lxg_alpha_param);
    serialize(f, &ptr->lxg_beta_param);
    serialize(f, &ptr->clx_alpha_param);
    serialize(f, &ptr->filter_gain);
    serialize(f, &ptr->kmeans_flt7);
    serialize(f, &ptr->kmeans_int0);
    serialize(f, &ptr->kmeans_int1);
    serialize(f, &ptr->kmeans_int2);
    serialize(f, &ptr->kmeans_int3);
    serialize(f, &ptr->kmeans_int4);
    serialize(f, &ptr->kmeans_int5);
    serialize(f, &ptr->kmeans_int6);
    serialize(f, &ptr->kmeans_int7);
    serialize(f, &ptr->kmeans_wht_matrix);
    write_magic8(f, 219);
}

void serialize(FILE *f, const net_v1layer *ptr)
{
    write_magic8(f, 22);
    serialize(f, &ptr->realms);
    write_magic8(f, 41);
}

void unserialize(FILE *f, net_v1layer::v1_realm *ptr)
{
    check_magic8(f, 66);
    unserialize_as_int64(f, &ptr->realm_type);
    unserialize(f, &ptr->realm_name);
    unserialize(f, &ptr->filters);
    unserialize_as_int64(f, &ptr->flags);
    unserialize(f, &ptr->scale_idx);
    unserialize(f, &ptr->projected_size);
    unserialize(f, &ptr->downsample_min);
    unserialize(f, &ptr->downsample_max);
    unserialize(f, &ptr->kmeans_wht_stab);
    unserialize(f, &ptr->kmeans_hardness);
    unserialize(f, &ptr->kmeans_rad_param);
    unserialize(f, &ptr->lxg_alpha_param);
    unserialize(f, &ptr->lxg_beta_param);
    unserialize(f, &ptr->clx_alpha_param);
    unserialize(f, &ptr->filter_gain);
    unserialize(f, &ptr->kmeans_flt7);
    unserialize(f, &ptr->kmeans_int0);
    unserialize(f, &ptr->kmeans_int1);
    unserialize(f, &ptr->kmeans_int2);
    unserialize(f, &ptr->kmeans_int3);
    unserialize(f, &ptr->kmeans_int4);
    unserialize(f, &ptr->kmeans_int5);
    unserialize(f, &ptr->kmeans_int6);
    unserialize(f, &ptr->kmeans_int7);
    unserialize(f, &ptr->kmeans_wht_matrix);
    check_magic8(f, 219);
}

void unserialize(FILE *f, net_v1layer *ptr)
{
    check_magic8(f, 22);
    unserialize(f, &ptr->realms);
    check_magic8(f, 41);
}


filter_family generate_uni_color_area(int size)
{
    filter_family r;
    r.size = size;
    r.cnt  = 0;
    if( size>=2 )
    {
        std::vector<float> patch;
        for(int i=0; i<size*size; i++)
            patch.push_back(+1.0f);
        ASSERT(patch.size()==size*size);
        r.coeffs.insert(r.coeffs.end(), patch.begin(), patch.end());
        r.cnt += 1;
        normalize_sums(r);
        return r;
    }
    ASSERT(0);
    abort();

}

filter_family generate_ea_colorless(int size)
{
    filter_family r;
    r.size = size;
    r.cnt  = 0;

    // size=3,5,7,9
    if( size==3 || size==5 || size==7 || size==9 )
    {
        std::vector<float> patch, edge_000, edge_045, edge_090, edge_135;

        // first, generate vertical and horizontal edges
        patch.clear();
        for(int i=0; i<(size/2)*size; i++)
            patch.push_back(+1.0f);
        for(int i=0; i<size; i++)
            patch.push_back( 0.0f);
        for(int i=0; i<(size/2)*size; i++)
            patch.push_back(-1.0f);
        ASSERT(patch.size()==size*size);
        edge_000 = patch;
        edge_090 = rotate90(patch, size);

        // second, generate diagonal edges
        patch.clear();
        for(int k=0; k<size; k++)
        {
            for(int i=0; i<size-1-k; i++)
                patch.push_back(+1.0f);
            patch.push_back(0.0f);
            for(int i=0; i<k; i++)
                patch.push_back(-1.0f);
        }
        ASSERT(patch.size()==size*size);
        edge_045 = patch;
        edge_135 = rotate90(patch, size);

        // append
        r.coeffs.insert(r.coeffs.end(), edge_000.begin(), edge_000.end());
        r.coeffs.insert(r.coeffs.end(), edge_045.begin(), edge_045.end());
        r.cnt += 2;
        r.coeffs.insert(r.coeffs.end(), edge_090.begin(), edge_090.end());
        r.coeffs.insert(r.coeffs.end(), edge_135.begin(), edge_135.end());
        r.cnt += 2;

        /*// generate right angles with 0/90/180/270 orientations
        patch.clear();
        for(int i=0; i<size/2; i++)
            for(int j=0; j<size; j++)
                patch.push_back(-1.0f);
        for(int i=size/2; i<size; i++)
        {
            for(int j=0; j<size/2; j++)
                patch.push_back(-1.0f);
            for(int j=size/2; j<size; j++)
                patch.push_back(+1.0f);
        }
        ASSERT(patch.size()==size*size);
        r.coeffs.insert(r.coeffs.end(), patch.begin(), patch.end());
        patch = rotate90(patch, size);
        r.coeffs.insert(r.coeffs.end(), patch.begin(), patch.end());
        patch = rotate90(patch, size);
        r.coeffs.insert(r.coeffs.end(), patch.begin(), patch.end());
        patch = rotate90(patch, size);
        r.coeffs.insert(r.coeffs.end(), patch.begin(), patch.end());
        r.cnt += 4;*/

        // done
        normalize_sums(r);
        return r;
    }
    ASSERT(0);
    abort();
}

filter_family generate_ea_colored(int size)
{
    filter_family r;
    r.size = size;
    r.cnt  = 0;

    // size=3,5,7,9
    if( size==3 || size==5 || size==7 || size==9 )
    {
        std::vector<float> patch, edge_000, edge_045, edge_090, edge_135, edge_180, edge_225, edge_270, edge_315;

        // first, generate vertical and horizontal edges
        patch.clear();
        for(int i=0; i<(size/2)*size; i++)
            patch.push_back(+1.0f);
        for(int i=0; i<size; i++)
            patch.push_back( 0.0f);
        for(int i=0; i<(size/2)*size; i++)
            patch.push_back(-1.0f);
        ASSERT(patch.size()==size*size);
        edge_000 = patch;
        edge_090 = (patch=rotate90(patch, size));
        edge_180 = (patch=rotate90(patch, size));
        edge_270 = (patch=rotate90(patch, size));

        // second, generate diagonal edges
        patch.clear();
        for(int k=0; k<size; k++)
        {
            for(int i=0; i<size-1-k; i++)
                patch.push_back(+1.0f);
            patch.push_back(0.0f);
            for(int i=0; i<k; i++)
                patch.push_back(-1.0f);
        }
        ASSERT(patch.size()==size*size);
        edge_045 = patch;
        edge_135 = (patch=rotate90(patch, size));
        edge_225 = (patch=rotate90(patch, size));
        edge_315 = (patch=rotate90(patch, size));

        // append
        r.coeffs.insert(r.coeffs.end(), edge_000.begin(), edge_000.end());
        r.coeffs.insert(r.coeffs.end(), edge_045.begin(), edge_045.end());
        r.coeffs.insert(r.coeffs.end(), edge_090.begin(), edge_090.end());
        r.coeffs.insert(r.coeffs.end(), edge_135.begin(), edge_135.end());
        r.coeffs.insert(r.coeffs.end(), edge_180.begin(), edge_180.end());
        r.coeffs.insert(r.coeffs.end(), edge_225.begin(), edge_225.end());
        r.coeffs.insert(r.coeffs.end(), edge_270.begin(), edge_270.end());
        r.coeffs.insert(r.coeffs.end(), edge_315.begin(), edge_315.end());
        r.cnt += 8;

        /*// generate right angles with 0/90/180/270 orientations
        patch.clear();
        for(int i=0; i<size/2; i++)
            for(int j=0; j<size; j++)
                patch.push_back(-1.0f);
        for(int i=size/2; i<size; i++)
        {
            for(int j=0; j<size/2; j++)
                patch.push_back(-1.0f);
            for(int j=size/2; j<size; j++)
                patch.push_back(+1.0f);
        }
        ASSERT(patch.size()==size*size);
        r.coeffs.insert(r.coeffs.end(), patch.begin(), patch.end());
        patch = rotate90(patch, size);
        r.coeffs.insert(r.coeffs.end(), patch.begin(), patch.end());
        patch = rotate90(patch, size);
        r.coeffs.insert(r.coeffs.end(), patch.begin(), patch.end());
        patch = rotate90(patch, size);
        r.coeffs.insert(r.coeffs.end(), patch.begin(), patch.end());
        r.cnt += 4;*/

        // done
        normalize_sums(r);
        return r;
    }
    ASSERT(0);
    abort();
}

filter_family generate_edges_blobs_colored(int size)
{
    filter_family r = generate_ea_colored(size);

    if( size==3 || size==5 || size==7 || size==9)
    {
        std::vector<float> blob_pos, blob_neg;
        if( size==3 )
        {
            blob_pos = {-1, -1, -1,
                        -1, +1, -1,
                        -1, -1, -1};
        }
        else if( size==5 )
        {
            blob_pos = {-1, -1, -1, -1, -1,
                        -1, -1, +1, -1, -1,
                        -1, +1, +1, +1, -1,
                        -1, -1, +1, -1, -1,
                        -1, -1, -1, -1, -1};
        }
        else
            ASSERT(0);
        ASSERT(blob_pos.size()==size*size);
        for(float v: blob_pos)
            blob_neg.push_back(-v);
        r.coeffs.insert(r.coeffs.end(), blob_pos.begin(), blob_pos.end());
        r.coeffs.insert(r.coeffs.end(), blob_neg.begin(), blob_neg.end());
        r.cnt += 2;

        // done
        normalize_sums(r);
        return r;
    }
    ASSERT(0);
    abort();
}

filter_family generate_femto_colored()
{
    int size = 2;

    filter_family r;
    r.size = size;
    r.cnt  = 0;

    std::vector<float> patch, edge_000, edge_090, edge_180, edge_270;

    patch.clear();
    patch.push_back(+1.0f);
    patch.push_back(+1.0f);
    patch.push_back(-1.0f);
    patch.push_back(-1.0f);
    edge_000 = patch;
    edge_090 = (patch=rotate90(patch, size));
    edge_180 = (patch=rotate90(patch, size));
    edge_270 = (patch=rotate90(patch, size));

    // append
    r.coeffs.insert(r.coeffs.end(), edge_000.begin(), edge_000.end());
    r.coeffs.insert(r.coeffs.end(), edge_090.begin(), edge_090.end());
    r.coeffs.insert(r.coeffs.end(), edge_180.begin(), edge_180.end());
    r.coeffs.insert(r.coeffs.end(), edge_270.begin(), edge_270.end());
    r.cnt += 4;

    // done
    normalize_sums(r);
    return r;
}

filter_family generate_e_colorless_micro(int size)
{
    filter_family r;
    r.size = size;
    r.cnt  = 0;

    // size=3
    if( size==3 )
    {
        std::vector<float> patch;

        // first, generate vertical and horizontal edges
        patch = {+1, +1, 0, -1, -1, 0, 0, 0, 0};
        ASSERT(patch.size()==size*size);
        r.coeffs.insert(r.coeffs.end(), patch.begin(), patch.end());
        patch = {+1, -1, 0, +1, -1, 0, 0, 0, 0};
        ASSERT(patch.size()==size*size);
        r.coeffs.insert(r.coeffs.end(), patch.begin(), patch.end());
        r.cnt += 2;

        // second, generate diagonal edges
        patch = {0, +1, 0, +1, 0, -1, 0, -1, 0};
        ASSERT(patch.size()==size*size);
        r.coeffs.insert(r.coeffs.end(), patch.begin(), patch.end());
        patch = {0, -1, 0, +1, 0, -1, 0, +1, 0};
        ASSERT(patch.size()==size*size);
        r.coeffs.insert(r.coeffs.end(), patch.begin(), patch.end());
        r.cnt += 2;

        // done
        return r;
    }

    // size=4
    if( size==4 )
    {
        std::vector<float> patch;

        // first, generate vertical and horizontal edges
        patch = {+1, +1, +1, +1, +1, +1, +1, +1, -1, -1, -1, -1, -1, -1, -1, -1};
        ASSERT(patch.size()==size*size);
        r.coeffs.insert(r.coeffs.end(), patch.begin(), patch.end());
        patch = rotate90(patch, size);
        r.coeffs.insert(r.coeffs.end(), patch.begin(), patch.end());
        r.cnt += 2;

        // second, generate diagonal edges
        patch = {+1, +1, +1, 0,   +1, +1, 0, -1,   +1, 0, -1, -1,  0, -1, -1, -1};
        ASSERT(patch.size()==size*size);
        r.coeffs.insert(r.coeffs.end(), patch.begin(), patch.end());
        patch = rotate90(patch, size);
        r.coeffs.insert(r.coeffs.end(), patch.begin(), patch.end());
        r.cnt += 2;

        // done
        return r;
    }

    // bad size
    ASSERT(0);
    abort();
}

filter_family generate_ea_colorless_micro(int size)
{
    filter_family r;
    r.size = size;
    r.cnt  = 0;

    // size=4
    if( size==4 )
    {
        std::vector<float> patch;

        // first, generate vertical and horizontal edges
        patch = {+1, +1, +1, +1, +1, +1, +1, +1, -1, -1, -1, -1, -1, -1, -1, -1};
        ASSERT(patch.size()==size*size);
        r.coeffs.insert(r.coeffs.end(), patch.begin(), patch.end());
        patch = rotate90(patch, size);
        r.coeffs.insert(r.coeffs.end(), patch.begin(), patch.end());
        r.cnt += 2;

        // second, generate diagonal edges
        patch = {+1, +1, +1, 0,   +1, +1, 0, -1,   +1, 0, -1, -1,  0, -1, -1, -1};
        ASSERT(patch.size()==size*size);
        r.coeffs.insert(r.coeffs.end(), patch.begin(), patch.end());
        patch = rotate90(patch, size);
        r.coeffs.insert(r.coeffs.end(), patch.begin(), patch.end());
        r.cnt += 2;

        // generate angles
        patch = {+1, +1, -3, -3,   +1, +1, -3, -3,   +1, +1, +1, +1,  +1, +1, +1, +1};
        ASSERT(patch.size()==size*size);
        r.coeffs.insert(r.coeffs.end(), patch.begin(), patch.end());
        patch = rotate90(patch, size);
        r.coeffs.insert(r.coeffs.end(), patch.begin(), patch.end());
        patch = rotate90(patch, size);
        r.coeffs.insert(r.coeffs.end(), patch.begin(), patch.end());
        patch = rotate90(patch, size);
        r.coeffs.insert(r.coeffs.end(), patch.begin(), patch.end());
        patch = rotate90(patch, size);
        r.cnt += 4;

        // done
        return r;
    }

    // bad size
    ASSERT(0);
    abort();
}

filter_family generate_eab_colorless_micro(int size)
{
    filter_family r;
    r.size = size;
    r.cnt  = 0;

    // size=4
    if( size==4 )
    {
        std::vector<float> patch;

        // first, generate vertical and horizontal edges
        patch = {+1, +1, +1, +1, +1, +1, +1, +1, -1, -1, -1, -1, -1, -1, -1, -1};
        ASSERT(patch.size()==size*size);
        r.coeffs.insert(r.coeffs.end(), patch.begin(), patch.end());
        patch = rotate90(patch, size);
        r.coeffs.insert(r.coeffs.end(), patch.begin(), patch.end());
        r.cnt += 2;

        // second, generate diagonal edges
        patch = {+1, +1, +1, 0,   +1, +1, 0, -1,   +1, 0, -1, -1,  0, -1, -1, -1};
        ASSERT(patch.size()==size*size);
        r.coeffs.insert(r.coeffs.end(), patch.begin(), patch.end());
        patch = rotate90(patch, size);
        r.coeffs.insert(r.coeffs.end(), patch.begin(), patch.end());
        r.cnt += 2;

        // generate angles
        patch = {+1, +1, -3, -3,   +1, +1, -3, -3,   +1, +1, +1, +1,  +1, +1, +1, +1};
        ASSERT(patch.size()==size*size);
        r.coeffs.insert(r.coeffs.end(), patch.begin(), patch.end());
        patch = rotate90(patch, size);
        r.coeffs.insert(r.coeffs.end(), patch.begin(), patch.end());
        patch = rotate90(patch, size);
        r.coeffs.insert(r.coeffs.end(), patch.begin(), patch.end());
        patch = rotate90(patch, size);
        r.coeffs.insert(r.coeffs.end(), patch.begin(), patch.end());
        patch = rotate90(patch, size);
        r.cnt += 4;

        // finally, generate bars
        patch = {+1, +1, +1, +1,   -1, -1, -1, -1,   -1, -1, -1, -1,   +1, +1, +1, +1};
        ASSERT(patch.size()==size*size);
        r.coeffs.insert(r.coeffs.end(), patch.begin(), patch.end());
        patch = rotate90(patch, size);
        r.coeffs.insert(r.coeffs.end(), patch.begin(), patch.end());
        r.cnt += 2;

        // done
        return r;
    }

    // bad size
    ASSERT(0);
    abort();
}

static float patch_lgt_variance(const bitmap &img, const int offs0, const int offs1, const int patch_size)
{
    ASSERT(patch_size>=3 && patch_size<=16);

    // fetch lightness
    float lgt[patch_size*patch_size];
    for(int i0=0; i0<patch_size; i0++)
        for(int i1=0; i1<patch_size; i1++)
        {
            int r, g, b;
            img.get_rgb(offs0+i0, offs1+i1, &r, &g, &b);
            lgt[i0*patch_size+i1] = (r+g+b)/(3.0f*255.0f);
        }

    // mean
    float lgt_mean = 0;
    for(int i0=0; i0<patch_size; i0++)
        for(int i1=0; i1<patch_size; i1++)
            lgt_mean += lgt[i0*patch_size+i1];
    lgt_mean /= patch_size*patch_size;

    // variance
    float lgt_var = 0;
    for(int i0=0; i0<patch_size; i0++)
        for(int i1=0; i1<patch_size; i1++)
        {
            float v = lgt[i0*patch_size+i1]-lgt_mean;
            lgt_var += v*v;
        }
    lgt_var /= patch_size*patch_size;
    return lgt_var;
}

template<class pixel_t>
void whiten_patch(pixel_t *p_patch, int patch_size, int nchannels, float wht_stab)
{
    const int ndim = patch_size*patch_size*nchannels;
    float mean = 0.0f;
    for(int j=0; j<ndim; j++)
        mean += p_patch[j];
    mean /= ndim;
    float variance = 0;
    for(int j=0; j<ndim; j++)
    {
        float v = p_patch[j]-mean;
        p_patch[j] = v;
        variance += v*v;
    }
    variance /= ndim;
    float scale = sqrtf(variance+wht_stab);
    for(int j=0; j<ndim; j++)
        p_patch[j] /= scale;
}

static float patch_clr_variance(const bitmap &img, const int offs0, const int offs1, const int patch_size)
{
    ASSERT(patch_size>=3);
    float clr_mean_r = 0;
    float clr_mean_g = 0;
    float clr_mean_b = 0;
    for(int i0=0; i0<patch_size; i0++)
        for(int i1=0; i1<patch_size; i1++)
        {
            int r, g, b;
            img.get_rgb(offs0+i0, offs1+i1, &r, &g, &b);
            float lgt = (r+g+b)/(3.0f*255.0f);
            clr_mean_r += r/255.0f-lgt;
            clr_mean_g += g/255.0f-lgt;
            clr_mean_b += b/255.0f-lgt;
        }
    clr_mean_r /= patch_size*patch_size;
    clr_mean_g /= patch_size*patch_size;
    clr_mean_b /= patch_size*patch_size;
    float clr_var = 0;
    for(int i0=0; i0<patch_size; i0++)
        for(int i1=0; i1<patch_size; i1++)
        {
            int r, g, b;
            img.get_rgb(offs0+i0, offs1+i1, &r, &g, &b);
            float lgt = (r+g+b)/(3.0f*255.0f);
            float vr = (r/255.0f-lgt)-clr_mean_r;
            float vg = (g/255.0f-lgt)-clr_mean_g;
            float vb = (b/255.0f-lgt)-clr_mean_b;
            clr_var += vr*vr+vg*vg+vb*vb;
        }
    clr_var /= 3*patch_size*patch_size;
    return clr_var;
}

static float patch_wht_clr_variance(const bitmap &img, const int offs0, const int offs1, const int patch_size, const float wht_stab)
{
    ASSERT(patch_size>=3);
    float patch[patch_size*patch_size*3];
    for(int i0=0; i0<patch_size; i0++)
        for(int i1=0; i1<patch_size; i1++)
        {
            int r, g, b;
            img.get_rgb(offs0+i0, offs1+i1, &r, &g, &b);
            float lgt = (r+g+b)/(3.0f*255.0f);
            patch[3*(i0*patch_size+i1)+0] = r/255.0f-lgt;
            patch[3*(i0*patch_size+i1)+1] = g/255.0f-lgt;
            patch[3*(i0*patch_size+i1)+2] = b/255.0f-lgt;
        }
    whiten_patch(patch, patch_size, 3, wht_stab);
    float clr_mean_r = 0;
    float clr_mean_g = 0;
    float clr_mean_b = 0;
    for(int i0=0; i0<patch_size; i0++)
        for(int i1=0; i1<patch_size; i1++)
        {
            clr_mean_r += patch[3*(i0*patch_size+i1)+0];
            clr_mean_g += patch[3*(i0*patch_size+i1)+1];
            clr_mean_b += patch[3*(i0*patch_size+i1)+2];
        }
    clr_mean_r /= patch_size*patch_size;
    clr_mean_g /= patch_size*patch_size;
    clr_mean_b /= patch_size*patch_size;
    float clr_var = 0;
    for(int i0=0; i0<patch_size; i0++)
        for(int i1=0; i1<patch_size; i1++)
        {
            float vr = patch[3*(i0*patch_size+i1)+0]-clr_mean_r;
            float vg = patch[3*(i0*patch_size+i1)+1]-clr_mean_g;
            float vb = patch[3*(i0*patch_size+i1)+2]-clr_mean_b;
            clr_var += vr*vr+vg*vg+vb*vb;
        }
    clr_var /= 3*patch_size*patch_size;
    return clr_var;
}

filter_family train_kmeans_filters(const raw_dataset &ds_raw,
                                   const int scale_idx,
                                   net_v1layer::realm_t kmeans_type,
                                   int patch_size,
                                   int sample_size,
                                   int nfilters,
                                   const v1_settings s,
                                   const char *rep_suffix)
{
    ASSERT(kmeans_type==net_v1layer::realm_t::kmeans ||
           kmeans_type==net_v1layer::realm_t::kmeans_lgt ||
           kmeans_type==net_v1layer::realm_t::kmeans_clr ||
           kmeans_type==net_v1layer::realm_t::kmeans_gry ||
           kmeans_type==net_v1layer::realm_t::kmeans_clg ||
           kmeans_type==net_v1layer::realm_t::kmeans_clu);
    int channels_per_filter = 0;
    if( kmeans_type==net_v1layer::realm_t::kmeans )
        channels_per_filter = 3;
    if( kmeans_type==net_v1layer::realm_t::kmeans_lgt)
        channels_per_filter = 1;
    if( kmeans_type==net_v1layer::realm_t::kmeans_clr)
        channels_per_filter = 3;
    if( kmeans_type==net_v1layer::realm_t::kmeans_gry)
        channels_per_filter = 1;
    if( kmeans_type==net_v1layer::realm_t::kmeans_clg)
        channels_per_filter = 3;
    if( kmeans_type==net_v1layer::realm_t::kmeans_clu)
        channels_per_filter = 3;
    const int ndim = channels_per_filter*patch_size*patch_size;
    std::mt19937  mt_rnd;
    std::seed_seq mt_seed{(unsigned int)1, (unsigned int)117};
    mt_rnd.seed(mt_seed);

    //
    // Report
    //
    int t0 = dtickcount();
    printf("Started training K-means\n");
    fflush(stdout);

    //
    // First, generate training sample
    //
    std::uniform_int_distribution<> rnd_img(0, ds_raw.ntrn-1);
    alglib::real_2d_array xy;
    std::vector<int> labels;
    xy.setlength(sample_size, ndim);
    for(int k=0; k<sample_size; k++)
    {
        int img_idx = rnd_img(mt_rnd);
        raw_dataset::entry entry = ds_raw.entries[img_idx];
        for(int s=0; s<scale_idx; s++)
            entry.img = entry.img.downsample2();
        ASSERT(entry.img.height>=patch_size && entry.img.width>=patch_size);

        // select patch
        std::uniform_int_distribution<> rnd_offs0(0, entry.img.height-patch_size);
        std::uniform_int_distribution<> rnd_offs1(0, entry.img.width-patch_size);
        const int nsearch = 50;
        int offs0 = -1, offs1 = -1;
        if( kmeans_type==net_v1layer::realm_t::kmeans )
        {
            offs0 = rnd_offs0(mt_rnd);
            offs1 = rnd_offs1(mt_rnd);
        }
        else if( kmeans_type==net_v1layer::realm_t::kmeans_lgt )
        {
            offs0 = rnd_offs0(mt_rnd);
            offs1 = rnd_offs1(mt_rnd);
        }
        else if( kmeans_type==net_v1layer::realm_t::kmeans_clr )
        {
            offs0 = rnd_offs0(mt_rnd);
            offs1 = rnd_offs1(mt_rnd);
            /*ASSERT(patch_size>=3);
            float winner_rating = -1;
            for(int ns=0; ns<nsearch; ns++)
            {
                int cand_offs0 = rnd_offs0(mt_rnd);
                int cand_offs1 = rnd_offs1(mt_rnd);
                float lgt_var = patch_lgt_variance(e.img, cand_offs0, cand_offs1, patch_size);
                float clr_var = patch_clr_variance(e.img, cand_offs0, cand_offs1, patch_size);
                float cand_rating;
                if( k%3==0 )
                    cand_rating = clr_var;
                if( k%3==1 )
                    cand_rating = 1.0f/(1.0f+clr_var);
                if( k%3==2 )
                    cand_rating = lgt_var;
                if( cand_rating>winner_rating )
                {
                    winner_rating = cand_rating;
                    offs0 = cand_offs0;
                    offs1 = cand_offs1;
                }
            }*/
        }
        else if( kmeans_type==net_v1layer::realm_t::kmeans_gry )
        {
            offs0 = rnd_offs0(mt_rnd);
            offs1 = rnd_offs1(mt_rnd);
        }
        else if( kmeans_type==net_v1layer::realm_t::kmeans_clg )
        {
            ASSERT(patch_size>=3);
            offs0 = rnd_offs0(mt_rnd);
            offs1 = rnd_offs1(mt_rnd);
            float winner_rating = -1;
            for(int ns=0; ns<nsearch; ns++)
            {
                int cand_offs0 = rnd_offs0(mt_rnd);
                int cand_offs1 = rnd_offs1(mt_rnd);
                float cand_rating = patch_wht_clr_variance(entry.img, cand_offs0, cand_offs1, patch_size, s.kmeans_wht_stab);
                if( cand_rating>winner_rating )
                {
                    winner_rating = cand_rating;
                    offs0 = cand_offs0;
                    offs1 = cand_offs1;
                }
            }
        }
        else if( kmeans_type==net_v1layer::realm_t::kmeans_clu )
        {
            ASSERT(patch_size>=3);
            offs0 = rnd_offs0(mt_rnd);
            offs1 = rnd_offs1(mt_rnd);
            float winner_rating = -1;
            for(int ns=0; ns<nsearch; ns++)
            {
                int cand_offs0 = rnd_offs0(mt_rnd);
                int cand_offs1 = rnd_offs1(mt_rnd);
                float cand_rating = 1.0f/(1.0f+patch_wht_clr_variance(entry.img, cand_offs0, cand_offs1, patch_size, s.kmeans_wht_stab));
                if( cand_rating>winner_rating )
                {
                    winner_rating = cand_rating;
                    offs0 = cand_offs0;
                    offs1 = cand_offs1;
                }
            }
        }
        else
            ASSERT(0);
        ASSERT(offs0>=0 && offs1>=0);

        // process patch
        labels.push_back(entry.class_id);
        for(int i0=0; i0<patch_size; i0++)
            for(int i1=0; i1<patch_size; i1++)
            {
                if( kmeans_type==net_v1layer::realm_t::kmeans )
                {
                    int dst = 3*(i0*patch_size+i1), r, g, b;
                    entry.img.get_rgb(offs0+i0, offs1+i1, &r, &g, &b);
                    xy(k,dst+0) = r/255.0;
                    xy(k,dst+1) = g/255.0;
                    xy(k,dst+2) = b/255.0;
                }
                else if( kmeans_type==net_v1layer::realm_t::kmeans_lgt ||
                         kmeans_type==net_v1layer::realm_t::kmeans_gry )
                {
                    int dst = i0*patch_size+i1, r, g, b;
                    entry.img.get_rgb(offs0+i0, offs1+i1, &r, &g, &b);
                    xy(k,dst) = (r+g+b)/(3.0f*255.0f);
                }
                else if( kmeans_type==net_v1layer::realm_t::kmeans_clr ||
                         kmeans_type==net_v1layer::realm_t::kmeans_clu ||
                         kmeans_type==net_v1layer::realm_t::kmeans_clg )
                {
                    int dst = 3*(i0*patch_size+i1), r, g, b;

                    entry.img.get_rgb(offs0+i0, offs1+i1, &r, &g, &b);

                    float fr = r/255.0f, fg = g/255.0f, fb = b/255.0f;
                    float flgt = (fr+fg+fb)/3.0f;
                    xy(k,dst+0) = fr-flgt;
                    xy(k,dst+1) = fg-flgt;
                    xy(k,dst+2) = fb-flgt;
                }
                else
                    ASSERT(0);
            }
    }

    //
    // Normalization
    //
    for(int k=0; k<sample_size; k++)
    {
        double *p_row = &xy(k,0);
        double mean = 0.0f;
        for(int j=0; j<ndim; j++)
            mean += p_row[j]/ndim;
        double variance = 0;
        for(int j=0; j<ndim; j++)
        {
            double v = p_row[j]-mean;
            p_row[j] = v;
            variance += v*v/ndim;
        }
        double scale = sqrt(variance+s.kmeans_wht_stab);
        for(int j=0; j<ndim; j++)
            p_row[j] /= scale;
    }

    //
    // Determine number of work filters
    //
    int nwork_filters = nfilters;
    if( kmeans_type==net_v1layer::realm_t::kmeans_clr )
        nwork_filters *= s.kmeans_clr_super;
    if( kmeans_type==net_v1layer::realm_t::kmeans_gry )
        nwork_filters *= 2;
    if( kmeans_type==net_v1layer::realm_t::kmeans_clg )
        nwork_filters *= 4;
    ASSERT(nwork_filters>=nfilters);

    //
    // Clustering
    //
    alglib::clusterizerstate cs;
    alglib::kmeansreport rep;
    alglib::clusterizercreate(cs);
    alglib::clusterizersetpoints(cs, xy, 2);
    alglib::clusterizersetkmeanslimits(cs, 1, 100);
    alglib::clusterizerrunkmeans(cs, nwork_filters, rep, alglib::parallel);
    ASSERT(rep.terminationtype>0);

    //
    // Analysis
    //
    std::vector<int> cluster_sizes(nwork_filters,0);
    for(int k=0; k<rep.cidx.length(); k++)
        cluster_sizes[rep.cidx[k]]++;
    int kmin = 0, kmax = 0;
    for(int k=0; k<nwork_filters; k++)
    {
        if( cluster_sizes[k]<cluster_sizes[kmin] )
            kmin = k;
        if( cluster_sizes[k]>cluster_sizes[kmax] )
            kmax = k;
    }

    //
    // Conversion
    //
    auto dist_pos = [](double *p0, double *p1, int64_t n)
    { double r = 0; for(int64_t k=0; k<n; k++) { double vv = p0[k]-p1[k]; r += vv*vv; } return r; };
    auto dist_neg = [](double *p0, double *p1, int64_t n)
    { double r = 0; for(int64_t k=0; k<n; k++) { double vv = p0[k]+p1[k]; r += vv*vv; } return r; };
    filter_family out;
    alglib::real_2d_array centers;
    if( kmeans_type==net_v1layer::realm_t::kmeans || kmeans_type==net_v1layer::realm_t::kmeans_lgt )
    {
        ASSERT(nwork_filters==nfilters);
        centers = rep.c;
    }
    else if( kmeans_type==net_v1layer::realm_t::kmeans_clr )
    {
        if( s.kmeans_clr_super>1 )
        {
            const float mean_decay = 1.0f;
            int t0 = dtickcount();
            printf("Smart CLR k-means... ");
            fflush(stdout);

            predictions baseline(sample_size, 0, ds_raw.nclasses, labels);
            activityf_vector vact(nwork_filters*sample_size);
            linear_evaluator eval(baseline, nfilters, 0.00001);
            for(int64_t idx=0; idx<sample_size; idx++)
            {
                float dist2[nwork_filters];
                float mean_dist2 = 1.0e-20f;
                for(int k=0; k<nwork_filters; k++)
                {
                    dist2[k] = dist_pos(&xy(idx,0),&rep.c(k,0),ndim);
                    mean_dist2 += dist2[k]/nwork_filters;
                }
                for(int k=0; k<nwork_filters; k++)
                    vact[k*sample_size+idx] = fmaxf(1.0f-dist2[k]/(mean_decay*mean_dist2),0.0f);
            }
            centers.setlength(nfilters, ndim);
            for(int fidx=0; fidx<nfilters; fidx++)
            {
                // select best filter
                int idx_best = -1;
                float rat_best = 0.0f;
                for(int k=0; k<nwork_filters; k++)
                {
                    float cv2 = eval.estimate_trn_covariation(&vact[k*sample_size]);
                    float nrm2 = vnrm2(&vact[k*sample_size],sample_size);
                    float rat = cv2/(nrm2+sample_size*0.00001);
                    //float rat = eval.deltar2(&vact[k*sample_size]);
                    if( idx_best<0 || rat>rat_best )
                    {
                        idx_best = k;
                        rat_best = rat;
                    }
                }
                ASSERT(idx_best>=0);

                // store
                for(int j=0; j<ndim; j++)
                    centers(fidx,j) = rep.c(idx_best,j);

                // add to evaluator
                eval.append_activity(activityf_vector(&vact[idx_best*sample_size],&vact[(idx_best+1)*sample_size]));
            }
            printf("done in %d ms\n", int(dtickcount()-t0));
            fflush(stdout);
        }
        else
        {
            ASSERT(nwork_filters==nfilters);
            centers = rep.c;
        }
    }
    else if( kmeans_type==net_v1layer::realm_t::kmeans_gry )
    {
        ASSERT(nwork_filters==2*nfilters);
        const float big_dist = 1.0e10f;
        std::vector<float> mindist(nwork_filters,big_dist);
        mindist[kmax] = 2*big_dist;
        centers.setlength(nfilters, ndim);
        for(int k=0; k<nfilters; k++)
        {
            // select cluster with maximum distance
            int idx = 0;
            for(int i=0; i<nwork_filters; i++)
                if( mindist[i]>mindist[idx] )
                    idx = i;

            // store
            for(int j=0; j<ndim; j++)
                centers(k,j) = rep.c(idx,j);

            // update distances
            for(int i=0; i<nwork_filters; i++)
            {
                float dp = dist_pos(&rep.c(idx,0), &rep.c(i,0), ndim);
                float dn = dist_neg(&rep.c(idx,0), &rep.c(i,0), ndim);
                mindist[i] = fminf(mindist[i],fminf(dp,dn));
            }
        }
    }
    else if( kmeans_type==net_v1layer::realm_t::kmeans_clg )
    {
        ASSERT(nwork_filters==4*nfilters);
        const float big_dist = 1.0e10f;
        std::vector<float> mindist(nwork_filters,big_dist);
        mindist[kmax] = 2*big_dist;
        centers.setlength(nfilters, ndim);
        for(int k=0; k<nfilters; k++)
        {
            // select cluster with maximum distance
            int idx = 0;
            for(int i=0; i<nwork_filters; i++)
                if( mindist[i]>mindist[idx] )
                    idx = i;

            // store
            for(int j=0; j<ndim; j++)
                centers(k,j) = rep.c(idx,j);

            // update distances
            double cur_filter[ndim];
            for(int j=0; j<ndim; j++)
                cur_filter[j] = rep.c(idx,j);
            for(int nrot=0; nrot<4; nrot++)
            {
                for(int i=0; i<nwork_filters; i++)
                    mindist[i] = fminf(mindist[i],dist_pos(cur_filter, &rep.c(i,0), ndim));
                rotate90(cur_filter, patch_size, channels_per_filter);
            }
        }
    }
    else if( kmeans_type==net_v1layer::realm_t::kmeans_clu )
    {
        ASSERT(nwork_filters==nfilters);
        centers = rep.c;
    }
    else
        ASSERT(0);
    out.size = patch_size;
    out.cnt  = nfilters;
    for(int k=0; k<nfilters; k++)
        for(int j=0; j<ndim; j++)
            out.coeffs.push_back(centers(k,j));

    //
    // Report
    //
    printf("Cluster %d is the smallest one (%d elements in the training sample)\n", kmin, cluster_sizes[kmin]);
#if 0
    for(int k=0; k<nfilters; k++)
    {
        const int scale = 128;
        bitmap img;
        img.resize(patch_size, patch_size);
        for(int i=0; i<patch_size; i++)
            for(int j=0; j<patch_size; j++)
            {
                if( kmeans_type==net_v1layer::realm_t::kmeans )
                {
                    img.draw_rgb(i, j,
                                 centers(k,3*(i*patch_size+j)+0)*scale+128,
                                 centers(k,3*(i*patch_size+j)+1)*scale+128,
                                 centers(k,3*(i*patch_size+j)+2)*scale+128);
                }
                else if( kmeans_type==net_v1layer::realm_t::kmeans_lgt ||
                         kmeans_type==net_v1layer::realm_t::kmeans_gry )
                {
                    int v = centers(k,i*patch_size+j)*scale+128;
                    img.draw_rgb(i, j, v, v, v);
                }
                else if( kmeans_type==net_v1layer::realm_t::kmeans_clr ||
                         kmeans_type==net_v1layer::realm_t::kmeans_clg ||
                         kmeans_type==net_v1layer::realm_t::kmeans_clu )
                {
                    img.draw_rgb(i, j,
                                 centers(k,3*(i*patch_size+j)+0)*scale+128,
                                 centers(k,3*(i*patch_size+j)+1)*scale+128,
                                 centers(k,3*(i*patch_size+j)+2)*scale+128);
                }
                else
                    ASSERT(0);
            }
        const char *skmtype = "unknown";
        if( kmeans_type==net_v1layer::realm_t::kmeans )
            skmtype = "kmn";
        if( kmeans_type==net_v1layer::realm_t::kmeans_lgt )
            skmtype = "lgt";
        if( kmeans_type==net_v1layer::realm_t::kmeans_clr )
            skmtype = "clr";
        if( kmeans_type==net_v1layer::realm_t::kmeans_gry )
            skmtype = "gry";
        if( kmeans_type==net_v1layer::realm_t::kmeans_clg )
            skmtype = "clg";
        if( kmeans_type==net_v1layer::realm_t::kmeans_clu )
            skmtype = "clu";
        char out_name[1024];
        sprintf(out_name, "%s/kmeans-%s-%s-c%03d.png", std::string(OUT_DIR).c_str(), skmtype, rep_suffix, k);
        img.to_png(out_name);
    }
#endif

    //
    // Done
    //
    t0 = dtickcount()-t0;
    printf("K-means training is done in %d ms\n", t0);
    fflush(stdout);
    return out;
}

filter_family train_kmeans_filters_lxg(const raw_dataset &ds_raw,
                                       const int scale_idx,
                                       int patch_size,
                                       int sample_size,
                                       int nfilters,
                                       const v1_settings s,
                                       const char *rep_suffix)
{
    const int patch_size2 = patch_size*patch_size;
    std::mt19937  mt_rnd;
    std::seed_seq mt_seed{(unsigned int)1, (unsigned int)117};
    mt_rnd.seed(mt_seed);

    //
    // Report
    //
    int t0 = dtickcount();
    printf("Started training K-means\n");
    fflush(stdout);

    //
    // First, generate training sample
    //
    std::uniform_int_distribution<> rnd_img(0, ds_raw.ntrn-1);
    alglib::real_2d_array xy;
    std::vector<int> labels;
    xy.setlength(sample_size, patch_size2);
    for(int k=0; k<sample_size; k++)
    {
        int img_idx = rnd_img(mt_rnd);
        raw_dataset::entry entry = ds_raw.entries[img_idx];
        for(int s=0; s<scale_idx; s++)
            entry.img = entry.img.downsample2();
        ASSERT(entry.img.height>=patch_size && entry.img.width>=patch_size);

        // select patch
        std::uniform_int_distribution<> rnd_offs0(0, entry.img.height-patch_size);
        std::uniform_int_distribution<> rnd_offs1(0, entry.img.width-patch_size);
        int offs0 = rnd_offs0(mt_rnd);
        int offs1 = rnd_offs1(mt_rnd);

        // process patch
        labels.push_back(entry.class_id);
        for(int i0=0; i0<patch_size; i0++)
            for(int i1=0; i1<patch_size; i1++)
            {
                int dst = i0*patch_size+i1, r, g, b;
                entry.img.get_rgb(offs0+i0, offs1+i1, &r, &g, &b);
                xy(k,dst) = (r+g+b)/(3.0f*255.0f);
            }
    }

    //
    // Normalization
    //
    for(int k=0; k<sample_size; k++)
    {
        double *p_row = &xy(k,0);
        double mean = 0.0f;
        for(int j=0; j<patch_size2; j++)
            mean += p_row[j]/patch_size2;
        double norm2 = 0;
        for(int j=0; j<patch_size2; j++)
        {
            double v = p_row[j]-mean;
            p_row[j] = v;
            norm2 += v*v;
        }
        double scale = sqrt(norm2+powf(0.03f,2));
        for(int j=0; j<patch_size2; j++)
            p_row[j] /= scale;
    }

    //
    // Determine number of work filters
    //
    int nwork_filters = 2*nfilters;

    //
    // Clustering
    //
    alglib::clusterizerstate cs;
    alglib::kmeansreport rep;
    alglib::clusterizercreate(cs);
    alglib::clusterizersetpoints(cs, xy, 2);
    alglib::clusterizersetkmeanslimits(cs, 1, 100);
    alglib::clusterizerrunkmeans(cs, nwork_filters, rep, alglib::parallel);
    ASSERT(rep.terminationtype>0);

    //
    // Analysis
    //
    std::vector<int> cluster_sizes(nwork_filters,0);
    for(int k=0; k<rep.cidx.length(); k++)
        cluster_sizes[rep.cidx[k]]++;
    int kmin = 0, kmax = 0;
    for(int k=0; k<nwork_filters; k++)
    {
        if( cluster_sizes[k]<cluster_sizes[kmin] )
            kmin = k;
        if( cluster_sizes[k]>cluster_sizes[kmax] )
            kmax = k;
    }

    //
    // Conversion
    //
    auto dist_pos = [](double *p0, double *p1, int64_t n)
    { double r = 0; for(int64_t k=0; k<n; k++) { double vv = p0[k]-p1[k]; r += vv*vv; } return r; };
    auto dist_neg = [](double *p0, double *p1, int64_t n)
    { double r = 0; for(int64_t k=0; k<n; k++) { double vv = p0[k]+p1[k]; r += vv*vv; } return r; };
    filter_family out;
    alglib::real_2d_array centers;
    ASSERT(nwork_filters==2*nfilters);
    const float big_dist = 1.0e10f;
    std::vector<float> mindist(nwork_filters,big_dist);
    mindist[kmax] = 2*big_dist;
    centers.setlength(nfilters, patch_size2);
    for(int k=0; k<nfilters; k++)
    {
        // select cluster with maximum distance
        int idx = 0;
        for(int i=0; i<nwork_filters; i++)
            if( mindist[i]>mindist[idx] )
                idx = i;

        // store
        for(int j=0; j<patch_size2; j++)
            centers(k,j) = rep.c(idx,j);

        // update distances
        for(int i=0; i<nwork_filters; i++)
        {
            float dp = dist_pos(&rep.c(idx,0), &rep.c(i,0), patch_size2);
            float dn = dist_neg(&rep.c(idx,0), &rep.c(i,0), patch_size2);
            mindist[i] = fminf(mindist[i],fminf(dp,dn));
        }
    }
    for(int k=0; k<nfilters; k++)
    {
        double *p_center = &centers(k,0);
        double mean = 0;
        for(int j=0; j<patch_size2; j++)
            mean += centers(k,j)/patch_size2;
        double norm2 = 0;
        for(int j=0; j<patch_size2; j++)
        {
            centers(k,j) -= mean;
            norm2 += centers(k,j)*centers(k,j);
        }
        for(int j=0; j<patch_size2; j++)
            centers(k,j) /= sqrt(norm2)+0.001;
    }
    out.size = patch_size;
    out.cnt  = nfilters;
    for(int k=0; k<nfilters; k++)
        for(int j=0; j<patch_size2; j++)
            out.coeffs.push_back(centers(k,j));

    //
    // Report
    //
    printf("Cluster %d is the smallest one (%d elements in the training sample)\n", kmin, cluster_sizes[kmin]);
#if 0
    for(int k=0; k<nfilters; k++)
    {
        const int scale = 96*int(sqrt(patch_size2));
        bitmap img;
        img.resize(patch_size, patch_size);
        for(int i=0; i<patch_size; i++)
            for(int j=0; j<patch_size; j++)
            {
                int v = centers(k,i*patch_size+j)*scale+128;
                img.draw_rgb(i, j, v, v, v);
            }
        char out_name[1024];
        sprintf(out_name, "%s/kmeans-lxg-%s-c%03d.png", std::string(OUT_DIR).c_str(), rep_suffix, k);
        img.to_png(out_name);
    }
#endif

    //
    // Done
    //
    t0 = dtickcount()-t0;
    printf("K-means training is done in %d ms\n", t0);
    fflush(stdout);
    return out;
}

void convolve(const std::vector<float> &input, int input_h, int input_w,
              const std::vector<float> &filter, int filter_h, int filter_w,
              std::vector<float> &out)
{
    int out_h = input_h-filter_h+1;
    int out_w = input_w-filter_w+1;
    ASSERT(0<filter_h && filter_h<=input_h);
    ASSERT(0<filter_w && filter_w<=input_w);

    out.resize(out_h*out_w);
    memset(&out[0], 0, out.size()*sizeof(out[0]));
    for(int i1=0; i1<filter_h; i1++)
        for(int j1=0; j1<filter_w; j1++)
        {
            float w = filter[i1*filter_w+j1];
            if( w==0 )
                continue;
            float *p_out = &out[0];
            const float *p_in = &input[i1*input_w+j1];
            for(int i0=0; i0<out_h; i0++)
            {
                //for(int j0=0; j0<out_w; j0++)
                //    p_out[i0*out_w+j0] += w*p_in[i0*input_w+j0];
                fadd_simd(p_out+i0*out_w, p_in+i0*input_w, out_w, w);
            }
        }
}

void apply_family(const bitmap &bmp, int pos0, int pos1, const filter_family &family, std::vector<float> &rr)
{
    int fsize =family.size;
    ASSERT(fsize<20); // integrity check
    float patch[3*fsize*fsize];

    // fetch patch
    float ss = 1.0f/255.0f;
    ASSERT(pos0+fsize-1<bmp.height);
    ASSERT(pos1+fsize-1<bmp.width);
    for(int i=0; i<fsize; i++)
        for(int j=0; j<fsize; j++)
        {
            int r, g, b;
            bmp.get_rgb(pos0+i, pos1+j, &r, &g, &b);
            patch[0*fsize*fsize+i*fsize+j] = r*ss;
            patch[1*fsize*fsize+i*fsize+j] = g*ss;
            patch[2*fsize*fsize+i*fsize+j] = b*ss;
        }

    // apply filters from family
    rr.resize(family.cnt);
    for(int fidx=0; fidx<family.cnt; fidx++)
    {
        const float *p_filter = &family.coeffs[fsize*fsize*fidx];
        float vmax = 0.0f;
        for(int cidx=0; cidx<3; cidx++)
        {
            float v = 0.0f;
            for(int k=0; k<fsize*fsize; k++)
                v += patch[cidx*fsize*fsize+k]*p_filter[k];
            vmax = fmaxf(vmax, fabsf(v));
        }
        rr[fidx] = vmax;
    }
}

class v1_parallel_colored_uniform
{
public:
    const raw_dataset &ds_raw;
    const filter_set &filters;
    const int projected_size;
    const v1_flags_t flags;
    const int batch_size;
    const int img_count;
    const int cgain_levels;
    const int downsample_min;
    const int downsample_max;
    std::atomic<int32_t> batch_idx;
    std::vector<activities> &vec_out;
    v1_parallel_colored_uniform(const raw_dataset &rds,
                              const filter_set &fs,
                              int ps, int cgl, v1_flags_t f, int bs, int ic,
                              int dsmin, int dsmax, std::vector<activities> &vo):
        ds_raw(rds),filters(fs),projected_size(ps),
        flags(f),batch_size(bs),img_count(ic),batch_idx(0),
        downsample_min(dsmin),downsample_max(dsmax),vec_out(vo),
        cgain_levels(cgl)
    {}
};

void* parallel_apply_colored_uniform_v1(void *xptr)
{
    const int nfilters = 8;
    v1_parallel_colored_uniform &env = *((v1_parallel_colored_uniform*)xptr);
    const raw_dataset   &ds_in = env.ds_raw;
    const filter_set    &filters = env.filters;
    const int           projected_size = env.projected_size;
    const int64_t       projected_size2 = projected_size*projected_size;
    const v1_flags_t    flags = env.flags;
    std::vector<activities> &vec_out = env.vec_out;

    // prepare internal output buffer
    downsampler dspl(nfilters, projected_size, projected_size, env.downsample_max);

    // perform integrity check for filter set, determine filters count
    ASSERT(filters.rules.size()>=1);
    for(const filter_set::rule &rr : filters.rules)
    {
        ASSERT(rr.filters.cnt==1);
        ASSERT(rr.filters.coeffs.size()==rr.filters.size*rr.filters.size*rr.filters.cnt);
    }

    // process dataset (note: int64_t is used to make sure no overflow takes place)
    std::vector<float> v_conv_r, v_conv_g, v_conv_b;
    std::vector<float> v_r, v_g, v_b;
    std::vector<float> out, gain_out, mx_out;
    for(int32_t batch_idx=env.batch_idx.fetch_add(1);
                batch_idx*env.batch_size<env.img_count;
                batch_idx=env.batch_idx.fetch_add(1))
        for(int64_t img_idx=batch_idx*env.batch_size; img_idx<imin((batch_idx+1)*env.batch_size,env.img_count); img_idx++)
        {
            // fetch image, subtract mean
            const raw_dataset::entry &ee = ds_in.entries[img_idx];
            v_r.resize(ee.img.height*ee.img.width);
            v_g.resize(ee.img.height*ee.img.width);
            v_b.resize(ee.img.height*ee.img.width);
            {
                const float inv255 = 1.0f/255.0f;
                float v_max = 0.0f, v_min = 999.0f;
                for(int64_t pos0=0; pos0<ee.img.height; pos0++)
                    for(int64_t pos1=0; pos1<ee.img.width; pos1++)
                    {
                        int r, g, b;
                        ee.img.get_rgb(pos0, pos1, &r, &g, &b);
                        v_r[pos0*ee.img.width+pos1] = r*inv255;
                        v_g[pos0*ee.img.width+pos1] = g*inv255;
                        v_b[pos0*ee.img.width+pos1] = b*inv255;
                        v_max = fmaxf(v_max,imax3(r,g,b)*inv255);
                        v_min = fminf(v_min,imin3(r,g,b)*inv255);
                    }
                float s = 1.0f/(v_max-v_min+1.0e-20f);
                for(float &v: v_r)
                    v = s*(v-v_min);
                for(float &v: v_g)
                    v = s*(v-v_min);
                for(float &v: v_b)
                    v = s*(v-v_min);
            }
            //white_balance(v_r, v_g, v_b);

            // compute convolution
            const filter_family &family = filters.find_family(imax(ee.img.width,ee.img.height));
            const int shift = family.size/2;
            const float *p_filter = &family.coeffs[0];
            std::vector<float> filter(p_filter, p_filter+family.size*family.size);
            float filter_sum = 0.0f;
            for(float f: filter)
                filter_sum += fabsf(f);
            for(float &v: filter)
                    v /= filter_sum;
            convolve(v_r, ee.img.height, ee.img.width,
                     filter, family.size, family.size,
                     v_conv_r);
            convolve(v_g, ee.img.height, ee.img.width,
                     filter, family.size, family.size,
                     v_conv_g);
            convolve(v_b, ee.img.height, ee.img.width,
                     filter, family.size, family.size,
                     v_conv_b);

            // prepare output locations
            int64_t conv_size0 = ee.img.height-family.size+1;
            int64_t conv_size1 = ee.img.width-family.size+1;
            int64_t arr_dst0[conv_size0], arr_dst1[conv_size1];
            ASSERT(ee.img.width==ee.img.height);
            for(int64_t pos0=0; pos0<conv_size0; pos0++)
                arr_dst0[pos0] = roundf(float(pos0+shift)/float(ee.img.height-1)*(projected_size-1));
            for(int64_t pos1=0; pos1<conv_size1; pos1++)
                arr_dst1[pos1] = roundf(float(pos1+shift)/float(ee.img.width -1)*(projected_size-1));


            // scan image, compute 8 projections (6 colors, black and white)
            out.clear();
            out.resize(nfilters*projected_size2,0.0f);
            for(int64_t pos0=0; pos0<conv_size0; pos0++)
                for(int64_t pos1=0; pos1<conv_size1; pos1++)
                {
                    // compute product with filter
                    float r = v_conv_r[pos0*(ee.img.width-family.size+1)+pos1];
                    float g = v_conv_g[pos0*(ee.img.width-family.size+1)+pos1];
                    float b = v_conv_b[pos0*(ee.img.width-family.size+1)+pos1];

                    // compute projections onto 8 directions in RGB space
                    float proj_red    = relu(2*r-g-b)/2.0f;
                    float proj_yellow = relu(r+g-2*b)/2.0f;
                    float proj_green  = relu(2*g-r-b)/2.0f;
                    float proj_teal   = relu(g+b-2*r)/2.0f;
                    float proj_blue   = relu(2*b-r-g)/2.0f;
                    float proj_purple = relu(r+b-2*g)/2.0f;
                    float proj_white  = relu(r+g+b-1.5f)/1.5f;
                    float proj_black  = relu(1.5f-r-g-b)/1.5f;
                    //float proj_white  = relu(r+g+b)/3;
                    //float proj_black  = relu(3-r-g-b)/3;

                    // determine output location for storage, store
                    int64_t dst  = arr_dst0[pos0]*projected_size+arr_dst1[pos1];
                    out[0*projected_size2+dst] = fmaxf(out[0*projected_size2+dst], proj_red);
                    out[1*projected_size2+dst] = fmaxf(out[1*projected_size2+dst], proj_yellow);
                    out[2*projected_size2+dst] = fmaxf(out[2*projected_size2+dst], proj_green);
                    out[3*projected_size2+dst] = fmaxf(out[3*projected_size2+dst], proj_teal);
                    out[4*projected_size2+dst] = fmaxf(out[4*projected_size2+dst], proj_blue);
                    out[5*projected_size2+dst] = fmaxf(out[5*projected_size2+dst], proj_purple);
                    out[6*projected_size2+dst] = fmaxf(out[6*projected_size2+dst], proj_white);
                    out[7*projected_size2+dst] = fmaxf(out[7*projected_size2+dst], proj_black);
                }

            // process each gain level
            for(int gain_idx=0; gain_idx<env.cgain_levels; gain_idx++)
            {
                float gain_val = powf(2.0f, gain_idx);
                float global_mx = 0;
                gain_out.clear();
                gain_out.resize(nfilters*projected_size2, 0.0f);
                mx_out.clear();
                mx_out.resize(projected_size2, 0.0f);

                // copy with gain applied
                for(int filter_idx=0; filter_idx<nfilters; filter_idx++)
                {
                    float s = filter_idx<6 ? gain_val : 1.0f;
                    const float *p_src = &out[filter_idx*projected_size2];
                    float       *p_dst = &gain_out[filter_idx*projected_size2];
                    for(int i=0; i<projected_size2; i++)
                    {
                        float v = s*p_src[i];
                        p_dst[i] = v;
                        mx_out[i] = fmaxf(mx_out[i], v);
                        global_mx = fmaxf(global_mx, v);
                    }
                }

                // competition
                for(int filter_idx=0; filter_idx<nfilters; filter_idx++)
                {
                    float s = 1.0f/global_mx;
                    const float *p_mx  = &mx_out[0];
                    float       *p_wrk = &gain_out[filter_idx*projected_size2];
                    for(int i=0; i<projected_size2; i++)
                        p_wrk[i] = s*(p_wrk[i]==p_mx[i] ? p_wrk[i] : 0.0f);
                }

                // downsample, offload to activities vector
                dspl.apply_downsampling(gain_out);
                for(int ds_k=0, ds=1; ds<=env.downsample_max; ds*=2,ds_k++)
                    if( ds>=env.downsample_min && ds<=env.downsample_max )
                    {
                        ASSERT(ds_k<vec_out.size());
                        activities &act_out = vec_out[ds_k];
                        int64_t cur_size = projected_size/ds;
                        dspl.output_kth_layer(ds_k, act_out, img_idx, gain_idx*nfilters*cur_size*cur_size);
                    }
                    else
                    {
                        // integrity check
                        activities &act_out = vec_out[ds_k];
                        ASSERT(act_out.nchannels==0);
                    }
            }
        }
    return NULL;
}

void apply_colored_uniform_v1(const raw_dataset &ds_in,
        const filter_set &filters,
        int projected_size,
        v1_flags_t flags,
        std::string realm_name,
        int downsample_min,
        int downsample_max,
        std::vector<activities> &vec_out)
{
    const int ncolors = 8;
    const int cgain_levels = 3;

    // perform integrity check for filter set, determine filters count
    ASSERT(filters.rules.size()>=1);
    for(const filter_set::rule &rr : filters.rules)
    {
        ASSERT(rr.filters.cnt==1);
        ASSERT(rr.filters.coeffs.size()==rr.filters.size*rr.filters.size*rr.filters.cnt);
    }

    // allocation, basic initialization by zeros
    vec_out = std::vector<activities>();
    for(int ds=1; ds<=downsample_max; ds*=2)
    {
        ASSERT(ds==downsample_max || ds<=downsample_max/2);
        ASSERT(projected_size%ds==0);
        vec_out.push_back(activities(ds_in.ntrn, ds_in.ntst, storage_format::i8));
    }
    for(int k=0, ds=1, ps=projected_size; ds<=downsample_max; ds*=2, ps/=2, k++)
        if( ds>=downsample_min && ds<=downsample_max )
        {
            for(int i=0; i<cgain_levels; i++)
            {
                char name[1024];
                sprintf(name, "%s%d", realm_name.c_str(), int(powf(2.0f,i)));
                vec_out[k].append_realm(realm<channel>::create_grid(name, ps, ps, ncolors));
            }
        }

    // run parallel processing
    const int batch_size = 64;
    v1_parallel_colored_uniform env(ds_in, filters, projected_size, cgain_levels, flags, batch_size, ds_in.ntrn+ds_in.ntst, downsample_min, downsample_max, vec_out);
    parallel_invoke(parallel_apply_colored_uniform_v1, &env);
}

class v1_parallel_colored_edges
{
public:
    const raw_dataset &ds_raw;
    const filter_set  &filters;
    const int projected_size;
    const bool need_black_and_white;
    const v1_flags_t  flags;
    const int batch_size;
    const int img_count;
    std::atomic<int32_t> batch_idx;
    activities &act_out;

    v1_parallel_colored_edges(const raw_dataset &rds,
                              const filter_set &fs,
                              int ps, bool nbw, v1_flags_t f, int bs, int ic,
                              activities &ao):
        ds_raw(rds),filters(fs),projected_size(ps),need_black_and_white(nbw),
        flags(f),batch_size(bs),img_count(ic),batch_idx(0),act_out(ao)
    {}
};

void* apply_colored_edges_v1_parallel(void *xptr)
{
    v1_parallel_colored_edges &env = *((v1_parallel_colored_edges*)xptr);
    const raw_dataset   &ds_in = env.ds_raw;
    const filter_set    &filters = env.filters;
    const int           projected_size = env.projected_size;
    const int64_t       projected_size2 = projected_size*projected_size;
    const bool          need_black_and_white = env.need_black_and_white;
    const v1_flags_t    flags = env.flags;
    activities          &act_out = env.act_out;

    // perform integrity check for filter set, determine filters count
    ASSERT(filters.rules.size()>=1);
    for(const filter_set::rule &rr : filters.rules)
    {
        ASSERT(rr.filters.cnt==filters.rules[0].filters.cnt);
        ASSERT(rr.filters.coeffs.size()==rr.filters.size*rr.filters.size*rr.filters.cnt);
    }
    int nedges  = filters.rules.front().filters.cnt;
    int ncolors_wrk = 8;
    int ncolors_out = need_black_and_white ? 8 : 6;

    // process dataset (note: int64_t is used to make sure no overflow takes place)
    std::vector<float> v_r, v_g, v_b;
    std::vector<float> conv_rp, conv_rn, conv_gp, conv_gn, conv_bp, conv_bn;
    std::vector<float> out;
    std::vector<float> mx;
    //for(int64_t img_idx=0; img_idx<ds_in.ntrn+ds_in.ntst; img_idx++)
    for(int32_t batch_idx=env.batch_idx.fetch_add(1);
                batch_idx*env.batch_size<env.img_count;
                batch_idx=env.batch_idx.fetch_add(1))
        for(int64_t img_idx=batch_idx*env.batch_size; img_idx<imin((batch_idx+1)*env.batch_size,env.img_count); img_idx++)
        {
            // fetch image, subtract mean
            const raw_dataset::entry &ee = ds_in.entries[img_idx];
            v_r.resize(ee.img.height*ee.img.width);
            v_g.resize(ee.img.height*ee.img.width);
            v_b.resize(ee.img.height*ee.img.width);
            for(int64_t pos0=0; pos0<ee.img.height; pos0++)
            {
                float s = 1.0f/255.0f;
                for(int64_t pos1=0; pos1<ee.img.width; pos1++)
                {
                    int r, g, b;
                    ee.img.get_rgb(pos0, pos1, &r, &g, &b);
                    v_r[pos0*ee.img.width+pos1] = r*s;
                    v_g[pos0*ee.img.width+pos1] = g*s;
                    v_b[pos0*ee.img.width+pos1] = b*s;
                }
            }

            // white balance
            white_balance(v_r, v_g, v_b);


            // scan image, compute projections onto color axes
            const filter_family &family = filters.find_family(imax(ee.img.width,ee.img.height));
            const int shift = family.size/2;
            out.clear();
            out.resize(projected_size2*nedges*ncolors_wrk,0.0f);
            float mx_colored = 1.0e-20f, mx_blackwhite = 1.0e-20f;
            for(int64_t edge_idx=0; edge_idx<nedges; edge_idx++)
            {
                // fetch filter coefficients, split into positive and negative parts
                std::vector<float> coeffs_pos(family.size*family.size,0.0f);
                std::vector<float> coeffs_neg(family.size*family.size,0.0f);
                float sum_coeff_pos = 0, sum_coeff_neg = 0;
                const float *p_coeffs = &family.coeffs[family.size*family.size*edge_idx];
                for(int64_t k=0; k<family.size*family.size; k++)
                {
                    coeffs_pos[k] = fmaxf(+p_coeffs[k],0.0f);
                    coeffs_neg[k] = fmaxf(-p_coeffs[k],0.0f);
                    sum_coeff_pos += coeffs_pos[k];
                    sum_coeff_neg += coeffs_neg[k];
                }

                // perform convolution of positive and negative parts with reg, gree, blue planes
                convolve(v_r, ee.img.height, ee.img.width, coeffs_pos, family.size, family.size, conv_rp);
                convolve(v_r, ee.img.height, ee.img.width, coeffs_neg, family.size, family.size, conv_rn);
                convolve(v_g, ee.img.height, ee.img.width, coeffs_pos, family.size, family.size, conv_gp);
                convolve(v_g, ee.img.height, ee.img.width, coeffs_neg, family.size, family.size, conv_gn);
                convolve(v_b, ee.img.height, ee.img.width, coeffs_pos, family.size, family.size, conv_bp);
                convolve(v_b, ee.img.height, ee.img.width, coeffs_neg, family.size, family.size, conv_bn);

                // apply filter to the image
                int64_t conv_height = ee.img.height-family.size+1;
                int64_t conv_width  = ee.img.width -family.size+1;
                for(int64_t pos0=0; pos0<conv_height; pos0++)
                    for(int64_t pos1=0; pos1<conv_width; pos1++)
                    {
                        int64_t src_idx = pos0*conv_width+pos1;

                        float conv_pos_r = conv_rp[src_idx];
                        float conv_pos_g = conv_gp[src_idx];
                        float conv_pos_b = conv_bp[src_idx];
                        float conv_neg_r = conv_rn[src_idx];
                        float conv_neg_g = conv_gn[src_idx];
                        float conv_neg_b = conv_bn[src_idx];
                        float conv_pos_s = conv_pos_r+conv_pos_g+conv_pos_b;
                        float conv_neg_s = conv_neg_r+conv_neg_g+conv_neg_b;

                        float proj_pos_red    = +(3*conv_pos_r-conv_pos_s)*0.5f;
                        float proj_pos_green  = +(3*conv_pos_g-conv_pos_s)*0.5f;
                        float proj_pos_blue   = +(3*conv_pos_b-conv_pos_s)*0.5f;
                        float proj_pos_yellow = -proj_pos_blue; //-(3*conv_pos_b-conv_pos_s)*0.5f;
                        float proj_pos_teal   = -proj_pos_red;  //-(3*conv_pos_r-conv_pos_s)*0.5f;
                        float proj_pos_purple = -proj_pos_green;//-(3*conv_pos_g-conv_pos_s)*0.5f;
                        float proj_pos_white  = conv_pos_s*0.333333333f;
                        float proj_pos_black  = sum_coeff_pos-conv_pos_s*0.333333333f;

                        float proj_neg_red    = +(3*conv_neg_r-conv_neg_s)*0.5f;
                        float proj_neg_green  = +(3*conv_neg_g-conv_neg_s)*0.5f;
                        float proj_neg_blue   = +(3*conv_neg_b-conv_neg_s)*0.5f;
                        float proj_neg_yellow = -proj_neg_blue; //-(3*conv_neg_b-conv_neg_s)*0.5f;
                        float proj_neg_teal   = -proj_neg_red;  //-(3*conv_neg_r-conv_neg_s)*0.5f;
                        float proj_neg_purple = -proj_neg_green;//-(3*conv_neg_g-conv_neg_s)*0.5f;
                        float proj_neg_white  = conv_neg_s*0.333333333f;
                        float proj_neg_black  = sum_coeff_neg-conv_neg_s*0.333333333f;

                        // compute activations
                        float act_red    = relu(relu(proj_pos_red   )-relu(proj_neg_red   ));
                        float act_yellow = relu(relu(proj_pos_yellow)-relu(proj_neg_yellow));
                        float act_green  = relu(relu(proj_pos_green )-relu(proj_neg_green ));
                        float act_teal   = relu(relu(proj_pos_teal  )-relu(proj_neg_teal  ));
                        float act_blue   = relu(relu(proj_pos_blue  )-relu(proj_neg_blue  ));
                        float act_purple = relu(relu(proj_pos_purple)-relu(proj_neg_purple));
                        float act_white  = relu(relu(proj_pos_white )-relu(proj_neg_white ));
                        float act_black  = relu(relu(proj_pos_black )-relu(proj_neg_black ));

                        // determine output location for storage, store
                        ASSERT(ee.img.width==ee.img.height);
                        int64_t dst0 = roundf(float(pos0+shift)/float(ee.img.height-1)*(projected_size-1));
                        int64_t dst1 = roundf(float(pos1+shift)/float(ee.img.width -1)*(projected_size-1));
                        int64_t dst  = dst0*projected_size+dst1;
                        int64_t dst_red    = 0*projected_size2*nedges+edge_idx*projected_size2+dst;
                        int64_t dst_yellow = dst_red+projected_size2*nedges;
                        int64_t dst_green  = dst_yellow+projected_size2*nedges;
                        int64_t dst_teal   = dst_green+projected_size2*nedges;
                        int64_t dst_blue   = dst_teal+projected_size2*nedges;
                        int64_t dst_purple = dst_blue+projected_size2*nedges;
                        int64_t dst_white  = dst_purple+projected_size2*nedges;
                        int64_t dst_black  = dst_white+projected_size2*nedges;
                        out[dst_red]    = fmaxf(out[dst_red],    act_red);
                        out[dst_yellow] = fmaxf(out[dst_yellow], act_yellow);
                        out[dst_green]  = fmaxf(out[dst_green],  act_green);
                        out[dst_teal]   = fmaxf(out[dst_teal],   act_teal);
                        out[dst_blue]   = fmaxf(out[dst_blue],   act_blue);
                        out[dst_purple] = fmaxf(out[dst_purple], act_purple);
                        out[dst_white]  = fmaxf(out[dst_white],  act_white);
                        out[dst_black]  = fmaxf(out[dst_black],  act_black);

                        // update maximums
                        mx_colored = fmaxf(mx_colored, act_red);
                        mx_colored = fmaxf(mx_colored, act_yellow);
                        mx_colored = fmaxf(mx_colored, act_green);
                        mx_colored = fmaxf(mx_colored, act_teal);
                        mx_colored = fmaxf(mx_colored, act_blue);
                        mx_colored = fmaxf(mx_colored, act_purple);
                        mx_blackwhite = fmaxf(mx_blackwhite, act_white);
                        mx_blackwhite = fmaxf(mx_blackwhite, act_black);
                    }
            }

            // competition, normalization
            mx.clear();
            mx.resize(projected_size2*ncolors_wrk,0.0f);
            for(int64_t color_idx=0; color_idx<ncolors_wrk; color_idx++)
            {
                float *p_mx = &mx[color_idx*projected_size2];
                for(int64_t edge_idx=0; edge_idx<nedges; edge_idx++)
                {
                    float *p_plane = &out[color_idx*projected_size2*nedges+edge_idx*projected_size2];
                    for(int64_t pos=0; pos<projected_size2; pos++)
                        p_mx[pos] = fmaxf(p_mx[pos],p_plane[pos]);
                }
            }
            for(int64_t color_idx=0; color_idx<ncolors_wrk; color_idx++)
            {
                float scale = 1.0f/(color_idx<6 ? fmaxf(mx_colored,mx_blackwhite/SNET_UCOLOR_AMPLIFY) : mx_blackwhite);
                for(int64_t edge_idx=0; edge_idx<nedges; edge_idx++)
                {
                    float *p_plane = &out[color_idx*projected_size2*nedges+edge_idx*projected_size2];
                    if( (flags&v1_flags_t::compete)==v1_flags_t::compete )
                    {
                        // competition, hardened output
                        float *p_mx = &mx[color_idx*projected_size2];
                        for(int64_t pos=0; pos<projected_size2; pos++)
                        {
                            float a = p_plane[pos];
                            float m = p_mx[pos];
                            p_plane[pos] = scale*(a<m ? 0.0f : a);
                        }
                    }
                    else if( (flags&v1_flags_t::not_compete)==v1_flags_t::not_compete )
                    {
                        // no competition, smooth output
                        for(int64_t pos=0; pos<projected_size2; pos++)
                            p_plane[pos] = scale*p_plane[pos];
                    }
                    else
                        ASSERT(0);
                }
            }
            soft_normalization(out, projected_size2);

            // offload to activities vector
            int64_t out_stride = act_out.get_stride();
            int8_t *p_dst      = act_out.get_i8()+img_idx;
            for(int64_t k=0; k<ncolors_out*nedges*projected_size2; k++)
                p_dst[k*out_stride] = posflt_to_int8(out[k]);
        }
    return NULL;
}

void apply_colored_edges_v1(const raw_dataset &ds_in,
        const filter_set &filters,
        int projected_size, v1_flags_t flags,
        std::string realm_name, bool need_black_and_white,
        activities &ds_out)
{
    // perform integrity check for filter set, determine filters count
    ASSERT(filters.rules.size()>=1);
    for(const filter_set::rule &rr : filters.rules)
    {
        ASSERT(rr.filters.cnt==filters.rules[0].filters.cnt);
        ASSERT(rr.filters.coeffs.size()==rr.filters.size*rr.filters.size*rr.filters.cnt);
    }
    int nedges      = filters.rules.front().filters.cnt;
    int ncolors_out = need_black_and_white ? 8 : 6;

    // allocation, basic initialization by zeros
    ds_out = activities(ds_in.ntrn, ds_in.ntst, storage_format::i8);
    for(int k=0; k<ncolors_out; k++)
    {
        char name[128];
        sprintf(name, "%s%02d", realm_name.c_str(), k);
        ds_out.append_realm(realm<channel>::create_grid(name, projected_size, projected_size, nedges));
    }

    // run parallel processing
    const int batch_size = 64;
    v1_parallel_colored_edges env(ds_in, filters, projected_size, need_black_and_white, flags, batch_size,
                                  ds_in.ntrn+ds_in.ntst, ds_out);
    //apply_colored_edges_v1_parallel(&env);
    parallel_invoke(apply_colored_edges_v1_parallel, &env);
}

class v1_parallel_lightness
{
public:
    const raw_dataset &ds_raw;
    const filter_set &filters;
    const int projected_size;
    const v1_flags_t flags;
    const int batch_size;
    const int img_count;
    const int downsample_min;
    const int downsample_max;
    std::atomic<int32_t> batch_idx;
    std::vector<activities> &vec_out;
    v1_parallel_lightness(const raw_dataset &rds,
                              const filter_set &fs,
                              int ps, v1_flags_t f, int bs, int ic,
                              int dsmin, int dsmax,
                              std::vector<activities> &vo):
        ds_raw(rds),filters(fs),projected_size(ps),
        flags(f),batch_size(bs),img_count(ic),batch_idx(0),downsample_min(dsmin),downsample_max(dsmax),vec_out(vo)
    {}
};

void* apply_lightness_v1_parallel(void *xptr)
{
    v1_parallel_lightness &env = *((v1_parallel_lightness*)xptr);
    const raw_dataset   &ds_in = env.ds_raw;
    const filter_set    &filters = env.filters;
    const int           projected_size = env.projected_size;
    const int64_t       projected_size2 = projected_size*projected_size;
    const v1_flags_t    flags = env.flags;
    std::vector<activities> &vec_out = env.vec_out;

    // perform integrity check for filter set, determine filters count
    ASSERT(filters.rules.size()>=1);
    for(const filter_set::rule &rr : filters.rules)
    {
        ASSERT(rr.filters.cnt==filters.rules[0].filters.cnt);
        ASSERT(rr.filters.coeffs.size()==rr.filters.size*rr.filters.size*rr.filters.cnt);
    }
    int nedges  = filters.rules.front().filters.cnt;

    // prepare internal output buffer
    downsampler dspl(nedges, projected_size, projected_size, env.downsample_max);

    // process dataset (note: int64_t is used to make sure no overflow takes place)
    std::vector<float> v_lgt, v_lgt_conv, mx, v_out;
    for(int32_t batch_idx=env.batch_idx.fetch_add(1);
                batch_idx*env.batch_size<env.img_count;
                batch_idx=env.batch_idx.fetch_add(1))
        for(int64_t img_idx=batch_idx*env.batch_size; img_idx<imin((batch_idx+1)*env.batch_size,env.img_count); img_idx++)
        {
            // fetch image
            const raw_dataset::entry &ee = ds_in.entries[img_idx];
            v_lgt.clear();
            v_lgt.resize(ee.img.height*ee.img.width,0.0f);
            v_out.clear();
            v_out.resize(projected_size2*nedges,0.0f);
            for(int64_t pos0=0; pos0<ee.img.height; pos0++)
            {
                float s = 1.0f/(3.0f*255.0f);
                for(int64_t pos1=0; pos1<ee.img.width; pos1++)
                {
                    int r, g, b;
                    ee.img.get_rgb(pos0, pos1, &r, &g, &b);
                    v_lgt[pos0*ee.img.width+pos1] = (r+g+b)*s;
                }
            }

            // scan image, compute projections onto color axes
            const filter_family &family = filters.find_family(imax(ee.img.width,ee.img.height));
            const int shift = family.size/2;
            float mx_lightness = 1.0e-20f;
            for(int64_t edge_idx=0; edge_idx<nedges; edge_idx++)
            {
                // perform convolution of filter and lightness
                const float *p_filter = &family.coeffs[family.size*family.size*edge_idx];
                std::vector<float> filter(p_filter, p_filter+family.size*family.size);
                convolve(v_lgt, ee.img.height, ee.img.width,
                         filter, family.size, family.size,
                         v_lgt_conv);

                // apply filter to the image
                ASSERT(ee.img.width==ee.img.height);
                int64_t conv_height = ee.img.height-family.size+1;
                int64_t conv_width  = ee.img.width -family.size+1;
                int64_t arr_dst0[conv_height], arr_dst1[conv_width];
                for(int64_t pos0=0; pos0<conv_height; pos0++)
                    arr_dst0[pos0] = roundf(float(pos0+shift)/float(ee.img.height-1)*(projected_size-1));
                for(int64_t pos1=0; pos1<conv_width; pos1++)
                    arr_dst1[pos1] = roundf(float(pos1+shift)/float(ee.img.width -1)*(projected_size-1));
                for(int64_t pos0=0; pos0<conv_height; pos0++)
                    for(int64_t pos1=0; pos1<conv_width; pos1++)
                    {
                        float   act_lgt = relu(v_lgt_conv[pos0*conv_width+pos1]);
                        int64_t dst     = edge_idx*projected_size2+arr_dst0[pos0]*projected_size+arr_dst1[pos1];
                        v_out[dst]   = fmaxf(v_out[dst],   act_lgt);
                        mx_lightness = fmaxf(mx_lightness, act_lgt);
                    }
            }

            // competition, normalization
            mx.clear();
            mx.resize(projected_size2,0.0f);
            for(int64_t edge_idx=0; edge_idx<nedges; edge_idx++)
            {
                float *p_mx = &mx[0];
                float *p_plane = &v_out[edge_idx*projected_size2];
                for(int64_t pos=0; pos<projected_size2; pos++)
                    p_mx[pos] = fmaxf(p_mx[pos],p_plane[pos]);
            }
            for(int64_t edge_idx=0; edge_idx<nedges; edge_idx++)
            {
                float scale = 1.0f/mx_lightness;
                float *p_plane = &v_out[edge_idx*projected_size2];
                if( (flags&v1_flags_t::compete)==v1_flags_t::compete )
                {
                    // competition, hardened output
                    float *p_mx = &mx[0];
                    for(int64_t pos=0; pos<projected_size2; pos++)
                    {
                        float a = p_plane[pos];
                        float m = p_mx[pos];
                        p_plane[pos] = scale*(a<m ? 0.0f : a);
                    }
                }
                else if( (flags&v1_flags_t::not_compete)==v1_flags_t::not_compete )
                {
                    // no competition, smooth output
                    for(int64_t pos=0; pos<projected_size2; pos++)
                        p_plane[pos] = scale*p_plane[pos];
                }
                else
                    ASSERT(0);
            }

            // downsample and offload to activities vector
            dspl.apply_downsampling(v_out);
            for(int ds_k=0, ds=1; ds<=env.downsample_max; ds*=2,ds_k++)
            {
                ASSERT(ds_k<vec_out.size());
                activities &act_out = vec_out[ds_k];
                if( ds>=env.downsample_min && ds<=env.downsample_max )
                    dspl.output_kth_layer(ds_k, act_out, img_idx, 0);
                else
                    ASSERT(act_out.nchannels==0);
            }
        }
    return NULL;
}

void apply_lightness_v1(
        const raw_dataset &ds_in,
        const filter_set &filters,
        int projected_size,
        v1_flags_t flags,
        std::string realm_name,
        int downsample_min,
        int downsample_max,
        std::vector<activities> &vec_out)
{
    // perform integrity check for filter set, determine filters count
    ASSERT(filters.rules.size()>=1);
    for(const filter_set::rule &rr : filters.rules)
    {
        ASSERT(rr.filters.cnt==filters.rules[0].filters.cnt);
        ASSERT(rr.filters.coeffs.size()==rr.filters.size*rr.filters.size*rr.filters.cnt);
    }
    int nedges  = filters.rules.front().filters.cnt;

    // allocation, basic initialization by zeros
    vec_out = std::vector<activities>();
    for(int ds=1; ds<=downsample_max; ds*=2)
    {
        ASSERT(ds==downsample_max || ds<=downsample_max/2);
        ASSERT(projected_size%ds==0);
        vec_out.push_back(activities(ds_in.ntrn, ds_in.ntst, storage_format::i8));
    }
    for(int k=0, ds=1, ps=projected_size; ds<=downsample_max; ds*=2, ps/=2, k++)
        if( ds>=downsample_min && ds<=downsample_max )
            vec_out[k].append_realm(realm<channel>::create_grid(realm_name, ps, ps, nedges));

    // run parallel processing
    const int batch_size = 64;
    v1_parallel_lightness env(ds_in, filters, projected_size, flags, batch_size, ds_in.ntrn+ds_in.ntst, downsample_min, downsample_max, vec_out);
    parallel_invoke(apply_lightness_v1_parallel, &env);
}

class v1_parallel_gray
{
public:
    const raw_dataset &ds_raw;
    const filter_set &filters;
    const int projected_size;
    const v1_flags_t flags;
    const int batch_size;
    const int img_count;
    const int downsample_min;
    const int downsample_max;
    std::atomic<int32_t> batch_idx;
    std::vector<activities> &vec_out;
    v1_parallel_gray(const raw_dataset &rds,
                              const filter_set &fs,
                              int ps, v1_flags_t f, int bs, int ic,
                              int dsmin, int dsmax,
                              std::vector<activities> &vo):
        ds_raw(rds),filters(fs),projected_size(ps),
        flags(f),batch_size(bs),img_count(ic),batch_idx(0),downsample_min(dsmin),downsample_max(dsmax),vec_out(vo)
    {}
};

void* apply_gray_v1_parallel(void *xptr)
{
    v1_parallel_gray    &env = *((v1_parallel_gray*)xptr);
    const raw_dataset   &ds_in = env.ds_raw;
    const filter_set    &filters = env.filters;
    const int           projected_size = env.projected_size;
    const int64_t       projected_size2 = projected_size*projected_size;
    const v1_flags_t    flags = env.flags;
    std::vector<activities> &vec_out = env.vec_out;

    // perform integrity check for filter set, determine filters count
    ASSERT(filters.rules.size()>=1);
    for(const filter_set::rule &rr : filters.rules)
    {
        ASSERT(rr.filters.cnt==filters.rules[0].filters.cnt);
        ASSERT(rr.filters.coeffs.size()==rr.filters.size*rr.filters.size*rr.filters.cnt);
    }
    int nedges  = filters.rules.front().filters.cnt;

    // prepare internal output buffer
    downsampler dspl(nedges, projected_size, projected_size, env.downsample_max);

    // process dataset (note: int64_t is used to make sure no overflow takes place)
    std::vector<float> v_r, v_g, v_b, v_r_conv, v_g_conv, v_b_conv, mx, v_out;
    for(int32_t batch_idx=env.batch_idx.fetch_add(1);
                batch_idx*env.batch_size<env.img_count;
                batch_idx=env.batch_idx.fetch_add(1))
        for(int64_t img_idx=batch_idx*env.batch_size; img_idx<imin((batch_idx+1)*env.batch_size,env.img_count); img_idx++)
        {
            // fetch image
            const raw_dataset::entry &ee = ds_in.entries[img_idx];
            v_r.clear();
            v_g.clear();
            v_b.clear();
            v_r.resize(ee.img.height*ee.img.width,0.0f);
            v_g.resize(ee.img.height*ee.img.width,0.0f);
            v_b.resize(ee.img.height*ee.img.width,0.0f);
            v_out.clear();
            v_out.resize(projected_size2*nedges,0.0f);
            for(int64_t pos0=0; pos0<ee.img.height; pos0++)
            {
                float s = 1.0f/255.0f;
                for(int64_t pos1=0; pos1<ee.img.width; pos1++)
                {
                    int r, g, b;
                    ee.img.get_rgb(pos0, pos1, &r, &g, &b);
                    v_r[pos0*ee.img.width+pos1] = r*s;
                    v_g[pos0*ee.img.width+pos1] = g*s;
                    v_b[pos0*ee.img.width+pos1] = b*s;
                }
            }

            // scan image, compute projections onto color axes
            const filter_family &family = filters.find_family(imax(ee.img.width,ee.img.height));
            const int shift = family.size/2;
            float mx_gray = 1.0e-20f;
            for(int64_t edge_idx=0; edge_idx<nedges; edge_idx++)
            {
                // perform convolution of filter and lightness
                const float *p_filter = &family.coeffs[family.size*family.size*edge_idx];
                std::vector<float> filter(p_filter, p_filter+family.size*family.size);
                convolve(v_r, ee.img.height, ee.img.width,
                         filter, family.size, family.size,
                         v_r_conv);
                convolve(v_g, ee.img.height, ee.img.width,
                         filter, family.size, family.size,
                         v_g_conv);
                convolve(v_b, ee.img.height, ee.img.width,
                         filter, family.size, family.size,
                         v_b_conv);

                // apply filter to the image
                ASSERT(ee.img.width==ee.img.height);
                int64_t conv_height = ee.img.height-family.size+1;
                int64_t conv_width  = ee.img.width -family.size+1;
                int64_t arr_dst0[conv_height], arr_dst1[conv_width];
                for(int64_t pos0=0; pos0<conv_height; pos0++)
                    arr_dst0[pos0] = roundf(float(pos0+shift)/float(ee.img.height-1)*(projected_size-1));
                for(int64_t pos1=0; pos1<conv_width; pos1++)
                    arr_dst1[pos1] = roundf(float(pos1+shift)/float(ee.img.width -1)*(projected_size-1));
                for(int64_t pos0=0; pos0<conv_height; pos0++)
                    for(int64_t pos1=0; pos1<conv_width; pos1++)
                    {
                        int64_t src = pos0*conv_width+pos1;
                        float   act_gry = 0;
                        act_gry = fmaxf(act_gry, fabsf(v_r_conv[src]));
                        act_gry = fmaxf(act_gry, fabsf(v_g_conv[src]));
                        act_gry = fmaxf(act_gry, fabsf(v_b_conv[src]));
                        int64_t dst = edge_idx*projected_size2+arr_dst0[pos0]*projected_size+arr_dst1[pos1];
                        v_out[dst]  = fmaxf(v_out[dst],   act_gry);
                        mx_gray     = fmaxf(mx_gray, act_gry);
                    }
            }

            // competition, normalization
            mx.clear();
            mx.resize(projected_size2,0.0f);
            for(int64_t edge_idx=0; edge_idx<nedges; edge_idx++)
            {
                float *p_mx = &mx[0];
                float *p_plane = &v_out[edge_idx*projected_size2];
                for(int64_t pos=0; pos<projected_size2; pos++)
                    p_mx[pos] = fmaxf(p_mx[pos],p_plane[pos]);
            }
            for(int64_t edge_idx=0; edge_idx<nedges; edge_idx++)
            {
                float scale = 1.0f/mx_gray;
                float *p_plane = &v_out[edge_idx*projected_size2];
                if( (flags&v1_flags_t::compete)==v1_flags_t::compete )
                {
                    // competition, hardened output
                    float *p_mx = &mx[0];
                    for(int64_t pos=0; pos<projected_size2; pos++)
                    {
                        float a = p_plane[pos];
                        float m = p_mx[pos];
                        p_plane[pos] = scale*(a<m ? 0.0f : a);
                    }
                }
                else if( (flags&v1_flags_t::not_compete)==v1_flags_t::not_compete )
                {
                    // no competition, smooth output
                    for(int64_t pos=0; pos<projected_size2; pos++)
                        p_plane[pos] = scale*p_plane[pos];
                }
                else
                    ASSERT(0);
            }

            // downsample and offload to activities vector
            dspl.apply_downsampling(v_out);
            for(int ds_k=0, ds=1; ds<=env.downsample_max; ds*=2,ds_k++)
            {
                ASSERT(ds_k<vec_out.size());
                activities &act_out = vec_out[ds_k];
                if( ds>=env.downsample_min && ds<=env.downsample_max )
                    dspl.output_kth_layer(ds_k, act_out, img_idx, 0);
                else
                    ASSERT(act_out.nchannels==0);
            }
        }
    return NULL;
}

void apply_gray_v1(
        const raw_dataset &ds_in,
        const filter_set &filters,
        int projected_size,
        v1_flags_t flags,
        std::string realm_name,
        int downsample_min,
        int downsample_max,
        std::vector<activities> &vec_out)
{
    // perform integrity check for filter set, determine filters count
    ASSERT(filters.rules.size()>=1);
    for(const filter_set::rule &rr : filters.rules)
    {
        ASSERT(rr.filters.cnt==filters.rules[0].filters.cnt);
        ASSERT(rr.filters.coeffs.size()==rr.filters.size*rr.filters.size*rr.filters.cnt);
    }
    int nedges  = filters.rules.front().filters.cnt;

    // allocation, basic initialization by zeros
    vec_out = std::vector<activities>();
    for(int ds=1; ds<=downsample_max; ds*=2)
    {
        ASSERT(ds==downsample_max || ds<=downsample_max/2);
        ASSERT(projected_size%ds==0);
        vec_out.push_back(activities(ds_in.ntrn, ds_in.ntst, storage_format::i8));
    }
    for(int k=0, ds=1, ps=projected_size; ds<=downsample_max; ds*=2, ps/=2, k++)
        if( ds>=downsample_min && ds<=downsample_max )
            vec_out[k].append_realm(realm<channel>::create_grid(realm_name, ps, ps, nedges));

    // run parallel processing
    const int batch_size = 64;
    v1_parallel_gray env(ds_in, filters, projected_size, flags, batch_size, ds_in.ntrn+ds_in.ntst, downsample_min, downsample_max, vec_out);
    parallel_invoke(apply_gray_v1_parallel, &env);
}

void apply_colorless_v1(
        const raw_dataset &ds_in,
        const filter_set &filters,
        int projected_size,
        std::string realm_name,
        activities &ds_out)
{
    // perform integrity check for filter set, determine filters count
    ASSERT(filters.rules.size()>=1);
    for(const filter_set::rule &rr : filters.rules)
    {
        ASSERT(rr.filters.cnt==filters.rules[0].filters.cnt);
        ASSERT(rr.filters.coeffs.size()==rr.filters.size*rr.filters.size*rr.filters.cnt);
    }
    int nfilters = filters.rules.front().filters.cnt;

    // allocation, basic initialization by zeros
    ds_out = activities(ds_in.ntrn, ds_in.ntst, storage_format::i8);
    ds_out.append_realm(realm<channel>::create_grid(realm_name, projected_size, projected_size, nfilters));

    // process dataset (note: int64_t is used to make sure no overflow takes place)
    for(int64_t img_idx=0; img_idx<ds_in.ntrn+ds_in.ntst; img_idx++)
    {
        // output: application of filters to the dataset and projection into fixed-size window
        std::vector<float> out(projected_size*projected_size*nfilters,0.0f);
        std::vector<float> mx(projected_size*projected_size,0.0f);
        float mx_total = 1.0e-20f;

        // process and project
        const raw_dataset::entry &ee = ds_in.entries[img_idx];
        const filter_family &family = filters.find_family(imax(ee.img.width,ee.img.height));
        std::vector<float> tmp_out;
        int shift = family.size/2;
        for(int64_t pos0=0; pos0<ee.img.height-family.size+1; pos0++)
            for(int64_t pos1=0; pos1<ee.img.width-family.size+1; pos1++)
            {
                // process
                apply_family(ee.img, pos0, pos1, family, tmp_out);

                // determine output location for storage, store
                ASSERT(ee.img.width==ee.img.height);
                int64_t dst0 = roundf(float(pos0+shift)/float(ee.img.height-1)*(projected_size-1));
                int64_t dst1 = roundf(float(pos1+shift)/float(ee.img.width -1)*(projected_size-1));
                for(int64_t k=0; k<nfilters; k++)
                {
                    int64_t dst = k*projected_size*projected_size+dst0*projected_size+dst1;
                    out[dst] = fmaxf(out[dst], tmp_out[k]);
                }
            }

        // competition, normalization
        for(int64_t k=0; k<nfilters; k++)
        {
            float *p_plane = &out[projected_size*projected_size*k];
            for(int64_t pos=0; pos<projected_size*projected_size; pos++)
            {
                float v = p_plane[pos];
                mx_total = fmaxf(mx_total, v);
                mx[pos] = fmaxf(mx[pos],v);
            }
        }
        float ss = 1.0f/mx_total;
        for(int64_t k=0; k<nfilters; k++)
        {
            float *p_plane = &out[projected_size*projected_size*k];
            for(int64_t pos=0; pos<projected_size*projected_size; pos++)
            {
                float v = p_plane[pos];
                float vm= mx[pos];
                p_plane[pos] = (v<vm ? 0.0f : v)*ss;
            }
        }

        // offload to activities vector
        int64_t out_stride = ds_out.get_stride();
        int8_t *p_dst      = ds_out.get_i8()+img_idx;
        for(int64_t k=0; k<nfilters*projected_size*projected_size; k++)
            p_dst[k*out_stride] = posflt_to_int8(out[k]);
    }
}

void downsample2(const activities &ds_in, const char *pfx, bool max_pool, activities &ds_out)
{
    const int64_t ntotal = ds_in.ntrn+ds_in.ntst;
    std::vector<int> buf(ntotal);

    ds_out = activities(ds_in.ntrn, ds_in.ntst, storage_format::i8);
    for(const realm<channel> &r_src: ds_in.realms)
    {
        // integrity checks
        ASSERT(r_src.grid_height%2==0 && r_src.grid_width%2==0);

        // append new realm
        ds_out.append_realm(realm<channel>::create_grid(pfx+r_src.name,r_src.grid_height/2,r_src.grid_width/2,r_src.grid_depth));
        realm<channel> &r_dst = ds_out.realms.back();

        // perform average-pooling or max-pooling
        for(int64_t cidx=0; cidx<r_src.grid_depth; cidx++)
        {
            for(int64_t i0=0; i0<r_src.grid_height/2; i0++)
                for(int64_t j0=0; j0<r_src.grid_width/2; j0++)
                {
                    // perform pooling to temporary buffer
                    int *p_intbuf = &buf[0];
                    memset(&buf[0], 0, buf.size()*sizeof(buf[0]));
                    for(int64_t i1=0; i1<2; i1++)
                        for(int64_t j1=0; j1<2; j1++)
                        {
                            int64_t src_cidx = r_src.get_cidx(cidx, i0*2+i1, j0*2+j1);
                            const int8_t *p_src = ds_in.get_i8(src_cidx);
                            if( max_pool )
                            {
                                for(int64_t k=0; k<ntotal; k++)
                                {
                                    int a = p_intbuf[k], s = p_src[k];
                                    p_intbuf[k] = a>s ? a : s;
                                }
                            }
                            else
                            {
                                for(int64_t k=0; k<ntotal; k++)
                                    p_intbuf[k] += p_src[k];
                            }
                        }

                    // output results
                    int64_t dst_cidx = r_dst.get_cidx(cidx, i0, j0);
                    int8_t *p_dst = ds_out.get_i8(dst_cidx);
                    if( max_pool )
                    {
                        for(int64_t k=0; k<ntotal; k++)
                            p_dst[k] = p_intbuf[k];
                    }
                    else
                    {
                        for(int64_t k=0; k<ntotal; k++)
                            p_dst[k] = int8_t(roundf(0.25f*p_intbuf[k]));
                    }
                }
        }
    }
}

void downsample2_max(const activities &ds_in, const char *pfx, activities &ds_out)
{
    downsample2(ds_in, pfx, true, ds_out);
}

void downsample2_avg(const activities &ds_in, const char *pfx, activities &ds_out)
{
    downsample2(ds_in, pfx, false, ds_out);
}

class v1_parallel_kmeans
{
public:
    const raw_dataset &ds_raw;
    const filter_set &filters;
    const int projected_size;
    const net_v1layer::realm_t kmeans_type;
    const float wht_stab;
    const float hardness;
    const float rad_param;
    const int batch_size;
    const int img_count;
    const int downsample_min;
    const int downsample_max;
    std::atomic<int32_t> batch_idx;
    std::vector<activities> &vec_out;
    v1_parallel_kmeans(const raw_dataset &rds,
                              const filter_set &fs,
                              int ps, net_v1layer::realm_t kmt, float ws, float hrd, float rp, int bs, int ic,
                              int dsmin, int dsmax,
                              std::vector<activities> &vo):
        ds_raw(rds),filters(fs),projected_size(ps),
        kmeans_type(kmt),wht_stab(ws),hardness(hrd),rad_param(rp),batch_size(bs),img_count(ic),batch_idx(0),downsample_min(dsmin),downsample_max(dsmax),vec_out(vo)
    {}
};

void* apply_kmeans_v1_parallel(void *xptr)
{
    v1_parallel_kmeans  &env = *((v1_parallel_kmeans*)xptr);
    const raw_dataset   &ds_in = env.ds_raw;
    const filter_set    &filters = env.filters;
    const int           projected_size = env.projected_size;
    const int64_t       projected_size2 = projected_size*projected_size;
    int channels_per_filter = 0;
    std::vector<activities> &vec_out = env.vec_out;
    if( env.kmeans_type==net_v1layer::realm_t::kmeans )
        channels_per_filter = 3;
    if( env.kmeans_type==net_v1layer::realm_t::kmeans_lgt )
        channels_per_filter = 1;
    if( env.kmeans_type==net_v1layer::realm_t::kmeans_gry )
        channels_per_filter = 1;
    if( env.kmeans_type==net_v1layer::realm_t::kmeans_clr )
        channels_per_filter = 3;
    if( env.kmeans_type==net_v1layer::realm_t::kmeans_clu )
        channels_per_filter = 3;
    if( env.kmeans_type==net_v1layer::realm_t::kmeans_clg )
        channels_per_filter = 3;

    // perform integrity check for filter set, determine filters count
    ASSERT(channels_per_filter>=1);
    ASSERT(filters.rules.size()>=1);
    for(const filter_set::rule &rr : filters.rules)
    {
        ASSERT(rr.filters.cnt==filters.rules[0].filters.cnt);
        ASSERT(rr.filters.coeffs.size()==channels_per_filter*rr.filters.size*rr.filters.size*rr.filters.cnt);
    }
    int nfilters = filters.rules.front().filters.cnt;

    // prepare internal output buffer
    downsampler dspl(nfilters, projected_size, projected_size, env.downsample_max);

    // process dataset (note: int64_t is used to make sure no overflow takes place)
    std::vector<float> _v_r, _v_g, _v_b, vv_out;
    activityf_vector v_shifted, xx_meanlgt, xx_varlgt, xx_scl, xx_dist, xx_dist1, xx_dist2, xx_dist3, xx_meandist, xx_c0, xx_c1;
    for(int32_t batch_idx=env.batch_idx.fetch_add(1);
                batch_idx*env.batch_size<env.img_count;
                batch_idx=env.batch_idx.fetch_add(1))
        for(int64_t img_idx=batch_idx*env.batch_size; img_idx<imin((batch_idx+1)*env.batch_size,env.img_count); img_idx++)
        {
            // fetch image
            const raw_dataset::entry &ee = ds_in.entries[img_idx];
            const filter_family &family = filters.find_family(imax(ee.img.width,ee.img.height));
            const int patch_size = family.size;
            const int pdim = patch_size*patch_size*channels_per_filter;
            const int patches0 =ee.img.height-patch_size+1, patches1 = ee.img.width-patch_size+1;
            const int sdim = patches0*patches1;

            // arrays which do not need zero-initialization
            xx_c0.resize(sdim);
            xx_c1.resize(sdim);
            xx_scl.resize(sdim);
            _v_r.resize(ee.img.height*ee.img.width);
            _v_g.resize(ee.img.height*ee.img.width);
            _v_b.resize(ee.img.height*ee.img.width);
            v_shifted.resize(pdim*sdim);

            // arrays which need zero-initialization
            xx_meanlgt.resize(sdim);
            xx_varlgt.resize(sdim);
            xx_dist.resize(sdim*nfilters);
            xx_dist1.resize(sdim*nfilters);
            xx_dist2.resize(sdim*nfilters);
            xx_dist3.resize(sdim*nfilters);
            xx_meandist.resize(sdim);
            memset(&xx_meanlgt[0],   0, xx_meanlgt.size()*sizeof(xx_meanlgt[0]));
            memset(&xx_varlgt[0],    0, xx_varlgt.size()*sizeof(xx_varlgt[0]));
            memset(&xx_dist[0],      0, xx_dist.size()*sizeof(xx_dist[0]));
            memset(&xx_dist1[0],     0, xx_dist1.size()*sizeof(xx_dist1[0]));
            memset(&xx_dist2[0],     0, xx_dist2.size()*sizeof(xx_dist2[0]));
            memset(&xx_dist3[0],     0, xx_dist3.size()*sizeof(xx_dist3[0]));
            memset(&xx_meandist[0],  0, xx_meandist.size()*sizeof(xx_meandist[0]));

            // fetch RGB
            for(int64_t pos0=0; pos0<ee.img.height; pos0++)
            {
                float s = 1.0f/255.0f;
                for(int64_t pos1=0; pos1<ee.img.width; pos1++)
                {
                    int ir, ig, ib;
                    ee.img.get_rgb(pos0, pos1, &ir, &ig, &ib);
                    float fr = ir*s, fg = ig*s, fb = ib*s;
                    if( env.kmeans_type==net_v1layer::realm_t::kmeans )
                    {
                        // do nothing
                    }
                    else if( env.kmeans_type==net_v1layer::realm_t::kmeans_lgt || env.kmeans_type==net_v1layer::realm_t::kmeans_gry )
                    {
                        float lgt = (fr+fg+fb)*0.333333333333333333f;
                        fr = lgt;
                        fg = lgt;
                        fb = lgt;
                    }
                    else if( env.kmeans_type==net_v1layer::realm_t::kmeans_clr ||
                             env.kmeans_type==net_v1layer::realm_t::kmeans_clu ||
                             env.kmeans_type==net_v1layer::realm_t::kmeans_clg )
                    {
                        float lgt = (fr+fg+fb)*0.333333333333333333f;
                        fr -= lgt;
                        fg -= lgt;
                        fb -= lgt;
                    }
                    else
                        ASSERT(0);
                    _v_r[pos0*ee.img.width+pos1] = fr;
                    _v_g[pos0*ee.img.width+pos1] = fg;
                    _v_b[pos0*ee.img.width+pos1] = fb;
                }
            }

            // split into patches
            for(int p0=0; p0<patches0; p0++)
                for(int p1=0; p1<patches1; p1++)
                {
                    // fetch patch
                    float p_patch[pdim];
                    if( channels_per_filter==3 )
                    {
                        // fetch 3-channel patch
                        for(int offs0=0; offs0<patch_size; offs0++)
                            for(int offs1=0; offs1<patch_size; offs1++)
                            {
                                int src = (p0+offs0)*ee.img.width+(p1+offs1);
                                int dst = 3*(offs0*patch_size+offs1);
                                p_patch[dst+0] = _v_r[src];
                                p_patch[dst+1] = _v_g[src];
                                p_patch[dst+2] = _v_b[src];
                            }
                    }
                    else if( channels_per_filter==1 )
                    {
                        // fetch 1-channel patch (R, G and B are expected to be equal, we use R)
                        for(int offs0=0; offs0<patch_size; offs0++)
                            for(int offs1=0; offs1<patch_size; offs1++)
                            {
                                int src = (p0+offs0)*ee.img.width+(p1+offs1);
                                int dst = offs0*patch_size+offs1;
                                p_patch[dst] = _v_r[src];
                            }
                    }
                    else
                        ASSERT(0);

                    // offload to v_shifted
                    float *p_dst = &v_shifted[p0*patches1+p1];
                    for(int k=0; k<pdim; k++)
                        p_dst[k*sdim] = p_patch[k];
                }

            // contrast-normalize patches
            for(int j=0; j<pdim; j++)
                fadd_simd(&xx_meanlgt[0], &v_shifted[j*sdim], sdim, 1.0f/pdim);
            for(int j=0; j<pdim; j++)
            {
                fadd_simd(&v_shifted[j*sdim], &xx_meanlgt[0], sdim, -1.0f);
                fadd_diffc2(&xx_varlgt[0], &v_shifted[j*sdim], 0.0f, sdim);
            }
            for(int k=0; k<sdim; k++)
                xx_scl[k] = 1.0f/sqrtf(xx_varlgt[k]/pdim+env.wht_stab);
            for(int j=0; j<pdim; j++)
                fmul_inplace(&v_shifted[j*sdim], &xx_scl[0], sdim);

            // compute squared distances and mean squared distances
            for(int k=0; k<nfilters; k++)
            {
                float *p_dst  = &xx_dist[k*sdim];
                if( env.kmeans_type==net_v1layer::realm_t::kmeans_gry )
                {
                    float *p_dst1 = &xx_dist1[k*sdim];
                    for(int j=0; j<pdim; j++)
                        fadd_diffc2(p_dst,  &v_shifted[j*sdim], +family.coeffs[k*pdim+j], sdim);
                    for(int j=0; j<pdim; j++)
                        fadd_diffc2(p_dst1, &v_shifted[j*sdim], -family.coeffs[k*pdim+j], sdim);
                    fmerge_min(p_dst, p_dst1, sdim);
                }
                else if( env.kmeans_type==net_v1layer::realm_t::kmeans_clg )
                {
                    float *p_dst1 = &xx_dist1[k*sdim];
                    float *p_dst2 = &xx_dist2[k*sdim];
                    float *p_dst3 = &xx_dist3[k*sdim];
                    float *pa[4] = {p_dst, p_dst1, p_dst2, p_dst3};
                    float patch[pdim];
                    for(int j=0; j<pdim; j++)
                        patch[j] = family.coeffs[k*pdim+j];
                    for(int nrot=0; nrot<4; nrot++)
                    {
                        for(int j=0; j<pdim; j++)
                            fadd_diffc2(pa[nrot],  &v_shifted[j*sdim], patch[j], sdim);
                        rotate90(patch, patch_size, channels_per_filter);
                    }
                    fmerge_min(p_dst, p_dst1, sdim);
                    fmerge_min(p_dst, p_dst2, sdim);
                    fmerge_min(p_dst, p_dst3, sdim);
                }
                else
                {
                    for(int j=0; j<pdim; j++)
                        fadd_diffc2(p_dst, &v_shifted[j*sdim], family.coeffs[k*pdim+j], sdim);
                }

                /*for(int j=0; j<pdim; j++)
                {
                    fadd_diffc2(p_dst, &v_shifted[j*sdim], family.coeffs[k*pdim+j], sdim);
                    if( env.kmeans_type==net_v1layer::realm_t::kmeans_gry )
                        fadd_diffc2(p_dstn, &v_shifted[j*sdim], -family.coeffs[k*pdim+j], sdim);
                }
                if( env.kmeans_type==net_v1layer::realm_t::kmeans_gry )
                    fmerge_min(p_dst, p_dstn, sdim);*/
                fadd_simd(&xx_meandist[0], p_dst, sdim, 1.0f/nfilters);
            }

            // compute activities
            ASSERT(env.hardness==0);
            for(int i=0; i<sdim; i++)
            {
                xx_scl[i] = -1.0f/(xx_meandist[i]*env.rad_param+0.0001f);
                xx_c0[i]  =  0.0f;
                xx_c1[i]  =  1.0f;
            }
            for(int k=0; k<nfilters; k++)
            {
                fmul_add_inplace(&xx_dist[k*sdim], &xx_scl[0], &xx_c1[0], sdim);
                fmerge_max(&xx_dist[k*sdim], &xx_c0[0], sdim);
            }

            // project to window
            ASSERT(ee.img.width==ee.img.height);
            const int shift = patch_size/2;
            vv_out.resize(projected_size2*nfilters);
            memset(&vv_out[0], 0, sizeof(vv_out[0])*vv_out.size());
            int64_t arr_dst0[patches0], arr_dst1[patches1];
            for(int64_t pos0=0; pos0<patches0; pos0++)
                arr_dst0[pos0] = roundf(float(pos0+shift)/float(ee.img.height-1)*(projected_size-1));
            for(int64_t pos1=0; pos1<patches1; pos1++)
                arr_dst1[pos1] = roundf(float(pos1+shift)/float(ee.img.width -1)*(projected_size-1));
            for(int64_t k=0; k<nfilters; k++)
                for(int64_t pos0=0; pos0<patches0; pos0++)
                {
                    int64_t _dst = k*projected_size2+arr_dst0[pos0]*projected_size;
                    int64_t _src = k*sdim+pos0*patches1;
                    for(int64_t pos1=0; pos1<patches1; pos1++)
                    {
                        int64_t dst = _dst+arr_dst1[pos1];
                        int64_t src = _src+pos1;
                        vv_out[dst] = fmaxf(xx_dist[src],vv_out[dst]);
                    }
                }

            // downsample and offload to activities vector
            dspl.apply_downsampling(vv_out);
            for(int ds_k=0, ds=1; ds<=env.downsample_max; ds*=2,ds_k++)
            {
                ASSERT(ds_k<vec_out.size());
                activities &act_out = vec_out[ds_k];
                if( ds>=env.downsample_min && ds<=env.downsample_max )
                    dspl.output_kth_layer(ds_k, act_out, img_idx, 0);
                else
                    ASSERT(act_out.nchannels==0);
            }
        }
    return NULL;
}

class v1_parallel_lxg
{
public:
    const raw_dataset &ds_raw;
    const filter_set &filters;
    const int expected_size;
    const net_v1layer::realm_t realm_type;
    const float lxg_alpha;
    const float lxg_beta;
    const float filter_gain;
    const int batch_size;
    const int img_count;
    const int downsample_min;
    const int downsample_max;
    std::atomic<int32_t> batch_idx;
    std::vector<activities> &vec_out;
    v1_parallel_lxg(const raw_dataset &rds,
                    const filter_set &fs,
                    int es, net_v1layer::realm_t rt, float lga, float lgb, float fg, int bs, int ic,
                    int dsmin, int dsmax,
                    std::vector<activities> &vo):
        ds_raw(rds),filters(fs),expected_size(es),
        realm_type(rt), lxg_alpha(lga),lxg_beta(lgb),filter_gain(fg),batch_size(bs),img_count(ic),batch_idx(0),
        downsample_min(dsmin),downsample_max(dsmax), vec_out(vo)
    {}
};

void* apply_kmeans_v1_parallel_xg(void *xptr)
{
    v1_parallel_lxg     &env = *((v1_parallel_lxg*)xptr);
    const raw_dataset   &ds_in = env.ds_raw;
    const filter_set    &filters = env.filters;
    const int           expected_size  = env.expected_size;
    const int64_t       expected_size2 = expected_size*expected_size;
    std::vector<activities> &vec_out = env.vec_out;

    // perform integrity check for filter set, determine filters count
    ASSERT(filters.rules.size()>=1);
    for(const filter_set::rule &rr : filters.rules)
    {
        ASSERT(rr.filters.cnt==filters.rules[0].filters.cnt);
        ASSERT(rr.filters.coeffs.size()==rr.filters.size*rr.filters.size*rr.filters.cnt);
    }
    int nkernels = filters.rules.front().filters.cnt;

    // prepare internal output buffer
    downsampler dspl_lgt(2*nkernels, expected_size, expected_size, env.downsample_max);
    downsampler dspl_gry(  nkernels, expected_size, expected_size, env.downsample_max);

    // process dataset (note: int64_t is used to make sure no overflow takes place)
    std::vector<float> _v_r, _v_g, _v_b, wrk_lgt, wrk_gry, padded_lgt, padded_gry;
    activityf_vector mean_lgt, mean_gry;
    activityf_vector dotk_r, dotk_g, dotk_b;
    activityf_vector im2col_r, im2col_g, im2col_b, im2col_l;
    for(int32_t batch_idx=env.batch_idx.fetch_add(1);
                batch_idx*env.batch_size<env.img_count;
                batch_idx=env.batch_idx.fetch_add(1))
        for(int64_t img_idx=batch_idx*env.batch_size; img_idx<imin((batch_idx+1)*env.batch_size,env.img_count); img_idx++)
        {
            // fetch image
            const raw_dataset::entry &ee = ds_in.entries[img_idx];
            ASSERT(ee.img.width==expected_size);
            ASSERT(ee.img.height==expected_size);
            const filter_family family = filters.find_family(imax(ee.img.width,ee.img.height));//.normalize_sum1();
            const int patch_size = family.size;
            const int patch_size2 = patch_size*patch_size;
            const int patches0 =ee.img.height-patch_size+1, patches1 = ee.img.width-patch_size+1;
            const int sdim = patches0*patches1;

            // arrays which do not need zero-initialization
            _v_r.resize(ee.img.height*ee.img.width);
            _v_g.resize(ee.img.height*ee.img.width);
            _v_b.resize(ee.img.height*ee.img.width);
            im2col_r.resize(patch_size2*sdim);
            im2col_g.resize(patch_size2*sdim);
            im2col_b.resize(patch_size2*sdim);
            im2col_l.resize(patch_size2*sdim);
            wrk_lgt.resize(2*nkernels*sdim);
            wrk_gry.resize(nkernels*sdim);

            // arrays which need zero-initialization
            dotk_r.resize(nkernels*sdim);
            dotk_g.resize(nkernels*sdim);
            dotk_b.resize(nkernels*sdim);
            mean_lgt.resize(sdim);
            mean_gry.resize(sdim);
            memset(&dotk_r[0],       0, dotk_r.size()*sizeof(dotk_r[0]));
            memset(&dotk_g[0],       0, dotk_g.size()*sizeof(dotk_g[0]));
            memset(&dotk_b[0],       0, dotk_b.size()*sizeof(dotk_b[0]));
            memset(&mean_lgt[0],     0, mean_lgt.size()*sizeof(mean_lgt[0]));
            memset(&mean_gry[0],     0, mean_gry.size()*sizeof(mean_gry[0]));

            // fetch RGB
            for(int64_t pos0=0; pos0<ee.img.height; pos0++)
            {
                float s = 1.0f/255.0f;
                for(int64_t pos1=0; pos1<ee.img.width; pos1++)
                {
                    int ir, ig, ib;
                    ee.img.get_rgb(pos0, pos1, &ir, &ig, &ib);
                    _v_r[pos0*ee.img.width+pos1] = ir*s;
                    _v_g[pos0*ee.img.width+pos1] = ig*s;
                    _v_b[pos0*ee.img.width+pos1] = ib*s;
                }
            }

            // split into patches
            for(int p0=0; p0<patches0; p0++)
                for(int p1=0; p1<patches1; p1++)
                {
                    float *p_dst_r = &im2col_r[p0*patches1+p1];
                    float *p_dst_g = &im2col_g[p0*patches1+p1];
                    float *p_dst_b = &im2col_b[p0*patches1+p1];
                    float *p_dst_l = &im2col_l[p0*patches1+p1];
                    for(int offs0=0; offs0<patch_size; offs0++)
                        for(int offs1=0; offs1<patch_size; offs1++)
                        {
                            int src = (p0+offs0)*ee.img.width+(p1+offs1);
                            int dst = offs0*patch_size+offs1;
                            p_dst_r[dst*sdim] = _v_r[src];
                            p_dst_g[dst*sdim] = _v_g[src];
                            p_dst_b[dst*sdim] = _v_b[src];
                            p_dst_l[dst*sdim] = 0.3333333f*(_v_r[src]+_v_g[src]+_v_b[src]);
                        }
                }

            // compute dot products of r/g/b with kernels
            for(int k=0; k<nkernels; k++)
                for(int j=0; j<patch_size2; j++)
                {
                    fadd_simd(&dotk_r[k*sdim], &im2col_r[j*sdim], sdim, family.coeffs[k*patch_size2+j]);
                    fadd_simd(&dotk_g[k*sdim], &im2col_g[j*sdim], sdim, family.coeffs[k*patch_size2+j]);
                    fadd_simd(&dotk_b[k*sdim], &im2col_b[j*sdim], sdim, family.coeffs[k*patch_size2+j]);
                }

            // produce activities sampled at working grid
            for(int k=0; k<nkernels; k++)
            {
                const float *p_src_r = &dotk_r[k*sdim];
                const float *p_src_g = &dotk_g[k*sdim];
                const float *p_src_b = &dotk_b[k*sdim];
                float *p_dst_lgt0 = &wrk_lgt[(2*k+0)*sdim];
                float *p_dst_lgt1 = &wrk_lgt[(2*k+1)*sdim];
                float *p_dst_gry  = &wrk_gry[k*sdim];
                for(int i=0; i<sdim; i++)
                {
                    //float w_lgt = env.lxg_alpha*patch_nrm_l[i];
                    float v_lgt = 0.333333333f*(p_src_r[i]+p_src_g[i]+p_src_b[i]);
                    float v_lgt_0 = fmaxf(+v_lgt, 0.0f);
                    float v_lgt_1 = fmaxf(-v_lgt, 0.0f);
                    p_dst_lgt0[i] = v_lgt_0;
                    p_dst_lgt1[i] = v_lgt_1;
                    //scl_lgt = fmaxf(scl_lgt, v_lgt_0);
                    //scl_lgt = fmaxf(scl_lgt, v_lgt_1);

                    float v_gry = fmax3f(fabsf(p_src_r[i]), fabsf(p_src_g[i]), fabsf(p_src_b[i]));
                    p_dst_gry[i]  = v_gry;
                    //scl_gry = fmaxf(scl_gry, v_gry);
                }
                fadd_simd(&mean_lgt[0], p_dst_lgt0, sdim, 0.5f/nkernels);
                fadd_simd(&mean_lgt[0], p_dst_lgt1, sdim, 0.5f/nkernels);
                fadd_simd(&mean_gry[0], p_dst_gry,  sdim, 1.0f/nkernels);
            }
            float mx_lgt=1.0e-20f, mx_gry=1.0e-20f;
            for(int k=0; k<nkernels; k++)
            {
                const float lxg_alpha = env.lxg_alpha;
                const float lxg_beta  = env.lxg_beta;
                const float *p_mean_lgt = &mean_lgt[0];
                const float *p_mean_gry = &mean_gry[0];
                float *p_dst_lgt0 = &wrk_lgt[(2*k+0)*sdim];
                float *p_dst_lgt1 = &wrk_lgt[(2*k+1)*sdim];
                float *p_dst_gry  = &wrk_gry[k*sdim];
                for(int i=0; i<sdim; i++)
                {
                    float v0 = fmaxf(p_dst_lgt0[i]-lxg_beta*p_mean_lgt[i], 0.0f);
                    float v1 = fmaxf(p_dst_lgt1[i]-lxg_beta*p_mean_lgt[i], 0.0f);
                    float vg = fmaxf( p_dst_gry[i]-lxg_alpha*p_mean_gry[i], 0.0f);
                    p_dst_lgt0[i] = v0;
                    p_dst_lgt1[i] = v1;
                    p_dst_gry[i]  = vg;
                    mx_lgt = fmaxf(mx_lgt, v0);
                    mx_lgt = fmaxf(mx_lgt, v1);
                    mx_gry = fmaxf(mx_gry, vg);
                }
            }
            if( env.filter_gain==0.0 )
            {
                mx_lgt = 1.0f/mx_lgt;
                mx_gry = 1.0f/mx_gry;
                fmul_inplace(&wrk_lgt[0], mx_lgt, wrk_lgt.size());
                fmul_inplace(&wrk_gry[0], mx_gry, wrk_gry.size());
            }
            else
            {
                fmul_inplace(&wrk_lgt[0], env.filter_gain, wrk_lgt.size());
                fmul_inplace(&wrk_gry[0], env.filter_gain, wrk_gry.size());
                fmin_inplace(&wrk_lgt[0], 1.0f, wrk_lgt.size());
                fmin_inplace(&wrk_gry[0], 1.0f, wrk_gry.size());
            }

            // project to window
            const int shift = patch_size/2;
            padded_lgt.resize(expected_size2*nkernels*2);
            padded_gry.resize(expected_size2*nkernels);
            memset(&padded_lgt[0], 0, sizeof(padded_lgt[0])*padded_lgt.size());
            memset(&padded_gry[0], 0, sizeof(padded_gry[0])*padded_gry.size());
            for(int64_t k=0; k<2*nkernels; k++)
                for(int64_t pos0=0; pos0<patches0; pos0++)
                {
                    int64_t _dst_lgt = k*expected_size2+(pos0+shift)*expected_size;
                    int64_t _src_lgt = k*sdim+pos0*patches1;
                    for(int64_t pos1=0; pos1<patches1; pos1++)
                    {
                        int64_t dst_lgt = _dst_lgt+(pos1+shift);
                        int64_t src_lgt = _src_lgt+pos1;
                        padded_lgt[dst_lgt] = fmaxf(wrk_lgt[src_lgt],padded_lgt[dst_lgt]);
                    }
                }
            for(int64_t k=0; k<nkernels; k++)
                for(int64_t pos0=0; pos0<patches0; pos0++)
                {
                    int64_t _dst_gry = k*expected_size2+(pos0+shift)*expected_size;
                    int64_t _src_gry = k*sdim+pos0*patches1;
                    for(int64_t pos1=0; pos1<patches1; pos1++)
                    {
                        int64_t dst_gry = _dst_gry+(pos1+shift);
                        int64_t src_gry = _src_gry+pos1;
                        padded_gry[dst_gry] = fmaxf(wrk_gry[src_gry],padded_gry[dst_gry]);
                    }
                }

            // downsample and offload to activities vector
            dspl_lgt.apply_downsampling(padded_lgt);
            dspl_gry.apply_downsampling(padded_gry);
            for(int ds_k=0, ds=1; ds<=env.downsample_max; ds*=2,ds_k++)
            {
                ASSERT(ds_k<vec_out.size());
                activities &act_out = vec_out[ds_k];
                if( ds>=env.downsample_min && ds<=env.downsample_max )
                {
                    // gry realm
                    ASSERT(act_out.realms.size()==1 || act_out.realms.size()==2);
                    ASSERT(act_out.realms[0].grid_depth==nkernels);
                    const realm<channel> &rr_gry = act_out.realms[0];
                    dspl_gry.output_kth_layer(ds_k, act_out, img_idx, rr_gry, false);

                    // lgt realm, if present
                    if( env.realm_type==net_v1layer::realm_t::kmeans_lxg )
                    {
                        ASSERT(act_out.realms.size()==2);
                        ASSERT(act_out.realms[1].grid_depth==2*nkernels);
                        const realm<channel> &rr_lgt = act_out.realms[1];
                        dspl_lgt.output_kth_layer(ds_k, act_out, img_idx, rr_lgt, false);
                    }
                }
                else
                    ASSERT(act_out.nchannels==0);
            }
        }
    return NULL;
}

void apply_kmeans_v1(const raw_dataset &ds_in,
                     const filter_set &filters,
                     int projected_size,
                     net_v1layer::realm_t kmeans_type,
                     const float wht_stab,
                     const float hardness,
                     const float rad_param,
                     std::string realm_name,
                     int downsample_min,
                     int downsample_max,
                     std::vector<activities> &vec_out)
{
    int channels_per_filter = 0;
    if( kmeans_type==net_v1layer::realm_t::kmeans )
        channels_per_filter = 3;
    if( kmeans_type==net_v1layer::realm_t::kmeans_lgt )
        channels_per_filter = 1;
    if( kmeans_type==net_v1layer::realm_t::kmeans_gry )
        channels_per_filter = 1;
    if( kmeans_type==net_v1layer::realm_t::kmeans_clr )
        channels_per_filter = 3;
    if( kmeans_type==net_v1layer::realm_t::kmeans_clu )
        channels_per_filter = 3;
    if( kmeans_type==net_v1layer::realm_t::kmeans_clg )
        channels_per_filter = 3;

    // perform integrity check for filter set, determine filters count
    ASSERT(kmeans_type==net_v1layer::realm_t::kmeans ||
           kmeans_type==net_v1layer::realm_t::kmeans_lgt ||
           kmeans_type==net_v1layer::realm_t::kmeans_gry ||
           kmeans_type==net_v1layer::realm_t::kmeans_clr ||
           kmeans_type==net_v1layer::realm_t::kmeans_clu ||
           kmeans_type==net_v1layer::realm_t::kmeans_clg);
    ASSERT(channels_per_filter>=1);
    ASSERT(filters.rules.size()>=1);
    for(const filter_set::rule &rr : filters.rules)
    {
        ASSERT(rr.filters.cnt==filters.rules[0].filters.cnt);
        ASSERT(rr.filters.coeffs.size()==channels_per_filter*rr.filters.size*rr.filters.size*rr.filters.cnt);
    }
    int nfilters = filters.rules.front().filters.cnt;

    // allocation, basic initialization by zeros
    vec_out = std::vector<activities>();
    for(int ds=1; ds<=downsample_max; ds*=2)
    {
        ASSERT(ds==downsample_max || ds<=downsample_max/2);
        ASSERT(projected_size%ds==0);
        vec_out.push_back(activities(ds_in.ntrn, ds_in.ntst, storage_format::i8));
    }
    for(int k=0, ds=1, ps=projected_size; ds<=downsample_max; ds*=2, ps/=2, k++)
        if( ds>=downsample_min && ds<=downsample_max )
            vec_out[k].append_realm(realm<channel>::create_grid(realm_name, ps, ps, nfilters));

    // run parallel processing
    const int batch_size = 64;
    v1_parallel_kmeans env(ds_in, filters, projected_size, kmeans_type, wht_stab, hardness, rad_param, batch_size, ds_in.ntrn+ds_in.ntst, downsample_min, downsample_max, vec_out);
    //int dbg = dtickcount();
    parallel_invoke(apply_kmeans_v1_parallel, &env);
    //dbg = dtickcount()-dbg;
    //printf("done in %d ms\n", dbg);
    //fflush(stdout);
    //ASSERT(0);
}

void apply_kmeans_v1_xg(const raw_dataset &ds_in,
                         const filter_set &filters,
                         int projected_size,
                         const net_v1layer::realm_t rtype,
                         const float lxg_alpha,
                         const float lxg_beta,
                         const float filter_gain,
                         std::string realm_name,
                         int downsample_min,
                         int downsample_max,
                         storage_format fmt,
                         std::vector<activities> &vec_out)
{
    // perform integrity check for filter set, determine filters count
    ASSERT(rtype==net_v1layer::realm_t::kmeans_lxg || rtype==net_v1layer::realm_t::kmeans_xxg);
    ASSERT(filters.rules.size()>=1);
    for(const filter_set::rule &rr : filters.rules)
    {
        ASSERT(rr.filters.cnt==filters.rules[0].filters.cnt);
        ASSERT(rr.filters.coeffs.size()==rr.filters.size*rr.filters.size*rr.filters.cnt);
    }
    int nkernels = filters.rules.front().filters.cnt;

    // allocation, basic initialization by zeros
    vec_out = std::vector<activities>();
    for(int ds=1; ds<=downsample_max; ds*=2)
    {
        ASSERT(ds==downsample_max || ds<=downsample_max/2);
        ASSERT(projected_size%ds==0);
        vec_out.push_back(activities(ds_in.ntrn, ds_in.ntst, fmt));
    }
    for(int k=0, ds=1, ps=projected_size; ds<=downsample_max; ds*=2, ps/=2, k++)
        if( ds>=downsample_min && ds<=downsample_max )
        {
            vec_out[k].append_realm(realm<channel>::create_grid(realm_name+".gry", ps, ps, nkernels));
            if( rtype==net_v1layer::realm_t::kmeans_lxg )
                vec_out[k].append_realm(realm<channel>::create_grid(realm_name+".lgt", ps, ps, 2*nkernels));
        }

    // run parallel processing
    const int batch_size = 64;
    v1_parallel_lxg env(ds_in, filters, projected_size, rtype, lxg_alpha, lxg_beta, filter_gain, batch_size, ds_in.ntrn+ds_in.ntst, downsample_min, downsample_max, vec_out);
    parallel_invoke(apply_kmeans_v1_parallel_xg, &env);
}

class v1_parallel_clx
{
public:
    const raw_dataset &ds_raw;
    const filter_set &filters;
    const int expected_size;
    const net_v1layer::realm_t realm_type;
    const float clx_alpha;
    const int batch_size;
    const int img_count;
    const int downsample_min;
    const int downsample_max;
    std::atomic<int32_t> batch_idx;
    std::vector<activities> &vec_out;
    v1_parallel_clx(const raw_dataset &rds,
                    const filter_set &fs,
                    int es, net_v1layer::realm_t rt, float lga, int bs, int ic,
                    int dsmin, int dsmax,
                    std::vector<activities> &vo):
        ds_raw(rds),filters(fs),expected_size(es),
        realm_type(rt), clx_alpha(lga),batch_size(bs),img_count(ic),batch_idx(0),
        downsample_min(dsmin),downsample_max(dsmax), vec_out(vo)
    {}
};

void* apply_kmeans_v1_parallel_xc(void *xptr)
{
    v1_parallel_clx     &env = *((v1_parallel_clx*)xptr);
    const raw_dataset   &ds_in = env.ds_raw;
    const filter_set    &filters = env.filters;
    const int           expected_size  = env.expected_size;
    const int64_t       expected_size2 = expected_size*expected_size;
    std::vector<activities> &vec_out = env.vec_out;

    // perform integrity check for filter set, determine filters count
    ASSERT(filters.rules.size()>=1);
    for(const filter_set::rule &rr : filters.rules)
    {
        ASSERT(rr.filters.cnt==filters.rules[0].filters.cnt);
        ASSERT(rr.filters.coeffs.size()==3*rr.filters.size*rr.filters.size*rr.filters.cnt);
    }
    int nkernels = filters.rules.front().filters.cnt;

    // prepare internal output buffer
    downsampler dspl_clx(nkernels, expected_size, expected_size, env.downsample_max);

    // process dataset (note: int64_t is used to make sure no overflow takes place)
    std::vector<float> _v_r, _v_g, _v_b, wrk_clx, padded_clx;
    activityf_vector mean_clx;
    activityf_vector dot_kern;
    activityf_vector im2col_r, im2col_g, im2col_b;
    for(int32_t batch_idx=env.batch_idx.fetch_add(1);
                batch_idx*env.batch_size<env.img_count;
                batch_idx=env.batch_idx.fetch_add(1))
        for(int64_t img_idx=batch_idx*env.batch_size; img_idx<imin((batch_idx+1)*env.batch_size,env.img_count); img_idx++)
        {
            // fetch image
            const raw_dataset::entry &ee = ds_in.entries[img_idx];
            ASSERT(ee.img.width==expected_size);
            ASSERT(ee.img.height==expected_size);
            const filter_family &family = filters.find_family(imax(ee.img.width,ee.img.height));
            const int patch_size = family.size;
            const int patch_size2 = patch_size*patch_size;
            const int patches0 =ee.img.height-patch_size+1, patches1 = ee.img.width-patch_size+1;
            const int sdim = patches0*patches1;

            // arrays which do not need zero-initialization
            _v_r.resize(ee.img.height*ee.img.width);
            _v_g.resize(ee.img.height*ee.img.width);
            _v_b.resize(ee.img.height*ee.img.width);
            im2col_r.resize(patch_size2*sdim);
            im2col_g.resize(patch_size2*sdim);
            im2col_b.resize(patch_size2*sdim);
            wrk_clx.resize(nkernels*sdim);

            // arrays which need zero-initialization
            dot_kern.resize(nkernels*sdim);
            mean_clx.resize(sdim);
            memset(&dot_kern[0],     0, dot_kern.size()*sizeof(dot_kern[0]));
            memset(&mean_clx[0],     0, mean_clx.size()*sizeof(mean_clx[0]));

            // fetch RGB
            for(int64_t pos0=0; pos0<ee.img.height; pos0++)
            {
                float s = 1.0f/255.0f;
                for(int64_t pos1=0; pos1<ee.img.width; pos1++)
                {
                    int ir, ig, ib;
                    ee.img.get_rgb(pos0, pos1, &ir, &ig, &ib);
                    _v_r[pos0*ee.img.width+pos1] = ir*s;
                    _v_g[pos0*ee.img.width+pos1] = ig*s;
                    _v_b[pos0*ee.img.width+pos1] = ib*s;
                }
            }

            // split into patches
            for(int p0=0; p0<patches0; p0++)
                for(int p1=0; p1<patches1; p1++)
                {
                    float *p_dst_r = &im2col_r[p0*patches1+p1];
                    float *p_dst_g = &im2col_g[p0*patches1+p1];
                    float *p_dst_b = &im2col_b[p0*patches1+p1];
                    for(int offs0=0; offs0<patch_size; offs0++)
                        for(int offs1=0; offs1<patch_size; offs1++)
                        {
                            int src = (p0+offs0)*ee.img.width+(p1+offs1);
                            int dst = offs0*patch_size+offs1;
                            p_dst_r[dst*sdim] = _v_r[src];
                            p_dst_g[dst*sdim] = _v_g[src];
                            p_dst_b[dst*sdim] = _v_b[src];
                        }
                }

            // compute dot products of r/g/b with kernels
            for(int k=0; k<nkernels; k++)
                for(int j=0; j<patch_size2; j++)
                {
                    fadd_simd(&dot_kern[k*sdim], &im2col_r[j*sdim], sdim, family.coeffs[3*(k*patch_size2+j)+0]);
                    fadd_simd(&dot_kern[k*sdim], &im2col_g[j*sdim], sdim, family.coeffs[3*(k*patch_size2+j)+1]);
                    fadd_simd(&dot_kern[k*sdim], &im2col_b[j*sdim], sdim, family.coeffs[3*(k*patch_size2+j)+2]);
                }

            // produce activities sampled at working grid
            for(int k=0; k<nkernels; k++)
            {
                const float *p_src = &dot_kern[k*sdim];
                float *p_dst_clx   = &wrk_clx[k*sdim];
                for(int i=0; i<sdim; i++)
                {
                    float v_clx = fmaxf(p_src[i], 0.0f);
                    p_dst_clx[i] = v_clx;
                }
                fadd_simd(&mean_clx[0], p_dst_clx,  sdim, 1.0f/nkernels);
            }
            float scl_clx=1.0e-20f;
            for(int k=0; k<nkernels; k++)
            {
                const float clx_alpha   = env.clx_alpha;
                const float *p_mean_clx = &mean_clx[0];
                float *p_dst_clx        = &wrk_clx[k*sdim];
                for(int i=0; i<sdim; i++)
                {
                    float vc = fmaxf(p_dst_clx[i]-clx_alpha*p_mean_clx[i], 0.0f);
                    p_dst_clx[i] = vc;
                    scl_clx = fmaxf(scl_clx, vc);
                }
            }
            scl_clx = 1.0f/scl_clx;
            fmul_inplace(&wrk_clx[0], scl_clx, wrk_clx.size());

            // project to window
            const int shift = patch_size/2;
            padded_clx.resize(expected_size2*nkernels);
            memset(&padded_clx[0], 0, sizeof(padded_clx[0])*padded_clx.size());
            for(int64_t k=0; k<nkernels; k++)
                for(int64_t pos0=0; pos0<patches0; pos0++)
                {
                    int64_t _dst_clx = k*expected_size2+(pos0+shift)*expected_size;
                    int64_t _src_clx = k*sdim+pos0*patches1;
                    for(int64_t pos1=0; pos1<patches1; pos1++)
                    {
                        int64_t dst_clx = _dst_clx+(pos1+shift);
                        int64_t src_clx = _src_clx+pos1;
                        padded_clx[dst_clx] = fmaxf(wrk_clx[src_clx],padded_clx[dst_clx]);
                    }
                }

            // downsample and offload to activities vector
            dspl_clx.apply_downsampling(padded_clx);
            for(int ds_k=0, ds=1; ds<=env.downsample_max; ds*=2,ds_k++)
            {
                ASSERT(ds_k<vec_out.size());
                activities &act_out = vec_out[ds_k];
                if( ds>=env.downsample_min && ds<=env.downsample_max )
                {
                    // cx1 realm
                    ASSERT(act_out.realms.size()==1);
                    ASSERT(act_out.realms[0].grid_depth==nkernels);
                    const realm<channel> &rr_clx = act_out.realms[0];
                    dspl_clx.output_kth_layer(ds_k, act_out, img_idx, rr_clx, false);
                }
                else
                    ASSERT(act_out.nchannels==0);
            }
        }
    return NULL;
}

void apply_kmeans_v1_xc(
        const raw_dataset &ds_in,
        const filter_set &filters,
        int expected_size,
        const net_v1layer::realm_t rtype,
        const float clx_alpha,
        std::string realm_name,
        int downsample_min,
        int downsample_max,
        storage_format fmt,
        std::vector<activities> &vec_out)
{
    // perform integrity check for filter set, determine filters count
    ASSERT(rtype==net_v1layer::realm_t::kmeans_clx1);
    ASSERT(filters.rules.size()>=1);
    for(const filter_set::rule &rr : filters.rules)
    {
        ASSERT(rr.filters.cnt==filters.rules[0].filters.cnt);
        ASSERT(rr.filters.coeffs.size()==3*rr.filters.size*rr.filters.size*rr.filters.cnt);
    }
    int nkernels = filters.rules.front().filters.cnt;

    // allocation, basic initialization by zeros
    vec_out = std::vector<activities>();
    for(int ds=1; ds<=downsample_max; ds*=2)
    {
        ASSERT(ds==downsample_max || ds<=downsample_max/2);
        ASSERT(expected_size%ds==0);
        vec_out.push_back(activities(ds_in.ntrn, ds_in.ntst, fmt));
    }
    for(int k=0, ds=1, ps=expected_size; ds<=downsample_max; ds*=2, ps/=2, k++)
        if( ds>=downsample_min && ds<=downsample_max )
            vec_out[k].append_realm(realm<channel>::create_grid(realm_name+".cx1", ps, ps, nkernels));

    // run parallel processing
    const int batch_size = 64;
    v1_parallel_clx env(ds_in, filters, expected_size, rtype, clx_alpha, batch_size, ds_in.ntrn+ds_in.ntst, downsample_min, downsample_max, vec_out);
    parallel_invoke(apply_kmeans_v1_parallel_xc, &env);
}

net_v1layer get_v1basic4x8_lxg(
        int img_width,
        const int scale_idx,
        const std::string layer_name,
        const int downsample_lo, const int downsample_hi,
        const v1_settings s,
        const float filter_gain)
{
    ASSERT(img_width>=1);
    ASSERT(scale_idx>=0);
    ASSERT(filter_gain>=0);
    net_v1layer out;
    net_v1layer::v1_realm v1_rlm;
    img_width = iDivUp(img_width,1UL<<scale_idx);
    v1_rlm.realm_name = "S["+std::to_string(scale_idx)+"]"+layer_name;
    v1_rlm.realm_type = net_v1layer::realm_t::kmeans_lxg;
    v1_rlm.filters.rules.resize(1);
    v1_rlm.filters.rules[0].chsize0 = img_width;
    v1_rlm.filters.rules[0].chsize1 = img_width+1;
    v1_rlm.filters.rules[0].filters = generate_e_colorless_micro(4).normalize_sum2nz();
    v1_rlm.scale_idx      = scale_idx;
    v1_rlm.projected_size = img_width;
    v1_rlm.downsample_min = downsample_lo;
    v1_rlm.downsample_max = downsample_hi;
    v1_rlm.lxg_alpha_param = s.lxg_alpha_param;
    v1_rlm.lxg_beta_param = s.lxg_beta_param;
    v1_rlm.filter_gain = filter_gain;
    out.append_realm(v1_rlm);
    return out;
}

net_v1layer get_v1basic4x12_lxg(
        int img_width,
        const int scale_idx,
        const std::string layer_name,
        const int downsample_lo, const int downsample_hi,
        const v1_settings s,
        const float filter_gain)
{
    ASSERT(img_width>=1);
    ASSERT(scale_idx>=0);
    ASSERT(filter_gain>=0);
    net_v1layer out;
    net_v1layer::v1_realm v1_rlm;
    img_width = iDivUp(img_width,1UL<<scale_idx);
    v1_rlm.realm_name = "S["+std::to_string(scale_idx)+"]"+layer_name;
    v1_rlm.realm_type = net_v1layer::realm_t::kmeans_lxg;
    v1_rlm.filters.rules.resize(1);
    v1_rlm.filters.rules[0].chsize0 = img_width;
    v1_rlm.filters.rules[0].chsize1 = img_width+1;
    v1_rlm.filters.rules[0].filters = generate_ea_colorless_micro(4).normalize_sum2nz();
    v1_rlm.scale_idx      = scale_idx;
    v1_rlm.projected_size = img_width;
    v1_rlm.downsample_min = downsample_lo;
    v1_rlm.downsample_max = downsample_hi;
    v1_rlm.lxg_alpha_param = s.lxg_alpha_param;
    v1_rlm.lxg_beta_param = s.lxg_beta_param;
    v1_rlm.filter_gain = filter_gain;
    out.append_realm(v1_rlm);
    return out;
}

net_v1layer get_v1basic4x14_lxg(
        int img_width,
        const int scale_idx,
        const std::string layer_name,
        const int downsample_lo, const int downsample_hi,
        const v1_settings s,
        const float filter_gain)
{
    ASSERT(img_width>=1);
    ASSERT(scale_idx>=0);
    ASSERT(filter_gain>=0);
    net_v1layer out;
    net_v1layer::v1_realm v1_rlm;
    img_width = iDivUp(img_width,1UL<<scale_idx);
    v1_rlm.realm_name = "S["+std::to_string(scale_idx)+"]"+layer_name;
    v1_rlm.realm_type = net_v1layer::realm_t::kmeans_lxg;
    v1_rlm.filters.rules.resize(1);
    v1_rlm.filters.rules[0].chsize0 = img_width;
    v1_rlm.filters.rules[0].chsize1 = img_width+1;
    v1_rlm.filters.rules[0].filters = generate_eab_colorless_micro(4).normalize_sum2nz();
    v1_rlm.scale_idx      = scale_idx;
    v1_rlm.projected_size = img_width;
    v1_rlm.downsample_min = downsample_lo;
    v1_rlm.downsample_max = downsample_hi;
    v1_rlm.lxg_alpha_param = s.lxg_alpha_param;
    v1_rlm.lxg_beta_param = s.lxg_beta_param;
    v1_rlm.filter_gain = filter_gain;
    out.append_realm(v1_rlm);
    return out;
}

net_v1layer get_v1km_xg(
        const raw_dataset &ds,
        const net_v1layer::realm_t rtype,
        const int scale_idx,
        const std::string layer_name,
        const std::string rep_suffix,
        const int patch_size, const int nkernels,
        const int downsample_lo, const int downsample_hi,
        const v1_settings s)
{
    ASSERT(rtype==net_v1layer::realm_t::kmeans_lxg || rtype==net_v1layer::realm_t::kmeans_xxg);
    ASSERT(scale_idx>=0);
    ASSERT(ds.entries.size()>=1);
    for(const raw_dataset::entry &e: ds.entries)
        ASSERT(e.img.width==ds.entries[0].img.width && e.img.height==ds.entries[0].img.height);
    const int img_width = iDivUp(ds.entries[0].img.width,1UL<<scale_idx);
    net_v1layer out;
    net_v1layer::v1_realm v1_rlm;
    v1_rlm.realm_name = "S["+std::to_string(scale_idx)+"]"+layer_name;
    v1_rlm.realm_type = rtype;
    v1_rlm.filters.rules.resize(1);
    v1_rlm.filters.rules[0].chsize0 = img_width;
    v1_rlm.filters.rules[0].chsize1 = img_width+1;
    v1_rlm.filters.rules[0].filters = train_kmeans_filters_lxg(ds, scale_idx, patch_size,
                                                               s.kmeans_trn_size, nkernels, s, rep_suffix.c_str());
    v1_rlm.scale_idx      = scale_idx;
    v1_rlm.projected_size = img_width;
    v1_rlm.downsample_min = downsample_lo;
    v1_rlm.downsample_max = downsample_hi;
    v1_rlm.lxg_alpha_param = s.lxg_alpha_param;
    v1_rlm.lxg_beta_param = s.lxg_beta_param;
    out.append_realm(v1_rlm);
    return out;
}

net_v1layer get_v1km_clr(const raw_dataset &ds,
                         const int scale_idx,
                         const std::string layer_name,
                         const std::string rep_suffix,
                         const int patch_size, const int nfilters,
                         const int downsample_lo, const int downsample_hi,
                         const v1_settings s)
{
    ASSERT(scale_idx>=0);
    ASSERT(ds.entries.size()>=1);
    for(const raw_dataset::entry &e: ds.entries)
        ASSERT(e.img.width==ds.entries[0].img.width && e.img.height==ds.entries[0].img.height);
    const int img_width = iDivUp(ds.entries[0].img.width,1UL<<scale_idx);
    net_v1layer out;
    net_v1layer::v1_realm km_med;
    km_med.realm_name = "S["+std::to_string(scale_idx)+"]"+layer_name;
    km_med.realm_type = net_v1layer::realm_t::kmeans_clx1;
    km_med.filters.rules.resize(1);
    km_med.filters.rules[0].chsize0 = img_width;
    km_med.filters.rules[0].chsize1 = img_width+1;
    km_med.filters.rules[0].filters = train_kmeans_filters(ds, scale_idx, net_v1layer::realm_t::kmeans_clr, patch_size,
                                                           s.kmeans_trn_size, nfilters, s, rep_suffix.c_str());
    km_med.scale_idx      = scale_idx;
    km_med.projected_size = img_width;
    km_med.downsample_min = downsample_lo;
    km_med.downsample_max = downsample_hi;
    km_med.clx_alpha_param = s.clx_alpha_param;
    out.append_realm(km_med);
    return out;
}

net_v1layer get_v1c_lxg(const raw_dataset &ds,
                        std::tuple<int,int> kernels,
                        int kextra,
                        bool just_top_level)
{
    net_v1layer v1layer;

    //
    v1_settings s1_sml, s1_xxx;
#ifndef CTRENG_KMEANS_TRN_SIZE
    s1_sml.kmeans_trn_size = 100000;
#else
    s1_sml.kmeans_trn_size = CTRENG_KMEANS_TRN_SIZE;
#endif

    //
    const int dslo[2] = {just_top_level ? 4 : 2, just_top_level ? 2 : 1};
    const int dshi[2] = {                     4,                      2};
    const int dsxx[2] = { 4, 2};

    //
    for(int scale_idx=0; scale_idx<=1; scale_idx++)
    {
        // sml layer
        if( std::get<0>(kernels)>0 )
            v1layer.append_layer(get_v1km_xg(ds,
                                             net_v1layer::realm_t::kmeans_lxg,
                                             scale_idx,
                                             "kx_sml",                             // layer name
                                             "kxsml_l"+std::to_string(scale_idx),  // rep suffix
                                             4,                                    // patch size
                                             std::get<0>(kernels),                     // nkernels
                                             dslo[scale_idx], dshi[scale_idx],     // downsample lo/hi
                                             s1_sml));

        // med layer
        if( std::get<1>(kernels)>0 )
            v1layer.append_layer(get_v1km_xg(ds,
                                             net_v1layer::realm_t::kmeans_lxg,
                                             scale_idx,
                                             "kx_med",                             // layer name
                                             "kxmed_l"+std::to_string(scale_idx),  // rep suffix
                                             6,                                    // patch size
                                             std::get<1>(kernels),                     // nkernels
                                             dslo[scale_idx], dshi[scale_idx],     // downsample lo/hi
                                             s1_sml));

        // mgx layer
        if( kextra>0 )
            v1layer.append_layer(get_v1km_xg(ds,
                                             net_v1layer::realm_t::kmeans_xxg,
                                             scale_idx,
                                             "kx_mgx",                             // layer name
                                             "kxmgx_l"+std::to_string(scale_idx),  // rep suffix
                                             6,                                    // patch size
                                             kextra,                               // nkernels
                                             dsxx[scale_idx], dsxx[scale_idx],     // downsample lo/hi
                                             s1_xxx));
    }
    return v1layer;
}

net_v1layer get_v1c_clr(const raw_dataset &ds,
                        std::tuple<int,int> kernels,
                        bool just_top_level)
{
    net_v1layer v1layer;

    //
    v1_settings s1_sml, s1_xxx;
#ifndef CTRENG_KMEANS_TRN_SIZE
    s1_sml.kmeans_trn_size = 300000;
    s1_xxx.kmeans_trn_size = 300000;
#else
    s1_sml.kmeans_trn_size = CTRENG_KMEANS_TRN_SIZE;
    s1_xxx.kmeans_trn_size = CTRENG_KMEANS_TRN_SIZE;
#endif

    //
    const int dslo[2] = {just_top_level ? 4 : 2, just_top_level ? 2 : 1};
    const int dshi[2] = {                     4,                      2};
    const int dsxx[2] = { 4, 2};

    //
    for(int scale_idx=0; scale_idx<=1; scale_idx++)
    {
        // sml layer
        if( std::get<0>(kernels)>0 )
            v1layer.append_layer(get_v1km_clr(ds,
                                              scale_idx,
                                              "clr_sml",                             // layer name
                                              "clrsml_l"+std::to_string(scale_idx),  // rep suffix
                                              4,                                    // patch size
                                              std::get<0>(kernels),                     // nkernels
                                              dslo[scale_idx], dshi[scale_idx],     // downsample lo/hi
                                              s1_sml));

        // med layer
        if( std::get<1>(kernels)>0 )
            v1layer.append_layer(get_v1km_clr(ds,
                                              scale_idx,
                                              "clr_med",                             // layer name
                                              "clrmed_l"+std::to_string(scale_idx),  // rep suffix
                                              6,                                    // patch size
                                              std::get<1>(kernels),                     // nkernels
                                              dslo[scale_idx], dshi[scale_idx],     // downsample lo/hi
                                              s1_sml));
    }
    return v1layer;
}

net_v1layer get_v1_w32p4_lxg(const raw_dataset &ds,
                             std::tuple<int,int> kernels,
                             bool just_8x8_level)
{
    net_v1layer v1layer;

    //
    // Integrity check
    //
    for(const raw_dataset::entry &e: ds.entries)
        ASSERT(e.img.width==32 && e.img.height==32);


    //
    v1_settings s1;
#ifndef CTRENG_KMEANS_TRN_SIZE
    s1.kmeans_trn_size = 200000;
#else
    s1.kmeans_trn_size = CTRENG_KMEANS_TRN_SIZE;
#endif

    //
    // S0 layer
    //
    if( std::get<0>(kernels)>0 )
        v1layer.append_layer(get_v1km_xg(
                                 ds,
                                 net_v1layer::realm_t::kmeans_lxg,
                                 0,                                     // scale_idx
                                 "kx_sml",                              // layer name
                                 "kxsml_l0",                            // rep suffix
                                 4,                                     // patch size
                                 std::get<0>(kernels),                  // nkernels
                                 just_8x8_level ? 4 : 2, 4,             // downsample lo/hi
                                 s1));

    //
    // S1 layer
    //
    if( std::get<1>(kernels)>0 )
        v1layer.append_layer(get_v1km_xg(
                                 ds,
                                 net_v1layer::realm_t::kmeans_lxg,
                                 1,                                     // scale_idx
                                 "kx_sml",                              // layer name
                                 "kxsml_l1",                            // rep suffix
                                 4,                                     // patch size
                                 std::get<1>(kernels),                  // nkernels
                                 just_8x8_level ? 2 : 1, 2,             // downsample lo/hi
                                 s1));

    return v1layer;
}

net_v1layer get_v1_wnnp4_bas8lxg(int w, float filter_gain, int scale_hi, bool just_top_level)
{
    net_v1layer v1layer;

    //
    // Integrity check
    //
    ASSERT(w>=1);
    ASSERT(scale_hi==0 || scale_hi==1);


    //
    v1_settings s1;

    //
    // S0 layer
    //
    if( scale_hi>=0 )
        v1layer.append_layer(get_v1basic4x8_lxg(
                                 w,
                                 0,                                     // scale_idx
                                 "b4_sml",                              // layer name
                                 just_top_level ? 4 : 2, 4,             // downsample lo/hi
                                 s1,
                                 filter_gain));

    //
    // S1 layer
    //
    if( scale_hi>=1 )
        v1layer.append_layer(get_v1basic4x8_lxg(
                                 w,
                                 1,                                     // scale_idx
                                 "b4_sml",                              // layer name
                                 just_top_level ? 2 : 1, 2,             // downsample lo/hi
                                 s1,
                                 filter_gain));

    return v1layer;
}

net_v1layer get_v1_wnnp4_bas12lxg(int w, float filter_gain, int scale_hi, bool just_top_level)
{
    net_v1layer v1layer;

    //
    // Integrity check
    //
    ASSERT(w>=1);
    ASSERT(scale_hi==0 || scale_hi==1);


    //
    v1_settings s1;

    //
    // S0 layer
    //
    if( scale_hi>=0 )
        v1layer.append_layer(get_v1basic4x12_lxg(
                                 w,
                                 0,                                     // scale_idx
                                 "b4_sml",                              // layer name
                                 just_top_level ? 4 : 2, 4,             // downsample lo/hi
                                 s1,
                                 filter_gain));

    //
    // S1 layer
    //
    if( scale_hi>=1 )
        v1layer.append_layer(get_v1basic4x12_lxg(
                                 w,
                                 1,                                     // scale_idx
                                 "b4_sml",                              // layer name
                                 just_top_level ? 2 : 1, 2,             // downsample lo/hi
                                 s1,
                                 filter_gain));

    return v1layer;
}

net_v1layer get_v1_wnnp4_bas14lxg(int w, float filter_gain, int scale_hi, bool just_top_level)
{
    net_v1layer v1layer;

    //
    // Integrity check
    //
    ASSERT(w>=1);
    ASSERT(scale_hi==0 || scale_hi==1);


    //
    v1_settings s1;

    //
    // S0 layer
    //
    if( scale_hi>=0 )
        v1layer.append_layer(get_v1basic4x14_lxg(
                                 w,
                                 0,                                     // scale_idx
                                 "b4_sml",                              // layer name
                                 just_top_level ? 4 : 2, 4,             // downsample lo/hi
                                 s1,
                                 filter_gain));

    //
    // S1 layer
    //
    if( scale_hi>=1 )
        v1layer.append_layer(get_v1basic4x14_lxg(
                                 w,
                                 1,                                     // scale_idx
                                 "b4_sml",                              // layer name
                                 just_top_level ? 2 : 1, 2,             // downsample lo/hi
                                 s1,
                                 filter_gain));

    return v1layer;
}

net_v1layer get_v1_w32p4_clx(const raw_dataset &ds,
                             std::tuple<int,int> kernels,
                             bool just_8x8_level)
{
    net_v1layer v1layer;

    //
    // Integrity check
    //
    for(const raw_dataset::entry &e: ds.entries)
        ASSERT(e.img.width==32 && e.img.height==32);


    //
    v1_settings s1;
#ifndef CTRENG_KMEANS_TRN_SIZE
    s1.kmeans_trn_size = 200000;
#else
    s1.kmeans_trn_size = CTRENG_KMEANS_TRN_SIZE;
#endif

    //
    // S0 layer
    //
    if( std::get<0>(kernels)>0 )
        v1layer.append_layer(get_v1km_clr(
                                 ds,
                                 0,                                     // scale_idx,
                                 "clr_sml",                             // layer name
                                 "clrsml_l0",                           // rep suffix
                                 4,                                     // patch size
                                 std::get<0>(kernels),                  // nkernels
                                 just_8x8_level ? 4 : 2, 4,             // downsample lo/hi
                                 s1));

    //
    // S1 layer
    //
    if( std::get<1>(kernels)>0 )
        v1layer.append_layer(get_v1km_clr(
                                 ds,
                                 1,                                     // scale_idx,
                                 "clr_sml",                             // layer name
                                 "clrsml_l1",                           // rep suffix
                                 4,                                     // patch size
                                 std::get<1>(kernels),                  // nkernels
                                 just_8x8_level ? 2 : 1, 2,             // downsample lo/hi
                                 s1));

    return v1layer;
}

net_v1layer get_v1_w32p6_lxg(const raw_dataset &ds,
                             std::tuple<int,int> kernels,
                             bool just_8x8_level)
{
    net_v1layer v1layer;

    //
    // Integrity check
    //
    for(const raw_dataset::entry &e: ds.entries)
        ASSERT(e.img.width==32 && e.img.height==32);


    //
    v1_settings s1;
#ifndef CTRENG_KMEANS_TRN_SIZE
    s1.kmeans_trn_size = 200000;
#else
    s1.kmeans_trn_size = CTRENG_KMEANS_TRN_SIZE;
#endif

    //
    // S0 layer
    //
    if( std::get<0>(kernels)>0 )
        v1layer.append_layer(get_v1km_xg(
                                 ds,
                                 net_v1layer::realm_t::kmeans_lxg,
                                 0,                                     // scale_idx
                                 "kx_med",                              // layer name
                                 "kxmed_l0",                            // rep suffix
                                 6,                                     // patch size
                                 std::get<0>(kernels),                  // nkernels
                                 just_8x8_level ? 4 : 2, 4,             // downsample lo/hi
                                 s1));

    //
    // S1 layer
    //
    if( std::get<1>(kernels)>0 )
        v1layer.append_layer(get_v1km_xg(
                                 ds,
                                 net_v1layer::realm_t::kmeans_lxg,
                                 1,                                     // scale_idx
                                 "kx_med",                              // layer name
                                 "kxmed_l1",                            // rep suffix
                                 6,                                     // patch size
                                 std::get<1>(kernels),                  // nkernels
                                 just_8x8_level ? 2 : 1, 2,             // downsample lo/hi
                                 s1));

    return v1layer;
}

net_v1layer get_v1_w32p6_clx(const raw_dataset &ds,
                             std::tuple<int,int> kernels,
                             bool just_8x8_level)
{
    net_v1layer v1layer;

    //
    // Integrity check
    //
    for(const raw_dataset::entry &e: ds.entries)
        ASSERT(e.img.width==32 && e.img.height==32);


    //
    v1_settings s1;
#ifndef CTRENG_KMEANS_TRN_SIZE
    s1.kmeans_trn_size = 200000;
#else
    s1.kmeans_trn_size = CTRENG_KMEANS_TRN_SIZE;
#endif

    //
    // S0 layer
    //
    if( std::get<0>(kernels)>0 )
        v1layer.append_layer(get_v1km_clr(
                                 ds,
                                 0,                                     // scale_idx,
                                 "clr_med",                             // layer name
                                 "clrmed_l0",                           // rep suffix
                                 6,                                     // patch size
                                 std::get<0>(kernels),                  // nkernels
                                 just_8x8_level ? 4 : 2, 4,             // downsample lo/hi
                                 s1));

    //
    // S1 layer
    //
    if( std::get<1>(kernels)>0 )
        v1layer.append_layer(get_v1km_clr(
                                 ds,
                                 1,                                     // scale_idx,
                                 "clr_med",                             // layer name
                                 "clrmed_l1",                           // rep suffix
                                 6,                                     // patch size
                                 std::get<1>(kernels),                  // nkernels
                                 just_8x8_level ? 2 : 1, 2,             // downsample lo/hi
                                 s1));

    return v1layer;
}

void v1_analyze(const net_v1layer &layer)
{
    for(const net_v1layer::v1_realm &rr: layer.realms)
    {
        if( rr.realm_type!=net_v1layer::realm_t::kmeans_lxg )
            continue;
        ASSERT(rr.filters.rules.size()==1);
        const filter_family &ff = rr.filters.rules[0].filters;
        ASSERT(ff.size%2==0);
        ASSERT(ff.cnt>=1);
        int fsize  = ff.size;
        int fsize2 = fsize/2;

        printf("Analyzing realm %s, count is %d, filter size is %d^2\n", rr.realm_name.c_str(), ff.cnt, fsize);
        std::vector<std::vector<float>> all_chunks;
        for(int k=0; k<ff.cnt; k++)
        {
            const float *coeffs = &ff.coeffs[k*ff.size*ff.size];
            for(int k0=0; k0<2; k0++)
                for(int k1=0; k1<2; k1++)
                {
                    std::vector<float> chunk;
                    for(int i=0; i<fsize2; i++)
                        for(int j=0; j<fsize2; j++)
                            chunk.push_back(coeffs[(k0*fsize2+i)*fsize+(k1*fsize2+j)]);
                    all_chunks.push_back(chunk);
                }
        }
        printf("> all_chunks = %d items\n", int(all_chunks.size()));

        /*std::vector<std::vector<float>> unique_chunks;
        for(std::vector<float> chunk: all_chunks)
        {
            const int n = fsize2*fsize2;
            float v;

            // normalize chunk, skip nearly zero ones
            v = sqrtf(vnrm2(&chunk[0], n));
            if( v<1.0e-3f )
                continue;
            fmul_inplace(&chunk[0], 1/v, n);

            // find best approximation
            float best_cos = -1.0e20f;
            const std::vector<float> *p_best_chunk = NULL;
            for(const std::vector<float> &cand_chunk: unique_chunks)
            {
                float cand_cos = fdot_simd(&chunk[0], &cand_chunk[0], n);
                if( cand_cos>best_cos && cand_cos>0.75f )
                {
                    best_cos = cand_cos;
                    p_best_chunk = &cand_chunk;
                }
            }
            if( p_best_chunk==NULL )
                unique_chunks.push_back(chunk);
        }*/
        std::vector<std::vector<float>> unique_chunks;
        const int   MAX_CNT = ff.cnt;
        const float MAX_ERR = 0.01f;
        std::vector<float> max_errors(all_chunks.size(),1.0e20f);
        std::vector<int>   ref_indexes(all_chunks.size(),-1);
        std::vector<float> ref_scales(all_chunks.size(),0.0f);
        float mx_err = 0;
        for(;;)
        {
            const int n = fsize2*fsize2;

            // find chunk with minimum error
            int mx_idx = 0;
            mx_err = max_errors[mx_idx];
            for(int i=0; i<all_chunks.size(); i++)
                if( max_errors[i]>max_errors[mx_idx] )
                {
                    mx_idx = i;
                    mx_err = max_errors[mx_idx];
                }
            if( unique_chunks.size()>=MAX_CNT || mx_err<MAX_ERR )
                break;
            std::vector<float> worst_chunk = all_chunks[mx_idx];

            // add normalized worst chunk to unique list
            {
                std::vector<float> nrm_chunk = worst_chunk;
                fmul_inplace(&nrm_chunk[0], 1/sqrtf(vnrm2(&nrm_chunk[0], n)), n);
                unique_chunks.push_back(nrm_chunk);
            }

            // update errors
            for(int k=0; k<all_chunks.size(); k++)
            {
                float best_err  = max_errors[k];
                int   ref_idx   = ref_indexes[k];
                float ref_scale = ref_scales[k];

                // find best single-vector approximation
                for(int uidx=0; uidx<unique_chunks.size(); uidx++)
                {
                    const std::vector<float> &cand0 = unique_chunks[uidx];
                    std::vector<float> wrk_chunk = all_chunks[k];
                    float cand_scale = fdot_simd(&cand0[0], &wrk_chunk[0], n);
                    fadd_simd(&wrk_chunk[0], &cand0[0], n, -cand_scale);
                    float cand_err = 0;
                    for(int i=0; i<n; i++)
                        cand_err = fmaxf(cand_err, fabsf(wrk_chunk[i]));
                    if( cand_err<best_err )
                    {
                        best_err  = cand_err;
                        ref_idx   = uidx;
                        ref_scale = cand_scale;
                    }
                }

                /*// find best two-vector approximation
                for(std::vector<float> cand0: unique_chunks)
                    for(std::vector<float> cand1: unique_chunks)
                    {
                        // orthogonalize cand1 wrt cand0
                        fadd_simd(&cand1[0], &cand0[0], n, -fdot_simd(&cand0[0], &cand1[0], n));
                        float c1nrm = sqrtf(vnrm2(&cand1[0], n));
                        fmul_inplace(&cand1[0], c1nrm>1.0e-5f ? 1.0f/c1nrm : 0.0f, n);

                        // compute residual from fitting by cand0/cand1
                        std::vector<float> wrk_chunk = all_chunks[k];
                        fadd_simd(&wrk_chunk[0], &cand0[0], n, -fdot_simd(&cand0[0], &wrk_chunk[0], n));
                        fadd_simd(&wrk_chunk[0], &cand1[0], n, -fdot_simd(&cand1[0], &wrk_chunk[0], n));
                        float cand_err = 0;
                        for(int i=0; i<n; i++)
                            cand_err = fmaxf(cand_err, fabsf(wrk_chunk[i]));
                        if( cand_err<best_err )
                        {
                            best_err = cand_err;
                        }
                    }*/

                // update
                max_errors[k]  = best_err;
                ref_indexes[k] = ref_idx;
                ref_scales[k]  = ref_scale;
            }
        }


        float global_mx = 0;
        for(const auto &c: all_chunks)
            for(float v: c)
                global_mx = fmaxf(global_mx, fabsf(v));
        printf("> uni_chunks  = %d items\n", int(unique_chunks.size()));
        printf("> max_err     = %.3f\n",     mx_err/global_mx);

        float max_restored_error = 0.0f;
        for(int k=0; k<ff.cnt; k++)
        {
            const float *coeffs = &ff.coeffs[k*ff.size*ff.size];
            int base_cidx = 4*k;
            std::vector<float> raw_coeffs(coeffs, coeffs+fsize*fsize);
            std::vector<float> wrk_coeffs(fsize*fsize,-1.0e20f);
            for(int k0=0; k0<2; k0++)
                for(int k1=0; k1<2; k1++)
                {
                    ASSERT(ref_indexes[base_cidx]>=0);
                    float scl = ref_scales[base_cidx];
                    std::vector<float> ref_chunk = unique_chunks[ref_indexes[base_cidx]];
                    for(int i=0; i<fsize2; i++)
                        for(int j=0; j<fsize2; j++)
                            wrk_coeffs[(k0*fsize2+i)*fsize+(k1*fsize2+j)] = scl*ref_chunk[i*fsize2+j];
                    base_cidx++;
                }
            for(int i=0; i<fsize*fsize; i++)
                max_restored_error = fmaxf(max_restored_error, fabsf(raw_coeffs[i]-wrk_coeffs[i]));
        }
        printf("> max_rest_err = %.3f\n", max_restored_error/global_mx);
    }

}

}

