/*************************************************************************
StrongNet/Contour Engine training framework.
Copyright (c) 2020 Sergey Bochkanov (ALGLIB project).

>>> SOURCE LICENSE >>>
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation (www.fsf.org); either version 2 of the 
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

A copy of the GNU General Public License is available at
http://www.fsf.org/licensing/licenses
>>> END OF LICENSE >>>
*************************************************************************/
#include "snetcommon.h"

#include <unistd.h>
#include <sys/time.h>
#include "3rdparty/alglib/linalg.h"
#include <math.h>
#include <map>
#include <set>
#include <numeric>
#include <algorithm>
#include <random>
#include <emmintrin.h>
#include <smmintrin.h>
#ifdef SNET_AVX2
#include <immintrin.h>
#endif
extern "C"
{
#include <jpeglib.h>
}
#include "3rdparty/alglib/interpolation.h"

#define ALIGN_FLOATS_BY 8
#define ALIGN_ACTIVITIES_BY 32

#define BYTES_PER_SIMD 16
#define FLOATS_PER_SIMD 4

namespace  strongnet {

template<class T>
T* coalesce_ptr(T *a, T *b)
{
    return a!=NULL ? a : b;
}

template<class T>
const T* coalesce_const_ptr(const T *a, const T *b)
{
    return a!=NULL ? a : b;
}

void* ae_align(void *ptr, size_t alignment)
{
    char *result = (char*)ptr;
    if( (result-(char*)0)%alignment!=0 )
        result += alignment - (result-(char*)0)%alignment;
    return result;
}

void* aligned_malloc(size_t size, size_t alignment)
{
    char *result = NULL;

    if( size==0 )
        return NULL;

    /* allocate */
    if( alignment<=1 )
    {
        /* no alignment, just call alloc */
        void *block;
        void **p; ;
        block = malloc(sizeof(void*)+size);
        if( block==NULL )
            return NULL;
        p = (void**)block;
        *p = block;
        result = (char*)((char*)block+sizeof(void*));
    }
    else
    {
        /* align */
        void *block;
        block = malloc(alignment-1+sizeof(void*)+size);
        if( block==NULL )
            return NULL;
        result = (char*)block+sizeof(void*);
        result = (char*)ae_align(result, alignment);
        *((void**)(result-sizeof(void*))) = block;
    }

    /* return */
    return (void*)result;
}

void aligned_free(void *block)
{
    void *p;
    if( block==NULL )
        return;
    p = *((void**)((char*)block-sizeof(void*)));
    free(p);
}

int dtickcount()
{
    struct timeval now;
    int64_t r, v;
    gettimeofday(&now, NULL);
    v = now.tv_sec;
    r = v*1000;
    v = now.tv_usec/1000;
    r = r+v;
    return (int)r;
}

template<>
void serialize<int8_t>(FILE *f, const std::vector<int8_t,aligned_allocator<int8_t> > *vec)
{
    serialize64(f, vec->size());
    serialize_long_blob(f, &((*vec)[0]), vec->size()*sizeof(int8_t));
}

template<>
void serialize<float>(FILE *f, const std::vector<float,aligned_allocator<float> > *vec)
{
    serialize64(f, vec->size());
    serialize_long_blob(f, &((*vec)[0]), vec->size()*sizeof(float));
}

template<>
void serialize<float>(FILE *f, const std::vector<float> *vec)
{
    serialize64(f, vec->size());
    serialize_long_blob(f, &((*vec)[0]), vec->size()*sizeof(float));
}

template<>
void serialize<int>(FILE *f, const std::vector<int> *vec)
{
    serialize64(f, vec->size());
    serialize_long_blob(f, &((*vec)[0]), vec->size()*sizeof(int));
}

template<>
void unserialize<int8_t>(FILE *f, std::vector<int8_t, aligned_allocator<int8_t> > *vec)
{
    int64_t vs = unserialize64(f);
    vec->resize(vs);
    unserialize_long_blob(f, &((*vec)[0]), vs*sizeof(int8_t));
}

template<>
void unserialize<float>(FILE *f, std::vector<float, aligned_allocator<float> > *vec)
{
    int64_t vs = unserialize64(f);
    vec->resize(vs);
    unserialize_long_blob(f, &((*vec)[0]), vs*sizeof(float));
}

template<>
void unserialize<float>(FILE *f, std::vector<float> *vec)
{
    int64_t vs = unserialize64(f);
    vec->resize(vs);
    unserialize_long_blob(f, &((*vec)[0]), vs*sizeof(float));
}

template<>
void unserialize<int>(FILE *f, std::vector<int> *vec)
{
    int64_t vs = unserialize64(f);
    vec->resize(vs);
    unserialize_long_blob(f, &((*vec)[0]), vs*sizeof(int));
}

static float lanczos2(float x)
{
    float pi_x;
    if( x==0 )
        return 1.0f;
    if( x<=-2 || x>=+2 )
        return 0.0f;
    pi_x = FLT_PI*x;
    return 2.0f*sinf(pi_x)*sinf(pi_x/2.0f)/(pi_x*pi_x);
}

static int misalignment(const void *ptr, size_t alignment)
{
    union _u
    {
        const void *ptr;
        std::ptrdiff_t iptr;
    } u;
    u.ptr = ptr;
    return (int)(u.iptr%alignment);
}

void apply_gain(const int8_t *act, int64_t n, int gain, int8_t *out)
{
#ifndef SNET_AVX2
#error apply_gain(): needs AVX2 support!
#endif
    if( gain==1 )
    {
        if( out!=act )
            memmove(out, act, n);
        return;
    }
    ASSERT(gain>1 && gain<=8);
    ASSERT(SNET_IMAX/gain+1<=127);
    const int64_t CHUNK_SIZE = 32;
    const int64_t nchunks = n/CHUNK_SIZE;
    __m256i mm_cap0   = _mm256_set1_epi8(SNET_IMAX/gain+1);
    __m256i mm_cap1   = _mm256_set1_epi8(SNET_IMAX);
    __m256i mm_gain16 = _mm256_set1_epi16(gain);
    for(int64_t k=0; k<nchunks; k++)
    {
        __m256i mm_chunk;
        mm_chunk = _mm256_load_si256((__m256i*)(act+CHUNK_SIZE*k));
        mm_chunk = _mm256_min_epi8(mm_chunk, mm_cap0);
        mm_chunk = _mm256_mullo_epi16(mm_chunk, mm_gain16);
        mm_chunk = _mm256_min_epi8(mm_chunk, mm_cap1);
        _mm256_store_si256((__m256i*)(out+CHUNK_SIZE*k), mm_chunk);
    }
    for(int64_t i=nchunks*CHUNK_SIZE; i<n; i++)
    {
        int v = int(act[i])*gain;
        out[i] = v>SNET_IMAX ? SNET_IMAX : v;
    }
}

void apply_gain(int8_t *act, int64_t n, int gain)
{
    // we know that it will work correctly
    apply_gain(act, n, gain, act);
}

void apply_gain(const float *act, int64_t n, int gain, float *out)
{
    if( gain==1 )
    {
        if( out!=act )
            memmove(out, act, n*sizeof(*act));
        return;
    }
#ifndef SNET_AVX2
#error apply_gain(): needs AVX2 support!
#endif
    __m256 mm_gain = _mm256_set1_ps(gain), mm_cap = _mm256_set1_ps(1.0f);
    const int CHUNK_SIZE = 8;
    const int nchunks = n/8;
    for(int64_t j=0; j<nchunks; j++)
    {
        __m256 mm_act = _mm256_loadu_ps(act+CHUNK_SIZE*j);
        mm_act = _mm256_mul_ps(mm_act, mm_gain);
        mm_act = _mm256_min_ps(mm_act, mm_cap);
        _mm256_storeu_ps(out+CHUNK_SIZE*j, mm_act);
    }
    for(int64_t i=nchunks*CHUNK_SIZE; i<n; i++)
    {
        float v = act[i]*gain;
        out[i] = v>1.0f ? 1.0f : v;
    }
}

void apply_gain(float *act, int64_t n, int gain)
{
    // we know that it will work correctly
    apply_gain(act, n, gain, act);
}

// both p0 and p1 must be aligned for SIMD reads
float fdot_simd(const float *p0, const float *p1, int64_t n)
{
    float v = 0.0f;

    //
    // Dataset sizes and generic code
    //
#ifdef SNET_AVX2
    const int64_t CHUNK_SIZE = 8;
#else
    const int64_t CHUNK_SIZE = 4;
#endif
    const int64_t nchunks = n/CHUNK_SIZE;
    bool both_aligned = misalignment(p0,CHUNK_SIZE*sizeof(float))==0 && misalignment(p1,CHUNK_SIZE*sizeof(float))==0;

    //
    // SIMD-specific processing
    //
#ifdef SNET_AVX2
    float vv[CHUNK_SIZE];
    __m256 mm_dot = _mm256_set1_ps(0.0f);
    if( both_aligned )
    {
        //
        // Properly aligned operands
        //
        for(int64_t j=0; j<nchunks; j++)
            mm_dot = _mm256_fmadd_ps(_mm256_load_ps(p0+CHUNK_SIZE*j),_mm256_load_ps(p1+CHUNK_SIZE*j),mm_dot);
    }
    else
    {
        //
        // Unaligned loads
        //
        for(int64_t j=0; j<nchunks; j++)
            mm_dot = _mm256_fmadd_ps(_mm256_loadu_ps(p0+CHUNK_SIZE*j),_mm256_loadu_ps(p1+CHUNK_SIZE*j),mm_dot);
    }
    _mm256_storeu_ps(vv, mm_dot);
    for(int64_t j=0; j<CHUNK_SIZE; j++)
        v += vv[j];
#else
    __m128 mm_dot = _mm_set1_ps(0.0f);
    if( both_aligned )
    {
        //
        // Properly aligned operands
        //
        for(int64_t j=0; j<nchunks; j++)
            mm_dot = _mm_add_ps(_mm_mul_ps(_mm_load_ps(p0+CHUNK_SIZE*j),_mm_load_ps(p1+CHUNK_SIZE*j)),mm_dot);
    }
    else
    {
        //
        // Unaligned loads
        //
        for(int64_t j=0; j<nchunks; j++)
            mm_dot = _mm_add_ps(_mm_mul_ps(_mm_loadu_ps(p0+CHUNK_SIZE*j),_mm_loadu_ps(p1+CHUNK_SIZE*j)),mm_dot);
    }
    mm_dot = _mm_hadd_ps(mm_dot,mm_dot);
    mm_dot = _mm_hadd_ps(mm_dot,mm_dot);
    _mm_store_ss(&v, mm_dot);
#endif

    //
    // Process tail
    //
    for(int64_t j=nchunks*CHUNK_SIZE; j<n; j++)
        v += p0[j]*p1[j];
    return v;
}

void fadd_diffc2(float *dst, const float *p0, float c, int64_t n)
{
#ifndef SNET_AVX2
#error SSE4 version is not supported
#endif
    const int64_t CHUNK_SIZE = 8;
    const int64_t nchunks = n/CHUNK_SIZE;
    __m256 mm_c = _mm256_set1_ps(c);
    for(int64_t j=0; j<nchunks; j++)
    {
        __m256 mm_diff = _mm256_sub_ps(_mm256_loadu_ps(p0+j*CHUNK_SIZE),mm_c);
        __m256 mm_acc  = _mm256_loadu_ps(dst+CHUNK_SIZE*j);
        _mm256_storeu_ps(dst+CHUNK_SIZE*j, _mm256_fmadd_ps(mm_diff,mm_diff,mm_acc));
    }
    for(int64_t j=nchunks*CHUNK_SIZE; j<n; j++)
    {
        float v = p0[j]-c;
        dst[j] += v*v;
    }
}

void fadd_diff2(float *dst, const float *p0, const float *p1, int64_t n)
{
#ifndef SNET_AVX2
#error SSE4 version is not supported
#endif
    const int64_t CHUNK_SIZE = 8;
    const int64_t nchunks = n/CHUNK_SIZE;
    for(int64_t j=0; j<nchunks; j++)
    {
        __m256 mm_diff = _mm256_sub_ps(_mm256_loadu_ps(p0+j*CHUNK_SIZE),_mm256_loadu_ps(p1+j*CHUNK_SIZE));
        __m256 mm_acc  = _mm256_loadu_ps(dst+CHUNK_SIZE*j);
        _mm256_storeu_ps(dst+CHUNK_SIZE*j, _mm256_fmadd_ps(mm_diff,mm_diff,mm_acc));
    }
    for(int64_t j=nchunks*CHUNK_SIZE; j<n; j++)
    {
        float v = p0[j]-p1[j];
        dst[j] += v*v;
    }
}

void fmerge_min(float *dst, const float *src, int64_t n)
{
#ifndef SNET_AVX2
#error SSE4 version is not supported
#endif
    const int64_t CHUNK_SIZE = 8;
    const int64_t nchunks = n/CHUNK_SIZE;
    for(int64_t j=0; j<nchunks; j++)
    {
        __m256 mm_a = _mm256_loadu_ps(dst+j*CHUNK_SIZE);
        __m256 mm_b = _mm256_loadu_ps(src+j*CHUNK_SIZE);
        _mm256_storeu_ps(dst+CHUNK_SIZE*j, _mm256_min_ps(mm_a,mm_b));
    }
    for(int64_t j=nchunks*CHUNK_SIZE; j<n; j++)
        dst[j] = fminf(dst[j],src[j]);
}

void fmerge_max(float *dst, const float *src, int64_t n)
{
#ifndef SNET_AVX2
#error SSE4 version is not supported
#endif
    const int64_t CHUNK_SIZE = 8;
    const int64_t nchunks = n/CHUNK_SIZE;
    for(int64_t j=0; j<nchunks; j++)
    {
        __m256 mm_a = _mm256_loadu_ps(dst+j*CHUNK_SIZE);
        __m256 mm_b = _mm256_loadu_ps(src+j*CHUNK_SIZE);
        _mm256_storeu_ps(dst+CHUNK_SIZE*j, _mm256_max_ps(mm_a,mm_b));
    }
    for(int64_t j=nchunks*CHUNK_SIZE; j<n; j++)
        dst[j] = fmaxf(dst[j],src[j]);
}

void fmerge_mul(float *dst, const float *src, int64_t n)
{
#ifndef SNET_AVX2
#error SSE4 version is not supported
#endif
    const int64_t CHUNK_SIZE = 8;
    const int64_t nchunks = n/CHUNK_SIZE;
    for(int64_t j=0; j<nchunks; j++)
    {
        __m256 mm_a = _mm256_loadu_ps(dst+j*CHUNK_SIZE);
        __m256 mm_b = _mm256_loadu_ps(src+j*CHUNK_SIZE);
        _mm256_storeu_ps(dst+CHUNK_SIZE*j, _mm256_mul_ps(mm_a,mm_b));
    }
    for(int64_t j=nchunks*CHUNK_SIZE; j<n; j++)
        dst[j] *= src[j];
}

void fmerge_wmax(float *dst, const float *src, float w, int64_t n)
{
#ifndef SNET_AVX2
#error SSE4 version is not supported
#endif
    if( w==1.0f )
    {
        fmerge_max(dst, src, n);
        return;
    }
    const int64_t CHUNK_SIZE = 8;
    const int64_t nchunks = n/CHUNK_SIZE;
    const __m256 mm_w = _mm256_set1_ps(w);
    for(int64_t j=0; j<nchunks; j++)
    {
        __m256 mm_a = _mm256_loadu_ps(dst+j*CHUNK_SIZE);
        __m256 mm_b = _mm256_loadu_ps(src+j*CHUNK_SIZE);
        _mm256_storeu_ps(dst+CHUNK_SIZE*j, _mm256_max_ps(mm_a,_mm256_mul_ps(mm_b,mm_w)));
    }
    for(int64_t j=nchunks*CHUNK_SIZE; j<n; j++)
        dst[j] = fmaxf(dst[j],w*src[j]);
}

void fmin_inplace(float *dst, float v, int64_t n)
{
#ifndef SNET_AVX2
#error SSE4 version is not supported
#endif
    const int64_t CHUNK_SIZE = 8;
    const int64_t nchunks = n/CHUNK_SIZE;
    __m256 mm_b = _mm256_set1_ps(v);
    for(int64_t j=0; j<nchunks; j++)
    {
        __m256 mm_a = _mm256_loadu_ps(dst+j*CHUNK_SIZE);
        _mm256_storeu_ps(dst+CHUNK_SIZE*j, _mm256_min_ps(mm_a,mm_b));
    }
    for(int64_t j=nchunks*CHUNK_SIZE; j<n; j++)
        dst[j] = fminf(dst[j],v);
}

void fmul_inplace(float *dst, float v, int64_t n)
{
#ifndef SNET_AVX2
#error SSE4 version is not supported
#endif
    const int64_t CHUNK_SIZE = 8;
    const int64_t nchunks = n/CHUNK_SIZE;
    __m256 mm_b = _mm256_set1_ps(v);
    for(int64_t j=0; j<nchunks; j++)
    {
        __m256 mm_a = _mm256_loadu_ps(dst+j*CHUNK_SIZE);
        _mm256_storeu_ps(dst+CHUNK_SIZE*j, _mm256_mul_ps(mm_a,mm_b));
    }
    for(int64_t j=nchunks*CHUNK_SIZE; j<n; j++)
        dst[j] *= v;
}

void fmul_inplace(float *dst, const float *src, int64_t n)
{
#ifndef SNET_AVX2
#error SSE4 version is not supported
#endif
    const int64_t CHUNK_SIZE = 8;
    const int64_t nchunks = n/CHUNK_SIZE;
    for(int64_t j=0; j<nchunks; j++)
    {
        __m256 mm_a = _mm256_loadu_ps(dst+j*CHUNK_SIZE);
        __m256 mm_b = _mm256_loadu_ps(src+j*CHUNK_SIZE);
        _mm256_storeu_ps(dst+CHUNK_SIZE*j, _mm256_mul_ps(mm_a,mm_b));
    }
    for(int64_t j=nchunks*CHUNK_SIZE; j<n; j++)
        dst[j] *= src[j];
}

void fmul_add_inplace(float *srcdst, const float *mul, const float *add, int64_t n)
{
#ifndef SNET_AVX2
#error SSE4 version is not supported
#endif
    const int64_t CHUNK_SIZE = 8;
    const int64_t nchunks = n/CHUNK_SIZE;
    for(int64_t j=0; j<nchunks; j++)
    {
        __m256 mm_a = _mm256_loadu_ps(srcdst+j*CHUNK_SIZE);
        __m256 mm_b = _mm256_loadu_ps(   mul+j*CHUNK_SIZE);
        __m256 mm_c = _mm256_loadu_ps(   add+j*CHUNK_SIZE);
        _mm256_storeu_ps(srcdst+CHUNK_SIZE*j, _mm256_fmadd_ps(mm_a,mm_b,mm_c));
    }
    for(int64_t j=nchunks*CHUNK_SIZE; j<n; j++)
        srcdst[j] = srcdst[j]*mul[j]+add[j];
}

void fadd_mul_inplace(float *srcdst, const float *add0, const float *add1, int64_t n)
{
#ifndef SNET_AVX2
#error SSE4 version is not supported
#endif
    const int64_t CHUNK_SIZE = 8;
    const int64_t nchunks = n/CHUNK_SIZE;
    for(int64_t j=0; j<nchunks; j++)
    {
        __m256 mm_a = _mm256_loadu_ps(srcdst+j*CHUNK_SIZE);
        __m256 mm_b = _mm256_loadu_ps(  add0+j*CHUNK_SIZE);
        __m256 mm_c = _mm256_loadu_ps(  add1+j*CHUNK_SIZE);
        _mm256_storeu_ps(srcdst+CHUNK_SIZE*j, _mm256_fmadd_ps(mm_b,mm_c,mm_a));
    }
    for(int64_t j=nchunks*CHUNK_SIZE; j<n; j++)
        srcdst[j] += add0[j]*add1[j];
}

float vnrm2(const int8_t *p, int64_t n)
{
    return i8nrm2(p,n);
}

float vnrm2(const float *p, int64_t n)
{
    return fdot_simd(p,p,n);
}

float vdist2(const float *p0, const float *p1, int64_t n)
{
/*
#ifndef SNET_AVX2
#error AVX2 support is required (SSE4 code is not implemented yet)
#endif
    float v = 0.0f;
    const int64_t CHUNK_SIZE = 8;
    const int64_t nchunks = n/CHUNK_SIZE;
    bool both_aligned = misalignment(p0,CHUNK_SIZE*sizeof(float))==0 && misalignment(p1,CHUNK_SIZE*sizeof(float))==0;
    float vv[CHUNK_SIZE];
    __m256 mm_dist2 = _mm256_set1_ps(0.0f);
    if( both_aligned )
    {
        //
        // Properly aligned operands
        //
        for(int64_t j=0; j<nchunks; j++)
        {
            __m256 mm_diff = _mm256_sub_ps(_mm256_load_ps(p0+CHUNK_SIZE*j), _mm256_load_ps(p1+CHUNK_SIZE*j));
            mm_dist2 = _mm256_fmadd_ps(mm_diff,mm_diff,mm_dist2);
        }
    }
    else
    {
        //
        // Unaligned loads
        //
        for(int64_t j=0; j<nchunks; j++)
        {
            __m256 mm_diff = _mm256_sub_ps(_mm256_loadu_ps(p0+CHUNK_SIZE*j), _mm256_loadu_ps(p1+CHUNK_SIZE*j));
            mm_dist2 = _mm256_fmadd_ps(mm_diff,mm_diff,mm_dist2);
        }
    }
    _mm256_storeu_ps(vv, mm_dist2);
    for(int64_t j=0; j<CHUNK_SIZE; j++)
        v += vv[j];
    for(int64_t j=nchunks*CHUNK_SIZE; j<n; j++)
    {
        float vv = p0[j]-p1[j];
        v += vv*vv;
    }
    return v;
*/
    float v = 0;
    for(int64_t k=0; k<n; k++)
    {
        float vv = p0[k]-p1[k];
        v += vv*vv;
    }
    return v;
}

float i8nrm2(const int8_t *p, int64_t n)
{
#ifndef SNET_AVX2
#error AVX2 support is required (SSE4 code is not implemented yet)
#endif
    const int64_t CHUNK_SIZE = 8;
    const int64_t nchunks = n/CHUNK_SIZE;
    __m256 mm_nrm2 = _mm256_set1_ps(0.0f);
    for(int64_t i=0; i<nchunks; i++)
    {
        __m256 mm_v = _mm256_cvtepi32_ps(_mm256_cvtepi8_epi32(_mm_set1_epi64x(((const int64_t*)p)[i])));
        mm_nrm2 = _mm256_fmadd_ps(mm_v,mm_v,mm_nrm2);
    }
    float r = 0.0f;
    float va[CHUNK_SIZE];
    _mm256_storeu_ps(va, mm_nrm2);
    for(int k=0; k<CHUNK_SIZE; k++)
        r += va[k];
    for(int64_t i=CHUNK_SIZE*nchunks; i<n; i++)
    {
        float a = p[i];
        r += a*a;
    }
    return r*SNET_I2F*SNET_I2F;
}

// both p0 and p1 must be aligned for SIMD reads
void fadd_simd(float *p0, const float *p1, int64_t n, float s)
{
    //
    // Dataset sizes and generic code
    //
#ifdef SNET_AVX2
    const int64_t CHUNK_SIZE = 8;
#else
    const int64_t CHUNK_SIZE = 4;
#endif
    const int64_t nchunks = n/CHUNK_SIZE;
    bool both_aligned = misalignment(p0,CHUNK_SIZE*sizeof(float))==0 && misalignment(p1,CHUNK_SIZE*sizeof(float))==0;

    //
    // SIMD-specific code
    //
#ifdef SNET_AVX2
    __m256 mm_s = _mm256_set1_ps(s);
    if( both_aligned )
    {
        //
        // Both operands are properly aligned
        //
        for(int64_t j=0; j<nchunks; j++)
        {
            float *p_target = p0+CHUNK_SIZE*j;
            __m256 mm_v1;
            mm_v1 = _mm256_load_ps(p1+j*CHUNK_SIZE);
            mm_v1 = _mm256_fmadd_ps(mm_v1, mm_s, _mm256_load_ps(p_target));
            _mm256_store_ps(p_target, mm_v1);
        }
    }
    else
    {
        //
        // Unlianged loads are used
        //
        for(int64_t j=0; j<nchunks; j++)
        {
            float *p_target = p0+CHUNK_SIZE*j;
            __m256 mm_v1;
            mm_v1 = _mm256_loadu_ps(p1+j*CHUNK_SIZE);
            mm_v1 = _mm256_fmadd_ps(mm_v1, mm_s, _mm256_loadu_ps(p_target));
            _mm256_storeu_ps(p_target, mm_v1);
        }
    }
#else
    __m128 mm_s = _mm_set1_ps(s);
    if( both_aligned )
    {
        //
        // Both operands are properly aligned
        //
        for(int64_t j=0; j<nchunks; j++)
        {
            __m128 mm_v1;
            float *p_target = p0+CHUNK_SIZE*j;
            mm_v1 = _mm_load_ps(p1+CHUNK_SIZE*j);
            mm_v1 = _mm_mul_ps(mm_v1,mm_s);
            _mm_store_ps(p_target, _mm_add_ps(_mm_load_ps(p_target), mm_v1));
        }
    }
    else
    {
        //
        // Unlianged loads are used
        //
        for(int64_t j=0; j<nchunks; j++)
        {
            __m128 mm_v1;
            float *p_target = p0+CHUNK_SIZE*j;
            mm_v1 = _mm_loadu_ps(p1+CHUNK_SIZE*j);
            mm_v1 = _mm_mul_ps(mm_v1,mm_s);
            _mm_storeu_ps(p_target, _mm_add_ps(_mm_loadu_ps(p_target), mm_v1));
        }
    }
#endif
    for(int64_t j=nchunks*CHUNK_SIZE; j<n; j++)
        p0[j] += s*p1[j];
}

/*void act_f32_to_i8(int8_t *p_target, const float *p_source, int64_t n)
{
#ifndef SNET_AVX2
#error AVX2 support is required (SSE4 code is not implemented yet)
#endif
    for(int64_t k=0; k<n; k++)
        p_target[k] = posflt_to_int8(p_source[k]);
}*/

void act8_merge_max_simd(int8_t *p_target, const int8_t *p_merged, int64_t n)
{
#ifndef SNET_AVX2
#error AVX2 support is required (SSE4 code is not implemented yet)
#endif
    const int64_t chunk_size = 32;
    const int64_t nchunks = n/chunk_size;
    const int8_t *pb = p_merged;
    int8_t *pa = p_target;
    for(int64_t i=0; i<nchunks; i++, pa+=chunk_size, pb+=chunk_size)
    {
        __m256i mm_a8 = _mm256_loadu_si256((__m256i*)pa);
        __m256i mm_b8 = _mm256_loadu_si256((__m256i*)pb);
        _mm256_storeu_si256((__m256i*)pa, _mm256_max_epi8(mm_a8,mm_b8));
    }
    for(int64_t i=nchunks*chunk_size; i<n; i++)
    {
        int8_t va = p_merged[i], vb = p_target[i];
        int8_t v = va>=vb ? va : vb;
        p_target[i] = v;
    }
    /*for(int64_t i=0; i<n; i++)
    {
        int8_t va = p_merged[i], vb = p_target[i];
        int8_t v = va>=vb ? va : vb;
        p_target[i] = v;
    }*/
}

void act8_merge_wmax_simd(int8_t *p_target, const int8_t *p_merged, weight_type w, int64_t n)
{
#ifndef SNET_AVX2
#error AVX2 support is required (SSE4 code is not implemented yet)
#endif
    if( w==weight_type::w100 )
    {
        act8_merge_max_simd(p_target, p_merged, n);
        return;
    }
    if( w==weight_type::w075 )
    {
        const __m256i mm_mask7 = _mm256_set1_epi16(0x7F7Ful);
        const __m256i mm_mask6 = _mm256_set1_epi16(0x3F3Ful);
        const int64_t chunk_size = 32;
        const int64_t nchunks = n/chunk_size;
        const int8_t *pb = p_merged;
        int8_t *pa = p_target;
        for(int64_t i=0; i<nchunks; i++, pa+=chunk_size, pb+=chunk_size)
        {
            __m256i mm_a8 = _mm256_loadu_si256((__m256i*)pa);
            __m256i mm_b8 = _mm256_loadu_si256((__m256i*)pb);
            __m256i mm_050 = _mm256_and_si256(_mm256_srli_epi16(mm_b8, 0x1), mm_mask7);
            __m256i mm_025 = _mm256_and_si256(_mm256_srli_epi16(mm_b8, 0x2), mm_mask6);
            mm_b8 = _mm256_add_epi8(mm_050, mm_025);
            _mm256_storeu_si256((__m256i*)pa, _mm256_max_epi8(mm_a8,mm_b8));
        }
        for(int64_t i=nchunks*chunk_size; i<n; i++)
        {
            int8_t va = p_target[i], vb = (p_merged[i]>>1)+(p_merged[i]>>2);
            int8_t v = va>=vb ? va : vb;
            p_target[i] = v;
        }
        return;
    }
    if( w==weight_type::w050 )
    {
        const __m256i mm_mask7 = _mm256_set1_epi16(0x7F7Ful);
        const int64_t chunk_size = 32;
        const int64_t nchunks = n/chunk_size;
        const int8_t *pb = p_merged;
        int8_t *pa = p_target;
        for(int64_t i=0; i<nchunks; i++, pa+=chunk_size, pb+=chunk_size)
        {
            __m256i mm_a8 = _mm256_loadu_si256((__m256i*)pa);
            __m256i mm_b8 = _mm256_loadu_si256((__m256i*)pb);
            mm_b8 = _mm256_srli_epi16(mm_b8, 0x1);
            mm_b8 = _mm256_and_si256(mm_b8, mm_mask7);
            _mm256_storeu_si256((__m256i*)pa, _mm256_max_epi8(mm_a8,mm_b8));
        }
        for(int64_t i=nchunks*chunk_size; i<n; i++)
        {
            int8_t va = p_target[i], vb = p_merged[i]>>1;
            int8_t v = va>=vb ? va : vb;
            p_target[i] = v;
        }
        return;
    }
    ASSERT(0);
}

void act8_merge_gmax_simd(int8_t *p_target, const int8_t *p_merged, gain_t g, int64_t n)
{
#ifndef SNET_AVX2
#error AVX2 support is required (SSE4 code is not implemented yet)
#endif
    if( g==gain_t::gain1 )
    {
        act8_merge_max_simd(p_target, p_merged, n);
        return;
    }
    ASSERT(g==gain_t::gain1 || g==gain_t::gain2 || g==gain_t::gain4 || g==gain_t::gain8);
    int ig = int(g);
#warning Need normal gained code!!!
    /*const int64_t chunk_size = 32;
    const int64_t nchunks = n/chunk_size;
    const int8_t *pb = p_merged;
    int8_t *pa = p_target;
    __m256i mm_cap0  = _mm256_set1_epi16(SNET_IMAX);
    for(int64_t i=0; i<nchunks; i++, pa+=chunk_size, pb+=chunk_size)
    {
        __m256i mm_a8 = _mm256_loadu_si256((__m256i*)pa);
        __m256i mm_b8 = _mm256_loadu_si256((__m256i*)pb);
        __m256i mm_v  = _mm256_max_epi8(mm_a8,mm_b8);
        _mm256_storeu_si256((__m256i*)pa, mm_v);
    }
    for(int64_t i=nchunks*chunk_size; i<n; i++)
    {
        int8_t va = p_merged[i], vb = p_target[i];
        int8_t v = va>=vb ? va : vb;
        p_target[i] = v;
    }*/
    for(int64_t i=0; i<n; i++)
    {
        int8_t va = p_merged[i], vb = p_target[i];
        int8_t v = va>=vb ? va : vb;
        v *= ig;
        v = v<=SNET_IMAX ? v : SNET_IMAX;
        p_target[i] = v;
    }
}

void act8_merge_min_simd(int8_t *p_target, const int8_t *p_merged, int64_t n)
{
#ifndef SNET_AVX2
#error AVX2 support is required (SSE4 code is not implemented yet)
#endif
#ifndef SNET_AVX2
#error AVX2 support is required (SSE4 code is not implemented yet)
#endif
    const int64_t chunk_size = 32;
    const int64_t nchunks = n/chunk_size;
    const int8_t *pb = p_merged;
    int8_t *pa = p_target;
    for(int64_t i=0; i<nchunks; i++, pa+=chunk_size, pb+=chunk_size)
    {
        __m256i mm_a8 = _mm256_loadu_si256((__m256i*)pa);
        __m256i mm_b8 = _mm256_loadu_si256((__m256i*)pb);
        _mm256_storeu_si256((__m256i*)pa, _mm256_min_epi8(mm_a8,mm_b8));
    }
    for(int64_t i=nchunks*chunk_size; i<n; i++)
    {
        int8_t va = p_merged[i], vb = p_target[i];
        int8_t v = va<vb ? va : vb;
        p_target[i] = v;
    }
    /*for(int64_t i=0; i<n; i++)
    {
        int8_t va = p_merged[i], vb = p_target[i];
        int8_t v = va<vb ? va : vb;
        p_target[i] = v;
    }*/
}

hsv hsvfromi24(int irgb)
{
    float s = 1.0f/255.0f;
    return rgb2hsv((irgb&0xFF)*s, ((irgb&0xFF00)>>8)*s, ((irgb&0xFF0000)>>16)*s);
}

int i24fromhsv(hsv v)
{
    rgb t = hsv2rgb(v);
    int ir = irange(t.r*255.0f, 0, 255);
    int ig = irange(t.g*255.0f, 0, 255);
    int ib = irange(t.b*255.0f, 0, 255);
    return ir | ig<<8 | ib<<16;
}

hsv rgb2hsv(float r, float g, float b)
{
    rgb tmp;
    tmp.r = r;
    tmp.g = g;
    tmp.b = b;
    return rgb2hsv(tmp);
}

rgb hsv2rgb(float h, float s, float v)
{
    hsv tmp;
    tmp.h = h;
    tmp.s = s;
    tmp.v = v;
    return hsv2rgb(tmp);
}

hsv rgb2hsv(const rgb &in)
{
    hsv   out;
    float min, max, delta;

    min = in.r < in.g ? in.r : in.g;
    min = min  < in.b ? min  : in.b;

    max = in.r > in.g ? in.r : in.g;
    max = max  > in.b ? max  : in.b;

    out.v = max;                                // v
    delta = max - min;
    if (delta < 0.00001f)
    {
        out.s = 0.0f;
        out.h = 0.0f; // undefined, maybe nan?
        return out;
    }
    if( max > 0.0f ) { // NOTE: if Max is == 0, this divide would cause a crash
        out.s = (delta / max);                  // s
    } else {
        // if max is 0, then r = g = b = 0
            // s = 0, v is undefined
        out.s = 0.0f;
        out.h = 0.0f;                            // its now undefined
        return out;
    }
    if( in.r >= max )                           // > is bogus, just keeps compilor happy
        out.h = ( in.g - in.b ) / delta;        // between yellow & magenta
    else
    if( in.g >= max )
        out.h = 2.0f + ( in.b - in.r ) / delta;  // between cyan & yellow
    else
        out.h = 4.0f + ( in.r - in.g ) / delta;  // between magenta & cyan

    out.h *= 60.0f;                              // degrees

    if( out.h < 0.0f )
        out.h += 360.0f;

    return out;
}

rgb hsv2rgb(const hsv &in)
{
    float      hh, p, q, t, ff;
    long        i;
    rgb out;

    if(in.s <= 0.0f) {       // < is bogus, just shuts up warnings
        out.r = in.v;
        out.g = in.v;
        out.b = in.v;
        return out;
    }
    hh = in.h;
    if(hh >= 360.0f) hh = 0.0f;
    hh /= 60.0f;
    i = (long)hh;
    ff = hh - i;
    p = in.v * (1.0f - in.s);
    q = in.v * (1.0f - (in.s * ff));
    t = in.v * (1.0f - (in.s * (1.0f - ff)));

    switch(i) {
    case 0:
        out.r = in.v;
        out.g = t;
        out.b = p;
        break;
    case 1:
        out.r = q;
        out.g = in.v;
        out.b = p;
        break;
    case 2:
        out.r = p;
        out.g = in.v;
        out.b = t;
        break;

    case 3:
        out.r = p;
        out.g = q;
        out.b = in.v;
        break;
    case 4:
        out.r = t;
        out.g = p;
        out.b = in.v;
        break;
    case 5:
    default:
        out.r = in.v;
        out.g = p;
        out.b = q;
        break;
    }
    return out;
}

bitmap::trf::trf()
{
    trf_type = type_t::none;

    flip_h = flip_v = false;
    rot = 0;
    fscale0 = fscale1 = 1;
    fshift0 = fshift1 = 0;
    crop_h = crop_w = crop_offs0 = crop_offs1 = -1;

    elastic_rad = 1;
    elastic_mag = 0;
    elastic_fixed_bnd = true;
    elastic_seed = 1;
}

void bitmap::trf::set_flips(bool h, bool v)
{
    ASSERT(trf_type==type_t::none || trf_type==type_t::pipeline1);
    trf_type = type_t::pipeline1;
    flip_h = h;
    flip_v = v;
}

void bitmap::trf::set_scale(float s)
{
    ASSERT(s>0);
    ASSERT(trf_type==type_t::none || trf_type==type_t::pipeline1);
    trf_type = type_t::pipeline1;
    fscale0 = fscale1 = s;
}

void bitmap::trf::set_fshift(float s0, float s1)
{
    ASSERT(trf_type==type_t::none || trf_type==type_t::pipeline1);
    trf_type = type_t::pipeline1;
    fshift0 = s0;
    fshift1 = s1;
}

void bitmap::trf::set_rot(float rad_rot)
{
    ASSERT(trf_type==type_t::none || trf_type==type_t::pipeline1);
    trf_type = type_t::pipeline1;
    rot = rad_rot;
}

void bitmap::trf::set_crop(int h, int w, int offs0, int offs1)
{
    ASSERT(trf_type==type_t::none || trf_type==type_t::pipeline1);
    trf_type = type_t::pipeline1;
    crop_h = h;
    crop_w = w;
    crop_offs0 = offs0;
    crop_offs1 = offs1;
}

bitmap bitmap::from_jpg(std::string sfilename)
{
    bitmap bmp;
    unsigned char r, g, b;
    struct jpeg_decompress_struct cinfo;
    struct jpeg_error_mgr jerr;

    FILE * infile;        /* source file */
    JSAMPARRAY pJpegBuffer;       /* Output row buffer */
    int row_stride;       /* physical row width in output buffer */
    infile = fopen(sfilename.c_str(), "rb");
    ASSERT(infile!=NULL);
    cinfo.err = jpeg_std_error(&jerr);
    jpeg_create_decompress(&cinfo);
    jpeg_stdio_src(&cinfo, infile);
    (void) jpeg_read_header(&cinfo, TRUE);
    (void) jpeg_start_decompress(&cinfo);
    bmp.resize(cinfo.output_height, cinfo.output_width);

    row_stride = bmp.width * cinfo.output_components;
    pJpegBuffer = (*cinfo.mem->alloc_sarray)((j_common_ptr) &cinfo, JPOOL_IMAGE, row_stride, 1);
    int y=0;
    while(cinfo.output_scanline < cinfo.output_height)
    {
        (void)jpeg_read_scanlines(&cinfo, pJpegBuffer, 1);
        for (int x = 0; x <bmp.width; x++)
        {
            r = pJpegBuffer[0][cinfo.output_components * x];
            if (cinfo.output_components > 2)
            {
                g = pJpegBuffer[0][cinfo.output_components * x + 1];
                b = pJpegBuffer[0][cinfo.output_components * x + 2];
            }
            else
            {
                g = r;
                b = r;
            }
            bmp.draw_rgb(y, x, r, g, b);
        }
        y++;
    }
    fclose(infile);
    (void) jpeg_finish_decompress(&cinfo);
    jpeg_destroy_decompress(&cinfo);
    return bmp;
}

bitmap bitmap::from_png(std::string sfilename)
{
    unsigned error;
    unsigned char* image;
    unsigned width, height;

    error = lodepng_decode24_file(&image, &width, &height, sfilename.c_str());
    ASSERT(error==0);
    bitmap out;
    out.resize(height, width);
    for(int i=0; i<height; i++)
        for(int j=0; j<width; j++)
            out.draw_rgb(i, j, image[3*(i*width+j)+0], image[3*(i*width+j)+1], image[3*(i*width+j)+2]);
    free(image);
    return out;
}

void bitmap::to_jpg(std::string filename, int compress) const
{
    ASSERT(compress>0 && compress<=100);
    ASSERT(sizeof(JSAMPLE)==1);
    jpeg_compress_struct cinfo;
    jpeg_error_mgr jerr;
    JSAMPROW row_pointer[1];

    //
    // Step 1: allocate and initialize JPEG compression object
    //
    cinfo.err = jpeg_std_error(&jerr);
    jpeg_create_compress(&cinfo);

    //
    // Step 2: specify data destination (eg, a file)
    //
    FILE *outfile = fopen(filename.c_str(), "wb");
    ASSERT(outfile!=NULL);
    jpeg_stdio_dest(&cinfo, outfile);

    //
    // Step 3: set parameters for compression
    //
    cinfo.image_width = width;
    cinfo.image_height = height;
    cinfo.input_components = 3;
    cinfo.in_color_space = JCS_RGB;
    jpeg_set_defaults(&cinfo);
    jpeg_set_quality(&cinfo, compress, TRUE /* limit to baseline-JPEG values */);

    //
    // Compress
    //
    jpeg_start_compress(&cinfo, TRUE);
    while( cinfo.next_scanline < cinfo.image_height )
    {
        row_pointer[0] = (JSAMPLE*)(const_cast<unsigned char*>(&data[cinfo.next_scanline*width*3]));
        jpeg_write_scanlines(&cinfo, row_pointer, 1);
    }
    jpeg_finish_compress(&cinfo);
    fclose(outfile);
    jpeg_destroy_compress(&cinfo);
}

bitmap bitmap::from_ppm(const char *filename)
{
    char buff[16];
    FILE *f;
    bitmap r;

    // try opening file
    f = fopen(filename, "rb");
    ASSERT(f!=NULL);

    // read and check image format
    if( !fgets(buff,sizeof(buff),f) )
        ASSERT(false);
    ASSERT(buff[0]=='P' && buff[1]=='6');

    //check for comments
    char c = getc(f);
    while (c == '#') {
    while (getc(f) != '\n') ;
         c = getc(f);
    }
    ungetc(c, f);

    //read image size information, allocate space, read data
    if( fscanf(f, "%d %d", &r.width, &r.height)!=2 )
        ASSERT(false);
    int depth;
    if( fscanf(f, "%d", &depth)!=1)
        ASSERT(false);
    ASSERT(depth==255);
    while (fgetc(f) != '\n') ;

    //memory allocation for pixel data
    r.data.resize(3*r.height*r.width);
    if( fread(&(r.data[0]), 3*r.height*r.width, 1, f)!=1 )
        ASSERT(false);

    // close and return
    fclose(f);
    return r;
}

int bitmap::get_24_lanczos(float pos_h, float pos_w) const
{
    // normalize coordinates
    if( pos_h<0 )
        pos_h = 0;
    if( pos_w<0 )
        pos_w = 0;
    if( pos_h>height-1 )
        pos_h = height-1;
    if( pos_w>width-1 )
        pos_w = width-1;

    // integer coordinates
    int ih = int(pos_h), iw = int(pos_w);
    if( ih==pos_h && iw==pos_w )
        return get_24(ih, iw);

    // too close to borders, use bilinear interpolation
    if( ih-1<0 || ih+2>height-1 || iw-1<0 || iw+2>width-1 )
    {
        // handle a few boundary cases
        if( ih==height-1 && iw==width-1 )
            return get_24(ih, iw);
        if( ih==height-1 && iw<width-1)
        {
            int r0, r1, g0, g1, b0, b1;
            float w0 = 1.0f-(pos_w-iw);
            float w1 = 1.0f-w0;
            get_rgb(ih, iw+0, &r0, &g0, &b0);
            get_rgb(ih, iw+1, &r1, &g1, &b1);
            r0 = irange(roundf(w0*r0+w1*r1), 0, 255);
            g0 = irange(roundf(w0*g0+w1*g1), 0, 255);
            b0 = irange(roundf(w0*b0+w1*b1), 0, 255);
            return r0 | (g0<<8) | (b0<<16);
        }
        if( ih<height-1  && iw==width-1 )
        {
            int r0, r1, g0, g1, b0, b1;
            float w0 = 1.0f-(pos_h-ih);
            float w1 = 1.0f-w0;
            get_rgb(ih+0, iw, &r0, &g0, &b0);
            get_rgb(ih+1, iw, &r1, &g1, &b1);
            r0 = irange(roundf(w0*r0+w1*r1), 0, 255);
            g0 = irange(roundf(w0*g0+w1*g1), 0, 255);
            b0 = irange(roundf(w0*b0+w1*b1), 0, 255);
            return r0 | (g0<<8) | (b0<<16);
        }
        if( ih<height-1  && iw<width-1 )
        {
            int r0, r1, r2, r3, g0, g1, g2, g3, b0, b1, b2, b3;
            float w0 = (1.0f-(pos_h-ih))*(1.0f-(pos_w-iw));
            float w1 = (1.0f-(pos_h-ih))*(pos_w-iw);
            float w2 = (pos_h-ih)*(1.0f-(pos_w-iw));
            float w3 = (pos_h-ih)*(pos_w-iw);
            get_rgb(ih+0, iw+0, &r0, &g0, &b0);
            get_rgb(ih+0, iw+1, &r1, &g1, &b1);
            get_rgb(ih+1, iw+0, &r2, &g2, &b2);
            get_rgb(ih+1, iw+1, &r3, &g3, &b3);
            r0 = irange(roundf(w0*r0+w1*r1+w2*r2+w3*r3), 0, 255);
            g0 = irange(roundf(w0*g0+w1*g1+w2*g2+w3*g3), 0, 255);
            b0 = irange(roundf(w0*b0+w1*b1+w2*b2+w3*b3), 0, 255);
            return r0 | (g0<<8) | (b0<<16);
        }
        ASSERT(false);
        return 0;
    }

    // use Lanczos resampling
    float v0 = 0, v1 = 0, v2 = 0;
    for(int i=ih-1; i<=ih+2; i++)
        for(int j=iw-1; j<=iw+2; j++)
        {
            int r, g, b;
            float w;
            get_rgb(i, j, &r, &g, &b);
            w = lanczos2(pos_h-i)*lanczos2(pos_w-j);
            v0 += w*r;
            v1 += w*g;
            v2 += w*b;
        }
    v0 = roundf(v0);
    v1 = roundf(v1);
    v2 = roundf(v2);
    return irange(v0,0,255) | (irange(v1,0,255)<<8) | (irange(v2,0,255)<<16);
}

int bitmap::get_24_bilin(float pos_h, float pos_w) const
{
    // normalize coordinates
    if( pos_h<0 )
        pos_h = 0;
    if( pos_w<0 )
        pos_w = 0;
    if( pos_h>height-1 )
        pos_h = height-1;
    if( pos_w>width-1 )
        pos_w = width-1;

    // integer coordinates
    int ih = int(pos_h), iw = int(pos_w);
    if( ih==pos_h && iw==pos_w )
        return get_24(ih, iw);

    // use bilinear interpolation
    if( ih==height-1 && iw==width-1 )
        return get_24(ih, iw);
    if( ih==height-1 && iw<width-1)
    {
        int r0, r1, g0, g1, b0, b1;
        float w0 = 1.0f-(pos_w-iw);
        float w1 = 1.0f-w0;
        get_rgb(ih, iw+0, &r0, &g0, &b0);
        get_rgb(ih, iw+1, &r1, &g1, &b1);
        r0 = irange(roundf(w0*r0+w1*r1), 0, 255);
        g0 = irange(roundf(w0*g0+w1*g1), 0, 255);
        b0 = irange(roundf(w0*b0+w1*b1), 0, 255);
        return r0 | (g0<<8) | (b0<<16);
    }
    if( ih<height-1  && iw==width-1 )
    {
        int r0, r1, g0, g1, b0, b1;
        float w0 = 1.0f-(pos_h-ih);
        float w1 = 1.0f-w0;
        get_rgb(ih+0, iw, &r0, &g0, &b0);
        get_rgb(ih+1, iw, &r1, &g1, &b1);
        r0 = irange(roundf(w0*r0+w1*r1), 0, 255);
        g0 = irange(roundf(w0*g0+w1*g1), 0, 255);
        b0 = irange(roundf(w0*b0+w1*b1), 0, 255);
        return r0 | (g0<<8) | (b0<<16);
    }
    if( ih<height-1  && iw<width-1 )
    {
        int r0, r1, r2, r3, g0, g1, g2, g3, b0, b1, b2, b3;
        float w0 = (1.0f-(pos_h-ih))*(1.0f-(pos_w-iw));
        float w1 = (1.0f-(pos_h-ih))*(pos_w-iw);
        float w2 = (pos_h-ih)*(1.0f-(pos_w-iw));
        float w3 = (pos_h-ih)*(pos_w-iw);
        get_rgb(ih+0, iw+0, &r0, &g0, &b0);
        get_rgb(ih+0, iw+1, &r1, &g1, &b1);
        get_rgb(ih+1, iw+0, &r2, &g2, &b2);
        get_rgb(ih+1, iw+1, &r3, &g3, &b3);
        r0 = irange(roundf(w0*r0+w1*r1+w2*r2+w3*r3), 0, 255);
        g0 = irange(roundf(w0*g0+w1*g1+w2*g2+w3*g3), 0, 255);
        b0 = irange(roundf(w0*b0+w1*b1+w2*b2+w3*b3), 0, 255);
        return r0 | (g0<<8) | (b0<<16);
    }
    ASSERT(false);
    return 0;
}

bitmap bitmap::rough_resize_min(int min_width, int min_height) const
{
    // determine image size
    int new_width, new_height;
    if( float(min_width)/float(min_height)>=float(width)/float(height) )
    {
        // new width is minimum width, height is computed
        new_width = min_width;
        new_height = roundf(float(height)/float(width)*new_width);
    }
    else
    {
        // new heightis minimum height, width is computed
        new_height = min_height;
        new_width = roundf(float(width)/float(height)*new_height);
    }

    // allocate and write
    bitmap r_img;
    r_img.resize(new_height, new_width);
    for(int i=0; i<new_height; i++)
        for(int j=0; j<new_width; j++)
        {
            int i0 = roundf(float(i)/(new_height-1)*(height-1));
            int j0 = roundf(float(j)/(new_width-1)*(width-1));
            int r, g, b;
            get_rgb(i0,j0,&r, &g, &b);
            r_img.draw_rgb(i, j, r, g, b);

            /*float fi = float(i)/(new_height-1)*(height-1);
            float fj = float(j)/(new_width-1)*(width-1);
            r_img.draw_24(i, j, get_24_lanczos(fi, fj));*/
        }
    return r_img;
}

bitmap bitmap::rough_resize_to (int new_width, int new_height) const
{
    bitmap r_img;
    r_img.resize(new_height, new_width);
    for(int i=0; i<new_height; i++)
        for(int j=0; j<new_width; j++)
        {
            int i0 = roundf(float(i)/(new_height-1)*(height-1));
            int j0 = roundf(float(j)/(new_width-1)*(width-1));
            int r, g, b;
            get_rgb(i0,j0,&r, &g, &b);
            r_img.draw_rgb(i, j, r, g, b);
        }
    return r_img;
}

bitmap bitmap::lanczos_resize_to (int new_width, int new_height) const
{
    bitmap r_img;
    r_img.resize(new_height, new_width);
    for(int i=0; i<new_height; i++)
        for(int j=0; j<new_width; j++)
            r_img.draw_24(i, j, get_24_lanczos(float(i)/(new_height-1)*(height-1), float(j)/(new_width-1)*(width-1)));
    return r_img;
}

bitmap bitmap::bilin_resize_to (int new_width, int new_height) const
{
    bitmap r_img;
    r_img.resize(new_height, new_width);
    for(int i=0; i<new_height; i++)
        for(int j=0; j<new_width; j++)
            r_img.draw_24(i, j, get_24_bilin(float(i)/(new_height-1)*(height-1), float(j)/(new_width-1)*(width-1)));
    return r_img;
}

bitmap bitmap::smooth_resize_to(int new_width, int new_height) const
{
    bitmap r = *this;
    while( r.height>=2*new_height && r.width>=2*new_width )
        r = r.downsample2();
    return r.lanczos_resize_to(new_width, new_height);
}

bitmap bitmap::append_border_v(int border, float v) const
{
    bitmap out;
    out.resize_v(height+2*border,width+2*border, v);
    for(int i=0; i<height; i++)
        for(int j=0; j<width; j++)
            out.draw_24(border+i, border+j, get_24(i,j));
    return out;
}

bitmap bitmap::append_border_v(int top, int bottom, int left, int right, float v) const
{
    bitmap out;
    out.resize_v(height+top+bottom,width+left+right, v);
    for(int i=0; i<height; i++)
        for(int j=0; j<width; j++)
            out.draw_24(top+i, left+j, get_24(i,j));
    return out;
}


bitmap bitmap::central_transform_f(float fshift0, float fshift1, float fscale0, float fscale1, float rad_rot) const
{
    bitmap out = *this;
    float c0 = 0.5f*(height-1), c1 = 0.5f*(width-1);
    float v_cos = cosf(rad_rot), v_sin = sinf(rad_rot);
    for(int i0=0; i0<height; i0++)
        for(int i1=0; i1<width; i1++)
        {
            float x0 = i0-c0, x1=i1-c1;
            float t0 = x0, t1 = x1;
            x0 = t0*v_cos - t1*v_sin;
            x1 = t0*v_sin + t1*v_cos;
            x0 /= fscale0;
            x1 /= fscale1;
            x0 -= fshift0*height;
            x1 -= fshift1*width;
            out.draw_24(i0, i1, get_24_lanczos(x0+c0, x1+c1));
        }
    return out;
}

bitmap bitmap::central_crop(int size) const
{
    ASSERT(width==height);
    ASSERT(size>=1 && size<=width);
    int pad0 = (width-size)/2;
    int pad1 = (width-size)-pad0;
    return select_pos_xy(pad0,pad0,width-1-pad1,height-1-pad1);
}

bitmap bitmap::invert() const
{
    bitmap r = *this;
    for(int k=0; k<r.data.size(); k++)
        r.data[k] = 255-r.data[k];
    return r;
}

bitmap bitmap::normalize_contrast() const
{
    bitmap r = *this;
    int v_min = 256, v_max = 0;
    for(int k=0; k<r.data.size(); k++)
    {
        int v = r.data[k];
        v_min = imin(v_min, v);
        v_max = imax(v_max, v);
    }
    for(int k=0; k<r.data.size(); k++)
        r.data[k] = irange(roundf(float(r.data[k]-v_min)/float(v_max-v_min)*255.0f), 0, 255);
    return r;
}

bitmap bitmap::transform(const trf &t) const
{
    bitmap out = *this;

    if( t.trf_type==trf::type_t::none )
        return out;
    if( t.trf_type==trf::type_t::pipeline1 )
    {
        //
        // Pipeline #1
        //
        int src_offs0=0, src_offs1=0;
        if( t.crop_h>0 && t.crop_w>0 )
        {
            ASSERT(height>=t.crop_h);
            ASSERT(width>=t.crop_w);
            ASSERT(t.crop_offs0>=0);
            ASSERT(t.crop_offs1>=0);
            out.resize(t.crop_h, t.crop_w);
            src_offs0 = t.crop_offs0;
            src_offs1 = t.crop_offs1;
        }
        else
            out.resize(height, width);
        float c0 = 0.5f*(height-1), c1 = 0.5f*(width-1);
        float v_cos = cosf(t.rot), v_sin = sinf(t.rot);
        for(int dst_i0=0; dst_i0<out.height; dst_i0++)
            for(int dst_i1=0; dst_i1<out.width; dst_i1++)
            {
                int src_i0 = dst_i0+src_offs0;
                int src_i1 = dst_i1+src_offs1;

                // centered coordinates
                float x0 = src_i0-c0, x1=src_i1-c1;

                // flips
                if( t.flip_h )
                    x1 *= -1;
                if( t.flip_v )
                    x0 *= -1;

                // rotation
                float t0 = x0, t1 = x1;
                x0 = t0*v_cos - t1*v_sin;
                x1 = t0*v_sin + t1*v_cos;

                // scaling
                x0 /= t.fscale0;
                x1 /= t.fscale1;

                // shifting
                x0 -= t.fshift0*height;
                x1 -= t.fshift1*width;

                // output
                out.draw_24(dst_i0, dst_i1, get_24_lanczos(x0+c0, x1+c1));
            }
        return out;
    }
    ASSERT(0);
}

bitmap bitmap::shift_by(int shift0, int shift1) const
{
    bitmap out = *this;
    for(int i0=0; i0<height; i0++)
        for(int i1=0; i1<width; i1++)
            out.draw_24(i0, i1, get_24(i0-shift0, i1-shift1));
    return out;
}

bitmap bitmap::select_pos_xy(int x0, int y0, int x1, int y1) const
{
    ASSERT(x0<=x1 && y0<=y1);

    bitmap out;
    out.resize(y1-y0+1, x1-x0+1);
    for(int i0=0; i0<out.height; i0++)
        for(int i1=0; i1<out.width; i1++)
            out.draw_24(i0, i1, get_24(i0+y0, i1+x0));
    return out;
}

bitmap bitmap::even_odd_upsample2() const
{
    bitmap r_img;
    r_img.resize(2*height-1, 2*width-1);
    for(int i=0; i<r_img.height; i++)
        for(int j=0; j<r_img.width; j++)
            r_img.draw_24(i, j, get_24_lanczos(0.5f*i, 0.5f*j));
    return r_img;
}

bitmap bitmap::downsample2() const
{
    bitmap r_img;
    int rh = (height+1)/2;
    int rw = (width+1)/2;
    ASSERT(rh>0 && rw>0);
    std::vector<int> counts(rh*rw,0);
    std::vector<int> vals(3*rh*rw,0);

    for(int i=0; i<height; i++)
        for(int j=0; j<width; j++)
        {
            int dst = (i/2)*rw+(j/2);
            int r, g, b;
            get_rgb(i, j, &r, &g, &b);
            vals[3*dst+0] += r;
            vals[3*dst+1] += g;
            vals[3*dst+2] += b;
            counts[dst]++;
        }
    r_img.resize(rh, rw);
    for(int i=0; i<rh; i++)
        for(int j=0; j<rw; j++)
        {
            int src = i*rw+j;
            int r = vals[3*src+0]/counts[src];
            int g = vals[3*src+1]/counts[src];
            int b = vals[3*src+2]/counts[src];
            r_img.draw_rgb(i, j, r, g, b);
        }
    return r_img;
}

bitmap bitmap::downsample2k(int k) const
{
    ASSERT(k>=0);
    bitmap r = *this;
    for(int i=0; i<k; i++)
        r = r.downsample2();
    return r;
}

bitmap bitmap::combine_ltr(const std::vector<bitmap> &v, int border, int border_color)
{
    ASSERT(v.size()>=1);

    int out_height = 0, out_width = 0;
    for(const bitmap &b: v)
    {
        out_height =  imax(out_height, b.height);
        out_width  += b.width;
    }
    out_height += 2*border;
    out_width += border*(v.size()+1);

    // prepare output bitmap, draw border
    bitmap out;
    out.resize(out_height, out_width);
    for(int i=0; i<out_height; i++)
        for(int j=0; j<out_width; j++)
            out.draw_24(i, j, border_color);

    // start output
    int out_col = border;
    for(const bitmap &b: v)
    {
        for(int i=0; i<b.height; i++)
            for(int j=0; j<b.width; j++)
                out.draw_24(border+i, out_col+j, b.get_24(i, j));
        out_col += b.width+border;
    }
    return out;
}

bitmap bitmap::combine_grid_w(const std::vector<bitmap> &v, int grid_width, int border, int border_color)
{
    // determine metrics
    const int grid_height = iDivUp(v.size(),grid_width);
    int max_height = 1, max_width = 1;
    for(const bitmap &b: v)
    {
        max_height =  imax(max_height, b.height);
        max_width  =  imax(max_width,  b.width);
    }
    const int out_height = (max_height+border)*grid_height+border, out_width = (max_width+border)*grid_width+border;
    const int cell_height = max_height+border, cell_width = max_width+border;

    // prepare output bitmap, draw border
    bitmap out;
    out.resize(out_height, out_width);
    for(int i=0; i<out_height; i++)
        for(int j=0; j<out_width; j++)
            out.draw_24(i, j, border_color);

    // start output
    int out_idx = 0;
    for(const bitmap &b: v)
    {
        int out_cell0 = out_idx/grid_width, out_cell1 = out_idx%grid_width;
        for(int i=0; i<b.height; i++)
            for(int j=0; j<b.width; j++)
                out.draw_24(border+out_cell0*cell_height+i,
                            border+out_cell1*cell_width+j,
                            b.get_24(i,j));
        out_idx++;
    }
    return out;
}

float bitmap::project(float x, project_t prj, float a, float b)
{
    if( prj==project_t::mirror )
    {
        for(;;)
        {
            if( x>=a && x<=b )
                return x;
            if( x<a )
                x = a + (a-x);
            if( x>b )
                x = b - (x-b);
        }
    }
    ASSERT(0);
}

void serialize(FILE *f, const bitmap *ptr)
{
    serialize(f, &ptr->height);
    serialize(f, &ptr->width);
    serialize(f, &ptr->data);
}

void unserialize(FILE *f, bitmap *ptr)
{
    unserialize(f, &ptr->height);
    unserialize(f, &ptr->width);
    unserialize(f, &ptr->data);
}

raw_dataset::raw_dataset(std::string filename)
{
    unserialize_from(filename, this);
}

bitmap raw_dataset::transform(const bitmap &img, float rad_rot)
{
    bitmap out_img;
    float c0 = 0.5f*(img.height-1), c1 = 0.5f*(img.width-1);
    float cos_rot = cosf(rad_rot);
    float sin_rot = sinf(rad_rot);
    out_img.resize(img.height, img.width);
    for(int i0=0; i0<img.height; i0++)
        for(int i1=0; i1<img.width; i1++)
        {
            float x0 = i0, x1 = i1;

            // center
            x0 -= c0;
            x1 -= c1;

            // rotate
            float t0 = x0*cos_rot-x1*sin_rot, t1 = x0*sin_rot + x1*cos_rot;

            // uncenter
            x0 = t0+c0;
            x1 = t1+c1;

            // interpolate
            out_img.draw_24(i0, i1, img.get_24_lanczos(x0, x1));
        }
    return out_img;
}

void raw_dataset::append(const raw_dataset &ds)
{
    ASSERT(nclasses==ds.nclasses);
    ASSERT(ds.ntrn==0 || ds.ntst==0 || ntst==0);
    for(int i=0; i<ds.ntrn; i++)
        append_trn(ds.entries[i]);
    for(int i=0; i<ds.ntst; i++)
        append_tst(ds.entries[ds.ntrn+i]);
}

void raw_dataset::append_trn(const entry &e)
{
    ASSERT(e.class_id>=0 && e.class_id<nclasses);
    ASSERT(ntst==0);
    entries.push_back(e);
    ntrn++;
}

void raw_dataset::append_tst(const entry &e)
{
    ASSERT(e.class_id>=0 && e.class_id<nclasses);
    entries.push_back(e);
    ntst++;
}

raw_dataset raw_dataset::merge(std::vector<const raw_dataset*> datasets)
{
    if( datasets.size()==0 )
        return raw_dataset();

    raw_dataset r(datasets[0]->nclasses);
    for(const raw_dataset* p: datasets)
        ASSERT(p->nclasses==r.nclasses);
    for(const raw_dataset* p: datasets)
        for(int64_t i=0; i<p->ntrn; i++)
            r.append_trn(p->entries[i]);
    for(const raw_dataset* p: datasets)
        for(int64_t i=p->ntrn; i<p->ntrn+p->ntst; i++)
            r.append_tst(p->entries[i]);
    return r;
}

int64_t raw_dataset::get_ntotal() const
{
    return ntrn+ntst;
}

int64_t raw_dataset::count_trn_class(int class_id) const
{
    int64_t r = 0;
    for(int k=0; k<ntrn; k++)
        if( entries[k].class_id==class_id )
            r++;
    return r;
}

void raw_dataset::debug_visualize(std::string spath, int max_files)
{
    alglib::hqrndstate rs;
    alglib::hqrndseed(343, 356, rs);

    if( max_files<=0 )
        max_files = ntrn+ntst;
    for(int i=0; i<max_files; i++)
    {
        int idx = alglib::hqrnduniformi(rs,ntrn);
        bitmap img = entries[idx].img;//.rough_resize_min(64,64);

        char filename[1024];
        sprintf(filename, "%s/%06d.png", spath.c_str(), idx);
        img.to_png(filename);
    }
}

void raw_dataset::debug_visualize_all(const char *path)
{
    for(int i=0; i<ntrn+ntst; i++)
    {
        bitmap img = entries[i].img;

        char filename[1024];
        sprintf(filename, "%s/ds-%s-%06d.png", path, i<ntrn ? "trn" : "tst", int(i<ntrn ? i : i-ntrn));
        img.to_png(filename);
    }

}

class raw_dataset_transformer
{
public:
    raw_dataset &ds_raw;
    const std::vector<bitmap::trf> &trf;
    std::atomic<int32_t> entry_idx;

    raw_dataset_transformer(raw_dataset &ds, const std::vector<bitmap::trf> &t):
        ds_raw(ds),trf(t),entry_idx(0){}
};

void* apply_transform_to_dataset(void *xptr)
{
    raw_dataset_transformer &env = *((raw_dataset_transformer*)xptr);
    raw_dataset                     &ds_raw= env.ds_raw;
    const std::vector<bitmap::trf>  &trf = env.trf;
    for(int32_t entry_idx=env.entry_idx.fetch_add(1);
                entry_idx<ds_raw.entries.size();
                entry_idx=env.entry_idx.fetch_add(1))
    {
        raw_dataset::entry &e = ds_raw.entries[entry_idx];
        e.img = e.img.transform(trf[entry_idx]);
    }
    return NULL;
}

void raw_dataset::transform_inplace(const std::vector<bitmap::trf> &v)
{
    ASSERT(v.size()==entries.size());
    //for(int k=0; k<entries.size(); k++)
    //    entries[k].img = entries[k].img.transform(v[k]);
    raw_dataset_transformer env(*this, v);
    parallel_invoke(apply_transform_to_dataset, &env);
}

class raw_dataset_rsi_transformer
{
public:
    raw_dataset &ds_raw;
    const float deg_hue;
    const float deg_rot;
    const float scale_dev;
    const float fshift_mag;
    const bitmap::project_t prj;
    const int seed;
    std::atomic<int32_t> entry_idx;

    raw_dataset_rsi_transformer(raw_dataset &ds, float dh, float dr, float sm, float fs, bitmap::project_t p, int sd):
        ds_raw(ds),deg_hue(dh),deg_rot(dr),scale_dev(sm),fshift_mag(fs),prj(p),seed(sd),entry_idx(0){}
};

void* apply_rsi_transform_to_dataset(void *xptr)
{
    raw_dataset_rsi_transformer &env = *((raw_dataset_rsi_transformer*)xptr);
    raw_dataset                 &ds_raw= env.ds_raw;
    const float deg_hue   = env.deg_hue + 1.0e-20f;
    const float rad_rot   = env.deg_rot/180.0f*3.1415f + 1.0e-20f;
    const float scale_dev = env.scale_dev + 1.0e-20f;
    const float fshift    = env.fshift_mag + 1.0e-20f;

    ASSERT(env.prj==bitmap::project_t::mirror);

    for(int32_t entry_idx=env.entry_idx.fetch_add(1);
                entry_idx<ds_raw.ntrn;
                entry_idx=env.entry_idx.fetch_add(1))
    {
        raw_dataset::entry &e = ds_raw.entries[entry_idx];

        // RNG used to distort image
        std::mt19937 mt_rnd;
        std::normal_distribution<> _rnd_nrm(0, 1);
        auto truncated_normal = [&]()->float
        { float r = _rnd_nrm(mt_rnd); while( r<-3 || r>+3 ) r =_rnd_nrm(mt_rnd); return r; };
        auto generate_hue_shift = [&]()->float
        { return deg_hue*truncated_normal(); };
        auto generate_rot = [&]()->float
        { return rad_rot*truncated_normal(); };
        auto generate_scl = [&]()->float
        { return 1.0f+scale_dev*truncated_normal();  };
        auto generate_sft = [&]()->float
        { return fshift*truncated_normal(); };
        std::seed_seq mt_seed{(unsigned int)env.seed, (unsigned int)entry_idx};
        mt_rnd.seed(mt_seed);

        // distortion
        bitmap raw = e.img;
        bitmap &out = e.img;
        float c0 = 0.5f*(raw.height-1), c1 = 0.5f*(raw.width-1);
        float hue_shift = generate_hue_shift();
        float rot = generate_rot(), fshift0 = generate_sft(), fshift1 = generate_sft();
        float scale0 = generate_scl(), scale1 = generate_scl();
        float v_cos = cosf(rot), v_sin = sinf(rot);
        for(int dst_i0=0; dst_i0<out.height; dst_i0++)
            for(int dst_i1=0; dst_i1<out.width; dst_i1++)
            {
                // centered coordinates
                float x0 = dst_i0-c0, x1=dst_i1-c1;

                // rotation
                float t0 = x0, t1 = x1;
                x0 = t0*v_cos - t1*v_sin;
                x1 = t0*v_sin + t1*v_cos;

                // scaling
                x0 /= scale0;
                x1 /= scale1;

                // shifting
                x0 -= fshift0*(out.height-1);
                x1 -= fshift1*(out.width-1);

                // centering and projecting
                x0 = bitmap::project(x0+c0, env.prj, 0.0f, out.height-1);
                x1 = bitmap::project(x1+c1, env.prj, 0.0f, out.width-1);

                // reading color value, applying hue shift
                hsv v_hsv = hsvfromi24(raw.get_24_lanczos(x0, x1));
                v_hsv.h += hue_shift;
                while( v_hsv.h<0 || v_hsv.h>360.0f )
                {
                    if( v_hsv.h<0 )
                        v_hsv.h += 360.0f;
                    if( v_hsv.h>360.0f )
                        v_hsv.h -= 360.0f;
                }

                // output
                out.draw_24(dst_i0, dst_i1, i24fromhsv(v_hsv));
            }
    }
    return NULL;
}

void raw_dataset::rotate_shift_inplace(float deg_rot, float fshift_mag, bitmap::project_t prj, int seed)
{
    hue_rotate_scale_shift_inplace(0.0f, deg_rot, 0.0f, fshift_mag, prj, seed);
}

void raw_dataset::hue_rotate_scale_shift_inplace(float deg_hue, float deg_rot, float scale_dev, float fshift_mag, bitmap::project_t prj, int seed)
{
    raw_dataset_rsi_transformer env(*this, deg_hue, deg_rot, scale_dev, fshift_mag, prj, seed);
    parallel_invoke(apply_rsi_transform_to_dataset, &env);
}

raw_dataset raw_dataset::duplicate_trn(int factor) const
{
    ASSERT(factor>=1);
    raw_dataset r;
    r.nclasses = nclasses;
    for(int k=0; k<ntrn; k++)
        for(int t=0; t<factor; t++)
            r.append_trn(entries[k]);
    for(int k=0; k<ntst; k++)
        r.append_tst(entries[ntrn+k]);
    return r;
}

raw_dataset raw_dataset::duplicate_all(int factor) const
{
    ASSERT(factor>=1);
    raw_dataset r;
    r.nclasses = nclasses;
    for(int k=0; k<ntrn; k++)
        for(int t=0; t<factor; t++)
            r.append_trn(entries[k]);
    for(int k=0; k<ntst; k++)
        for(int t=0; t<factor; t++)
            r.append_tst(entries[ntrn+k]);
    return r;
}

raw_dataset raw_dataset::shift_by(int shift0, int shift1) const
{
    raw_dataset r = *this;
    for(entry &e: r.entries)
        e.img = e.img.shift_by(shift0, shift1);
    return r;
}

raw_dataset raw_dataset::subsample(int new_trn, int new_tst, int seed) const
{
    ASSERT(new_trn>=0 && new_trn<=ntrn && new_tst>=0 && new_tst<=ntst);

    //
    // Normalize
    //
    if( new_trn==0 )
        new_trn = ntrn;
    if( new_tst==0 )
        new_tst = ntst;

    //
    // Select subsample of dataset to process, no duplication
    //
    std::vector<int> final_sample, cur_sample;
    alglib::hqrndstate rs;
    alglib::hqrndseed(seed, seed+1001, rs);
    cur_sample = std::vector<int>(ntrn);
    std::iota(cur_sample.begin(), cur_sample.end(), 0);
    for(int i=0; i<ntrn; i++)
    {
        int j = i+alglib::hqrnduniformi(rs,ntrn-i);
        std::swap(cur_sample[i], cur_sample[j]);
    }
    for(int i=0; i<new_trn; i++)
        final_sample.push_back(cur_sample[i]);
    cur_sample = std::vector<int>(ntst);
    std::iota(cur_sample.begin(), cur_sample.end(), 0);
    for(int i=0; i<ntst; i++)
    {
        int j = i+alglib::hqrnduniformi(rs,ntst-i);
        std::swap(cur_sample[i], cur_sample[j]);
    }
    for(int i=0; i<new_tst; i++)
        final_sample.push_back(ntrn+cur_sample[i]);

    raw_dataset rr;
    rr.ntrn = new_trn;
    rr.ntst = new_tst;
    rr.nclasses = nclasses;
    for(int i=0; i<final_sample.size(); i++)
        rr.entries.push_back(entries[final_sample[i]]);
    return rr;
}

raw_dataset raw_dataset::augment(int afactor, float shift_mag, float scale_mag_lo, float deg_rot) const
{
    //
    // Integrity checks and normalization
    //
    ASSERT(afactor>=2);
    ASSERT(shift_mag>=0);
    ASSERT(scale_mag_lo>0 && scale_mag_lo<=1);
    ASSERT(deg_rot>=0);

    //
    // Prepare RNG
    //
    float rad_rot = deg_rot/180.f*3.1415f;
    std::uniform_real_distribution<> uniform_shift(-shift_mag, +shift_mag);
    std::uniform_real_distribution<> uniform_scale(scale_mag_lo, 1.0f);
    std::uniform_real_distribution<> uniform_rot(-rad_rot, +rad_rot);
    std::mt19937  mt_rnd;
    std::seed_seq mt_seed{(unsigned int)1, (unsigned int)2};
    mt_rnd.seed(mt_seed);


    //
    // Augment
    //
    raw_dataset ds_augmented;
    std::vector<bitmap::trf> transform;
    ds_augmented.nclasses = nclasses;
    for(int idx=0; idx<ntrn; idx++)
    {
        //
        // First image in augmentation group is added "as is"
        //
        ds_augmented.append_trn(entries[idx]);
        transform.push_back(bitmap::trf());

        //
        // Subsequent images are augmented
        //
        for(int aidx=1; aidx<afactor; aidx++)
        {
            // image
            ds_augmented.append_trn(entries[idx]);

            // transform
            bitmap::trf trf;
            if( shift_mag>0 )
                trf.set_fshift(uniform_shift(mt_rnd), uniform_shift(mt_rnd));
            if( scale_mag_lo<1 )
                trf.set_scale(uniform_scale(mt_rnd));
            if( deg_rot>0 )
                trf.set_rot(uniform_rot(mt_rnd));
            transform.push_back(trf);
        }
    }
    for(int idx=0; idx<ntst; idx++)
    {
        ds_augmented.append_tst(entries[ntrn+idx]);
        transform.push_back(bitmap::trf());
    }
    ds_augmented.transform_inplace(transform);
    return ds_augmented;
}

raw_dataset raw_dataset::augment_hflip() const
{
    //
    // Augment
    //
    raw_dataset ds_augmented(nclasses);
    for(int idx=0; idx<ntrn; idx++)
    {
        bitmap::trf trf_hflip;
        trf_hflip.set_flips(true, false);

        entry e = entries[idx];
        ds_augmented.append_trn(e);
        e.img = e.img.transform(trf_hflip);
        ds_augmented.append_trn(e);
    }
    for(int idx=0; idx<ntst; idx++)
        ds_augmented.append_tst(entries[ntrn+idx]);
    return ds_augmented;
}

raw_dataset raw_dataset::augment_hflip_crop(int afactor, int crop_to) const
{
    //
    // Integrity checks and normalization
    //
    ASSERT(afactor>=1);
    for(const entry &e: entries)
    {
        ASSERT(e.img.width>=crop_to);
        ASSERT(e.img.height>=crop_to);
    }

    //
    // Prepare RNG
    //
    std::mt19937  mt_rnd;
    std::seed_seq mt_seed{(unsigned int)1, (unsigned int)2};
    mt_rnd.seed(mt_seed);


    //
    // Augment
    //
    raw_dataset ds_augmented;
    std::vector<bitmap::trf> transform;
    ds_augmented.nclasses = nclasses;
    for(int idx=0; idx<ntrn; idx++)
    {
        const bitmap &b = entries[idx].img;

        //
        // First image in augmentation group is added "as is"
        //
        bitmap::trf trf;
        trf.set_crop(crop_to, crop_to, (b.height-crop_to)/2, (b.width-crop_to)/2);
        ds_augmented.append_trn(entries[idx]);
        transform.push_back(trf);

        //
        // Subsequent images are augmented
        //
        for(int aidx=1; aidx<afactor; aidx++)
        {
            // image
            ds_augmented.append_trn(entries[idx]);

            // transform
            std::uniform_int_distribution<> uniform_shift0(0, b.height-crop_to);
            std::uniform_int_distribution<> uniform_shift1(0, b.width-crop_to);
            bitmap::trf trf;
            trf.set_flips(aidx%2==1, false);
            trf.set_crop(crop_to, crop_to, uniform_shift0(mt_rnd), uniform_shift1(mt_rnd));
            transform.push_back(trf);
        }
    }
    for(int idx=0; idx<ntst; idx++)
    {
        const bitmap &b = entries[ntrn+idx].img;
        ds_augmented.append_tst(entries[ntrn+idx]);
        bitmap::trf trf;
        trf.set_crop(crop_to, crop_to, (b.height-crop_to)/2, (b.width-crop_to)/2);
        transform.push_back(trf);
    }
    ds_augmented.transform_inplace(transform);
    return ds_augmented;
}

raw_dataset raw_dataset::augment_hflip_crop_rot(int afactor, int crop_to, float deg_rot) const
{
    float rad_rot = deg_rot/180.f*3.1415f;

    //
    // Integrity checks and normalization
    //
    ASSERT(afactor>=1);
    for(const entry &e: entries)
    {
        ASSERT(e.img.width>=crop_to);
        ASSERT(e.img.height>=crop_to);
    }

    //
    // Prepare RNG
    //
    std::mt19937  mt_rnd;
    std::seed_seq mt_seed{(unsigned int)1, (unsigned int)2};
    mt_rnd.seed(mt_seed);


    //
    // Augment
    //
    raw_dataset ds_augmented;
    std::vector<bitmap::trf> transform;
    ds_augmented.nclasses = nclasses;
    for(int idx=0; idx<ntrn; idx++)
    {
        const bitmap &b = entries[idx].img;

        //
        // First image in augmentation group is added "as is"
        //
        bitmap::trf trf;
        trf.set_crop(crop_to, crop_to, (b.height-crop_to)/2, (b.width-crop_to)/2);
        ds_augmented.append_trn(entries[idx]);
        transform.push_back(trf);

        //
        // Subsequent images are augmented
        //
        for(int aidx=1; aidx<afactor; aidx++)
        {
            // image
            ds_augmented.append_trn(entries[idx]);

            // transform
            std::uniform_int_distribution<> uniform_shift0(0, b.height-crop_to);
            std::uniform_int_distribution<> uniform_shift1(0, b.width-crop_to);
            std::uniform_real_distribution<> uniform_rot(-rad_rot, +rad_rot);
            bitmap::trf trf;
            trf.set_flips(aidx%2==1, false);
            trf.set_crop(crop_to, crop_to, uniform_shift0(mt_rnd), uniform_shift1(mt_rnd));
            trf.set_rot(uniform_rot(mt_rnd));
            transform.push_back(trf);
        }
    }
    for(int idx=0; idx<ntst; idx++)
    {
        const bitmap &b = entries[ntrn+idx].img;
        ds_augmented.append_tst(entries[ntrn+idx]);
        bitmap::trf trf;
        trf.set_crop(crop_to, crop_to, (b.height-crop_to)/2, (b.width-crop_to)/2);
        transform.push_back(trf);
    }
    ds_augmented.transform_inplace(transform);
    return ds_augmented;
}

raw_dataset raw_dataset::augment_shear(int afactor, float mag, float deg_rot) const
{
    //
    // Integrity checks and normalization
    //
    ASSERT(afactor>=1);
    ASSERT(mag>=0);
    ASSERT(deg_rot>=0);

    //
    // RNG
    //
    mag     += 1.0e-20f;
    deg_rot += 1.0e-20f;
    std::uniform_real_distribution<> rnd_mag(mag/3, mag);
    std::uniform_real_distribution<> rnd_rot(-deg_rot/180*3.1415f, +deg_rot/180*3.1415f);
    std::uniform_int_distribution<>  rnd_dir(0, 1);
    std::uniform_int_distribution<>  rnd_shear_kind(0, 2);
    std::uniform_int_distribution<>  rnd_rot_kind(0, 1);
    std::mt19937  mt_rnd;
    std::seed_seq mt_seed{(unsigned int)1, (unsigned int)2};
    mt_rnd.seed(mt_seed);

    //
    // Augment
    //
    raw_dataset ds_augmented;
    ds_augmented.nclasses = nclasses;
    for(int idx=0; idx<ntrn; idx++)
    {
        const bitmap &b = entries[idx].img;

        //
        // First image in augmentation group is added "as is"
        //
        ds_augmented.append_trn(entries[idx]);

        //
        // Subsequent images are augmented
        //
        for(int aidx=1; aidx<afactor; aidx++)
        {
            entry e = entries[idx];
            bitmap img_raw = e.img;

            int   dist_kind = rnd_shear_kind(mt_rnd);
            float dist_mag0 = rnd_mag(mt_rnd)*(2*rnd_dir(mt_rnd)-1)*(dist_kind!=1 ? 1 : 0);
            float dist_mag1 = rnd_mag(mt_rnd)*(2*rnd_dir(mt_rnd)-1)*(dist_kind!=0 ? 1 : 0);
            float rad_rot = rnd_rot(mt_rnd)*rnd_rot_kind(mt_rnd);
            float rot_c = cosf(rad_rot), rot_s = sinf(rad_rot);
            for(int i0=0; i0<img_raw.height; i0++)
                for(int i1=0; i1<img_raw.width; i1++)
                {
                    float x0=i0/float(img_raw.height-1), x1=i1/float(img_raw.width-1);

                    // rotate
                    x0 -= 0.5f;
                    x1 -= 0.5f;
                    float r0 = x0*rot_c+x1*rot_s, r1 = x0*(-rot_s)+x1*rot_c;
                    x0 = r0+0.5f;
                    x1 = r1+0.5f;
                    x0 = fmaxf(x0, 0);
                    x0 = fminf(x0, 1);
                    x1 = fmaxf(x1, 0);
                    x1 = fminf(x1, 1);

                    // then shear
                    x0 = x0+dist_mag0*x0*(1-x0);
                    x1 = x1+dist_mag1*x1*(1-x1);

                    // read/write
                    e.img.draw_24(i0, i1, img_raw.get_24_lanczos(x0*(img_raw.height-1), x1*(img_raw.width-1)));
                }
            ds_augmented.append_trn(e);
        }
    }
    for(int idx=0; idx<ntst; idx++)
        ds_augmented.append_tst(entries[ntrn+idx]);
    return ds_augmented;
}

class rds_affine_augmenter
{
public:
    raw_dataset &ds_aug;
    std::vector<std::vector<float>> distortions;
    std::atomic<int32_t> entry_idx;
    const int batch_size;
    const float mag_ctr;
    const float mag_sft;

    rds_affine_augmenter(raw_dataset &ds, int bs, float mc, float ms, int seed):
        ds_aug(ds),batch_size(bs),mag_ctr(mc),mag_sft(ms),entry_idx(0)
    {
        const int NDIST = 12;
        //const int NDIST = 8;

        std::mt19937  mt_rnd;
        std::seed_seq mt_seed{(unsigned int)seed};
        mt_rnd.seed(mt_seed);
        std::uniform_int_distribution<>  rnd_ctri(0,NDIST-1);
        std::uniform_real_distribution<> rnd_ctrm(-mag_ctr,+mag_ctr);
        std::uniform_real_distribution<>  rnd_sft(-mag_sft,+mag_sft);
        std::uniform_int_distribution<>  rnd_01(0,1);
        distortions.resize(ds_aug.entries.size());
        for(int i=0; i<ds_aug.entries.size(); i++)
        {

            // produce distorions matrix
            const float mag = rnd_ctrm(mt_rnd);
            const float dist_mat[4*NDIST] =
                {
                cosf(mag),  sinf(mag), -sinf(mag), cosf(mag),     // rotate by +mag
                cosf(mag), -sinf(mag),  sinf(mag), cosf(mag),     // rotate by -mag
                +1.0f,     0.0f, +mag, +1.0f,     // horizontal skew
                +1.0f,     0.0f, -mag, +1.0f,     // horizontal skew
                +1.0f,     +mag, 0.0f, +1.0f,     // vertical skew
                +1.0f,     -mag, 0.0f, +1.0f,     // vertical skew
                +1.0f,     0.0f, 0.0f, +1.0f+mag, // horizontal scale
                +1.0f,     0.0f, 0.0f, +1.0f-mag, // horizontal scale
                +1.0f+mag, 0.0f, 0.0f, +1.0f,     // vertical scale
                +1.0f-mag, 0.0f, 0.0f, +1.0f,     // vertical scale
                +1.0f+mag, 0.0f, 0.0f, +1.0f+mag, // grand scale
                +1.0f-mag, 0.0f, 0.0f, +1.0f-mag, // grand scale
                };

            //
            std::vector<float>       &d = distortions[i];
            const raw_dataset::entry &e = ds_aug.entries[i];
            const float *trf = dist_mat+rnd_ctri(mt_rnd)*4;
            const float sft_flg = rnd_01(mt_rnd);
            const float sft0 = sft_flg*rnd_sft(mt_rnd)*e.img.width;
            const float sft1 = sft_flg*rnd_sft(mt_rnd)*e.img.height;
            d.resize(6);
            d[0] = trf[0];
            d[1] = trf[1];
            d[2] = trf[2];
            d[3] = trf[3];
            d[4] = sft0;
            d[5] = sft1;
        }
    }
};

void* rds_apply_affine_augmentation(void *xptr)
{
    rds_affine_augmenter &env = *((rds_affine_augmenter*)xptr);
    for(int32_t bidx=env.entry_idx.fetch_add(1);
                bidx<env.ds_aug.entries.size()/env.batch_size;
                bidx=env.entry_idx.fetch_add(1))
    {
        int img_idx = bidx*env.batch_size;

        // apply transformation (first image is not changed)
        for(int t=1; t<env.batch_size; t++)
        {
            raw_dataset::entry &e = env.ds_aug.entries[img_idx+t];
            bitmap img_new = e.img;
            const float c0 = 0.5f*(e.img.height-1), c1 = 0.5f*(e.img.width-1);
            for(int i=0; i<e.img.height; i++)
                for(int j=0; j<e.img.width; j++)
                {
                    const float *trf = &env.distortions[img_idx+t][0];
                    float x0 = i-c0, x1 = j-c1;
                    float y0 = trf[0]*x0+trf[1]*x1, y1 = trf[2]*x0+trf[3]*x1;
                    y0 += trf[4];
                    y1 += trf[5];
                    y0 += c0;
                    y1 += c1;
                    img_new.draw_24(i, j, e.img.get_24_lanczos(y0,y1));
                }
            e.img = img_new;
        }
    }
    return NULL;
}

raw_dataset raw_dataset::augment_affine(int afactor, std::tuple<float,float> mag_ctr_sft, int seed) const
{
    // quick exit
    if( afactor==1 )
        return *this;

    raw_dataset out;
    out.nclasses = nclasses;
    for(int img_idx=0; img_idx<ntrn; img_idx++)
        for(int t=0; t<afactor; t++)
            out.append_trn(entries[img_idx]);
    rds_affine_augmenter env(out,afactor,std::get<0>(mag_ctr_sft),std::get<1>(mag_ctr_sft),seed);
    parallel_invoke(rds_apply_affine_augmentation, &env);
    for(int img_idx=ntrn; img_idx<ntrn+ntst; img_idx++)
        out.append_tst(entries[img_idx]);
    return out;
}


raw_dataset raw_dataset::downsample2() const
{
    raw_dataset r = *this;
    for(raw_dataset::entry &e: r.entries)
        e.img = e.img.downsample2();
    return r;
}

raw_dataset raw_dataset::multiply_class_trn(int class_idx, int factor) const
{
    raw_dataset r;
    r.nclasses = nclasses;
    ASSERT(factor>=1);

    // training set
    for(int k=0; k<ntrn; k++)
    {
        int cnt = entries[k].class_id==class_idx ? factor : 1;
        for(int t=0; t<cnt; t++)
            r.append_trn(entries[k]);
    }

    // test set
    for(int k=0; k<ntst; k++)
        r.append_tst(entries[ntrn+k]);

    // done
    return r;

}

raw_dataset raw_dataset::normalize_contrast() const
{
    raw_dataset r = *this;
    for(raw_dataset::entry &e: r.entries)
        e.img = e.img.normalize_contrast();
    return r;
}

raw_dataset raw_dataset::reorder_class_to_beginning(int class_idx) const
{
    raw_dataset r;
    r.nclasses = nclasses;

    // training set
    for(int k=0; k<ntrn; k++)
        if( entries[k].class_id==class_idx )
            r.append_trn(entries[k]);
    for(int k=0; k<ntrn; k++)
        if( entries[k].class_id!=class_idx )
            r.append_trn(entries[k]);

    // test set
    for(int k=0; k<ntst; k++)
        if( entries[ntrn+k].class_id==class_idx )
            r.append_tst(entries[ntrn+k]);
    for(int k=0; k<ntst; k++)
        if( entries[ntrn+k].class_id!=class_idx )
            r.append_tst(entries[ntrn+k]);
    return r;
}

raw_dataset raw_dataset::select_entries_of_class(int c) const
{
    ASSERT(c>=0 && c<nclasses);
    raw_dataset r(nclasses);
    for(int i=0; i<ntrn+ntst; i++)
        if( entries[i].class_id==c )
        {
            if( i<ntrn )
                r.append_trn(entries[i]);
            else
                r.append_tst(entries[i]);
        }
    return r;
}

raw_dataset raw_dataset::select_classes(std::vector<int> classes) const
{
    ASSERT(classes.size()>=1);

    raw_dataset r;
    r.nclasses = classes.size();
    for(int i=0; i<ntrn+ntst; i++)
    {
        int class_idx = std::find(classes.begin(), classes.end(), entries[i].class_id)-classes.begin();
        if( class_idx==classes.size() )
            continue;
        entry e = entries[i];
        e.class_id = class_idx;
        if(i<ntrn )
            r.append_trn(e);
        else
            r.append_tst(e);
    }
    return r;
}

raw_dataset raw_dataset::select_subsample(const std::vector<int> &idx) const
{
    // integrity checks
    if( idx.size()>=2 )
    {
        for(int i=1; i<idx.size(); i++)
            ASSERT(idx[i-1]<idx[i] && idx[i-1]>=0 && idx[i]<ntrn+ntst);
    }

    // output
    raw_dataset r;
    r.ntrn = 0;
    r.ntst = 0;
    r.nclasses = nclasses;
    for(int i=0; i<idx.size(); i++)
        if( idx[i]<ntrn )
            r.append_trn(entries[idx[i]]);
        else
            r.append_tst(entries[idx[i]]);
    return r;
}

raw_dataset raw_dataset::select_trn() const
{
    raw_dataset r;
    r.nclasses = nclasses;
    for(int i=0; i<ntrn; i++)
        r.append_trn(entries[i]);
    return r;
}

raw_dataset raw_dataset::select_tst() const
{
    raw_dataset r;
    r.nclasses = nclasses;
    for(int i=ntrn; i<ntrn+ntst; i++)
        r.append_tst(entries[i]);
    return r;
}

raw_dataset raw_dataset::set_class_to(int cls) const
{
    raw_dataset r = *this;
    for(raw_dataset::entry &e: r.entries)
        e.class_id = cls;
    return r;
}

raw_dataset raw_dataset::shuffle(int seed) const
{
    std::mt19937  mt_rnd;
    std::seed_seq mt_seed{(unsigned int)seed};
    mt_rnd.seed(mt_seed);

    raw_dataset r = *this;
    std::shuffle(r.entries.begin(), r.entries.begin()+ntrn, mt_rnd);
    std::shuffle(r.entries.begin()+ntrn, r.entries.end(), mt_rnd);
    return r;
}

raw_dataset raw_dataset::shuffle_trn(int seed) const
{
    std::mt19937  mt_rnd;
    std::seed_seq mt_seed{(unsigned int)seed};
    mt_rnd.seed(mt_seed);

    raw_dataset r = *this;
    std::shuffle(r.entries.begin(), r.entries.begin()+ntrn, mt_rnd);
    return r;
}

raw_dataset raw_dataset::select_kth_fold(int folds_count, int fold) const
{
    ASSERT(folds_count>=1 && fold>=0 && fold<folds_count);
    raw_dataset out;
    out.nclasses = nclasses;
    for(int i=0; i<ntrn+ntst; i++)
        if( i%folds_count==fold )
            if( i<ntrn )
                out.append_trn(entries[i]);
            else
                out.append_tst(entries[i]);
    return out;
}

raw_dataset raw_dataset::recombine_folds(const std::vector<raw_dataset> &datasets)
{
    ASSERT(datasets.size()>=1);
    for(const raw_dataset &ds: datasets)
        ASSERT(ds.nclasses==datasets[0].nclasses);

    int total_trn = 0, total_tst = 0;
    for(const raw_dataset &ds: datasets)
    {
        total_trn += ds.ntrn;
        total_tst += ds.ntst;
    }
    const int folds_count = datasets.size();
    std::vector<bool> is_set(total_trn+total_tst,false);
    raw_dataset out;
    out.nclasses = datasets[0].nclasses;
    out.entries.resize(total_trn+total_tst);
    out.ntrn = total_trn;
    out.ntst = total_tst;
    for(int fold=0; fold<datasets.size(); fold++)
    {
        const raw_dataset &ds = datasets[fold];
        for(int i=0; i<ds.ntrn+ds.ntst; i++)
        {
            const int dst = i*folds_count+fold;
            ASSERT(dst<total_trn+total_tst && !is_set[dst]);
            out.entries[dst] = ds.entries[i];
            is_set[dst] = true;
        }
    }
    return out;
}

void serialize(FILE *f, const raw_dataset::entry *ptr)
{
    ASSERT(ptr->flavor.length()==0); // present version does not support serialization with non-empty flavor
    serialize(f, &ptr->img);
    serialize(f, &ptr->class_id);
}

void serialize(FILE *f, const raw_dataset *ptr)
{
    serialize(f, &ptr->ntrn);
    serialize(f, &ptr->ntst);
    serialize(f, &ptr->nclasses);
    serialize(f, &ptr->entries);
}

void unserialize(FILE *f, raw_dataset::entry *ptr)
{
    *ptr = raw_dataset::entry();
    unserialize(f, &ptr->img);
    unserialize(f, &ptr->class_id);
}

void unserialize(FILE *f, raw_dataset *ptr)
{
    unserialize(f, &ptr->ntrn);
    unserialize(f, &ptr->ntst);
    unserialize(f, &ptr->nclasses);
    unserialize(f, &ptr->entries);
}

calibrator::calibrator()
{
}

calibrator::calibrator(const std::vector<float> &_x, const std::vector<float> &_y):
    x(_x),y(_y)
{
    ASSERT(x.size()==y.size());
    for(int k=0; k<x.size()-1; k++)
        ASSERT(x[k]<x[k+1]);
}

bool  calibrator::is_identity() const
{
    return x.size()==0 && y.size()==0;
}

float calibrator::calibrate(float vx) const
{
    if( x.size()==0 && y.size()==0 )
        return vx;

    ASSERT(x.size()==y.size());
    const int nbins = x.size();
    if( vx<=x[0] )
        return y[0];
    if( vx>=x[nbins-1] )
        return y[nbins-1];
    for(int k=0; k<nbins-1; k++)
        if( vx>=x[k] && vx<=x[k+1] )
        {
            float t = (vx-x[k])/(x[k+1]-x[k]);
            return (1-t)*y[k] + t*y[k+1];
        }
    ASSERT(0);
}

void serialize(FILE *f, const calibrator *ptr)
{
    write_magic8(f, 217);
    serialize(f, &ptr->x);
    serialize(f, &ptr->y);
    write_magic8(f, 192);
}

void unserialize(FILE *f, calibrator *ptr)
{
    check_magic8(f, 217);
    unserialize(f, &ptr->x);
    unserialize(f, &ptr->y);
    check_magic8(f, 192);
}

predictions::predictions(int64_t trn, int64_t tst, int64_t classes, const std::vector<int> &lbl)
{
    ntrn = trn;
    ntst = tst;
    nclasses = classes;
    stride = trn+tst;
    labels = lbl;
    fpredictions.resize(nclasses*stride, 0.0f);
}

predictions predictions::zero_baseline(int64_t trn, int64_t tst, int64_t classes)
{
    predictions r;
    r.ntrn = trn;
    r.ntst = tst;
    r.nclasses = classes;
    r.stride = trn+tst;
    r.labels.resize(trn+tst, 0);
    r.fpredictions.resize(classes*r.stride, 0.0f);
    return r;
}

predictions::predictions(const raw_dataset &ds)
{
    ntrn = ds.ntrn;
    ntst = ds.ntst;
    nclasses = ds.nclasses;
    stride = ds.ntrn+ds.ntst;

    //
    labels.resize(ntrn+ntst);
    for(int i=0; i<ntrn+ntst; i++)
        labels[i] = ds.entries[i].class_id;
    fpredictions.resize(nclasses*stride, 0.0f);
}

predictions::predictions(std::string filename)
{
    unserialize_from(filename, this);
}

void predictions::append(const predictions &pred)
{
    ASSERT(nclasses==pred.nclasses);

    int64_t new_trn = ntrn+pred.ntrn;
    int64_t new_tst = ntst+pred.ntst;
    int64_t new_stride = new_trn+new_tst;
    std::vector<int> new_labels(new_stride,-1);
    fvector new_fpredictions(new_stride*nclasses,0);

    // present activities
    for(int64_t i=0; i<ntrn; i++)
        new_labels[i] = labels[i];
    for(int64_t i=0; i<ntst; i++)
        new_labels[new_trn+i] = labels[ntrn+i];
    for(int64_t k=0; k<nclasses; k++)
    {
        for(int64_t i=0; i<ntrn; i++)
            new_fpredictions[k*new_stride+i] = predicted_prob(i,k);
        for(int64_t i=0; i<ntst; i++)
            new_fpredictions[k*new_stride+new_trn+i] = predicted_prob(ntrn+i,k);
    }

    // appended activities
    for(int64_t i=0; i<pred.ntrn; i++)
        new_labels[ntrn+i] = pred.labels[i];
    for(int64_t i=0; i<pred.ntst; i++)
        new_labels[new_trn+ntst+i] = pred.labels[pred.ntrn+i];
    for(int64_t k=0; k<nclasses; k++)
    {
        for(int64_t i=0; i<pred.ntrn; i++)
            new_fpredictions[k*new_stride+ntrn+i] = pred.predicted_prob(i,k);
        for(int64_t i=0; i<pred.ntst; i++)
            new_fpredictions[k*new_stride+new_trn+ntst+i] = pred.predicted_prob(pred.ntrn+i,k);
    }

    // replace
    ntrn = new_trn;
    ntst = new_tst;
    stride = new_stride;
    labels = new_labels;
    fpredictions = new_fpredictions;
}

activityf_vector predictions::kth_as_activity(int64_t k) const
{
    ASSERT(k>=0 && k<nclasses);
    const float *p = &fpredictions[k*stride];
    return activityf_vector(p, p+ntrn+ntst);
}

std::vector<float> predictions::kth_as_fvector(int64_t k) const
{
    ASSERT(k>=0 && k<nclasses);
    const float *p = &fpredictions[k*stride];
    return std::vector<float>(p, p+ntrn+ntst);
}

void predictions::compute_errors(float *trn_rms, float *tst_rms, float *tst_cls, std::vector<int> *p_class_errors) const
{
    alglib::real_1d_array probs;
    probs.setlength(nclasses);
    *trn_rms = 0;
    *tst_rms = 0;
    *tst_cls = 0;
    for(int64_t k=0; k<nclasses; k++)
        for(int64_t idx=0; idx<ntrn; idx++)
        {
            float prediction = fpredictions[idx+k*stride];
            float target     = labels[idx]==k ? 1.0f : 0.0f;
            float residual   = prediction-target;
            *trn_rms += residual*residual;
        }
    *trn_rms = sqrt(*trn_rms/(ntrn*nclasses));
    if( p_class_errors!=NULL )
    {
        p_class_errors->clear();
        p_class_errors->resize(nclasses, 0);
    }
    for(int64_t idx=0; idx<ntst; idx++)
    {
        for(int64_t k=0; k<nclasses; k++)
            probs(k) = fpredictions[(ntrn+idx)+k*stride];
        int64_t mx = 0;
        int itarget = labels[ntrn+idx];
        for(int64_t k=0; k<nclasses; k++)
        {
            *tst_rms += pow(probs(k)-(k==itarget ? 1 : 0),2.0);
            if( probs(k)>probs(mx) )
                mx = k;
        }
        *tst_cls += mx!=itarget ? 1 : 0;
        ASSERT(itarget>=0 && itarget<nclasses);
        if( p_class_errors!=NULL && mx!=itarget )
            (*p_class_errors)[itarget]++;
    }
    *tst_rms = sqrt(*tst_rms/(ntst*nclasses));
    *tst_cls = *tst_cls/ntst;
}

float predictions::trn_rms() const
{
    float trn, tst, cls;
    compute_errors(&trn, &tst, &cls);
    return trn;
}

float predictions::tst_rms() const
{
    float trn, tst, cls;
    compute_errors(&trn, &tst, &cls);
    return tst;
}

float predictions::tst_cls() const
{
    float trn, tst, cls;
    compute_errors(&trn, &tst, &cls);
    return cls;
}

int predictions::tst_false_positives(int class_idx) const
{
    ASSERT(class_idx>=0 && class_idx<nclasses);
    int fp = 0;
    for(int64_t idx=0; idx<ntst; idx++)
    {
        float probs[nclasses];
        for(int64_t k=0; k<nclasses; k++)
            probs[k] = fpredictions[(ntrn+idx)+k*stride];
        int64_t mx = 0;
        int itarget = labels[ntrn+idx];
        for(int64_t k=0; k<nclasses; k++)
            if( probs[k]>probs[mx] )
                mx = k;
        if( itarget!=class_idx && mx==class_idx )
            fp++;
    }
    return fp;
}

int predictions::tst_false_negatives(int class_idx) const
{
    ASSERT(class_idx>=0 && class_idx<nclasses);
    int fn = 0;
    for(int64_t idx=0; idx<ntst; idx++)
    {
        float probs[nclasses];
        for(int64_t k=0; k<nclasses; k++)
            probs[k] = fpredictions[(ntrn+idx)+k*stride];
        int64_t mx = 0;
        int itarget = labels[ntrn+idx];
        for(int64_t k=0; k<nclasses; k++)
            if( probs[k]>probs[mx] )
                mx = k;
        if( itarget==class_idx && mx!=class_idx )
            fn++;
    }
    return fn;
}

predictions::tfpn_rec predictions::tst_class_detection_by_threshold(float threshold, int class_idx) const
{
    ASSERT(class_idx>=0 && class_idx<nclasses);

    tfpn_rec rec;
    memset(&rec, 0, sizeof(rec));

    const float *p_probs = &fpredictions[ntrn+class_idx*stride];
    for(int64_t idx=0; idx<ntst; idx++)
    {
        bool is_target = labels[ntrn+idx]==class_idx;
        bool is_pred   = p_probs[idx]>threshold;

        if( is_target && is_pred )
            rec.true_pos_cnt++;
        if( !is_target && !is_pred )
            rec.true_neg_cnt++;
        if( is_target && !is_pred )
            rec.false_neg_cnt++;
        if( !is_target && is_pred )
            rec.false_pos_cnt++;
        rec.total_cnt++;
    }
    rec.true_pos_rate = rec.true_pos_cnt/float(1.0e-20f+rec.true_pos_cnt+rec.false_neg_cnt);
    rec.true_neg_rate = rec.true_neg_cnt/float(1.0e-20f+rec.true_neg_cnt+rec.false_pos_cnt);
    rec.false_neg_rate = rec.false_neg_cnt/float(1.0e-20f+rec.true_pos_cnt+rec.false_neg_cnt);
    rec.false_pos_rate = rec.false_pos_cnt/float(1.0e-20f+rec.true_neg_cnt+rec.false_pos_cnt);
    return rec;
}

void predictions::output_bad_cases(const raw_dataset &raw_ds, std::string dir) const
{
    ASSERT(nclasses<=1024);
    float probs[nclasses];
    int cls_err = 0;
    for(int64_t idx=0; idx<ntst; idx++)
    {
        int itarget = labels[ntrn+idx];
        for(int64_t k=0; k<nclasses; k++)
            probs[k] = fpredictions[(ntrn+idx)+k*stride];
        int64_t mx = 0;
        for(int64_t k=0; k<nclasses; k++)
            if( probs[k]>probs[mx] )
                mx = k;
        if( mx!=itarget )
        {
            char filename[1024];
            int p100 = irange(roundf(probs[mx]*100), 0, 100);
            sprintf(filename, "%s/class%02d_tst%06d_as%02d_p%02d.png", dir.c_str(), (int)itarget, (int)idx, (int)mx, (int)p100);
            raw_ds.entries[ntrn+idx].img.to_png(filename);
            cls_err++;
        }
    }
    printf("ClsErr=%.4f\n", double(cls_err)/ntst);
}

void predictions::print_calibration_curve(int class_idx) const
{
    const int nbins = 21;
    std::vector<int> counts(nbins, 0), totals(nbins,0);
    const float *p_pred = &fpredictions[class_idx*stride];
    for(int64_t k=0; k<ntrn; k++)
    {
        int bin_idx = irange(roundf(p_pred[k]*(nbins-1)), 0, nbins-1);
        if( labels[k]==class_idx )
            counts[bin_idx]++;
        totals[bin_idx]++;
    }
    for(int k=0; k<nbins; k++)
    {
        printf("%6d %6d      ", counts[k], totals[k]);
        printf("%.4f\n", double(counts[k])/double(totals[k]+1));
    }
}

float predictions::predicted_prob(int img_idx, int class_idx) const
{
    ASSERT(img_idx>=0 && img_idx<ntrn+ntst);
    ASSERT(class_idx>=0 && class_idx<nclasses);
    return fpredictions[img_idx+class_idx*stride];
}

std::vector<int> predictions::predicted_classes() const
{
    std::vector<int> r;
    ASSERT(nclasses<=1024);
    float probs[nclasses];
    for(int64_t idx=0; idx<ntrn+ntst; idx++)
    {
        for(int64_t k=0; k<nclasses; k++)
            probs[k] = fpredictions[idx+k*stride];
        int64_t mx = 0;
        for(int64_t k=0; k<nclasses; k++)
            if( probs[k]>probs[mx] )
                mx = k;
        r.push_back(mx);
    }
    return r;
}

predictions predictions::select_subsample(const std::vector<int> items) const
{
    // integrity checks
    ASSERT(items.size()>=1);
    for(int k=1; k<items.size(); k++)
    {
        ASSERT(items[k-1]>=0 && items[k-1]<ntrn+ntst);
        ASSERT(items[k]>=0   && items[k]<ntrn+ntst);
        ASSERT(items[k-1]<=items[k]);
    }

    // determine output sample size, prepare labels
    int new_trn=0, new_tst=0;
    std::vector<int> new_labels;
    for(int k=0; k<items.size(); k++)
    {
        if( items[k]<ntrn )
            new_trn++;
        else
            new_tst++;
        new_labels.push_back(labels[items[k]]);
    }

    // allocate new output, copy data
    predictions r(new_trn, new_tst, nclasses, new_labels);
    for(int k=0; k<items.size(); k++)
    {
        int j = items[k];
        for(int i=0; i<nclasses; i++)
            r.fpredictions[k+i*r.stride] = fpredictions[j+i*stride];
    }
    return r;
}

std::vector<calibrator> predictions::recalibrate(int nbins)
{
    ASSERT(nbins>=2);
    std::vector<calibrator> result;
    for(int64_t class_idx=0; class_idx<nclasses; class_idx++)
    {
        std::vector<float> vx, vy;
        float *p_pred = &fpredictions[class_idx*stride];

        //
        // Build calibrator for idx-th class
        //
        std::vector<int> counts(nbins, 0), totals(nbins,0);
        for(int64_t k=0; k<ntrn; k++)
        {
            int bin_idx = irange(roundf(p_pred[k]*(nbins-1)), 0, nbins-1);
            if( labels[k]==class_idx )
                counts[bin_idx]++;
            totals[bin_idx]++;
        }
        alglib::spline1dinterpolant scalibrator;
        alglib::real_1d_array cx, cy;
        cx.setlength(nbins);
        cy.setlength(nbins);
        for(int k=0; k<nbins; k++)
        {
            cx(k) = double(k)/double(nbins-1);
            cy(k) = double(counts[k])/double(totals[k]+1);
            vx.push_back(cx(k));
            vy.push_back(cy(k));
        }
        alglib::spline1dbuildlinear(cx, cy, scalibrator);

        //
        // Calibrate
        //
        for(int64_t k=0; k<ntrn+ntst; k++)
            p_pred[k] = alglib::spline1dcalc(scalibrator, p_pred[k]);
        result.push_back(calibrator(vx,vy));
    }
    return result;
}

predictions predictions::calibrate_onevsall(const predictions &p, const calibrator &c)
{
    predictions r = p;
    ASSERT(r.nclasses==2);
    for(int64_t i=0; i<r.ntrn+r.ntst; i++)
    {
        float v = c.calibrate(r.fpredictions[i]);
        r.fpredictions[i+0*r.stride] = v;
        r.fpredictions[i+1*r.stride] = 1-v;
    }
    return r;
}

predictions predictions::calibrate(const predictions &p, const std::vector<calibrator> &c)
{
    predictions r = p;
    ASSERT(r.nclasses==c.size());
    for(int k=0; k<r.nclasses; k++)
        for(int64_t i=0; i<r.ntrn+r.ntst; i++)
            r.fpredictions[i+k*r.stride] = c[k].calibrate(r.fpredictions[i+k*r.stride]);
    return r;
}

predictions predictions::average(const std::vector<predictions> &vec)
{
    predictions r;

    // integrity checks
    ASSERT(vec.size()>=1);
    for(const predictions &p: vec)
    {
        ASSERT(p.ntrn==vec[0].ntrn);
        ASSERT(p.ntst==vec[0].ntst);
        ASSERT(p.nclasses==vec[0].nclasses);
        ASSERT(p.stride==vec[0].stride);
        ASSERT(p.fpredictions.size()==vec[0].fpredictions.size());
    }

    // average
    r = predictions(vec[0].ntrn, vec[0].ntst, vec[0].nclasses, vec[0].labels);
    ASSERT(r.fpredictions.size()==vec[0].fpredictions.size());
    for(const predictions &p: vec)
    {
        const int64_t cnt    = p.fpredictions.size();
        const float   *p_src = &p.fpredictions[0];
        float         *p_dst = &r.fpredictions[0];
        float v = 1.0f/vec.size();
        for(int64_t k=0; k<cnt; k++)
            p_dst[k] += v*p_src[k];
    }
    return r;
}

predictions predictions::onevsall(const predictions &all, int class_id)
{
    predictions out;
    std::vector<int> lbl_out;

    // integrity checks
    ASSERT(class_id>=0 && class_id<all.nclasses);
    ASSERT(all.ntrn+all.ntst==all.labels.size());

    // output labels
    for(int64_t i=0; i<all.labels.size(); i++)
        lbl_out.push_back(all.labels[i]==class_id ? 0 : 1);

    // output activities
    out = predictions(all.ntrn, all.ntst, 2, lbl_out);
    for(int64_t k=0; k<all.nclasses; k++)
    {
        const float *p_src = &all.fpredictions[k*all.stride];
        float *p_dst = k==class_id ? &out.fpredictions[0*out.stride] : &out.fpredictions[1*out.stride];
        for(int64_t i=0; i<all.ntrn+all.ntst; i++)
            p_dst[i] += p_src[i];
    }

    //
    return out;
}

predictions predictions::merge_onevsall(const std::vector<predictions> &vec)
{
    predictions r;

    // integrity checks
    ASSERT(vec.size()>=2);
    for(const predictions &p: vec)
    {
        ASSERT(p.ntrn==vec[0].ntrn);
        ASSERT(p.ntst==vec[0].ntst);
        ASSERT(p.nclasses==1 || p.nclasses==2);
        ASSERT(p.stride==vec[0].stride);
    }
    const int64_t ntotal   = vec[0].ntrn+vec[0].ntst;
    const int64_t nclasses = vec.size();

    // compute label vector
    std::vector<int> out_labels(ntotal,-1);
    for(int64_t class_idx=0; class_idx<nclasses; class_idx++)
        for(int64_t i=0; i<ntotal; i++)
            if( vec[class_idx].labels[i]==0 )
            {
                ASSERT(out_labels[i]<0);
                out_labels[i] = class_idx;
            }
    for(int v: out_labels)
        ASSERT(v>=0);

    // compute prediction vector
    r = predictions(vec[0].ntrn, vec[0].ntst, nclasses, out_labels);
    for(int64_t k=0; k<nclasses; k++)
    {
        const float *p_src = &vec[k].fpredictions[0];
        float       *p_dst = &r.fpredictions[k*r.stride];
        for(int64_t i=0; i<ntotal; i++)
            p_dst[i] = p_src[i];
    }
    return r;
}

predictions predictions::merge_onevsall_partial(const std::vector<predictions> &vec)
{
    predictions r;

    // integrity checks
    ASSERT(vec.size()>=1);
    for(const predictions &p: vec)
    {
        ASSERT(p.ntrn==vec[0].ntrn);
        ASSERT(p.ntst==vec[0].ntst);
        ASSERT(p.nclasses==2);
        ASSERT(p.stride==vec[0].stride);
        ASSERT(p.fpredictions.size()==vec[0].fpredictions.size());
    }
    const int64_t ntotal   = vec[0].ntrn+vec[0].ntst;
    const int64_t nclasses = vec.size()+1;

    // compute label vector
    std::vector<int> out_labels(ntotal,-1);
    for(int64_t class_idx=0; class_idx<nclasses-1; class_idx++)
        for(int64_t i=0; i<ntotal; i++)
            if( vec[class_idx].labels[i]==0 )
            {
                ASSERT(out_labels[i]<0);
                out_labels[i] = class_idx;
            }
    for(int &v: out_labels)
        if( v<0 )
            v = nclasses-1;

    // compute prediction vector
    r = predictions(vec[0].ntrn, vec[0].ntst, nclasses, out_labels);
    for(int64_t i=0; i<ntotal; i++)
        r.fpredictions[i+(nclasses-1)*r.stride] = 1.0f;
    for(int64_t k=0; k<nclasses-1; k++)
    {
        const float *p_src      = &vec[k].fpredictions[0];
        float       *p_dst      = &r.fpredictions[k*r.stride];
        float       *p_dst_last = &r.fpredictions[(nclasses-1)*r.stride];
        for(int64_t i=0; i<ntotal; i++)
        {
            p_dst[i] = p_src[i];
            p_dst_last[i] -= p_src[i];
        }
    }
    return r;

}

void serialize(FILE *f, const predictions *ptr)
{
    write_magic8(f, 89);
    serialize(f, &ptr->ntrn);
    serialize(f, &ptr->ntst);
    serialize(f, &ptr->nclasses);
    serialize(f, &ptr->labels);
    serialize(f, &ptr->fpredictions);
    serialize(f, &ptr->stride);
    write_magic8(f, 121);
}

void unserialize(FILE *f, predictions *ptr)
{
    check_magic8(f, 89);
    unserialize(f, &ptr->ntrn);
    unserialize(f, &ptr->ntst);
    unserialize(f, &ptr->nclasses);
    unserialize(f, &ptr->labels);
    unserialize(f, &ptr->fpredictions);
    unserialize(f, &ptr->stride);
    check_magic8(f, 121);
}

conv_predictions::conv_predictions(int ih, int iw, int nc):
    conv_predictions(std::make_tuple(ih,iw), 1, std::make_tuple(ih,iw), std::make_tuple(0,0), nc)
{
}


conv_predictions::conv_predictions(std::tuple<int,int> img_hw, int ds, std::tuple<int,int> pred_hw, std::tuple<int,int> shift_hw, int nc)
{
    ASSERT(ds>=1);
    ASSERT(std::get<0>(img_hw)>=1    && std::get<1>(img_hw)>=1);
    ASSERT(std::get<0>(img_hw)%ds==0 && std::get<1>(img_hw)%ds==0);
    ASSERT(std::get<0>(pred_hw)>=1   && std::get<1>(pred_hw)>=1);
    ASSERT(std::get<0>(shift_hw)>=0  && std::get<1>(shift_hw)>=0);
    ASSERT(nc>=1);

    nclasses      = nc;
    ds_factor     = ds;
    img_height    = std::get<0>(img_hw);
    img_width     = std::get<1>(img_hw);
    pred_height   = std::get<0>(pred_hw);
    pred_width    = std::get<1>(pred_hw);
    shift_height  = std::get<0>(shift_hw);
    shift_width   = std::get<1>(shift_hw);
    fpredictions.resize(nclasses*pred_width*pred_height, 0.0f);
}

int64_t conv_predictions::get_img_width() const
{
    return img_width;
}

int64_t conv_predictions::get_img_height() const
{
    return img_height;
}

int64_t conv_predictions::get_pred_width() const
{
    return pred_width;
}

int64_t conv_predictions::get_pred_height() const
{
    return pred_height;
}

float* conv_predictions::get_ptr()
{
    return &fpredictions[0];
}

const float* conv_predictions::get_ptr() const
{
    return &fpredictions[0];
}

float conv_predictions::get_pred(int pos0, int pos1, int class_idx) const
{
    ASSERT(pos0>=0 && pos0<pred_height);
    ASSERT(pos1>=0 && pos1<pred_width);
    ASSERT(class_idx>=0 && class_idx<nclasses);
    return fpredictions[class_idx*pred_height*pred_width+pos0*pred_width+pos1];
}


void conv_predictions::set_pred(int pos0, int pos1, int class_idx, float pred)
{
    ASSERT(pos0>=0 && pos0<pred_height);
    ASSERT(pos1>=0 && pos1<pred_width);
    ASSERT(class_idx>=0 && class_idx<nclasses);
    fpredictions[class_idx*pred_height*pred_width+pos0*pred_width+pos1] = pred;
}

void conv_predictions::set_pred01(int pos_h, int pos_w, float pred)
{
    ASSERT(nclasses==2);
    ASSERT(pos_h>=0 && pos_h<pred_height);
    ASSERT(pos_w>=0 && pos_w<pred_width);
    fpredictions[0*pred_height*pred_width+pos_h*pred_width+pos_w] = pred;
    fpredictions[1*pred_height*pred_width+pos_h*pred_width+pos_w] = 1-pred;
}

float conv_predictions::get_original_pixel(int pos_h, int pos_w, int class_idx) const
{
    ASSERT(pos_h>=0 && pos_h<img_height);
    ASSERT(pos_w>=0 && pos_w<img_width);
    ASSERT(class_idx>=0 && class_idx<nclasses);

    int pos0 = pos_h/ds_factor-shift_height;
    int pos1 = pos_w/ds_factor-shift_width;
    if( pos0<0 || pos0>=pred_height || pos1<0 || pos1>=pred_width )
        return 0.0f;
    return fpredictions[class_idx*pred_height*pred_width+pos0*pred_width+pos1];
}

void conv_predictions::set_original_pixel01(int pos_h, int pos_w, float pred)
{
    ASSERT(pos_h>=0 && pos_h<img_height);
    ASSERT(pos_w>=0 && pos_w<img_width);
    ASSERT(nclasses==2);

    int pos0 = pos_h/ds_factor-shift_height;
    int pos1 = pos_w/ds_factor-shift_width;
    if( pos0<0 || pos0>=pred_height || pos1<0 || pos1>=pred_width )
        return;
    fpredictions[0*pred_height*pred_width+pos0*pred_width+pos1] = pred;
    fpredictions[1*pred_height*pred_width+pos0*pred_width+pos1] = 1-pred;
}

std::tuple<int,int> conv_predictions::ds2ori(int pos_h, int pos_w) const
{
    return std::make_tuple(pos_h*ds_factor,pos_w*ds_factor);
}

void conv_predictions::merge_max(const conv_predictions &src)
{
    ASSERT(nclasses==src.nclasses);
    ASSERT(img_height==src.img_height && img_width==src.img_width);
    ASSERT(pred_height==src.pred_height && pred_width==src.pred_width);
    fmerge_max(&fpredictions[0], &src.fpredictions[0], nclasses*pred_height*pred_width);
}

conv_predictions conv_predictions::upscale2k(int k) const
{
    ASSERT(k>=0);
    conv_predictions r = *this;
    r.img_height *= 1<<k;
    r.img_width  *= 1<<k;
    r.ds_factor  *= 1<<k;
    return r;
}

void conv_predictions::debug_output(const bitmap &orig_image, std::string filename)
{
    ASSERT(orig_image.height==img_height && orig_image.width==img_width);
    bitmap img = orig_image;
    for(int i=0; i<img.height; i++)
        for(int j=0; j<img.width; j++)
        {
            const float gray_f = 0.5f;
            hsv v = img.get_hsv(i,j);
            v.s = 0;
            v.v = gray_f*0.25f + (1-gray_f)*v.v;
            rgb w = hsv2rgb(v);
            float p = get_original_pixel(i,j);
            w.r = (1-p)*w.r;
            w.g = (1-p)*w.g + p*1.0f;
            w.b = (1-p)*w.b;
            img.draw_rgb(i, j, w);
        }
    img.to_png(filename);
}

void conv_predictions::debug_output(const bitmap &orig_image, float mark_above, std::string filename)
{
    ASSERT(orig_image.height==img_height && orig_image.width==img_width);
    bitmap img = orig_image;
    for(int i=0; i<img.height; i++)
        for(int j=0; j<img.width; j++)
        {
            const float gray_f = 0.5f;
            hsv v = img.get_hsv(i,j);
            v.s = 0;
            v.v = gray_f*0.25f + (1-gray_f)*v.v;
            rgb w = hsv2rgb(v);
            float p = get_original_pixel(i,j);
            p = p>mark_above ? 0.5f : 0.0f;
            w.r = (1-p)*w.r;
            w.g = (1-p)*w.g + p*1.0f;
            w.b = (1-p)*w.b;
            img.draw_rgb(i, j, w);
        }
    img.to_png(filename);
}

void serialize(FILE *f, const channel *ptr)
{
    serialize(f, &ptr->desc);
}

void unserialize(FILE *f, channel *ptr)
{
    unserialize(f, &ptr->desc);
}

activities::activities(int trn, int tst, storage_format fmt):
    ntrn(trn),ntst(tst),nchannels(0),format(fmt),
    pwrk_activities8(NULL),pwrk_activitiesf(NULL)
{
    stride = iDivUp(ntrn+ntst,ALIGN_ACTIVITIES_BY)*ALIGN_ACTIVITIES_BY;
}

activities::activities(std::string filename):
    pwrk_activities8(NULL),pwrk_activitiesf(NULL)
{
    unserialize_from(filename, this);
}


const realm<channel>& activities::realm_by_channel(int cidx) const
{
    const realm<channel> *result = NULL;
    for(const realm<channel> &rr: realms)
    {
        ASSERT(rr.is_densely_packed());
        if( cidx>=rr.ch_offs && cidx<rr.ch_offs+rr.grid_depth*rr.grid_height*rr.grid_width )
        {
            ASSERT(result==NULL);
            result = &rr;
        }
    }
    ASSERT(result!=NULL);
    return *result;
}

const realm<channel>& activities::realm_by_name(std::string name) const
{
    for(const realm<channel> &rr: realms)
        if( rr.name==name )
            return rr;
    ASSERT(0);
    abort();
}

bool activities::is_i8() const
{ return format==storage_format::i8; }

bool activities::is_f32() const
{ return format==storage_format::f32; }

int64_t activities::get_stride() const
{ return stride; }

float activities::get_normalized(int64_t cidx, int64_t img_idx) const
{
    ASSERT(cidx>=0 && cidx<nchannels);
    ASSERT(img_idx>=0 && img_idx<ntrn+ntst);
    if( is_i8() )
    {
        const activity8_vector &act8 = *(coalesce_const_ptr(pwrk_activities8, &lcl_activities8));
        return SNET_I2F*act8[img_idx+cidx*stride];
    }
    if( is_f32() )
    {
        const activityf_vector &actf = *(coalesce_const_ptr(pwrk_activitiesf, &lcl_activitiesf));
        return actf[img_idx+cidx*stride];
    }
    ASSERT(0);
    return 0;
}

int64_t activities::get_memory_usage() const
{
    if( is_i8() )
    {
        const activity8_vector &act8 = *(coalesce_const_ptr(pwrk_activities8, &lcl_activities8));
        return act8.size()*sizeof(int8_t);
    }
    if( is_f32() )
    {
        const activityf_vector &actf = *(coalesce_const_ptr(pwrk_activitiesf, &lcl_activitiesf));
        return actf.size()*sizeof(float);
    }
    ASSERT(0);
    return 0;
}

int8_t* activities::get_i8(int64_t cidx)
{
    ASSERT(format==storage_format::i8 && cidx>=0 && cidx<nchannels);
    activity8_vector &act8 = *(coalesce_ptr(pwrk_activities8, &lcl_activities8));
    return &act8[cidx*stride];
}

float* activities::get_f32(int64_t cidx)
{
    ASSERT(format==storage_format::f32 && cidx>=0 && cidx<nchannels);
    activityf_vector &actf = *(coalesce_ptr(pwrk_activitiesf, &lcl_activitiesf));
    return &actf[cidx*stride];
}

const int8_t* activities::get_i8(int64_t cidx) const
{
    ASSERT(format==storage_format::i8 && cidx>=0 && cidx<nchannels);
    const activity8_vector &act8 = *(coalesce_const_ptr(pwrk_activities8, &lcl_activities8));
    return &act8[cidx*stride];
}

const float* activities::get_f32(int64_t cidx) const
{
    ASSERT(format==storage_format::f32 && cidx>=0 && cidx<nchannels);
    const activityf_vector &actf = *(coalesce_const_ptr(pwrk_activitiesf, &lcl_activitiesf));
    return &actf[cidx*stride];
}

void activities::set_activity(int64_t cidx, const activity8_vector &a)
{
    ASSERT(pwrk_activities8==NULL && pwrk_activitiesf==NULL);
    ASSERT(a.size()==ntrn+ntst);
    ASSERT(cidx>=0 && cidx<nchannels);

    if( format==storage_format::i8 )
    {
        int8_t *p_dst = &lcl_activities8[cidx*stride];
        for(int64_t k=0; k<ntrn+ntst; k++)
            p_dst[k] = a[k];
        return;
    }
    ASSERT(0);
}

void activities::set_activity(int64_t cidx, const activityf_vector &a)
{
    ASSERT(pwrk_activities8==NULL && pwrk_activitiesf==NULL);
    ASSERT(a.size()==ntrn+ntst);
    ASSERT(cidx>=0 && cidx<nchannels);

    if( format==storage_format::f32 )
    {
        float *p_dst = &lcl_activitiesf[cidx*stride];
        for(int64_t k=0; k<ntrn+ntst; k++)
            p_dst[k] = a[k];
        return;
    }
    ASSERT(0);
}

void activities::append1(const activity8_vector &a, const std::string &desc)
{
    ASSERT(realms.size()==1 && realms[0].grid_width==1 && realms[0].grid_height==1);
    ASSERT(a.size()==ntrn+ntst);

    // copy activities
    if( format==storage_format::i8 )
    {
        ASSERT(pwrk_activities8==NULL);
        ASSERT(lcl_activities8.size()==nchannels*stride);
        int64_t old_act_cnt = lcl_activities8.size();
        lcl_activities8.resize(old_act_cnt+stride,0x0);
        memcpy(&lcl_activities8[old_act_cnt], &a[0], (ntrn+ntst)*sizeof(a[0]));
    }
    else
        ASSERT(0);

    // update realm
    realms[0].append_element(channel(desc));

    // update nchannels
    nchannels++;
}

void activities::append1(const activityf_vector &a, const std::string &desc)
{
    ASSERT(realms.size()==1 && realms[0].grid_width==1 && realms[0].grid_height==1);
    ASSERT(a.size()==ntrn+ntst);

    // copy activities
    if( format==storage_format::f32 )
    {
        ASSERT(pwrk_activitiesf==NULL);
        ASSERT(lcl_activitiesf.size()==nchannels*stride);
        int64_t old_act_cnt = lcl_activitiesf.size();
        lcl_activitiesf.resize(old_act_cnt+stride,0x0);
        memcpy(&lcl_activitiesf[old_act_cnt], &a[0], (ntrn+ntst)*sizeof(a[0]));
    }
    else
        ASSERT(0);

    // update realm
    realms[0].append_element(channel(desc));

    // update nchannels
    nchannels++;
}

void activities::append_plane1(const activities &a, const std::string desc)
{
    ASSERT(pwrk_activities8==NULL && pwrk_activitiesf==NULL);
    ASSERT(a.ntrn==ntrn);
    ASSERT(a.ntst==ntst);
    ASSERT(a.stride==stride);
    ASSERT(realms.size()==1 && realms[0].ch_offs==0);
    ASSERT(a.realms.size()==1 && a.realms[0].ch_offs==0 && a.realms[0].grid_depth==1);
    ASSERT(realms[0].grid_width==a.realms[0].grid_width && realms[0].grid_height==a.realms[0].grid_height);
    ASSERT(a.realms[0].is_densely_packed());
    ASSERT(realms[0].is_densely_packed());
    ASSERT(format==storage_format::i8 && a.format==storage_format::i8);

    // copy activities
    const int64_t grid_height = realms[0].grid_height, grid_width = realms[0].grid_width;
    ASSERT(lcl_activities8.size()==nchannels*stride);
    int64_t old_act_cnt = lcl_activities8.size();
    lcl_activities8.resize(old_act_cnt+stride*grid_height*grid_width,0x0);
    memcpy(&lcl_activities8[old_act_cnt], a.get_i8(), stride*grid_height*grid_width*sizeof(lcl_activities8[0]));

    // update realm
    realms[0].append_plane(channel(desc));

    // update nchannels
    nchannels+= grid_height*grid_width;
}

void activities::append_plane1(const activities &a, int downsample, const std::string desc)
{
    ASSERT(pwrk_activities8==NULL && pwrk_activitiesf==NULL);
    ASSERT(downsample>=1);
    ASSERT(a.ntrn==ntrn);
    ASSERT(a.ntst==ntst);
    ASSERT(a.stride==stride);
    ASSERT(realms.size()==1 && realms[0].ch_offs==0);
    ASSERT(a.realms.size()==1 && a.realms[0].ch_offs==0 && a.realms[0].grid_depth==1);
    ASSERT(realms[0].grid_width ==iDivUp(a.realms[0].grid_width,downsample) &&
           realms[0].grid_height==iDivUp(a.realms[0].grid_height,downsample));
    ASSERT(format==a.format);
    ASSERT(a.realms[0].is_densely_packed());
    ASSERT(realms[0].is_densely_packed());

    // copy activities
    const int64_t grid_height = realms[0].grid_height, grid_width = realms[0].grid_width;
    const int64_t srcgrid_height = a.realms[0].grid_height, srcgrid_width = a.realms[0].grid_width;
    if( format==storage_format::i8 )
    {
        ASSERT(lcl_activities8.size()==nchannels*stride);
        int64_t old_act_cnt = lcl_activities8.size();
        lcl_activities8.resize(old_act_cnt+stride*grid_height*grid_width, 0x0);
        for(int i=0; i<srcgrid_height; i++)
            for(int j=0; j<srcgrid_width; j++)
            {
                const int8_t *p_src = a.get_i8(i*srcgrid_width+j);
                int8_t *p_dst = &lcl_activities8[(nchannels+(i/downsample)*grid_width+(j/downsample))*stride];
                for(int64_t k=0; k<stride; k++)
                {
                    int8_t va = p_src[k], vb=p_dst[k];
                    p_dst[k] = va>vb ? va : vb;
                }
            }
    }
    else if( format==storage_format::f32 )
    {
        ASSERT(lcl_activitiesf.size()==nchannels*stride);
        int64_t old_act_cnt = lcl_activitiesf.size();
        lcl_activitiesf.resize(old_act_cnt+stride*grid_height*grid_width, 0.0f);
        for(int i=0; i<srcgrid_height; i++)
            for(int j=0; j<srcgrid_width; j++)
            {
                const float *p_src = a.get_f32(i*srcgrid_width+j);
                float *p_dst = &lcl_activitiesf[(nchannels+(i/downsample)*grid_width+(j/downsample))*stride];
                for(int64_t k=0; k<stride; k++)
                {
                    float va = p_src[k], vb=p_dst[k];
                    p_dst[k] = va>vb ? va : vb;
                }
            }
    }
    else
        ASSERT(0);

    // update realm
    realms[0].append_plane(channel(desc));

    // update nchannels
    nchannels+= grid_height*grid_width;

}

void activities::append_planes_from_realm(const activities &a, const std::string realm_name)
{
    const realm<channel> &r_src = a.realm_by_name(realm_name);

    // integrity checks
    ASSERT(pwrk_activities8==NULL && pwrk_activitiesf==NULL);
    ASSERT(format==a.format);
    ASSERT(a.stride==stride);
    ASSERT((format==storage_format::i8  && lcl_activities8.size()==nchannels*stride) ||
           (format==storage_format::f32 && lcl_activitiesf.size()==nchannels*stride));
    ASSERT(realms.size()==0 || realms.size()==1);
    if( realms.size()==1 )
    {
        ASSERT(realms[0].grid_height==r_src.grid_height && realms[0].grid_width==r_src.grid_width);
        ASSERT(nchannels==realms[0].grid_depth*realms[0].grid_height*realms[0].grid_width);
    }

    // insertion
    if( realms.size()==0 )
        realms.push_back(realm<channel>::create_grid(realm_name, r_src.grid_height, r_src.grid_width, 0));
    ASSERT(realms[0].is_densely_packed());

    // copy activities
    const int64_t grid_height = realms[0].grid_height, grid_width = realms[0].grid_width;
    if( format==storage_format::i8 )
    {
        int64_t old_act_cnt = lcl_activities8.size();
        int64_t new_total   = stride*r_src.grid_depth*grid_height*grid_width;
        lcl_activities8.resize(old_act_cnt+new_total, 0x0);
        memcpy(&lcl_activities8[old_act_cnt], a.get_i8(r_src.ch_offs), new_total*sizeof(lcl_activities8[0]));
    }
    else if( format==storage_format::f32 )
    {
        int64_t old_act_cnt = lcl_activitiesf.size();
        int64_t new_total   = stride*r_src.grid_depth*grid_height*grid_width;
        lcl_activitiesf.resize(old_act_cnt+new_total, 0x0);
        memcpy(&lcl_activitiesf[old_act_cnt], a.get_f32(r_src.ch_offs), new_total*sizeof(lcl_activitiesf[0]));
    }
    else
        ASSERT(0);

    // update realm
    for(int k=0; k<r_src.grid_depth; k++)
        realms[0].append_plane(channel());

    // update nchannels
    nchannels+= r_src.grid_depth*grid_height*grid_width;
}

int activities::append_realm(const realm<channel> &r)
{
    ASSERT(pwrk_activities8==NULL && pwrk_activitiesf==NULL);
    ASSERT(r.ch_offs==0 && r.is_densely_packed());
    int64_t new_total = r.grid_depth*r.grid_height*r.grid_width;

    // reallocate storage
    if( format==storage_format::i8 )
    {
        ASSERT(lcl_activities8.size()==stride*nchannels);
        int64_t old_size = lcl_activities8.size();
        int64_t nact64 = new_total*stride;
        lcl_activities8.resize(old_size+nact64);
        memset(&lcl_activities8[old_size], 0, sizeof(lcl_activities8[0])*nact64);
    }
    else if( format==storage_format::f32 )
    {
        ASSERT(lcl_activitiesf.size()==stride*nchannels);
        int64_t old_size = lcl_activitiesf.size();
        int64_t nact64 = new_total*stride;
        lcl_activitiesf.resize(old_size+nact64);
        memset(&lcl_activitiesf[old_size], 0, sizeof(lcl_activitiesf[0])*nact64);
    }
    else
        ASSERT(0);

    // add realm
    realms.push_back(r.rebase(nchannels));
    nchannels += new_total;

    //
    return realms.back().ch_offs;
}

void activities::extract_realm(std::string realm_name, activities &act_out) const
{
    ASSERT(is_i8());

    const realm<channel> &rr = realm_by_name(realm_name);
    act_out = activities(ntrn, ntst, format);
    act_out.append_realm(realm<channel>::create_grid(rr.name, rr.grid_height, rr.grid_width, rr.grid_depth));
    for(int d=0; d<rr.grid_depth; d++)
        for(int i0=0; i0<rr.grid_height; i0++)
            for(int i1=0; i1<rr.grid_width; i1++)
            {
                const int8_t *p_src = get_i8(rr.get_cidx(d,i0,i1));
                int8_t *p_dst = act_out.get_i8(act_out.realms[0].get_cidx(d,i0,i1));
                memcpy(p_dst, p_src, (ntrn+ntst)*sizeof(p_dst[0]));
            }
}

void activities::merge_in(const activities &merged, std::string name_prefix, std::string name_suffix)
{
    ASSERT(pwrk_activities8==NULL && pwrk_activitiesf==NULL);
    ASSERT(merged.pwrk_activities8==NULL && merged.pwrk_activitiesf==NULL);
    ASSERT(merged.ntrn==ntrn);
    ASSERT(merged.ntst==ntst);
    ASSERT(merged.stride==stride);

    // copy activities
    ASSERT(format==merged.format);
    if( format==storage_format::i8 )
    {
        // integrity checks
        ASSERT(lcl_activities8.size()==nchannels*stride);
        ASSERT(merged.lcl_activities8.size()==merged.nchannels*stride);

        // reallocation
        int64_t old_act_cnt = lcl_activities8.size();
        lcl_activities8.resize(old_act_cnt+merged.lcl_activities8.size());
        memcpy(&lcl_activities8[old_act_cnt], &merged.lcl_activities8[0], merged.lcl_activities8.size()*sizeof(merged.lcl_activities8[0]));
    }
    else if( format==storage_format::f32 )
    {
        // integrity checks
        ASSERT(lcl_activitiesf.size()==nchannels*stride);
        ASSERT(merged.lcl_activitiesf.size()==merged.nchannels*stride);

        // reallocation
        int64_t old_act_cnt = lcl_activitiesf.size();
        lcl_activitiesf.resize(old_act_cnt+merged.lcl_activitiesf.size());
        memcpy(&lcl_activitiesf[old_act_cnt], &merged.lcl_activitiesf[0], merged.lcl_activitiesf.size()*sizeof(merged.lcl_activitiesf[0]));
    }
    else
        ASSERT(0);

    // rebase realms
    int64_t cur_offs = nchannels;
    for(const realm<channel> &r: merged.realms)
    {
        realms.push_back(r.rebase(cur_offs));
        realms.back().name = name_prefix+realms.back().name+name_suffix;
        cur_offs += r.elements.size();
    }
    ASSERT(format!=storage_format::i8  || cur_offs*stride==lcl_activities8.size());
    ASSERT(format!=storage_format::f32 || cur_offs*stride==lcl_activitiesf.size());

    // update nchannels
    nchannels += merged.nchannels;
}

void activities::merge_in_with_gain(const activities &merged, const int gain, std::string name_prefix, std::string name_suffix)
{
    ASSERT(pwrk_activities8==NULL && pwrk_activitiesf==NULL);
    ASSERT(merged.pwrk_activities8==NULL && merged.pwrk_activitiesf==NULL);
    ASSERT(merged.ntrn==ntrn);
    ASSERT(merged.ntst==ntst);
    ASSERT(merged.stride==stride);

    // copy activities
    ASSERT(format==merged.format);
    if( format==storage_format::i8 )
    {
        // integrity checks
        ASSERT(lcl_activities8.size()==nchannels*stride);
        ASSERT(merged.lcl_activities8.size()==merged.nchannels*stride);

        // reallocation
        int64_t old_act_cnt = lcl_activities8.size();
        lcl_activities8.resize(old_act_cnt+merged.lcl_activities8.size());

        // append with gain
        const int64_t append_len = merged.lcl_activities8.size();
        const int8_t *p_src = &merged.lcl_activities8[0];
        int8_t *p_dst = &lcl_activities8[old_act_cnt];
        for(int64_t i=0; i<append_len; i++)
        {
            int vv = int(p_src[i])*gain;
            vv = vv>=+SNET_IMAX ? +SNET_IMAX : vv;
            vv = vv<=-SNET_IMAX ? -SNET_IMAX : vv;
            p_dst[i] = vv;
        }
    }
    else
        ASSERT(0);

    // rebase realms
    int64_t cur_offs = nchannels;
    for(const realm<channel> &r: merged.realms)
    {
        realms.push_back(r.rebase(cur_offs));
        realms.back().name = name_prefix+realms.back().name+name_suffix;
        cur_offs += r.elements.size();
    }
    ASSERT(format!=storage_format::i8 || cur_offs*stride==lcl_activities8.size());

    // update nchannels
    nchannels += merged.nchannels;
}

void activities::convert_to_f32(activities &out) const
{
    ASSERT(pwrk_activities8==NULL && pwrk_activitiesf==NULL);

    if( format==storage_format::f32 )
    {
        out = *this;
        return;
    }
    ASSERT(format==storage_format::i8);

    //
    // Copy what can be just copied
    //
    // NOTE: although it is possible to reduce stride for f32,
    //       we prefer to stick to same stride for all kinds
    //       of storage.
    //
    out.stride   = stride;
    out.ntrn     = ntrn;
    out.ntst     = ntst;
    out.nchannels= nchannels;
    out.realms   = realms;
    out.pwrk_activities8 = NULL;
    out.pwrk_activitiesf = NULL;

    //
    // Convert activities
    //
    out.format = storage_format::f32;
    out.lcl_activitiesf.resize(out.stride*out.nchannels, 0.0f);
    for(int64_t cidx=0; cidx<nchannels; cidx++)
    {
        const int8_t *p_src = get_i8(cidx);
        float *p_dst = out.get_f32(cidx);
        for(int64_t k=0; k<ntrn+ntst; k++)
            p_dst[k] = SNET_I2F*p_src[k];
    }
}

void activities::convert_to_f32()
{
    if( format==storage_format::f32 )
        return;
    activities tmp;
    convert_to_f32(tmp);
    *this = tmp;
}

void activities::split_q6(int overlap)
{
    ASSERT(is_i8());
    std::vector<realm<channel>> _prev_realms = realms;
    realms.clear();

    //
    // Output 5 quadrants
    //
    for(const realm<channel> &rr_prev: _prev_realms)
    {
        ASSERT(rr_prev.grid_height%4==0 && rr_prev.grid_width%4==0);
        ASSERT(rr_prev.grid_height/2 + overlap<=rr_prev.grid_height);
        ASSERT(rr_prev.grid_width/2  + overlap<=rr_prev.grid_width);
        const int grid_height = rr_prev.grid_height, grid_width = rr_prev.grid_width;
        const int64_t ntotal = ntrn+ntst;
        for(int qidx=0; qidx<5; qidx++)
        {
            // new realm name, size, location
            const std::string new_name = rr_prev.name+"[q"+std::to_string(qidx)+"]";
            const int new_height = grid_height/2+overlap;
            const int new_width  = grid_width/2+overlap;
            int offs0=0, offs1=0;
            if( qidx==1 || qidx==3 )
                offs1 = grid_width-new_width;
            if( qidx==2 || qidx==3 )
                offs0 = grid_height-new_height;
            if( qidx==4 )
            {
                offs0 = (grid_height-new_height)/2;
                offs1 = (grid_width-new_width)/2;
            }

            // new realm itself
            realm<channel> rr_new = realm<channel>::create_grid(new_name, new_height, new_width, rr_prev.grid_depth);
            rr_new.grid_stride0 = rr_prev.grid_height;
            rr_new.grid_stride1 = rr_prev.grid_width;
            rr_new.ch_offs = rr_prev.get_cidx(0, offs0, offs1);
            realms.push_back(rr_new);
        }
    }

    //
    // Output entire area
    //
    for(realm<channel> rr: _prev_realms)
    {
        rr.name = rr.name+"[qa]";
        realms.push_back(rr);
    }
}

void activities::split_q10(int max_overlap)
{
    ASSERT(is_i8());
    std::vector<realm<channel>> _prev_realms = realms;
    realms.clear();

    //
    // Output 9 quadrants
    //
    for(const realm<channel> &rr_prev: _prev_realms)
    {
        const int grid_height = rr_prev.grid_height, grid_width = rr_prev.grid_width;
        if( grid_height<4 || grid_width<4 )
            continue;
        ASSERT(grid_height%2==0 && grid_width%2==0);
        for(int qidx=0; qidx<9; qidx++)
        {
            // new realm name, size, location
            const std::string new_name = rr_prev.name+"[q"+std::to_string(qidx)+"]";
            const int q_height = grid_height/2;
            const int q_width  = grid_width/2;
            const int overlap_h = imin(max_overlap,grid_height/8), overlap_w = imin(max_overlap,grid_width/8);
            ASSERT(q_height+2*overlap_h<grid_height && q_width+2*overlap_w<grid_width);
            int offs0=0, offs1=0;
            int new_height = q_height+overlap_h, new_width = q_width+overlap_w;

            if( qidx/3==1 )
            {
                new_height = q_height;
                if( new_height%2!=0 )
                    new_height++;
                ASSERT(new_height<=grid_height);
                offs0 = (grid_height-new_height)/2;
                offs0 -= overlap_h;
                ASSERT(offs0>=0);
                new_height += 2*overlap_h;
                ASSERT(2*offs0+new_height==grid_height);
            }
            if( qidx/3==2 )
            {
                offs0      = grid_height-q_height-overlap_h;
                new_height = q_height+overlap_h;
            }

            if( qidx%3==1 )
            {
                new_width = q_width;
                if( new_width%2!=0 )
                    new_width++;
                ASSERT(new_width<=grid_width);
                offs1 = (grid_width-new_width)/2;
                offs1 -= overlap_w;
                ASSERT(offs1>=0);
                new_width += 2*overlap_w;
                ASSERT(2*offs1+new_width==grid_width);
            }
            if( qidx%3==2 )
            {
                offs1     = grid_width-q_width-overlap_w;
                new_width = q_width+overlap_w;
            }

            // new realm itself
            realm<channel> rr_new = realm<channel>::create_grid(new_name, new_height, new_width, rr_prev.grid_depth);
            rr_new.grid_stride0 = rr_prev.grid_height;
            rr_new.grid_stride1 = rr_prev.grid_width;
            rr_new.ch_offs = rr_prev.get_cidx(0, offs0, offs1);
            realms.push_back(rr_new);
        }
    }

    //
    // Output entire area
    //
    for(realm<channel> rr: _prev_realms)
    {
        rr.name = rr.name+"[qa]";
        realms.push_back(rr);
    }
}

void activities::split_and_sum_h22(activities &act_out) const
{
    const int64_t ntotal = ntrn+ntst;
    ASSERT(is_i8());

    //
    // Output 5 quadrants
    //
    act_out = activities(ntrn, ntst, storage_format::f32);
    for(const realm<channel> &rr: realms)
    {
        ASSERT(rr.grid_height%4==0 && rr.grid_width%4==0);
        const int grid_height = rr.grid_height, grid_width = rr.grid_width;

        // compute sums over 4x4 grid
        for(int hidx=0; hidx<16; hidx++)
        {
            activities act_tmp = activities(ntrn, ntst, storage_format::f32);
            act_tmp.append_realm(realm<channel>::create_grid(rr.name+"[h"+std::to_string(hidx/4)+std::to_string(hidx%4)+"]",1,1,0));
            for(int pidx=0; pidx<rr.grid_depth; pidx++)
            {
                activityf_vector sum(ntotal,0.0f);
                float *p_dst = &sum[0];
                const int i0 = (hidx/4)*(grid_height/4), i1 = i0+(grid_height/4);
                const int j0 = (hidx%4)*( grid_width/4), j1 = j0+( grid_width/4);
                for(int i=i0; i<i1; i++)
                    for(int j=j0; j<j1; j++)
                    {
                        const int8_t *p_src = get_i8(rr.get_cidx(pidx,i,j));
                        for(int64_t k=0; k<ntotal; k++)
                            p_dst[k] += p_src[k];
                    }
                for(int64_t k=0; k<ntotal; k++)
                    p_dst[k] *= SNET_I2F;
                act_tmp.append1(sum,"");
            }
            act_out.merge_in(act_tmp);
        }

        // compute sums over 2x2 grid + central element
        for(int qidx=0; qidx<4; qidx++)
        {
            activities act_tmp = activities(ntrn, ntst, storage_format::f32);
            act_tmp.append_realm(realm<channel>::create_grid(rr.name+"[q"+std::to_string(qidx/2)+std::to_string(qidx%2)+"]",1,1,0));
            for(int pidx=0; pidx<rr.grid_depth; pidx++)
            {
                activityf_vector sum(ntotal,0.0f);
                float *p_dst = &sum[0];
                const int i0 = (qidx/2)*(grid_height/2), i1 = i0+(grid_height/2);
                const int j0 = (qidx%2)*( grid_width/2), j1 = j0+( grid_width/2);
                for(int i=i0; i<i1; i++)
                    for(int j=j0; j<j1; j++)
                    {
                        const int8_t *p_src = get_i8(rr.get_cidx(pidx,i,j));
                        for(int64_t k=0; k<ntotal; k++)
                            p_dst[k] += p_src[k];
                    }
                for(int64_t k=0; k<ntotal; k++)
                    p_dst[k] *= SNET_I2F;
                act_tmp.append1(sum,"");
            }
            act_out.merge_in(act_tmp);
        }
    }
}

void activities::split_and_accumulate(split_t split, accumulate_t accumulate, activities &act_out) const
{
    const int64_t ntotal = ntrn+ntst;

    //
    // entire area is accumulated
    //
    if( split==split_t::entire )
    {
        ASSERT(is_i8() && (accumulate==accumulate_t::sum||accumulate==accumulate_t::max));

        act_out = activities(ntrn, ntst, storage_format::f32);
        for(const realm<channel> &rr: realms)
        {
            const int grid_height = rr.grid_height, grid_width = rr.grid_width;
            activities act_tmp = activities(ntrn, ntst, storage_format::f32);
            act_tmp.append_realm(realm<channel>::create_grid(rr.name,1,1,0));
            for(int pidx=0; pidx<rr.grid_depth; pidx++)
            {
                activityf_vector acc(ntotal,0.0f);
                float *p_dst = &acc[0];
                for(int i=0; i<grid_height; i++)
                    for(int j=0; j<grid_width; j++)
                    {
                        const int8_t *p_src = get_i8(rr.get_cidx(pidx,i,j));
                        if( accumulate==accumulate_t::sum )
                        {
                            for(int64_t k=0; k<ntotal; k++)
                                p_dst[k] += p_src[k];
                        }
                        else if( accumulate==accumulate_t::max )
                        {
                            for(int64_t k=0; k<ntotal; k++)
                                p_dst[k] = fmaxf(p_dst[k], p_src[k]);
                        }
                        else
                            ASSERT(0);
                    }
                for(int64_t k=0; k<ntotal; k++)
                    p_dst[k] *= SNET_I2F;
                act_tmp.append1(acc,"");
            }
            act_out.merge_in(act_tmp);
        }
        return;
    }

    //
    // Q4 split-and-accumulate
    //
    if( split==split_t::q4 )
    {
        ASSERT(is_i8() && accumulate==accumulate_t::sum);
        auto determine_range = [](int prj_idx, int entire_size)->std::tuple<int,int>
        {
            int new_size = iDivUp(entire_size,2);
            if( prj_idx==0 )
                return std::make_tuple(0, new_size);
            if( prj_idx==1 )
                return std::make_tuple(entire_size-new_size, entire_size);
            ASSERT(0);
        };

        act_out = activities(ntrn, ntst, storage_format::f32);
        for(const realm<channel> &rr: realms)
        {
            ASSERT(rr.grid_height>=4 && rr.grid_width>=4);
            const int grid_height = rr.grid_height, grid_width = rr.grid_width;

            // compute sums over 2x2 grid of nonoverlapping quadrants
            for(int qidx=0; qidx<4; qidx++)
            {
                activities act_tmp = activities(ntrn, ntst, storage_format::f32);
                act_tmp.append_realm(realm<channel>::create_grid(rr.name+"[q"+std::to_string(qidx/2)+std::to_string(qidx%2)+"]",1,1,0));
                for(int pidx=0; pidx<rr.grid_depth; pidx++)
                {
                    activityf_vector sum(ntotal,0.0f);
                    float *p_dst = &sum[0];
                    int i0, i1, j0, j1;
                    std::tie(i0,i1) = determine_range(qidx/2,grid_height);
                    std::tie(j0,j1) = determine_range(qidx%2,grid_width);
                    for(int i=i0; i<i1; i++)
                        for(int j=j0; j<j1; j++)
                        {
                            const int8_t *p_src = get_i8(rr.get_cidx(pidx,i,j));
                            for(int64_t k=0; k<ntotal; k++)
                                p_dst[k] += p_src[k];
                        }
                    for(int64_t k=0; k<ntotal; k++)
                        p_dst[k] *= SNET_I2F;
                    act_tmp.append1(sum,"");
                }
                act_out.merge_in(act_tmp);
            }
        }
        return;
    }

    //
    // Q9 split-and-accumulate
    //
    if( split==split_t::q9 )
    {
        ASSERT(is_i8() && accumulate==accumulate_t::sum);

        act_out = activities(ntrn, ntst, storage_format::f32);
        for(const realm<channel> &rr: realms)
        {
            ASSERT(rr.grid_height%4==0 && rr.grid_width%4==0);
            const int grid_height = rr.grid_height, grid_width = rr.grid_width;

            // compute sums over 3x3 grid of overlapping quadrants
            for(int qidx=0; qidx<9; qidx++)
            {
                activities act_tmp = activities(ntrn, ntst, storage_format::f32);
                act_tmp.append_realm(realm<channel>::create_grid(rr.name+"[q"+std::to_string(qidx/3)+std::to_string(qidx%3)+"]",1,1,0));
                for(int pidx=0; pidx<rr.grid_depth; pidx++)
                {
                    activityf_vector sum(ntotal,0.0f);
                    float *p_dst = &sum[0];
                    const int i0 = (qidx/3)*(grid_height/4), i1 = i0+(grid_height/2);
                    const int j0 = (qidx%3)*( grid_width/4), j1 = j0+( grid_width/2);
                    for(int i=i0; i<i1; i++)
                        for(int j=j0; j<j1; j++)
                        {
                            const int8_t *p_src = get_i8(rr.get_cidx(pidx,i,j));
                            for(int64_t k=0; k<ntotal; k++)
                                p_dst[k] += p_src[k];
                        }
                    for(int64_t k=0; k<ntotal; k++)
                        p_dst[k] *= SNET_I2F;
                    act_tmp.append1(sum,"");
                }
                act_out.merge_in(act_tmp);
            }
        }
        return;
    }

    //
    // Unexpected
    //
    ASSERT(0);
}

void activities::subdivide(split_t split, std::map<std::string,std::string> &realms_map, int max_overlap)
{
    //
    // Quick exit if possible
    //
    if( split==split_t::entire )
    {
        realms_map.clear();
        for(const realm<channel> &rr: realms)
            realms_map.insert(std::make_pair(rr.name, rr.name));
        return;
    }

    //
    // Determine which splits are needed
    //
    std::set<split_t> supported_splits{split_t::q9,
                                       split_t::q4,
                                       split_t::h16,
                                       split_t::p25,
                                       split_t::g64,
                                       split_t::q9h16,
                                       split_t::q9h16p25,
                                       split_t::q9h16g64,
                                       split_t::h16g64};
    ASSERT(supported_splits.count(split)==1);
    const bool need_q9  = (int(split)&int(split_t::q9))!=0;
    const bool need_h16 = (int(split)&int(split_t::h16))!=0;
    const bool need_p25 = (int(split)&int(split_t::p25))!=0;
    const bool need_g64 = (int(split)&int(split_t::g64))!=0;
    auto determine_q9offs_size = [](int prj_idx, int rec_size, int entire_size, int overlap)->std::tuple<int,int>
    {
        if( prj_idx==0 )
            return std::make_tuple(0, rec_size+overlap);
        if( prj_idx==1 )
        {
            int new_offs = (entire_size-rec_size)/2-overlap;
            int new_size = rec_size+2*overlap;
            ASSERT(new_size>0);
            ASSERT(new_offs>0);
            ASSERT(2*new_offs+new_size==entire_size);
            return std::make_tuple(new_offs,new_size);
        }
        if( prj_idx==2 )
            return std::make_tuple(entire_size-rec_size-overlap,rec_size+overlap);
        ASSERT(0);
    };
    auto determine_h16offs_size = [](int prj_idx, int rec_size, int entire_size, int overlap)->std::tuple<int,int>
    {
        ASSERT(entire_size==4*rec_size);
        ASSERT(overlap<rec_size);
        if( prj_idx==0 )
            return std::make_tuple(0, rec_size+overlap);
        if( prj_idx==1 || prj_idx==2 )
        {
            int new_offs = prj_idx*rec_size-overlap;
            int new_size = rec_size+2*overlap;
            ASSERT(new_size>0);
            ASSERT(new_offs>0);
            return std::make_tuple(new_offs,new_size);
        }
        if( prj_idx==3 )
            return std::make_tuple(entire_size-rec_size-overlap,rec_size+overlap);
        ASSERT(0);
    };
    auto determine_p25offs_size = [](int prj_idx, int entire_size, int overlap)->std::tuple<int,int>
    {
        ASSERT(entire_size>=10);
        ASSERT(overlap<=entire_size/5);
        ASSERT(prj_idx>=0 && prj_idx<=4);
        int i0 = roundf(    prj_idx/5.0f*entire_size);
        int i1 = roundf((prj_idx+1)/5.0f*entire_size);
        if( prj_idx==0 )
            return std::make_tuple(0, i1+overlap);
        if( prj_idx>=1 && prj_idx<=3 )
        {
            overlap = imin3(overlap,i0,entire_size-i1);
            int new_offs = i0-overlap;
            int new_size = (i1-i0)+2*overlap;
            ASSERT(new_size>0);
            ASSERT(new_offs>0);
            return std::make_tuple(new_offs,new_size);
        }
        if( prj_idx==4 )
            return std::make_tuple(i0-overlap,(i1-i0)+overlap);
        ASSERT(0);
    };
    auto determine_g64offs_size = [](int prj_idx, int rec_size, int entire_size, int overlap)->std::tuple<int,int>
    {
        ASSERT(entire_size==8*rec_size);
        ASSERT(overlap<=rec_size);
        if( prj_idx==0 )
            return std::make_tuple(0, rec_size+overlap);
        if( prj_idx>=1 && prj_idx<=6 )
        {
            int new_offs = prj_idx*rec_size-overlap;
            int new_size = rec_size+2*overlap;
            ASSERT(new_size>0);
            ASSERT(new_offs>=0 && new_offs+new_size-1<=entire_size-1);
            return std::make_tuple(new_offs,new_size);
        }
        if( prj_idx==7 )
            return std::make_tuple(entire_size-rec_size-overlap,rec_size+overlap);
        ASSERT(0);
    };
    std::vector<realm<channel>> _prev_realms = realms;
    realms.clear();
    realms_map.clear();

    //
    // Q9 splits
    //
    if( need_q9 )
    {
        for(const realm<channel> &rr_prev: _prev_realms)
        {
            const int grid_height = rr_prev.grid_height, grid_width = rr_prev.grid_width;
            if( grid_height%4!=0 || grid_width%4!=0 )
                continue;
            for(int qidx=0; qidx<9; qidx++)
            {
                // new realm name, size, location
                const std::string new_name = rr_prev.name+"[q"+std::to_string(qidx/3)+std::to_string(qidx%3)+"]";
                const int q_height = grid_height/2;
                const int q_width  = grid_width/2;
                const int overlap_h = imin(max_overlap,grid_height/8), overlap_w = imin(max_overlap,grid_width/8);
                ASSERT(q_height+2*overlap_h<grid_height && q_width+2*overlap_w<grid_width);

                int offs0, offs1, new_height, new_width;
                std::tie(offs0,new_height) = determine_q9offs_size(qidx/3, q_height, grid_height, overlap_h);
                std::tie(offs1,new_width)  = determine_q9offs_size(qidx%3, q_width,  grid_width,  overlap_w);

                // new realm itself
                realm<channel> rr_new = realm<channel>::create_grid(new_name, new_height, new_width, rr_prev.grid_depth);
                rr_new.grid_stride0 = rr_prev.grid_height;
                rr_new.grid_stride1 = rr_prev.grid_width;
                rr_new.ch_offs = rr_prev.get_cidx(0, offs0, offs1);
                realms.push_back(rr_new);

                // map of new name to old name
                realms_map.insert(std::make_pair(new_name, rr_prev.name));
            }
        }
    }

    //
    // H16 splits
    //
    if( need_h16 )
    {
        for(const realm<channel> &rr_prev: _prev_realms)
        {
            const int grid_height = rr_prev.grid_height, grid_width = rr_prev.grid_width;
            if( grid_height%4!=0 || grid_width%4!=0 )
                continue;
            for(int hidx=0; hidx<16; hidx++)
            {
                // new realm name, size, location
                const std::string new_name = rr_prev.name+"[h"+std::to_string(hidx/4)+std::to_string(hidx%4)+"]";
                const int q_height = grid_height/4;
                const int q_width  = grid_width/4;
                const int overlap_h = imin(max_overlap,grid_height/8), overlap_w = imin(max_overlap,grid_width/8);
                ASSERT(q_height+2*overlap_h<grid_height && q_width+2*overlap_w<grid_width);

                int offs0, offs1, new_height, new_width;
                std::tie(offs0,new_height) = determine_h16offs_size(hidx/4, q_height, grid_height, overlap_h);
                std::tie(offs1,new_width)  = determine_h16offs_size(hidx%4, q_width,  grid_width,  overlap_w);

                // new realm itself
                realm<channel> rr_new = realm<channel>::create_grid(new_name, new_height, new_width, rr_prev.grid_depth);
                rr_new.grid_stride0 = rr_prev.grid_height;
                rr_new.grid_stride1 = rr_prev.grid_width;
                rr_new.ch_offs = rr_prev.get_cidx(0, offs0, offs1);
                realms.push_back(rr_new);

                // map of new name to old name
                realms_map.insert(std::make_pair(new_name, rr_prev.name));
            }

        }
    }

    //
    // P25 split
    //
    if( need_p25 )
    {
        for(const realm<channel> &rr_prev: _prev_realms)
        {
            const int grid_height = rr_prev.grid_height, grid_width = rr_prev.grid_width;
            if( grid_height<10 || grid_width<10 )
                continue;
            for(int gidx=0; gidx<25; gidx++)
            {
                // new realm name, size, location
                const std::string new_name = rr_prev.name+"[p"+std::to_string(gidx/5)+std::to_string(gidx%5)+"]";
                const int overlap_h = imin(max_overlap,1+grid_height/10), overlap_w = imin(max_overlap,1+grid_width/10);

                int offs0, offs1, new_height, new_width;
                std::tie(offs0,new_height) = determine_p25offs_size(gidx/5, grid_height, overlap_h);
                std::tie(offs1,new_width)  = determine_p25offs_size(gidx%5, grid_width,  overlap_w);

                // new realm itself
                realm<channel> rr_new = realm<channel>::create_grid(new_name, new_height, new_width, rr_prev.grid_depth);
                rr_new.grid_stride0 = rr_prev.grid_height;
                rr_new.grid_stride1 = rr_prev.grid_width;
                rr_new.ch_offs = rr_prev.get_cidx(0, offs0, offs1);
                realms.push_back(rr_new);

                // map of new name to old name
                realms_map.insert(std::make_pair(new_name, rr_prev.name));
            }
        }
    }

    //
    // G64 split
    //
    if( need_g64 )
    {
        for(const realm<channel> &rr_prev: _prev_realms)
        {
            const int grid_height = rr_prev.grid_height, grid_width = rr_prev.grid_width;
            ASSERT(grid_height%8==0 && grid_width%8==0);
            for(int gidx=0; gidx<64; gidx++)
            {
                // new realm name, size, location
                const std::string new_name = rr_prev.name+"[g"+std::to_string(gidx/8)+std::to_string(gidx%8)+"]";
                const int q_height = grid_height/8;
                const int q_width  = grid_width/8;
                const int overlap_h = imin(max_overlap,1+grid_height/16), overlap_w = imin(max_overlap,1+grid_width/16);
                ASSERT(q_height+2*overlap_h<grid_height && q_width+2*overlap_w<grid_width);

                int offs0, offs1, new_height, new_width;
                std::tie(offs0,new_height) = determine_g64offs_size(gidx/8, q_height, grid_height, overlap_h);
                std::tie(offs1,new_width)  = determine_g64offs_size(gidx%8, q_width,  grid_width,  overlap_w);

                // new realm itself
                realm<channel> rr_new = realm<channel>::create_grid(new_name, new_height, new_width, rr_prev.grid_depth);
                rr_new.grid_stride0 = rr_prev.grid_height;
                rr_new.grid_stride1 = rr_prev.grid_width;
                rr_new.ch_offs = rr_prev.get_cidx(0, offs0, offs1);
                realms.push_back(rr_new);

                // map of new name to old name
                realms_map.insert(std::make_pair(new_name, rr_prev.name));
            }
        }
    }
}

void activities::subdivide(split_t split, int max_overlap)
{
    std::map<std::string,std::string> realms_map;
    subdivide(split, realms_map, max_overlap);
}

void activities::select_subsample(const std::vector<int> items, activities &act_out) const
{
    // integrity checks
    ASSERT(pwrk_activities8==NULL && pwrk_activitiesf==NULL);
    ASSERT(items.size()>=1);
    for(int k=1; k<items.size(); k++)
    {
        ASSERT(items[k-1]>=0 && items[k-1]<ntrn+ntst);
        ASSERT(items[k]>=0   && items[k]<ntrn+ntst);
        ASSERT(items[k-1]<=items[k]);
    }

    // determine output sample sizes
    int new_trn=0, new_tst=0;
    for(int k=0; k<items.size(); k++)
        if( items[k]<ntrn )
            new_trn++;
        else
            new_tst++;

    // init output metrics
    ASSERT(format==storage_format::i8 || format==storage_format::f32);
    act_out.format = format;
    act_out.lcl_activities8.clear();
    act_out.lcl_activities8.shrink_to_fit();
    act_out.lcl_activitiesf.clear();
    act_out.lcl_activitiesf.shrink_to_fit();
    act_out.ntrn = new_trn;
    act_out.ntst = new_tst;
    act_out.stride = iDivUp(new_trn+new_tst,ALIGN_ACTIVITIES_BY)*ALIGN_ACTIVITIES_BY;
    act_out.nchannels = nchannels;
    act_out.realms = realms;
    act_out.pwrk_activities8 = NULL;
    act_out.pwrk_activitiesf = NULL;

    // copy subsampled dataset
    if( format==storage_format::i8 )
    {
        act_out.lcl_activities8.resize(act_out.stride*act_out.nchannels);
        memset(&act_out.lcl_activities8[0], 0, sizeof(act_out.lcl_activities8[0])*act_out.lcl_activities8.size());
        for(int64_t k=0; k<nchannels; k++)
        {
            const int8_t *p_src = &lcl_activities8[k*stride];
            int8_t       *p_dst = &act_out.lcl_activities8[k*act_out.stride];
            for(int i=0; i<new_trn+new_tst; i++)
                p_dst[i] = p_src[items[i]];
        }
    }
    else if( format==storage_format::f32 )
    {
        act_out.lcl_activitiesf.resize(act_out.stride*act_out.nchannels);
        memset(&act_out.lcl_activitiesf[0], 0, sizeof(act_out.lcl_activitiesf[0])*act_out.lcl_activitiesf.size());
        for(int64_t k=0; k<nchannels; k++)
        {
            const float *p_src = &lcl_activitiesf[k*stride];
            float       *p_dst = &act_out.lcl_activitiesf[k*act_out.stride];
            for(int i=0; i<new_trn+new_tst; i++)
                p_dst[i] = p_src[items[i]];
        }
    }
    else
        ASSERT(0);
}

void activities::create_view(activities &act_out)
{
    // init output metrics
    ASSERT(format==storage_format::i8 || format==storage_format::f32);
    act_out.format = format;
    act_out.lcl_activities8.clear();
    act_out.lcl_activities8.shrink_to_fit();
    act_out.lcl_activitiesf.clear();
    act_out.lcl_activitiesf.shrink_to_fit();
    act_out.ntrn = ntrn;
    act_out.ntst = ntst;
    act_out.stride = stride;
    act_out.nchannels = nchannels;
    act_out.realms = realms;
    act_out.pwrk_activities8 = pwrk_activities8==NULL ? &lcl_activities8 : pwrk_activities8;
    act_out.pwrk_activitiesf = pwrk_activitiesf==NULL ? &lcl_activitiesf : pwrk_activitiesf;
}

void activities::create_view_window(int plane_height, int plane_width,
                                    int w_offs0, int w_offs1,
                                    int w_height, int w_width,
                                    activities &act_out)
{
    create_view(act_out);
    for(realm<channel> &rr: act_out.realms)
    {
        ASSERT(rr.grid_height%plane_height==0 && rr.grid_width%plane_width==0);
        ASSERT(rr.grid_height/plane_height==rr.grid_width/plane_width);
        int scale = rr.grid_height/plane_height;
        int offs0 = w_offs0*scale;
        int offs1 = w_offs1*scale;
        int new_height = w_height*scale;
        int new_width = w_width*scale;
        realm<channel> rr_new = realm<channel>::create_grid(rr.name, new_height, new_width, rr.grid_depth);
        rr_new.grid_stride0 = rr.grid_height;
        rr_new.grid_stride1 = rr.grid_width;
        rr_new.ch_offs = rr.get_cidx(0, offs0, offs1);
        rr = rr_new;
    }
}

void activities::select_view(select_t select, int window_height, int window_width, activities &act_out)
{
    create_view(act_out);
    act_out.realms.clear();
    for(const realm<channel> &rr: realms)
    {
        if( select==select_t::none )
            continue;
        if( select==select_t::all )
        {
            act_out.realms.push_back(rr);
            continue;
        }
        if( select==select_t::larger )
        {
            if( rr.grid_height>window_height || rr.grid_width>window_width )
                act_out.realms.push_back(rr);
            continue;
        }
        if( select==select_t::not_larger )
        {
            if( rr.grid_height<=window_height && rr.grid_width<=window_width )
                act_out.realms.push_back(rr);
            continue;
        }
        if( select==select_t::smaller )
        {
            if( rr.grid_height<window_height && rr.grid_width<window_width )
                act_out.realms.push_back(rr);
            continue;
        }
        if( select==select_t::not_smaller )
        {
            if( rr.grid_height>=window_height || rr.grid_width>=window_width )
                act_out.realms.push_back(rr);
            continue;
        }
        ASSERT(0);
    }
}

void activities::detach_view()
{
    ASSERT(pwrk_activities8!=NULL && pwrk_activitiesf!=NULL);
    lcl_activities8 = *pwrk_activities8;
    lcl_activitiesf = *pwrk_activitiesf;
    pwrk_activities8 = NULL;
    pwrk_activitiesf = NULL;
}

void activities::gain2(activities &out, std::string prefix) const
{
    ASSERT(pwrk_activities8==NULL && pwrk_activitiesf==NULL);
    ASSERT(is_i8());

    //
    out = *this;
    for(int8_t &v: out.lcl_activities8)
    {
        int8_t vv;
        vv = v;
        vv = vv>=+50 ? +50 : vv;
        vv = vv<=-50 ? -50 : vv;
        v = 2*vv;
    }
    for(realm<channel> &rr: out.realms)
        rr.name = prefix+rr.name;
}

void activities::gaink(int k, activities &out) const
{
    ASSERT(pwrk_activities8==NULL && pwrk_activitiesf==NULL);
    ASSERT(is_i8());

    //
    out = *this;
    for(int64_t i=0; i<out.lcl_activities8.size(); i++)
    {
        int vv = int(out.lcl_activities8[i])*k;
        vv = vv>=+SNET_IMAX ? +SNET_IMAX : vv;
        vv = vv<=-SNET_IMAX ? -SNET_IMAX : vv;
        out.lcl_activities8[i] = vv;
    }
}

void activities::gain2()
{
    ASSERT(pwrk_activities8==NULL && pwrk_activitiesf==NULL);
    ASSERT(is_i8());
    for(int8_t &v: lcl_activities8)
    {
        int8_t vv;
        vv = v;
        vv = vv>=+50 ? +50 : vv;
        vv = vv<=-50 ? -50 : vv;
        v = 2*vv;
    }
}

void activities::add_prefix(std::string s)
{
    for(realm<channel> &rr: realms)
        rr.name = s+rr.name;
}

void activities::add_suffix(std::string s)
{
    for(realm<channel> &rr: realms)
        rr.name = rr.name+s;
}

/*void activities::append_downsampled(int nlevels)
{
    const int64_t ntotal = ntrn+ntst;

    ASSERT(nlevels>=1);
    ASSERT(is_i8());
    ASSERT(vactivities8.size()==nchannels*stride);

    //
    // Determine number of channels to add and new location, resize storage
    //
    int64_t new_channels = 0;
    for(int64_t level=0, cc = nchannels; level<nlevels; level++)
    {
        ASSERT(cc%4==0);
        cc /= 4;
        new_channels += cc;
    }
    int64_t new_location = vactivities8.size();
    vactivities8.reserve(new_location+new_channels*stride);
    //memset(&vactivities8[new_location], 0, sizeof(vactivities8[0])*new_channels*stride);

    //
    // Process realms one by one
    //
    std::vector<realm<channel>> old_realms = realms;
    for(const realm<channel> &_rr: old_realms)
    {
        realm<channel> r_cur = _rr;
        for(int64_t level=0; level<nlevels; level++)
        {
            // integrity checks
            ASSERT(r_cur.grid_height%2==0 && r_cur.grid_width%2==0);

            // append new realm
            char pfx[128];
            sprintf(pfx, "ds%.0f.", pow(2,1+level));
            append_realm(realm<channel>::create_grid(pfx+_rr.name,r_cur.grid_width/2,r_cur.grid_height/2,r_cur.grid_depth));
            realm<channel> r_new = realms.back();

            // perform pooling
            for(int64_t cidx=0; cidx<r_new.grid_depth; cidx++)
                for(int64_t i0=0; i0<r_new.grid_height; i0++)
                    for(int64_t j0=0; j0<r_new.grid_width; j0++)
                    {
                        // determine destination location, zero-fill storage
                        int64_t dst_cidx = r_new.ch_offs;
                        dst_cidx += cidx*r_new.grid_height*r_new.grid_width;
                        dst_cidx += i0*r_new.grid_width + j0;
                        int8_t  *p_dst = get_i8(dst_cidx);
                        memset(p_dst, 0, ntotal*sizeof(*p_dst));

                        // perform pooling
                        for(int64_t i1=0; i1<2; i1++)
                            for(int64_t j1=0; j1<2; j1++)
                            {
                                int64_t src_cidx = r_cur.ch_offs;
                                src_cidx += cidx*r_cur.grid_height*r_cur.grid_width;
                                src_cidx += (i0*2+i1)*r_cur.grid_width+(j0*2+j1);
                                const int8_t *p_src = get_i8(src_cidx);
                                for(int64_t k=0; k<ntotal; k++)
                                {
                                    int a = p_dst[k], s = p_src[k];
                                    p_dst[k] = a>s ? a : s;
                                }
                            }
                    }

            // update realm info
            r_cur = r_new;
        }
    }
}*/

void activities::reserve_storage(int add_channels)
{
    ASSERT(pwrk_activities8==NULL && pwrk_activitiesf==NULL);
    if( is_i8() )
        lcl_activities8.reserve(lcl_activities8.size()+add_channels*stride);
    if( is_f32() )
        lcl_activitiesf.reserve(lcl_activitiesf.size()+add_channels*stride);
}

void activities::drop_storage()
{
    lcl_activities8.clear();
    lcl_activities8.shrink_to_fit();
    lcl_activitiesf.clear();
    lcl_activitiesf.shrink_to_fit();
}

void activities::estimate_separability(const predictions &baseline, double lambdav, float *trn_rms, float *tst_rms, float *tst_cls, bool report)
{
    linear_evaluator eval(baseline, nchannels, lambdav);
    if( format==storage_format::i8 )
    {
        activity8_vector activity(ntrn+ntst);
        for(int k=0; k<nchannels; k++)
        {
            memcpy(&activity[0], get_i8(k), (ntrn+ntst)*sizeof(activity[0]));
            eval.append_activity(activity);
            if( report )
            {
                printf("[C%04d]  TrnRms=%.4f  TstRms=%.4f  TstCls=%.4f\n",
                    int(k),
                    (double)eval.trn_rms(), (double)eval.tst_rms(), (double)eval.tst_cls());
                fflush(stdout);
            }
        }
    }
    else
        ASSERT(0);
    *trn_rms = eval.trn_rms();
    *tst_rms = eval.tst_rms();
    *tst_cls = eval.tst_cls();
}

class sbf_data_source
{
public:
    std::string     realm_name;
    const int8_t    *p_next_row;
    const channel   *p_next_channel;
    int             elements_left;
};

void activities::visualize_item(int item_idx, std::string filename_prefix) const
{
    for(const realm<channel> &rr: realms)
    {
        std::vector<bitmap> planes;
        for(int pidx=0; pidx<rr.grid_depth; pidx++)
        {
            bitmap img;
            img.resize(rr.grid_height, rr.grid_width);
            for(int i=0; i<rr.grid_height; i++)
                for(int j=0; j<rr.grid_width; j++)
                {
                    int src_idx = rr.get_cidx(pidx, i, j);
                    img.draw_v(i, j, get_normalized(src_idx, item_idx));
                }
            planes.push_back(img);
        }
        bitmap out = bitmap::combine_ltr(planes, 2);
        char filename[1024];
        sprintf(filename, "%s-%s.png", filename_prefix.c_str(), rr.name.c_str());
        out.to_png(filename);
    }
}

void serialize(FILE *f, const activities *ptr)
{
    ASSERT(ptr->pwrk_activities8==NULL && ptr->pwrk_activitiesf==NULL);
    int64_t dummy_nclasses = 0; // older version of activities stored number of classes, have to maintain compatibility
    if( ptr->format==storage_format::i8 )
        write_magic8(f, 19);
    else if( ptr->format==storage_format::f32 )
        write_magic8(f, 20);
    else
        ASSERT(0);
    serialize(f, &ptr->ntrn);
    serialize(f, &ptr->ntst);
    serialize(f, &ptr->stride);
    serialize(f, &dummy_nclasses);
    serialize(f, &ptr->nchannels);
    serialize(f, &ptr->realms);
    if( ptr->format==storage_format::i8 )
        serialize(f, &ptr->lcl_activities8);
    else if( ptr->format==storage_format::f32 )
        serialize(f, &ptr->lcl_activitiesf);
    else
        ASSERT(0);
    write_magic8(f, 117);
}

void unserialize(FILE *f, activities *ptr)
{
    int64_t dummy_nclasses = 0; // older version of activities stored number of classes, have to maintain compatibility
    int8_t magic = check_magic8(f, 19, 20);
    ptr->format = magic==19 ? storage_format::i8 : storage_format::f32;
    unserialize(f, &ptr->ntrn);
    unserialize(f, &ptr->ntst);
    unserialize(f, &ptr->stride);
    unserialize(f, &dummy_nclasses);
    unserialize(f, &ptr->nchannels);
    unserialize(f, &ptr->realms);
    ptr->lcl_activities8 = activity8_vector();
    ptr->lcl_activitiesf = activityf_vector();
    if( ptr->format==storage_format::i8 )
        unserialize(f, &ptr->lcl_activities8);
    else if( ptr->format==storage_format::f32 )
        unserialize(f, &ptr->lcl_activitiesf);
    else
        ASSERT(0);
    check_magic8(f, 117);
    ptr->pwrk_activities8 = NULL;
    ptr->pwrk_activitiesf = NULL;
}

int  conv_activities::conv_realm::get_ndata() const
{
    return grid_depth*grid_dim0*grid_dim1;
}

bool conv_activities::conv_realm::is_densely_packed() const
{
    return grid_dim0==grid_stride0 && grid_dim1==grid_stride1;
}

int conv_activities::conv_realm::get_offs(int plane_idx, int offs0, int offs1) const
{
    return offs+plane_idx*grid_stride0*grid_stride1+offs0*grid_stride1+offs1;
}

conv_activities::conv_realm conv_activities::conv_realm::create_grid(const std::string &sname, int gh, int gw, int sidx, int nk)
{
    return create_grid(sname, gh, gw, sidx, 1, 1, nk);
}

conv_activities::conv_realm conv_activities::conv_realm::create_grid(const std::string &sname, int gh, int gw, int sidx, int wh, int ww, int nk)
{
    conv_realm rr;
    ASSERT(gh>=wh);
    ASSERT(gw>=ww);
    rr.name = sname;
    rr.offs = 0;
    rr.grid_stride0 = gh;
    rr.grid_stride1 = gw;
    rr.grid_dim0    = gh;
    rr.grid_dim1    = gw;
    rr.window_dim0  = wh;
    rr.window_dim1  = ww;
    rr.grid_depth   = nk;
    rr.scale_idx    = sidx;
    return rr;
}

conv_activities::conv_realm conv_activities::conv_realm::rebase(int new_offs) const
{
    ASSERT(is_densely_packed());
    conv_realm r = *this;
    r.offs = new_offs;
    return r;
}

int conv_activities::get_img_width() const
{
    return img_width;
}

int conv_activities::get_img_height() const
{
    return img_height;
}

storage_format conv_activities::get_format() const
{
    return format;
}

int64_t conv_activities::get_memory_usage() const
{
    return ndata;
}

bool conv_activities::is_f32() const
{
    return format==storage_format::f32;
}

float* conv_activities::get_f32(int64_t offs)
{
    ASSERT(offs>=0 && offs<ndata);
    ASSERT(format==storage_format::f32);
    activityf_vector &actf = *(coalesce_ptr(pwrk_activitiesf, &lcl_activitiesf));
    return &actf[offs];
}

const float* conv_activities::get_f32(int64_t offs) const
{
    ASSERT(offs>=0 && offs<ndata);
    ASSERT(format==storage_format::f32);
    const activityf_vector &actf = *(coalesce_ptr((const activityf_vector*)pwrk_activitiesf, &lcl_activitiesf));
    return &actf[offs];
}

const conv_activities::conv_realm& conv_activities::realm_by_name(const std::string &name) const
{
    for(const conv_realm &rr: realms)
        if( rr.name==name )
            return rr;
    ASSERT(0);
    abort();
}

bool conv_activities::has_realm(const std::string &name) const
{
    for(const conv_realm &rr: realms)
        if( rr.name==name )
            return true;
    return false;
}

int conv_activities::get_grid_dim0(const std::vector<std::string> &names) const
{
    ASSERT(names.size()>=1);
    int grid_dim0 = realm_by_name(names.front()).grid_dim0;
    for(const std::string &name: names)
        ASSERT(realm_by_name(name).grid_dim0==grid_dim0);
    return grid_dim0;
}

int conv_activities::get_grid_dim1(const std::vector<std::string> &names) const
{
    ASSERT(names.size()>=1);
    int grid_dim1 = realm_by_name(names.front()).grid_dim1;
    for(const std::string &name: names)
        ASSERT(realm_by_name(name).grid_dim1==grid_dim1);
    return grid_dim1;
}

int conv_activities::get_window_dim0(const std::vector<std::string> &names) const
{
    ASSERT(names.size()>=1);
    int window_dim0 = realm_by_name(names.front()).window_dim0;
    for(const std::string &name: names)
        ASSERT(realm_by_name(name).window_dim0==window_dim0);
    return window_dim0;
}

int conv_activities::get_window_dim1(const std::vector<std::string> &names) const
{
    ASSERT(names.size()>=1);
    int window_dim1 = realm_by_name(names.front()).window_dim1;
    for(const std::string &name: names)
        ASSERT(realm_by_name(name).window_dim1==window_dim1);
    return window_dim1;
}


void conv_activities::append_planes_from_realm(const conv_activities &a, const std::string realm_name)
{
    const conv_realm &r_src = a.realm_by_name(realm_name);

    // integrity checks
    ASSERT(pwrk_activitiesf==NULL);
    ASSERT(format==a.format);
    ASSERT(format==storage_format::f32 && lcl_activitiesf.size()==ndata);
    ASSERT(realms.size()==0 || realms.size()==1);
    ASSERT(r_src.is_densely_packed());
    if( realms.size()==1 )
    {
        conv_realm &r_dst = realms[0];
        ASSERT(r_dst.is_densely_packed());
        ASSERT(r_dst.grid_dim0==r_src.grid_dim0 && r_dst.grid_dim1==r_src.grid_dim1);
        ASSERT(r_dst.window_dim0==r_src.window_dim0 && r_dst.window_dim1==r_src.window_dim1);
        ASSERT(ndata==r_dst.grid_depth*r_dst.grid_dim0*r_dst.grid_dim1);
    }

    // insertion
    if( realms.size()==0 )
        realms.push_back(conv_realm::create_grid(realm_name, r_src.grid_dim0, r_src.grid_dim1, r_src.scale_idx, r_src.window_dim0, r_src.window_dim1, 0));
    ASSERT(realms[0].is_densely_packed());

    // copy activities
    const int64_t grid_height = realms[0].grid_dim0, grid_width = realms[0].grid_dim1;
    if( format==storage_format::f32 )
    {
        int64_t old_act_cnt = ndata;
        int64_t new_total   = r_src.grid_depth*grid_height*grid_width;
        lcl_activitiesf.resize(old_act_cnt+new_total, 0x0);
        memcpy(&lcl_activitiesf[old_act_cnt], a.get_f32(r_src.offs), new_total*sizeof(lcl_activitiesf[0]));
    }
    else
        ASSERT(0);

    // update realm
    realms[0].grid_depth += r_src.grid_depth;

    // update data counter
    ndata += r_src.grid_depth*grid_height*grid_width;
}

void conv_activities::append_planes_from1(const conv_activities &a)
{
    ASSERT(a.realms.size()==1);
    append_planes_from_realm(a, a.realms[0].name);
}

int conv_activities::append_realm(const conv_realm &r)
{
    ASSERT(pwrk_activitiesf==NULL);
    ASSERT(r.offs==0 && r.is_densely_packed());
    int64_t new_total = r.grid_depth*r.grid_dim0*r.grid_dim1;

    // reallocate storage
    if( format==storage_format::f32 )
    {
        ASSERT(lcl_activitiesf.size()==ndata);
        int64_t old_size = ndata;
        lcl_activitiesf.resize(old_size+new_total);
        memset(&lcl_activitiesf[old_size], 0, sizeof(lcl_activitiesf[0])*new_total);
    }
    else
        ASSERT(0);

    // add realm
    realms.push_back(r.rebase(ndata));
    ndata += new_total;

    //
    return realms.back().offs;
}

void conv_activities::merge_in(const conv_activities &merged, std::string name_prefix, std::string name_suffix)
{
    ASSERT(pwrk_activitiesf==NULL);
    ASSERT(merged.pwrk_activitiesf==NULL);
    ASSERT(merged.img_height>=img_height && merged.img_height%img_height==0);
    ASSERT(merged.img_width>=img_width   && merged.img_width%img_width==0);
    ASSERT(format==merged.format);

    // copy activities
    if( format==storage_format::f32 )
    {
        // integrity checks
        ASSERT(lcl_activitiesf.size()==ndata);
        ASSERT(merged.lcl_activitiesf.size()==merged.ndata);

        // reallocation
        int64_t old_act_cnt = ndata;
        lcl_activitiesf.resize(old_act_cnt+merged.ndata);
        memcpy(&lcl_activitiesf[old_act_cnt], &merged.lcl_activitiesf[0], merged.ndata*sizeof(merged.lcl_activitiesf[0]));
    }
    else
        ASSERT(0);

    // rebase realms
    int64_t cur_offs = ndata;
    for(const conv_realm &r: merged.realms)
    {
        std::string new_name = name_prefix+r.name+name_suffix;
        ASSERT(r.is_densely_packed());
        ASSERT(!has_realm(new_name));
        realms.push_back(r.rebase(cur_offs));
        realms.back().name = new_name;
        cur_offs += r.get_ndata();
    }
    ASSERT(format!=storage_format::f32 || cur_offs==lcl_activitiesf.size());

    // update nchannels
    ndata += merged.ndata;
}

void conv_activities::reserve_storage(int add_data)
{
    ASSERT(is_f32());
    if( is_f32() && pwrk_activitiesf==NULL )
        lcl_activitiesf.reserve(lcl_activitiesf.size()+add_data);
}

void conv_activities::add_prefix(std::string s)
{
    for(conv_realm &rr: realms)
        rr.name = s+rr.name;
}

void conv_activities::add_suffix(std::string s)
{
    for(conv_realm &rr: realms)
        rr.name = rr.name+s;
}

void conv_activities::convert_to_f32(conv_activities &out) const
{
    ASSERT(is_f32());
    ASSERT(pwrk_activitiesf==NULL);

    if( format==storage_format::f32 )
    {
        out = *this;
        return;
    }
    ASSERT(0);
}

void conv_activities::convert_to_f32()
{
    ASSERT(is_f32());
    ASSERT(pwrk_activitiesf==NULL);
}

void conv_activities::select_view(select_t select, int window_height, int window_width, conv_activities &act_out)
{
    // create view
    ASSERT(format==storage_format::f32);
    act_out = conv_activities();
    act_out.format = format;
    act_out.img_height = img_height;
    act_out.img_width  = img_width;
    act_out.ndata      = ndata;
    act_out.pwrk_activitiesf = pwrk_activitiesf==NULL ? &lcl_activitiesf : pwrk_activitiesf;
    for(const conv_realm &rr: realms)
    {
        if( select==select_t::none )
            continue;
        if( select==select_t::all )
        {
            act_out.realms.push_back(rr);
            continue;
        }
        if( select==select_t::larger )
        {
            if( rr.window_dim0>window_height || rr.window_dim1>window_width )
                act_out.realms.push_back(rr);
            continue;
        }
        if( select==select_t::not_larger )
        {
            if( rr.window_dim0<=window_height && rr.window_dim1<=window_width )
                act_out.realms.push_back(rr);
            continue;
        }
        if( select==select_t::smaller )
        {
            if( rr.window_dim0<window_height && rr.window_dim1<window_width )
                act_out.realms.push_back(rr);
            continue;
        }
        if( select==select_t::not_smaller )
        {
            if( rr.window_dim0>=window_height || rr.window_dim1>=window_width )
                act_out.realms.push_back(rr);
            continue;
        }
        ASSERT(0);
    }
}

conv_activities conv_activities::synchronize_scales() const
{
    conv_activities act_out = conv_activities(img_height, img_width, format);
    if( realms.size()==0 )
        return act_out;

    // determine output size
    int lowest_scale = 999999;
    for(const conv_realm &rr: realms)
        lowest_scale = imin(lowest_scale, rr.scale_idx);
    ASSERT(lowest_scale<999999);
    int out_grid0=0, out_grid1=0;
    for(const conv_realm &rr: realms)
        if( rr.scale_idx==lowest_scale )
        {
            out_grid0 = rr.grid_dim0;
            out_grid1 = rr.grid_dim1;
            break;
        }
    ASSERT(out_grid0>0 && out_grid1>0);

    // perform integrity check
    for(const conv_realm &rr: realms)
    {
        ASSERT(rr.is_densely_packed() && rr.window_dim0==1 && rr.window_dim1==1);
        ASSERT(out_grid0-1==(rr.grid_dim0-1)*(1<<(rr.scale_idx-lowest_scale)));
        ASSERT(out_grid1-1==(rr.grid_dim1-1)*(1<<(rr.scale_idx-lowest_scale)));
    }

    // output
    ASSERT(format==storage_format::f32);
    for(const conv_realm &rr: realms)
    {
        act_out.append_realm(conv_realm::create_grid(rr.name,out_grid0,out_grid1,0,rr.grid_depth));
        const conv_realm &rr_out = act_out.realms.back();
        for(int pidx=0; pidx<rr_out.grid_depth; pidx++)
            for(int i0=0; i0<rr_out.grid_dim0; i0++)
            {
                int shift_factor = rr.scale_idx-lowest_scale;
                const float *p_src_row = get_f32(rr.get_offs(pidx,i0>>shift_factor,0));
                float       *p_dst_row = act_out.get_f32(rr_out.get_offs(pidx,i0,0));
                for(int i1=0; i1<rr_out.grid_dim1; i1++)
                    p_dst_row[i1] = p_src_row[i1>>shift_factor];
            }
    }
    return act_out;
}

linear_predictor::linear_predictor()
{
    is_logit = false;
    nchannels = 0;
    nclasses  = 0;
}

linear_predictor::linear_predictor(const predictions &_baseline)
{
    ASSERT(_baseline.nclasses>=1);
    is_logit = false;
    nchannels = 0;
    nclasses = _baseline.nclasses;
    coeffs = std::vector<float>(nclasses,0.0f);
}

float linear_predictor::get_coeff(int cidx, int cls_idx) const
{
    ASSERT(cidx>=-1 && cidx<nchannels+1);
    if( cidx==-1 )
        cidx = nchannels;
    if( is_logit )
    {
        ASSERT(cls_idx>=0 && cls_idx<nclasses-1);
        return coeffs[cidx*(nclasses-1)+cls_idx];
    }
    else
    {
        ASSERT(cls_idx>=0 && cls_idx<nclasses);
        return coeffs[cidx*nclasses+cls_idx];
    }
}

float linear_predictor::get_l1_factor() const
{
    const int effective_nclasses = is_logit ? nclasses-1 : nclasses;
    std::vector<float> sum_abs(effective_nclasses, 0.0f);
    for(int i=0; i<nchannels; i++)
        for(int j=0; j<effective_nclasses; j++)
            sum_abs[j] += fabsf(get_coeff(i,j));
    float r = 0.0f;
    for(int j=0; j<effective_nclasses; j++)
        r = fmaxf(r, sum_abs[j]);
    return r;
}

bool linear_predictor::get_logit_flag() const
{
    return is_logit;
}

linear_predictor linear_predictor::create_logit(const std::vector<float> &c, int nclasses, int nchannels)
{
    ASSERT(c.size()==(nchannels+1)*(nclasses-1));
    linear_predictor r;
    r.is_logit = true;
    r.nchannels = nchannels;
    r.nclasses = nclasses;
    r.coeffs = c;
    return r;
}

void linear_predictor::update_coeffs(const std::vector<float> &c, int nchannels_new)
{
    ASSERT(nclasses>=1);
    ASSERT(c.size()==(nchannels_new+1)*nclasses);
    coeffs  = c;
    nchannels = nchannels_new;
}

predictions linear_predictor::predict(const predictions &baseline, const activities& act) const
{
    return _predict(baseline, std::vector<const activities*>{&act}, predict_t::prob);
}

predictions linear_predictor::predict(const predictions &baseline, const std::vector<const activities*> v_act) const
{
    return _predict(baseline, v_act, predict_t::prob);
}

predictions linear_predictor::predict_log(const predictions &baseline, const std::vector<const activities*> v_act) const
{
    return _predict(baseline, v_act, predict_t::logprob);
}

predictions linear_predictor::predict_logit(const predictions &baseline, const std::vector<const activities*> v_act) const
{
    return _predict(baseline, v_act, predict_t::logit);
}

conv_predictions linear_predictor::conv_predict1(const conv_activities &act, const calibrator &cal) const
{
    // integrity checks
    ASSERT(nclasses==2);
    int ch_total = 0;
    for(const conv_activities::conv_realm &rr: act.realms)
        ch_total += rr.grid_depth;
    ASSERT(ch_total==nchannels);
    ASSERT(act.realms.size()>=1);
    int64_t dim0 = act.realms[0].grid_dim0;
    int64_t dim1 = act.realms[0].grid_dim1;
    int64_t out_total = dim0*dim1;

    // generate predictions
    conv_predictions r_pred = conv_predictions(dim0, dim1, 1);
    ASSERT(is_logit);
    ASSERT(coeffs.size()==nchannels+1);
    float *p_pred = r_pred.get_ptr();
    float c0 = coeffs[nchannels];
    for(int64_t i=0; i<out_total; i++)
        p_pred[i] = c0;
    int cidx = 0;
    for(const conv_activities::conv_realm &rr: act.realms)
    {
        ASSERT(rr.grid_dim0==dim0 && rr.grid_dim1==dim1);
        ASSERT(rr.window_dim0==1 && rr.window_dim1==1);
        for(int plane_idx=0; plane_idx<rr.grid_depth; plane_idx++)
        {
            fadd_simd(p_pred, act.get_f32(rr.get_offs(plane_idx,0,0)), out_total, coeffs[cidx]);
            cidx++;
        }
    }
    ASSERT(cidx==nchannels);
    for(int64_t i=0; i<out_total; i++)
        p_pred[i] = sigma0(p_pred[i]);
    if( !cal.is_identity() )
    {
        for(int64_t i=0; i<out_total; i++)
            p_pred[i] = cal.calibrate(p_pred[i]);
    }
    return r_pred;
}

predictions linear_predictor::_predict(const predictions &baseline, const std::vector<const activities*> v_act, predict_t pred) const
{
    // integrity checks
    ASSERT(v_act.size()>=1);
    int ch_total = 0;
    for(const activities *a: v_act)
    {
        ASSERT(a->ntrn==v_act[0]->ntrn);
        ASSERT(a->ntst==v_act[0]->ntst);
        ch_total += a->nchannels;
    }
    ASSERT(ch_total==nchannels);
    ASSERT(baseline.nclasses==nclasses);
    ASSERT(baseline.ntrn==v_act[0]->ntrn);
    ASSERT(baseline.ntst==v_act[0]->ntst);

    //
    int64_t ntotal = v_act[0]->ntrn+v_act[0]->ntst;

    // generate predictions
    predictions r = baseline;
    if( is_logit )
    {
        ASSERT(coeffs.size()==(nchannels+1)*(nclasses-1));

        //
        // Logit predictor
        //
        for(int64_t k=0; k<nclasses-1; k++)
        {
            float *p_dst = &r.fpredictions[k*r.stride];
            float v = coeffs[nchannels*(nclasses-1)+k];
            for(int64_t i=0; i<ntotal; i++)
                p_dst[i] = v;
        }
        for(int64_t i=0; i<ntotal; i++)
            r.fpredictions[i+(nclasses-1)*r.stride] = 0;
        int cidx = 0;
        for(const activities *a: v_act)
        {
            for(int64_t ic=0; ic<a->nchannels; ic++)
            {
                for(int64_t k=0; k<nclasses-1; k++)
                {
                    float *p_dst = &r.fpredictions[k*r.stride];
                    float v      = coeffs[cidx*(nclasses-1)+k];
                    if( a->is_i8() )
                    {
                        const int8_t *p_src = a->get_i8(ic);
                        float vv = SNET_I2F*v;
                        for(int64_t i=0; i<ntotal; i++)
                            p_dst[i] += vv*p_src[i];
                    }
                    else if( a->is_f32() )
                    {
                        const float *p_src = a->get_f32(ic);
                        for(int64_t i=0; i<ntotal; i++)
                            p_dst[i] += v*p_src[i];

                    }
                    else
                        ASSERT(0);
                }
                cidx++;
            }
        }
        ASSERT(cidx==nchannels);
        activityf_vector mx(ntotal,0.0f), sum(ntotal,0.0f);
        for(int64_t k=0; k<nclasses; k++)
        {
            float *p = &r.fpredictions[k*r.stride];
            for(int64_t i=0; i<ntotal; i++)
                mx[i] = fmaxf(mx[i],p[i]);
        }
        if( pred==predict_t::logprob )
        {
            //
            // Predict log-probabilities
            //
            for(int64_t k=0; k<nclasses; k++)
            {
                float *p = &r.fpredictions[k*r.stride];
                for(int64_t i=0; i<ntotal; i++)
                {
                    float v = p[i]-mx[i];
                    p[i] = v;
                    sum[i] += expf(v);
                }
            }
            for(int64_t i=0; i<ntotal; i++)
                sum[i] = log(sum[i]);
            for(int64_t k=0; k<nclasses; k++)
            {
                float *p = &r.fpredictions[k*r.stride];
                for(int64_t i=0; i<ntotal; i++)
                    p[i] -= sum[i];
            }
        }
        else if( pred==predict_t::prob )
        {
            //
            // Predict probabilities
            //
            for(int64_t k=0; k<nclasses; k++)
            {
                float *p = &r.fpredictions[k*r.stride];
                for(int64_t i=0; i<ntotal; i++)
                {
                    float v = expf(p[i]-mx[i]);
                    p[i] = v;
                    sum[i] += v;
                }
            }
            for(int64_t i=0; i<ntotal; i++)
                sum[i] = 1.0f/sum[i];
            for(int64_t k=0; k<nclasses; k++)
            {
                float *p = &r.fpredictions[k*r.stride];
                for(int64_t i=0; i<ntotal; i++)
                    p[i] *= sum[i];
            }
        }
        else if( pred==predict_t::logit )
        {
            ASSERT(nclasses==2);

            //
            // Predict logits - they are already here
            //
        }
        else
            ASSERT(0)
    }
    else
    {
        ASSERT(coeffs.size()==(nchannels+1)*nclasses);
        ASSERT(pred==predict_t::prob);

        //
        // Linear predictor
        //
        for(int64_t k=0; k<nclasses; k++)
        {
            float *p_dst = &r.fpredictions[k*r.stride];
            float v = coeffs[nchannels*nclasses+k];
            for(int64_t i=0; i<ntotal; i++)
                p_dst[i] += v;
        }
        int cidx = 0;
        for(const activities *a: v_act)
        {
            for(int64_t ic=0; ic<a->nchannels; ic++)
            {
                for(int64_t k=0; k<nclasses; k++)
                {
                    float *p_dst = &r.fpredictions[k*r.stride];
                    float v      = coeffs[cidx*nclasses+k];
                    if( a->is_i8() )
                    {
                        const int8_t *p_src = a->get_i8(ic);
                        float vv = SNET_I2F*v;
                        for(int64_t i=0; i<ntotal; i++)
                            p_dst[i] += vv*p_src[i];
                    }
                    else if( a->is_f32() )
                    {
                        const float *p_src = a->get_f32(ic);
                        for(int64_t i=0; i<ntotal; i++)
                            p_dst[i] += v*p_src[i];

                    }
                    else
                        ASSERT(0);
                }
                cidx++;
            }
        }
        ASSERT(cidx==nchannels);
    }
    return r;
}

void serialize(FILE *f, const linear_predictor *ptr)
{
    int is_logit = ptr->is_logit ? 1 : 0;
    write_magic8(f, 111);
    serialize(f, &is_logit);
    serialize(f, &ptr->nchannels);
    serialize(f, &ptr->nclasses);
    serialize(f, &ptr->coeffs);
    write_magic8(f, 124);
}

void unserialize(FILE *f, linear_predictor *ptr)
{
    int is_logit;
    check_magic8(f, 111);
    unserialize(f, &is_logit);
    ptr->is_logit = is_logit!=0;
    unserialize(f, &ptr->nchannels);
    unserialize(f, &ptr->nclasses);
    unserialize(f, &ptr->coeffs);
    check_magic8(f, 124);

}

float compute_rating(rating_t r, float omega_rat, const activity8_vector &act, float cov2, float nrm2, const linear_evaluator &eval)
{
    if( r==rating_t::cov2 )
        return cov2;
    else if( r==rating_t::deltar2 )
        return eval.deltar2(&act[0]);
    else if( r==rating_t::fastdr2 )
        return cov2/(omega_rat*eval.get_ntrn()+nrm2);
    else
        ASSERT(0);
    return 0;
}

float compute_rating(rating_t r, float omega_rat, const activityf_vector &act, float cov2, float nrm2, const linear_evaluator &eval)
{
    if( r==rating_t::cov2 )
        return cov2;
    else if( r==rating_t::deltar2 )
        return eval.deltar2(&act[0]);
    else if( r==rating_t::fastdr2 )
        return cov2/(omega_rat*eval.get_ntrn()+nrm2);
    else
        ASSERT(0);
    return 0;
}

float compute_rating(rating_t r, float omega_rat, const activity8_vector &act, const linear_evaluator &eval)
{
    if( r==rating_t::cov2 )
        return eval.estimate_trn_covariation(&act[0]);
    else if( r==rating_t::deltar2 )
        return eval.deltar2(&act[0]);
    else if( r==rating_t::fastdr2 )
        return eval.estimate_trn_covariation(&act[0])/(omega_rat*eval.get_ntrn()+vnrm2(&act[0],eval.get_ntrn()));
    else
        ASSERT(0);
    return 0;
}

float compute_rating(rating_t r, float omega_rat, const activityf_vector &act, const linear_evaluator &eval)
{
    if( r==rating_t::cov2 )
        return eval.estimate_trn_covariation(&act[0]);
    else if( r==rating_t::deltar2 )
        return eval.deltar2(&act[0]);
    else if( r==rating_t::fastdr2 )
        return eval.estimate_trn_covariation(&act[0])/(omega_rat*eval.get_ntrn()+vnrm2(&act[0],eval.get_ntrn()));
    else
        ASSERT(0);
    return 0;
}

void serialize(FILE *f, const and_type_t *ptr)
{
    int iv = (int)(*ptr);
    serialize(f, &iv);
}

void serialize(FILE *f, const or_type_t  *ptr)
{
    int iv = (int)(*ptr);
    serialize(f, &iv);
}

void unserialize(FILE *f, and_type_t *ptr)
{
    int iv;
    unserialize(f, &iv);
    *ptr = (and_type_t)iv;
}

void unserialize(FILE *f, or_type_t  *ptr)
{
    int iv;
    unserialize(f, &iv);
    *ptr = (or_type_t)iv;
}

const int detector::WEIGHT_100 = 100;

detector::conn::conn(int pp, int p0, int p1, int w, const realm<detector> &rr)
{
    ASSERT(w==WEIGHT_100);
    plane_idx   = pp;
    pos0        = p0;
    pos1        = p1;
    weight      = w;
    realm_depth = rr.grid_depth;
    realm_dim0  = rr.grid_height;
    realm_dim1  = rr.grid_width;
    realm_name  = rr.name;
}

detector::conn::conn(int cc, int w, const activities &ref_act)
{
    ASSERT(w==WEIGHT_100);
    const realm<channel> &rr = ref_act.realm_by_channel(cc);
    rr.get_element_info(cc, &plane_idx, &pos0, &pos1);
    weight      = w;
    realm_depth = rr.grid_depth;
    realm_dim0  = rr.grid_height;
    realm_dim1  = rr.grid_width;
    realm_name  = rr.name;
}

detector::detector()
{
    and_type = and_type_t::AND_MIN;
    or_type  = or_type_t::OR_MAX;
    transform = {trf()};
    gain = 1;
}

detector::detector(const seed_detector &seed, and_type_t andtype, or_type_t ortype, const activities &ref_act):and_type(andtype),or_type(ortype)
{
    ASSERT(seed.channel_idx0>=0);
    ASSERT(andtype==and_type_t::AND_MIN || andtype==and_type_t::AND_MUL);
    ASSERT(ortype ==or_type_t::OR_MAX   || ortype ==or_type_t::OR_ADD);
    ASSERT(andtype==seed.and_type);

    gain = int(seed.ggain);
    if( seed.channel_idx0>=0 )
        input0 = {conn(seed.channel_idx0,WEIGHT_100,ref_act)};
    if( seed.channel_idx1>=0 )
        input1 = {conn(seed.channel_idx1,WEIGHT_100,ref_act)};
    if( seed.channel_idx2>=0 )
        input2 = {conn(seed.channel_idx2,WEIGHT_100,ref_act)};
    if( seed.channel_idx3>=0 )
        input3 = {conn(seed.channel_idx3,WEIGHT_100,ref_act)};
    if( seed.channel_idx4>=0 )
        input4 = {conn(seed.channel_idx4,WEIGHT_100,ref_act)};
    transform = {trf()};
}

int detector::inputs_cnt() const
{
    if( input0.size()==0 )
        return 0;
    if( input1.size()==0 )
        return 1;
    if( input2.size()==0 )
        return 2;
    if( input3.size()==0 )
        return 3;
    if( input4.size()==0 )
        return 4;
    return 5;
}

std::vector<std::string> detector::get_realm_names() const
{
    ASSERT(input0.size()>=1);
    std::set<std::string> names;
    names.insert(input0[0].realm_name);
    for(const conn &c: input0)
        names.insert(c.realm_name);
    for(const conn &c: input1)
        names.insert(c.realm_name);
    for(const conn &c: input2)
        names.insert(c.realm_name);
    for(const conn &c: input3)
        names.insert(c.realm_name);
    for(const conn &c: input4)
        names.insert(c.realm_name);
    return std::vector<std::string>(names.begin(),names.end());

}

std::string detector::get_realm_name() const
{
    ASSERT(input0.size()>=1);
    std::string name = input0[0].realm_name;
    for(const conn &c: input0)
        ASSERT(c.realm_name==name);
    for(const conn &c: input1)
        ASSERT(c.realm_name==name);
    for(const conn &c: input2)
        ASSERT(c.realm_name==name);
    for(const conn &c: input3)
        ASSERT(c.realm_name==name);
    for(const conn &c: input4)
        ASSERT(c.realm_name==name);
    return name;
}

int  detector::get_realm_dim0() const
{
    ASSERT(input0.size()>=1);
    int dim0 = input0[0].realm_dim0;
    for(const conn &c: input0)
        ASSERT(c.realm_dim0==dim0);
    for(const conn &c: input1)
        ASSERT(c.realm_dim0==dim0);
    for(const conn &c: input2)
        ASSERT(c.realm_dim0==dim0);
    for(const conn &c: input3)
        ASSERT(c.realm_dim0==dim0);
    for(const conn &c: input4)
        ASSERT(c.realm_dim0==dim0);
    return dim0;
}

int  detector::get_realm_dim1() const
{
    ASSERT(input0.size()>=1);
    int dim1 = input0[0].realm_dim1;
    for(const conn &c: input0)
        ASSERT(c.realm_dim1==dim1);
    for(const conn &c: input1)
        ASSERT(c.realm_dim1==dim1);
    for(const conn &c: input2)
        ASSERT(c.realm_dim1==dim1);
    for(const conn &c: input3)
        ASSERT(c.realm_dim1==dim1);
    for(const conn &c: input4)
        ASSERT(c.realm_dim1==dim1);
    return dim1;
}

const std::vector<detector::conn>& detector::get_input(int i) const
{
    ASSERT(i<inputs_cnt());
    if( i==0 )
        return input0;
    if( i==1 )
        return input1;
    if( i==2 )
        return input2;
    if( i==3 )
        return input3;
    return input4;
}

std::vector<detector::conn>& detector::get_input(int i)
{
    ASSERT(i<inputs_cnt());
    if( i==0 )
        return input0;
    if( i==1 )
        return input1;
    if( i==2 )
        return input2;
    if( i==3 )
        return input3;
    return input4;
}

void detector::add_conn(int input_idx, const conn &c)
{
    ASSERT(input_idx>=0 && input_idx<5);
    if( input_idx==0 )
        input0.push_back(c);
    if( input_idx==1 )
        input1.push_back(c);
    if( input_idx==2 )
        input2.push_back(c);
    if( input_idx==3 )
        input3.push_back(c);
    if( input_idx==4 )
        input4.push_back(c);
}

void detector::generate_inputs5(const activities &ds, int8_t *p0, int8_t *p1, int8_t *p2, int8_t *p3, int8_t *p4) const
{
    ASSERT(transform.size()==1 && transform[0].shift0==0 && transform[0].shift1==0);
    generate_input(ds, input0, transform[0], p0);
    generate_input(ds, input1, transform[0], p1);
    generate_input(ds, input2, transform[0], p2);
    generate_input(ds, input3, transform[0], p3);
    generate_input(ds, input4, transform[0], p4);
}

void detector::generate_inputs5(const activities &ds, float  *p0, float  *p1, float  *p2, float  *p3, float  *p4) const
{
    ASSERT(transform.size()==1 && transform[0].shift0==0 && transform[0].shift1==0);
    generate_input(ds, input0, transform[0], p0);
    generate_input(ds, input1, transform[0], p1);
    generate_input(ds, input2, transform[0], p2);
    generate_input(ds, input3, transform[0], p3);
    generate_input(ds, input4, transform[0], p4);
}

void detector::generate_activity_slow(const activities &ds, activity8_vector &activity) const
{
    // integrity checks
    ASSERT(ds.is_i8());
    ASSERT(and_type==and_type_t::AND_MIN);
    ASSERT(or_type==or_type_t::OR_MAX);

    // process
    const int64_t ntotal = ds.ntrn+ds.ntst;
    activity8_vector x(ntotal,0), r(ntotal,0);
    activity.resize(ntotal);
    int8_t *p_x = &x[0];
    int8_t *p_r = &r[0];
    int8_t *p_out = &activity[0];
    memset(p_out, 0, activity.size()*sizeof(activity[0]));
    for(trf channel_transform: transform)
    {
        // process first input, store directly to r
        generate_input(ds, input0, channel_transform, p_r);

        // second input, if present
        if( input1.size()>0 )
        {
            generate_input(ds, input1, channel_transform, p_x);
            for(int64_t i=0; i<ntotal; i++)
            {
                int8_t s = p_x[i], a = p_r[i];
                p_r[i] = s<a ? s : a;
            }
        }

        // third input, if present
        if( input2.size()>0 )
        {
            generate_input(ds, input2, channel_transform, p_x);
            for(int64_t i=0; i<ntotal; i++)
            {
                int8_t s = p_x[i], a = p_r[i];
                p_r[i] = s<a ? s : a;
            }
        }

        // fourth input, if present
        if( input3.size()>0 )
        {
            generate_input(ds, input3, channel_transform, p_x);
            for(int64_t i=0; i<ntotal; i++)
            {
                int8_t s = p_x[i], a = p_r[i];
                p_r[i] = s<a ? s : a;
            }
        }

        // fifth input, if present
        if( input4.size()>0 )
        {
            generate_input(ds, input4, channel_transform, p_x);
            for(int64_t i=0; i<ntotal; i++)
            {
                int8_t s = p_x[i], a = p_r[i];
                p_r[i] = s<a ? s : a;
            }
        }

        // append to final activity
        for(int64_t i=0; i<ntotal; i++)
        {
            int8_t s = p_r[i], a = p_out[i];
            p_out[i] = s>a ? s : a;
        }
    }
}

void detector::generate_activity_slow(const activities &ds, activityf_vector &activity) const
{
    // integrity checks
    ASSERT(ds.is_f32());
    ASSERT(and_type==and_type_t::AND_MIN || and_type==and_type_t::AND_MUL);
    ASSERT(or_type==or_type_t::OR_MAX);

    // process
    const int64_t ntotal = ds.ntrn+ds.ntst;
    activityf_vector x(ntotal,0.0f), r(ntotal,0.0f);
    activity.resize(ntotal);
    float *p_x = &x[0];
    float *p_r = &r[0];
    float *p_out = &activity[0];
    memset(p_out, 0, activity.size()*sizeof(activity[0]));
    for(trf channel_transform: transform)
    {
        // process first input, store directly to r
        generate_input(ds, input0, channel_transform, p_r);

        // second input, if present
        if( input1.size()>0 )
        {
            generate_input(ds, input1, channel_transform, p_x);
            if( and_type==and_type_t::AND_MIN )
            {
                for(int64_t i=0; i<ntotal; i++)
                {
                    float s = p_x[i], a = p_r[i];
                    p_r[i] = s<a ? s : a;
                }
            }
            else if( and_type==and_type_t::AND_MUL )
            {
                for(int64_t i=0; i<ntotal; i++)
                    p_r[i] *= p_x[i];
            }
            else
                ASSERT(0)
        }

        // third input, if present
        if( input2.size()>0 )
        {
            generate_input(ds, input2, channel_transform, p_x);
            if( and_type==and_type_t::AND_MIN )
            {
                for(int64_t i=0; i<ntotal; i++)
                {
                    float s = p_x[i], a = p_r[i];
                    p_r[i] = s<a ? s : a;
                }
            }
            else if( and_type==and_type_t::AND_MUL )
            {
                for(int64_t i=0; i<ntotal; i++)
                    p_r[i] *= p_x[i];
            }
            else
                ASSERT(0)
        }

        // fourth input, if present
        if( input3.size()>0 )
        {
            generate_input(ds, input3, channel_transform, p_x);
            if( and_type==and_type_t::AND_MIN )
            {
                for(int64_t i=0; i<ntotal; i++)
                {
                    float s = p_x[i], a = p_r[i];
                    p_r[i] = s<a ? s : a;
                }
            }
            else if( and_type==and_type_t::AND_MUL )
            {
                for(int64_t i=0; i<ntotal; i++)
                    p_r[i] *= p_x[i];
            }
            else
                ASSERT(0)
        }

        // fifth input, if present
        if( input4.size()>0 )
        {
            generate_input(ds, input4, channel_transform, p_x);
            if( and_type==and_type_t::AND_MIN )
            {
                for(int64_t i=0; i<ntotal; i++)
                {
                    float s = p_x[i], a = p_r[i];
                    p_r[i] = s<a ? s : a;
                }
            }
            else if( and_type==and_type_t::AND_MUL )
            {
                for(int64_t i=0; i<ntotal; i++)
                    p_r[i] *= p_x[i];
            }
            else
                ASSERT(0)
        }

        // append to final activity
        for(int64_t i=0; i<ntotal; i++)
        {
            float s = p_r[i], a = p_out[i];
            p_out[i] = s>a ? s : a;
        }
    }
}

conv_activities detector::conv_generate_activity_slow(const conv_activities &act_src, std::string out_realm_name) const
{
    // integrity checks
    ASSERT(act_src.is_f32());
    ASSERT(and_type==and_type_t::AND_MIN || and_type==and_type_t::AND_MUL);
    ASSERT(or_type==or_type_t::OR_MAX);

    // process
    const std::vector<std::string> realm_names = get_realm_names();
    //const conv_activities::conv_realm &rr_src = act_src.realm_by_name(realm_name);
    const int src_dim0 = act_src.get_grid_dim0(realm_names);// rr_src.grid_dim0;
    const int src_dim1 = act_src.get_grid_dim1(realm_names);// rr_src.grid_dim1;
    const int window_dim0 = get_realm_dim0();
    const int window_dim1 = get_realm_dim1();
    const int out_dim0 = src_dim0-window_dim0+1;
    const int out_dim1 = src_dim1-window_dim1+1;
    const int out_total = out_dim0*out_dim1;
    ASSERT(window_dim0==act_src.get_window_dim0(realm_names) && window_dim1==act_src.get_window_dim1(realm_names));
    activityf_vector buf0(out_total,0.0f), buf1(out_total,0.0f);
    conv_activities act_out = conv_activities(act_src.get_img_height(), act_src.get_img_width(), act_src.get_format());
    act_out.append_realm(conv_activities::conv_realm::create_grid(out_realm_name,out_dim0,out_dim1,0,1));
    float *p_out_buf = act_out.get_f32();
    float *p_buf0 = &buf0[0];
    float *p_buf1 = &buf1[0];
    for(trf channel_transform: transform)
    {
        // process first input, store directly to r
        conv_generate_input(act_src, src_dim0, src_dim1, input0, channel_transform, window_dim0, window_dim1, p_buf0);

        // second input, if present
        if( input1.size()>0 )
        {
            conv_generate_input(act_src, src_dim0, src_dim1, input1, channel_transform, window_dim0, window_dim1, p_buf1);
            if( and_type==and_type_t::AND_MIN )
                fmerge_min(p_buf0, p_buf1, out_total);
            else if( and_type==and_type_t::AND_MUL )
                fmerge_mul(p_buf0, p_buf1, out_total);
            else
                ASSERT(0);
        }

        // third input, if present
        if( input2.size()>0 )
        {
            conv_generate_input(act_src, src_dim0, src_dim1, input2, channel_transform, window_dim0, window_dim1, p_buf1);
            if( and_type==and_type_t::AND_MIN )
                fmerge_min(p_buf0, p_buf1, out_total);
            else if( and_type==and_type_t::AND_MUL )
                fmerge_mul(p_buf0, p_buf1, out_total);
            else
                ASSERT(0);
        }

        // fourth input, if present
        if( input3.size()>0 )
        {
            conv_generate_input(act_src, src_dim0, src_dim1, input3, channel_transform, window_dim0, window_dim1, p_buf1);
            if( and_type==and_type_t::AND_MIN )
                fmerge_min(p_buf0, p_buf1, out_total);
            else if( and_type==and_type_t::AND_MUL )
                fmerge_mul(p_buf0, p_buf1, out_total);
            else
                ASSERT(0);
        }

        // fifth input, if present
        if( input4.size()>0 )
        {
            conv_generate_input(act_src, src_dim0, src_dim1, input4, channel_transform, window_dim0, window_dim1, p_buf1);
            if( and_type==and_type_t::AND_MIN )
                fmerge_min(p_buf0, p_buf1, out_total);
            else if( and_type==and_type_t::AND_MUL )
                fmerge_mul(p_buf0, p_buf1, out_total);
            else
                ASSERT(0);
        }

        // append to final activity
        fmerge_max(p_out_buf, p_buf0, out_total);
    }
    return act_out;
}

void detector::generate_input(const activities &ds, const std::vector<conn> &input, trf channel_transform, int8_t *activity) const
{
    // integrity checks
    ASSERT(ds.is_i8());
    ASSERT(or_type==or_type_t::OR_MAX);

    //
    const int64_t ntotal = ds.ntrn+ds.ntst;

    //
    // Quick exit for empty input
    //
    if( input.size()==0 )
    {
        memset(activity, flt_to_int8(1.0f), ntotal*sizeof(activity[0]));
        return;
    }

    //
    // Non-empty input
    //
    memset(activity, 0, ntotal*sizeof(activity[0]));
    for(const conn &c: input)
    {
        const realm<channel> &rr = ds.realm_by_name(c.realm_name);
        int cidx = rr.get_cidx(c.plane_idx,c.pos0+channel_transform.shift0,c.pos1+channel_transform.shift1);
        ASSERT(c.weight==WEIGHT_100);
        ASSERT(cidx>=0 && cidx<ds.nchannels);
        const int8_t *p_src = ds.get_i8(cidx);
        for(int64_t i=0; i<ntotal; i++)
        {
            int8_t s = p_src[i], a = activity[i];
            activity[i] = s>a ? s : a;
        }
    }
    apply_gain(activity, ntotal, gain);
}

void detector::generate_input(const activities &ds, const std::vector<conn> &input, trf channel_transform, float  *activity) const
{
    // integrity checks
    ASSERT(ds.is_f32());
    ASSERT(or_type==or_type_t::OR_MAX);

    //
    const int64_t ntotal = ds.ntrn+ds.ntst;

    //
    // Quick exit for empty input
    //
    if( input.size()==0 )
    {
        for(int64_t k=0; k<ntotal; k++)
            activity[k] = 1.0f;
        return;
    }

    //
    // Non-empty input
    //
    memset(activity, 0, ntotal*sizeof(activity[0]));
    for(const conn &c: input)
    {
        const realm<channel> &rr = ds.realm_by_name(c.realm_name);
        int cidx = rr.get_cidx(c.plane_idx,c.pos0+channel_transform.shift0,c.pos1+channel_transform.shift1);
        ASSERT(c.weight==WEIGHT_100);
        ASSERT(cidx>=0 && cidx<ds.nchannels);
        const float *p_src = ds.get_f32(cidx);
        for(int64_t i=0; i<ntotal; i++)
        {
            float s = p_src[i], a = activity[i];
            activity[i] = s>a ? s : a;
        }
    }
    apply_gain(activity, ntotal, gain);
}

void detector::conv_generate_input(
        const conv_activities &ds, int src_dim0, int src_dim1,
        const std::vector<conn> &input, trf channel_transform, int window_dim0, int window_dim1,
        float *activity) const
{
    // integrity checks
    ASSERT(ds.is_f32());
    ASSERT(or_type==or_type_t::OR_MAX);

    //
    const int64_t out_dim0  = src_dim0-window_dim0+1;
    const int64_t out_dim1  = src_dim1-window_dim1+1;
    const int64_t out_total = out_dim0*out_dim1;

    //
    // Quick exit for empty input
    //
    if( input.size()==0 )
    {
        for(int64_t k=0; k<out_total; k++)
            activity[k] = 1.0f;
        return;
    }

    //
    // Non-empty input
    //
    memset(activity, 0, out_total*sizeof(activity[0]));
    for(const conn &c: input)
    {
        ASSERT(c.weight==WEIGHT_100);
        const conv_activities::conv_realm &rr_src = ds.realm_by_name(c.realm_name);
        const float *p_src = ds.get_f32(rr_src.get_offs(c.plane_idx,c.pos0+channel_transform.shift0,c.pos1+channel_transform.shift1));
        for(int64_t i0=0; i0<out_dim0; i0++)
        {
            //for(int64_t i1=0; i1<out_dim1; i1++) //!!!!!!!!!!!!!!!!!!
            //    activity[i0*out_dim1+i1] = fmaxf(activity[i0*out_dim1+i1], p_src[i0*rr_src.grid_stride1+i1]);
            fmerge_max(activity+i0*out_dim1, p_src+i0*rr_src.grid_stride1, out_dim1);
        }
    }
    apply_gain(activity, out_total, gain);
}


void serialize(FILE *f, const detector::conn *ptr)
{
    write_magic8(f, 117);
    serialize(f, &ptr->realm_name);
    serialize(f, &ptr->realm_depth);
    serialize(f, &ptr->realm_dim0);
    serialize(f, &ptr->realm_dim1);
    serialize(f, &ptr->plane_idx);
    serialize(f, &ptr->pos0);
    serialize(f, &ptr->pos1);
    serialize(f, &ptr->weight);
    write_magic8(f, 11);
}

void serialize(FILE *f, const detector::trf *ptr)
{
    write_magic8(f, 24);
    serialize(f, &ptr->shift0);
    serialize(f, &ptr->shift1);
    write_magic8(f, 69);
}

void serialize(FILE *f, const detector *ptr)
{
    write_magic8(f, 83);
    serialize(f, &ptr->desc);
    serialize(f, &ptr->and_type);
    serialize(f, &ptr->or_type);
    serialize(f, &ptr->transform);
    serialize(f, &ptr->gain);
    serialize(f, &ptr->input0);
    serialize(f, &ptr->input1);
    serialize(f, &ptr->input2);
    serialize(f, &ptr->input3);
    serialize(f, &ptr->input4);
    write_magic8(f, 38);
}

void unserialize(FILE *f, detector::trf *ptr)
{
    check_magic8(f, 24);
    unserialize(f, &ptr->shift0);
    unserialize(f, &ptr->shift1);
    check_magic8(f, 69);
}

void unserialize(FILE *f, detector::conn *ptr)
{
    check_magic8(f, 117);
    unserialize(f, &ptr->realm_name);
    unserialize(f, &ptr->realm_depth);
    unserialize(f, &ptr->realm_dim0);
    unserialize(f, &ptr->realm_dim1);
    unserialize(f, &ptr->plane_idx);
    unserialize(f, &ptr->pos0);
    unserialize(f, &ptr->pos1);
    unserialize(f, &ptr->weight);
    check_magic8(f, 11);
}

void unserialize(FILE *f, detector *ptr)
{
    check_magic8(f, 83);
    unserialize(f, &ptr->desc);
    unserialize(f, &ptr->and_type);
    unserialize(f, &ptr->or_type);
    unserialize(f, &ptr->transform);
    unserialize(f, &ptr->gain);
    unserialize(f, &ptr->input0);
    unserialize(f, &ptr->input1);
    unserialize(f, &ptr->input2);
    unserialize(f, &ptr->input3);
    unserialize(f, &ptr->input4);
    check_magic8(f, 38);
}

void linear_evaluator::daa::alloc(int trn, int tst, int baserows, int qcols, int nclasses)
{
    activity.resize(trn+tst);
    qu.resize(qcols);
    rqu.resize(nclasses);
    coeffs.resize(baserows);
    invalidate();
}

void linear_evaluator::daa::invalidate()
{
    signature = 0;
    signed_by = NULL;
    memset(&activity[0], 0, activity.size()*sizeof(activity[0]));
    memset(&coeffs[0],   0, coeffs.size()*sizeof(coeffs[0]));
    memset(&rqu[0],      0, rqu.size()*sizeof(rqu[0]));
    memset(&qu[0],       0, qu.size()*sizeof(qu[0]));
    deltar2 = 0.0f;
    qnrm = 0.0f;
}

float cov11(const int8_t *res, const int8_t *activity, int64_t ntrn)
{
    float v = 0;
#ifdef SNET_AVX2
    const int64_t CHUNK_SIZE = 8;
    const int64_t nchunks = ntrn/CHUNK_SIZE;
    float va[CHUNK_SIZE];
    __m256 mm_cov = _mm256_set1_ps(0.0f);
    for(int64_t i=0; i<nchunks; i++)
    {
        __m256 mm_res = _mm256_cvtepi32_ps(_mm256_cvtepi8_epi32(_mm_set1_epi64x(((const int64_t*)res)[i])));
        __m256 mm_act = _mm256_cvtepi32_ps(_mm256_cvtepi8_epi32(_mm_set1_epi64x(((const int64_t*)activity)[i])));
        mm_cov = _mm256_fmadd_ps(mm_res,mm_act,mm_cov);
    }
    _mm256_storeu_ps(va, mm_cov);
    for(int k=0; k<CHUNK_SIZE; k++)
        v += va[k];
#else
    const int64_t CHUNK_SIZE = 4;
    const int64_t nchunks = ntrn/CHUNK_SIZE;
    __m128 mm_cov = _mm_set1_ps(0.0f);
    for(int64_t i=0; i<nchunks; i++)
    {
        __m128 mm_res = _mm_cvtepi32_ps(_mm_cvtepi8_epi32(_mm_set1_epi32(((const int32_t*)res)[i])));
        __m128 mm_act = _mm_cvtepi32_ps(_mm_cvtepi8_epi32(_mm_set1_epi32(((const int32_t*)activity)[i])));
        mm_cov = _mm_add_ps(_mm_mul_ps(mm_res,mm_act),mm_cov);
    }
    mm_cov = _mm_hadd_ps(mm_cov,mm_cov);
    mm_cov = _mm_hadd_ps(mm_cov,mm_cov);
    _mm_store_ss(&v, mm_cov);
#endif
    for(int64_t i=nchunks*CHUNK_SIZE; i<ntrn; i++)
        v += float(res[i])*float(activity[i]);
    return v*(SNET_I2F*SNET_I2F);
}

float cov11(const float *res, const float *activity, int64_t ntrn)
{
    float v = 0;
    const int64_t CHUNK_SIZE = 4;
    const int64_t nchunks = ntrn/CHUNK_SIZE;
    __m128 mm_cov = _mm_set1_ps(0.0f);
    for(int64_t i=0; i<nchunks; i++)
    {
        __m128 mm_res = _mm_load_ps(     res+i*CHUNK_SIZE);
        __m128 mm_act = _mm_load_ps(activity+i*CHUNK_SIZE);
        mm_cov = _mm_add_ps(_mm_mul_ps(mm_res,mm_act),mm_cov);
    }
    mm_cov = _mm_hadd_ps(mm_cov,mm_cov);
    mm_cov = _mm_hadd_ps(mm_cov,mm_cov);
    _mm_store_ss(&v, mm_cov);
    for(int64_t i=nchunks*CHUNK_SIZE; i<ntrn; i++)
        v += res[i]*activity[i];
    return v;
}

void cov21(const int8_t *res0, const int8_t *res1,
                   const int8_t *activity,
                   int64_t ntrn,
                   float &r0, float &r1)
{
    //r0 = tasklet_cov11(residuals0, activity, ntrn);
    //r1 = tasklet_cov11(residuals1, activity, ntrn);
    float v0=0, v1=0;
#ifdef SNET_AVX2
    const int64_t CHUNK_SIZE = 8;
    const int64_t nchunks = ntrn/CHUNK_SIZE;
    float va0[CHUNK_SIZE], va1[CHUNK_SIZE];
    __m256 mm_cov0 = _mm256_set1_ps(0.0f);
    __m256 mm_cov1 = _mm256_set1_ps(0.0f);
    for(int64_t i=0; i<nchunks; i++)
    {
        __m256 mm_res0 = _mm256_cvtepi32_ps(_mm256_cvtepi8_epi32(_mm_set1_epi64x(((const int64_t*)res0)[i])));
        __m256 mm_res1 = _mm256_cvtepi32_ps(_mm256_cvtepi8_epi32(_mm_set1_epi64x(((const int64_t*)res1)[i])));
        __m256 mm_act  = _mm256_cvtepi32_ps(_mm256_cvtepi8_epi32(_mm_set1_epi64x(((const int64_t*)activity)[i])));
        mm_cov0 = _mm256_fmadd_ps(mm_res0,mm_act,mm_cov0);
        mm_cov1 = _mm256_fmadd_ps(mm_res1,mm_act,mm_cov1);
    }
    _mm256_storeu_ps(va0, mm_cov0);
    _mm256_storeu_ps(va1, mm_cov1);
    for(int k=0; k<CHUNK_SIZE; k++)
    {
        v0 += va0[k];
        v1 += va1[k];
    }
#else
    const int64_t CHUNK_SIZE = 4;
    const int64_t nchunks = ntrn/CHUNK_SIZE;
    __m128 mm_cov0 = _mm_set1_ps(0.0f);
    __m128 mm_cov1 = _mm_set1_ps(0.0f);
    for(int64_t i=0; i<nchunks; i++)
    {
        __m128 mm_res0 = _mm_cvtepi32_ps(_mm_cvtepi8_epi32(_mm_set1_epi32(((const int32_t*)res0)[i])));
        __m128 mm_res1 = _mm_cvtepi32_ps(_mm_cvtepi8_epi32(_mm_set1_epi32(((const int32_t*)res1)[i])));
        __m128 mm_act  = _mm_cvtepi32_ps(_mm_cvtepi8_epi32(_mm_set1_epi32(((const int32_t*)activity)[i])));
        mm_cov0 = _mm_add_ps(_mm_mul_ps(mm_res0,mm_act),mm_cov0);
        mm_cov1 = _mm_add_ps(_mm_mul_ps(mm_res1,mm_act),mm_cov1);
    }
    mm_cov0 = _mm_hadd_ps(mm_cov0,mm_cov0);
    mm_cov0 = _mm_hadd_ps(mm_cov0,mm_cov0);
    mm_cov1 = _mm_hadd_ps(mm_cov1,mm_cov1);
    mm_cov1 = _mm_hadd_ps(mm_cov1,mm_cov1);
    _mm_store_ss(&v0, mm_cov0);
    _mm_store_ss(&v1, mm_cov1);
#endif
    for(int64_t i=nchunks*CHUNK_SIZE; i<ntrn; i++)
    {
        v0 += float(res0[i])*float(activity[i]);
        v1 += float(res1[i])*float(activity[i]);
    }
    r0 = v0*(SNET_I2F*SNET_I2F);
    r1 = v1*(SNET_I2F*SNET_I2F);
}

std::atomic<int32_t> linear_evaluator::pcnt_truerms(0);
std::atomic<int32_t> linear_evaluator::pcnt_append(0);
std::atomic<int32_t> linear_evaluator::pcnt_cov(0);
std::atomic<int32_t> linear_evaluator::pcnt_da(0);
std::atomic<int32_t> linear_evaluator::pcnt_ar(0);

linear_evaluator::linear_evaluator(const predictions &_pred, int64_t _max_channels, double _lambdav):
    baseline(_pred)
{
    const int64_t ntrn      = baseline.ntrn;
    const int64_t ntst      = baseline.ntst;
    const int64_t nclasses  = baseline.nclasses;

    lambdav = _lambdav;
    max_channels = _max_channels;
    max_rows = max_channels+1;
    ncols = ntrn+max_rows;
    nrows = 0;
    signature = rand();

    //
    // Initialize residuals
    //
    residuals_stride = iDivUp(ncols,ALIGN_ACTIVITIES_BY)*ALIGN_ACTIVITIES_BY;
    residuals_f32.resize(nclasses*residuals_stride, 0.0f);
    for(int i=0; i<ntrn; i++)
        residuals_f32[i+baseline.labels[i]*residuals_stride] = 1.0f;
    for(int64_t k=0; k<nclasses; k++)
        for(int64_t i=0; i<ntrn; i++)
            residuals_f32[i+k*residuals_stride] -= baseline.fpredictions[i+k*baseline.stride];
    residuals_i8.resize(residuals_stride*nclasses, 0);
    residuals_squared.resize(nclasses, 0);

    //
    // Allocate place for QR factorization, etc.
    // Zero-fill everything.
    //
    lq_trf.setlength(max_rows, max_rows);
    for(int i=0; i<lq_trf.rows(); i++)
        for(int j=0; j<lq_trf.cols(); j++)
            lq_trf(i,j) = 0.0;
    lq_stride = iDivUp(ncols,ALIGN_FLOATS_BY)*ALIGN_FLOATS_BY;
    lq_ort32.resize(max_rows*lq_stride, 0.0f);
    qb.setlength(max_rows, nclasses);
    for(int i=0; i<qb.rows(); i++)
        for(int j=0; j<qb.cols(); j++)
            qb(i,j) = 0.0;

    //
    // Test set activities
    //
    if( ntst>0 )
    {
        test_set.setlength(max_rows, ntst);
        for(int i=0; i<test_set.rows(); i++)
            for(int j=0; j<test_set.cols(); j++)
                test_set(i,j) = 0.0;
    }

    //
    // Add first row, constant term, computed as average
    //
    double nrm = 0;
    for(int64_t i=0; i<ntrn; i++)
    {
        lq_ort32[i] = 1.0f;
        nrm += 1.0;
    }
    lq_ort32[ntrn+nrows] = lambdav;
    nrm += lambdav*lambdav;
    nrm = sqrt(nrm);
    double ss = 1/nrm;
    for(int64_t i=0; i<ncols; i++)
        lq_ort32[i] *= ss;
    lq_trf(0,0) = nrm;
    for(int64_t i=0; i<ntst; i++)
        test_set(0,i) = 1.0;
    update_residuals(&lq_ort32[0]);
    nrows++;

    //
    // Allocate matrix of output predictions and output predictor
    //
    output = predictions(baseline.ntrn, baseline.ntst, baseline.nclasses, baseline.labels);
    predictor = linear_predictor(baseline);

    //
    // Update error metrics
    //
    update_predictor_output_errors();
}

linear_evaluator::~linear_evaluator()
{
}

float linear_evaluator::true_trn_rms(const float *activity_f32, const int8_t *activity_i8) const
{
    const int64_t ntrn      = baseline.ntrn;
    const int64_t ntst      = baseline.ntst;
    const int64_t nclasses  = baseline.nclasses;
    float v;
    timer counter(pcnt_truerms);

    //
    // Orthogonalize activity wrt to already present activities
    //
    activityf_vector new_row(ncols,0.0f);
    float *p_row = &new_row[0];
    if( activity_i8 )
    {
        ASSERT(activity_f32==NULL);
        for(int64_t i=0; i<ntrn; i++)
            p_row[i] =  int8_to_flt(activity_i8[i]);
    }
    else
    {
        ASSERT(activity_i8==NULL);
        for(int64_t i=0; i<ntrn; i++)
            p_row[i] =  activity_f32[i];
    }
    if( ntrn+nrows<ncols )
        p_row[ntrn+nrows] = lambdav;
    float tmpnrm;
    orthonormalize_row(p_row, &tmpnrm);

    //
    // For each row of residual matrix perform:
    // * dot product with orthogonalized activity
    // * subtraction and accumulation
    //
    float true_rms = 0;
    for(int64_t k=0; k<nclasses; k++)
    {
        const float *p_res = &residuals_f32[k*residuals_stride];
        v = 0.0f;
        for(int64_t i=0; i<ncols; i++)
            v += p_row[i]*p_res[i];
        for(int64_t j=0; j<ntrn; j++)
        {
            float vv = p_res[j]-v*p_row[j];
            true_rms += vv*vv;
        }
    }
    return true_rms;
}

void linear_evaluator::append_activity(const float *activity_f32, const int8_t *activity_i8, int64_t activity_size)
{
    const int64_t ntrn = baseline.ntrn;
    const int64_t ntst = baseline.ntst;
    const int64_t nclasses = baseline.nclasses;
    float v;
    timer counter(pcnt_append);

    //
    // Integrity checks
    //
    ASSERT(activity_size==ntrn+ntst);
    ASSERT(nrows<max_rows);

    //
    // Prepare new row of design matrix
    //
    activityf_vector new_row(ncols,0.0f);
    float *p_row = &new_row[0];
    if( activity_i8 )
    {
        ASSERT(activity_f32==NULL);
        for(int64_t i=0; i<ntrn; i++)
            p_row[i] =  int8_to_flt(activity_i8[i]);
    }
    else
    {
        ASSERT(activity_i8==NULL);
        for(int64_t i=0; i<ntrn; i++)
            p_row[i] =  activity_f32[i];
    }
    p_row[ntrn+nrows] = lambdav;

    //
    // Orthogonalize row with respect to Q, append to Q
    //
    std::vector<float> tmp_vec(nrows);
    float vnrmf;
    orthonormalize_row(p_row, &vnrmf, &tmp_vec[0]);
    for(int64_t i=0; i<nrows; i++)
        lq_trf[nrows][i] = tmp_vec[i];
    lq_trf[nrows][nrows] = vnrmf;
    for(int64_t i=0; i<ncols; i++)
        lq_ort32[i+nrows*lq_stride] = p_row[i];

    //
    // Update residuals and product of Q*residuals
    //
    update_residuals(p_row);
    if( activity_i8 )
    {
        ASSERT(activity_f32==NULL);
        for(int64_t i=0; i<ntst; i++)
            test_set(nrows,i) = int8_to_flt(activity_i8[ntrn+i]);
    }
    else
    {
        ASSERT(activity_i8==NULL);
        for(int64_t i=0; i<ntst; i++)
            test_set(nrows,i) = activity_f32[ntrn+i];
    }

    //
    // Row is added
    //
    nrows++;
    signature++;

    //
    // Update error metrics
    //
    update_predictor_output_errors();
}

const int8_t* linear_evaluator::get_res8() const
{
    return &residuals_i8[0];
}

const float* linear_evaluator::get_res32() const
{
    return &residuals_f32[0];
}

int64_t linear_evaluator::get_res_stride() const
{
    return residuals_stride;
}

int linear_evaluator::get_nclasses() const
{
    ASSERT(baseline.nclasses>=2);
    return baseline.nclasses;
}

int linear_evaluator::get_ntrn() const
{
    return baseline.ntrn;
}

void linear_evaluator::analyze(const float  *activity, daa *p_result) const
{
    p_result->alloc(baseline.ntrn, baseline.ntst, nrows, ncols, baseline.nclasses);
    for(int64_t k=0; k<baseline.ntrn+baseline.ntst; k++)
        p_result->activity[k] = activity[k];
    run_analysis(p_result);
}

void linear_evaluator::analyze(const int8_t *activity, daa *p_result) const
{
    p_result->alloc(baseline.ntrn, baseline.ntst, nrows, ncols, baseline.nclasses);
    for(int64_t k=0; k<baseline.ntrn+baseline.ntst; k++)
        p_result->activity[k] = SNET_I2F*activity[k];
    run_analysis(p_result);
}

float linear_evaluator::deltar2(const float  *activity) const
{
    daa a;
    analyze(activity, &a);
    return a.deltar2;
}

float linear_evaluator::deltar2(const int8_t *activity) const
{
    daa a;
    analyze(activity, &a);
    return a.deltar2;
}

float linear_evaluator::true_trn_rms(const int8_t *activity) const
{
    return true_trn_rms(NULL, activity);
}

float linear_evaluator::true_trn_rms(const float *activity_f32) const
{
    return true_trn_rms(activity_f32, NULL);
}

float linear_evaluator::estimate_trn_covariation(const int8_t *activity) const
{
    timer counter(pcnt_cov);
    int64_t widx;
    float   wcov;
    estimate_trn_covariation(activity, 1, baseline.ntrn, &widx, &wcov);
    return wcov;
}

float linear_evaluator::estimate_trn_covariation(const float *activity) const
{
    timer counter(pcnt_cov);
    int64_t widx;
    float   wcov;
    estimate_trn_covariation(activity, 1, baseline.ntrn, &widx, &wcov);
    return wcov;
}

void linear_evaluator::estimate_trn_covariation(const int8_t *activity_batch, int64_t batch_size, int64_t stride, int64_t *winner_idx, float *winner_cov) const
{
    estimate_trn_covariation(NULL, activity_batch, batch_size, stride, winner_idx, winner_cov);
}

void linear_evaluator::estimate_trn_covariation(const float *activity_batch, int64_t batch_size, int64_t stride, int64_t *winner_idx, float *winner_cov) const
{
    estimate_trn_covariation(activity_batch, NULL, batch_size, stride, winner_idx, winner_cov);
}

void linear_evaluator::estimate_trn_covariation(const float *activity_batch_f32, const int8_t *activity_batch_i8,
                                                int64_t batch_size, int64_t stride,
                                                int64_t *winner_idx, float *winner_cov) const
{
    timer counter(pcnt_cov);
    int64_t ntrn = baseline.ntrn;
    ASSERT(activity_batch_f32!=NULL || activity_batch_i8!=NULL);
    ASSERT(activity_batch_f32==NULL || activity_batch_i8==NULL);

    bool use_i8 = activity_batch_i8!=NULL;
    *winner_cov = 0.0f;
    *winner_idx = -1;
    for(int64_t eidx=0; eidx<batch_size; eidx++)
    {
        // first, compute covariations and normalized activities
        ASSERT(baseline.nclasses<1000);
        float cov_vec[baseline.nclasses];
        if( use_i8 )
            compute_covariations(activity_batch_i8+eidx*stride, cov_vec);
        else
            compute_covariations(activity_batch_f32+eidx*stride, cov_vec);
        float cov2 = fastmetric(cov_vec);
        if( *winner_idx<0 || cov2>*winner_cov )
        {
            *winner_idx = eidx;
            *winner_cov = cov2;
        }
    }
}

void linear_evaluator::compute_covariations(const float *activity, float *cov_vec) const
{
    for(int64_t class_idx=0; class_idx<baseline.nclasses; class_idx++)
        cov_vec[class_idx] = cov11(&residuals_f32[class_idx*residuals_stride], activity, baseline.ntrn);
}

void linear_evaluator::compute_covariations(const int8_t *activity, float *cov_vec) const
{
    int64_t nclasses = baseline.nclasses, ntrn = baseline.ntrn;
    if( nclasses!=2 )
    {
        //
        // Generic code
        //
        for(int64_t chunk_idx=0; chunk_idx<nclasses/2; chunk_idx++)
        {
            cov21(&residuals_i8[(chunk_idx*2+0)*residuals_stride],
                  &residuals_i8[(chunk_idx*2+1)*residuals_stride],
                  activity, ntrn, cov_vec[chunk_idx*2+0], cov_vec[chunk_idx*2+1]);
        }
        if( nclasses%2!=0 )
            cov_vec[nclasses-1] = cov11(&residuals_i8[(nclasses-1)*residuals_stride], activity, ntrn);
    }
    else
    {
        //
        // Special code for 2-class problems
        //
        cov_vec[0] = cov11(&residuals_i8[0], activity, ntrn);
        cov_vec[1] = -cov_vec[0];
    }
}



float linear_evaluator::fastmetric(const float *cov_vec) const
{
    float cov2 = 0.0f;
    for(int j=0; j<baseline.nclasses; j++)
        cov2 += cov_vec[j]*cov_vec[j];
    return cov2;

}

void linear_evaluator::update_predictor_output_errors()
{
    const int64_t ntrn      = baseline.ntrn;
    const int64_t ntst      = baseline.ntst;
    const int64_t nclasses  = baseline.nclasses;
    alglib::real_1d_array probs;
    probs.setlength(nclasses);

    //
    // Compute coefficients of the linear model
    //
    alglib::real_2d_array coeffs = qb;
    alglib::rmatrixlefttrsm(nrows, nclasses, lq_trf, 0, 0, false, false, 1, coeffs, 0, 0);
    std::vector<float> coeffs32;
    for(int i=1; i<nrows; i++)
        for(int j=0; j<nclasses; j++)
            coeffs32.push_back(coeffs(i,j));
    for(int j=0; j<nclasses; j++)
        coeffs32.push_back(coeffs(0,j));
    predictor.update_coeffs(coeffs32, nrows-1);

    //
    // Compute predictions for training and test set elements
    //
    _trn_rms = 0;
    _tst_rms = 0;
    _tst_cls = 0;
    for(int64_t k=0; k<nclasses; k++)
        for(int64_t idx=0; idx<ntrn; idx++)
        {
            // prediction = target-residual
            float residual = residuals_f32[idx+k*residuals_stride];
            float target   = output.labels[idx]==k ? 1.0f : 0.0f;
            output.fpredictions[idx+k*output.stride] = target-residual;
            _trn_rms += residual*residual;
        }
    _trn_rms = sqrt(_trn_rms/(ntrn*nclasses));
    for(int64_t idx=0; idx<ntst; idx++)
    {
        for(int64_t k=0; k<nclasses; k++)
            probs(k) = baseline.fpredictions[(ntrn+idx)+k*baseline.stride];
        for(int64_t i=0; i<nrows; i++)
        {
            double va = test_set(i,idx);
            if( va!=0 )
            {
                for(int64_t k=0; k<nclasses; k++)
                    probs(k) += coeffs(i,k)*va;
            }
        }
        for(int64_t k=0; k<nclasses; k++)
            output.fpredictions[(ntrn+idx)+k*output.stride] = probs(k);
        int64_t mx = 0;
        int target = baseline.labels[ntrn+idx];
        for(int64_t k=0; k<nclasses; k++)
        {
            _tst_rms += pow(probs(k)-(k==target ? 1 : 0),2.0);
            if( probs(k)>probs(mx) )
                mx = k;
        }
        _tst_cls += mx!=target ? 1 : 0;
    }
    _tst_rms = sqrt(_tst_rms/(ntst*nclasses+1.0e-20f));
    _tst_cls = _tst_cls/(ntst+1.0e-20f);
}

void linear_evaluator::update_residuals(const float *p_q, const float *p_qr)
{
    const int64_t ntrn     = baseline.ntrn;
    const int64_t nclasses = baseline.nclasses;
    for(int64_t k=0; k<nclasses; k++)
    {
        //
        // Update f32 representation of residuals
        //
        float *p_res = &residuals_f32[k*residuals_stride];
        float v = p_qr==NULL ? fdot_simd(p_res, p_q, ncols) : p_qr[k];
        fadd_simd(p_res, p_q, ncols, -v);
        qb(nrows,k) = v;

        //
        // Update residuals_squared
        //
        residuals_squared[k] = fdot_simd(p_res, p_res, ncols);

        //
        // Update i8 copy
        //
        int8_t *res8 = &residuals_i8[k*residuals_stride];
        for(int64_t i=0; i<ntrn; i++)
            res8[i] = flt_to_int8(p_res[i]);
    }
}

void linear_evaluator::run_analysis(daa *p_result) const
{
    int64_t ntrn = baseline.ntrn;
    int64_t ntst = baseline.ntst;
    int64_t nclasses = baseline.nclasses;
    timer counter(pcnt_da);

    ASSERT(p_result->activity.size()==ntrn+ntst);
    ASSERT(p_result->qu.size()==ncols);
    ASSERT(p_result->rqu.size()==nclasses);
    ASSERT(p_result->coeffs.size()==nrows);

    //
    // First orthogonalize activity wrt Q factor
    //
    for(int64_t k=0; k<ntrn; k++)
        p_result->qu[k] = p_result->activity[k];
    for(int64_t k=ntrn; k<ncols; k++)
        p_result->qu[k] = 0.0f;
    orthogonalize_row(&p_result->qu[0], &p_result->coeffs[0]);
    p_result->qnrm = sqrtf(fdot_simd(&p_result->qu[0], &p_result->qu[0], ncols));

    //
    // For each row of residual matrix perform:
    // * dot product with orthogonalized activity
    // * subtraction and accumulation
    //
    p_result->deltar2 = 0.0f;
    for(int64_t k=0; k<nclasses; k++)
    {
        float v = fdot_simd(&residuals_f32[k*residuals_stride], &p_result->qu[0], ncols);
        float invnrm = 1.0f/(p_result->qnrm>0 ? p_result->qnrm : 1.0f);
        p_result->rqu[k] = v;
        p_result->deltar2 += v*v*invnrm*invnrm;
    }

    //
    // Finally, sign result
    //
    p_result->signature = signature;
    p_result->signed_by = this;
}

void linear_evaluator::orthogonalize_row(float *p_row32, float *p_coeffs) const
{
    float *p_buf = p_row32;
    for(int64_t i=0; i<nrows; i++)
    {
        const float *p_lq_row = &lq_ort32[i*lq_stride];
        float v = 0.0f;

        //
        // Depending on specific SIMD set being used, determine chunk size and count
        //
#ifdef SNET_AVX2
        const int64_t CHUNK_SIZE = 8;
        const int64_t nchunks = ncols/CHUNK_SIZE;
#else
        const int64_t CHUNK_SIZE = 4;
        const int64_t nchunks = ncols/CHUNK_SIZE;
#endif

        //
        // Dot product (row,Qi) - process leading chunks with SIMD.
        // The tail is processed with generic C
        //
#ifdef SNET_AVX2
        float vv[CHUNK_SIZE];
        __m256 mm_dot = _mm256_set1_ps(0.0f);
        for(int64_t j=0; j<nchunks; j++)
        {
            __m256 mm_v0 = _mm256_load_ps(   p_buf+CHUNK_SIZE*j);
            __m256 mm_v1 = _mm256_load_ps(p_lq_row+CHUNK_SIZE*j);
            mm_dot = _mm256_fmadd_ps(mm_v0,mm_v1,mm_dot);
        }
        _mm256_storeu_ps(vv, mm_dot);
        for(int64_t j=0; j<CHUNK_SIZE; j++)
            v += vv[j];
#else
        __m128 mm_dot = _mm_set1_ps(0.0f);
        for(int64_t j=0; j<nchunks; j++)
        {
            __m128 mm_v0 = _mm_load_ps(   p_buf+CHUNK_SIZE*j);
            __m128 mm_v1 = _mm_load_ps(p_lq_row+CHUNK_SIZE*j);
            mm_dot = _mm_add_ps(_mm_mul_ps(mm_v0,mm_v1),mm_dot);
        }
        mm_dot = _mm_hadd_ps(mm_dot,mm_dot);
        mm_dot = _mm_hadd_ps(mm_dot,mm_dot);
        _mm_store_ss(&v, mm_dot);
#endif
        for(int64_t j=nchunks*CHUNK_SIZE; j<ncols; j++)
            v += p_buf[j]*p_lq_row[j];

        //
        // Store coefficients if needed
        //
        if( p_coeffs!=NULL )
            p_coeffs[i] = v;

        //
        // Subtraction row -= V*Qi - process leading chunks with SIMD,
        // the tail is processed with generic C
        //
#ifdef SNET_AVX2
        __m256 mm_v = _mm256_set1_ps(-v);
        for(int64_t j=0; j<nchunks; j++)
        {
            float *pp = p_buf+CHUNK_SIZE*j;
            __m256 mm_v1 = _mm256_load_ps(p_lq_row+CHUNK_SIZE*j);
            _mm256_store_ps(pp, _mm256_fmadd_ps(mm_v1, mm_v, _mm256_load_ps(pp)));
        }
#else
        __m128 mm_v = _mm_set1_ps(v);
        for(int64_t j=0; j<nchunks; j++)
        {
            float *pp = p_buf+CHUNK_SIZE*j;
            __m128 mm_v1 = _mm_load_ps(p_lq_row+CHUNK_SIZE*j);
            _mm_store_ps(pp, _mm_sub_ps(_mm_load_ps(pp), _mm_mul_ps(mm_v1,mm_v)));
        }
#endif
        for(int64_t j=nchunks*CHUNK_SIZE; j<ncols; j++)
            p_buf[j] -= v*p_lq_row[j];
    }
}

void linear_evaluator::orthonormalize_row(float *p_row32, float *nrm, float *p_coeffs) const
{
    float *p_buf = p_row32;
    orthogonalize_row(p_row32, p_coeffs);
    ASSERT(nrm!=NULL);
    float vs = sqrtf(fdot_simd(p_buf, p_buf, ncols));
    *nrm = vs;
    vs = 1.0f/(vs>0 ? vs : 1.0f);
    for(int64_t k=0; k<ncols; k++)
        p_buf[k] *= vs;
}

void linear_evaluator::append_activity(const activity8_vector &activity)
{
    ASSERT(activity.size()>0);
    append_activity(NULL, &activity[0], activity.size());
}

void linear_evaluator::append_activity(const activityf_vector &activity)
{
    ASSERT(activity.size()>0);
    append_activity(&activity[0], NULL, activity.size());
}

void linear_evaluator::append_and_refresh(const daa &append, const std::vector<daa*> &refresh)
{
    const int64_t ntrn = baseline.ntrn;
    const int64_t ntst = baseline.ntst;
    const int64_t nclasses = baseline.nclasses;
    timer t(pcnt_ar);

    // integrity checks
    ASSERT(nrows<max_rows);
    ASSERT(append.signed_by==this && append.signature==signature);
    ASSERT(append.qu.size()==ncols);
    ASSERT(append.rqu.size()==nclasses);
    ASSERT(append.activity.size()==ntrn+ntst);
    for(daa *p: refresh)
    {
        ASSERT(p->signed_by==this && p->signature==signature);
        ASSERT(p->qu.size()==ncols);
        ASSERT(p->rqu.size()==nclasses);
        ASSERT(p->activity.size()==ntrn+ntst);
    }

    // append row to Q, normalize
    float *new_qrow = &lq_ort32[nrows*lq_stride];
    for(int64_t i=0; i<ncols; i++)
        new_qrow[i] = append.qu[i];
    new_qrow[ntrn+nrows] = lambdav;
    float nrm = sqrtf(fdot_simd(new_qrow,new_qrow,ncols));
    ASSERT(nrm>0.0f);
    for(int64_t k=0; k<nrows; k++)
        lq_trf[nrows][k] = append.coeffs[k];
    lq_trf[nrows][nrows] = nrm;
    float invnrm = 1.0f/nrm;
    for(int64_t i=0; i<ncols; i++)
        new_qrow[i] *= invnrm;

    // compute products of normalized q_new with residuals, stored them to rq, update residuals
    ASSERT(nclasses<1024);
    float rq[nclasses];
    for(int k=0; k<nclasses; k++)
        rq[k] = fdot_simd(&residuals_f32[k*residuals_stride], new_qrow, ncols);
    update_residuals(new_qrow, rq);

    // append to test set
    for(int64_t i=0; i<ntst; i++)
        test_set(nrows,i) = append.activity[ntrn+i];

    // increase rows count, signature, update error metrics
    nrows++;
    signature++;
    update_predictor_output_errors();

    // update elements to be refreshed
    for(daa *_elem: refresh)
    {
        daa &elem = *_elem;
        const float *p_qa = new_qrow;
        float       *p_qb = &elem.qu[0];

        //
        // compute product of qa (row being appended) and qb (row being updated), store to coeffs
        //
        float qaqb = fdot_simd(p_qa, p_qb, ncols);
        elem.coeffs.push_back(qaqb);

        //
        // reorthogonalize qb, update qnrm
        //
        fadd_simd(p_qb, p_qa, ncols, -qaqb);
        elem.qnrm = sqrtf(fdot_simd(p_qb,p_qb,ncols));

        //
        // update rq elements of ELEM and deltar2.
        //
        // We use following formula:
        //
        // (r , q_new) = (r , q_old-(q_new,qa)*qa) = (r,q_old) - (q_new,qa)*(r,qa)
        //
        float invnrm = 1.0f/(elem.qnrm>0 ? elem.qnrm : 1.0f);
        elem.deltar2 = 0.0f;
        for(int64_t k=0; k<nclasses; k++)
        {
            elem.rqu[k] -= qaqb*rq[k];
            elem.deltar2 += elem.rqu[k]*elem.rqu[k]*invnrm*invnrm;
        }

        //
        // Sign elem again
        //
        elem.signature = signature;
    }
}

/*void linear_evaluator::_dbg_print()
{
    for(int64_t k=0; k<baseline.nclasses; k++)
    {
        float v2 = fdot_simd(&residuals_f32[k*residuals_stride], &residuals_f32[k*residuals_stride], baseline.ntrn);
        printf("r[] = %8.3f\n", (double)v2);
    }
    fflush(stdout);
}*/

std::atomic<int32_t> seed_evaluator::pcnt_seedeval(0);

seed_evaluator::seed_evaluator(const activities &_ds, const linear_evaluator &_lin_eval):
    ds(_ds),lin_eval(_lin_eval)
{
    best_actf = activityf_vector(ds.ntrn+ds.ntst,0);
    eval_stride = iDivUp(ds.ntrn+ds.ntst,ALIGN_ACTIVITIES_BY)*ALIGN_ACTIVITIES_BY;
    eval_act8.resize(eval_stride, 0x0);
    eval_act32.resize(eval_stride, 0.0f);
    ASSERT(eval_stride==ds.get_stride());
    start_evaluation(and_type_t::AND_MIN);
}

seed_evaluator::~seed_evaluator()
{
}

void seed_evaluator::start_evaluation(and_type_t atype)
{
    ASSERT(atype==and_type_t::AND_MIN || atype==and_type_t::AND_MUL);

    int64_t ntotal = ds.ntrn+ds.ntst;
    eval_and_type = atype;
    best_cov2 = 0.0f;
    best_seed = seed_detector();
    for(int64_t i=0; i<ntotal; i++)
        best_actf[i] = 0;
}

void seed_evaluator::evaluate_seed(const seed_detector &seed)
{
    timer counter(pcnt_seedeval);
    const int64_t CHUNK_SIZE = ds.is_i8() ? 32 : 8;

    //
    // Integrity checks
    //
    ASSERT(seed.and_type==eval_and_type);
    ASSERT((ds.get_stride()%CHUNK_SIZE)==0);
    ASSERT(seed.channel_idx0>=0);

    //
    // Dataset metrics
    //
    const int64_t nchunks = ds.get_stride()/CHUNK_SIZE;
    const int inputs_cnt = seed.inputs_cnt();
    const int64_t res_stride = lin_eval.get_res_stride();
    const int64_t nres = lin_eval.get_nclasses();
#ifndef SNET_AVX2
#error AVX2 support is required (SSE4 code is not implemented yet)
#endif

    //
    // Compute squared covariation
    //
    float cov2;
    if( ds.is_i8() && eval_and_type==and_type_t::AND_MIN )
    {
        //
        // int8_t code
        //
        const int8_t *p_res8  = lin_eval.get_res8();
        const int8_t *v8_act0 = ds.get_i8(seed.channel_idx0);
        const int8_t *v8_act1 = seed.channel_idx1>=0 ? ds.get_i8(seed.channel_idx1) : NULL;
        const int8_t *v8_act2 = seed.channel_idx2>=0 ? ds.get_i8(seed.channel_idx2) : NULL;
        const int8_t *v8_act3 = seed.channel_idx3>=0 ? ds.get_i8(seed.channel_idx3) : NULL;
        const int8_t *v8_act4 = seed.channel_idx4>=0 ? ds.get_i8(seed.channel_idx4) : NULL;
        int8_t *p_dst = &eval_act8[0];
        for(int64_t i=0; i<nchunks; i++)
        {
            // compute detector activity with unit gain
            __m256i v_act_g1 = _mm256_load_si256((__m256i*)(v8_act0+CHUNK_SIZE*i));
            if( inputs_cnt>=2 )
            {
                v_act_g1 = _mm256_min_epi8(v_act_g1,            _mm256_load_si256((__m256i*)(v8_act1+CHUNK_SIZE*i)));
                if( inputs_cnt>=3 )
                {
                    v_act_g1 = _mm256_min_epi8(v_act_g1,        _mm256_load_si256((__m256i*)(v8_act2+CHUNK_SIZE*i)));
                    if( inputs_cnt>=4 )
                    {
                        v_act_g1 = _mm256_min_epi8(v_act_g1,    _mm256_load_si256((__m256i*)(v8_act3+CHUNK_SIZE*i)));
                        if( inputs_cnt>=5 )
                            v_act_g1 = _mm256_min_epi8(v_act_g1,_mm256_load_si256((__m256i*)(v8_act4+CHUNK_SIZE*i)));
                    }
                }
            }
            _mm256_store_si256((__m256i*)(p_dst+CHUNK_SIZE*i), v_act_g1);
        }
        if( int(seed.ggain)>1 )
            apply_gain(p_dst, eval_stride, int(seed.ggain));
        cov2 = 0.0f;
        for(int64_t k=0; k<nres; k++)
        {
            float v = cov11(p_res8+k*res_stride, p_dst, ds.ntrn);
            cov2 += v*v;
        }
    }
    else if( ds.is_f32() && eval_and_type==and_type_t::AND_MUL )
    {
        //
        // fp32 code
        //
        ASSERT(ds.is_f32() && eval_and_type==and_type_t::AND_MUL);
        const float *p_res32 = lin_eval.get_res32();
        const float *v32_act0 = ds.get_f32(seed.channel_idx0);
        const float *v32_act1 = seed.channel_idx1>=0 ? ds.get_f32(seed.channel_idx1) : NULL;
        const float *v32_act2 = seed.channel_idx2>=0 ? ds.get_f32(seed.channel_idx2) : NULL;
        const float *v32_act3 = seed.channel_idx3>=0 ? ds.get_f32(seed.channel_idx3) : NULL;
        const float *v32_act4 = seed.channel_idx4>=0 ? ds.get_f32(seed.channel_idx4) : NULL;
        float *p_dst = &eval_act32[0];
        for(int64_t i=0; i<nchunks; i++)
        {
            // compute detector activity with unit gain
            __m256 v_act_g1 = _mm256_load_ps(v32_act0+CHUNK_SIZE*i);
            if( inputs_cnt>=2 )
            {
                v_act_g1 = _mm256_mul_ps(v_act_g1,            _mm256_load_ps(v32_act1+CHUNK_SIZE*i));
                if( inputs_cnt>=3 )
                {
                    v_act_g1 = _mm256_mul_ps(v_act_g1,        _mm256_load_ps(v32_act2+CHUNK_SIZE*i));
                    if( inputs_cnt>=4 )
                    {
                        v_act_g1 = _mm256_mul_ps(v_act_g1,    _mm256_load_ps(v32_act3+CHUNK_SIZE*i));
                        if( inputs_cnt>=5 )
                            v_act_g1 = _mm256_mul_ps(v_act_g1,_mm256_load_ps(v32_act4+CHUNK_SIZE*i));
                    }
                }
            }
            _mm256_store_ps(p_dst+CHUNK_SIZE*i, v_act_g1);
        }
        if( int(seed.ggain)>1 )
            apply_gain(p_dst, eval_stride, int(seed.ggain));
        cov2 = 0.0f;
        for(int64_t k=0; k<nres; k++)
        {
            float v = cov11(p_res32+k*res_stride, p_dst, ds.ntrn);
            cov2 += v*v;
        }
    }
    else if( ds.is_f32() && eval_and_type==and_type_t::AND_MIN )
    {
        //
        // fp32 code
        //
        const float *p_res32 = lin_eval.get_res32();
        const float *v32_act0 = ds.get_f32(seed.channel_idx0);
        const float *v32_act1 = seed.channel_idx1>=0 ? ds.get_f32(seed.channel_idx1) : NULL;
        const float *v32_act2 = seed.channel_idx2>=0 ? ds.get_f32(seed.channel_idx2) : NULL;
        const float *v32_act3 = seed.channel_idx3>=0 ? ds.get_f32(seed.channel_idx3) : NULL;
        const float *v32_act4 = seed.channel_idx4>=0 ? ds.get_f32(seed.channel_idx4) : NULL;
        float *p_dst = &eval_act32[0];
        for(int64_t i=0; i<nchunks; i++)
        {
            // compute detector activity with unit gain
            __m256 v_act_g1 = _mm256_load_ps(v32_act0+CHUNK_SIZE*i);
            if( inputs_cnt>=2 )
            {
                v_act_g1 = _mm256_min_ps(v_act_g1,            _mm256_load_ps(v32_act1+CHUNK_SIZE*i));
                if( inputs_cnt>=3 )
                {
                    v_act_g1 = _mm256_min_ps(v_act_g1,        _mm256_load_ps(v32_act2+CHUNK_SIZE*i));
                    if( inputs_cnt>=4 )
                    {
                        v_act_g1 = _mm256_min_ps(v_act_g1,    _mm256_load_ps(v32_act3+CHUNK_SIZE*i));
                        if( inputs_cnt>=5 )
                            v_act_g1 = _mm256_min_ps(v_act_g1,_mm256_load_ps(v32_act4+CHUNK_SIZE*i));
                    }
                }
            }
            _mm256_store_ps(p_dst+CHUNK_SIZE*i, v_act_g1);
        }
        if( int(seed.ggain)>1 )
            apply_gain(p_dst, eval_stride, int(seed.ggain));
        cov2 = 0.0f;
        for(int64_t k=0; k<nres; k++)
        {
            float v = cov11(p_res32+k*res_stride, p_dst, ds.ntrn);
            cov2 += v*v;
        }
    }
    else
        ASSERT(0);

#if 0
    //
    // One code for multiclass problems, another one (better optimized) for two-class ones
    //
    int8_t *p_dst = &eval_act8[0];
    float cov2;
    if( nres>2 )
    {
        //
        // Compute activity value with unit gain
        //
        for(int64_t i=0; i<nchunks; i++)
        {
            // compute detector activity with unit gain
            __m256i v_act_g1 = _mm256_load_si256((__m256i*)(v8_act0+CHUNK_SIZE*i));
            if( inputs_cnt>=2 )
            {
                v_act_g1 = _mm256_min_epi8(v_act_g1,            _mm256_load_si256((__m256i*)(v8_act1+CHUNK_SIZE*i)));
                if( inputs_cnt>=3 )
                {
                    v_act_g1 = _mm256_min_epi8(v_act_g1,        _mm256_load_si256((__m256i*)(v8_act2+CHUNK_SIZE*i)));
                    if( inputs_cnt>=4 )
                    {
                        v_act_g1 = _mm256_min_epi8(v_act_g1,    _mm256_load_si256((__m256i*)(v8_act3+CHUNK_SIZE*i)));
                        if( inputs_cnt>=5 )
                            v_act_g1 = _mm256_min_epi8(v_act_g1,_mm256_load_si256((__m256i*)(v8_act4+CHUNK_SIZE*i)));
                    }
                }
            }
            _mm256_store_si256((__m256i*)(p_dst+CHUNK_SIZE*i), v_act_g1);
        }

        //
        // Apply gain if needed
        //
        if( int(seed.ggain)>1 )
            apply_gain(p_dst, eval_stride, int(seed.ggain));

        //
        // Compute squared covariations
        // One code for multiclass problem, more optimized one for 1-residual
        //
        cov2 = 0.0f;
        for(int64_t k=0; k<nres; k++)
        {
            float v = cov11(p_res8+k*res_stride, p_dst, ds.ntrn);
            cov2 += v*v;
        }
    }
    else
    {
        ASSERT(nres==2);

        //
        // Activity computation, gain and residual covariation are deeply integrated
        //
        ASSERT(int(seed.ggain)>=1 && int(seed.ggain)<=8);
        ASSERT(SNET_IMAX/int(seed.ggain)+1<=127);
        __m256i mm_cap0   = _mm256_set1_epi8(SNET_IMAX/int(seed.ggain)+1);
        __m256i mm_cap1   = _mm256_set1_epi8(SNET_IMAX);
        __m256i mm_gain16 = _mm256_set1_epi16(int(seed.ggain));
        __m256  mm_cov    = _mm256_set1_ps(0.0f);
        bool gain_needed = int(seed.ggain)>1;
        const int64_t ntrn_chunks = ds.ntrn/CHUNK_SIZE;
        for(int64_t i=0; i<nchunks; i++)
        {
            // compute detector activity with unit gain
            __m256i mm_act = _mm256_load_si256((__m256i*)(v8_act0+CHUNK_SIZE*i));
            if( inputs_cnt>=2 )
            {
                mm_act = _mm256_min_epi8(mm_act,            _mm256_load_si256((__m256i*)(v8_act1+CHUNK_SIZE*i)));
                if( inputs_cnt>=3 )
                {
                    mm_act = _mm256_min_epi8(mm_act,        _mm256_load_si256((__m256i*)(v8_act2+CHUNK_SIZE*i)));
                    if( inputs_cnt>=4 )
                    {
                        mm_act = _mm256_min_epi8(mm_act,    _mm256_load_si256((__m256i*)(v8_act3+CHUNK_SIZE*i)));
                        if( inputs_cnt>=5 )
                            mm_act = _mm256_min_epi8(mm_act,_mm256_load_si256((__m256i*)(v8_act4+CHUNK_SIZE*i)));
                    }
                }
            }

            // apply gain if needed
            if( gain_needed )
            {
                mm_act = _mm256_min_epi8(mm_act, mm_cap0);
                mm_act = _mm256_mullo_epi16(mm_act, mm_gain16);
                mm_act = _mm256_min_epi8(mm_act, mm_cap1);
            }

            // compute trn covariation if needed
            if( i<ntrn_chunks )
            {
                const int8_t *p_res8_chunk  = p_res8+CHUNK_SIZE*i;
                const float  *p_res32_chunk = p_res32+CHUNK_SIZE*i;
                alignas(32) int8_t v32[32];
                __m256 mm_res_f32, mm_act_f32;

                _mm256_store_si256((__m256i*)v32, mm_act);
                for(int j=0; j<4; j++)
                {
                    //mm_res_f32 = _mm256_load_ps(p_res32_chunk+2*8);// TODO: handle SNET_I2F below in case you load floats
                    mm_res_f32 = _mm256_cvtepi32_ps(_mm256_cvtepi8_epi32(_mm_set1_epi64x(((const int64_t*)p_res8_chunk)[j])));
                    mm_act_f32 = _mm256_cvtepi32_ps(_mm256_cvtepi8_epi32(_mm_set1_epi64x(((const int64_t*)v32)[j])));
                    mm_cov     = _mm256_fmadd_ps(mm_res_f32,mm_act_f32,mm_cov);
                }
            }

            // store activity
            _mm256_store_si256((__m256i*)(p_dst+CHUNK_SIZE*i), mm_act);
        }

        //
        // Compute squared covariations
        // One code for multiclass problem, more optimized one for 1-residual
        //
        //float v = cov11(p_res8, p_dst, ds.ntrn); cov2 = 2*v*v;
        float vv[8];
        float v = 0;
        mm_cov = _mm256_mul_ps(mm_cov, _mm256_set1_ps(SNET_I2F*SNET_I2F));
        _mm256_storeu_ps(vv, mm_cov);
        for(int k=0; k<8; k++)
            v += vv[k];
        cov2 = 2*v*v;
    }
#endif

    //
    // Compare with best
    //
    if( cov2>best_cov2 )
    {
        best_cov2 = cov2;
        best_seed = seed;
        if( ds.is_i8() )
        {
            for(int64_t k=0; k<ds.ntrn+ds.ntst; k++)
                best_actf[k] = SNET_I2F*eval_act8[k];
        }
        else if( ds.is_f32() )
        {
            for(int64_t k=0; k<ds.ntrn+ds.ntst; k++)
                best_actf[k] = eval_act32[k];
        }
        else
            ASSERT(0);
    }
}

void seed_evaluator::get_winner(float *cov, seed_detector *seed, activityf_vector *activity_f32, activity8_vector *activity_i8)
{
    // offload results
    int64_t ntotal = ds.ntrn+ds.ntst;
    *cov  = best_cov2;
    *seed = best_seed;
    if( activity_i8!=NULL )
    {
        activity_i8->resize(ntotal);
        for(int64_t k=0; k<ntotal; k++)
            (*activity_i8)[k] = flt_to_int8(best_actf[k]);
    }
    if( activity_f32!=NULL )
    {
        activity_f32->resize(ntotal);
        for(int64_t k=0; k<ntotal; k++)
            (*activity_f32)[k] = best_actf[k];
    }

}

void seed_evaluator::get_winner(float *cov, seed_detector *seed, activity8_vector *activity)
{
    get_winner(cov, seed, NULL, activity);
}

void seed_evaluator::get_winner(float *cov, seed_detector *seed, activityf_vector *activity)
{
    get_winner(cov, seed, activity, NULL);
}

std::atomic<int32_t> generalizer_evaluator::pcnt_gen(0);

generalizer_evaluator::generalizer_evaluator(const activities &_ds, const linear_evaluator &_lin_eval):
    ds(_ds), lin_eval(_lin_eval)
{
    max_queue_size = 1024;
    best_act_i8    = activity8_vector(ds.ntrn+ds.ntst,0);
    best_act_f32   = activityf_vector(ds.ntrn+ds.ntst,0);
    base_act_a_i8  = activity8_vector(ds.get_stride(),0);
    base_act_g_i8  = activity8_vector(ds.get_stride(),0);
    base_act_a_f32 = activityf_vector(ds.get_stride(),0);
    base_act_g_f32 = activityf_vector(ds.get_stride(),0);
    tmpa0_i8       = activity8_vector(ds.get_stride(),0);
    tmpa1_i8       = activity8_vector(ds.get_stride(),0);
    tmpa2_i8       = activity8_vector(ds.get_stride(),0);
    tmpa3_i8       = activity8_vector(ds.get_stride(),0);
    tmpa4_i8       = activity8_vector(ds.get_stride(),0);
    tmpa0_f32      = activityf_vector(ds.get_stride(),0);
    tmpa1_f32      = activityf_vector(ds.get_stride(),0);
    tmpa2_f32      = activityf_vector(ds.get_stride(),0);
    tmpa3_f32      = activityf_vector(ds.get_stride(),0);
    tmpa4_f32      = activityf_vector(ds.get_stride(),0);
    if( ds.is_i8() )
        queue_cand_act_i8  = activity8_vector(ds.get_stride()*max_queue_size,0);
    if( ds.is_f32() )
        queue_cand_act_f32 = activityf_vector(ds.get_stride()*max_queue_size,0);
    best_cov = 0.0f;
}

void generalizer_process_queue_item_adapter(void *p, int idx)
{
    generalizer_evaluator &eval = *((generalizer_evaluator*)p);
    eval.process_queue_item(idx);
}

void generalizer_evaluator::process_queue()
{
    timer tmr(pcnt_gen);
    //for(int entry_idx=0; entry_idx<eval_queue.size(); entry_idx++)
    //    process_queue_item(entry_idx);
    parallel_for(generalizer_process_queue_item_adapter, this, eval_queue.size());
    int best_idx = -1;
    for(int k=0; k<eval_queue.size(); k++)
        if( best_idx<0 || (eval_queue[k].cov2>eval_queue[best_idx].cov2 && eval_queue[k].cov2>0) )
            best_idx = k;
    if( best_idx>=0 && eval_queue[best_idx].cov2>best_cov )
    {
        best_cov  = eval_queue[best_idx].cov2;
        best_detector = base_detector;
        best_detector.add_conn(input_studied, detector::conn(eval_queue[best_idx].merge_idx,detector::WEIGHT_100, ds));
        if( ds.is_i8() )
            best_detector.generate_activity_slow(ds, best_act_i8);
        else if( ds.is_f32() )
            best_detector.generate_activity_slow(ds, best_act_f32);
        else
            ASSERT(0);
    }
    eval_queue.clear();
}

void generalizer_evaluator::process_queue_item(int item_idx)
{
#ifndef SNET_AVX2
#error generalizer_evaluator(): AVX2 is required!
#endif
    const int64_t CHUNK_SIZE = ds.is_i8() ? 32 : 8;
    const int64_t nchunks = ds.get_stride()/CHUNK_SIZE;
    queue_entry &entry = eval_queue[item_idx];
    ASSERT(ds.get_stride()%CHUNK_SIZE==0);
    ASSERT((ds.is_i8() && and_type==and_type_t::AND_MIN)||(ds.is_f32() && and_type==and_type_t::AND_MUL));
    ASSERT(or_type==or_type_t::OR_MAX);
    ASSERT(base_detector.gain>=1 && base_detector.gain<=8);
    ASSERT(SNET_IMAX/base_detector.gain+1<=127);

    if( ds.is_i8() )
    {
        //
        // int8_t implementation
        //
        ASSERT(and_type==and_type_t::AND_MIN);
        const int gain = base_detector.gain;
        const int8_t *p_a = &base_act_a_i8[0];
        const int8_t *p_g = &base_act_g_i8[0];
        const int8_t *p_m = ds.get_i8(entry.merge_idx);
        int8_t *p_cand_act = &queue_cand_act_i8[item_idx*ds.get_stride()];
        __m256i mm_cap0   = _mm256_set1_epi8(SNET_IMAX/gain+1);
        __m256i mm_cap1   = _mm256_set1_epi8(SNET_IMAX);
        __m256i mm_gain16 = _mm256_set1_epi16(gain);
        for(int64_t i=0; i<nchunks; i++)
        {
            __m256i va = _mm256_load_si256((__m256i*)(p_a+CHUNK_SIZE*i));
            __m256i vg = _mm256_load_si256((__m256i*)(p_g+CHUNK_SIZE*i));
            __m256i vm = _mm256_load_si256((__m256i*)(p_m+CHUNK_SIZE*i));
            if( gain>1 )
            {
                vm = _mm256_min_epi8(vm, mm_cap0);
                vm = _mm256_mullo_epi16(vm, mm_gain16);
                vm = _mm256_min_epi8(vm, mm_cap1);
            }
            __m256i vr = _mm256_min_epi8(va,_mm256_max_epi8(vg,vm));
            _mm256_store_si256((__m256i*)(p_cand_act+CHUNK_SIZE*i), vr);
        }
        entry.cov2 = lin_eval.estimate_trn_covariation(p_cand_act);
    }
    else
    {
        //
        // fp32 implementation
        //
        ASSERT(ds.is_f32());
        ASSERT(and_type==and_type_t::AND_MUL);
        ASSERT(base_detector.gain==1);
        const float *p_a = &base_act_a_f32[0];
        const float *p_g = &base_act_g_f32[0];
        const float *p_m = ds.get_f32(entry.merge_idx);
        float *p_cand_act = &queue_cand_act_f32[item_idx*ds.get_stride()];
        for(int64_t i=0; i<nchunks; i++)
        {
            __m256 va = _mm256_load_ps(p_a+CHUNK_SIZE*i);
            __m256 vg = _mm256_load_ps(p_g+CHUNK_SIZE*i);
            __m256 vm = _mm256_load_ps(p_m+CHUNK_SIZE*i);
            __m256 vr = _mm256_mul_ps(va,_mm256_max_ps(vg,vm));
            _mm256_store_ps(p_cand_act+CHUNK_SIZE*i, vr);
        }
        entry.cov2 = lin_eval.estimate_trn_covariation(p_cand_act);
    }

#if 0
    //
    // Depending on classes count, two code variants are used
    //
    const int nres = lin_eval.get_nclasses();
    const int gain = base_detector.gain;
    const int8_t *p_a = &base_act_a_i8[0];
    const int8_t *p_g = &base_act_g_i8[0];
    const int8_t *p_m = ds.get_i8(entry.merge_idx);
    int8_t *p_cand_act = &queue_cand_act_i8[item_idx*ds.get_stride()];
    __m256i mm_cap0   = _mm256_set1_epi8(SNET_IMAX/gain+1);
    __m256i mm_cap1   = _mm256_set1_epi8(SNET_IMAX);
    __m256i mm_gain16 = _mm256_set1_epi16(gain);
    if( nres>2 )
    {
        //
        // Generate activity, store to buffer, compute covariation using linear_evaluator
        //
        for(int64_t i=0; i<nchunks; i++)
        {
            __m256i va = _mm256_load_si256((__m256i*)(p_a+CHUNK_SIZE*i));
            __m256i vg = _mm256_load_si256((__m256i*)(p_g+CHUNK_SIZE*i));
            __m256i vm = _mm256_load_si256((__m256i*)(p_m+CHUNK_SIZE*i));
            if( gain>1 )
            {
                vm = _mm256_min_epi8(vm, mm_cap0);
                vm = _mm256_mullo_epi16(vm, mm_gain16);
                vm = _mm256_min_epi8(vm, mm_cap1);
            }
            __m256i vr = _mm256_min_epi8(va,_mm256_max_epi8(vg,vm));
            _mm256_store_si256((__m256i*)(p_cand_act+CHUNK_SIZE*i), vr);
        }
        entry.cov2 = lin_eval.estimate_trn_covariation(p_cand_act);
    }
    else
    {
        ASSERT(nres==2);

        //
        // 2-class problem, compute activity and covariation in-place
        //
        const int8_t *p_res8  = lin_eval.get_res8();
        const int64_t ntrn_chunks = ds.ntrn/CHUNK_SIZE;
        __m256  mm_cov    = _mm256_set1_ps(0.0f);
        for(int64_t i=0; i<ntrn_chunks; i++)
        {
            //
            // Compute activity
            //
            __m256i va = _mm256_load_si256((__m256i*)(p_a+CHUNK_SIZE*i));
            __m256i vg = _mm256_load_si256((__m256i*)(p_g+CHUNK_SIZE*i));
            __m256i vm = _mm256_load_si256((__m256i*)(p_m+CHUNK_SIZE*i));
            if( gain>1 )
            {
                vm = _mm256_min_epi8(vm, mm_cap0);
                vm = _mm256_mullo_epi16(vm, mm_gain16);
                vm = _mm256_min_epi8(vm, mm_cap1);
            }
            __m256i mm_act = _mm256_min_epi8(va,_mm256_max_epi8(vg,vm));

            //
            // Compute covariation
            //
            const int8_t *p_res8_chunk  = p_res8+CHUNK_SIZE*i;
            alignas(32) int8_t v32[32];
            __m256 mm_res_f32, mm_act_f32;
            _mm256_store_si256((__m256i*)v32, mm_act);
            for(int j=0; j<4; j++)
            {
                mm_res_f32 = _mm256_cvtepi32_ps(_mm256_cvtepi8_epi32(_mm_set1_epi64x(((const int64_t*)p_res8_chunk)[j])));
                mm_act_f32 = _mm256_cvtepi32_ps(_mm256_cvtepi8_epi32(_mm_set1_epi64x(((const int64_t*)v32)[j])));
                mm_cov     = _mm256_fmadd_ps(mm_res_f32,mm_act_f32,mm_cov);
            }
        }
        float vv[8];
        float v = 0;
        mm_cov = _mm256_mul_ps(mm_cov, _mm256_set1_ps(SNET_I2F*SNET_I2F));
        _mm256_storeu_ps(vv, mm_cov);
        for(int k=0; k<8; k++)
            v += vv[k];
        entry.cov2 = 2*v*v;
    }
#endif
}

void generalizer_evaluator::start_evaluation(const detector &d, int input_idx, and_type_t andtype, or_type_t ortype)
{
    // integrity checks
    ASSERT(d.input0.size()>0);
    ASSERT(d.input2.size()==0 || d.input1.size()>0);
    ASSERT(d.input3.size()==0 || d.input2.size()>0);
    ASSERT(d.input4.size()==0 || d.input3.size()>0);
    ASSERT(input_idx>=0 && input_idx<5);
    if( input_idx==1 )
        ASSERT(d.input1.size()>0);
    if( input_idx==2 )
        ASSERT(d.input2.size()>0);
    if( input_idx==3 )
        ASSERT(d.input3.size()>0);
    if( input_idx==4 )
        ASSERT(d.input4.size()>0);
    ASSERT(d.transform.size()==1 && d.transform[0].shift0==0 && d.transform[0].shift1==0);
    ASSERT(d.and_type==andtype);
    ASSERT(d.or_type==ortype);

    // clear
    eval_queue.clear();

    // store evaluation settings
    and_type = andtype;
    or_type  = ortype;
    input_studied = input_idx;

    // baseline
    int64_t stride = ds.get_stride();
    base_detector = d;
    if( ds.is_i8() )
    {
        ASSERT(and_type==and_type_t::AND_MIN);
        d.generate_inputs5(ds, &tmpa0_i8[0], &tmpa1_i8[0], &tmpa2_i8[0], &tmpa3_i8[0], &tmpa4_i8[0]);
        for(int64_t i=0; i<stride; i++)
        {
            if( input_idx==0 )
            {
                base_act_a_i8[i] = imin(imin(tmpa1_i8[i],tmpa2_i8[i]),imin(tmpa3_i8[i],tmpa4_i8[i]));
                base_act_g_i8[i] = tmpa0_i8[i];
            }
            if( input_idx==1 )
            {
                base_act_a_i8[i] = imin(imin(tmpa0_i8[i],tmpa2_i8[i]),imin(tmpa3_i8[i],tmpa4_i8[i]));
                base_act_g_i8[i] = tmpa1_i8[i];
            }
            if( input_idx==2 )
            {
                base_act_a_i8[i] = imin(imin(tmpa0_i8[i],tmpa1_i8[i]),imin(tmpa3_i8[i],tmpa4_i8[i]));
                base_act_g_i8[i] = tmpa2_i8[i];
            }
            if( input_idx==3 )
            {
                base_act_a_i8[i] = imin(imin(tmpa0_i8[i],tmpa1_i8[i]),imin(tmpa2_i8[i],tmpa4_i8[i]));
                base_act_g_i8[i] = tmpa3_i8[i];
            }
            if( input_idx==4 )
            {
                base_act_a_i8[i] = imin(imin(tmpa0_i8[i],tmpa1_i8[i]),imin(tmpa2_i8[i],tmpa3_i8[i]));
                base_act_g_i8[i] = tmpa4_i8[i];
            }
        }
    }
    else
    {
        ASSERT(ds.is_f32());
        ASSERT(and_type==and_type_t::AND_MUL);
        d.generate_inputs5(ds, &tmpa0_f32[0], &tmpa1_f32[0], &tmpa2_f32[0], &tmpa3_f32[0], &tmpa4_f32[0]);
        for(int64_t i=0; i<stride; i++)
        {
            if( input_idx==0 )
            {
                base_act_a_f32[i] = tmpa1_f32[i]*tmpa2_f32[i]*tmpa3_f32[i]*tmpa4_f32[i];
                base_act_g_f32[i] = tmpa0_f32[i];
            }
            if( input_idx==1 )
            {
                base_act_a_f32[i] = tmpa0_f32[i]*tmpa2_f32[i]*tmpa3_f32[i]*tmpa4_f32[i];
                base_act_g_f32[i] = tmpa1_f32[i];
            }
            if( input_idx==2 )
            {
                base_act_a_f32[i] = tmpa0_f32[i]*tmpa1_f32[i]*tmpa3_f32[i]*tmpa4_f32[i];
                base_act_g_f32[i] = tmpa2_f32[i];
            }
            if( input_idx==3 )
            {
                base_act_a_f32[i] = tmpa0_f32[i]*tmpa1_f32[i]*tmpa2_f32[i]*tmpa4_f32[i];
                base_act_g_f32[i] = tmpa3_f32[i];
            }
            if( input_idx==4 )
            {
                base_act_a_f32[i] = tmpa0_f32[i]*tmpa1_f32[i]*tmpa2_f32[i]*tmpa3_f32[i];
                base_act_g_f32[i] = tmpa4_f32[i];
            }
        }
    }
#if 0
    ASSERT(ds.is_i8());
    ASSERT(and_type==and_type_t::AND_MIN);
    int64_t stride = ds.get_stride();
    base_detector = d;
    d.generate_inputs5(ds, &tmpa0_i8[0], &tmpa1_i8[0], &tmpa2_i8[0], &tmpa3_i8[0], &tmpa4_i8[0]);
    for(int64_t i=0; i<stride; i++)
    {
        if( input_idx==0 )
        {
            base_act_a_i8[i] = imin(imin(tmpa1_i8[i],tmpa2_i8[i]),imin(tmpa3_i8[i],tmpa4_i8[i]));
            base_act_g_i8[i] = tmpa0_i8[i];
        }
        if( input_idx==1 )
        {
            base_act_a_i8[i] = imin(imin(tmpa0_i8[i],tmpa2_i8[i]),imin(tmpa3_i8[i],tmpa4_i8[i]));
            base_act_g_i8[i] = tmpa1_i8[i];
        }
        if( input_idx==2 )
        {
            base_act_a_i8[i] = imin(imin(tmpa0_i8[i],tmpa1_i8[i]),imin(tmpa3_i8[i],tmpa4_i8[i]));
            base_act_g_i8[i] = tmpa2_i8[i];
        }
        if( input_idx==3 )
        {
            base_act_a_i8[i] = imin(imin(tmpa0_i8[i],tmpa1_i8[i]),imin(tmpa2_i8[i],tmpa4_i8[i]));
            base_act_g_i8[i] = tmpa3_i8[i];
        }
        if( input_idx==4 )
        {
            base_act_a_i8[i] = imin(imin(tmpa0_i8[i],tmpa1_i8[i]),imin(tmpa2_i8[i],tmpa3_i8[i]));
            base_act_g_i8[i] = tmpa4_i8[i];
        }
    }
#endif

    // results
    best_cov = 0;
    best_detector = detector();
    memset(&best_act_i8[0],  0, best_act_i8.size()*sizeof(best_act_i8[0]));
    memset(&best_act_f32[0], 0, best_act_f32.size()*sizeof(best_act_f32[0]));
}

void generalizer_evaluator::evaluate_generalization(int channel_idx)
{
    ASSERT(eval_queue.size()<=max_queue_size);
    if( eval_queue.size()==max_queue_size )
        process_queue();
    eval_queue.push_back(queue_entry(channel_idx));
}

void generalizer_evaluator::get_winner(float *cov, detector *winner, activity8_vector *activity)
{
    ASSERT(ds.is_i8());

    // finalize queue if needed
    if( eval_queue.size()>0 )
        process_queue();

    // offload results
    *cov      = best_cov;
    *winner   = best_detector;
    *activity = best_act_i8;
}

void generalizer_evaluator::get_winner(float *cov, detector *winner, activityf_vector *activity)
{
    ASSERT(ds.is_f32());

    // finalize queue if needed
    if( eval_queue.size()>0 )
        process_queue();

    // offload results
    *cov      = best_cov;
    *winner   = best_detector;
    *activity = best_act_f32;
}

float sigma0(float x)
{
    float mx = fmaxf(x,0);
    float ex = expf(x-mx), e0 = expf(-mx);
    return ex/(ex+e0);
}

int get_scale_idx_safe(int win_dim, int base_dim)
{
    ASSERT(base_dim>=1 && win_dim>=1 && win_dim<=base_dim);
    for(int scale_idx=0; win_dim<=base_dim; win_dim*=2, scale_idx++)
        if( win_dim==base_dim )
            return scale_idx;
    ASSERT(0);
}

std::vector<std::string> explode(const std::string &s, char sep)
{
    std::vector<std::string> r;
    std::string a;
    for(const char *p = s.c_str(); *p!=0; p++)
        if( *p==sep )
        {
            r.push_back(a);
            a.clear();
        }
        else
            a.append(p, 1);
    r.push_back(a);
    return r;
}

std::string trim(const std::string &s)
{
    const char *p = s.c_str();
    while(*p==' ' || *p=='\t' || *p=='\r' || *p=='\n' )
        p++;
    int len = strlen(p);
    while( len>0 && (p[len-1]==' ' || p[len-1]=='\t' || p[len-1]=='\r' || p[len-1]=='\n' ) )
        len--;
    return std::string(p, len);
}

std::vector<std::string> trim(const std::vector<std::string> &s)
{
    std::vector<std::string> r;
    for(const std::string &a: s)
        r.push_back(trim(a));
    return r;
}

bool file_exists(std::string filename)
{
    FILE *f = fopen(filename.c_str(),"rb");
    if( f!=NULL )
    {
        fclose(f);
        return true;
    }
    return false;
}

std::string file_get_contents(std::string filename)
{
    FILE *f = fopen(filename.c_str(),"rb");
    ASSERT(f!=NULL);

    fseek(f, 0, SEEK_END);
    int64_t filesize = ftell(f);
    fseek(f, 0, SEEK_SET);

    char *buf = (char*)malloc(filesize+1);
    ASSERT(buf!=NULL);
    unserialize_long_blob(f, buf, filesize);
    buf[filesize] = 0;
    std::string r = buf;
    free(buf);

    fclose(f);

    return r;
}

void file_put_contents(std::string filename, std::string data)
{
    FILE *f = fopen(filename.c_str(),"wb");
    ASSERT(f!=NULL);
    serialize_long_blob(f, data.c_str(), data.length());
    fclose(f);
}

void file_append_contents(std::string filename, std::string data)
{
    FILE *f = fopen(filename.c_str(),"ab");
    ASSERT(f!=NULL);
    serialize_long_blob(f, data.c_str(), data.length());
    fclose(f);
}

void serialize_long_blob(FILE *f, const void *blob, int64_t len)
{
    const char *p_blob = (const char*)blob;
    while(len>0)
    {
        int64_t chunk_size = imin64(1024*1024,len);
        size_t  r          = fwrite(p_blob, (size_t)chunk_size, 1, f);
        ASSERT(r==1);
        p_blob += chunk_size;
        len    -= chunk_size;
        fflush(f);
    }
}

void serialize64(FILE *f, int64_t v)
{
    fwrite(&v, sizeof(v), 1, f);
}

void write_magic8(FILE *f, int8_t magic)
{
    fwrite(&magic, 1, 1, f);
}

void serialize(FILE *f, const std::string *s)
{
    serialize64(f, s->size());
    serialize_long_blob(f, s->c_str(), s->size());
}

void serialize(FILE *f, const unsigned char *p_i)
{
    fwrite(p_i, sizeof(*p_i), 1, f);
}

void serialize(FILE *f, const int8_t *p_i)
{
    fwrite(p_i, sizeof(*p_i), 1, f);
}

void serialize(FILE *f, const int *p_i)
{
    fwrite(p_i, sizeof(*p_i), 1, f);
}

void serialize(FILE *f, const int64_t *p_i)
{
    fwrite(p_i, sizeof(*p_i), 1, f);
}

void serialize(FILE *f, const float *p_v)
{
    fwrite(p_v, sizeof(*p_v), 1, f);
}

void unserialize_long_blob(FILE *f, void *blob, int64_t len)
{
    char *p_blob = (char*)blob;
    while(len>0)
    {
        int64_t chunk_size = imin64(1024*1024,len);
        size_t  r          = fread(p_blob, (size_t)chunk_size, 1, f);
        ASSERT(r==1);
        p_blob += chunk_size;
        len    -= chunk_size;
    }
}

int64_t unserialize64(FILE *f)
{
    int64_t r;
    int t = fread(&r, sizeof(r), 1, f);
    ASSERT(t==1);
    return r;
}

int8_t check_magic8(FILE *f, int8_t magic)
{
    int8_t v;
    int t = fread(&v, 1, 1, f);
    ASSERT(t==1);
    ASSERT(v==magic);
    return v;
}

int8_t check_magic8(FILE *f, int8_t magic0, int8_t magic1)
{
    int8_t v;
    int t = fread(&v, 1, 1, f);
    ASSERT(t==1);
    ASSERT(v==magic0 || v==magic1);
    return v;
}

void unserialize(FILE *f, std::string *s)
{
    const int buf_len = 128;
    char buf[buf_len];

    memset(buf, 0, buf_len);
    int64_t slen = unserialize64(f);
    ASSERT(slen<buf_len-1);
    unserialize_long_blob(f, buf, slen);
    *s = buf;
}

void unserialize(FILE *f, unsigned char *p_i)
{
    int t = fread(p_i, sizeof(*p_i), 1, f);
    ASSERT(t==1);
}

void unserialize(FILE *f, int8_t *p_i)
{
    int t = fread(p_i, sizeof(*p_i), 1, f);
    ASSERT(t==1);
}

void unserialize(FILE *f, int *p_i)
{
    int t = fread(p_i, sizeof(*p_i), 1, f);
    ASSERT(t==1);
}

void unserialize(FILE *f, int64_t *p_i)
{
    int t = fread(p_i, sizeof(*p_i), 1, f);
    ASSERT(t==1);
}

void unserialize(FILE *f, float *p_v)
{
    int t = fread(p_v, sizeof(*p_v), 1, f);
    ASSERT(t==1);
}

thread_local int _parallel_flag = 0;

void parallelism_activate()
{
    _parallel_flag = 1;
}

int get_cores_count()
{
    return sysconf(_SC_NPROCESSORS_ONLN);
}

void parallel_invoke(void* (*func)(void*), void *param)
{
    //
    // Quick exit
    //
    // NOTE: checking _parallel_flag also prevents us from spawning new workers
    //       from already running worker threads.
    //
    if( _parallel_flag==0 )
    {
        func(param);
        return;
    }

    //
    // Try to process
    //
#if !defined(SNET_SERIAL)
    ASSERT(get_cores_count()>=2);
    int nthreads = get_cores_count();
    std::vector<pthread_t> threads(nthreads);
    for(int k=0; k<nthreads; k++)
    {
        int iret1 = pthread_create(&threads[k], NULL, func, param);
        ASSERT(iret1==0);
    }
    for(int k=0; k<nthreads; k++)
        pthread_join(threads[k], NULL);
#else
    func(param);
#endif
}

}
